//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/ADLCompiler.java,v $
// $Revision: 1.14 $
// $Date: 1998/10/27 13:02:03 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL Compiler Main
// NAME:        ADLCompiler.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 19, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.compiler;

import ORG.opengroup.adl.compiler.j.JCompiler;
import ORG.opengroup.adl.compiler.cpp.CPPCompiler;
import ORG.opengroup.adl.compiler.c.CCompiler;
import ORG.opengroup.adl.nlengine.*;
import ORG.opengroup.adl.tools.FilterProcess;
import java.util.StringTokenizer;
import java.io.*;

/**
 * Abstract factory class which is used to build a language dependent
 * ADL/TDD compiler.
 * 
 * @see ORG.opengroup.adl.compiler.j.JCompiler
 * @see ORG.opengroup.adl.compiler.cpp.CPPCompiler
 * @see ORG.opengroup.adl.compiler.c.CCompiler
 */
public abstract class ADLCompiler implements ADLCompiler.ReturnCodes {

  /** the name of the tool using this compiler */
  protected String toolName;

  /** nlengine, necessary when docRequested */
  protected NLEngine nlengine = null;

  /** debug mode indicator */
  static private final boolean debug = false;

  public interface ReturnCodes {
    /** compilation OK return code */
    short compilationOK = 0;

    /** doc generation failed return code */
    short docFailed = 1;

    /** code generation failed return code */
    short codeFailed = 2;
  }

  /**
   * Creates an ADLCompiler suitable for the requested binding language.
   *
   * <p>Note that for IDL, the Java binding language is assumed.
   *
   * @param toolName the name of the tool using the ADLCompiler
   * @param bindingLanguage the requested binding language
   * @return an instance of a concrete subclass of ADLCompiler
   */
  public static ADLCompiler newCompiler(String toolName,
					String bindingLanguage) {

    ADLCompiler compiler = null;

    if (bindingLanguage.equals("java") || bindingLanguage.equals("idl")) {
      compiler = new JCompiler(toolName);
    }
    else if (bindingLanguage.equals("cpp")) {
      compiler = new CPPCompiler(toolName);
    }
    else if (bindingLanguage.equals("c")) {
      compiler =  new CCompiler(toolName);
    }

    // create appropriate NL engine to deal with documentation processing
    if (Adlt.docRequested()) {
      try {
        compiler.nlengine = new NLEngine();
      }
      catch (Throwable t) {
        System.err.println();
        System.err.println("WARNING: unable to load the NL engine: " + t.getMessage());
        System.err.println("Will not perform any documentation processing.");
        if (debug) t.printStackTrace(System.err);
      }
    }

    return compiler;
  }


  /**
    * Close the NLEngine object (needed for a "clean" termination,
    * as the NLEngine has opened channels with the Prolog engine).
    */
  public void closeNLEngine() throws IOException {

    if (nlengine != null)
      nlengine.close();
  }

  /**
   * Returns an input stream for a given file: apply an extension filter
   * beforehand if one has been specified.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the <code>adl.extension_filter</code> property to check if
   *     an extension filter command has been specified
   * <li>if yes:<ul>
   * <li>create a new FilterProcess for the filter command and file
   * <li>build a <code>DataInputStream</code> out of the filter process
   *     output stream
   * </ul><li>else just build a <code>DataInputStream</code> out of a
   *     <code>FileInputStream</code> for the file
   * </ul>
   *
   * @param fileName the name of the file to filter or open
   * @return the built stream or null if anything failed
   *
   * @exception FileNotFoundException when there is no extension filter
   *            and the file was not found.
   * @exception IOException when a preprocessing error occured.
   */
  static public DataInputStream getStream(String fileName)
  throws FileNotFoundException, IOException {
    DataInputStream fileStream = null;

    // check if there is an extension preprocessor to be applied before
    // parsing
    String extensionFilter = null;
    try {
      extensionFilter = System.getProperty("adl.extension_filter");
    } catch (Exception e) {
      // exception ignored
    }
    if (extensionFilter != null) {
      // there is an extension preprocessor, use output stream of
      // that filter as input stream for parser
      String extensionCommand = extensionFilter + " " + fileName;
      if (Adlt.verboseMode())
	System.out.println("[ preprocessing file with '" + 
			   extensionCommand + "' ]");

      fileStream = new DataInputStream
	(new FilterProcess
	 (extensionCommand, fileName).getInputStream());
    } else {
      // no filter, just open the file stream
      fileStream = new DataInputStream(new FileInputStream(fileName));
    }

    return fileStream;
  }

  /**
   * Abstract method that should be provided in concrete subclasses.
   *
   * @param fileName the name of the file to compile
   * @return compilationOK if the compilation was OK
   *         docFailed if the documentation generation failed
   *         codeFailed if the code generation failed
   * @see ORG.opengroup.adl.compiler.ADLCompiler.ReturnCodes
   * @see ORG.opengroup.adl.compiler.j.JCompiler#compile
   * @see ORG.opengroup.adl.compiler.cpp.CPPCompiler#compile
   */
  public abstract short compile(String fileName);
}
// ***********************************************************************
// History Log:
// $Log: ADLCompiler.java,v $
// Revision 1.14  1998/10/27 13:02:03  mallet
// Added fileName as the last command line argument to the extension filter.
//
// Revision 1.13  1998/10/23 14:59:36  mallet
// Bug fixed: put back extension filter preprocessing by making the
// getStream() method static. It now uses the FilterProcess class and
// throws either a FileNotFoundException or an IOException instead of
// returning null.
//
// Revision 1.12  1998/05/15 14:18:51  mallet
// Added nested interface to define compiler return codes (0 -> OK, 1 ->
// docFailed, 2-> codeFailed).
// Changed signature of compile method to return a short.
//
// Revision 1.11  1998/04/28 13:53:37  cinard
// Updated to take ADL for C compiler into account.
//
// Revision 1.10  1998/04/17 14:10:41  bareau
// Transfer the initialization of the NL engine in ADLCompiler
// (method newCompiler).
// Bug fix in Adlt: newCompiler must be called with property
// adl.language.
//
// Revision 1.9  1998/04/15 13:49:44  bareau
// Added a handle "nlengine" (of class nlengine.NLEngine) associated
// to the compiler, to deal with NL processing.
//
// Revision 1.8  1998/02/27 13:49:25  mallet
// Updated copyright notice.
//
// Revision 1.7  1998/02/25 08:30:17  mallet
// Moved private field toolName in ADLCompiler: now protected.
//
// Revision 1.6  1998/02/12 16:01:49  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.5  1998/01/26 15:35:35  mallet
// Changed reporting of extension filtering to be done only in verbose mode.
//
// Revision 1.4  1998/01/14 11:05:53  wassel
// Added call for CPPCompiler when language option is C++.
//
// Revision 1.3  1998/01/07 17:21:48  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.2  1998/01/05 16:40:18  mallet
// Added a generic getStream method to be able to pre-process input file
// through an extension filter when required without affecting the
// language dependent compilers: the extension filter command is
// specified in the adl.extension_filter property.
//
// Revision 1.1  1997/12/19 14:01:16  mallet
// Language independent ADL/TDD compiler: abstract factory creating a
// concrete compiler according to the selected language binding.
//
// ***********************************************************************
