//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/Adlt.java,v $
// $Revision: 1.23 $
// $Date: 1998/11/25 07:47:10 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADLT Main program
// NAME:        Adlt.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 18, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.compiler;

import java.util.*;
import java.io.*;

/**
 * Main class for the ADL Translation System.
 * Its purpose is to parse command line options and operate
 * language-dependent compilers to generate output.
 *
 * ADLT cannot be operated without either setting the ADL2HOME environment
 * variable or an adl.home java property.
 *
 * The ADLT environment can be modified by either the system wide
 * configuration file <code>$ADL2HOME/Adlt.rc</code>, or by a user defined
 * configuration file through the <code>-config</code> option or in 
 * <code>./adlt.rc</code>, or by environment variable or by command line
 * options, in increasing order of precedence.
 */
public class Adlt {
  /** RCS revision number of the ADL Translation System */
  private static final String id = "$Revision: 1.23 $";
  /** Tool name to be passed to other modules */
  private static final String toolName = "ADLT 2.1";

  /** Local reference to system properties */
  private static Properties properties = null;
 
  /** Verbose mode indicator */
  private static boolean _verboseMode = false;
  /** language option */
  private static Option langOption = new Option("language opt",
						"only one of -java, -idl, -cpp or -c",
						"adl.language");
  /** documentation generation option */
  private static Option docOption = new Option("doc opt",
					       "only one of -doc");
  /** Boolean indicating whether an output option has been set */
  private static boolean _outputOptSet = false;
  /** ACO generation option */
  private static Option acoOption = new Option("output opts",
					       "only one of -aco, -acf or -nocode");
  /** test cases generation option */
  private static Option tcOption = new Option("output opts",
					       "only one of -tc or -nocode");
  /** test data and factory generation option */
  private static Option tdOption = new Option("output opts",
					       "only one of -td or -nocode");
  /** makefile generation option */
  private static Option mkOption = new Option("output opts",
					      "only one of -mk or -nocode");
  /** development makefile generation option */
  private static Option devOption = new Option("[-dev]", "-dev");

  /** TET scenario generation option */
  private static Option scenOption = new Option("output opts",
						"only one of -scen or -nocode");
  /** TET included scenario generation option */
  private static Option noIncludeOption = new Option("[-noinc]",
						   "-noinc");

  /** Config file option */
  private static Option configOption = new Option("env opts",
						  "only one -config",
						  "adl.config");
  /** Locale option */
  private static Option localeOption = new Option("env opts",
						  "only one -locale",
						  "adl.locale");
  /** Inc path option */
  private static Option incpathOption = new Option("env opts",
						   "only one -incpath",
						   "adl.incpath");
  
  /** Debug flag. */
  private final static boolean debug = false;

  /**
   * Indicates whether verbose mode has been requested.
   */
  public static boolean verboseMode() {
    return _verboseMode;
  }

  /**
    * Indicates if the file is a nld file
    * @param fileName the name of a file
    */
  public static boolean isNLDFile(String fileName) {

    return getFileType(fileName) == FileType.nldFile;
  }

  /**
   * Sets a given option to a specified value.
   *
   * When the option has already been set, an error message is displayed,
   * and the program stops.
   */
  private static void setOption(Option opt, boolean val) {
    try {
      opt.set(val);
    } catch (Exception e) {
      System.err.println(e.getMessage());
      usage(null);
    }
  }

  /**
   * Sets a given option to a specified value and parameter.
   *
   * When the option has already been set, an error message is displayed,
   * and the program stops.
   */
  private static void setOption(Option opt, boolean val, String param) {
    try {
      opt.set(val, param);
    } catch (Exception e) {
      System.err.println(e.getMessage());
      usage(null);
    }
  }

  /**
   * Gets the value of a given option.
   */
  private static boolean getOption(Option opt) {
    try {
      return opt.get();
    } catch (Exception e) {
      System.err.println(e.getMessage());
      System.exit(11);
      return false;
    }
  }

  /**
   * Gets the property value of a given option.
   */
  private static String getOptionProperty(Option opt) {
    try {
      return opt.getProperty();
    } catch (Exception e) {
      System.err.println(e.getMessage());
      System.exit(11);
      return null;
    }
  }

  /**
   * Indicates whether ACO generation has been requested.
   */
  public static final boolean acoRequested() {
    return acoOption.isSet() && getOption(acoOption);
  }

  /**
   * Indicates whether test case classes has been requested.
   */
  public static final boolean testCasesRequested() {
    return tcOption.isSet() && getOption(tcOption);
  }

  /**
   * Indicates whether test data and factory generation has been
   * requested.
   */
  public static final boolean testDataRequested() {
    return tdOption.isSet() && getOption(tdOption);
  }

  /**
   * Indicates whether the makefile generation has been requested.
   */
  public static final boolean makefileRequested() {
    return mkOption.isSet() && getOption(mkOption);
  }

  /**
   * Indicates whether the generated makefile should include test suite
   * development targets.
   */
  public static final boolean devMakefileRequested() {
    return devOption.isSet() && getOption(devOption);
  }

  /**
   * Indicates whether the TET scenario generation has been requested.
   */
  public static final boolean scenarioRequested() {
    return scenOption.isSet() && getOption(scenOption);
  }

  /**
   * Indicates whether the generated TET scenario will be included
   * in a test suite level one.
   */
  public static final boolean includedScenarioRequested() {
    return !(noIncludeOption.isSet() && getOption(noIncludeOption));
  }

  /**
   * Indicates whether code generation has been requested.
   */
  public static final boolean codeRequested() {
    return (acoRequested() || testCasesRequested() || testDataRequested()
      || makefileRequested() || scenarioRequested());
  }

  /**
   * Specifies the binding language that has been requested.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if the language option has already been set, then:<ul>
   * <li>if the requested language does not match the previous one in the
   *     system property <code>adl.language</code>, print an error
   *     message and exit with status 11.
   * </ul><li>else:<ul>
   * <li>set <code>langOptSet</code> to true
   * <li>set system property <code>adl.language</code> to the requested
   *     language
   * </ul></ul>
   *
   * @param langOpt the requested binding language
   */
  private static void setLangOpt(String langOpt) {
    if (langOption.isSet()) {
      if (!langOpt.equals(properties.getProperty("adl.language"))) {
	System.err.println("Incompatible languages between input files");
	System.exit(11);
      }
    } else {
      setOption(langOption, true, langOpt);
    }
  }

  /**
   * Indicates whether the documentation generation has been requested.
   */
  public static final boolean docRequested() {
    return docOption.isSet() && getOption(docOption);
  }

  /** File type code interface */
  private interface FileType {
    /** NLD file */
    short nldFile = 0;

    /** ADL file */
    short adlFile = 1;

    /** TDD file */
    short tddFile = 2;

    /** Illegal file type */
    short illegalFile = 100;
  }

  /**
   * Verifies that the suffix of a given input file is valid.
   *
   * <p>When the suffix provides information about the binding language to
   * use, this binding language is requested (call
   * <code>setLangOpt</code>).
   *
   * <p>Valid suffixes are: <code>.adl</code>, <code>.tdd</code>
   * and <code>.nld</code>.
   *
   * @param fileName the name of the file to verify
   * @return the recognized file type as defined in the FileType interface
   *
   * @see ORG.opengroup.adl.compiler.Adlt.FileType
   */
  private static short getFileType(String fileName) {
    String suffix = fileName.substring(fileName.lastIndexOf('.') + 1);
    if (suffix.equals("nld")) {
      return FileType.nldFile;
    } else if (suffix.equals("adl")) {
      return FileType.adlFile;
    } else if (suffix.equals("tdd")) {
      return FileType.tddFile;
    } else
      return FileType.illegalFile;
  }
      
  /**
   * Prints a usage message, possibly preceded by a bad option message
   * if a non-null bad option name is provided.
   *
   * @param badOption the option that was not recognized as valid
   */
  private static void usage(String badOption) {

    if (badOption != null && !badOption.equals(""))
      System.err.println("Incorrect option: " + badOption);

    System.err.println("Usage: adlt [-v] [language opt] [doc opt] " +
		       "[-dev] [-noinc] " +
		       "[output opts] [env opts] <input file>+");
    System.err.println("where:");
    System.err.println("  * [language opt] is one of -java, -cpp, " +
		       "-idl or -c");
    System.err.println("  * [doc opt] is optionally -doc");
    System.err.println("  * [output opts] is either -nocode " +
		       "or any combination of " +
		       "-aco|-acf, -tc, -td,");
    System.err.println("\t-mk and -scen");
    System.err.println("  * [env opts] may include " +
		       "-config <config file>, -incpath <path>");
    System.err.println("\tor -locale <locale>");
    System.err.println("  * <input file> can be any file with the " +
		       "following suffixes: .adl, .tdd or .nld");
    System.exit(11);
  }

  /**
   * Starts the ADL Translation System.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>print tool name and version
   * <li>verify that the property <code>adl.home</code> is defined and
   *     accessible, otherwise exit with an error message
   * <li>load system wide configuration file <code>$ADL2HOME/Adlt.rc</code>
   *     in default properties and replace system properties
   * <li>decode command line options
   * <li>consider each unknown argument not starting with '-' as a file
   *     name and:<ul>
   * <li>verify that its suffix is valid
   * <li>verify that the file exists, is a file and can be read
   * <li>add this file in the list of input files to compile
   * </ul>
   * <li>try to load user defined configuration file (<code>-config</code>
   *     command line option or <code>./adlt.rc</code>) in default
   *     properties
   * <li>if no output option has been set, select default values for these
   * <li>if no language option has been set, print a warning message and
   *     select "java" as the default binding language
   * <li>otherwise print the selected binding language
   * <li>create an <code>ADLCompiler</code> for the selected binding
   *     language (call <code>ADLCompiler.newCompiler</code>)
   * <li>for each input file to compile, call
   *     <code>ADLCompiler.compile</code> on it
   * </ul>
   *
   * @param args the command line options
   * @see ORG.opengroup.adl.compiler.ADLCompiler#newCompiler
   * @see ORG.opengroup.adl.compiler.ADLCompiler#compile
   */
  public static void main(String[] args) {
    System.out.println(toolName + " - " + id);
    System.out.println();

    // be sure we know where ADLT has been installed
    String adlHome = System.getProperty("adl.home");
    if (adlHome == null || adlHome.trim().equals("")) {
      System.err.println("Unable to locate ADL home directory " +
			 "(no ADL2HOME)");
      System.exit(10);
    }

    // load system wide configuration file and replace system properties
    Properties defaultProps = new Properties();
    String systemConfigFile = adlHome + File.separator + "Adlt.rc";
    try {
      defaultProps.load(new FileInputStream(systemConfigFile));
    }
    catch (SecurityException e) {
      System.out.println("Unable to read system configuration file: " +
			 systemConfigFile);
    }
    catch (IOException e) {
      System.out.println("Unable to load system configuration file: " +
			 systemConfigFile);
    }
    properties = replaceSystemProperties(defaultProps);
    
    // path to user defined configuration file (either specified via 
    // -config or guessed in "adlt.rc")
    String configFile = null;

    // vector of input files to compile
    Vector nldFiles = new Vector();
    Vector adlFiles = new Vector();
    Vector tddFiles = new Vector();

    // decode command line options
    for (int ai = 0; ai < args.length; ai++) {
      if (args[ai].equals("-v")) {
	_verboseMode = true;
      } else if (args[ai].equals("-cpp")) {
	setOption(langOption, true, args[ai].substring(1));
      } else if (args[ai].equals("-java")) {
	setOption(langOption, true, args[ai].substring(1));
      } else if (args[ai].equals("-idl")) {
	setOption(langOption, true, args[ai].substring(1));
      } else if (args[ai].equals("-c")) {
	setOption(langOption, true, args[ai].substring(1));
      } else if (args[ai].equals("-nocode")) {
	_outputOptSet = true;
	setOption(acoOption, false);
	setOption(tcOption, false);
	setOption(tdOption, false);
	setOption(mkOption, false);
	setOption(scenOption, false);
      } else if (args[ai].equals("-aco")) {
	_outputOptSet = true;
	setOption(acoOption, true);
      } else if (args[ai].equals("-acf")) {
	_outputOptSet = true;
	setOption(acoOption, true);
      } else if (args[ai].equals("-tc")) {
	_outputOptSet = true;
	setOption(tcOption, true);
      } else if (args[ai].equals("-td")) {
	_outputOptSet = true;
	setOption(tdOption, true);
      } else if (args[ai].equals("-mk")) {
	_outputOptSet = true;
	setOption(mkOption, true);
      } else if (args[ai].equals("-dev")) {
	setOption(devOption, true);
      } else if (args[ai].equals("-scen")) {
	_outputOptSet = true;
	setOption(scenOption, true);
      } else if (args[ai].equals("-noinc")) {
	setOption(noIncludeOption, true);
      } else if (args[ai].equals("-doc")) {
	setOption(docOption, true);
      } else if (args[ai].equals("-config")) {
	setOption(configOption, true, args[++ai]);
      } else if (args[ai].equals("-locale")) {
	setOption(localeOption, true, args[++ai]);
      } else if (args[ai].equals("-incpath")) {
	setOption(incpathOption, true, args[++ai]);
      } else {
	if (args[ai].startsWith("-"))
	  usage(args[ai]);
	else {
	  // probably an input file, verify that the file is
	  // accessible and its suffix is valid
	  short fileType = getFileType(args[ai]);
	  if (fileType != FileType.illegalFile) {
	    File inputFile = new File(args[ai]);
	    
	    if (inputFile.exists() && inputFile.isFile() && 
		inputFile.canRead()) {
	      if (fileType == FileType.nldFile) 
		nldFiles.addElement(args[ai]);
	      else if (fileType == FileType.adlFile)
		adlFiles.addElement(args[ai]);
	      else if (fileType == FileType.tddFile)
		tddFiles.addElement(args[ai]);
	      else {
		System.err.println("Internal compiler error: " +
				   "illegal file type (" +
				   fileType + ")");
		System.exit(200);
	      }
	      
	    } else {
	      System.err.println(args[ai] +
				 ": No such file or file unreadable!");
	      System.exit(12);
	    }
	  } else {
	    System.err.println(args[ai] + 
			       ": Unknown suffix!");
	    usage(null);
	  }
	}
      }
    }

    // load the configuration file in the default properties
    configFile = properties.getProperty("adl.config");
    if (configFile == null) {
      configFile = new String("adlt.rc");
    }

    try {
      defaultProps.load(new FileInputStream(configFile));
      if (_verboseMode) {
	System.out.println("Loaded local configuration file: " +
			   configFile);
      }

    }
    catch (FileNotFoundException e) {
      if (_verboseMode) {
	System.out.println("No local configuration file: " + configFile);
      }
    }
    catch (SecurityException e) {
      if (_verboseMode) {
	System.out.println("Unable to read configuration file: " + 
			   configFile);
      }
    }
    catch (IOException e) {
      if (_verboseMode) {
	System.out.println("Unable to load configuration file: " +
			   configFile);
      }
    }

    // in case there was no output_opts specified, set default
    // values so that everything will be generated
    if (!_outputOptSet) {
      setOption(acoOption, true);
      setOption(tcOption, true);
      setOption(tdOption, true);
      setOption(mkOption, true);
      setOption(scenOption, true);
    }

    // be sure we know what language to use
    String bindingLanguage = properties.getProperty("adl.language");
    if (bindingLanguage == null || bindingLanguage.trim().equals("") ) {
      if (_verboseMode)
	System.out.println("Warning: " + 
			   "No language specification, using java");
      setLangOpt("java");
    } else {
      if (_verboseMode)
	System.out.println("Binding language: " + bindingLanguage);
    }

    // be sure we have a fallback locale
    String userLocale = properties.getProperty("adl.locale");
    if (userLocale == null || userLocale.trim().equals("")) {
      if (_verboseMode)
	System.out.println("Warning: " + 
			   "No user locale specification, using C");
      setOption(localeOption, true, "C");
    } else {
      if (_verboseMode)
	System.out.println("User locale: " + userLocale);
    }

    // create appropriate compiler according to language specification
    ADLCompiler compiler = 
      ADLCompiler.newCompiler(toolName, properties.getProperty("adl.language"));

    // now that everything has been checked and validated, loop on each
    // input file to parse it, and generated corresponding output
    int compilerStatus = 0;
    Vector inputFiles[] = {nldFiles, adlFiles, tddFiles};
    for (short i = 0; i < inputFiles.length; i++) {
      if ((inputFiles[i] == nldFiles) && !nldFiles.isEmpty() &&
	  !docRequested()) {
	if (_verboseMode)
	  System.out.println("No documentation requested: " +
			     "skipping NLD files " + nldFiles);
	continue;
      }

      for (Enumeration enum = inputFiles[i].elements();
	   enum.hasMoreElements();) {
	try {
	  int status = compiler.compile((String)enum.nextElement());
	  if (status > compilerStatus) {
	    compilerStatus = status;
	  }
	}
	catch (Throwable t) {
	  System.err.println();
	  System.err.println("Internal compiler error: " + t.getMessage());
	  System.err.println();
	  System.err.println("Please report to ADL team.");
	  if (debug) t.printStackTrace(System.err);
	  System.exit(100);
	}
      }
    }

    try {
      compiler.closeNLEngine();
    }
    catch(IOException e) {
      System.err.println();
      System.err.println("Internal compiler error: could not terminate NL process: " +
                         e.getMessage());
      System.err.println();
      System.err.println("Please report to ADL team.");
      if (debug) e.printStackTrace(System.err);
      System.exit(100);
    }

    System.exit(compilerStatus);
  }



  /**
   * Replaces the system properties with a new Properties object with the
   * specified default properties.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>create a new Properties object with the specified default
   *     properties
   * <li>copy the system properties in the new object
   * <li>replace the system properties object with the new one
   * </ul>
   *
   * @param defaultProps the default properties for the application
   * @return the new system properties object
   */
  private static  Properties replaceSystemProperties(Properties defaultProps) {
    Properties appProps = new Properties(defaultProps);
    Properties systemProps = System.getProperties();
    String key;

    // copy properties from system to new object
    for (Enumeration e = systemProps.propertyNames(); 
	 e.hasMoreElements(); ) {
      key = (String)e.nextElement();
      appProps.put(key, systemProps.getProperty(key));
    }

    // replace system properties with new object
    System.setProperties(appProps);

    return appProps;
  }

}

// ***********************************************************************
// History Log:
// $Log: Adlt.java,v $
// Revision 1.23  1998/11/25 07:47:10  mallet
// Updated version number for public delivery
//
// Revision 1.22  1998/08/06 13:42:10  mallet
// Incremented version number: now 2.0.3.
//
// Revision 1.21  1998/07/30 07:14:44  mallet
// Added a new method codeRequested() so that *Compiler classes can
// determine when to translate the parsed AST or not.
//
// Revision 1.20.2.1  1998/07/03 07:43:51  mallet
// Incremented version number.
//
// Revision 1.20  1998/06/30 07:25:14  mallet
// Turned debug mode off for delivery.
// Removed useless suffixes support: retain only .adl, .tdd and .nld.
//
// Revision 1.19  1998/06/16 09:09:26  mallet
// Changed role of -dev and -scen_include options: they are now just
// modifiers to the output options.
// Renamed -scen_include as -noinc: default is still to generate an
// includable TET scenario.
//
// Revision 1.18  1998/06/04 16:30:23  bareau
// Added a public static method "isNLDFile(String fileName)" so that
// compilers just parse "pure nld" files and don't try to transform and
// unparse them.
//
// Revision 1.17  1998/05/19 13:22:16  mallet
// Revised exit status management.
// Removed -sgml -ascii and replaced -nodoc with -doc (default inverted).
// Fixed bug in *Requested() methods: when the option was never set, the
// program stopped where it should only report `false'.
// Added private interface to define input file type: renamed
// suffixIsValid as getFileType which returns a FileType.
// Added sorting of input files: NLD first, then ADL, then TDD.
// Added skip of NLD files when no documentation has been requested and
// some NLD files have been passed on the command line.
//
// Revision 1.16  1998/04/28 13:53:37  cinard
// Updated to take ADL for C compiler into account.
//
// Revision 1.15  1998/04/17 14:10:41  bareau
// Transfer the initialization of the NL engine in ADLCompiler
// (method newCompiler).
// Bug fix in Adlt: newCompiler must be called with property
// adl.language.
//
// Revision 1.14  1998/04/17 13:43:21  mallet
// Fixed bug in check of adl.home: trimming could be applied to null.
// Fixed bug in check of binding language: check property rather than
// just option for the case where the default binding language is defined
// in the system/user configuration file (Adlt.rc).
// Added check of user locale: default value set to C.
// Removed call to setNLEngine on the newly created compiler: this should
// rather be done in ADLCompiler.newCompiler().
//
// Revision 1.13  1998/04/15 13:49:44  bareau
// Added a handle "nlengine" (of class nlengine.NLEngine) associated
// to the compiler, to deal with NL processing.
//
// Revision 1.12  1998/04/06 13:40:43  mallet
// Replaced boolean representation of command line options with instances
// of the Option class (easier handling of conflictual options).
// Added trimming of adl.home property.
// Added support for C binding.
//
// Revision 1.11  1998/03/10 17:26:28  mallet
// Merged v1.1 with development trunk
//
// Revision 1.10.2.1  1998/03/09 09:16:24  mallet
// Updated toolName field: now version 2.0.1.
//
// Revision 1.10  1998/02/27 13:49:25  mallet
// Updated copyright notice.
//
// Revision 1.9  1998/02/17 10:57:05  mallet
// Merged v1.0 with development trunk
//
// Revision 1.8.2.1  1998/02/16 08:21:38  mallet
// Fixed small bug in usage message: -mk and -dev are kind of mutually
// exclusive.
//
// Revision 1.8  1998/02/12 16:01:50  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.7  1998/02/02 12:42:06  cinard
// In debug mode stack trace is printed out in case an interal compiler
// error occurs.
//
// Revision 1.6  1998/01/28 13:53:32  mallet
// Changed toolname to include ADL version (currently 2.0). CVS version
// id is now the revision number.
//
// Revision 1.5  1998/01/26 15:56:48  mallet
// Fixed processing of -config command line option: order of precedence
// was not implemented correctly.
// Added a new fallback configuration file in local directory.
// Added a new method replaceSystemProperties to copy the system
// properties into a new object in which we can access the default
// properties object as well (to load configuration files): this is
// required to ensure environment variables take precedence over the
// configuration files (environment variable are passed as system
// properties by the C driver).
// Added more detailed verbose messages.
//
// Revision 1.4  1998/01/23 17:46:10  mallet
// Added -v option for verbose mode: all other modules should call
// verboseMode() to determine whether the compiler is operated in verbose
// mode or not.
// Added processing of -config option (loading configuration file -
// default is $ADL2HOME/Adlt.ad).
// Added verification of property adl.home: exit when it is not set.
// Fixed processing of -ascii and -sgml options.
// Renamed -o option to -doc.
// Added a try/catch block around compiler.compile to catch internal
// compiler errors and report them as such.
//
// Revision 1.3  1998/01/07 17:21:49  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.2  1998/01/05 16:37:27  mallet
// Minor fixes.
//
// Revision 1.1  1997/12/19 14:00:23  mallet
// ADLT main class: parse command line options and drive language
// independent compiler.
//
// ***********************************************************************
