//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/Option.java,v $
// $Revision: 1.1 $
// $Date: 1998/04/06 13:36:38 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADLT Main program
// NAME:        Option.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 18, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.compiler;

/**
 * This class represents a command line option.
 */
public final class Option {
  /** Mask of option state indicator (set/unset) */
  private static int _isSet;
  /** Mask of option settings indicator (on/off) */
  private static int _value;
  /** last assigned option number */
  private static int _lastOptionMask = 1;

  /** option mask */
  private int optionMask;
  /** name of the property associated to the option */
  private String property;
  /** option usage message */
  private String usageMsg;
  /** option name */
  private String name;

  /**
   * Basic constructor: assigns an option number, intializes the option
   * name and usage string.
   *
   * @param optionName the option name
   * @param optionUsage the usage message string
   */
  protected Option(String optionName, String optionUsage) {
    optionMask = _lastOptionMask;
    _lastOptionMask *= 2;
    name = optionName;
    usageMsg = optionUsage;
  }

  /**
   * Constructor for options associated to a property.
   *
   * @param optionName the option name
   * @param optionUsage the usage message string
   * @param optionProperty the name of the associated property
   */
  protected Option(String optionName, String optionUsage,
		 String optionProperty) {
    this(optionName, optionUsage);
    property = optionProperty.trim();
  }
      
  /**
   * Checks if an option is set.
   *
   * @return true if the option is set, false otherwise
   */
  protected boolean isSet() {
    return ((_isSet & optionMask) != 0);
  }

  /**
   * Sets a value to an option.
   *
   * @param val the value of the option
   * @exception Exception the option is already set
   */
  protected void set(boolean val) throws Exception {
    if (isSet()) {
      throw new Exception("Option \"" + name + "\" is already set" +
			  ((usageMsg != null) ? (": " + usageMsg) : ""));
    } else {
      _isSet |= optionMask;
      if (val) {
	_value |= optionMask;
      } else {
	_value &= ~optionMask;
      }
      if (property != null && !property.equals("")) {
	System.getProperties().put(property,
				   String.valueOf(val));
      }
    }
  }
    
  /**
   * Sets a value to an option with a parameter.
   *
   * @param val The value of the option
   * @param propValue the value of the parameter
   * @exception Exception the option is already set or it does not have
   *                      any associated property
   */
  protected void set(boolean val, String propValue) throws Exception {
    set(val);
    if (property != null && !property.equals("")) {
      System.getProperties().put(property, propValue.toString());
    } else {
      throw new Exception("No property for option \"" + name + "\"");
    }
  }

  /**
   * Gets the value of an option.
   *
   * @return the boolean value of the option
   * @exception Exception the option was never set 
   */
  protected boolean get() throws Exception {
    if (!isSet()) {
      throw new Exception("Option \"" + name + "\" was never set");
    } else {
      return ((_value & optionMask) != 0);
    }
  }

  /**
   * Gets the value of the parameter of an option.
   *
   * @return the value of the property associated to the option
   * @exception Exception the option was never set or it does not have
   *                      any associated property
   */
  protected String getProperty() throws Exception {
    if (!isSet()) {
      throw new Exception("Option \"" + name + "\" was never set");
    }
    if (property != null && !property.equals("")) {
      return System.getProperties().getProperty(property);
    } else {
      throw new Exception("No property for option \"" + name + "\"");
    }
  }
}

// ***********************************************************************
// History Log:
// $Log: Option.java,v $
// Revision 1.1  1998/04/06 13:36:38  mallet
// Added new class Option to manage settings of command line options.
//
// ***********************************************************************
