//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/c/CCompiler.java,v $
// $Revision: 1.10 $
// $Date: 1998/07/30 07:43:56 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/C Compiler
// NAME:        CCompiler.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Apr 23, 1998: Created.
//
// ***********************************************************************

package ORG.opengroup.adl.compiler.c;

import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.nlengine.*;
import ORG.opengroup.adl.unparsers.ScenarioGenerator;
import ORG.opengroup.adl.tools.FilterProcess;
import ORG.opengroup.adl.unparsers.c.*;
import ORG.opengroup.adl.translators.c.*;

import java.util.*;
import java.io.*;

/**
 * Concrete implementation of an ADL/TDD compiler for the C
 * language binding
 *
 * @see ORG.opengroup.adl.compiler.ADLCompiler
 * @see ORG.opengroup.adl.parsers.adlc.ADLCParser
 * @see ORG.opengroup.adl.translators.c.ADLCTransformVisitor
 * @see ORG.opengroup.adl.translators.c.TDDCTransformVisitor
 * @see ORG.opengroup.adl.unparsers.c.CUnparserVisitor
 */

public class CCompiler extends ORG.opengroup.adl.compiler.ADLCompiler 
{
  /** the unparser component of this compiler */
  private CUnparserVisitor unparser;

  /** makefile generator */
  private CMakefileGenerator makefileGenerator;

  /** scenario generator */
  private ScenarioGenerator scenarioGenerator;

  /** the parser component of this compiler */
  private ADLCParser parser = null;

  /** the ADL translator component of this compiler */
  private ADLCTransformVisitor adlTranslator = new ADLCTransformVisitor();

  /** the TDD translator component of this compiler */
  private CTDDTransformVisitor tddTranslator = new CTDDTransformVisitor();

  /** the NL unparser component of this compiler */
  private CNLDUnparserVisitor nlunparser;

  /** the directory for generated documentation files */
  private String docOutputDirectory;

  /** debug mode indicator */
  private static final boolean debug = false ;

  /**
   * Basic constructor: initializes the name of the tool using the
   * services of this class and the unparser module
   *
   * @param toolname the name of the tool using the CCompiler
   */

  public CCompiler(String toolname)
  {
     toolName = toolname;
     unparser = new CUnparserVisitor(toolname);
     makefileGenerator = new CMakefileGenerator(toolname);
     scenarioGenerator = new ScenarioGenerator(toolname);    
  }

  /**
   * Compiles a given ADL/TDD file, i.e., generate all possible output
   * from this input file.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>extract leading path from file name and set unparser output
   *     directory
   * <li>get input stream for the given file name
   * <li>if stream is not null:<ul>
   * <li>parse the file
   * <li>proceed with AST translation
   * <li>and finally unparse each resulting AST
   * </ul>
   *
   * <p>When the compilation failed, an error message is printed on the
   *    standard error stream but no exception is raised.
   *
   * @param fileName the name of the file to parse
   * @return compilationOK if the compilation was OK
   *         docFailed if the documentation generation failed
   *         codeFailed if the code generation failed
   * @see ORG.opengroup.adl.compiler.ADLCompiler.ReturnCodes
   * @see ORG.opengroup.adl.compiler.ADLCompiler#getStream
   * @see ORG.opengroup.adl.parsers.adlc.ADLCParser#parse
   * @see #translate
   * @see ORG.opengroup.adl.unparsers.c.CUnparserVisitor#unparse
   * @see #getStream
   */
  
  public short compile(String fileName)
  {
     if (Adlt.verboseMode()) 
     {
        System.out.println("*** compiling " + fileName + ":");
     }

     String parsedfileDirectory = new File(fileName).getParent();
     String outputDirectory = parsedfileDirectory ;
     if (outputDirectory == null) 
     { 
        // no path: local file
        outputDirectory = ".";
     }
     docOutputDirectory = new String(outputDirectory);
     outputDirectory +=  File.separator + "adltest";

     unparser.setDirectory(outputDirectory);
     makefileGenerator.setDirectory(outputDirectory);
     scenarioGenerator.setDirectory(outputDirectory);
     // Set tag used to distinguish sections in the generated scenarii file.
     scenarioGenerator.setTag(fileName);
     
     short status = compilationOK;
     Node topNode = null;

     ParseUnit parseUnit = ParseUnit.getParseUnit(fileName, nlengine);

     // parsing...
     if ( parseUnit.canParse() )
     {
        topNode = parseUnit.parse();

        if ( topNode == null ) return codeFailed;

	if (Adlt.isNLDFile(fileName)) 
	{
	  return compilationOK;
	}
        else 
	{
	  if (Adlt.docRequested() && (nlengine != null)) 
	  {
	    status = generateDoc(topNode);
	  }

	  if (Adlt.codeRequested()) {
	    // translating...
	    Vector asts = translate(topNode);
	    
	    if (asts == null) return codeFailed;

	    // unparsing...
	    if (!asts.isEmpty()) {
	      for (Enumeration enumAsts = asts.elements(); 
		   enumAsts.hasMoreElements();) {
		try {
		  unparser.unparse((ASTTranslationUnit)enumAsts.nextElement());
		}
		catch (Throwable t) {
		  System.out.println("Internal error: ");
		  t.printStackTrace();
		}
	      }
	      
	    } 
	    else 
	      {
		if (Adlt.verboseMode() && 
		    (Adlt.acoRequested() ||
		     Adlt.testDataRequested() ||
		     Adlt.testCasesRequested()))
		  System.out.println("Nothing to generate!");
	      }
	  }
	  return status;
	} 
     }
     else 
     {
        System.err.println("Failed to compile " + fileName + "!");
	return codeFailed;
     }

  }
  
  /**
   * Applies the appropriate translator to an AST given its top node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>starting from last child, jump back over all NLD annotation to
   *     locate the class declaration node
   * <li>if ACO generation has been requested and the class declaration
   *     node is an <code>ADL_AnnotatedDeclaration</code>:<ul>
   * <li>visit the top node with the ADL translator
   * <li>add resulting AST to the vector of transformed ASTs
   * </ul>
   * <li>else if test cases and/or test data generation have been
   *     requested:<ul>
   * <li>visit the top node with the TDD translator
   * <li>add resulting AST to the vector of transformed ASTs
   * <li>if the constants interface AST has been generated, add it to
   *     the vector of transformed ASTs
   * <li>if test cases generation has been requested and some have been
   *     generated, add them to the vector of transformed ASTs
   * <li>if test data generation has been requested:<ul>
   * <li>if some datasets have been generated, add them to the vector
   *     of transformed ASTs
   * <li>if some factories have been generated, add them to the vector
   *     of transformed ASTs
   * </ul>
   * <li>reset TDD translator fields
   * </ul>
   * </ul>
   *
   * @param topNode the top node of the AST to translate
   * @return a vector of generated ASTs
   */
  private Vector translate(Node topNode) 
  {
    Vector returnedNodes = new Vector();
    int classIndex = topNode.jjtGetNumChildren() - 1;
    Node transformedAST = null;

    // if there is no ADL or TDD class child, don't go further
    if (classIndex < 0)
      return null;

    // check type of last child of TranslationUnit
    if (topNode.jjtGetChild(classIndex) instanceof ASTADL_AnnotatedDeclaration)
    {
       // ASTADL_AnnotatedDeclaration
      if (Adlt.acoRequested()) {
        try {
          transformedAST = adlTranslator.translate(topNode);
          returnedNodes.addElement(transformedAST);
        } 
        catch (Throwable e) {
          if (debug)
            e.printStackTrace(System.err);
          return null;
        }
      } 
    }
    else if (topNode.jjtGetChild(classIndex) 
             instanceof ASTTDD_AnnotatedDeclaration) 
    {
       // ASTTDD_AnnotatedDeclaration
       if (debug)
          System.out.println("tdd class: run the tdd translator");
     
        try {
	  transformedAST = tddTranslator.translate(topNode);
        } 
        catch (Throwable e) {
          if (debug)
            e.printStackTrace(System.err);
          return null;
        }

       if (debug)
          System.out.println("After tdd translation : " + transformedAST);

       if (Adlt.testCasesRequested() || Adlt.testDataRequested()) 
       {
	 // in any case, we generate both the main tdd class and the 
	 // constants interface
	 if (transformedAST != null) {
	   returnedNodes.addElement(transformedAST);
	 }

          if (tddTranslator.constantsClass != null)
          {
             returnedNodes.addElement(tddTranslator.constantsClass);
          }
          else if (debug)
          {
             System.out.println("constantsClass null");
          }
       }

       if (Adlt.testCasesRequested() &&
           tddTranslator.directives != null ) 
       {
          // append generated test directive ASTs
          Enumeration DIR = tddTranslator.directives.elements();
          while (DIR.hasMoreElements()) {
             if (debug)
                System.out.println("directive");
             returnedNodes.addElement(DIR.nextElement());
          }
       }
      
    if (Adlt.testDataRequested()) {
       if (tddTranslator.datasets != null) {
          // append generated dataset ASTs
          Enumeration DS = tddTranslator.datasets.elements();
          while (DS.hasMoreElements()) {
             if (debug)
                System.out.println("dataset");
             returnedNodes.addElement(DS.nextElement());
          }
       }
	
       if (tddTranslator.factories != null) {
          // append generated factory ASTs
          Enumeration F = tddTranslator.factories.elements();
          while (F.hasMoreElements()) {
             if (debug)
                System.out.println("factory");
             returnedNodes.addElement(F.nextElement());
          }
       }
    }

       // Generate makefile if requested.
       if (    Adlt.makefileRequested() 
            && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
	 makefileGenerator.generateMakefile
	   ((String)(tddTranslator.getTDDClassNames().elementAt(0)),
	    tddTranslator.getACONames(),
	    tddTranslator.getTDDConstantNames(),
	    tddTranslator.getTDDDatasetNames(),
	    tddTranslator.getTDDFactoryNames(),
	    tddTranslator.getTDDTestDirectiveNames(),
 	    Adlt.devMakefileRequested());
       }

       // generate scenario if requested
       if ( Adlt.scenarioRequested()
            && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
          scenarioGenerator.generateScenario
          (tddTranslator.getTDDTestDirectiveNames(),
           Adlt.includedScenarioRequested());
       }
     }

    return returnedNodes;
  }

  /** Returns the preprocessing command.
    * <p> Reads the system property "adl.filter", adds to it the incpath,
    * and returns its value.
    */
  private String getPreProcessorCommand() 
  {
 
    String preProcessor = null ;
    StringTokenizer incpath = null;

    try 
    {
       preProcessor = System.getProperty("adl.filter");
       incpath = 
       new StringTokenizer(System.getProperty("adl.incpath", "."),
                           File.pathSeparator);
    } 
    catch (Exception e) 
    {
      // exception ignored
    }

    if (preProcessor != null) 
    {
       // append incpath directories to preprocessing command
       while (incpath != null && incpath.hasMoreTokens()) 
       {
          preProcessor += " -I" + incpath.nextToken();
       }

       if (debug) 
       {
          System.out.println("Preprocessing command: "+ preProcessor );
       }
    }

    return preProcessor ;
  }

  /**
   * Run the NLD unparser on the ADL/TDD AST
   */
  private short generateDoc(Node topNode) 
  {
    try 
    {
      nlunparser = new CNLDUnparserVisitor(toolName, nlengine);
      nlunparser.setDirectory(docOutputDirectory);

      nlunparser.unparse((ASTADL_TranslationUnit)topNode);
    }
    catch(Throwable e) 
    {
      System.err.println("Error: " +
			 "no documentation generated for this file, " +
			 "proceeding anyway");
      if (debug)
        e.printStackTrace(System.err);

      return docFailed;
    }
    
    return compilationOK;
  }
  
}


// ***********************************************************************
// History Log:
// $Log: CCompiler.java,v $
// Revision 1.10  1998/07/30 07:43:56  mallet
// Turned on TDD translation in method translate() and initialization of
// related fields.
// Encapsulated the second part of the compilation within a test that
// verifies that the code generation has been requested: fixes a bug
// where the translator was started even when only the documentation has
// been requested.
// Fixed a bug in method translate() that was returning null instead of
// an empty vector when there is no ADL class or TDD class declaration in
// the compiled file.
// Fixed a bug in method translate() where an exception raised in the TDD
// translator was simply ignored: now return null to indicate a failure.
// Added a test that the vector returned by the method translate() is not
// null, otherwise return codeFailed.
//
// Revision 1.9  1998/06/30 07:26:25  mallet
// Turned debug mode off for delivery.
// Fixed cast of AST for unparser.unparse(): ASTTranslationUnit expected.
//
// Revision 1.8  1998/06/29 11:47:12  bareau
// Integrated the ADL/C translator.
//
// Revision 1.7  1998/06/23 07:16:48  cinard
// Fixed call to the nld unparser.
//
// Revision 1.6  1998/06/17 16:15:02  cinard
// Added support for nld unparsing.
//
// Revision 1.5  1998/06/05 15:55:35  mallet
// Added test of number of test directives to determine whether the
// makefiles and scenario have to be generated.
//
// Revision 1.4  1998/05/27 13:30:55  cinard
// Temporary fix for a problem due to missing ADL/C translator.
//
// Revision 1.3  1998/05/15 14:22:56  mallet
// Changed signature of method compile to return a short: updated the
// code to return as soon as an error is detected (except for docFailed).
// Fixed small bug with reporting "Nothing to generate" when no code
// generation was requested.
//
// Revision 1.2  1998/05/11 12:13:27  cinard
// Updated call to the parser : class ParseUnit is used now.
//
// Revision 1.1  1998/04/28 13:52:22  cinard
// First version.
//
// ***********************************************************************
