//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/cpp/CPPCompiler.java,v $
// $Revision: 1.22 $
// $Date: 1998/07/30 07:27:50 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/C++ Compiler
// NAME:        CPPCompiler.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 18, 1997: Created for Java
//      Wassel Berrayana, The Open Group RI, Dec 29 : Updated for ADL/ C++
//
// ***********************************************************************

package ORG.opengroup.adl.compiler.cpp;

import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.parsers.ParseException;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.unparsers.ScenarioGenerator;
import ORG.opengroup.adl.unparsers.cpp.*;
import ORG.opengroup.adl.translators.cpp.*;
import ORG.opengroup.adl.unparsers.cpp.*;

import java.util.*;
import java.io.*;

/**
 * Concrete implementation of an ADL/TDD compiler for the C++
 * language binding
 *
 * @see ORG.opengroup.adl.compiler.ADLCompiler
 * @see ORG.opengroup.adl.parsers.adlcpp.ADLCPPParser
 * @see ORG.opengroup.adl.translators.cpp.ADLCPPTransformVisitor
 * @see ORG.opengroup.adl.translators.cpp.TDDCPPTransformVisitor
 * @see ORG.opengroup.adl.unparsers.cpp.CPPUnparserVisitor
 */
public class CPPCompiler extends ORG.opengroup.adl.compiler.ADLCompiler {

  /** the unparser component of this compiler */
  private CPPUnparserVisitor unparser;

  /** makefile generator */
  private CPPMakefileGenerator makefileGenerator;

  /** scenario generator */
  private ScenarioGenerator scenarioGenerator;

  /** the parser component of this compiler */
  private ADLCPPParser parser = null;

  /** the ADL translator component of this compiler */
  private ADLCPPTransformVisitor adlTranslator = new ADLCPPTransformVisitor();

  /** the TDD translator component of this compiler */
  private CPPTDDTransformVisitor tddTranslator = new CPPTDDTransformVisitor();

  /** the NL unparser component of this compiler */
  private CPPNLDUnparserVisitor nlunparser;

  /** the directory for generated documentation files */
  private String docOutputDirectory;

  /** debug mode indicator */
  private static final boolean debug = false ;

  /** NLD to generate. */
  private boolean nld_required = true;

  /**
   * Basic constructor: initializes the name of the tool using the
   * services of this class and the unparser module
   *
   * @param toolname the name of the tool using the CPPCompiler
   */
  public CPPCompiler(String toolname) {
    toolName = toolname;
    unparser = new CPPUnparserVisitor(toolname);
    makefileGenerator = new CPPMakefileGenerator(toolname);
    scenarioGenerator = new ScenarioGenerator(toolname);

    nld_required = nlengine != null;
  }

  /**
   * Compiles a given ADL/TDD file, i.e., generate all possible output
   * from this input file.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>extract leading path from file name and set unparser output
   *     directory
   * <li>get input stream for the given file name
   * <li>if stream is not null:<ul>
   * <li>parse the file
   * <li>proceed with AST translation
   * <li>and finally unparse each resulting AST
   * </ul>
   * </ul>
   *
   * <p>When the compilation failed, an error message is printed on the
   *    standard error stream but no exception is raised.
   *
   * @param fileName the name of the file to parse
   * @return compilationOK if the compilation was OK
   *         docFailed if the documentation generation failed
   *         codeFailed if the code generation failed
   * @see ORG.opengroup.adl.compiler.ADLCompiler.ReturnCodes
   * @see ORG.opengroup.adl.compiler.ADLCompiler#getStream
   * @see ORG.opengroup.adl.parsers.adlcpp.ADLCPPParser#parse
   * @see #translate
   * @see ORG.opengroup.adl.unparsers.cpp.CPPUnparserVisitor#unparse
   * @see #getStream
   */
  public short compile(String fileName) {
    if (Adlt.verboseMode()) {
      System.out.println("*** compiling " + fileName + ":");
    }

    String parsedfileDirectory = new File(fileName).getParent();
    String outputDirectory = parsedfileDirectory ;
    if (outputDirectory == null) { // no path: local file
      outputDirectory = ".";
    }
    docOutputDirectory = new String(outputDirectory);
    outputDirectory +=  File.separator + "adltest";

    unparser.setDirectory(outputDirectory);
    makefileGenerator.setDirectory(outputDirectory);
    scenarioGenerator.setDirectory(outputDirectory);
    // Set tag used to distinguish sections in the generated scenarii file.
    scenarioGenerator.setTag(fileName);

    short status = compilationOK;
    Node topNode = null;
    
    // parsing...
    ParseUnit parseUnit = ParseUnit.getParseUnit(fileName, nlengine);

    // parsing...
    if ( parseUnit.canParse() )
    {
      topNode = parseUnit.parse();

      if (topNode == null) return codeFailed;

      if (debug)
        System.out.println("Parsing done");

      if (Adlt.isNLDFile(fileName)) {
        return compilationOK;
      }
      else {
        if (Adlt.docRequested() && (nlengine != null)) {
          status = generateDoc(topNode);
        }

	if (Adlt.codeRequested()) {
	  // translating...
	  Vector asts = translate(topNode);
	  
	  if (asts == null) return codeFailed;
	  
	  // unparsing...
	  if (!asts.isEmpty()) {
	    for ( Enumeration enumAsts = asts.elements(); 
		  enumAsts.hasMoreElements();) {
	      try {
		unparser.unparse((SimpleNode)enumAsts.nextElement());
	      } catch ( ParseException e) {
		// do nothing since exception message already displayed
		return codeFailed;
	      }
	    }
	  } else {
	    if (Adlt.verboseMode() && 
		(Adlt.acoRequested() ||
		 Adlt.testDataRequested() ||
		 Adlt.testCasesRequested()))
	      System.out.println("Nothing to generate!");
	  }
	}
        return status;
      }
    } else {
      System.err.println("Failed to compile " + fileName + "!");
      return codeFailed;
    }
  }

  /**
   * Applies the appropriate translator to an AST given its top node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>starting from last child, jump back over all NLD annotation to
   *     locate the class declaration node
   * <li>if ACO generation has been requested and the class declaration
   *     node is an <code>ADLClass</code>:<ul>
   * <li>visit the top node with the ADL translator
   * <li>add resulting AST to the vector of transformed ASTs
   * </ul>
   * <li>else if test cases and/or test data generation have been
   *     requested:<ul>
   * <li>visit the top node with the TDD translator
   * <li>add resulting AST to the vector of transformed ASTs
   * <li>if the constants interface AST has been generated, add it to
   *     the vector of transformed ASTs
   * <li>if test cases generation has been requested and some have been
   *     generated, add them to the vector of transformed ASTs
   * <li>if test data generation has been requested:<ul>
   * <li>if some datasets have been generated, add them to the vector
   *     of transformed ASTs
   * <li>if some factories have been generated, add them to the vector
   *     of transformed ASTs
   * </ul>
   * <li>reset TDD translator fields
   * </ul>
   * </ul>
   *
   * @param topNode the top node of the AST to translate
   * @return a vector of generated ASTs
   */
  private Vector translate(Node topNode) {
    Vector returnedNodes = new Vector();
    int classIndex = topNode.jjtGetNumChildren() - 1;
    Node transformedAST;

    // if there is no ADL or TDD class child, don't go further
    if (classIndex < 0)
      return returnedNodes;

    // check type of last child of TranslationUnit
    if (topNode.jjtGetChild(classIndex) instanceof ASTADLClass ) {
      // ASTADLClass
      if (Adlt.acoRequested()) {
	try {
	  transformedAST = adlTranslator.translate(topNode);
	  returnedNodes.addElement(transformedAST);
	} catch (Throwable e) {
          if (debug)
            e.printStackTrace(System.err);
          return null;
	}
      }
    } else if (topNode.jjtGetChild(classIndex) instanceof ASTTDDClassDecl) {
      // ASTTDDClassDecl
      if (debug)
        System.out.println("tdd class: run the tdd translator");

      try {
        transformedAST = tddTranslator.translate(topNode);
      } catch (ParseException e) {
        if (debug) {
          System.err.println("ParseException caught: "  + e.getMessage());
	  e.printStackTrace(System.err);
	} 

        return null;
      }

      if (debug)
        System.out.println("After tdd translation : " + transformedAST);

      if (Adlt.testCasesRequested() || Adlt.testDataRequested()) {
      // in any case, we generate both the main tdd class and the 
      // constants interface
        returnedNodes.addElement(transformedAST);

        if (tddTranslator.constantsClass != null)
          returnedNodes.addElement(tddTranslator.constantsClass);
        else if (debug)
          System.out.println("constantsClass null");
      }

      if (Adlt.testCasesRequested() &&
	  tddTranslator.directives != null) {
	// append generated test directive ASTs
	Enumeration DIR = tddTranslator.directives.elements();
	while (DIR.hasMoreElements()) {
          if (debug)
            System.out.println("directive");
	  returnedNodes.addElement(DIR.nextElement());
	}
      }
      
      if (Adlt.testDataRequested()) {
	if (tddTranslator.datasets != null) {
	  // append generated dataset ASTs
	  Enumeration DS = tddTranslator.datasets.elements();
	  while (DS.hasMoreElements()) {
          if (debug)
            System.out.println("dataset");
	    returnedNodes.addElement(DS.nextElement());
	  }
	}
	
	if (tddTranslator.factories != null) {
	  // append generated factory ASTs
	  Enumeration F = tddTranslator.factories.elements();
	  while (F.hasMoreElements()) {
          if (debug)
            System.out.println("factory");
	    returnedNodes.addElement(F.nextElement());
	  }
	}
      }

       // Generate makefile if requested.
       if (    Adlt.makefileRequested() 
            && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
	 makefileGenerator.generateMakefile
	   ((String)(tddTranslator.getTDDClassNames().elementAt(0)),
	    tddTranslator.getACONames(),
	    tddTranslator.getTDDConstantNames(),
	    tddTranslator.getTDDDatasetNames(),
	    tddTranslator.getTDDFactoryNames(),
	    tddTranslator.getTDDTestDirectiveNames(),
 	    Adlt.devMakefileRequested());
       }

       // generate scenario if requested
       if ( Adlt.scenarioRequested()
            && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
          scenarioGenerator.generateScenario
          (tddTranslator.getTDDTestDirectiveNames(),
           Adlt.includedScenarioRequested());
       }
    }

    return returnedNodes;
  }


  /**
    * Run the NLD unparser on the ADL/TDD AST
    */
  private short generateDoc(Node topNode) {

    try {
      nlunparser = new CPPNLDUnparserVisitor(toolName, nlengine);
      nlunparser.setDirectory(docOutputDirectory);

      nlunparser.unparse(topNode);
    }
    catch(Throwable e) {
      System.err.println("Error: " +
			 "no documentation generated for this file, " +
			 "proceeding anyway");
      if (debug)
        e.printStackTrace();

      return docFailed;
    }

    return compilationOK;
  }

}


// ***********************************************************************
// History Log:
// $Log: CPPCompiler.java,v $
// Revision 1.22  1998/07/30 07:27:50  mallet
// Encapsulated the second part of the compilation within a test that
// verifies that the code generation has been requested: fixes a bug
// where the translator was started even when only the documentation has
// been requested.
// Fixed a bug in method translate() that was returning null instead of
// an empty vector when there is no ADL class or TDD class declaration in
// the compiled file.
// Fixed a bug in method translate() where an exception raised in the TDD
// translator was simply ignored: now return null to indicate a failure.
// Added a test that the vector returned by the method translate() is not
// null, otherwise return codeFailed.
//
// Revision 1.21  1998/06/30 07:26:57  mallet
// Turned debug mode off for delivery.
// Added import of ORG.opengroup.adl.parsers.ParseException.
//
// Revision 1.20  1998/06/05 15:55:36  mallet
// Added test of number of test directives to determine whether the
// makefiles and scenario have to be generated.
//
// Revision 1.19  1998/06/04 16:30:24  bareau
// Added a public static method "isNLDFile(String fileName)" so that
// compilers just parse "pure nld" files and don't try to transform and
// unparse them.
//
// Revision 1.17  1998/05/27 12:40:33  cinard
// Compiler now uses ParseUnit class to parse instead of directly using
// the ADLCPPParser.
//
// Revision 1.16  1998/05/15 14:22:55  mallet
// Changed signature of method compile to return a short: updated the
// code to return as soon as an error is detected (except for docFailed).
// Fixed small bug with reporting "Nothing to generate" when no code
// generation was requested.
//
// Revision 1.15  1998/04/15 16:26:49  bareau
// Send the nlengine to the ADL/TDD parser.
//
// Revision 1.14  1998/04/15 13:49:45  bareau
// Added a handle "nlengine" (of class nlengine.NLEngine) associated
// to the compiler, to deal with NL processing.
//
// Revision 1.13  1998/03/31 14:17:22  wassel
// Merged back portable-code into trunk
//
// Revision 1.12.2.2  1998/03/25 08:50:15  wassel
// Added call, in method compile, to the parser method setParsedFileDirectory.
//
// Revision 1.12.2.1  1998/03/20 10:25:21  wassel
// Removed method getStream : Preprocessing mechanism moved into the parser.
//
// Revision 1.12  1998/03/10 17:26:28  mallet
// Merged v1.1 with development trunk
//
// Revision 1.11.2.3  1998/03/09 19:51:51  mallet
// Bug fixed: replaced call to jjtAccept by call to translate for ADL
// translation to reinitialize translator state information.
//
// Revision 1.11.2.2  1998/03/04 18:48:31  cinard
// Now a scenario can contains sections for development targets coming from
// several tdd files.
//
// Revision 1.11.2.1  1998/03/04 10:25:04  bareau
// The tdd translation is now performed via the new entry point method
// "translate" of the tdd visitor. No need to perform a "reset" anymore.
//
// Revision 1.11  1998/02/27 13:49:27  mallet
// Updated copyright notice.
//
// Revision 1.10  1998/02/25 08:29:40  mallet
// Added '/adltest' to outputDirectory in method compile to generate in
// the adltest subdirectory.
// Added TDD translation in method translate.
// Removed private field toolName: now in ADLCompiler.
// Changed splitting of adl.incpath to use system-dependent
// File.pathSeparator instead of hard-coded ':'.
//
// Revision 1.9  1998/02/17 10:57:06  mallet
// Merged v1.0 with development trunk
//
// Revision 1.8.2.2  1998/02/16 08:23:27  mallet
// Fixed small bug in translate method: now an explicit test of TDD class is
// made to avoid going into this branch for a file that does not contain
// a TDD class (e.g., with only an import declaration in Java).
//
// Revision 1.8.2.1  1998/02/13 16:06:20  mallet
// Fixed bug in translate: was trying to generate TDD releated files out
// of an ADL input when ACO was not requested.
//
// Revision 1.8  1998/02/12 16:01:51  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.7  1998/02/12 10:25:03  wassel
// Moved back cast of unparse with SimpleNode.
//
// Revision 1.6  1998/02/11 16:11:09  mallet
// Added temporary hack for NULL: append "-DNULL=NULL" to pre-processing
// command.
//
// Revision 1.5  1998/02/11 15:39:54  rami
// Fixed cast of unparse method arguments.
//
// Revision 1.4  1998/02/10 18:15:32  mallet
// Fixed some @see tags in class comments.
// Added ADL/CPP translator.
// Changed initialization of parser: now initialized only once in parse,
// when we have a real stream to pass in to constructor.
// Moved preprocessor handling into a separate method getStream
// overriding ADLCompiler.getStream.
// Updated translate method to skip NLD annotation nodes to find out the
// type of AST we have to translate.
//
// Revision 1.3  1998/02/10 15:30:00  cinard
// Now use generic scenario generator.
//
// Revision 1.2  1998/02/10 15:14:12  cinard
// Added hooks for Makefile generation. Actual generation not yet implemented.
//
// Revision 1.1  1998/01/14 11:12:16  wassel
// C++ compiler still lacking translator and javadocs.
//
// Revision 1.2  1997/12/24 11:26:54  mallet
// Added translation phase.
//
// Revision 1.1  1997/12/19 14:06:00  mallet
// Concrete implementation of the ADLCompiler for the Java language binding.
//
// ***********************************************************************
