//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/compiler/j/JCompiler.java,v $
// $Revision: 1.33 $
// $Date: 1998/07/30 07:19:48 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/Java Compiler
// NAME:        JCompiler.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 18, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.compiler.j;

import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.nlengine.*;
import ORG.opengroup.adl.unparsers.ScenarioGenerator;
import ORG.opengroup.adl.unparsers.j.*;
import ORG.opengroup.adl.translators.j.*;

import java.util.*;
import java.io.*;

import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.HashMap;

/**
 * Concrete implementation of an ADL/TDD compiler for the Java
 * binding language.
 *
 * <p>The compiler is activated by calling the
 *    <code><a href="#compile">compile</a></code> method.
 *
 * @see ORG.opengroup.adl.compiler.ADLCompiler
 * @see ORG.opengroup.adl.parsers.adljava.ADLJavaParser
 * @see ORG.opengroup.adl.translators.j.JADLTransformVisitor
 * @see ORG.opengroup.adl.translators.j.JTDDTransformVisitor
 * @see ORG.opengroup.adl.unparsers.j.JUnparserVisitor
 */
public class JCompiler extends ORG.opengroup.adl.compiler.ADLCompiler {

  /** the unparser component of this compiler */
  private JUnparserVisitor unparser;

  /** makefile generator */
  private JMakefileGenerator makefileGenerator;

  /** scenario generator */
  private ScenarioGenerator scenarioGenerator;

  /** the ADL translator component of this compiler */
  private JADLTransformVisitor adlTranslator = new JADLTransformVisitor();

  /** the TDD translator component of this compiler */
  private JTDDTransformVisitor tddTranslator = new JTDDTransformVisitor();

  /** the NL unparser component of this compiler */
  private JNLDUnparserVisitor nlunparser;

  /** the directory for generated documentation files */
  private String docOutputDirectory;

  /** debug mode indicator */
  private static final boolean debug = false;

  /**
   * Basic constructor: initializes the name of the tool using the
   * services of this compiler and the unparser module.
   *
   * @param toolname the name of the tool using the JCompiler
   */
  public JCompiler(String toolname) {
    toolName = toolname;
    unparser = new JUnparserVisitor(toolname);
    makefileGenerator = new JMakefileGenerator(toolname);
    scenarioGenerator = new ScenarioGenerator(toolname);    
  }

  /**
   * Compiles a given ADL/TDD file, i.e., generate all possible outputs
   * from this input file.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>extract leading path from file name and set unparser output
   *     directory
   * <li>get input stream for the given file name
   * <li>if stream is not null:<ul>
   * <li>parse the file
   * <li>proceed with AST translation
   * <li>and finally unparse each resulting AST
   * </ul>
   * </ul>
   *
   * <p>When the compilation failed, an error message is printed on the
   *    standard error stream but no exception is raised.
   *
   * @param fileName the name of the file to parse
   * @return compilationOK if the compilation was OK
   *         docFailed if the documentation generation failed
   *         codeFailed if the code generation failed
   * @see ORG.opengroup.adl.compiler.ADLCompiler.ReturnCodes
   * @see ORG.opengroup.adl.compiler.ADLCompiler#getStream
   * @see ORG.opengroup.adl.parsers.adljava.ADLJavaParser#parse
   * @see #translate
   * @see ORG.opengroup.adl.unparsers.j.JUnparserVisitor#unparse
   */
  public short compile(String fileName) {
    if (Adlt.verboseMode()) {
      System.out.println("*** compiling " + fileName);
    }

    String outputDirectory = new File(fileName).getParent();
    if (outputDirectory == null) { // no path: local file
      outputDirectory = ".";
    }
    docOutputDirectory = new String(outputDirectory);
    outputDirectory += File.separator + "adltest";

    unparser.setDirectory(outputDirectory);
    makefileGenerator.setDirectory(outputDirectory);
    scenarioGenerator.setDirectory(outputDirectory);
    // Set tag used to distinguish sections in the generated scenarii file.
    scenarioGenerator.setTag(fileName);

    short status = compilationOK;
    Node topNode = null;

    ParseUnit parseUnit = ParseUnit.getParseUnit(fileName, nlengine);

    // parsing...
    if ( parseUnit.canParse() )
    {
       topNode = parseUnit.parse();

       if ( topNode == null ) return codeFailed;

       if (Adlt.isNLDFile(fileName)) {
         return compilationOK;
       }
       else {
         if (Adlt.docRequested() && (nlengine != null)) {
           status = generateDoc(topNode);
         }

	 if (Adlt.codeRequested()) {
	   // translating...
	   Vector asts = translate(topNode);

	   // unparsing...
	   if (!asts.isEmpty()) {
	     for (Enumeration enumAsts = asts.elements(); 
		  enumAsts.hasMoreElements();) {
	       unparser.unparse((SimpleNode)enumAsts.nextElement());
	     }
	     
	   } else {
	     if (Adlt.verboseMode() && 
		 (Adlt.acoRequested() ||
		  Adlt.testDataRequested() ||
		  Adlt.testCasesRequested()))
	       System.out.println("Nothing to generate!");
	   }
	 }
         return status;
       }
    } else {
      System.err.println("Failed to compile " + fileName + "!");
      return codeFailed;
    }
  }

  /**
   * Applies the appropriate translator to an AST given its top node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>starting from last child, jump back over all NLD annotation to
   *     locate the class declaration node
   * <li>if ACO generation has been requested and the class declaration
   *     node is an <code>ADL_ClassDeclaration</code>:<ul>
   * <li>visit the top node with the ADL translator
   * <li>add resulting AST to the vector of transformed ASTs
   * </ul>
   * <li>else if test cases and/or test data generation have been
   *     requested:<ul>
   * <li>visit the top node with the TDD translator
   * <li>add resulting AST to the vector of transformed ASTs
   * <li>if the constants interface AST has been generated, add it to
   *     the vector of transformed ASTs
   * <li>if test cases generation has been requested and some have been
   *     generated, add them to the vector of transformed ASTs
   * <li>if test data generation has been requested:<ul>
   * <li>if some datasets have been generated, add them to the vector
   *     of transformed ASTs
   * <li>if some factories have been generated, add them to the vector
   *     of transformed ASTs
   * </ul>
   * <li>reset TDD translator fields
   * </ul>
   * </ul>
   *
   * @param topNode the top node of the AST to translate
   * @return a vector of generated ASTs
   */
  private Vector translate(Node topNode) {
    Vector returnedNodes = new Vector();
    int classIndex = topNode.jjtGetNumChildren() - 1;
    Node transformedAST;

    // if there is no ADL or TDD class child, don't go further
    if (classIndex < 0)
      return returnedNodes;

    // check type of last child of ADL_CompilationUnit
    if (topNode.jjtGetChild(classIndex) instanceof ASTADL_ClassDeclaration) {
      // ASTADL_ClassDeclaration
      if (Adlt.acoRequested()) {
	transformedAST = adlTranslator.translate(topNode);
	if (transformedAST != null) {
	  returnedNodes.addElement(transformedAST);
	}
      }
    }
    else if (topNode.jjtGetChild(classIndex) instanceof ASTTDD_ClassDeclaration) {
       // ASTTDD_ClassDeclaration
      transformedAST = tddTranslator.translate(topNode);

       if (Adlt.testCasesRequested() || Adlt.testDataRequested()) {
	 // in any case, we generate both the main tdd class and the 
	 // constants interface
	 if (transformedAST != null) {
	   returnedNodes.addElement(transformedAST);
	 }

	 if (tddTranslator.constantsInterface != null)
	   returnedNodes.addElement(tddTranslator.constantsInterface);
       }

       if (Adlt.testCasesRequested() &&
           tddTranslator.directives != null) {
          // append generated test directive ASTs
          Enumeration DIR = tddTranslator.directives.elements();
          while (DIR.hasMoreElements()) {
             returnedNodes.addElement(DIR.nextElement());
          }
       }
      
       if (Adlt.testDataRequested()) {
          if (tddTranslator.datasets != null) {
             // append generated dataset ASTs
             Enumeration DS = tddTranslator.datasets.elements();
             while (DS.hasMoreElements()) {
                returnedNodes.addElement(DS.nextElement());
             }
          }
	
          if (tddTranslator.factories != null) {
             // append generated factory ASTs
             Enumeration F = tddTranslator.factories.elements();
             while (F.hasMoreElements()) {
                returnedNodes.addElement(F.nextElement());
             }
          }
       }

       // Generate makefile if requested.
       if (    Adlt.makefileRequested() 
	    && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
	 makefileGenerator.generateMakefile
	   ((String)(tddTranslator.getTDDClassNames().elementAt(0)),
	    tddTranslator.getACONames(),
	    tddTranslator.getTDDConstantNames(),
	    tddTranslator.getTDDDatasetNames(),
	    tddTranslator.getTDDFactoryNames(),
	    tddTranslator.getTDDTestDirectiveNames(),
	    Adlt.devMakefileRequested());
       }

       // generate scenario if requested
       if ( Adlt.scenarioRequested() 
	    && tddTranslator.getTDDTestDirectiveNames().size() > 0 )
       {
          scenarioGenerator.generateScenario
          (tddTranslator.getTDDTestDirectiveNames(),
           Adlt.includedScenarioRequested());
       }
    }

    return returnedNodes;
  }


  /**
    * Run the NLD unparser on the ADL/TDD AST
    */
  private short generateDoc(Node topNode) {

    try {
      nlunparser = new JNLDUnparserVisitor(toolName, nlengine);
      nlunparser.setDirectory(docOutputDirectory);

      nlunparser.unparse(topNode);
    }
    catch(Throwable e) {
      System.err.println("Error: " +
			 "no documentation generated for this file, " +
			 "proceeding anyway");
      if (debug)
        e.printStackTrace();

      return docFailed;
    }

    return compilationOK;
  }

}

// ***********************************************************************
// History Log:
// $Log: JCompiler.java,v $
// Revision 1.33  1998/07/30 07:19:48  mallet
// Encapsulated the second part of the compilation within a test that
// verifies that the code generation has been requested: fixes a bug
// where the translator was started even when only the documentation has
// been requested.
// Fixed a bug in method translate() that was returning null instead of
// an empty vector when there is no ADL class or TDD class declaration in
// the compiled file.
//
// Revision 1.32  1998/06/30 07:23:58  mallet
// Turned debug mode off for delivery.
//
// Revision 1.31  1998/06/05 15:55:37  mallet
// Added test of number of test directives to determine whether the
// makefiles and scenario have to be generated.
//
// Revision 1.30  1998/06/04 16:30:23  bareau
// Added a public static method "isNLDFile(String fileName)" so that
// compilers just parse "pure nld" files and don't try to transform and
// unparse them.
//
// Revision 1.29  1998/05/26 15:21:35  bareau
// Bug fix: sgml files are no more generated in adltest directory.
//
// Revision 1.28  1998/05/15 16:21:17  bareau
// Must continue if no prolog engine.
//
// Revision 1.27  1998/05/15 14:22:54  mallet
// Changed signature of method compile to return a short: updated the
// code to return as soon as an error is detected (except for docFailed).
// Fixed small bug with reporting "Nothing to generate" when no code
// generation was requested.
//
// Revision 1.26  1998/05/15 13:26:47  bareau
// Added documentation generation.
//
// Revision 1.25  1998/05/12 12:40:41  shane
// Fixed references to docRequested.
//
// Revision 1.24  1998/04/15 13:49:46  bareau
// Added a handle "nlengine" (of class nlengine.NLEngine) associated
// to the compiler, to deal with NL processing.
//
// Revision 1.23  1998/03/25 15:00:46  cinard
// Parsing now use the ParseUnit class.
//
// Revision 1.22  1998/03/19 17:39:02  bareau
// Importations in case of unnamed packages are now dealt with using
// property "classesFromUnnamedPackage".
// (see translators.j.JADLTransformVisitor).
//
// Revision 1.21  1998/03/19 11:16:55  cinard
// Fixed a bug (procced set to false if type check fails).
//
// Revision 1.20  1998/03/19 09:19:50  cinard
// Moved parsing processing (especially parsing of used tdd files)
// into the parser.
//
// Revision 1.19  1998/03/18 16:20:49  cinard
// Added support for "Use" clause. Pure parsing and type checks are
// now distincts, so that all "used" files can be parsed before
// type check takes place.
//
// Revision 1.18  1998/03/10 17:26:29  mallet
// Merged v1.1 with development trunk
//
// Revision 1.17.2.3  1998/03/06 16:32:01  bareau
// Integrated the mechanism for unnamed packages: each used class will be
// imported in generated files.
//
// Revision 1.17.2.2  1998/03/04 18:48:31  cinard
// Now a scenario can contains sections for development targets coming from
// several tdd files.
//
// Revision 1.17.2.1  1998/03/04 10:08:48  bareau
// The adl/tdd translation is now performed via the new entry point method
// "translate" of the adl/tdd visitors. No need to perform a "reset" anymore.
//
// Revision 1.17  1998/02/27 13:49:28  mallet
// Updated copyright notice.
//
// Revision 1.16  1998/02/27 12:59:56  cinard
// Updated call to the parser parse method. No more ParseException is thrown
// anymore.
//
// Revision 1.15  1998/02/25 08:30:18  mallet
// Moved private field toolName in ADLCompiler: now protected.
//
// Revision 1.14  1998/02/25 08:09:24  mallet
// Added '/adltest' to outputDirectory in method compile to generate in
// the adltest subdirectory.
//
// Revision 1.13  1998/02/17 10:57:07  mallet
// Merged v1.0 with development trunk
//
// Revision 1.12.2.3  1998/02/16 08:23:26  mallet
// Fixed small bug in translate method: now an explicit test of TDD class is
// made to avoid going into this branch for a file that does not contain
// a TDD class (e.g., with only an import declaration in Java).
//
// Revision 1.12.2.2  1998/02/13 16:43:00  bareau
// Bug fix: the translator was not launched when the ADL_CompilationUnit
// after parsing has no "import" node (case of a test of a class with no
// explicit package).
//
// Revision 1.12.2.1  1998/02/13 16:04:50  mallet
// Fixed bug in translate: was trying to generate TDD releated files out
// of an ADL input when ACO was not requested.
// Fixed bug in translate: makefile was not generated when test data or
// test cases were not requested.
// Fixed bug in translate: confusion between makefileRequested and
// devMakefileRequested caused Makefile to be generated only when the
// latter was true.
//
// Revision 1.12  1998/02/12 16:01:52  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.11  1998/02/11 10:14:53  bareau
// The ADL translator can now return null when it detects an error:
// nothing to generate in this case.
//
// Revision 1.10  1998/02/10 18:17:55  mallet
// Moved initialization of parser in compiler after test of fileStream:
// it doesn't make sense to initialize the parser if the stream is null.
//
// Revision 1.9  1998/02/10 15:29:59  cinard
// Now use generic scenario generator.
//
// Revision 1.8  1998/02/10 15:11:57  cinard
// Added scenario generation through calls to method generateScenario()
// from class org.opengroup.adl.unparser.j.JScenarioGenerator.
//
// Revision 1.7  1998/02/02 12:38:29  cinard
// Delayed ADLJavaParser construction (to avoid initializing it with
// standard input).
// Added Makefile generation through calls to generateMakefile() and
// generateDevMakefile() in org.opengroup.adl.unparser.j.JMakefileGenerator
// class.
//
// Revision 1.6  1998/01/26 15:32:04  mallet
// Changed reporting to be done only in verbose mode.
//
// Revision 1.5  1998/01/16 13:24:36  bareau
// Bug fix: output directory set to "." when compiling local files.
//
// Revision 1.4  1998/01/08 08:32:09  mallet
// updated and fixed javadoc comments.
//
// Revision 1.3  1998/01/05 16:55:02  mallet
// Fixed javadoc comments.
// Added use of the setDirectory method in unparser to specify output
// directory.
// Added use of the getStream method for extension filters.
// Improved compile method.
// Added check in translate method to be sure that the AST returned by
// the parser contains at least one ADL or TDD class.
//
// Revision 1.2  1997/12/24 11:26:54  mallet
// Added translation phase.
//
// Revision 1.1  1997/12/19 14:06:00  mallet
// Concrete implementation of the ADLCompiler for the Java language binding.
//
// ***********************************************************************
