// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/jbcfile/ClassImporter.java,v $
// $Revision: 1.12 $
// $Date: 1998/09/04 14:54:04 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   jbcfile
// NAME:        ClassImporter.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Nov 06,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.jbcfile;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.lang.reflect.*;
import com.objectspace.jgl.HashSet;

/**
 * ClassImporter enable to read bytecode and create appropriate objects
 * accordingly. Using ClassImporter prevents the JVM to actually load
 * the classes for which bytecode file is to be read.
 */

public class ClassImporter
{

  /** 
   * Constructs class importer.Bytecode files are searched by default in
   * Java classpath.
   */
  public ClassImporter()
  {
     this(System.getProperty("java.class.path",""));
  }

  /** 
   * Constructs class importer.
   * @param classPath class path where bytecode shall be searched.
   */
  public ClassImporter(String classPath)
  {
     // Store class path in a more convenient way.
     StringTokenizer st = new StringTokenizer
                              (classPath,File.pathSeparator);
     _classPath = new String[st.countTokens()];
     for ( int i = 0; i < _classPath.length; i++ )
     {
        _classPath[i] = (String)st.nextToken();
     }

     // cache zip files to speed-up access
     cacheZipFiles();
  }

 /**
  * Can package be imported.
  * @param packageName java name of the package.
  */

  public boolean canImportPackage(String packageName)
  {
     if (debug)
     { 
        System.out.println
        ( "debug: ClassImporter.canImportPackage(" + packageName + ")");
     }
   
     boolean ok = false;

     // Replace . (java) with file separator.
     String packagePath = packageName.replace('.',File.separatorChar) 
                        + File.separator;
     // Entries in a Zip or Jar file always use the character '/' as a 
     // file separator
     String packageZipPath = packageName.replace('.','/') + '/';


     // First, search for package in class path.
     for (int i = 0; i < _classPath.length; i++)
     {
         String path = _classPath[i];
         if ( path.endsWith(".zip") || path.endsWith(".jar") )
         {
             ZipFile zipFile = (ZipFile)zipFiles.get(path);
             if ( zipFile != null  )
             {
                // archive file
                ZipEntry zipEntry = zipFile.getEntry
                                    (packageZipPath);

                if (zipEntry != null && zipEntry.isDirectory() )
                {
                   ok = true;
                   break;
                }
             }
         }
         else
         {
             // normal path
             File file = new File(path,packagePath);

             if ( file.isDirectory() )
             {
                ok = true;
                break;
             }
         }
     }

     // last chance to import the package
     if ( ! ok ) 
     {
	ok = ( _packageNames.count(packagePath) > 0 );
     }

     if (debug)
     { 
        System.out.println
        ( "debug: ClassImporter.canImportPackage(" + packageName + ")"
         + " return " + ok);
     }

     return ok;
  }

  /**
   * Import a class file.
   * @param className fully qualified name of the class.
   * @return ClassFile object created from the bytecode.
   */

  public ClassFile importClass(String className)
  {

     if (debug) 
     {
        System.out.println 
        ("debug: ClassImporter.importClass(" + className + ")");
     }

     ClassFile   cf         = null;
     InputStream is         = null;

     // Get a file name out of the class name.
     String fileName   = className.replace('.',File.separatorChar) + ".class";
     // Entries in a Zip or Jar file always use the character '/' as a 
     // file separator
     String zipFileName   = className.replace('.','/') + ".class";

     // First, search for class file in class path.

     try 
     {
     LOOP:
        for (int pathIdx = 0; pathIdx < _classPath.length; pathIdx++)
        {
           String path = _classPath[pathIdx];
           if ( path.endsWith(".zip") || path.endsWith(".jar") )
           {
             ZipFile zipFile = (ZipFile)zipFiles.get(path);
             if ( zipFile != null  )
             {
                // archive file
                ZipEntry zipEntry = zipFile.getEntry(zipFileName);
                if (zipEntry != null)
                {
                   is = zipFile.getInputStream(zipEntry);

                   // get class file object
                   try
                   {
                      DataInputStream dis = new DataInputStream 
                                            ( new BufferedInputStream (is));
                      cf =  new ClassFile(dis,className);
                      break LOOP;
                   }
                   catch ( ClassFileException e ) 
                   {
                      // keep searching  
                      cf = null;
                   }
                }
             }
           }
           else
           {
              // normal path
              File file = new File(path,fileName);
              if ( file.exists() )
              {
                 // get byte code input stream
                 is = new FileInputStream(file);

                 // get class file object
                 try
                 {
                    DataInputStream dis = new DataInputStream 
                                          ( new BufferedInputStream (is));
                    cf = new ClassFile(dis,className);
                    break LOOP;
                 }
                 catch ( ClassFileException e ) 
                 {
                    // keep searching
                    cf = null;
                 }
              }
           }
        }
     }
     catch ( ZipException e ) { cf = null; }
     catch ( IOException  e ) { cf = null; }

     
     if (debug) 
     {
        System.out.println 
        (  "debug: ClassImporter.importClass(" + className + ")"
         + " return " + ( cf != null ? cf.getScopedName() : null));
     }

     return cf;
  }

  // -- Unit testings --

  public static final void main(String[] args)
  {
     ClassImporter classImporter = new ClassImporter();
     String fileName = ( args.length == 0 ? "Object" : args[0] );
     ClassFile cf = classImporter.importClass(fileName);
     System.out.println("Importing " + fileName);
     System.out.println(cf);
  }

  // -- Private methods --

  /** 
   * Cache archive files (.zip, .jar)  that are in classpath.
   * This increase speed when searching into these archives.
   */

  private void cacheZipFiles()
  {
     // First, search for package in class path.
     for (int i = 0; i < _classPath.length; i++)
     {
         String path = _classPath[i];
         if ( path.endsWith(".zip") || path.endsWith(".jar") )
         {
             File file = new File(path);
             if ( file.exists() )
             {
                // archive file
                try
                {
                   ZipFile zipFile = new ZipFile(path);
                   if ( zipFile != null ) 
                   {  
                      zipFiles.put(path,zipFile);
                      // look for package entries in zip files
                      checkForPackageNames(zipFile);
                   }
                }
                catch ( java.io.IOException e ) { /* who cares */ }
             }
         }
     }
  }

  /** Check for package name entries in zip file. */
  private void checkForPackageNames(ZipFile zipFile)
  {
     Enumeration zipEntries = zipFile.entries();
     while ( zipEntries.hasMoreElements() )
     {
        ZipEntry zipEntry = (ZipEntry)zipEntries.nextElement();
	String name = zipEntry.getName();
	// Entries in a Zip or Jar file always use the character '/' as a 
	// file separator
        int index = name.indexOf('/');
        while ( index != -1 )
        {
	   String packageName = name.substring(0,index+1);
           _packageNames.put(packageName);
           index = name.indexOf('/',index+1);
        }
     }
  }

  // -- Private attributes --

  /** Cached opened java zip files. */

  private Hashtable zipFiles = new Hashtable();

  /** Cached package names for which no zip entries exists. */
  
  private HashSet _packageNames = new HashSet();

  /** Components of the class path. */

  private String[] _classPath;

  /** Debug flag. */

  private static final boolean debug = false;

}

// ***********************************************************************
// History Log:
// $Log: ClassImporter.java,v $
// Revision 1.12  1998/09/04 14:54:04  mallet
// Fixed bug in canImportPackage, importClass and checkForPackageNames:
// when looking for a class in a Jar or Zip file, don't use the
// File.separator but '/' - java.util.zip.ZipEntry abstracts the file
// system dependencies.
//
// Revision 1.11  1998/06/10 17:30:36  cinard
// Improved package look-up algorithm : now package entry can be detected
// in a jar file even if no specific entry for the package exists in
// the jar file.
//
// Revision 1.10  1998/02/27 12:34:15  cinard
// Cleaned up the API for jbcfile components by introducing a user friendly
// representation for bytecode types (BCType class).
// Updated copyright notice.
//
// Revision 1.9  1998/02/23 14:19:59  cinard
// Improved byte code import algorithm.
//
// Revision 1.8  1998/02/12 16:01:54  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.7  1998/01/28 18:11:28  cinard
// Added a check when importing java class bytecode (qualified name of java class
// must be equals to the qualified name in the bytecode).
//
// Revision 1.6  1998/01/26 14:59:03  mallet
// Prepended adl.incpath before java.class.path property for search path
// of class files.
//
// Revision 1.5  1998/01/12 13:18:25  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.4  1997/12/08 12:42:08  cinard
// Added caching of JAR and ZIP archive files.
//
// Revision 1.3  1997/11/26 15:49:09  cinard
// Improved Makefile. Changed package name.
//
// Revision 1.2  1997/11/26 14:17:56  cinard
// Distinguished class file check (where full class name is got) from
// class file import => optimization of the JSymbolLoader class.
//
// Revision 1.1  1997/11/12 12:14:38  cinard
// Tool classes  enabling to read java bytecode.
//
// ***********************************************************************
