// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/nlengine/NLEngine.java,v $
// $Revision: 1.8 $
// $Date: 1998/07/03 08:01:30 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   nlengine
// NAME:        NLEngine.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Apr 3,  1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.nlengine;

import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.tools.FilterProcess;
import java.io.*;
import java.util.Vector;
import java.util.Stack;
import java.util.Enumeration;
import java.util.Hashtable;

/**
  * Interface between the translation modules and the Prolog engine.
  * <P><UL>
  * <LI>An NLEngine is created by the ADLCompiler.
  * <LI> During parsing, the parsers send "feed" messages that
  * correspond to NL mapping definitions that enrich the Prolog
  * engine database. In the case of nld definitions intermapped into
  * ADL/TDD code, information on current scope will also be provided.
  * <LI>During unparsing, the NL unparsers will send "query" messages
  * that correspond to expressions to be translated; NLEngine is here a
  * 2 way interface: the Prolog engine sends back an answer to the query.
  * </UL>
  */

public class NLEngine {

  /** Very useful flag for debugging: will print on stderr any command
    * sent to the Prolog engine, and the returned result. */
  private static final boolean seePrologEngineInOut = false;

  private static final boolean debug = false;

  /** Link with the Prolog engine */
  private FilterProcess prologEngine;

  /** In-channel of the Prolog engine */
  private BufferedWriter prologEngineIn;

  /** Out-channel of the Prolog engine */
  private BufferedReader prologEngineOut;

  /** Vector of components for the current Prolog scope */
  private Vector prologScope;

  /** String for the Prolog scope of the current mapping */
  private String mapPrologScope;

  /** Flag for mappings of methods */
  private boolean mapfn;

  /** Number of parameters of the current method map */
  private int nbParams;

  /** The substring at the beginning of the answer of the Prolog
    * engine, when the translation is OK */
  private final String openNLD = ">>>NL Answer>>>";

  /** The substring at the end of the answer of the Prolog
    * engine, when the translation is OK */
  private final String endNLD = "<<<NL Answer<<<";

  /** Hashtable for user-defined SGML entities */
  private Hashtable entities;


  /**
    * Constructor of the NLEngine.
    * <P>Initializes the Prolog engine with the NL rules file,
    * and its in/out channels.
    * Initializes the map/mapfn databases with default values.
    */
  public NLEngine() throws FileNotFoundException, IOException {

    entities = new Hashtable();
    String nlpath = System.getProperty("adl.home") + File.separator;
    String nlrulesPath = nlpath + "lib" + File.separator + "nl" + File.separator +
                         getLocale() + File.separator + "nl_rules";
    File nlrulesFile = new File(nlrulesPath);
    if (! nlrulesFile.exists())
      throw new FileNotFoundException("No Prolog rules file found: " +
                                      nlrulesFile.getCanonicalPath());

    String prologExecutable = System.getProperty("adl.prolog");

    boolean ok = false;
    if (prologExecutable != null) {
      try {
        prologEngine = new FilterProcess(prologExecutable + " -f " + nlrulesPath);
        ok = true;
      } catch(IOException e) {
        System.err.println("Cannot run Prolog " + prologExecutable +
                           "... Running default $ADL2HOME/bin/pl");
      }
    }

    if (! ok) {
      prologEngine = new FilterProcess(nlpath + "bin" + File.separator + "pl" +
                                       " -f " + nlrulesPath);
    }

    prologEngineIn = new BufferedWriter(new OutputStreamWriter(prologEngine.getOutputStream()));
    prologEngineOut = new BufferedReader(new InputStreamReader(prologEngine.getInputStream()));

    toPrompt();

    feed("assertz(mapEntity(X, _, _, [])).\n");
    feed("assertz(map(X, _, X, [], _)).\n");
    feed("assertz(mapfn(X, _, X, _, _, _)).\n");
  }


  /**
    * Close the Prolog engine.
    */
  public void close() throws IOException {

    prologEngine.close();
  }


  /**
    * Set the current scope, in "Scope" format and in
    * "prolog" format.
    * @see #ORG.opengroup.adl.symboltable.Scope
    * @see setPrologScope
    */
  public void setScope(Scope s) {

    if (s instanceof LModule) {
      prologScope = new Vector();
    }
    else {
      prologScope = s.getNLName();
    }
    prologScope.insertElementAt("#<EXTERN>", 0);
  }


  /**
    * Set the scope for the current mapping
    * @see #ORG.opengroup.adl.symboltable.Scope
    * @see setPrologScope
    */
  public void setScopedName(Vector scopedName) {

    if (debug)
      System.out.println("setScopedName: " + scopedName);

    Vector vectScope = new Vector();
    Enumeration enum = scopedName.elements();
    String first = (String)enum.nextElement();
    if (first.equals("#current")) {
      for (int i = 0; i < prologScope.size(); i++) {
        vectScope.addElement(new String((String)prologScope.elementAt(i)));
      }
    }
    else {
        vectScope.addElement("#<EXTERN>");
    }

    while (enum.hasMoreElements()) {
      vectScope.addElement(enum.nextElement());
    }
    mapPrologScope = scope2Str(vectScope, true);
  }


  /**
    * Feed the Prolog database with a mapping ("asserta" predicate).
    * This is a one-way communication: no answer is expected.
    * @param mapping the NL mapping
    * @param userPreds user-defined predicates associated with the symbol
    * @param systemPreds pre-defined predicates associated with the symbol
    * @param paramPredicates selector predicates associated with the symbol
    * (only when the current mapping is for a method/inline/function).
    */
  public void map(Vector mapping, Vector userPreds,
                  Vector systemPreds, Vector paramPredicates) {

    String asserta = "asserta(map" + (mapfn ? "fn(" : "(") + mapPrologScope +
      ", " + vector2StrBuf(mapping, false);

    if (mapfn) {
      int nbp = paramPredicates.size();
      Enumeration enum = paramPredicates.elements();
      StringBuffer params = new StringBuffer("[");
      boolean notfirst = false;
      while (enum.hasMoreElements()) {
        if (notfirst)
          params.append(", ");
        Vector v = (Vector)enum.nextElement();
        if (v == null) {
          params.append("[]");
        }
        else {
          params.append(vector2StrBuf(v, false));
        }
        notfirst = true;
      }
      if ((nbp == 0) && (nbParams > 0)) {
        params.append("[]");
        nbp = 1;
      }
      for (int i = nbp; i < nbParams; i++) {
        params.append(", []");
      }
      asserta += ", " + params + "]";
    }

    asserta += ", " + vector2StrBuf(userPreds, false) + ", " +
               vector2StrBuf(systemPreds, false) + ")).";

    feed(asserta);
  }


  public void mapProperty(String entityName, String entityText) {

    feed("asserta(mapEntity(" + mapPrologScope + ", " +
         entityName + ", " + entityText + ")).");
  }


  public void mapIncludes(Vector v) {

    mapProperty("%includes", vector2StrBuf(v, false).toString());
  }


  public void mapUserEntity(String entityName, String entityText) {

    entities.put(entityName, entityText);
  }


  public Hashtable getEntities() {

    return entities;
  }


  /**
    * Queries the Prolog database for a translation ("xlate" predicate).
    * This is a two-way communication: an answer is expected.
    *
    * @param cmd The Prolog command
    * @param scope The scope of the expression to translate
    * @return the translation string
    */
  public String query(String cmd, Scope scope) throws NLException {

    return query(cmd, scope, null);
  }

  /**
    * Queries the Prolog database for a translation ("xlate" predicate,
    * or "xlateEntity" predicate if the third argument is not null).
    * This is a two-way communication: an answer is expected.
    *
    * @param cmd The Prolog command.
    * @param scope The scope of the expression to translate.
    * @param prop The property to search a mapping for ("description",
    * "includes", "seeAlso" or "purpose").
    * @return the translation string
    */
  public String query(String cmd, Scope scope, String prop) throws NLException {

    String prologCmd = null;
    if (prop == null || prop.equals("")) {
      prologCmd = "xlate(" + cmd + ", " + NLScopeScopeList(scope) + ").";
    }
    else {
      prologCmd = "xlateEntity('" + cmd + "', " + prop +
        ", " + NLScopeScopeList(scope) + ").";
    }

    write(prologCmd);

    String s1, s2; int i1, i2;
    s1 = toPrompt();
    i1 = s1.indexOf(openNLD);
    i2 = s1.indexOf(endNLD);
    if ((i1 == -1) || (i2 == -1)) {
      throw new NLException("translation failed: " + cmd);
    }
    s2 = s1.substring(i1 + openNLD.length(), i2);

    if (seePrologEngineInOut)
      System.err.println("TRANSLATION:\n" + s2);

    return s2;
  }


  /**
    * Returns a atring that corresponds to the search scope for a
    * method, in the prolog engine format, i.e. a couple
    * [Scope, ScopeList] where Scope is the current scope and
    * ScopeList corresponds to the ancestors of the current class.
    * @param scope the current scope
    * @return the string "[<scope>, [<scopelist>]]"
    */
  public String NLScopeScopeList(Scope scope) {

    if (scope == null) {
      return "['#<EXTERN>'], [[]]";
    }

    Scope ds = scope;
    while (! ((ds == null) ||
              (ds instanceof LClass) ||
              (ds instanceof TDDClass) ||
              (ds instanceof LCompilationUnit))) {
      if (ds instanceof ADLClass) {
        ds = ((ADLClass)ds).getLClass();
      }
      else {
        ds = ds.getParent();
      }
    }

    if (ds == null) {
      // package scope or "C aldmodule"
      Vector v = scope.getNLName();
      v.insertElementAt("#<EXTERN>", 0);
      return scope2Str(v, false) + ", [[]]";
    }
    if (ds instanceof LCompilationUnit) {
      return "['#<EXTERN>'], [[]]";
    }

    String scopeList = "[" + vector2StrBuf(ds.getNLName(), true);
    if (ds instanceof LClass) { // if ds instanceof TDDClass, no ancestor
      LClass currentClass = (LClass)ds;
      LClass[] ancestors = currentClass.getSuperClasses();
      for (int i = 0; i < ancestors.length; i++) {
        scopeList += ", " + vector2StrBuf(ancestors[i].getNLName(), true);
      }
    }
    scopeList += "]";
      
    Stack scopeStack = new Stack(); Scope ds2 = scope;
    while (ds2 != ds) {
      if (ds2 instanceof ADLClass) {
        ds2 = ((ADLClass)ds2).getLClass();
      }
      else {
        if (ds2 instanceof LBlockScope) {
          // do nothing
        }
        else if (ds2 instanceof LCallable) {
          scopeStack.push(((LCallable)ds2).getNLSignature());
          scopeStack.push(((LCallable)ds2).getName());
        }
        else if (ds2 instanceof LConstructedType) {
          scopeStack.push(((LConstructedType)ds2).getName());
        }
        ds2 = ds2.getParent();
      }
    }

    Vector v = new Vector();
    v.addElement("#<CLASS>");
    while (! scopeStack.empty()) {
      v.addElement(scopeStack.pop());
    }
    return scope2Str(v, false) + ", " + scopeList;
  }


  /**
    * Feed the Prolog database with a mapping ("asserta" predicate).
    * This is a one-way communication: no answer is expected.
    *
    * @param cmd The Prolog command
    */
  private void feed(String cmd) {

    write(prune(cmd));
    toPrompt();
  }


  /**
    * Send a string to the Prolog engine
    */
  private void write(String s) {

    if(seePrologEngineInOut)
      System.err.println("WRITING: \n" + s);

    try {
      prologEngineIn.write(s, 0, s.length());
      prologEngineIn.newLine();
      prologEngineIn.flush();
    }
    catch(IOException e) {
      System.err.println("I/O Error: Cannot write to the Prolog interpreter");
    }
  }


  /**
    * Wait for the Prolog engine's computation end, and returns
    * the result string of this computation.
    * <P><B>Note: </B>this is highly dependant on the Prolog
    * interpreter (the SWI Prolog's prompt is "?- ").
    * @return the string returned by the Prolog engine
    */
  private String toPrompt() {

    int i = 0; char c; int bufSize = 512;
    char[] buf = new char[bufSize];

    try {
      c = (char)prologEngineOut.read();
      while ((c != '?') && prologEngineOut.ready()) {
        if (i == bufSize) {
          char[] tmpbuf = new char[bufSize+512];
          System.arraycopy(buf, 0, tmpbuf, 0, bufSize);
          buf = tmpbuf;
          bufSize += 512;
        }
        buf[i++] = c;
        c = (char)prologEngineOut.read();
      }
      buf[i++] = (char)prologEngineOut.read();
      buf[i++] = (char)prologEngineOut.read();
      if ((buf[i-2] != '-') || (buf[i-1] != ' ')) {
        return new String(buf) + toPrompt();
      }
      return new String(buf);
    }
    catch(IOException e) {
      System.err.println("I/O Error: Cannot read from the Prolog interpreter");
      return "";
    }
  }


  /** Return the current locale */
  private String getLocale() {

    String locale = System.getProperty("adl.locale");
    if ((locale == null) || (locale.equals("")))
      locale = "C";
    return locale;
  }


  /**
    * Simple tool to transform a Vector into a StringBuffer with the format
    * "[component1, component2, ...]".
    */
  private StringBuffer vector2StrBuf(Vector v, boolean b) {

    Enumeration enum = v.elements();
    StringBuffer ret = new StringBuffer("[");
    if (enum.hasMoreElements()) {
      if (b)
        ret.append('\'').append(enum.nextElement()).append('\'');
      else
        ret.append(enum.nextElement());
      while (enum.hasMoreElements()) {
        if (b)
          ret.append(", '").append(enum.nextElement()).append('\'');
        else
          ret.append(", ").append(enum.nextElement());
      }
    }
    ret.append("]");
    return ret;
  }


  /**
    * Transforms a vector scope (i.e. a vector where each component
    * represents a refinement of the scope, from the most global to
    * the most local) into a string that represents this scope in the
    * format expected by the prolog engine.
    * <P>Two distinct element strings are created: theSymbol represents
    * the last element of the scope, and theScope the preceding ones.
    * When the boolean parameter of scope2Str is true (called for
    * mapping feeding), the return string is "theSymbol, [theScope]"
    * and otherwise (called for query), the return string is
    * "[theScope, theSymbol]" (or "[theSymbol]" if theScope is null).
    */
  private String scope2Str(Vector v, boolean mapp) {

    mapfn = false;
    nbParams = 0;

    int scopesize = v.size();
    if (scopesize == 0) {
      return "[]";
    }

    String sc1 = "", sc2, theScope = null, theSymbol = null, s1, s2;
    boolean notfirst = false;
    for (int i = 0; i < scopesize; i++) {
      s1 = (String)v.elementAt(i);
      if (i == scopesize-1) {
        theSymbol = "'" + s1 + "'";
        theScope = sc1;
      }
      else {
        s2 = (String)v.elementAt(i+1);
        if (i == scopesize-2) {
          if (s2.charAt(0) == '(') {
            if (mapp) {
              theSymbol = ("'" + s1 + s2 + "'");
              theScope = sc1;
              mapfn = true;
              if (s2.equals("()") || s2.equals("(*)"))
                nbParams = 0;
              else {
                nbParams = 1;
                int comma = s2.indexOf(',');
                while (comma != -1) {
                  nbParams += 1;
                  comma = s2.indexOf(',', comma+1);
                }
              }
            }
            else {
              theScope = "";
              theSymbol = sc1 + (notfirst ? ", ['" : "['") + s1 + "', '" + s2 + "']";
            }
          }
          else {
            theSymbol = "'" + s2 + "'";
            theScope = sc1 + (notfirst ? ", '" : "'") + s1 + "'";
          }
          break;
        }
        else {
          if (s2.charAt(0) == '(') {
            i += 1;
            sc1 += (notfirst ? ", ['" : "['") + s1 + "', '" + s2 + "']";
          }
          else {
            sc1 += (notfirst ? ", '" : "'") + s1 + "'";
          }
          notfirst = true;
        }
      }
    }

    return mapp ? (theSymbol + ", [" + theScope + "]") :
      ("[" + (theScope.equals("") ? "" : (theScope + ", ")) + theSymbol + "]");
  }


  private String prune(String str) {

    String ret = ""; int index1 = 0, index2;
    do {
      index2 = str.indexOf('?', index1);
      if (index2 != -1) {
        ret += str.substring(index1, index2) + "&quest;";
        index1 = index2+1;
      }
    } while (index2 != -1);
    ret += str.substring(index1);
    return ret;
  }

}

// ***********************************************************************
// History Log:
// $Log: NLEngine.java,v $
// Revision 1.8  1998/07/03 08:01:30  bareau
// Fixed the printing of the couple (Scope, ScopeList) in the case
// of a C adlmodule.
//
// Revision 1.7  1998/06/30 11:15:27  bareau
// Bug fix: mappings for Java external components (with fully qualified
// names) were not correctly supported. The fix consists in adding "extern"
// in scopes.
//
// Revision 1.6  1998/06/12 13:17:26  bareau
// Read Prolog interpreter as system property "adl.prolog"
// (see $ADL2HOME/Adlt.rc).
//
// Revision 1.5  1998/06/10 15:33:03  bareau
// Updated the extern/local mapping behavior according to the new
// scoping mechanism in the NLD parser ("." instead of #current).
//
// Revision 1.4  1998/05/26 15:18:55  bareau
// Changed the call to the Prolog engine (no SWI_HOME_DIR variable).
// Changed a few names and comments.
//
// Revision 1.3  1998/05/15 13:42:23  bareau
// Complete revision.
//
// Revision 1.2  1998/04/15 16:23:33  bareau
// Changed the path to the Prolog interpreter ($ADL2HOME/bin/nl_engine).
//
// Revision 1.1  1998/04/15 13:47:46  bareau
// Changed NL_Engine into NLEngine.
//
//
// ***********************************************************************
