// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/ADLNode.java,v $
// $Revision: 1.2 $
// $Date: 1998/06/29 16:07:51 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL Parsers
// NAME:        ADLNode.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Mar 31, 1998: Created from adljava/SimpleNode
//
// ************************************************************************

package ORG.opengroup.adl.parsers;

import java.io.OutputStream;
import java.io.PrintStream;

import java.util.Enumeration;
import java.util.Vector;

/**
 * The root class for the AST node hierarchy.  It implements all the
 * functionality for accessing node properties.
 */
public abstract class ADLNode implements Cloneable {

  /** The type identity number of the current node */
  protected int id;

  /** First parser token of the current node */
  public Token firstToken;

  /** Last parser token of the current node */
  public Token lastToken;

  /**
   * Field property containing a unique integer to identify the node.
   * Serial numbers are allocated sequentially.
   */
  public int serialNumber;	// property

  private static int last_serial_number = 0;

  /** Vector o properties of the current node */
  private PropertySet properties = null;

  /**
    * Each node type must implement a one argument constructor to
    * create an empty node.  The new node's association table will be
    * empty, and its field properties will be set to their initial
    * values.
    */
  public ADLNode(int i) {
    id = i;
    serialNumber = ++last_serial_number;
  }

  /**
    * Called by JJTree actions to notify the node that no more
    * children will be added. This routine is reserved for JJTree.
    * Each node type can override this routine to add properties for
    * its children. 
    *
    * @see #jjtGetChild
    */
  public void jjtClose() {
  }

  
  /**
    * This routine is reserved for JJTree.
    */
  public void jjtOpen() {
  }


  /**
    * @see #toString()
    */
  public String toString(String prefix) {
    return prefix + toString();
  }


  /**
   * Returns expression image associated a particular node
   */
  public final String getExpressionImage() {

    if (firstToken == null) {
      System.out.println("null token for node " + this);
      return "";
    } else {
      StringBuffer str = new StringBuffer();

      str.append(addEscapes(firstToken.image));
      
      if (firstToken != lastToken) {
	for (Token t = firstToken.next;
	     (t != null) && (t != lastToken);
	     t = t.next) {
	  // Put a blank in between token images.
	  str.append(" ");
	  
	  str.append(addEscapes(t.image));
	}
      }
      return str.toString();
    }
  }

  /**
   * Returns the source text for a given AST node.
   */
  public final String getSourceText() {

    StringBuffer str = new StringBuffer();
    Token t = firstToken, tt;

    while ((t != null) && (t != lastToken)) {
      str.append(addEscapes(t.sourceImage()));
      t = t.next;
    }

    return str.toString();
  }

  /**
   * Copy first and last tokens from one node to another.
   */
  public final void copySourceTokens(ADLNode n) {
    firstToken = n.firstToken;
    lastToken = n.lastToken;
  }

  /**
   * Prepend a new token at the beginning of the token chain.
   */
  public final void prependToken(Token t) {
    t.next = firstToken;
    firstToken = t;
  }
   
  /**
   * Append a new token at the end of the token chain.
   */
  public final void appendToken(Token t) {
    if (firstToken == null) {
      firstToken = t;
      lastToken = t.next;
    } else {
      Token tt = firstToken;

      while (tt.next != null && tt.next != lastToken) { tt = tt.next; }
      
      tt.next = t;
      lastToken = t.next;
    }
  }
   
  /**
   * Escape special characters in strings.
   */
  private final String addEscapes(String str) {
    StringBuffer retval = new StringBuffer();
    for (int i = 0; i < str.length(); i++) {
      switch (str.charAt(i))
        {
// 	case 0 :
// 	  continue;
// 	case '\b':
// 	  retval.append("\\b");
// 	  continue;
// 	case '\t':
// 	  retval.append("\\t");
// 	  continue;
// 	case '\n':
// 	  retval.append("\\n");
// 	  continue;
// 	case '\f':
// 	  retval.append("\\f");
// 	  continue;
// 	case '\r':
// 	  retval.append("\\r");
// 	  continue;
// 	case '\'':
// 	  retval.append("\\\'");
// 	  continue;
	case '\"':
	  retval.append("\\\"");
	  continue;
	case '\\':
	  retval.append("\\\\");
	  continue;
	default:
	  retval.append(str.charAt(i));
	  continue;
        }
      }
      return retval.toString();
  }

  /**
    * Make a copy of the source node's properties.  If a property
    * exists in the source node but does not exist in the current node
    * it is copied.  A new property is created and added to the current
    * node's association table with the current value of the source
    * node's property as its value.
    */
  public final void copyFrom(ADLNode source) { //throws PropertyAccessException {
    if (properties != source.properties) {
      Enumeration e = source.getKeys();
      if (e != null) {
	while (e.hasMoreElements()) {
	  Object key = e.nextElement();
	  if (!hasProperty(key)) {
	    try {
	      setProperty(key, source.getProperty(key));
	    }
	    catch(PropertyAccessException exc) {
	      System.out.println("PropertyAccessException in copyFrom for property: " + key +
				 " " + exc.getMessage());
	    }
	  }
	}
      }
    }
  }


  /**
    * Get the value of the specified property.  If necessary, follow
    * the delegation chain to retrieve the value.  It is an error to
    * attempt to get the value of a nonexistent property.
    */
  public final Object getProperty(Object key) throws PropertyAccessException {
    if (properties == null)
      throw new PropertyAccessException("No Properties!");

    return properties.get(key);
  }


  /**
    * Set the specified property to the specified value.  If the
    * property does not exist it is created and added to the current
    * node's association table.  If it does exist and the node is
    * additive an error is signaled, otherwise the property is updated
    * to the new value.
    * <P>
    * An exception to this is if a lazy property has a setter method, in
    * which case it can be set more than once, regardless of whether the
    * node is additive or not.
    */
  public final void setProperty(Object key, Object value)
    throws PropertyAccessException
  {
    if (properties == null)
      properties = new PropertySet(this);

    properties.set(key, value);
  }


  /**
    * Get the value of the indexed property.  It is an error if the
    * property is nonexistent or cannot be indexed.
    */
  public final Object getProperty(Object key, int index)
    throws PropertyAccessException, ArrayIndexOutOfBoundsException
  {
    if (properties == null)
      throw new PropertyAccessException("No Properties!");

    return properties.get(key, index);
  }
  

  /**
    * Get or set the value of the indexed property.  It is an error if
    * the property is nonexistent or cannot be indexed.
    */
  public final void setProperty(Object key, int i, Object value)
    throws PropertyAccessException, ArrayIndexOutOfBoundsException
  {
    if (properties == null)
      properties = new PropertySet(this);

    properties.set(key, i, value);
  }


  /**
    * Test to determine whether the property already exists on this
    * node, following the delegation chain if necessary.
    */
  public final boolean hasProperty(Object key) {
    if (properties == null)
      return false;
    else
      return properties.has(key);
  }


  /**
    * Remove the specified property from the node.  If the node is
    * additive an error is signaled.  Any kind of property can be
    * removed regardless of whether it is represented as a table entry,
    * a node field or method, or a delegated property.  Be careful.
    */
  public final void removeProperty(Object key)
    throws NoSuchPropertyException
  {
    if (properties != null)
      properties.remove(key);
  }


  /**
    * Return an enumeration that produces the keys of all the
    * properties of the node.
    */
  public final Enumeration getKeys() {
    if (properties == null)
      return null;
    else
      return properties.keys();
  }


  /**
    * Return a deep clone of the current node
    */
  public Object clone() throws CloneNotSupportedException {
    ADLNode copy = (ADLNode)super.clone();

    // be sure the clone has its own serial number
    copy.serialNumber = ++last_serial_number;

    // there is no need to copy the properties since both nodes share the
    // same PropertySet object.
    // Maybe when PropertySet implements Cloneable...

    return copy;
  }



  /**
    * Produce a textual representation of the nodes and properties of
    * the AST.  The level argument controls the amount of detail in the
    * output.
    */
  public void dump(OutputStream o, int level) {
    dumpNode((PrintStream)o, "", " ", level);
  }


  protected void dumpNode(PrintStream o, String label, String pad, int level)
  {
    o.print(pad + label + ": " + toString());
    o.println("");

    dumpNodeProperties(o, pad, level);
  }

  protected void dumpProperty(PrintStream o, String pad, Object key)
  {
    try {
      o.println(pad + " " + key + "=" + getProperty(key));
    } catch (PropertyAccessException e) {
      o.println(pad + " " + key + " !!! " + e.getMessage());
    }
  }

  protected void dumpNodeProperties(PrintStream o, String pad, int level)
  {   
    Enumeration keys;
    if (level == Integer.MAX_VALUE) {
      keys = getKeys();
    } else {
      keys = getDumpKeys(level).elements();
    }
    while (keys != null && keys.hasMoreElements()) {
      Object pn = keys.nextElement();
      if (hasProperty(pn)) {
	dumpProperty(o, pad, pn);
      }
    }
  }
  protected Vector getDumpKeys(int level)
  {
    Vector keys = new Vector();
    if (level > 1000) {
      keys.addElement("SerialNumber");
    }
    return keys;
  }

}

// ***********************************************************************
// History Log:
// $Log: ADLNode.java,v $
// Revision 1.2  1998/06/29 16:07:51  mallet
// Added firstToken and lastToken fields to replace expressionImage
// property by a method getExpressionImage() based on the parsed token
// chain. firstToken and lastToken should be set for all nodes in
// parsers. firstToken is the first token of the parsed construction, and
// lastToken is the token following the last token of the construction.
// Added some token chain management methods: prependToken(),
// appendToken(), copySourceTokens().
// Added method getSourceText() for NLD unparsing of verbatim code.
// Made most methods final to speed up AST processing.
//
// Revision 1.1  1998/04/01 16:32:06  mallet
// Created ADLNode to factor some code from binding-dependent SimpleNode,
// especially the code related to property access.
//
// ***********************************************************************
