// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/CycleDetector.java,v $
// $Revision: 1.1 $
// $Date: 1998/02/27 15:30:24 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL Parser
// NAME:        CycleDetector.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Dec 23, 1997: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers;

import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.Pair;
import com.objectspace.jgl.HashMap;
import com.objectspace.jgl.HashSet;

/**
 * Detects cycles among dependencies using a mark-up technique.
 */

public class CycleDetector
{

 /**
  * States that first object depends on second object, and detect cycle
  * in case these dependency along with the former ones contanis a 
  * cycle.
  * Algorithm :
  * All dependencies after the new one (note: dependency(b,c) is after 
  * dependency(a,b)) are marked. And if a dependency should be marked twice,
  * a cycle is detected.
  * 
  * @param first  first object involved.
  * @param second second object involved.
  * @exception CycleException thrown if this dependency introduces
  *                           a cycle 
  */

  public void addDependency(Object first,Object second)
  throws CycleException
  {
     // Clear marked dependencies.
     markedDependencies.clear();

     if ( first.equals(second) )
     {
        Vector cycle = new Vector();
        cycle.addElement(first);
        cycle.addElement(second);
        throw new CycleException(cycle);
     }

     try
     {
        // Create a vector (pair)  for easy manipulation of the dependency
        Pair dependency = new Pair(first,second);
     
        // Add the dependency
        // . update forwardMap
        // . update reverseMap
        // . update map
        // Check dependency does not already exists

        Enumeration existingDeps = forwardMap.values(first);
        while ( existingDeps.hasMoreElements() )
        {
           Object depPoint = existingDeps.nextElement();
           if ( depPoint.equals(second) )
           {
              // dependency is already here, just return
              return;
           }
        }

        forwardMap.add(first,second);
        reverseMap.add(second,first);

        // Mark all dependencies

        Vector markedDependencyStarts = new Vector();
        markedDependencyStarts.addElement(first);
        
        while ( markedDependencyStarts.size() > 0 )
        {
           // Propagate flows
           markDependencies(markedDependencyStarts.elements(),first);
           // Get next generation of dependency
           markedDependencyStarts = 
                   getNextDependencies(markedDependencyStarts.elements());
        }
     }
     catch ( CycleException e )
     {
        // remove the cycle
        removeCycle(first,second);
        // rethrow exception
        throw e;
     }
  }


  /**
   * Returns a vector containing all the points that depends from
   * points in supplied enumeration.
   * @param points points for which the depenent points are to be searched.
   * @return dependent points.
   */

  private Vector getNextDependencies(Enumeration points)
  {
     Vector nextPoints = new Vector();
     while ( points.hasMoreElements() )
     {
        Object start = points.nextElement();

        // Add all dependencies starting with end to vector
        Enumeration e = forwardMap.values(start);
        while ( e.hasMoreElements() )
        {
           nextPoints.addElement(e.nextElement());
        }
     }
     return nextPoints;
      
  }
  
  /**
   * Mark all dependencies below the added one, in case first object
   * is reached again a cycle is detected.
   * @param points origins of the dependencies to mark
   * @param first object that should not be reached again
   * @exception ORG.opengroup.adl.parsers.adljava.CycleException
   *            thrown if a cycle is detected 
   */

  private void markDependencies(Enumeration points,Object first)
  throws CycleException
  {
     while ( points.hasMoreElements() )
     {
        Object point = points.nextElement();
        Enumeration dependentPoints = forwardMap.values(point);
        while ( dependentPoints.hasMoreElements() )
        {
           Object dependentPoint = dependentPoints.nextElement();
           Pair dep = new Pair(point,dependentPoint);
           markedDependencies.add(dep);
           if ( dependentPoint.equals(first) )
           {
              // Cycle detected
              // Isolate cycle
              Vector cycle = isolateCycle(dep);
              throw new CycleException(cycle);
           }
        }
     }
  }

  /**
   * Isolate a cycle that has just been detected.
   * @param the dependency where thecycle was detected.
   * @return the detected cycle.
   */

  private Vector isolateCycle(Pair dependency)
  {
     Vector cycle = new Vector();
     // save first dependency
     Object endPoint = dependency.second;
     cycle.insertElementAt(endPoint,0);
     Object currentPoint = dependency.first;
     cycle.insertElementAt(currentPoint,0);
     boolean ok = false;
     do
     {
        Enumeration formerPoints = reverseMap.values(currentPoint);
        while ( formerPoints.hasMoreElements() )
        {
           Object formerPoint = formerPoints.nextElement();
           if ( markedDependencies.count
                                   (new Pair(formerPoint,currentPoint)) > 0 )
           {
              // dependency found
              // update first
              currentPoint = formerPoint;
              cycle.insertElementAt(currentPoint,0);
              ok = true;
              break;
           }
        }
      }
      while ( ok && ( ! currentPoint.equals(endPoint) ) );
      return cycle;
  }

 /**
  * Remove the cycle just detected, so that the CycleDetector can be used
  * to detect other cycles.
  *
  * @param first  first object involved in the cycle.
  * @param second second object involved in the cycle.
  */

  private void removeCycle(Object first, Object second)
  {
     Enumeration e = forwardMap.values(first);
     forwardMap.remove(first);
     while ( e.hasMoreElements() )
     {
        Object o = e.nextElement();
        if ( ! o.equals(second) )
        {
           forwardMap.add(first,o);
        }
     }

     Enumeration f = reverseMap.values(second);
     reverseMap.remove(second);
     while ( f.hasMoreElements() )
     {
        Object o = f.nextElement();
        if ( ! o.equals(first) )
        {
           forwardMap.add(second,o);
        }
     }

  }

  // Just for test purpose

  public static void main(String[] args)
  {
     if ( debug )
     {
        // Create some points
        Object a = "a";
        Object b = "b";
        Object c = "c";
        Object d = "d";
        Object e = "e";
        Object f = "f";
        Object g = "g";
        Object z = "z";
        Object t = "t";
        Object u = "u";
        Object v = "v";
        
        CycleDetector cd1 = new CycleDetector();
        System.out.println("Trying to detect cycle b, c , d, e , b : ");
        try {
           cd1.addDependency(a,c);
           cd1.addDependency(b,c);
           cd1.addDependency(c,d);
           cd1.addDependency(d,e);
           cd1.addDependency(d,f);
           cd1.addDependency(e,b);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }
        
        CycleDetector cd2 = new CycleDetector();
        System.out.println("Trying to detect cycle a, b, a : ");
        try {
           cd2.addDependency(a,b);
           cd2.addDependency(b,a);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }

        CycleDetector cd3 = new CycleDetector();
        System.out.println("Trying to detect cycle no cycle.");
        try {
           cd3.addDependency(a,c);
           cd3.addDependency(b,c);
           cd3.addDependency(c,d);
           cd3.addDependency(d,e);
           cd3.addDependency(d,f);
           System.out.println("... no cycle detected.");
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }

        CycleDetector cd4 = new CycleDetector();
        System.out.println("Trying to detect cycle a, b, c , d, e , a : ");
        try {
           cd4.addDependency(a,b);
           cd4.addDependency(b,c);
           cd4.addDependency(c,d);
           cd4.addDependency(d,e);
           cd4.addDependency(e,a);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }
        
        CycleDetector cd5 = new CycleDetector();
        System.out.println("Trying to detect cycle a, b, c , d, e , a : ");
        try {
           cd5.addDependency(a,b);
           cd5.addDependency(b,c);
           cd5.addDependency(c,d);
           cd5.addDependency(c,z);
           cd5.addDependency(z,t);
           cd5.addDependency(t,u);
           cd5.addDependency(u,e);
           cd5.addDependency(d,e);
           cd5.addDependency(e,a);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }

        CycleDetector cd6 = new CycleDetector();
        System.out.println("Trying to detect cycle b, d, b : ");
        try {
            cd6.addDependency(b,a);
            cd6.addDependency(b,d);
            cd6.addDependency(d,b);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }

        CycleDetector cd7 = new CycleDetector();
        System.out.println("Trying to detect two cycles abca and abcda.");
        try {
            cd7.addDependency(a,b);
            cd7.addDependency(b,c);
            cd7.addDependency(c,a);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }
        try {
            cd7.addDependency(c,d);
            cd7.addDependency(d,a);
        }
        catch ( CycleException ex ) 
        { 
           Enumeration en = ex.getCycle().elements();
           System.out.print("... cycle detected = ");
           while ( en.hasMoreElements() )
           {System.out.print(en.nextElement() + " "); }
           System.out.println("");
        }

     }
  }

  private static final boolean debug = true;

  private HashMap forwardMap         = new HashMap(true);
  private HashMap reverseMap         = new HashMap(true);
  private HashSet markedDependencies = new HashSet(false);

}

// ***********************************************************************
// History Log:
// $Log: CycleDetector.java,v $
// Revision 1.1  1998/02/27 15:30:24  cinard
// CycleDetector made more robust (several cycles can be detected now) and
// moved to parser module.
//
// ***********************************************************************
