// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1996-1997 Sun Microsystems Inc.
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/PropertySet.java,v $
// $Revision: 1.6 $
// $Date: 1998/04/01 16:28:53 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   AST
// NAME:        PropertySet.java
//
// HISTORY:
//      Rob Duncan, Sun Microsystems Inc. - Feb 22, 1997: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers;

import java.lang.reflect.*;

import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This class defines a manager for a list of properties that can be set
 * on an object just as Java Beans properties.
 *
 * <p>A property can be implemented in the object as a field, a couple of
 * setter/getter methods or be stored in a Hashtable.
 *
 * <p>Access to the property makes use of the Java Reflection API.
 *
 * @see java.lang.Class
 * @see java.lang.reflect.Field
 * @see java.lang.reflect.Method
 */
public class PropertySet
{
  /** debug mode indicator */
  private static final boolean DEBUG = false;
  /** value used to reflect a removed property */
  private static DeletedProperty deleted = new DeletedProperty();

  /** hashtable for properties that are neither field or setter/getter */
  private Hashtable table;
  /** initial size of the hashtable for properties */
  private static final int tableSize = 5;
  /** the object to which the properties are attached */
  private Object object;
  /** the class of the object to which the properties are attached */
  private Class clazz;

  /* These are fields so we don't have to reallocate them every
     time in getGetter() and friends. */
  /** preallocated array for getDeclaredMethod for getter */
  private final Class getterParams[] = new Class[0];
  /** preallocated array to invoke the getter */
  private final Object getterArgs[] = new Object[0];
  /** preallocated array for getDeclaredMethod for setter */
  private final Class setterParams[] = new Class[1];
  /** preallocated array to invoke the setter */
  private final Object setterArgs[] = new Object[1];

  /**
   * Basic constructor specifying the object to which the properties are
   * attached.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>initializes the hashtable
   * <li>set the <code>object</code>
   * <li>intializes the <code>clazz</code> to the object's class.
   * </ul>
   *
   * @param o the object to which the properties are attached
   */
  public PropertySet(Object o)
  {
    table = new Hashtable(tableSize);
    object = o;
    clazz = object.getClass();
  }


  /**
   * Gets the getter method for a method property with the specified
   * name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>build method name with "get" + name of the property
   * <li>starting from the class of the object, until last ancestor
   *     class:<ul>
   * <li>try to get a method of that name with no parameter
   * <li>if there is such a method and it is not private, return that
   *     method
   * <li>if there is such a method and it is private, return null
   * <li>else continue with next ancestor
   * </ul>
   * </ul>
   *
   * @param name the name of the method property
   * @return the getter method or null if it doesn't exist or is private
   * @exception PropertyAccessException access to the method is denied
   */
  private Method getGetter(String name) throws PropertyAccessException
  {
    String methodName = "get" + name;
    if (DEBUG) {
      System.err.println("reflecting for method name: " + methodName);
    }
    for (Class c = clazz; c != null; c = c.getSuperclass()) {
      try {
	Method getter = c.getDeclaredMethod(methodName, getterParams);
	if (Modifier.isPrivate(getter.getModifiers())) {
	  return null;
	} else {
	  return getter;
	}
      } catch (NoSuchMethodException e) {
      } catch (SecurityException e) {
	throw new PropertyAccessException(e);
      }
    }
    return null;
  }


  /**
   * Gets the setter method for a method property with the specified
   * name and type.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>build method name with "set" + name of the property
   * <li>set the setter parameter type to <code>returnType</code>
   * <li>starting from the class of the object, until last ancestor
   *     class:<ul>
   * <li>try to get a method of that name with the specified parameter
   * <li>if there is such a method and it is not private, return that
   *     method
   * <li>if there is such a method and it is private, return null
   * <li>else continue with next ancestor
   * </ul>
   * </ul>
   *
   * @param name the name of the method property
   * @param returnType the expected type for the property, that is the
   *        return type of the getter
   * @return the setter method or null if it doesn't exist or is private
   * @exception PropertyAccessException access to the method is denied
   */
  private Method getSetter(String name, Class returnType)
       throws PropertyAccessException
  {
    String methodName = "set" + name;
    if (DEBUG) {
      System.err.println("reflecting for method name: " + methodName);
    }
    setterParams[0] = returnType;
    for (Class c = clazz; c != null; c = c.getSuperclass()) {
      try {
	Method setter = c.getDeclaredMethod(methodName, setterParams);
	if (Modifier.isPrivate(setter.getModifiers())) {
	  return null;
	} else {
	  return setter;
	}
      } catch (NoSuchMethodException e) {
      } catch (SecurityException e) {
	throw new PropertyAccessException(e);
      }
    }
    return null;
  }


  /**
   * Gets the field for a field property with the specified name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>build field name by <em>decapitalizing</em> name of the property
   * <li>starting from the class of the object, until last ancestor
   *     class:<ul>
   * <li>try to get a field of that name
   * <li>if there is such a field and it is not private, return that
   *     field
   * <li>if there is such a field and it is private, return null
   * <li>else continue with next ancestor
   * </ul>
   * </ul>
   *
   * @param name the name of the field property
   * @return the field or null if it doesn't exist or is private.
   * @exception PropertyAccessException access to the field is denied
   */
  private Field getField(String name) throws PropertyAccessException
  {
    String fieldName = java.beans.Introspector.decapitalize(name);
    if (DEBUG) {
      System.err.println("reflecting for field name: " + fieldName);
    }
    for (Class c = clazz; c != null; c = c.getSuperclass()) {
      try {
	Field field = c.getDeclaredField(fieldName);
	if (Modifier.isPrivate(field.getModifiers())) {
	  return null;
	} else {
	  return field;
	}
      } catch (NoSuchFieldException e) {
      } catch (SecurityException e) {
	throw new PropertyAccessException(e);
      }
    }
    return null;
  }

  /**
   * Gets the value of a property specified by a given key.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>check if the property is defined in the hastable and if yes:<ul>
   * <li>get the value from the hashtable
   * <li>if the value has been deleted, throw a
   *     <code>NoSuchPropertyException</code>
   * <li>else return the value
   * </ul><li>if the key is a String:<ul>
   * <li>try to get a getter method
   * <li>if the getter exists, invoke it and return result
   * <li>else try to get a field
   * <li>if the field exists, get and return its value
   * </ul>
   * </ul>
   *
   * <p>Note: a deleted property is equal to <code>deleted</code>.
   *
   * @param key the key specifying the property
   * @return the value of the property
   * @exception NoSuchPropertyException the property has been removed or
   *            does not exist
   * @exception PropertyAccessException access to the field or method
   *            property is denied
   */
  public Object get(Object key) throws PropertyAccessException
  {
    if (table.containsKey(key)) {
      Object value = table.get(key);
      if (value == deleted) {
	throw new NoSuchPropertyException(object, key);
      }
      return value;
    }
    if (key instanceof String) {
      Method getter = getGetter((String)key);
      if (getter != null) {
	try {
	  return getter.invoke(object, getterArgs);
	} catch (InvocationTargetException e) {
	  throw new PropertyAccessException(e);
	} catch (IllegalAccessException e) {
	  throw new PropertyAccessException(e);
	}
      }
      Field field = getField((String)key);
      if (field != null) {
	try {
	  return field.get(object);
	} catch (IllegalAccessException e) {
	  throw new PropertyAccessException(e);
	}
      }
    }
    throw new NoSuchPropertyException(object, key);
  }

  /**
   * Sets the value of a property specified by a given key.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>check if the property is defined in the hastable and if yes:<ul>
   * <li>set the value in the hashtable and return
   * </ul><li>else if the key is a String:<ul>
   * <li>try to get a getter method
   * <li>if the getter exists:<ul>
   * <li>try to get a setter method, invoke it and return result
   * <li>if there is no setter, raise en exception
   * </ul>
   * <li>else try to get a field
   * <li>if the field exists, set its value
   * </ul>
   * <li>else set the value in the hashtable
   * </ul>
   *
   * @param key the key specifying the property
   * @param value the value of the property
   * @exception MissingSetterMethodException the method property has
   *            a getter method but no setter
   * @exception PropertyAccessException access to the field or method
   *            property is denied
   */
   public void set(Object key, Object value)
       throws PropertyAccessException
  {
    if (table.contains(key)) {
      table.put(key, value);
      return;
    }
    if (key instanceof String) {
      Method getter = getGetter((String)key);
      if (getter != null) {
	Method setter = getSetter((String)key, getter.getReturnType());
	if (setter != null) {
	  /* We can always call a setter method, even if the object is
             additive. */
	  setterArgs[0] = value;
	  try {
	    setter.invoke(object, setterArgs);
	    return;
	  } catch (InvocationTargetException e) {
	    throw new PropertyAccessException(e);
	  } catch (IllegalAccessException e) {
	    throw new PropertyAccessException(e);
	  }
	} else {
	  throw new MissingSetterMethodException(object, key);
	}
      } 
      Field field = getField((String)key);
      if (field != null) {
	try {
	  field.set(object, value);
	  return;
	} catch (IllegalAccessException e) {
	  throw new PropertyAccessException(e);
	}
      }
    }
    table.put(key, value);
  }


  /**
   * Gets the indexed value of a property specified by a given key: <b>not
   * yet implemented</b>.
   *
   * @param key the key specifying the property
   * @param index the index of the element in the value
   * @return the indexed value of the property
   * @exception PropertyAccessException not yet implemented
   */
  public Object get(Object key, int index) throws PropertyAccessException
  {
    throw new PropertyAccessException("Sorry, indexed properties aren't implemented yet.");
  }

  /**
   * Sets the indexed value of a property specified by a given key:
   * <b>not yet implemented</b>.
   *
   * @param key the key specifying the property
   * @param index the index of the element in the value
   * @param value the value of the property
   * @exception PropertyAccessException not yet implemented
   */
  public void set(Object key, int index, Object value)
       throws PropertyAccessException
  {
    throw new PropertyAccessException("Sorry, indexed properties aren't implemented yet.");
  }
  

  /**
   * Checks if a property specified by a given key exists.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the value from the hashtable:<ul>
   * <li>if the value has been deleted, return false
   * <li>else if the value is not null return true
   * </ul><li>if the key is a String:<ul>
   * <li>try to get a getter method
   * <li>if the getter exists, return true
   * <li>else try to get a field
   * <li>if the field exists, return true
   * </ul>
   * <li>else return false
   * </ul>
   *
   * <p>Note: a deleted property is equal to <code>deleted</code>.
   *
   * @param key the key specifying the property
   * @return true if the property exists, false otherwise
   */
  public boolean has(Object key) {
    Object value = table.get(key);
    if (value == deleted) {
      return false;
    } else if (value != null) {
      return true;
    }
    if (key instanceof String) {
      try {
	if (getGetter((String)key) != null) {
	  return true;
	}
	if (getField((String)key) != null) {
	  return true;
	}
      } catch (PropertyAccessException e) {
	return false;
      }
    }
    return false;
  }


  /**
   * Removes a property specified by a given key.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if the property doesn't exist throw an exception
   * <li>if the property is in the hashtable, remove it from there
   * <li>if the property still exists (method or field property), set
   *     <code>deleted</code> as its value in the hashtable
   * </ul>
   *
   * @param key the key specifying the property
   * @exception NoSuchPropertyException the property does not exist
   */
  public void remove(Object key) throws NoSuchPropertyException
  {
    if (!has(key)) {
      throw new NoSuchPropertyException(object, key);
    }
    if (table.containsKey(key)) {
      table.remove(key);
    }
    if (has(key)) {
      table.put(key, deleted);
    }
  }


  /**
   * Returns the key specifying a method property.
   *
   * <p>The name of the property is taken from the name of the getter
   *    without the "get" prefix.
   *
   * @param m the getter method for the property
   * @return the name of the property or null if the getter is private
   *         or there is no corresponding setter or the getter name
   *         does not start with "get" or it has any argument or
   *         access to the getter information is denied
   */
  private String keyFromGetterMethod(Method m)
  {
    try {
      if (Modifier.isPrivate(m.getModifiers())) {
	return null;
      }
      String n = m.getName();
      int nargs = m.getParameterTypes().length;
      if (n.startsWith("get") && nargs == 0) {
	Method setter = null;
	try {
	  setter = getSetter(n.substring(3), m.getReturnType());
	}
	catch (PropertyAccessException e) {
	  return null;
	}
	if (setter != null)
	  return n.substring(3);
	else
	  return null;
      } else {
	return null;
      }
    } catch (SecurityException e) {
      return  null;
    }
  }


  /**
   * Returns the key specifying a field property.
   *
   * <p>The name of the property is taken from the name of the field
   *    with first letter in uppercase.
   *
   * @param f the field for the property
   * @return the name of the property or null if the field is private or
   *         access to the field information is denied
   */
  private String keyFromField(Field f) {
    try {
      if (Modifier.isPrivate(f.getModifiers())) {
	return null;
      }
      String s = f.getName();
      if (Character.isLowerCase(s.charAt(0))) {
	s = Character.toUpperCase(s.charAt(0)) + s.substring(1);
      }
      return s;
    } catch (SecurityException e) {
      return null;
    }
  }


  /**
   * Returns the list of keys of the properties as an Enumeration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>build a local hashtable to store keys to avoid duplicates
   * <li>first add hashtable keys of non deleted properties
   * <li>then, in all ancestor classes:<ul>
   * <li>look for methods matching a property
   * <li>look for public fields
   * </ul></ul>
   *
   * <p>Note: a deleted property is equal to <code>deleted</code>.
   *
   * @return the list of keys without duplicates
   * @see #keyFromGetterMethod
   * @see #keyFromField
   */
  public Enumeration keys() {
    Hashtable keyTable = new Hashtable();

    Enumeration e = table.keys();
    while (e.hasMoreElements()) {
      Object k = e.nextElement();
      if (!(table.get(k) instanceof DeletedProperty)) {
	keyTable.put(k, k);
      }
    }

    for (Class c = clazz; c != null; c = c.getSuperclass()) {
      Method[] methods = c.getDeclaredMethods();
      for (int i = 0; i < methods.length; ++i) {
	String k = keyFromGetterMethod(methods[i]);
	if (k != null) {
	  keyTable.put(k, k);
	}
      }
      Field[] fields = c.getDeclaredFields();
      for (int i = 0; i < fields.length; ++i) {
	String k = keyFromField(fields[i]);
	if (k != null) {
	  keyTable.put(k, k);
	}
      }
    }
    return keyTable.keys();
  }
}

class DeletedProperty
{
}
