// Copyright(c) 1997-1998 The Open Group.
// Copyright(c) 1994-1997 Sun Microsystems Inc.
// Copyright(c) 1994-1998 Information-technology Promotion Agency, Japan
//
// Permissions to use, copy, modify and distribute this software are
// governed by the terms and conditions set forth in the file COPYRIGHT,
// located in the ADLT release package.
//
//
// ************************************************************************
//
// $Revision: 1.10 $
// $Date: 1998/07/28 15:25:51 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL for C parser
// NAME:        ADLTypeCheckVisitor.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI . - Apr 8, 1998: Created
// 	
// ************************************************************************

package ORG.opengroup.adl.parsers.adlc;

import java.util.Stack;
import java.util.Enumeration;
import java.util.Vector;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;

import ORG.opengroup.adl.nlengine.*;

/** 
 * This class defines a visitor which can be applied to an AST
 * generated from the ADLJavaParser grammar. It performs several
 * type checks over.
 *
 * <p>The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and performs type
 * checks for nodes where this is relevant.
 *
 * <p>A helper class ADLTypeCheckHelper is used by this visitor for basic 
 * type check actions such as type synthesis.
 *
 * <p>The type checking starts by accepting this visitor (accept method) from
 * the top node of the AST, that is the ASTADL_CompilationUnit.
 *
 * @see ADLTypeCheckVisitor#visit(ASTADL_CompilationUnit,Object)
 * @see ADLTypeCheckHelper
 */

public class ADLTypeCheckVisitor implements  ADLCParserVisitor
{

  // -- Inner class --
  
  /**
   * Stores useful information during the process of finding the
   * type of primary expressions.
   */

  class SearchTypeInfo
  {
    /** 
     * Constructs search type information. 
     * @param type type to stored.
     * @param scopedName name to stored.
     * @param scope scope to be stored.
     */

    SearchTypeInfo
    (LType type,String scopedName,Scope scope)
    {
       _type              = type;
       _scopedName        = scopedName;
       _scope = ( type instanceof Scope ? (Scope)type : scope );
       _astName           = null;
    }

    /**
     * Returns the ASTName (if there is one) coming first in the primary
     * expression. This is useful when information is to be attached to
     * that ASTName while the PrimarySuffixes are processed.
     */

    ASTName getASTName()                { return _astName; }

    /**
     * Saves a reference to the ASTName that may come in first position
     * in the processed primary expression.
     * @param astName the astName to save.
     */

    void    setASTName(ASTName astName) 
    {
       // name can be set only once
       if ( _astName == null ) { _astName = astName; }
    }

    /**
     * Get scoped name.
     */

    String getScopedName()              { return _scopedName; }

    /**
     * Set search type information scoped name. The effect is to
     * update type.
     * @param s scoped name.
     */

    void setScopedName(String s)
    {
       _type = null; // reset type
       _scopedName = s; 
    }
    
    /**
     * Get the type held in search type information object.
     */

    LType getType()
    {
       return _type;
    }

    /**
     * Set search type information type. The effect is to update
     * type, scope and scopedName fields.
     * @param type new type.
     */

    void setType(LType type)
    {
       _type = type;
       _scope = ( type instanceof Scope ? (Scope)type : null );
       _scopedName = "";
    }

    /**
     * Return current primary expression scope.
     */

    Scope getScope()
    {
       if ( _type instanceof Scope ) 
       { 
          return _type.getScope();
       }
       else 
       {
          return _scope;
       }
    }

    /**
     * Set scope. Side effect is to update the type and scopedName field.
     */

    void setScope(Scope scope)
    {
       _scope = scope;
       _type  = ( ( scope instanceof LType ) ? (LType)scope : null );
       _scopedName = null;
    }

    /** Type found while parsing the primary expression. */
    private LType   _type               = null;

    /** Current scoped name found while parsing the primary expression. */
    private String  _scopedName         = null;

    /**
     * Current scope where remaining suffixes are to be searched while
     * processing the primary expression.
     */
    private Scope   _scope              = null;

    /**
     * Reference that should be set to the first ASTName of the Primary
     * expression in case there is one.
     */
    private ASTName _astName    = null;

    /**
      * Flag indicating that the astNameKind property has been set.
      */
    boolean _astNameKindPropertySet = false;

  }

  // -- Public constructors --

  /**
   * Constructs the type check visitor.
   *
   * @param fileName name of the source file from which the AST was 
   *                 constructed. Used for error reporting only.
   */

  public ADLTypeCheckVisitor(String fileName, boolean nld)
  {
     // Reset symbol table scope
     SimpleNode.stm.setCurrentScope(SimpleNode.stm.getUnitScope());
     _scopeStack.push(SimpleNode.stm.getUnitScope());

     _fileName   = fileName;
     _sourceName = fileName.substring(fileName.lastIndexOf('/') + 1,
                                      fileName.lastIndexOf('.'));
     nld_required = nld;
  }


  /**
   * Visit SimpleNode.
   * <p> Just called in case acceptor not unimplemented in subclass.
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(SimpleNode node, Object data)
  {
     System.out.println( node + ": acceptor not unimplemented in subclass?");
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visit ASTTranslationUnit.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> not relevant for adl type check 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTranslationUnit node, Object data)
  {
    return null;
  }

  /**
   * Visit ASTADL_TranslationUnit.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> enter translation unit scope
   * <li> just visit children 
   * <li> exit translation unit scope
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_TranslationUnit node, Object data)
  {
    node.childrenAccept(this,data);
    return new Boolean(_typeCheckOk);
  }

  /**
   * Visit ASTIncludeFileList.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTIncludeFileList node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTIncludeFileDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTIncludeFileDeclaration node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTADL_AnnotatedDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push adl annotated declaration scope on top of scope stack
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_AnnotatedDeclaration node, Object data)
  {
    _scopeStack.push(node.getScope());  // Enter scope
    node.childrenAccept(this,data);
    _scopeStack.pop();                  // Exit scope
    return data;
  }

  /**
   * Visit ASTADL_InlineDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set inline declaration flag
   * <li> push inline scope (the inline itself) on top of scope stack
   * <li> visit children 
   * <li> check inline assertion group is compatible with inline return type
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_InlineDeclaration node, Object data)
  {
     _inADLInlineScope = true;
     LCallable operation = (LCallable)node.getScope();
     _scopeStack.push(operation);        // Enter operation scope
     node.childrenAccept(this,data);
     // Check third children ( ADL_AssertionGroup) type is 
     // compatible with inline return type
     TypedNode assertionGroup = (TypedNode)node.jjtGetChild(2);
     if ( ! operation.getReturnType().assignable(assertionGroup.getType()) )
     {
        errorMessage
        (  "Incompatible type for assertion group in inline. Can't convert "
           + assertionGroup.getTypeName()
           + " to " 
           + operation.getReturnType().getName(),
           assertionGroup
        ); 
     }
     _scopeStack.pop();                  // Exit scope
     _inADLInlineScope = false;
     return data;
  }

  /**
   * Visit ASTADL_BehaviorDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push behavior declaration (ADL method) scope on top of scope stack
   * <li> save ADL method return type for further reference in the children
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorDeclaration node, Object data)
  {
     _visitedADLMethod = (ADLMethod)node.getScope();
     _visitedOpeReturnType = _visitedADLMethod.getReturnType();
     _scopeStack.push(_visitedADLMethod); // Enter scope
     node.childrenAccept(this,data);
     _scopeStack.pop();                   // Exit scope
     _visitedOpeReturnType = null;
     _visitedADLMethod = null;
     return data;
  }

  /**
   * Visit ASTADL_Prolog.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set prolog scope flag
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Prolog node, Object data)
  {
     _inPrologueScope = true;
     // Enter scope if necessary
     if ( node.manageScope() ) { _scopeStack.push(node.getScope());  }
     node.childrenAccept(this,data);
     _inPrologueScope = false;
     return data;
  }

  /**
   * Visit ASTADL_Epilog.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set epilog scope flag
   * <li> push epilog scope on top of scope stack
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Epilog node, Object data)
  {
     _inEpilogueScope = true;
     _scopeStack.push(node.getScope());  // Enter scope
     node.childrenAccept(this,data);
     _scopeStack.pop();                  // Exit scope
     _inEpilogueScope = false;
     return data;
  }

  /**
   * Visit ASTADL_BehaviorSpecification.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorSpecification node, Object data)
  {
    node.childrenAccept(this,data);

    if (nld_required)
      node.setScope((Scope)_scopeStack.peek());

    return data;
  }

  /**
   * Visit ASTADL_BehaviorClassification.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorClassification node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

 /**
  * Visit ASTADL_NormalBehavior.
  * <p> <b> Algorithm : </b> <ul> 
  * <li> just visit children 
  * </ul>
  * @param node node to visit.
  * @param data arbitrary data received from the caller.
  */

  public Object visit(ASTADL_NormalBehavior node, Object data)
  {
    // TBD should check adl expression!
    node.childrenAccept(this,data);
    if (nld_required)
      node.setScope((Scope)_scopeStack.peek());
    return data;
  }

 /**
  * Visit ASTADL_AbnormalBehavior.
  * <p> <b> Algorithm : </b> <ul> 
  * <li> just visit children 
  * </ul>
  * @param node node to visit.
  * @param data arbitrary data received from the caller.
  */

  public Object visit(ASTADL_AbnormalBehavior node, Object data)
  {
    // TBD should check adl expression!
    node.childrenAccept(this,data);
    if (nld_required)
      node.setScope((Scope)_scopeStack.peek());
    return data;
  }

  /**
   * Visit ASTADL_AssertionGroup.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push prolog scope on top of scope stack
   * <li> visit children 
   * <li> get all ADL_Statement (ADL_If or ADL_Assertion) in 
   *      assertion group, in case there is several ADL_Statement 
   *      check their types are compatible with (scalar),
   * <li> set assertion group type to that of enclosed ADL_Statement
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_AssertionGroup node,Object data) 
  {
     if ( node.manageScope() )
     {
        _scopeStack.push(node.getScope());  // Enter scope
     }
     node.childrenAccept(this,data);
     // Find index of first and last ADL_Statement 
     int numChildren = node.jjtGetNumChildren();
     int firstADLStatement = numChildren; // initialize past last index
     int lastADLStatement  = numChildren - 1;
     for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
     {
        if ( node.jjtGetChild(i) instanceof ASTADL_Statement )
        {
           firstADLStatement = i;
           break;
        }
     }
     int numStatement = lastADLStatement - firstADLStatement + 1;
     if ( numStatement > 1 )
     {
        try
        {
           LType synthetizedType = 
                 ADLTypeCheckHelper.checkTypesAreScalar
                 (node,
                  firstADLStatement,
                  lastADLStatement);

           node.setType(synthetizedType); 
        }
        catch ( BadTypeException e )
        {
           String msg = "Assertion group with multiple assertions"
                      + " shall contain only scalar assertion: "
                      +  e.getMessage();
           errorMessage(msg,e.getErrorNode() );
        }
     }
     else if ( numStatement == 1 )
     {
        node.setType
        (((TypedNode)node.jjtGetChild(firstADLStatement)).getType());
     }

     if ( debug )
     {
        System.out.println
        ("debug: type of ADLAssertionGroup =  " + node.getTypeName());
     }

     if ( node.manageScope() )
     {
        _scopeStack.pop();                  // Exit scope
     }
     return data;
  }

  /**
   * Visit ASTADL_Binding.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push binding scope on top of scope stack
   * <li> visit last children (ASTExpression)
   * <li> check binding type and initializer expression are compatible
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Binding node, Object data)
  {
     _scopeStack.push(node.getScope());  // Enter scope
     // Only the last child (ASTExpression) is to be visited
     TypedNode expressionNode = (TypedNode)node.jjtGetChild
                                           (node.jjtGetNumChildren() - 1);
     expressionNode.jjtAccept(this,VISITING_EXPRESSION);
     try
     {
        LType bindingType = ((LExpression)node.getProperty("bindingInit")).getType();
        if (      bindingType == null 
             || ! bindingType.assignable(expressionNode.getType()) 
           )
        {
           errorMessage
           (  "Incompatible type for expression in binding. Can't convert "
              + expressionNode.getTypeName()
              + " to " 
              + bindingType.getName(),
              node
           );

        }
     } 
     catch ( PropertyAccessException e ) { }
     _scopeStack.pop();                  // Exit scope
     return data;
  }

  /**
   * Visit ASTADL_NamedParamList.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_NamedParamList node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTADL_NamedParam.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_NamedParam node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTADL_Statement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children 
   * <li> set type to that of child   
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Statement node, Object data)
  {
     // Visit children
     node.childrenAccept(this,data);

     // Set type to that of child
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     return data;
  }

  /**
   * Visit ASTADL_IfStatement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children 
   * <li> check first child is convertible to bool
   * <li> perform synthesis for 2nd and 3rd child types
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_IfStatement node, Object data)
  {
     boolean ok = true;
     LType type = null;

     // visit children

     node.childrenAccept(this,data);

     // check first child is convertible to bool
     LType firstChildType = ((TypedNode) node.jjtGetChild(0)).getType();
     if ( ! CType.isScalar(firstChildType))
     {
        errorMessage
        ("Illegal use of ADL If statement: can not convert "+
		 firstChildType + " to bool.",(SimpleNode)node.jjtGetChild(0));
        ok = false ;
     }

     // perform synthesis for 2nd and 3rd child types
     if ( ok )
     {
        try
        {
           type = ADLTypeCheckHelper.synthetizeTypes
                  (node,1,node.jjtGetNumChildren()-1);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage("incompatible types for adl if statement : "
                       + e.getMessage(),
                       e.getErrorNode());
        }
     }

     if (nld_required)
      node.setScope((Scope)_scopeStack.peek());

     return data;
  }

  /**
   * Visit ASTADL_Assertion.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children 
   * <li> set type to that of child
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Assertion node, Object data)
  {
     if (nld_required)
      node.setScope((Scope)_scopeStack.peek());

     // visist children
     node.childrenAccept(this,VISITING_EXPRESSION);
    
     // Set type to that of child
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());
    
     return data;
  }

  /**
   * Visit ASTADL_Labels.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Labels node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTADL_Tags.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Tags node, Object data)
  {
    return data;
  }

  /**
    * Visit ASTADL_QuantifiedAssertion.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> push quantified assertion scope on scope stack
    * <li> set type to scalar
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_QuantifiedAssertion node,Object data) 
  {
     _scopeStack.push(node.getScope());  // Enter scope
     node.childrenAccept(this,data);
     node.setType(CBuiltinType.intT);     // Set type to scalar
     _scopeStack.pop();                  // Exit scope
     return data;
  }

  /**
   * Visit ASTADL_Quantifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Quantifier node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTADL_DomainList.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_DomainList node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTADL_Domain.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Domain node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTADL_DomainExpression.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_DomainExpression node, Object data)
  {
    // visit children (indicate children can be expressions)
    node.childrenAccept(this,VISITING_EXPRESSION);

    // check 1th and 2nd children are of scalar type
     TypedNode erroneousNode = null;
     if (! CType.isScalar
           (((TypedNode)node.jjtGetChild(0)).getType()) )
     {
        erroneousNode = (TypedNode)node.jjtGetChild(0); 
     }
     if (! CType.isScalar
           (((TypedNode)node.jjtGetChild(1)).getType()) )
     {     
        erroneousNode = (TypedNode)node.jjtGetChild(2); 
     }
     if ( erroneousNode != null )
     {
        errorMessage
        (  "Incompatible type for " + node.getName()
         + ". Can't convert " + erroneousNode.getTypeName() + " to scalar.",
         erroneousNode
        );
     }
     else
     {
       if ( node.getName().equals("ADL_short_range") ) 
	 node.setType(CBuiltinType.shortT);
       else if ( node.getName().equals("ADL_int_range") ) 
	 node.setType(CBuiltinType.intT);
       else // node.getName().equals("ADL_long_range")
	 node.setType(CBuiltinType.longT);
     }

     return data;
  }

  /**
   * Visits an ASTADL_ImplExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check 0th and 2nd childrens (impl. op skipped) are of scalar type
   * <li>set expression type to scalar
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTADL_ImplExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check 0th and 2nd children (impl. operator skipped) are of scalar type
     TypedNode erroneousNode = null;
     if (! CType.isScalar
           (((TypedNode)node.jjtGetChild(0)).getType()) )
     {
        erroneousNode = (TypedNode)node.jjtGetChild(0); 
     }
     if (! CType.isScalar
           (((TypedNode)node.jjtGetChild(2)).getType()) )
     {     
        erroneousNode = (TypedNode)node.jjtGetChild(2); 
     }
     if ( erroneousNode != null )
     {
        String operator = "";
        try {
           operator = 
                (String)((SimpleNode)node.jjtGetChild(1)).getProperty("op");
        }
        catch (PropertyAccessException e) { }
        errorMessage
        (  "Incompatible type for " + operator
         + ". Can't convert " + erroneousNode.getTypeName() + " to scalar.",
         erroneousNode
        );
     }
     else
     {
        // set expression type to scalar
        node.setType(CBuiltinType.intT);
     }

     if ( debug )
        System.out.println("debug: (ADLImplExpression) - type "
                           + "(" + node.getType() + ") " 
                           + ((TypedNode)node.jjtGetChild(0)).getTypeName() 
                           + " " 
                           + ((TypedNode)node.jjtGetChild(2)).getTypeName()
                          );
     return data;
  }

  /**
   * Visit ASTADL_ImplOp.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_ImplOp node, Object data)
  {
    return data;
  }

  /**
   * Visits an ASTADL_CallStateExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>increase nested call state scope level
   * <li>visit children (indicate children can be expressions) 
   * <li>update expression type according to the child type
   * <li>decrease nested call state scope level
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  
  public Object visit(ASTADL_CallStateExpression node,Object data) 
  {
     // increase nested call state scope level
     _callStateScopeLevel++;

     // visit children (indicate children can be expressions) 
     node.childrenAccept(this,VISITING_EXPRESSION);

     // update expression type according to the child type
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());

     if ( debug )
     {
        System.out.print("debug: (ADL_CallStateExpression) - type "
                           + "(" + node.getTypeName() + ") ");
        System.out.print
        (" " + ((TypedNode)node.jjtGetChild(0)).getTypeName());

        System.out.println("");
     }

     // decrease nested call state scope level
     _callStateScopeLevel--;

     return data;
  }

  /**
   * Visits an ASTNormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set expression type to scalar
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Normal node,Object data) 
  {
    node.setType(CBuiltinType.intT);
    return data;
  }

  /**
   * Visits an ASTAbnormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set expression type to scalar
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTADL_Abnormal node,Object data) 
  {
    node.setType(CBuiltinType.intT);
    return data;
  }

  /**
   * Visits an ASTADL_Unchanged.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>set expression type to scalar
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTADL_Unchanged node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // set expression type to scalar
     node.setType(CBuiltinType.intT);

     return data;
  }

  /**
   * Visit ASTADL_ArgumentList.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_ArgumentList node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTExternalDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>   
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTExternalDeclaration node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTFunctionDefinition.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> enter function scope
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTFunctionDefinition node, Object data)
  {
     Scope functionScope = node.getScope();
     LCallable function = (LCallable)functionScope;
     _visitedOpeReturnType = function.getReturnType();
     _inTestFunction = true;

     _scopeStack.push(functionScope);    // enter scope
     node.childrenAccept(this,data);
     // exit function scope
     while ( functionScope != _scopeStack.pop() ) {}
     _inTestFunction = false;
     _visitedOpeReturnType = null;
    return data;
  }

  /**
   * Visit ASTDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> just visit children. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDeclaration node, Object data)
  {
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visit ASTDeclarationSpecifiers.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDeclarationSpecifiers node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTTypeQualifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTypeQualifier node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTStorageStructSpecifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTStorageStructSpecifier node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTBuiltinTypeSpecifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTBuiltinTypeSpecifier node, Object data)
  {
     return data;
  }

  /**
   * Visit ASTInitDeclaratorList.
   * <p> <b> Algorithm : </b> <ul> <li> just visit children. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTInitDeclaratorList node, Object data)
  {
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visit ASTInitDeclarator.
   * <p> <b> Algorithm : </b> <ul> 
   * <li>check declared type and initializer have same type
   *     note initializers of the form { ... } are not checked
   * </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTInitDeclarator node, Object data)
  {
     if ( debug )
     {
        System.out.println("debug: ADLTypeCheckVisitor(ASTInitDeclarator)");
     }
 
     // set scope to declaration scope
     _scopeStack.push(node.getScope());

     // check declared type and initializer have same type
     // note initializers of the form { ... } are not checked
     if (    node.jjtGetNumChildren() == 2
          && node.jjtGetChild(1) instanceof TypedNode
        )
     {
        if ( debug )
        {
           System.out.println("debug: ADLTypeCheckVisitor(ASTInitDeclarator)"
                              + " ... type checking");
        }

        TypedNode initializer = (TypedNode)(node.jjtGetChild(1));
        initializer.jjtAccept(this,VISITING_EXPRESSION);
        try 
        {
           LType declType = (LType)node.getProperty("declarationType");
           LType initType = initializer.getType();
           if ( ! declType.assignable(initType) )
           {
             errorMessage("bad initializer : "
                          + initializer.getTypeName()
                          + " not compatible with "
                          + ( declType != null ? declType.getName() 
                                               : "<no type>") 
                          , initializer); 
           }
        }
        catch ( PropertyAccessException e ) {}

     }
     if ( debug )
     {
        System.out.println("debug: ADLTypeCheckVisitor(ASTInitDeclarator)"
                          + " - return");
     }
     return data;
  }

  /**
   * Visit ASTInitializer.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTInitializer node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTStructSpecifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTStructSpecifier node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTMemberDeclaratorList.
   * <p> <b> Algorithm : </b> <ul> <li> just visit children. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTMemberDeclaratorList node, Object data)
  {
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visit ASTMemberDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTMemberDeclaration node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTEnumSpecifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTEnumSpecifier node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTEnumeratorList.
   * <p> <b> Algorithm : </b> <ul> <li> just visit children. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTEnumeratorList node, Object data)
  {
     node.childrenAccept(this,data);
     return data;
  }
  
  /**
   * Visit ASTEnumerator.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children, indicate they can be expression
   * <li> check children type is integral
   * </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTEnumerator node, Object data)
  {
     node.childrenAccept(this,data);
     if ( node.jjtGetNumChildren() > 0 )
     {
        if ( ! CType.isIntegral(((TypedNode)node.jjtGetChild(0)).getType()) )
        {
           errorMessage
           (  "enumerator value for `" + node.getName() 
              + "` not an integer constant",
              node
           );
        }
     }
     return data;
  }

  /**
   * Visit ASTPtrOperator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTPtrOperator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTCvQualifierSeq.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTCvQualifierSeq node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTDeclarator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDeclarator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTDirectDeclarator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDirectDeclarator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTDeclaratorSuffixes.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDeclaratorSuffixes node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTFunctionDeclarator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTFunctionDeclarator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTFunctionDirectDeclarator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTFunctionDirectDeclarator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTParameterList.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTParameterList node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTParameterDeclarationList.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTParameterDeclarationList node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTParameterDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTParameterDeclaration node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTTypeName.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTypeName node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTAbstractDeclarator.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTAbstractDeclarator node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTAbstractDeclaratorSuffix.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>  
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTAbstractDeclaratorSuffix node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTStatement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTStatement node, Object data)
  {
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visit ASTLabeledStatement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTLabeledStatement node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTCaseLabelStatement.
   * <p><b>Algorithm:</b><ul>
   * <li>check expression is compatible with int
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTCaseLabelStatement node, Object data)
  {
     // visit first children (indicate it can be an expression)
     TypedNode exprNode = (TypedNode)node.jjtGetChild(0);
     ((SimpleNode)exprNode).jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children 
     for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
     {
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
     }

     // and check expression is compatible with int
     LType   type = exprNode.getType();
     if ( ! CType.isIntegral(type) )
     {
        errorMessage
        (  "Incompatible type for case. Can't convert "
         + exprNode.getTypeName()
         + " to int.", exprNode
        );
     }
     return data;
  }

  /**
   * Visit ASTDefaultLabelStatement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDefaultLabelStatement node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTCompoundStatement.
   * <p><b>Algorithm:</b><ul>
   * <li>get block scope if block was responsible for scope management
   * <li>visit children
   * <li>exit block scope ( sevaral variable declaration scopes may have
   *     been opened and shall be exited - that is popped from the scope
   *     stack)
   * <li>
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTCompoundStatement node, Object data)
  {
     Scope blockScope = null;   // block scope

     if ( node.manageScope() )
     {
        blockScope = node.getScope();  // get block scope
        _scopeStack.push(blockScope);    // enter block scope
     }

     node.childrenAccept(this,data); // visit children

     if ( node.manageScope() )
     {
        // exit block scope ( sevaral variable declaration scopes may have
        // been opened and shall be exited - that is popped from the scope
        // stack)
        Scope popedScope;
        do { popedScope = (Scope)_scopeStack.pop(); }
        while ( popedScope != blockScope );
     }
     
     return data;
  }

  /**
   * Visits an ASTIfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first children (indicate it can be an expression)
   * <li>visit other children
   * <li>check first child is a scalar
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTIfStatement node,Object data) 
  {
     // visit first children (indicate it can be an expression)
     TypedNode firstNode = (TypedNode)node.jjtGetChild(0);
     ((SimpleNode)firstNode).jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children
     for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
     {
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
     }

     // check first child is scalar
     if ( ! CType.isScalar(firstNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for if. Can't convert " 
           + firstNode.getTypeName() +  " to scalar.",
           firstNode
          );
     }

     return data;
  }

  /**
   * Visits an ASTSwitchStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform type check : first expression shall be compatible with int
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTSwitchStatement node,Object data) 
  {
     // visit first children (indicate it can be an expression)
     TypedNode firstNode = (TypedNode)node.jjtGetChild(0);
     ((SimpleNode)firstNode).jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children
     ((SimpleNode)node.jjtGetChild(1)).jjtAccept(this,data);

     // check first child is integer
     if ( ! CType.isIntegral(firstNode.getType()) )
     {
          errorMessage("switch quantity not an integer",firstNode);
     }

     return data;
  }

  /**
   * Visit ASTWhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first children (indicate it can be an expression)
   * <li>visit other children
   * <li>check first child is a scalar
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTWhileStatement node, Object data)
  {
     // visit first children (indicate it can be an expression)
     TypedNode firstNode = (TypedNode)node.jjtGetChild(0);
     ((SimpleNode)firstNode).jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children
     for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
     {
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
     }

     // check first child is scalar
     if ( ! CType.isScalar(firstNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for while. Can't convert " 
           + firstNode.getTypeName() +  " to scalar.",
           firstNode
          );
     }

     return data;
  }

  /**
   * Visit ASTDoWhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child
   * <li>visit last child (indicate it can be an expression)
   * <li>check last child is a scalar
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDoWhileStatement node, Object data)
  {
     // visit first child 
     ((SimpleNode)node.jjtGetChild(0)).jjtAccept(this,data);

     // visit last child (indicate it can be an expression) 
     TypedNode lastNode = (TypedNode)node.jjtGetChild(1);
     lastNode.jjtAccept(this,VISITING_EXPRESSION);
     
     // check last child is scalar
     if ( ! CType.isScalar(lastNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for do ... while. Can't convert " 
           + lastNode.getTypeName() +  " to scalar.",
           lastNode
          );
     }
     return data;
  }

  /**
   * Visit ASTForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>enter "for" statement scope
   * <li>visit children (indicate child can be an expression)
   * <li>exit "for" statement scope
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTForStatement node, Object data)
  {
     _scopeStack.push(node.getScope());  // enter "for" statement scope

     // visit children (except but for ASTDeclaration and last child
     // (statement) indicate child can be an expression)
     SimpleNode child;
     for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
     {
        child = (SimpleNode)node.jjtGetChild(i);
        child.jjtAccept(this,VISITING_EXPRESSION);
     }
     _scopeStack.pop();                  // exit "for" statement scope

     return data;
  }

  /**
   * Visit ASTGotoStatement.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTGotoStatement node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTContinueStatement.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTContinueStatement node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTBreakStatement.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTBreakStatement node, Object data)
  {
    return data;
  }

  /**
   * Visits an ASTReturnStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>check if the statement contains an expression
   * <li>if there's an expression, visit the expression, and save its type
   * <li>if it's a stand alone return, save return type is void
   * <li>check saved type is compatible with return type
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTReturnStatement node,Object data) 
  {
     boolean ok = false;

     // check if the statement contains an expression
     LType   type;
     if ( node.jjtGetNumChildren() > 0 )
     {
        // there's an expression, visit the expression, and save its type 
        node.childrenAccept(this,VISITING_EXPRESSION);
        type = ((TypedNode)(node.jjtGetChild(0))).getType();  
     }
     else
     {
        // it's a stand alone return, save return type is void
        type = CBuiltinType.voidT;
     }

     // check type is compatible with return type
     if (  CBuiltinType.voidT.equals(_visitedOpeReturnType) )
     {
        ok = CBuiltinType.voidT.equals(type);
     }
     else
     {
        ok = _visitedOpeReturnType.assignable(type);
     }

     if ( ! ok )
     {
        errorMessage
        (  "incompatible type for return: can't convert "
         + ( type == null ? "<no type>" : type.getName() )
         + " to " 
         + _visitedOpeReturnType.getName(),
         node
        );
     }

     return data;
  }

  /**
   * Visit ASTExpression.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children stating they can be expressions
   * <li> set type to that of last visited childrer
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTExpression node, Object data)
  {
    node.childrenAccept(this,VISITING_EXPRESSION);
    node.setType
    (((TypedNode)(node.jjtGetChild(node.jjtGetNumChildren() - 1))).getType());
    return data;
  }

  /**
   * Visits an ASTAssignmentExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>get left and right operand
   * <li>get operator from node property "op"
   * <li>perform type check and type synthesis, as follows:
   * <li>for all operators : right hand side operand shall be assignable to
   *     left hand side operand
   * <li>for simple = operator : above condition is enough
   * <li>for operator += and -= : one of the following shall hold,
   *     left operand shall have pointer type and right operand shall have
   *     integral type, or both operand shall have arithmetic types
   * <li>for operators *= and /= : both operand shall have arithmetic types
   * <li>for other operators : both operand shall have integral types
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTAssignmentExpression node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     boolean ok = true;

     // get left and right operand nodes
     TypedNode left  = (TypedNode)node.jjtGetChild(0);
     TypedNode right = (TypedNode)node.jjtGetChild(1);

     // left and right operands type
     LType leftType  = left.getType();
     LType rightType = right.getType();

     // get operator

     String op = "";
     try { op = (String)node.getProperty("operator"); }
     catch (PropertyAccessException e) { }

     // for all operators : check operands are assignable
     if ( ( leftType == null ) || ! leftType.assignable( rightType ) )
     {
        // operands are not assignables
        ok = false;
     }
     else if ( ! op.equals("=") )
     {
        // operands are assignable : perform further operator dependant checks
        // for all operators except simple =
        if ( op.equals("+=") || op.equals("-=") )
        {
           // += and -= : one of the following shall hold:
           // left operand has pointer type and right operand has integral type
           // or both operands have arithmetic types
           ok =  (    CType.isPointer(leftType) 
                   && CType.isIntegral(rightType) ) 
              || (    CType.isArithmetic(leftType) 
                   && CType.isArithmetic(rightType) );
        }
        else if ( op.equals("/=") || op.equals("*=") )
        {
           // /= and *= : beside assignability, operands shall have 
           // arithmetic types
           ok = CType.isArithmetic(leftType) && CType.isArithmetic(rightType);
        }
        else // %= >>= >>= &= ^= |= : operands shall have integral type 
        {
           ok = CType.isIntegral(leftType) && CType.isIntegral(rightType);
        }
     }

     if ( ok )
     {
        node.setType(left.getType());
     }
     else
     {
        errorMessage("wrong types to " + op,node);
     }

     return data;
  }

  /**
   * Visits an ASTConditionalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check first operand has a scalar type 
   * <li>check that one of the following hold for second and third operands:
   * <li>both operands have arithmetic type
   * <li>both operands have compatible structure or union types
   * <li>both operands have void type
   * <li>both operands are pointers to qualified or unqualified versions
   *     of compatible types
   * <li>one operand is a pointer and the other is a null pointer constant
   * <li>one operand is a pointer to an object or incomplete type and the
   *     other is a pointer to a qualified or unqualified version of void
   * <li>and synthetize expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTConditionalExpression node,Object data) 
  {
     boolean ok = true;

     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // type of first operand
     LType type1 = ((TypedNode)node.jjtGetChild(0)).getType();

     // check first operand has scalar type
     if ( ! CType.isArithmetic(type1) )
     {
        errorMessage
        ("wrong type to '?:' : first operand shall have scalar type",
         (SimpleNode)node.jjtGetChild(0)
        );
        ok = false;
     }

     // types of second and third operands
     LType type2 = ((TypedNode)node.jjtGetChild(1)).getType(); 
     LType type3 = ((TypedNode)node.jjtGetChild(2)).getType(); 

     // synthetize expression type using the second and third expressions
     // nodes in conditional expression
     if ( ok )
     {
        if (  ( CType.isArithmetic(type2) && CType.isArithmetic(type3) )
           || ( type2.isVoid() && type3.isVoid() )
           )
        {
           // for arithmetic and void types, type synthesis is enough 
           try
           {
              LType type = ADLTypeCheckHelper.synthetizeTypes(node,1,2);
              node.setType(type);
           }
           catch ( BadTypeException e )
           {
              throw new Error("Error in type synthetizis.");
           }
        }
        else if (  ( type2.isA(SymbolId.STRUCT) && type3.isA(SymbolId.STRUCT) )
                || ( type2.isA(SymbolId.UNION) && type3.isA(SymbolId.UNION) )
                || ( CType.isPointer(type2) && CType.isPointer(type3) )
                )
        {
           // for struct, union and pointers : types may not be compatible
           // report an erro message in case type synthesis fail
           try
           {
              LType type = ADLTypeCheckHelper.synthetizeTypes(node,1,2);
              node.setType(type);
           }
           catch ( BadTypeException e )
           {
              errorMessage
              ("wrong type to '?:' : second and third operands types are not"
              +" compatibles",
               (TypedNode)node.jjtGetChild(1));
           }
        }
        else
        {
           errorMessage
           ("wrong type to '?:' : second and third operands not compatibles",
            (TypedNode)node.jjtGetChild(1));
        }
     }

     if ( debug )
        System.out.println("debug: (ConditionalExpression) - type "
                           + "(" + node.getTypeName() + ") " 
                           + ((TypedNode)node.jjtGetChild(0)).getTypeName() 
                           + " ? "
                           + ((TypedNode)node.jjtGetChild(1)).getTypeName()
                           + " : "
                           + ((TypedNode)node.jjtGetChild(2)).getTypeName()
                          );
    return data;
  }

  /**
   * Visit ASTConstantExpression.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children 
   * <li> set type to that of child
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTConstantExpression node, Object data)
  {
     // visit chidren
     node.childrenAccept(this,data);

     // set type to that of child
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());
    
    return data;
  }

  /**
   * Visit ASTLogicalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions are scalar
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTLogicalOrExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are scalar 
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreScalar(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for ||. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTLogicalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions types are scalar 
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTLogicalAndExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are scalar
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreScalar(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for &&. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTInclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions types are integral
   * <li>set expression type accordingly
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTInclusiveOrExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are integer
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreIntegral(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for |. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTExclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions are integral
   * <li>set expression type accordingly
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTExclusiveOrExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are integer
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreIntegral(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for ^. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions are integral
   * <li>set expression type accordingly
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTAndExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are integral
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreIntegral(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for &. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTEqualityExpression.
    *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>initialize synthetized expression type with type of first child
   *     (it will be updated by applying a left recursion algorithm
   *     on the child expression type)
   * <li>if both operand arithmetics, set synthetized type to int
   * <li>if both types are pointer, check pointed types are compatible
   *     set synthetized type to int
   * <li>set expression type accordingly, or report an error
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTEqualityExpression node, Object data)
  {

     // status
     boolean ok = true;
     
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // initialize synthetized expression type with type of first child
     // (it will be updated by applying a left recursion algorithm
     // on the child expression type)
     LType synthetizedType = ((TypedNode)node.jjtGetChild(0)).getType();

     // next child expression's type
     LType nextType = null;

     // get operators
     String[] operators = null;
     try
     {
        operators = (String[])node.getProperty("operators");
     }
     catch ( PropertyAccessException e ) {}

     for ( int i = 0; i < operators.length; i ++ )
     {
        nextType = ((TypedNode)node.jjtGetChild(i+1)).getType();

        if (    CType.isArithmetic(synthetizedType)  
             && CType.isArithmetic(nextType) ) 
        {
           // both operand arithmetics, set synthetized type to int
           synthetizedType = CBuiltinType.intT;
        }
        else if (    CType.isPointer(synthetizedType) 
                  && CType.isPointer(nextType) )
        {
           // both operands have pointer types check that either :
           // . one operand is a void pointer and the other a pointer to
           //   an object or incomplete type
           // . one operand is a null pointer constant and the other 
           //   a pointer
           // . both operand are pointers to compatible types 
           if ( (  CType.isPointerToObject(nextType) 
                && CType.isVoidPointer(synthetizedType))
              ||
                (  CType.isVoidPointer(nextType) 
                && CType.isPointerToObject(synthetizedType))
              ||
                (  CType.isNullPointerConstant(nextType)
                || CType.isNullPointerConstant(synthetizedType))
              )
           {
              // synthetized type to int
              synthetizedType = CBuiltinType.intT;
           }
           else // check both operands are pointer to compatible types
           {
              
              // check pointed types are compatible
              // TBC - this means type synthesis is possible
              try
              {
                 LType[] types = {synthetizedType,nextType};
                 ADLTypeCheckHelper.synthetizeTypes(types);
                 synthetizedType = CBuiltinType.intT;
              }
              catch ( BadTypeException e ) 
              {
                 errorMessage("invalid operands to binary " + operators[i],
                              (SimpleNode)(node.jjtGetChild(i)));
                 ok = false;
                 break;
              }
           }
        }
        else 
        {
           // error
           errorMessage("invalid operands to binary " + operators[i],
                        (SimpleNode)(node.jjtGetChild(i)));
           ok = false;
           break;
        }
     }

     if ( ok )
     {
        node.setType(synthetizedType);
     }

     return data;

  }

  /**
   * Visit ASTRelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>initialize synthetized expression type with type of first child
   *     (it will be updated by applying a left recursion algorithm
   *     on the child expression type)
   * <li>if both operand arithmetics, set synthetized type to int
   * <li>if both types are pointer, check pointed types are compatible
   *     set synthetized type to int
   * <li>set expression type accordingly, or report an error
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTRelationalExpression node, Object data)
  {
     // status
     boolean ok = true;
     
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // initialize synthetized expression type with type of first child
     // (it will be updated by applying a left recursion algorithm
     // on the child expression type)
     LType synthetizedType = ((TypedNode)node.jjtGetChild(0)).getType();

     // next child expression's type
     LType nextType = null;

     // get operators
     String[] operators = null;
     try
     {
        operators = (String[])node.getProperty("operators");
     }
     catch ( PropertyAccessException e ) {}

     for ( int i = 0; i < operators.length; i ++ )
     {
        nextType = ((TypedNode)node.jjtGetChild(i+1)).getType();
        if (    CType.isArithmetic(synthetizedType) 
             && CType.isArithmetic(nextType) ) 
        {
           // both operand arithmetics, set synthetized type to int
           synthetizedType = CBuiltinType.intT;
        }
        else if (    CType.isPointer(synthetizedType) 
                  && CType.isPointer(nextType) )
        {
           // both types are pointer
           // check pointed types are compatible
           // TBC - this means type synthesis is possible
           try
           {
              LType[] types = {synthetizedType,nextType};
              ADLTypeCheckHelper.synthetizeTypes(types);
              // synthetized type is synthetized to int
              synthetizedType = CBuiltinType.intT;
           }
           catch ( BadTypeException e ) 
           {
              errorMessage("invalid operands to binary " + operators[i],
                           (SimpleNode)(node.jjtGetChild(i)));
              ok = false;
              break;
           }
        }
        else 
        {
           // error
           errorMessage("invalid operands to binary " + operators[i],
                        (SimpleNode)(node.jjtGetChild(i)));
           ok = false;
           break;
        }
     }

     if ( ok )
     {
        node.setType(synthetizedType);
     }

     return data;

  }

  /**
   * Visit ASTShiftExpression.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children  (indicate children can be expressions)
   * <li> check all children expressions have integral types
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTShiftExpression node, Object data)
  {
     node.childrenAccept(this,VISITING_EXPRESSION);

     // Check all children expressions are integral
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreIntegral(node);
        node.setType(CBuiltinType.intT);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for shift operator. " + e.getMessage(),
         e.getErrorNode());
     }

     if ( debug )
     {
        System.out.print("debug: (ShiftExpression) - type "
                        + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }

     return data;
  }

  /**
   * Visit ASTAdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>initialize synthetized expression type with type of first child
   *     (it will be updated by applying a left recursion algorithm
   *      on the child expression type)
   * <li>get next operator and next children type
   * <li>if both synthetized type are arithmetic update synthetized type
   *     (perform type synthesis on the two types)
   * <li>only for "+" operator : if one type is pointer and the other
   *     integral : synthetized type is set to pointer
   * <li>only for "-" operator : if synthetized type is pointer and
   *     next type is integral : synthetized type is set to pointer
   * <li>only for "-" operator : if both types are pointer to
   *     compatible types : synthtized type is set to arithmetic
   * <li>report an error or set expression type to synthetized type
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTAdditiveExpression node, Object data)
  {
     // status
     boolean ok = true;
     
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // initialize synthetized expression type with type of first child
     // (it will be updated by applying a left recursion algorithm
     // on the child expression type)
     LType synthetizedType = ((TypedNode)node.jjtGetChild(0)).getType();

     // next child expression's type
     LType nextType = null;

     // get additive operators
     
     String[] operators = null;
     try
     {
        operators = (String[])node.getProperty("operators");
     }
     catch ( PropertyAccessException e ) {}

     for ( int i = 0; i < operators.length; i ++ )
     {
        nextType = ((TypedNode)(node.jjtGetChild(i+1))).getType();
        if (   CType.isArithmetic(synthetizedType) 
            && CType.isArithmetic(nextType) ) 
        {
           // both operand arithmetics, perform type synthesis
           try
           {
              LType[] types = {synthetizedType,nextType};
              synthetizedType = ADLTypeCheckHelper.synthetizeTypes(types);
           }
           catch ( BadTypeException e ) 
           {
              throw new Error
              ("Error in type synthesis : " + e.getMessage());
           }
        }
        else if ( operators[i].equals("+") )
        {
           // checks for "+" operator

           if (    CType.isIntegral(synthetizedType) 
                && CType.isPointer(nextType) )
           {
              // synthetized type is set to pointer
              synthetizedType = nextType;
           }
           else if ( ! (    CType.isPointer(synthetizedType) 
                         && CType.isIntegral(nextType) ) )
           {
              // error
              errorMessage("invalid operands to binary " + operators[i],
                           (SimpleNode)(node.jjtGetChild(i)));
              ok = false;
              break;
           }
           // at this point synthetized type has not been touched
           // or has been updated
        }
        else // operator -
        {
           if (    CType.isPointer(synthetizedType) 
                && CType.isPointer(nextType) )
           {
              // both types are pointer
              // check pointed types are compatible
              // TBC - this means type synthesis is possible
              try
              {
                 LType[] types = {synthetizedType,nextType};
                 ADLTypeCheckHelper.synthetizeTypes(types);
                 // synthetized type is synthetized to long
                 synthetizedType = CBuiltinType.longT;
              }
              catch ( BadTypeException e ) 
              {
                 errorMessage("invalid operands to binary " + operators[i],
                              (SimpleNode)(node.jjtGetChild(i)));
                 ok = false;
                 break;
              }
           }
           else if ( ! (    CType.isPointer(synthetizedType) 
                         && CType.isIntegral(nextType) ) )
           {
              // error
              errorMessage("invalid operands to binary " + operators[i],
                           (SimpleNode)(node.jjtGetChild(i)));
              ok = false;
              break;
           }
           // at this point synthetized type has not been touched
           // or has been updated
        }
     }

     if ( ok )
     {
        node.setType(synthetizedType);
     }

     return data;
  }

  /**
   * Visit ASTMultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>initialize synthetized expression type with type of first child
   *     (it will be updated by applying a left recursion algorithm
   *     on the child expression type)
   * <li>if operator is "%", check synthetizedType and nextType are integral
   * <li>if operator is "*" or "/", check synthetizedType and nextType are
   *     arithmetics
   * <li>update synthetized type accordingly and continue checking with next
   *     child type
   * <li>eventually, set expression type accordingly
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTMultiplicativeExpression node, Object data)
  {
     // status
     boolean ok = true;
     
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // initialize synthetized expression type with type of first child
     // (it will be updated by applying a left recursion algorithm
     // on the child expression type)
     LType synthetizedType = ((TypedNode)node.jjtGetChild(0)).getType();

     // next child expression's type
     LType nextType = null;

     // get operators
     String[] operators = null;
     try
     {
        operators = (String[])node.getProperty("operators");
     }
     catch ( PropertyAccessException e ) {}

     for ( int i = 0; i < operators.length; i ++ )
     {
        nextType = ((TypedNode)node.jjtGetChild(i+1)).getType();

        if ( operators[i].equals("%") )
        {
           // check operands have integral types
        
           ok =     CType.isIntegral(synthetizedType) 
                 && CType.isIntegral(nextType);
        }
        else // * or / operators
        {
           // check operands have arithmetic types
           ok =    CType.isArithmetic(synthetizedType) 
                && CType.isArithmetic(nextType);
        }

        if ( ok )
        {
           // perform type synthesis
           try
           {
              LType[] types = { synthetizedType, nextType};
              synthetizedType = ADLTypeCheckHelper.synthetizeTypes(types);
           }
           catch ( BadTypeException e ) 
           {
              throw new Error("Error in type synthesis.");
           }
        }
        else 
        {
           // error
           errorMessage("invalid operands to binary " + operators[i],
                        (SimpleNode)(node.jjtGetChild(i)));
           break;
        }
     }

     if ( ok )
     {
        node.setType(synthetizedType);
     }

     return data;

  }

  /**
   * Visit ASTCastExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child holding expression to cast
   * <li>get cast type
   * <li>unless cast type is void type, check cast type and operand type
   *     are scalar
   * <li>set expression type to cast type.
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTCastExpression node, Object data)
  {
     // visit child holding expression to cast

     TypedNode origTypeNode = (TypedNode)node.jjtGetChild(1);

     origTypeNode.jjtAccept(this,VISITING_EXPRESSION);

     // get cast type

     LType  castType     = null;
     try
     {
        castType = (LType)node.getProperty("type");
     } 
     catch ( PropertyAccessException e ) { }

     String castTypeName =   (castType != null) 
                           ? castType.getName() 
                           : "<no type>" ;
    
     // get type of expression to cast
     LType origType = origTypeNode.getType();

     // unless cast type is void type, check cast type and operand type
     // are scalar 
     if (  ! CBuiltinType.voidT.equals(castType) )
     {
        if ( ! ( CType.isScalar(castType) && CType.isScalar(origType) ) )
        {
           errorMessage("wrong type to typecast",node);
        }
        else
        {
           node.setType(castType);
        }
     }
     else
     {
        node.setType(castType);
     }

     if ( debug )
     {
        System.out.print("debug: (CastExpression) - type "
                          + "(" + castTypeName + ") ");
        System.out.print
        (" " + origTypeNode.getTypeName());

        System.out.println("");
     }
     
     return data;
  }

  /**
   * Visit ASTPreIncrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check children types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTPreIncrementExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // get expression type
     LType type = ((TypedNode)node.jjtGetChild(0)).getType();

     // check expression type is scalar
     if ( CType.isScalar(type) )
     {
        node.setType(type);
     }
     else
     {
        errorMessage
        ("wrong type argument to increment: " 
         + ((TypedNode)node.jjtGetChild(0)).getTypeName() + " not a scalar",
         node);
     }

     if ( debug )
     {
        System.out.println("debug: (PreIncrementExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visit ASTPreDecrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check children types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTPreDecrementExpression node, Object data)
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // get expression type
     LType type = ((TypedNode)node.jjtGetChild(0)).getType();

     // check expression type is scalar
     if ( CType.isScalar(type) )
     {
        node.setType(type);
     }
     else
     {
        errorMessage
        ("wrong type argument to decrement: " 
         + ((TypedNode)node.jjtGetChild(0)).getTypeName() + " not a scalar",
         node);
     }

     if ( debug )
     {
        System.out.println("debug: (PreDecrementExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visit ASTUnaryOperatorExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child (indicate child can be expressions)
   * <li>for * operator : expression shall have a pointer type
   * <li>for & operator : expression shall have function designator type
   *                      or lvalue that designates an object that is not 
   *                      a bit field and is not declared with the register
   *                      storage class specifier
   * <li> for ! operator : expression shall have scalar type
   * <li> for + and - operators : expression shall have arithmetic type
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTUnaryOperatorExpression node, Object data)
  {
     LType exprType = null; 

     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     String op = "";
     try { op = (String)node.getProperty("operator"); }
     catch (PropertyAccessException e) { }

     LType childType = ((TypedNode)node.jjtGetChild(0)).getType();

     if ( op.equals("*") )
     {
        // * operator : expression shall have a pointer type
        if ( childType instanceof LPointerToFunction )
        {
           exprType = ((LPointerToFunction) childType).getPointedToType();
        }
        else if ( childType  instanceof LPointer ) 
        {
           exprType = ((LPointer) childType).getPointedToType() ;
        }
        else 
        {
           errorMessage("invalid type argument of unary *",node);
        }
     }
     else if( op.equals("&")) 
     {
        // & operator : expression shall have function designator type
        //              or lvalue that designates an object that is not 
        //              a bit field and is not declared with the register
        //              storage class specifier
        if ( childType instanceof CFunctionType ) 
        {
           // function designator
           exprType = SimpleNode.sbf.newPointerToFunction
                      (((CFunctionType)childType).getFunctionTypeComparator(),
                       0x00);
        }
        else
        {
           // lvalue
           exprType = SimpleNode.sbf.newPointer( childType, 0x00 ) ;
        }
     }
     else if ( op.equals("!") )
     {
        // ! operator : expression shall have scalar type
        if ( CType.isScalar(childType) )
        {
           exprType = childType ;
        }
        else 
        {
           errorMessage("wrong type argument to unary exclamation mark",node);
        }
     }
     else if ( op.equals("~") )
     {
        // ~ operator : expression shall have integral type
        if ( CType.isIntegral(childType) )
        {
           exprType = childType ;
        }
        else 
        {
           errorMessage("wrong type argument to unary ~",node);
        }
     }
     else
     {
        // + and - operators : expression shall have arithmetic type
        if ( CType.isArithmetic(childType) )
        {
           exprType = childType ;
        }
        else 
        {
           errorMessage("wrong type argument to unary " + op,node);
        }
     }

     // now save recovered type as a node property
     node.setType( exprType ) ;

     if ( debug )
     {
        System.out.println("debug: (UnaryOperatorExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visit SizeOfExpression.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children if child is an expression
   * <li> get type to which sizeof operator is applied
   * <li> check type to which sizeof is applied is not a function type
   *      nor a bitfield
   * <li> set expression type to int
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTSizeOfExpression node, Object data)
  {
     // type to which sizeof is applied
     LType type = null;

     // visit children if necessary and get type to which 
     // sizeof operator is applied

     if ( ! ( node.jjtGetChild(0) instanceof ASTTypeName ) )
     {
        node.childrenAccept(this,VISITING_EXPRESSION);
        type = ((TypedNode)node.jjtGetChild(0)).getType();
     }
     else
     {
        String typeName = node.getName();
        type = SimpleNode.stm.getType(typeName,null);
     }

     // check type to which sizeof is applied is not a function type
     // nor a bitfield
     if (    ! ( type instanceof CFunctionType )
          && ! ( CType.isBitfield(type) )
        )
     {
        errorMessage("wrong type argument to sizeof operator",node);
     }

     // set sizeof expression type to int
     node.setType(CBuiltinType.intT);

     if ( debug )
     {
        System.out.println("debug: (SizeOfExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visits an ASTPostfixExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children applying a special processing : postfix expression 
   *     is constructed with a PrimaryExpression
   *     and zero or more PrimarySuffix. A search type information object
   *     is passed in between childs. Each child being in charge
   *     of updating the search type information object.
   * <li>process resulting search type information as detailed below :
   * <li>if search type information  contains a type set primary expression 
   *     type to that type
   * <li>if search type information contains no type, get expression
   *     corresponding to that information
   * <li>or if the expression is not found, search for a class
   * <li>update primary expression type accordingly
   * </ul>
   *
   * Note during this processing, the "isClass" property is set.
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTPostfixExpression node,Object data) 
  {
     // Primary expression is constructed with a PrimaryPrefix
     // and zero or more PrimarySuffix. A search type information object
     // is passed in between childs. Each child being in charge
     // of updating the search type information object.
     SearchTypeInfo searchTypeInfo = new SearchTypeInfo
                                         (  null,
                                            "",
                                            (Scope)_scopeStack.peek()
                                         );
     int numChild = node.jjtGetNumChildren();
     for (int i = 0; i < numChild ; i++)
     { 
        searchTypeInfo = (SearchTypeInfo) 
                         node.jjtGetChild(i).jjtAccept(this,searchTypeInfo);
     }

     // Process resulting search type information :
     // (1) if search type information  contains a type set primary expression 
     // type to that type
     // (2) if search type information contains no type, get expression
     //     corresponding to that information
     // (3) update primary expression type accordingly
       
     if ( searchTypeInfo.getType() != null )
     {
        // If search type information  contains a type set primary expression 
        // type to that type
        node.setType(searchTypeInfo.getType());
     }
     else
     {
        // If search type information contains no type, get expression
        // corresponding to that information and update type accordingly
        LExpression lExpr = SimpleNode.stm.getExpression
                            (searchTypeInfo.getScopedName(),
                             searchTypeInfo.getScope());
        if ( lExpr != null )
        {
           checkSymbolIsAllowed(lExpr,node);
           node.setType(lExpr.getType());
           decorateNode(searchTypeInfo.getASTName(),lExpr);
        }
        else
        {
           errorMessage
           ("Undefined expression " + searchTypeInfo.getScopedName(),node);
        }
     }

     if ( debug )
       System.out.println("debug: ADLTypeCheckVisitor(ASTPrimaryExpression) - "
                          + " found expression type = " 
                          + (   node.getType() != null 
                              ? node.getTypeName() : null ) );
     return data;
  }

  /**
   * Visit ASTPostDeIncrement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li>get expression
   * <li>check expression is convertible to numeric
   * <li>update search type information
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTPostDeIncrement node, Object data)
  {
     // check child expression is convertible to numeric
     SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
     LExpression expression = 
                 SimpleNode.stm.getExpression(searchTypeInfo.getScopedName(),
                                              searchTypeInfo.getScope()
                                             );
     if ( expression == null )
     {
        errorMessage
        ("Undefined expression " + searchTypeInfo.getScopedName() ,node);
     }
     else
     {
        checkSymbolIsAllowed(expression,node);
        LType type = expression.getType();
        if ( ! CType.isScalar(type) )
        {
           errorMessage("wrong type argument to increment/decrement",node);
        }
        else
        {
           decorateNode(searchTypeInfo.getASTName(),expression);
           searchTypeInfo.setType(type);
        }
     }
     return searchTypeInfo;
  }

  /**
   * Visit ASTArraySuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child expression
   * <li>get type of expression in bracket
   * <li>check expression inside brackets is of integer arithmetic type
   * <li>check expression before brackets is of array or pointer type
   * <li>update received search type information object
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTArraySuffix node, Object data)
  {
     // visit child expressions
     node.childrenAccept(this,VISITING_EXPRESSION);

     // get type of expression in bracket
     TypedNode exprInBracket = (TypedNode)node.jjtGetChild(0);

     if ( exprInBracket.getType() != null )
     {
        // check expression inside brackets is of integer type
        if ( ! CType.isIntegral(exprInBracket.getType()) )
        {
           errorMessage("array subscript is not an integer",node);
        }

        // check expression before brackets is of array type
        SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
        LExpression lExpression = 
                    SimpleNode.stm.getExpression
                    (searchTypeInfo.getScopedName(),searchTypeInfo.getScope());
        if ( lExpression != null )
        {
           checkSymbolIsAllowed(lExpression,node);
           LType lType = lExpression.getType();
           if ( lType.isA(SymbolId.ARRAY) )
           {
              decorateNode(searchTypeInfo.getASTName(),lExpression);
              searchTypeInfo.setType(((LArray)lType).getElementType());
           }
           else if ( lType.isA(SymbolId.POINTER) )
           {
              decorateNode(searchTypeInfo.getASTName(),lExpression);
              searchTypeInfo.setType(((LPointer)lType).getPointedToType());
           }
           else
           {
              errorMessage("subscripted value is neither array nor pointer",
                           node);
           }
        }
        else
        {
           errorMessage
           (searchTypeInfo.getScopedName() + " undeclared" ,node);
        }
     }
     return data;
  }

  /**
   * Visit ASTArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children. If ASTArguments is part of a PrimaryExpression 
   *     (data instanceof SearchTypeInfo) then indicates children may be
   *      expressions
   * <li>if ASTArguments is part of a PrimaryExpression, visited arguments
   *     together with preceding scoped names are a call to an operation
   *     (method, constructor...). Processing is as follows :
   * <li>retrieve useful information from received SearchTypeInfo object
   * <li>constructs an operation comparator from the arguments in
   *     the visited expression
   * <li>if operation is a constructor, retrieve the corresponding
   *     class (useful to set the search scope for the constructor)
   * <li>search for the operation
   * <li>update search type info object accordingly (type set to
   *     constructor's class for a constructor, or to operation
   *     return type for other operations
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTArgumentList node, Object data)
  {
     // visit children. If ASTArguments is part of a PrimaryExpression 
     // (data instanceof SearchTypeInfo) then indicate children may be 
     // expressions
     if ( data instanceof SearchTypeInfo )
     {
        node.childrenAccept(this,VISITING_EXPRESSION);
     }
     else
     {
        node.childrenAccept(this,data);
     }

     // if ASTArguments is part of a PrimaryExpression, visited arguments
     // together with preceding scoped names are a call to an operation
     // (method, constructor...)
     // Processing is as follows :
     // (1) retrieve useful information from received SearchTypeInfo object
     // (2) constructs an operation comparator from the arguments in
     //     the visited expression
     // (4) serch the corresponding function
     // (5) update search type info object accordingly (type set to
     //     constructor's class for a constructor, or to operation
     //     return type for other operations

     if ( data instanceof SearchTypeInfo )
     {
        SearchTypeInfo searchTypeInfo  = (SearchTypeInfo)data ;
        String scopedName              = searchTypeInfo.getScopedName();
        Scope  scope                   = searchTypeInfo.getScope();
        LCallableAssignComparator comp = new LCallableAssignComparator();

        // in case an argument list is present - get argument types
        // to construct operation comparator
        if ( node.jjtGetNumChildren() == 1 )
        { 
           Node argumentList = null;

           if ( node.jjtGetChild(0) instanceof ASTExpression )
           {
              // several arguments under ASTExpression child node
              argumentList = node.jjtGetChild(0);
           }
           else
           {
              // a single argument under node
              argumentList = node;
           }
           
           for (int i = 0; i < argumentList.jjtGetNumChildren() ; i++)
           {
              Node argExpression = argumentList.jjtGetChild(i);
              if ( argExpression instanceof TypedNode )
              {
                 LType lType = ((TypedNode)argExpression).getType();
                 if ( debug )
                    System.out.println
                    ("debug: ADLTypeCheckVisitor(ASTArgumentList) - "
                     + " adding param type = " 
                     + ( lType != null ? lType.getName() : null )
                    );
                 comp.addParameterType(lType);
              }
              else
              {
                 throw new Error
                       ("Child in Arguments is not a subclass of TypedNode");
              }
           }
        }
        if ( debug )
          System.out.println("debug: ADLTypeCheckVisitor(ASTArguments) - "
                             + " search operation = " + scopedName);
        
        // search for the operation
        // TBC - update this when direct lookup of operation in
        //       ADLClass will be setup
        LCallable ope = SimpleNode.stm.getMethod(scopedName,comp,null,scope);

        if ( ope != null )
        {
           checkSymbolIsAllowed(ope,node);
           decorateNode(searchTypeInfo.getASTName(),ope);
           // update search type info object accordingly (type set to
           // constructor's class for a constructor, or to operation
           // return type for other operations
           searchTypeInfo.setType(ope.getReturnType());
           // TBD - check if nld unparser really needs this
	   if (nld_required) 
	   {
	     try {node.setProperty("NLDScope", ope); }
	     catch(PropertyAccessException e) {}
	   }
        }
        else
        {
           errorMessage
           ("function " + scopedName 
            + " undeclared or wrong number of arguments in function call",
            node);
        }
        if ( debug )
          System.out.println("debug: ADLTypeCheckVisitor(ASTArgumentList) - "
                          + " found operation = " + ope);

	// if within a TDD_ADLExpression, try to locate the ADL module
	// which annotates this function
	Vector acoModules = SimpleNode.stm.getUsedADLClasses();
	for  (Enumeration e = acoModules.elements(); e.hasMoreElements();) {
	  ADLClass acoModule = (ADLClass)e.nextElement();
	  if (acoModule.getSymbol(scopedName, comp, null) != null) {
	    try {
	      searchTypeInfo.getASTName()
		.setProperty("acoName", acoModule.getName());
	    }
	    catch (PropertyAccessException exc) { }
	    break;
	  }
	}

        return searchTypeInfo ;
     }
     else
     {
        // Visit is finished - just return data
        return data;
     }
  }

  /**
   * Visit ASTDotAccessSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get SearchTypeInfo object
   * <li>update scoped name in SearchTypeInfo object
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTDotAccessSuffix node, Object data)
  {
     // update scoped name
     SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
     String newName = searchTypeInfo.getScopedName();
     if ( newName == null )
     {
        newName = "";
     }
     else if ( newName.length() > 0 )
     {
        newName += "." ;
     }
     newName += node.getName();
     searchTypeInfo.setScopedName(newName);
     return data;
  }

  /**
   * Visit ASTRefAccessSuffix.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> get SearchTypeInfo object
   * <li> check expression before ref access is of pointer or array type
   * <li> update search type info accordingly (type + name)
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTRefAccessSuffix node, Object data)
  {
     // check expression before ref access is of pointer or array type
     SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
     LExpression lExpression = 
                 SimpleNode.stm.getExpression
                 (searchTypeInfo.getScopedName(),searchTypeInfo.getScope());
     if ( lExpression != null )
     {
        checkSymbolIsAllowed(lExpression,node);
        LType lType = lExpression.getType();
        if ( lType != null && lType.isA(SymbolId.ARRAY) )
        {
           decorateNode(searchTypeInfo.getASTName(),lExpression);
           searchTypeInfo.setType(((LArray)lType).getElementType());
           // and update scoped name
           searchTypeInfo.setScopedName(node.getName());
        }
        else if ( lType != null && lType.isA(SymbolId.POINTER) )
        {
           decorateNode(searchTypeInfo.getASTName(),lExpression);
           searchTypeInfo.setType(((LPointer)lType).getPointedToType());
           // and update scoped name
           searchTypeInfo.setScopedName(node.getName());
        }
        else
        {
           errorMessage("invalid type argument of ->",node);
        }
     }
     else
     {
        errorMessage
        (searchTypeInfo.getScopedName() + " undeclared" ,node);
     }
     return data;
  }

  /**
   * Visit ASTName.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>processing depends on the name being part of a PostfixExpression
   *     or being a stand alone expression or dataset expression
   * <li>if name is part of a PostfixExpression, just update the name
   *     in the received search type information object
   * <li>if name is stand alone and is an expression, find the expression
   *     and set node's type accordingly
   * <li>if name is stand alone and is a dataset expression, find the 
   *     expression and set node's type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data is updated during the visit,
   *             or an Integer indicating VISITING_DATASET_EXPR or
   *             VISITING_EXPRESSION, otherwise data is unused.
   * @return received data object, that has been updated if it is a 
   *         SearchTypeInfo object.
   *
   */

  public Object visit(ASTName node, Object data)
  {
      // processing depends on the name being part of a PrimeryExpression
     // or being a stand alone expression or dataset expression
     if ( data instanceof SearchTypeInfo )
     {
        // name is part of a PrefixExpression just update the name in the 
        // received search type information object
        SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
        searchTypeInfo.setScopedName(node.getName());
        // store AST name in SearchTypeInfo for later decoration
        searchTypeInfo.setASTName(node);
        return data;
     }
     else if ( data instanceof Integer && data.equals(VISITING_EXPRESSION) )
     {
        // stand alone expression name, just find it and set expression type
        // accordingly
        LExpression lExpression = SimpleNode.stm.getExpression
                                  (node.getName(),(Scope)_scopeStack.peek());
        if ( lExpression != null ) 
        {
           checkSymbolIsAllowed(lExpression,node);
           node.setType(lExpression.getType());
           decorateNode(node,lExpression);
        }
        else 
        { 
           errorMessage
           ("undeclared expression " + node.getName() ,node);
        }
        return null;
     }
     else if ( data instanceof Integer && data.equals(VISITING_DATASET_EXPR) )
     {
        // dataset expression, just find it and set expression type
        // accordingly
        LExpression lExpression = SimpleNode.stm.getExpression
                                  (node.getName(),(Scope)_scopeStack.peek());
        // directly check symbol is allowed (no need to call for 
        // checkSymbolIsAllowed)
        if ( lExpression == null )
        {
           errorMessage
           ("undeclared dataset " + node.getName() + ". " ,node);
        }
        else
        {
           decorateNode(node,lExpression);
           node.setType(lExpression.getType());
        }
        return null;
     }

     // LM (07/17/1998): I think we never reach this code
     if ( nld_required && node.getScope() == null )
     {
       node.setScope((Scope)_scopeStack.peek());
     }
     
     return data;
  }

  /**
   * Visit ASTString.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set type to array of char 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTString node, Object data)
  {
    node.setType(SimpleNode.sbf.newArray(CBuiltinType.charT));
    return data;
  }

  /**
   * Visit ASTADL_Return.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get return type of the tested method
   * <li>check return expression is allowed here
   * <li>update expression type according to return type
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Return node, Object data)
  {
     // get return type of the tested method
     LExpression e = SimpleNode.stm.getExpression
                     ("return",(Scope)_scopeStack.peek());

     if ( e == null || e.getType() == null )
     {
        throw new Error("Cannot find return type of the tested method.");
     }

     // check return expression is allowed here
     checkSymbolIsAllowed(e,node);

     // update expression type according to return type
     node.setType(e.getType());
    
     // process search type info if any
     if ( data instanceof SearchTypeInfo )
     {
        // expression is part of a PrimaryExpression,
        // just add useful information (type of the return expression)
        // to the search type information
        SearchTypeInfo searchTypeInfo = (SearchTypeInfo)data;
        searchTypeInfo.setType(node.getType());
     }

     if ( debug )
     {
        System.out.println("debug: (Return) - type "
                           + "(" + node.getTypeName() + ") ");
     }
     
     return data;
  }

  /**
   * Visit ASTParentheizedExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child expression
   * <li>set type to that of child
   * <li>if expression is part of a PrimaryExpression,
   *     just add useful information to the search type information object
   *     received : that is the type of the child
   * <li>if it is a stand alone expression is alone, just find and set its type
   * </ul>
   *
   * @param node node to visit.
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data sis updated during the visit,
   *             otherwise data is unused.
   */

  public Object visit(ASTParentheizedExpression node, Object data)
  {
     // visit child expression
     node.childrenAccept(this,VISITING_EXPRESSION);

     // type set to type of expression in parenthesis.
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());

     // in case data is instance of search type info
     // update search type information
     if ( data instanceof SearchTypeInfo )
     {
        SearchTypeInfo sti = (SearchTypeInfo)data;
        sti.setType(node.getType());
     }

     return data;
 }

  /**
   * Visit ASTUnaryPlusMinusConstant.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> type set to constant (child 0) type
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTUnaryPlusMinusConstant node, Object data)
  {
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     return data;
  }

  /**
   * Visit ASTConstant.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> already typed by parser
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTConstant node, Object data)
  {
    return data;
  }

  /**
   * Visit ASTTDD_AnnotatedDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push tdd annotated declaration scope on top of scope stack
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_AnnotatedDeclaration node, Object data)
  {
    _scopeStack.push(node.getScope());  // Enter scope
    node.childrenAccept(this,data);
    _scopeStack.pop();                  // Exit scope
    return data;
  }

  /**
   * Visit ASTTDD_ImportDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_ImportDeclaration node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTTDD_UseDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_UseDeclaration node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTTDD_Declaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_Declaration node, Object data)
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTTDD_FieldDeclaration.
   * 
   * <p><b>Algorithm:</b><ul>
   * <li>retrieve field types property
   * <li>visit each initializer and check its type is assignable
   *     to the declared field
   * </ul>
   * 
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_FieldDeclaration node, Object data)
  {
     LType[] fieldTypes = null;

     // retrieve field types property
     try 
     { 
       LExpression[] fields = (LExpression[])node.getProperty("fields");
       fieldTypes = new LType[fields.length];
       for ( int i = 0; i < fields.length; i++ ) 
	 fieldTypes[i] = fields[i].getType();
     }
     catch ( PropertyAccessException e ) {}

     // visit initializers with the field types
     for ( int i = 2, j = 0; i < node.jjtGetNumChildren(); i = i + 2, j++ )
     {
        TypedNode initNode = (TypedNode)node.jjtGetChild(i);
        initNode.jjtAccept(this,VISITING_EXPRESSION);
        if ( ! fieldTypes[j].assignable(initNode.getType()) )
        {
           errorMessage("incompatible types in initialization",
                       initNode);
        }
     }
     return data;
  }

  /**
   * Visit ASTTDD_DatasetDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get dataset type and name form node properties
   * <li>indicate a dataset is being defined
   * <li>visit dataset initializer
   * <li>mark end of dataset definition
   * <li>check for assignability of initializer to defined dataset
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetDeclaration node, Object data)
  {
     // get dataset type and name form node properties
     LType datasetType = null;
     String datasetName = null;
     try
     {
        datasetType = (LType)node.getProperty("datasetType");
        datasetName = (String)node.getProperty("datasetName");
     }
     catch ( PropertyAccessException e ) { }

     // indicate a dataset is being defined
     _currentDataset = node;
     _currentDatasetName = datasetName;

     // visit dataset initializer
     TypedNode initNode = (TypedNode)node.jjtGetChild(1);
     initNode.jjtAccept(this,VISITING_DATASET_EXPR);

     // mark end of dataset definition
     _currentDataset = null;
     _currentDatasetName = null;

     // check for assignability of initializer to defined dataset
     if ( initNode.getType() == null )
     {
        errorMessage("Type check error in initializer for dataset "
                     + datasetName + ".",
		     (SimpleNode)node.jjtGetChild(1));
     }
     else if ( ! datasetType.assignable(initNode.getType()) )
     { 
        errorMessage
        ("Incompatible type for dataset initializer: Can't convert "
         + initNode.getTypeName()
         + " to " 
         + datasetType.getName() + "."
         ,initNode);
     }
     return data;
  }

  /**
   * Visit ASTTDD_FactoryDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>enter factory scope
   * <li>visit factory body
   * <li>if necessary visit factory relinquish clause
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_FactoryDefinition node, Object data)
  {
     // enter factory scope
     Scope factoryScope = node.getScope();
     _scopeStack.push(factoryScope);

     // set visited factory return type
     _visitedOpeReturnType = ((TDDFactory)factoryScope).getReturnType();

     // visit factory body
     node.jjtGetChild(2).jjtAccept(this,data);

     // exit factory scope
     _scopeStack.pop(); 

     // if necessary visit factory relinquish clause
     if ( node.jjtGetNumChildren() == 5 )
     {
        try
        {
           Scope relinquishScope = (Scope)node.getProperty("relinquishScope");
           _scopeStack.push(relinquishScope);  // enter relinquish clause scope
           ((SimpleNode)node.jjtGetChild(4)).jjtAccept(this,data);
           _scopeStack.pop();                  // exit relinquish clause scope
        }
        catch (PropertyAccessException e ) {}
     }

     // reset visited ope return type
     _visitedOpeReturnType = null;
     return data;
  }

  /**
   * Visit ASTTDD_TestDirective.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>indicate test directive visit
   * <li>set test directive scope
   * <li>visit children
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_TestDirective node, Object data)
  {
     _inTestDirective = true;
     _scopeStack.push(node.getScope());  // enter scope 
     node.childrenAccept(this,data);
     _scopeStack.pop();                  // exit scope
     _inTestDirective = false;
     return data;
  }

  /**
   * Visit ASTTDD_DatasetDomain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>type check is performed only if the node's got an initializer :
   * <li>get dataset type
   * <li>visit initializer
   * <li>get initializer type
   * <li>check initializer can be assigned to the dataset
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetDomain node, Object data)
  {
     // perform type check only if the node's got an initializer
     if ( node.jjtGetNumChildren() > 1 )
     {
        // there's an initializer, get it's type
        LType datasetType = null;
        try
        {
           datasetType = (LType)node.getProperty("datasetType");
        }
        catch ( PropertyAccessException e ) { }

        // visit initializer
        TypedNode initNode = (TypedNode)node.jjtGetChild(1);
        initNode.jjtAccept(this,VISITING_EXPRESSION);
        
        // check initializer can be assigned to the dataset
        if ( initNode.getType() == null )
        {
           errorMessage(  "undeclared dataset " 
                        + ((SimpleNode)node.jjtGetChild(0)).getName() ,
                        initNode  
                       );
        }
        else if ( ! datasetType.assignable(initNode.getType()) )
        { 
           errorMessage
           ("incompatible types in dataset initialisation: can't convert "
            + initNode.getTypeName()
            + " to " 
            + datasetType.getName() + "."
            ,initNode);
        }
     }
     else 
     {
        node.jjtGetChild(0).jjtAccept(this,VISITING_EXPRESSION);
     }
     return data;
  }

  /**
   * Visit ASTTDD_DatasetLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform children types synthesis
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetLiteral node, Object data)
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // perform children types synthesis and set expression type accordingly
     try
     {
        LType type = ADLTypeCheckHelper.synthetizeTypes(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage("incompatible types in dataset literal. "
                     + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visit ASTTDD_DatasetMember.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children (indicating they can be expresions)
   * <li>in case there's a range, check boundaries are integer arithmetics
   * <li>set member type to that obtained from child type synthesis
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetMember node, Object data)
  {
     // just visit children (indicating they can be expresions)
     node.childrenAccept(this,VISITING_EXPRESSION);
    
     // in case there's a range, check boundaries are integer arithmetics
     if ( node.jjtGetNumChildren() > 1 ) 
     {
        try
        { 
           LType type = ADLTypeCheckHelper.checkTypesAreIntegral(node);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage("only integral types allowed in range: " 
                        + e.getMessage(), e.getErrorNode());
        }
     }
     else
     {
        node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     }
     if ( debug )
     {
        System.out.print("debug: (DatasetMember) - type "
                          + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }

  /**
   * Visit ASTTDD_DatasetExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform children types synthesis
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetExpression node, Object data)
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // perform children types synthesis
     // and set expression type accordingly
     if ( node.jjtGetNumChildren() > 1 )
     {
        // several children
        try
        { 
           LType type = ADLTypeCheckHelper.synthetizeTypes(node);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage("Incompatible type for dataset concatenation. " 
                        +  e.getMessage(), e.getErrorNode());
        }
     }
     else
     {
        // one child
        node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     }

     if ( debug )
     {
        System.out.print("debug: (DatasetExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }

  /**
   * Visit ASTTDD_DatasetConcatenationExpr.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child (indicate it can be an expression)
   * <li>set expression type according to child type
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data)
  {
     node.childrenAccept(this,VISITING_EXPRESSION);
     node.setType( ((TypedNode)node.jjtGetChild(0)).getType() );
     if ( debug )
     {
        System.out.print("debug: (DatasetConcatenationExpr) - type "
                         + "(" + node.getTypeName() + ") ");
        System.out.println("");
     }

     return data;
  }

  /**
   * Visit ASTTDD_DatasetSingleton.
  *
  * <p><b>Algorithm:</b><ul>
  * <li>visit chilren (except child 0 which is a name)
  * <li>set expression type according to first child type
  * <li>if a dataset is currently being defined and this dataset singleton
  *     is a name, check for circular dependency into dataset definition,
  *     that is indicate new dependency (using the CycleDetector) and check
  *     it does not detects a cycle among dataset dependencies
  * </ul>
  *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_DatasetSingleton node, Object data)
  {
     // visit child (indicate a dataset expression is visited)
     node.childrenAccept(this,VISITING_DATASET_EXPR);
     
     // set expression type according to child type
     TypedNode child = (TypedNode)node.jjtGetChild(0);
     node.setType( child.getType() );

     // if a dataset is currently being defined and this dataset singleton
     // is a name, check for circular dependency into dataset definition:
     if (    (_currentDataset != null)
          && (child instanceof ASTName )
        )
     {
        // indicate new dependency (using the CycleDetector) and check
        // it does not detects a cycle among dataset dependencies
        try 
        {
           String childName = child.getName();
           String childTDDClassName = (String)child.getProperty
                                      ("enclosingConstructName");
           // TBD - modify this for multi tdd file parsing
           _cycleDetector.addDependency
                          (_sourceName + "." + _currentDatasetName,
                           childTDDClassName + "." + childName);
//            _cycleDetector.addDependency
//                           (_currentDatasetName,childName);
        }
        catch (PropertyAccessException e) {}
        catch (CycleException e)
        {
           Enumeration cycle = e.getCycle().elements();
           String s = "";
           while ( cycle.hasMoreElements() )
           { s += cycle.nextElement() + " "; }
           errorMessage(  "cycle detected for dataset " + _currentDatasetName
                          + ": " + s,
                          _currentDataset
                       );
        }
     }

     if ( debug )
     {
        System.out.print("debug: (DatasetConcatenationSingleton) - type "
                         + "(" + node.getTypeName() + ") ");
        System.out.println("");
     }

     return data;
  }

  /**
   * Visit ASTTDD_FactoryCall.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit chilren
   * <li>get factory name 
   * <li>get factory arguments : child 1 and following
   * <li>get the factory symbol and set node type to the factory's returned 
   *     type
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_FactoryCall node, Object data)
  {
    // visit chilren (except child 0 which is a name)
    for ( int i = 0; i < node.jjtGetNumChildren(); i++)
    {
       ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
    }

    // get factory name 
    String factoryName = ((SimpleNode)node).getName();
    LCallableAssignComparator comp = new LCallableAssignComparator();

    // get factory arguments : child 1 and following
    for ( int i = 0; i < node.jjtGetNumChildren(); i++ ) 
    {
       TypedNode pNode = (TypedNode)node.jjtGetChild(i);
       comp.addParameterType(pNode.getType());
    }

    // get the factory symbol and set node type to the factory's returned type
    TDDFactory factory = SimpleNode.stm.getTDDFactory
                         (factoryName,comp,(Scope)_scopeStack.peek());
    if ( factory == null )
    {
        errorMessage("undeclared factory " + factoryName,node);
    }
    else
    {
       // set node type to return type of returned factory
       node.setType(factory.getReturnType());
       // save ASTName kind
       decorateNode(node,factory);
    }

    return data;
  }

  /**
   * Visit ASTTDD_ADLExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the search type information object
   * <li>visit children (indicate children can be expressions)
   * <li>check first child kind is reference
   * <li>update search type information
   * <li>update node type
   * </ul>
   *
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTTDD_ADLExpression node, Object data)
  {
     // get the search type information object.
     SearchTypeInfo searchTypeInfo = null;
     if ( data instanceof SearchTypeInfo )
     {
        searchTypeInfo = (SearchTypeInfo)data ;
     }

     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     TypedNode childExpr = (TypedNode)node.jjtGetChild(0);
     
     // update search type information
     if ( searchTypeInfo != null )
     {
        searchTypeInfo.setType(childExpr.getType());
     }
        
     // update node type
     node.setType(childExpr.getType());

     return data;
  }

  // -- non visit public methods --

  /** State nld processing is required. */
  public void setNLDrequired(NLEngine nle) 
  {
    nld_required = true;
    nlengine = nle;
  }

  // -- private methods -

  /**
   * Display an error message.
   * @param msg the message to display.
   */

  private static void errorMessage(String msg)
  {
     errorMessage(msg,0);
  }

  // -- Private methods --

  /**
   * Check an expression can be used at the place it is found.
   * @param expression   the expression for which the check is to be performed.
   * @param relatedNode  node to use to get line number in case an 
   *                     error is to be reported.
   * @return true if the expression can be used, false otherwise.
   */

  private boolean checkSymbolIsAllowed
                  (Symbol symbol, SimpleNode relatedNode)
  {
     if ( symbol instanceof ADLBinding )
     {
        if (    ( _callStateScopeLevel > 0 )
             ||   _inPrologueScope 
             || ( _inEpilogueScope && ! _inADLMethodScope )
           )
        {
           errorMessage("A binding cannot be used in the scope of a call-state"
                        + " expression, nor in a prologue, nor in a global "
                        + " epilogue.", relatedNode);
           return false;
        }
     }
     else if ( symbol instanceof ADLInline )
     {
        if (    _inADLInlineScope
             || _inPrologueScope 
             || _inEpilogueScope
           )
        {
           errorMessage("An inline cannot be used in the scope of an other "
                        + " inline, nor in a prologue, nor in an epilogue.",
                       relatedNode);
           return false;
        }
     }
     else if ( symbol instanceof TDDDataset )
     {
        // Not only test directive is to be tested here because for the other
        // places where dataset expression are allowed the test is straight
        // forward since only ASTDatasetExpression are allowed.
        if ( ! _inTestDirective )
        {
           errorMessage
           (  "Dataset expression " + symbol.getName() + " cannot be"
            + " used here. Dataset expression are allowed only in dataset "
            + " definition, in factory invocation or in test directive.",
            relatedNode
           );
           return false;
        }
     }
     else if ( symbol instanceof LFunction )
     {
        // Test functions can only be used in test directives
        if (    ((LFunction)symbol).isTestFunction() 
             && ! ( _inTestDirective || _inTestFunction )
           )
        {
           errorMessage
           (  "Test function " + symbol.getName() + " cannot be"
            + " used outside a test directives or another test function.",
            relatedNode
           );
           return false;
        }
     }
     return true;
  }

  /**
   * Decorate a node with usefull information.
   * This information comprise:<ul> 
   * <li> the enclosing structure (TDDClass, ADLClass...) of the expression,
   *      property name is "enclosingConstructName"
   * <li> the class of the expression corresponding to the name
   *      associated to the node. It enables
   *      to know wether an encountered name is that of a variable, an 
   *      ADLInline, a function... The property name is "astNameKind".
   * </ul>
   * @param node AST node holding the expression name
   * @param symbol symbol corresponding to the expression
   */

  private void decorateNode(SimpleNode node,Symbol symbol)
  {
     // For TDD dataset and factories save symbol's enclosing structure name.

     if (   ( symbol instanceof TDDFactory ) 
	 || ( symbol instanceof TDDDataset )
         || (    ( symbol instanceof LAttribute ) 
	      && ( symbol.getEnclosingConstruct() instanceof TDDClass ))
        )
     {
        try
        {
           TDDClass tddClass = (TDDClass) symbol.getEnclosingConstruct();
           node.setProperty("enclosingConstructName",tddClass.getName());
        }
        catch(PropertyAccessException e) { }
     }

     try
     {
        Class symbolClass = symbol.getClass();

       // the following prevent setting the property twice
       if ( node != null)
       {
         boolean astNameKindPropertySet = false;
         try 
         {
           String unused = (String)node.getProperty("astNameKind");
           astNameKindPropertySet = true;
         }
         catch(PropertyAccessException e) { }
         if ( ! astNameKindPropertySet )
         {
           // hook for range methods
           if (     node.getName().equals("ADL_short_range")
                 || node.getName().equals("ADL_int_range")
                 || node.getName().equals("ADL_long_range")
              )
           {
              // range method - property is objectACO
              node.setProperty
              ("astNameKind","objectACO");
           }
           else
           {
              // normal behaviour
              // store last suffix of expression class
              String s = symbolClass.toString();
              node.setProperty
              ("astNameKind",s.substring(s.lastIndexOf('.')+1));
           }
         }
       }
     }
     catch ( PropertyAccessException e ) {}

  }

  /**
   * Display an error message.
   * @param msg the message to display.
   * @param erroneousNode node to use to get additional information 
   *                      (e.g. line number)
   */

  private static void errorMessage(String msg,SimpleNode erroneousNode)
  {
     int line = ( erroneousNode != null ) ? erroneousNode.getLineNumber() : 0;
     errorMessage(msg,line);
  }

  /**
   * Display an error message.
   * @param msg the message to display.
   * @param line number of the line where the error occured.
   */
  
  private static void errorMessage(String msg,int line)
  {
     System.out.println(_fileName + ":" + line  + ": " + msg + "\n");
     if ( _typeCheckOk) { _typeCheckOk = false; }
  }

  /**
   * Display a debug message.
   * @param msg the message to display.
   * @param node node to use to get additional information (e.g. line number)
   */

  private static void debugMessage(String msg,SimpleNode node)
  {
     int line = ( node != null ) ? node.getLineNumber() : 0;
     System.out.println("debug:" + line  + ": " + msg + "\n");
  }

  // -- Private fields --

  /** NLD to generate */
  private boolean nld_required;

  /** Type check status. */
  private static boolean _typeCheckOk = true;

  /** Name of the source file from which the AST was constructed. */
  private static String _fileName = null;

  /** Name of the source file from which the AST was constructed,
      without path nor extension. */
  private static String _sourceName = null;

  /** Constant indicating that an expression is being visited. */
  private static final Integer VISITING_EXPRESSION   = new Integer(0);

  /** Constant indicating that a dataset expression is being visited. */
  private static final Integer VISITING_DATASET_EXPR = new Integer(1);

  /** Adl superclass of the visited ADL class. */
  private ADLClass _adlSuperClass = null;

  /**  Return type of the operation (ADLMethod, Function, TDD Factory)
   *   currently being visited.
   */
  private LType _visitedOpeReturnType = null;

  /** Currently visited tested function. */
  private ADLMethod _visitedADLMethod = null;

  /** Stack containing the scopes where symbols shall be searched. */
  private Stack _scopeStack = new Stack();

  /** Level of nesting inside call states scopes */
  static private int _callStateScopeLevel = 0;

  /** Flag indicating if we're in a prologue scope */
  static private boolean _inPrologueScope = false;

  /** Flag indicating if we're in an epilogue scope */
  static private boolean _inEpilogueScope = false;

  /** Flag indicating if we're in an ADL function scope. */
  static private boolean _inADLMethodScope = false;

  /** Flag indicating if we're in an ADL inline scope. */
  static private boolean _inADLInlineScope = false;

  /** Flag indicating if we're in a TDD test directive. */
  static private boolean _inTestDirective = false;

  /** Flag indicating if we're in a TDD test function. */
  static private boolean _inTestFunction = false;

  /** Cycle detector used to detect cycles in dataset expressions. */
  private static CycleDetector _cycleDetector = new CycleDetector();

  /** Dataset currently being defined. */
  private SimpleNode _currentDataset = null;

  /** Name of the dataset currently being defined. */
  private String _currentDatasetName = null;

  /** Handle on the NL engine associated with the compiler */
  private NLEngine nlengine;  

  /** Debug flag. */
  private static boolean debug = false;
}

// ***********************************************************************
// History Log:
// $Log: ADLTypeCheckVisitor.java,v $
// Revision 1.10  1998/07/28 15:25:51  mallet
// Replaced ASTIdExpression with ASTName.
// Merged visit of ASTIdExpression and ASTName methods.
// Added case of PostfixExpression within a TDD_ADLExpression in visit
// method for ArgumentList: determine appropriate ACO and set property
// acoName on Name child of PostfixExpression.
// Improved error message in case of type check error for a dataset
// initializer: made it similar to ADL/Java.
//
// Revision 1.9  1998/06/29 12:13:34  bareau
// Suppressed unused prodcution SimpleTypeSpecifier.
// Added set/get value methods in SimpleNode.
//
// Revision 1.8  1998/06/23 13:24:17  cinard
// Changed the grammar to avoid conflicts in between ADL domains and bit
// fields declarations.
//
// Revision 1.7  1998/06/17 16:09:42  cinard
// Added visit methods for ASTADL_NormalBehavior and ASTADL_AbnormalBehavior.
//
// Revision 1.6  1998/06/16 12:35:28  cinard
// Added support for nld  processing.
//
// Revision 1.5  1998/06/10 12:04:54  cinard
// Fixed bug in type check : tdd field initializer are now visited (declarators
// were visited instead).
// Fixed bug in type check : property "enclosingConstructName" is now set
// only for factories, datasets and tdd fields. It was also set for all class
// attributes!
//
// Revision 1.4  1998/05/27 10:02:39  cinard
// Added support for use clause.
//
// Revision 1.3  1998/05/11 12:00:53  cinard
// Updated scope management for nodes with file scope.
// Perform type check for enumerator list.
//
// Revision 1.2  1998/04/28 15:43:22  cinard
// Removed visit method for "else" statememt AST node ("else" clause is now
// part of the AST node for the "if" statement).
//
// Revision 1.1  1998/04/28 13:32:52  cinard
// ADL for C parser has been updated, and type check added. Implementation
// is closer to ADL for Java parser (most processing is in parser actions).
// Code portability has also been implemented.
//
// ***********************************************************************
