// Copyright(c) 1997-1998 The Open Group.
// Copyright(c) 1994-1997 Sun Microsystems Inc.
// Copyright(c) 1994-1998 Information-technology Promotion Agency, Japan
//
// Permissions to use, copy, modify and distribute this software are
// governed by the terms and conditions set forth in the file COPYRIGHT,
// located in the ADLT release package.
//
//
// ************************************************************************
//
// $Revision: 1.1 $
// $Date: 1998/04/28 13:32:53 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL for C parser
// NAME:        DeclarationHelper.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI . - Apr 8, 1998: Created
// 	
// ************************************************************************

package ORG.opengroup.adl.parsers.adlc;

import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;

/** 
 * Helps in parsing C declarators.
 * This helper class is useful to simplify parsing of complex C declaration.
 * <P>For example when parsing "int ( * foo [20] ) [10]", "*foo [20]" is parsed
 * before knowing the base type will be int[10]. The declarator helper helps
 * in manipulating incomplete types that are encountered while parsing
 * C declarations.
 * <P>The components of a declarator helper are :<ul>
 * <li>a base type : from which the expression type will be derived,
 * <li>an updated type : that will be updated as the declaration is parsed
 *     and will be the expression type
 * <li>a partially defined type : the ultimate type that lack a base
 *     type in the above updated type. Ultimately, this partially defined
 *     type will be updated with the base type held in the declarator helper.
 * <li>an expression : currently being declared.
 * </ul>
 * <P>For example when parsing an expression such as "int * foo []", 
 * the base type is int, the updated type will be array of pointer
 * and the partially defined type is pointer. The ultimate step in
 * defining the expression's type is to "connect" partially defined pointer
 * to int!
 */

public class DeclaratorHelper 
{

  /** Disabled constructor. */
  private DeclaratorHelper() { _expression = null; }

  /** 
   * Constructs a declaration helper for supplied expression.
   * @param expression expressio to declare.
   * @param baseType initial type for the expression that is likely to be 
   *                 updated as the declarator is parsed.
   */

  public DeclaratorHelper(LExpression expression, LType baseType)
  {
     if ( debug )
     {
        System.out.println("debug: DeclaratorHelper"
                           + "<" + count + ">" 
                           + "(" + expression 
                           + ", " + baseType + ")");
     }

     _expression = expression;
     _baseType = baseType; 
     if ( expression != null )
     {
        _name = expression.getName();
     }
  }

  /** Expression being declared. */
  public LExpression getExpression()
  {
     return _expression;
  }

  /**
   * Update this declarator helper with a nested declarator helper.
   * @param nestetDeclarationHelper nested declarator helper.
   *        Nested declarator helper are used when nested declarator
   *        are parsed (e.g. in the declaration int ( * foo [10] ) [10]).
   */

  public void update(DeclaratorHelper nestetDeclarationHelper)
  {
     if ( nestetDeclarationHelper._expression != null )
     {
        throw new Error
                  ("Update not allowed from top level declaration helper");
     }

     // Get partially defined type.

     if ( _partialyDefinedType == null )
     {
        _partialyDefinedType = nestetDeclarationHelper._partialyDefinedType;
     }
     
     // Update type.

     if ( _updatedType != null )
     {
        nestetDeclarationHelper.setBaseType(_updatedType);
        nestetDeclarationHelper.updatePartialyDefinedType();
     }

     updateType(nestetDeclarationHelper.getType());
     
     // Update name.
     setName(nestetDeclarationHelper.getName());

     // Try to finish the work in case this is a top level Declaration Helper
     finishDefinition();
  }

  /**
   * Finish definition of the expression by completing partially defined
   * type, and updating expression held by the declarator helper.
   */

  public void finishDefinition()
  {
     // update expression type

     updatePartialyDefinedType();
     if ( _expression != null )
     {
        _expression.setName(_name);
        _expression.setType(_updatedType);
     }
  }

  /** Get the base type for the declaration helper. */
 
  public LType getBaseType() 
  {
     return _baseType;
  }
  
  /** Set the base type for the declaration helper. */

  public void  setBaseType(LType baseType)
  {
     if ( debug )
     {
        System.out.println
        ("debug: DeclaratorHelper<" + count + ">.setBaseType(" 
        + baseType + ")");
     }
     _baseType = baseType; 
  }

  /** Get the updated type for the declaration helper. */

  public LType getType() 
  {
     return _updatedType;
  }
  
  /** Update the type of the declarator helper. */

  public void  updateType(LType type) 
  {
     if ( debug ) 
     {
        System.out.println
        ("debug: DeclaratorHelper" 
         + "<" + count + ">" + ".updateType(" + type + ")");
     }

     _updatedType = type;

     if ( _partialyDefinedType == null )
     {
        _partialyDefinedType = type;
     }
  }

  /** 
   * Update parially defined type, that is set base type of the ultimate
   * partially defined type held in the declaration helper.
   * <P>E.g. if partially defined type is pointer and base type
   *    is int, final type will be pointer to pointer to int.
   */
  
  public void updatePartialyDefinedType()
  {
     if ( _baseType == null ) { return; }

     // finish partial type definition

     if ( _partialyDefinedType instanceof LPointer ) 
     {
        ((LPointer)_partialyDefinedType).setPointedToType(_baseType);
     }
     else if ( _partialyDefinedType instanceof LReference )
     {
        ((LReference)_partialyDefinedType).setReferenced(_baseType);
     }
     else if ( _partialyDefinedType instanceof LArray )
     {
        ((LArray)_partialyDefinedType).setElementType(_baseType);
     }
     else if ( _partialyDefinedType instanceof CFunctionType )
     {
        ((CFunctionType)_partialyDefinedType).setFunctionTypeReturnType
                                              (_baseType);
     }
     else 
     { 
        _updatedType = _baseType; 
     }
  }

  /** Get name held in the declaration helper. */

  public String getName()
  { 
     return _name;
  }

  /** Set name held in the declaraion helper. */
  
  public void  setName(String name) 
  {
     _name = name;
  }

  
  /** Declaration helper base type. */
  private LType  _baseType            = null;

  /** Declaration helper final updated type. */
  private LType  _updatedType         = null;

  /** Name of the expression being declared. */
  private String _name                = null;

  /** Partially defined type. This type is the ultimate type
   * obtained by deriving the final type being declared and for 
   * which base type is not defined (null).
   * <P>E.g. in array of pointer, the partially defined type
   *    is pointer.
   */
  private LType  _partialyDefinedType = null;

  /**
   * Expression being declared.
   */

  private final LExpression _expression;

  /**
   * Debug info.
   */

  private static long number_of_instances = 0;
  private long count = number_of_instances++;
  private final static boolean debug = false;
}


// ***********************************************************************
// History Log:
// $Log: DeclaratorHelper.java,v $
// Revision 1.1  1998/04/28 13:32:53  cinard
// ADL for C parser has been updated, and type check added. Implementation
// is closer to ADL for Java parser (most processing is in parser actions).
// Code portability has also been implemented.
//
// ***********************************************************************
