// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   adlcppparser
// NAME:        ADLCPPTypeCheckVisitor.java
//
// HISTORY:
//      Wassel BERRAYANA, TOG -October,97: Created.
//     
// This file implements the Type check visitor. Extends ADLCPPGenericVisitor.
// Called when parsing done AST nodes built.
// Overrides super methods to extend defauilt behavior.
// 
// ***********************************************************************

package ORG.opengroup.adl.parsers.adlcpp;

import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import ORG.opengroup.adl.nlengine.*;
import ORG.opengroup.adl.parsers.*;
import java.util.Vector ;
import java.util.Enumeration ;



/** 
 * This class defines a visitor which can be applied to an AST
 * generated from the ADLCPPParser grammar. It performs several
 * type checks over.
 *
 * The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and performs type
 * checks for nodes where this is relevant.
 *
 * <p>All visits are not implemented here. Some of them have their default 
 * behavior in ADLCPPGenericVisitor.
 *
 * The type checking starts by accepting this visitor (accept method) from
 * the top node of the AST.
 *
 */

public class ADLCPPTypeCheckVisitor extends ADLCPPGenericVisitor {


  /**
   * Constructs the type check visitor.
   * 
   * @param fileName name of the source file from which the AST was 
   *                 constructed. Used for error reporting only.
   */

  public ADLCPPTypeCheckVisitor ( String filename, boolean nld ) {
    this.typecheckOK = true ;
    // reset symbol table scope
    SimpleNode.stm.setCurrentScope(SimpleNode.stm.getUnitScope());
    this.fileName = filename ;
    nld_required = nld;
  }

  /** parsed file name */
  private static String fileName = null;

  /** booolean telling if type  check is Ok */
  private static boolean typecheckOK = true ;
  /** Tells if Id expression is a static field
   * Updated while visiting IdExpr and used while visiting UnaryExpr.
   */
  private static boolean idIsNotStaticField;
  /** Tells the parent class of the current Id expression.
   * Relevant only when <code>idIsNotStaticField</code> is true.
   */
  private static LConstructedType idExprParentClass ;
  /** Boolean to tell if in scope of a test directive or not */
  private static boolean inTestDirective = false ;

  public static void errorMessage (SimpleNode erroneousNode, String msg )
    throws ParseException {

    typecheckOK = false ;
    try {
      erroneousNode.errorMessage(msg);
    } catch ( ParseException e) {
      System.out.println(e.getMessage() ) ;

      // propagate exception so that type check ends
      // this could be improved by making the type checker robust on
      // null type: defining type as an attribute of Typed interface
      // instead of a property and taking care of each use of a type.
      throw e;
    }
 
  }



 /**
   * Return the synthesis type of some node children.
   * <p> Should be called only when related node children  ( indexes between 
   * begin and end) does exist and are Typed.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Set initial result type as the first child type. 
   * <li> loop on childen whose numbers are in the specified range setting
   * at each type result type as the synthesis type of current result type and
   * child type.
   * <li> return result type. 
   * </ul>
   * @param node the node whose type of children are required.
   * @param begin index of first child. It is up to caller to make sure  node
   * has typed child with that index.
   * @param end index of last child. It is up to caller to make sure node has
   * typed child with that index.
   * @result result type of the specified range node children. Is Null  when 
   * children are not compatible
   */
  private LType getChildrenTypesSynthesis(SimpleNode node ,int begin 
					  ,int end ) throws ParseException {

    LType childType ;
    LType saved;
    LType resultType = ((Typed) node.jjtGetChild(begin)).getType();
    if (debug)
      System.out.println("debug TypeCheckVisitor:in getChildrenTypesSynthesis"+
			 "1st child type is "+ resultType) ;
    // loop on children
    for ( int i = begin +1 ; i < end +1 ; i++ )
    {
      childType = (( Typed) node.jjtGetChild(i)).getType() ;
      saved = resultType ;
      // update result type
      resultType = synthetizeTypes ( resultType , childType );
      if (debug)
	System.out.println("debug TypeCheckVisitor:getChildrenTypesSynthesis"+
			   " child "+i+" type is "+childType+" and synthesis "+
			   " type is "+ resultType );
      // return null if child type is null
      if ( resultType == null ) 
	throw new ParseException("Types "+ saved+" and "+ childType+
				 " are not compatible");
    }

    return resultType ;
  }
    

  /**
   * Return syntheis type of both arguments.
   * <p><b>Algorithm:</b><ul>
   * <li> Transform as many times as possible second type according to the  
   * scheme "Reference to type T" into "type T". 
   * <li> Calls appropriate method synthesisType according to first and second
   * arguments.
   * </ul>
   *
   * @param first first type.
   * @param second second type
   * @result syntheis type of first and second.
   */
  static private LType synthetizeTypes( LType first , LType second ) {

    if ( ( first == null ) || ( second == null) ) 
      return null ;

    while ( second instanceof LReference )
    {
      second = ( (LReference) second).getReferenced() ;
    }

    if ( first instanceof CPPBuiltinType )
      return first.synthesisType ( second ) ;
    else if ( second instanceof CPPBuiltinType )
      return second.synthesisType ( first ) ;
    else
      return first.synthesisType ( second ) ;
  }


  /**
   * Tells if right hand type is assignable to left hand type.
   * <p> Takes into account the fact that "Reference to T" are assignable to
   * "T" type.
   * @param left type we are checking assignability to.
   * @param right type we checking if it is assignable
   * @result true only when right is assignable to left.
   */

  static private boolean isAssignable( LType left,LType right )
  {
    if (debug)
      System.out.println(left + " isAssignable " + right);

    if ( right == null || left == null )
      return false ;

    while ( right instanceof LReference )
    {
      right = ( (LReference) right).getReferenced() ;
    }

    return left.assignable( right ) ;
  }

  public void setNLDrequired(NLEngine nle) {
    nld_required = true;
    nlengine = nle;
  }

  /** NLD to generate */
  private boolean nld_required;



  /**
   * Visit the root node TranslationUnit.
   * Algorithm : <ul> <li>
   * simply visit children and return a boolean telling if type check was ok
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   */

  public Object visit(ASTTranslationUnit node,Object data) 
    throws ParseException {

    node.childrenAccept(this,data);
    return new Boolean(typecheckOK);
  }  


  /**
   * Visits an ADL assertion group node  ASTADLAssGrp.
   *
   * <p> Assertion group children types should be bool when it has more than
   * one assertion. Group type is bool in such case.
   * <p> When node  has 1 assertion child, group type is that assertion type.
   * <p> When the assertion group is empty => assertion group type is null.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children ( down-top visit ).
   * <li>get number of typed children : children implementing interface 
   * ORG.opengroup.adl.parsers.adlcpp.Typed.
   * <li> loop on node children : <ul>
   *  <li> If child is Typed : <ul>
   *   <li> If child is the first typed child, group type is the child type.
   *   <li> Else, make sure child type convertible to bool. group type = bool.
   *   </ul>
   *  <li> Else, no type check is done for non typed children
   *  </ul>
   * <li> When assertion group type not null, save it as a node property.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child
   */
  public Object visit(ASTADLAssGrp node, Object data) throws ParseException  {

    // first visit children
    data = node.childrenAccept(this, data);

    LType childType;
    LType groupType = null ;
    int nbChildren = node.jjtGetNumChildren();
    int nbTypedChildren = 0; // Number of Typed children
    Node child;   
    for ( int i = 0 ; i  <  nbChildren ; i++ )
    {
      child = node.jjtGetChild(i);
      // take into consideration Typed children only
      if( child instanceof Typed ) {
	childType = ((Typed) child).getType();
	if( nbTypedChildren == 0) {
	  groupType = childType;
	  nbTypedChildren++;
	}
	else {
	  // make sure child type convertible to bool.
	  if( ! CPPBuiltinType.convertibleToBool ( childType )  ) {
	    errorMessage(node," Error : Assertions must be boolean in "+
			      "multiple-assertions group");
	  }

	  nbTypedChildren++;
	  groupType= boolType;
	}
      }
    }
    // ASTADLAssGrp may have no children  { } => groupType = null; 
    if( groupType != null) {
      node.setType( groupType);
    }

  return data;
  }


  public Object visit(ASTADLAssertion node, Object data)
    throws ParseException { 

    if (nld_required)
      node.setScope(node.stm.getCurrentScope());

    return node.childrenAccept(this, data);
  }


  /**
   * Visits an ADL behavior declaration node ASTADLBhvDecl.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Save current scope then open (ADL method ) node scope.
   * <li> Save ADL method return type for further reference.
   * <li> Set ab/normal expression types to their default bool value.
   * <li> Visit children. 
   * <li> Restore the former scope of the visit beginning.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child
   */

  public Object visit( ASTADLBhvDecl node, Object data)
   throws ParseException {
     _visitedADLMethod = node.getADLMethod();
    // save scope.
    Scope formerScope = node.stm.getCurrentScope();
    // get node scope = adl method.
    // TBD LCallable adlMethod = (LCallable ) node.getScope();
    LCallable adlMethod = _visitedADLMethod;
    if (debug)
      System.out.println("debug TypeCheck: recovered adl method "+ adlMethod);
    // Update current Scope
    if( adlMethod  != null )
      node.stm.setCurrentScope(  adlMethod );
    // Save Adl method return type.
    adlMethodReturnType = adlMethod.getReturnType() ;
    // set ab/normal default type values.
    NormalExprType  = boolType;
    AbNormalExprType= boolType;
    // visit children. During such visit Abn/NormalExprType may change.
    data = node.childrenAccept(this, data);
    // restore previous scope. 
    node.stm.setCurrentScope(formerScope );
    _visitedADLMethod = null;
    return data;
  }
 

  /**
   * Visits an ADL behavior classification node ASTADLBhvClassification.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children.
   * <li>If node has only one child, normal and abnormal types = child type.
   * <li> Else => node has 2 children, node FirstChildKind method tells if
   * 1st child is the definition of normal or abnormal. Abn/Normal expression
   * types are set accordingly. 
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  
  public Object visit( ASTADLBhvClassification node, Object data)
    throws ParseException {

    int nbChild = node.jjtGetNumChildren();
    // Visit children.
    data = node.childrenAccept(this, data);

    if( nbChild == 1) {
      NormalExprType   = ( (Typed) node.jjtGetChild(0)).getType();
      AbNormalExprType = NormalExprType;
    }
    else if( nbChild ==2) {
      String firstChild = node.getFirstChildKind();
      if( firstChild.equals("NORMAL") ) {

	NormalExprType   = ( (Typed) node.jjtGetChild(0)).getType();

	AbNormalExprType = ( (Typed) node.jjtGetChild(1)).getType();      
      }
      else {
	NormalExprType   = ( (Typed) node.jjtGetChild(1)).getType();

	AbNormalExprType = ( (Typed) node.jjtGetChild(0)).getType();
      }
    }

    if (nld_required)
      node.setScope(node.stm.getCurrentScope());

    return data;

  }

  /**
   * Visits an ADL class node ASTADLClass.
   * <p><b>Algorithm:</b><ul>
   * <li> Save current scope then open node scope.
   * <li>set variable <a href="#currentADLClass">currentADLClass</a> to node
   * ADL class 
   * <li> Visits children.
   * (li> Restore the former scope of the visit beginning.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return returned by the visited child.
   */
  public Object visit( ASTADLClass node, Object data) throws ParseException {

    Scope formerScope = node.stm.getCurrentScope();
    Scope classScope = node.getScope();
    
    if(classScope == null  ||! (classScope instanceof ADLClass) )
       // should never occurs
       return data ;
 
    // Update current Scope
    node.stm.setCurrentScope(classScope );
    currentADLClass = (ADLClass) classScope;
    if(debug) {
      System.out.println("Entering ADL class "+
		 ((ADLClass) classScope).getName() );
    }     
    // visit children
    data = node.childrenAccept(this, data);
    // restore former scope.
    node.stm.setCurrentScope(formerScope );

    return data;
  }

  /**
   * Visits an ADLSuperClass node.
   * <p> <b> Algorithm : </b> <ul>
   * check all adl superclasses are accessible
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADLSuperClass node,Object data) 
  throws ParseException 
  {
    // check all adl superclasses are accessible
    Enumeration superClasses = node.getSuperClasses().elements();
    while ( superClasses.hasMoreElements() ) {
      LClass cppSuperClass = (LClass)superClasses.nextElement();
      ADLClass adlSuperClass = SimpleNode.stm.getADLClass
	                       (cppSuperClass.getName(),null);
      if ( adlSuperClass == null ) {
	errorMessage
	(node,
	 "cannot find adl class " 
        + cppSuperClass.getName()
        + ": adl superclass of tested class must be reachable!"
	);
      }
    }
    return data;
  }

  /**
   * Visits an ADL Epilogue  node ASTADLEpilog.
   * <p> <b> Algorithm : </b> <ul>
   * Simply visits children. Epilog scope opened by CompoundStatement child.
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADLEpilog node,Object data) throws ParseException 
  {
    return node.childrenAccept(this, data);
  }


  /**
   * Visits an ADL Prologue  node ASTADLProlog.
   * <p> <b> Algorithm : </b> <ul>
   * Simply visits children. Epilog scope opened by CompoundStatement child.
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADLProlog node,Object data) throws ParseException 
  {
    return node.childrenAccept(this, data);
  }


  /**
   * Visits an ASTADLInhSemantics node.
   * <p> <b> Algorithm : </b> <ul>
   * <li>check one adl super class contains tested method
   *     (ADL inheritance tree is recursively checked)
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADLInhSemantics node,Object data) 
  throws ParseException 
  {
    // Check one adl super class contains tested method

    boolean isConstructor = _visitedADLMethod.isConstructorSemantics();

    ADLClass  adlMethodSuperClass = null;
    ADLMethod adlMethod = null; 

    // several superclasses are likely to be explored (multiple
    // inheritance) let's use an arrray to store them

    Vector superClassNames = new Vector();
    superClassNames.addElement(node.getSuperClassName());

    // explore adl super classes one level at time
    while ( adlMethod == null && superClassNames.size() > 0 ) 
    {
      // this will store adl classes one level up in inheritance tree
      Vector oneLevelUpSuperClasses = new Vector();

      // now explore adl super classes for the method
      Enumeration e = superClassNames.elements();
      while ( e.hasMoreElements() )
      {
	String superClassName = (String)e.nextElement();

	// First, check in symbol table
	adlMethodSuperClass = SimpleNode.stm.getADLClass(superClassName,null);

	if ( adlMethodSuperClass == null )
	{
	  // Symbol table does not contain super class, force parsing.

	  String fileName = SimpleNode.stm.getFileNameForADLClass
                            (superClassName);

	  if ( fileName != null ) {
	    ParseUnit parseUnit = ParseUnit.getParseUnit(fileName, nlengine);
	    if ( parseUnit != null ) {
	      parseUnit.pureParse();
	    }
	  }

	  // ... now symbol should be in symbol table.

	  adlMethodSuperClass = SimpleNode.stm.getADLClass
	                        (superClassName,null);

	}

	// search for the method for which semantics is required

	if ( adlMethodSuperClass != null ) {
	  // collect this super class
	  oneLevelUpSuperClasses.addElement(adlMethodSuperClass);

	  // search for the method
	  adlMethod = (ADLMethod)SimpleNode.stm.getADLMethod
                      (isConstructor ? adlMethodSuperClass.getName()
		                     : _visitedADLMethod.getName(),
		       _visitedADLMethod.getOperationComparator(),
		       adlMethodSuperClass,
		       true);
	  // break if adlMethod is found
	  if ( adlMethod != null ) break;
	}
      }

      // semantics was not found, climb up one level in adl inheritance tree
      // and get new adl superclasses names
      superClassNames = new Vector();
      for ( int i = 0; i < oneLevelUpSuperClasses.size(); i++ )
      {
	  String[] names = ((ADLClass)oneLevelUpSuperClasses.elementAt(i))
	                 .getSuperClassNames();
	  for ( int j = 0; j < names.length; j++ )
	  {
	    superClassNames.addElement(names[j]);
	  }
      }

    }
     
    if ( adlMethod == null ) {
      errorMessage(node,
		    "Super.semantics cannot be used: no semantics defined"
                  + " for " 
                  + _visitedADLMethod.getSignature()
		  + " in adl super class " + node.getSuperClassName()
		  );
    }

    if ( isConstructor && adlMethod != null ) 
    {
      try { 
	node.setProperty("definingADLClass",adlMethodSuperClass.getName()); 
      } catch ( PropertyAccessException e ) { }
    }
    
    return data;
  }



  /**
   * Visits a normal expression node ASTNormalExpr.
   * <p><b>Algorithm:</b><ul>
   * Sets node type to visitor variable  <a href="#NormalExpr">NormalExpr</a>.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  public Object visit(ASTNormalExpr node, Object data) throws ParseException  {
    
    node.setType( NormalExprType);
    return data;
  }

  /**
   * Visits abnormal expression node ASTAbormalExpr.
   * <p><b>Algorithm:</b><ul>
   * Sets node type to variable  <a href="#AbNormalExpr">AbNormalExpr</a>.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  public Object visit(ASTAbnormalExpr node, Object data) throws ParseException  {
    node.setType( AbNormalExprType);
    return data;
  }

  /**
   * Visit ASTADLQuantifiedAssertion.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> save former scope
   * <li> set scope to quantified assertion 
   * <li> visit children
   * <li> restore former scope
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADLQuantifiedAssertion node,Object data) 
  throws ParseException {

    // save former scope
    Scope oldScope = node.stm.getCurrentScope();

    // set scope to quantified assertion 
    Scope quantifierScope = node.getScope();

    if ( quantifierScope != null)
      node.stm.setCurrentScope(quantifierScope);

    // visit children
    data = node.childrenAccept(this, data);

    // reset old scope
    node.stm.setCurrentScope( oldScope );

    return data;
  }

  /**
   * Visits this node ASTThis.
   * <p><b>Algorithm:</b><ul>
   * Set node type as pointer to ORG.opengroup.adl.symboltable.LClass of  
   * variable <a href="#currentADLClass">currentADLClass</a>
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  public Object visit(ASTThis node, Object data) throws ParseException  {

    if( currentADLClass == null) {
      errorMessage(node,"Error: Illegal use of \"this\" expression");
      return data ;
    }
    node.setType( node.sbf.newPointer(currentADLClass.getLClass(),0x00));
    return data;
  }



  /**
   * Visit a function definition node FunctionDefinition.
   * <p><b>Algorithm:</b><ul>
   * Visits last node child in case it is a CompoundStatement.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  public Object visit(ASTFunctionDefinition node, Object data)
    throws ParseException  {

    int nbchild = node.jjtGetNumChildren();
    
    if (node.jjtGetChild(nbchild -1 ) instanceof ASTCompoundStatement){
      // visit block statement node
      node.jjtGetChild(nbchild -1).jjtAccept( this, data );
    }

    if (nld_required)
      node.setScope(node.stm.getCurrentScope());

    return data;
  }

  /**
   * Visits a compound statement node ASTCompoundStatement.
   * <p><b>Algorithm:</b><ul>
   * <li> Save current scope then open node scope.
   * <li> Visit children.
   * <li> Restore old scope.
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTCompoundStatement node, Object data)
    throws ParseException  {  

    Scope oldScope = node.stm.getCurrentScope();
    Scope blockScope = node.getScope();

    if ( blockScope != null)
      node.stm.setCurrentScope( blockScope);

    // visit children
    data = node.childrenAccept(this, data);
    // reset old scope
    node.stm.setCurrentScope( oldScope );

    return data;
  }

  /**
   * Visits declaration node  ASTDeclaration.
   *
   * <p> A scope is created for each declaration coming next to statements in
   * the scope of a CompoundStatement. This is to prevents the use in 
   * statements of variables declared later.
   * <p><b>Algorithm:</b><ul>
   * <li> Opens declartions node scope if not null.
   * <li> Visit children.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTDeclaration node,Object data) throws ParseException {
    
    Scope declarationScope = node.getScope();
    if ( declarationScope != null ) 
      SimpleNode.stm.setCurrentScope( declarationScope );
    data = node.childrenAccept(this, data);

    return data;
  }
      
  /**
   * Visits a linkage specification node  ASTLinkageSpecification.
   * <p><b>Algorithm:</b><ul>
   * Simply return <code> data</code>. Do not visit children.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTLinkageSpecification node,Object data) 
  throws ParseException {
 
   return data ;
  }

  /**
   * Visits a catch block node  ASTCatchNode.
   * <p><b>Algorithm:</b><ul>
   * <li>Set scope to catch block scope and visit children.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTCatchNode node, Object data) 
  throws ParseException {
    Scope formerScope = SimpleNode.stm.getCurrentScope();
    SimpleNode.stm.setCurrentScope(node.getScope());
    data = node.childrenAccept(this, data);
    SimpleNode.stm.setCurrentScope(formerScope);
    return data;
  }

  /**
   * Visits an assignment expression node ASTAssignmentExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children first.
   * <li> Make sure right child type assignable to left child type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTAssignmentExpr node,Object data)
   throws ParseException { 
 
    data = node.childrenAccept(this, data);
    LType leftType = ((Typed) node.jjtGetChild(0)).getType();
    LType rightType = ((Typed) node.jjtGetChild(1)).getType();

    if ( (leftType == null ) || !( isAssignable(leftType, rightType) ) ) 
      errorMessage (node,"Error in assignment expression: can not convert "+
                          rightType+" to "+ leftType ); 
    return data;
  }



  /**
   * Visits a pointer to member expression node  PmExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> First visit children.
   * <li> get the vector of node operators.
   */

  public Object visit(ASTPmExpr node,Object data) throws ParseException 
  {
    String strOp;
    LType baseType ;
    LClass pointed ;
    boolean ok = true ;
    int childNum = node.jjtGetNumChildren() ;
    // First visit children
    data = node.childrenAccept(this , data) ;
    // get opretors vector.
    Vector operators = node.getOperatorsVector() ;
    // initial right hand type is last child type.
    LType pointedType =((Typed)node.jjtGetChild( operators.size())).getType() ;

    // loop
    for ( int i = 0 ; i < operators.size() && ok ; i++ ) 
    {
      strOp = (String) operators.elementAt(i);
      // get base type
      baseType = ((Typed)node.jjtGetChild(childNum-i-2 )).getType() ;
      // make sure pointed type is a pointer to member.
      if (( baseType == null ) ||
	  (!(pointedType instanceof LPointerToFieldMember) &&
	   !(pointedType instanceof LPointerToFunctionMember))) {
	errorMessage(node,"Error: Illegal pointer to member expression.. Left "
		     +"hand expression should be a pointer to member");
	ok = false ;
      }

      // case of pointer to function member
      if ( ok && pointedType instanceof LPointerToFunctionMember) {
	pointed = (LClass) ((LPointerToFunctionMember)pointedType).
	                                            getEnclosingClass() ;

 
	if (strOp.equals(".*"))
	{
	  // make sure base type is a subclass of the pointed class.
	  if (!(baseType instanceof LClass ) || 
	      !((LClass)baseType).subclassOf( pointed ))
	  {
	    errorMessage(node,"Error: Illegal pointer to member expression.."
			 +"base expression type is not subclass of the "+
			  "pointed class");
	    ok = false ;
	  }
	  // the new pointed type class is a CFunctionType.
	  if ( ok )
	    pointedType = node.sbf.newFunctionType(
		((LPointerToFunctionMember)pointedType).getPointedComparator(),
		null);
	  if ( debug && ok )
	    System.out.println("debug Type check PmExpr: direct pointer to "+
			       "function member: type found is "+pointedType);
	}
	else if (strOp.equals("->*")) {
	  // in this case base Type should be pointer to a subclass of the 
	  // pointed class.
	  if (!(baseType instanceof LPointer ) ||
	      !(((LPointer)baseType). getPointedToType() instanceof LClass ) ||
	      !(((LClass)(((LPointer)baseType).getPointedToType())).subclassOf
                                                         ( pointed))) 
	  {
	    errorMessage(node,"Error: Illegal pointer to member expression.."
			 +"base expression type should be  a pointer to "
			 +"a subclass of the pointed class");
	    ok = false ;
	  }
	  // the new pointed type class is a CFunctionType.
	  if ( ok) 
	    pointedType = node.sbf.newFunctionType(
	        ((LPointerToFunctionMember)pointedType).getPointedComparator()
		,null);
	  if ( debug  && ok )
	    System.out.println("debug Type check PmExpr: pointer to function"+
			       " member: Recovered type is "+pointedType) ;
	}			       
      }  
      else if ( ok && pointedType instanceof LPointerToFieldMember ) {
	// case of LPointerToFieldMember.
	pointed = (LClass)((LPointerToFieldMember)pointedType).getEnclosingClass();

	if (strOp.equals(".*")) {
	  // make sure base type is subclass of the pointed class.
	  if (!(baseType instanceof LClass ) || 
	      !((LClass)baseType).subclassOf((LClass) pointed)) 
	  {
	    errorMessage(node,"Error: Illegal pointer to member expression.. "+
			 "base expression type not subclass of pointed class");
	    ok = false ;
	  }
	  // new expression type is the type of pointedType is pointing to.
	  if ( ok)
	    pointedType=((LPointerToFieldMember)pointedType).getPointedToType();
	  if ( debug )
	    System.out.println("debug Type check PmExpr: direct pointer to "+
			       "field member: type found is "+pointedType);
	}
	else if (strOp.equals("->*")) {
	  if (!(baseType instanceof LPointer ) ||
	   !(((LPointer)baseType). getPointedToType() instanceof LClass ) ||
	   !(((LClass)(((LPointer)baseType).getPointedToType())).subclassOf
                                                         ( pointed))) 
	  {
	    errorMessage(node,"Error: Illegal pointer to member expression.. "+
			 "base expression type should be a pointer to a "+
			 "subclass of the pointed class");
	    ok = false ;
	  }
	  if ( ok) {
	    pointedType=((LPointerToFieldMember)pointedType).getPointedToType();
	  if ( debug )
	    System.out.println("debug Type check PmExpr: pointer to field"+
			       "member: found type "+pointedType) ;
	  }
	}
      }
    }


    // save recovered type as
    if ( ok )
      node.setType( pointedType ) ;
    return data ;
  }

  /**
   * Visits an Id expression node ASTIdExpr. 
   * <p><b>Algorithm:</b><ul>
   * <li> look for the identifier name and scope override.
   * <li> Look for identifier as a variable in current or its scope if any.
   * <li> When found see if it is a field of the tested C++ class.
   * <li> If not found look for a  method with the identifier name with no
   * comparator specified. Type is then a CFunctionType whose parent class is
   * not null only if the recovered method is a non static class member. 
   * <li> Save the found symbol type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTIdExpr node, Object data) throws ParseException  {

    LType idType = null ;
    boolean ok = true ;
    // get the scope ID override if any.
    Scope idScope = node.getScope();
    // get the identifier name.
    String name = node.getIdName();
    LExpression idExpr = null ;
    if ( localIDLookup == true ) 
    {
      if ( debug)
	System.out.println("debug type check IdExpr: local search of "+ name+
			   " in Scope "+ SimpleNode.stm.getCurrentScope() );
      // First make local search of symbol based on the (un)qualified name.
      // Symbol is first searched as an LExpression.
      idExpr = SimpleNode.stm.getExpression( name,null , true ) ;
      // case symbol was found as an LExpression.
   
      if ( idExpr != null )
      {

	// set id type if no errors 
	// set is field indicator to true if idExpr is static
	idIsNotStaticField = !((CModifier) idExpr.getModifier()).isStatic();
	idType = idExpr.getType() ;
	node.setType ( idType ) ;
	// decorate node
	decorateASTName(node,idExpr);

      } else
      // case LExpression search failed 
      // => look for symbol as a no arguments  function.
      // No comparator is used in search, there should be not more than 1 
      // method with that name.
      // the same algorithm is here again applied.
      {
	 LCallable searched = null ;
	 try {
	   searched = node.stm.getMethod(name, null, null,null, true ) ;
	 } catch ( java.lang.Error e) {
	   errorMessage(node,"Error ambiguous request: can't find symbol "+
			name);
	   ok = false ;
	 }
	 if ( searched == null )
	   {
	     errorMessage(node,"Error: symbol "+ name+" not found "+
			  " in scope "+ SimpleNode.stm.getCurrentScope() );
	     ok = false ;
	   }

	 if ( ok )
	 {
	   // if found method static idType is a CFunctionType.
	   if ( !((CModifier)searched.getModifier()).isStatic() ) {
	     idType=node.sbf.newFunctionType(
					     (LCallableEqualsComparator)
					     searched.getOperationComparator(),
					     (LConstructedType) idScope ) ;
	   }
	   else 
	     idType =node.sbf.newFunctionType((LCallableEqualsComparator)
					      searched.getOperationComparator()
					      ,null ) ;
	   // save current type 
	   node.setType ( idType ) ;
	   // decorate node
	   decorateASTName(node,searched);
	 }
    
      }
      return data ;
    }

	   
    // Case IDExpr not postfixed.
    // search for the symbol in the appropriate type.
    if ( idScope != null) {
      idExpr = node.stm.getExpression( name,idScope);
      if ( ( idScope instanceof LConstructedType ) && (idExpr != null )
	   && ! ( (CModifier) idExpr.getModifier()).isStatic() ) {
	idIsNotStaticField = true ;
	idExprParentClass = (LConstructedType) idScope ;
      }
      else
	idIsNotStaticField = false;
    }
    else 
      idExpr = node.stm.getExpression( name,node.stm.getCurrentScope());

    if ( idExpr != null) {
      // decorate node
      decorateASTName(node,idExpr);

      if (idExpr instanceof TDDDataset) {
	if (debug)
	  System.out.println("debug Type check: IdExpr "+ name +
			     " is dataset (current dataset : " +
			     currentDatasetName + ")");

	// this is a dataset
	node.isDataset = true;
	// add dataset dependency to cycle detector
	if ( currentDatasetName != null ) 
	  {
	    try {
	      cycleDetect.addDependency(currentDatasetName, name );
	    } catch (CycleException e) {
	      Enumeration cycle = e.getCycle().elements();
	      String s = "";
	      while ( cycle.hasMoreElements() )
		{ s += cycle.nextElement() + " "; }
	      errorMessage(node,"Cycle detected for dataset " + currentDatasetName
			   + ": " + s);
	    }
	  }

      } else {

	node.indicateIfClassAttribute ( idExpr.isLClassAttribute() ) ;
	if ( debug && idExpr.isLClassAttribute() )
	  System.out.println("debug Type check: IdExpr "+ name+" is field of the"
			     +" tested class");
	
	//Handle case of checkingBinding mode and idExpr part of bindings vector
	if ( checkingBinding && ( bindingsToCheck.indexOf(idExpr) != -1) ) {
	  if ( debug )
	    System.out.println("debug Typecheck IdExpr: found binding "+idExpr);
	  // node parent should be ArgumentList !
	  // side effect initialization valid only when binding in function args 
	  if ( ! (node.jjtGetParent() instanceof ASTArgumentList ) )
	    {
	      errorMessage(node,"Error: wrong side-effect initialization of bind"+
			   "ing "+idExpr.getName());
	      ok = false ;
	    }
	  // now remove idExpr from binding to check vector.
	  bindingsToCheck.removeElement( idExpr) ;
	}
      }

      idType = idExpr.getType();
    }
    else {
      LCallable searched = null ;
      // now look for  a method with name. Comapartor is null.
      try {
	searched = node.stm.getMethod(name, null, null,idScope ) ;
      } catch ( java.lang.Error e) {
	errorMessage(node,"Error ambiguous request: can't find symbol "+name);
	ok = false ;
      }
      if ( searched == null )
      {
	errorMessage(node,"Error: symbol "+ name+" not found");
	ok = false ;
      }
      if ( ok )
      {
	// decorate node
	decorateASTName(node,searched);

	if ( idScope != null && ( idScope instanceof LConstructedType) && 
	   !((CModifier)searched.getModifier()).isStatic() ) {
	  if (debug ) 
	  System.out.println("debug IdExpr Typecheck: Found method "+searched+ 
			    " is non static function member. "+"Its isStatic="+
			    ((CModifier)searched.getModifier()).isStatic() );
	  idType =node.sbf.newFunctionType((LCallableEqualsComparator)searched.
					  getOperationComparator(),
					  (LConstructedType) idScope ) ;
	}
	else 
	  idType =node.sbf.newFunctionType((LCallableEqualsComparator)searched.
					  getOperationComparator(),null ) ;

	if (debug)
	  System.out.println("debug IdExpr Typecheck: type is CFunctionType");
      }
    }

    // save node type.
    if ( ok )
      node.setType( idType );

    if (nld_required && (idScope == null)) {
      node.setScope(node.stm.getCurrentScope());
    }

    return data ;
  }


  /** 
   * Visits a conditional expression node ASTCondExpr.
   * 
   * <p> Node has 3 children. 
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children and make sure child0 has bool type.
   * <li> Node type is the synthesis type of 2nd and 3rd child types.
   * <li> save synthesis type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTCondExpr node, Object data) throws ParseException  {  
 
    boolean ok = true ;
    // visit children.
    data = node.childrenAccept(this, data);
    // check first type is boolean.
    LType firstchildType = ((Typed) node.jjtGetChild(0)).getType();
    LType secondType = ((Typed) node.jjtGetChild(1)).getType(); 

    if(!CPPBuiltinType.convertibleToBool(firstchildType) || (secondType == null))
    {
      errorMessage(node,"Illegal use of conditional expression");
      ok = false ;
    }

    if ( ok ) 
    {
      LType syntype = synthetizeTypes (((Typed) node.jjtGetChild(2)).getType(),
				        secondType ) ;
      if ( syntype == null ) {
	errorMessage(node,"Incompatible type for conditional expression");
	return data ;
      }
      node.setType( syntype );
    }

    return data;
  }   


  /* Visits an expression node ASTExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children.
   * <li> expression type is first set to first child expression.
   * <li> loop on remaining children setting expression type to synthesisType()
   * of child type and expression type.
   * <li> save expression type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.
   */
  public Object visit(ASTExpr node, Object data) throws ParseException  {

   node.childrenAccept(this, data);
   int nbchild = node.jjtGetNumChildren();
   LType childType;
   
   try {
     LType exprType = getChildrenTypesSynthesis( node, 0,nbchild-1) ;
     // save type as a node property.
     node.setType( exprType);
   } catch ( ParseException e) {
     errorMessage(node,"Type mismatch in expression: "+e.getMessage());
   }


   return data;
  }


  /**
   * Visits a binary arithmetic expression node ASTBinaryArithExpr.
   * 
   * <p> Tests children Type compatibility.
   * <p><b>Algorithm:</b><ul>
   * <li> visit children first.
   * <li> expression type node children synthesis type. Check compatibility.
   * <li> Save the resulting expression type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTBinaryArithExpr node, Object data)
    throws ParseException  {

    Node child;
    LType childType;
    LType exprType;
 
    int nbchildren = node.jjtGetNumChildren();
    // visit children
    node.childrenAccept(this, data);

    try {
      exprType = getChildrenTypesSynthesis( node,0, nbchildren-1) ;
      // save type as a node property
       node.setType(exprType);
    } catch ( ParseException e ) {
      errorMessage(node,"Type mismatch in arithmetic expression: "+
		   e.getMessage());
    }


    return data;
  }


  /**
   * Visits a binary logical expression node ASTBinaryLogicalExpr.
   *
   * <p> verifies all children types convertible to bool type.
   * <p><b>Algorithm:</b><ul>
   * <li> First visit children.
   * <li> getchildren synthesis type and make sure it is convertible to bool
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTBinaryLogicalExpr node, Object data)
    throws ParseException  {

    int nbchildren = node.jjtGetNumChildren();

    node.childrenAccept(this, data);
    // get node syntheis type
    LType nodeType = null ;
    try{
      nodeType = getChildrenTypesSynthesis ( node , 0,nbchildren -1) ; 
    } catch ( ParseException e) {
      errorMessage(node,"Type mismatch in logical expression: "+
		   e.getMessage());
    }
    	
    return data;
  }


  /**
   * Visits ADL if statement node ASTADLDomain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> get named parameter from node.
   * <li> Visit second child.
   * <li> Make sure named parameter type and expression type are compatible.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTADLDomain node, Object data) throws ParseException {

    boolean ok = true ;
    LExpression var = node.getNamedParam() ;
    if ( var == null )
    {
      errorMessage(node,"Illegal ADL domain expression");
      ok = false ;
    }
    // visit conditional expression child.
    node.jjtGetChild(1).jjtAccept( this , data ) ;
    LType exprType = ( (Typed)node.jjtGetChild(1)).getType() ;
  
    if ( ok &&  synthetizeTypes( var.getType() , exprType ) == null )
      errorMessage(node,"Type mismatch in ADL domain definition: types "+
		   var.getType()+" and "+ exprType+" are not compatible");

    return data ;
  }


  /**
   * Visits ADL if statement node ASTADLIfStmt.
   * 
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children first.
   * <li> make sure first child type convertible to bool. 
   * <li> verify second child type and third child type (if exists) compatible
   * <li> save their synthesis type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */

  public Object visit(ASTADLIfStmt node, Object data) throws ParseException {

    // visit children first. 
    data = node.childrenAccept(this, data);
    boolean ok = true ;
    int nbchildren = node.jjtGetNumChildren(); // equals 2 or 3 
    LType firstchildType = ((Typed) node.jjtGetChild(0)).getType();
    if ( ! CPPBuiltinType.convertibleToBool(firstchildType))
    {
      errorMessage(node,"Illegal use of ADL If statement: can not convert "+
		   firstchildType+" to bool.");
      ok = false ;
    }
    LType secondType = ((Typed) node.jjtGetChild(1)).getType();  ; 

    if ( ok ) { 
      try {
	secondType = getChildrenTypesSynthesis( node,1, nbchildren -1 );
      } catch ( ParseException e) {
	errorMessage(node,"Type mismatch in ADL If statement: "+
		     e.getMessage());
	ok = false ;
      }

    }

    // save type into the node
    if ( ok)
      node.setType( secondType);

    if (nld_required)
      node.setScope(node.stm.getCurrentScope());

    return data;
  }   


  /**
   * Visits an ADL binding node ASTADLBinding.
   *
   * <p> type check is done only when the bining is initialized.
   * <p><b>Algorithm:</b><ul>
   * <li> Type check children first.
   * <li> If Initialized binding : <ul>
   *  <li>get node binding type and initialization expression (last child)type
   *  <li> Make sure expression type assignable to node type.
   *  </ul>
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */

  public Object visit(ASTADLBinding node, Object data) throws ParseException 
  {  

    int nbchildren = node.jjtGetNumChildren();
    // visit last ( ADL_Expression ) child and set checkBinding mode and vetor
    checkingBinding = true ;
    bindingsToCheck = node.getBindings() ;
    if ( debug)
      System.out.println("debug Type check ADLBinding: before ADLExpr visit.."+
			 " Number of non initialized bindings "+
			 bindingsToCheck.size() );
    node.jjtGetChild(nbchildren -1).jjtAccept( this , data );
    // set flag off
    checkingBinding = false ;
    if ( debug)
      System.out.println("debug Type check ADLBinding: Visited ADLExpression.."
			 +" Number of not initialized bindings is "+
			 bindingsToCheck.size() );
    // make sure are bindings have been removed from vector => Error if not.
    if ( (bindingsToCheck != null) && (  bindingsToCheck.size() != 0 ) )
    {
      errorMessage(node,"Illegal binding definition: declared binding "+
		   ((LExpression)bindingsToCheck.elementAt(0)).getName()+
		   " not initialized");
    }
    // reset bingings vector.
    bindingsToCheck = null ;
    if ( node.getInitialized() != null ) {
      LType nodeType = node.getInitialized().getType();
      LType exprType = ( (Typed) node.jjtGetChild(nbchildren -1)).getType();
      // check type assignability
      if ( !isAssignable(nodeType, exprType))
	errorMessage(node,"Type mismatch in binding definition: "+
		     "can't convert "+ exprType+" to "+nodeType);

    }
    return data;
  }

  /**
   * Visits ADL inline declaration node ASTADLInlineDecl.
   * 
   * <p> Verifies Assertion group Type equals Inline Function return type
   * <p><b>Algorithm:</b><ul>
   * <li> Get ADLInline method stored as a node property.
   * <li> Save current scope then open ADL Inline scope.
   * <li> Uses the recovered ADL inline method to update the current scope
   * <li> Visit children.   
   * <li> get type ADL inline method return type and its assertion group type.
   * <li> Make sure both types are compatible.
   * <li> restore former scope at the end of the visit.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit (ASTADLInlineDecl node, Object data)
   throws ParseException  {

    // get ADL node inline method
    ADLInline myInline = node.getADLInline();
    // save current scope.
    Scope formerScope = node.stm.getCurrentScope();

    int nbchild = node.jjtGetNumChildren();

    node.stm.setCurrentScope( (Scope) myInline);
    // visit children
    node.childrenAccept(this, data); 
    LType inlineType = myInline.getReturnType();
    // get type of the assertion group.
    LType assgrpType = ((Typed) node.jjtGetChild( nbchild-1)).getType();

    if( debug) {
      System.out.println("debug type check: ADLInline Assertion Group type: "+
			 ( (assgrpType !=null)? assgrpType.getName():""));
      System.out.println("debug type check ADL inline return Type : "+
			 ( (inlineType !=null)? inlineType.getName():"") );
    }       

    // check assertion group type is compatible with inline return type. 
    if ( inlineType == null || !isAssignable ( inlineType,assgrpType ) ) {
      errorMessage(node,"Illegal adl Inline definition: return type of inline"+
		   ((myInline !=null)? myInline.getName():"")+" and related "+ 
		   "assertion group type "+assgrpType+" does not match");

    }
    // restore former scope
    node.stm.setCurrentScope( formerScope);
    return data;
  } 


  /**
   * Visits an unary expression node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Visit the child.
   * <li> If unary operator is a "*" => make sure child type is a pointer and 
   * expression type is  the type pointed to.
   * <li> If unary operator is "&" <ul>
   *  <li>If child type is function type <ul>
   *   <li> If parent class is null (case of static member or non enclosed
   *   variable) => node type is pointer to that function.
   *   <li> Else(non static class member)=>type is pointer to a function member
   *   </ul>
   *  <li>Else if child a non static field ( when<a href="#idIsNotStaticField"
   *  idIsNotStaticField </a> is true, node type is a pointer to field member
   *  <li>Else if it is a static field member, type is a pointer to child type.
   *  <li>Else => node type is a pointer to child type. 
   *  </ul>
   * <li> Else, expression type is the child type.
   * <li> Save recovered type as a node property.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTUnaryExpr node,Object data)
  throws ParseException {

    boolean ok = true ;
    LType exprType = null ;
    // visit the child.
    node.jjtGetChild(0).jjtAccept(this, data ) ;
    // get child type.
    LType childType = ( (Typed) node.jjtGetChild(0)).getType();
    if ( childType == null ) 
    {
      errorMessage(node,"Illegal unary expression type.");
      // leave node visit.
      return data ;
    }
   // get expression operator.
    String strOp = node.getOp();
    // case of "*"
    if( strOp.equals("*") ) {
      // expression type is the type pointed to.
      if ( childType  instanceof LPointer ) {
	exprType = ((LPointer) childType).getPointedToType() ;
	// IMPORTANT
	// C++ standard states that 'If  an  expression initially has the type 
	// "reference to T", the type is adjusted to T" prior to any analysis'
	// Hence according to C++ standard both types are equivalent.
	// Because of implementation raisons ( due to the fact that we generate
	// adl local variable ), unary expression type is A REFERENCE to the 
	// pointed to type when it ( the pointed type) an object.
	if ( exprType instanceof LClass ) 
	  exprType = node.sbf.newReference(exprType ) ;
      }
      else {
	ok = false ;
	errorMessage(node,"Operator \"*\" can be applied only to pointers");
      }
    }
    else if( strOp.equals("&")) {
      if ( childType instanceof CFunctionType ) {
	if ( ((CFunctionType)childType).getEnclosingClass() == null )
	  // type is a pointer to function
	  exprType = node.sbf.newPointerToFunction(((CFunctionType)childType).
						  getFunctionTypeComparator(),0x00);
	else
	  // type is LPointerToFunctionMember
	  exprType = node.sbf.newPointerToFunctionMember
	             (((CFunctionType)childType).getFunctionTypeComparator(),
		      ((CFunctionType)childType).getEnclosingClass() ) ;
      }
      else if ( idIsNotStaticField && ( idExprParentClass!= null)) {
	// type is pointer to field member of child type.
	exprType= node.sbf.newPointerToFieldMember(childType,
						   idExprParentClass);
      }
      else {
	// type is a pointer to child type.
	exprType = node.sbf.newPointer( childType, 0x00) ;
      }
    }
    else {
      exprType = childType ;
    }
    // now save recovered type as a node property
    if ( exprType != null && ok )
      node.setType( exprType ) ;

    return data ;
  }
							
	  

  /**
   * Visits a postfix expression node ASTPostfixExpr.
   *
   * <p> Operates semantic and type check of PostfixExpr.
   * <p> Possible node children are PrimaryExpr, IdExpr, This, String, Return, 
   * Constant, TDDADLExpr and ParenthExpr as 1st child and ArraySuffix,
   * DotAccessSuffix, RefAccessSuffix and PostDeIncrement for other children.
   * <p><b>Algorithm:</b><ul>
   * <li> visit first child.
   * <li> If child 0 is an IdExpr,a PrimaryExpr, a This or Resturn node : set 
   * current type to  child0 type and loop on remaining children : <ul>
   *  <li> If current child is an ArgumentList => Make sure current Type is
   *  CFunctionType. New current type is CFuntionType comparator return type.
   *  <li> If current child is ArraySuffix => verify current type instance of 
   *  ORG.opengroup.adl.symboltable.LArray and update current type into that
   * array element.
   *  <li> If current child is a DotAccessSuffix, update current type with the
   *  of the visit of current child.
   *  <li> If current child is a RefAccessSuffix, update current type with the
   *  of the visit of current child.
   *  <li> Else => current child is a PostDeIncrement. make sure current type 
   *  is not a ORG.opengroup.adl.symboltable.LConstructedType.
   *  </ul>
   * <li> Save the recovered type as the node type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTPostfixExpr node,Object data) throws ParseException {

    // get children number
    int nbChildren = node.jjtGetNumChildren();
    boolean ok = true ;
    Node child =  node.jjtGetChild(0);
    Node curChild ;
    LType curType;
    // update current post fix scope variable.
    postfixContextScope = SimpleNode.stm.getCurrentScope() ;

    child.jjtAccept(this,null);
    if( child instanceof ASTIdExpr || child instanceof ASTPrimaryExpr ||
	child instanceof ASTReturn || child instanceof ASTThis || 
	child instanceof ASTTDDADLExpr || child instanceof ASTString ||
	child instanceof ASTParenthExpr ) {

      curType = ((Typed) child).getType();

      for (int i =1; i< nbChildren && ok ;i++) {
	if ( i == nbChildren -1 )
	  idIsNotStaticField = false ;
	curChild = node.jjtGetChild(i);	
	if( curChild instanceof ASTArgumentList) {
	  // make sure current type is a function type.
	  // TBD in further release: Make sure ArgumentList comparator
	  // equals CFunctionType comparator.
	  if ( ! ( curType instanceof CFunctionType) )
	  {
	    errorMessage(node,"Illegal use of postfix expression");
	    ok = false ;
	  }
	  if ( ok )
	    curType = ( (CFunctionType)curType).getFunctionTypeReturnType();
	}
	else if( curChild instanceof ASTArraySuffix) {

	  // Verify current type is an LArray and update it
	  if(!((curType instanceof LArray) || (curType instanceof LPointer) ) ) 
	  {
	    errorMessage(node,"Error: Illegal use of array expression");
	    ok = false ;
	  }
	  if ( ok)    
	  {
	    if ( curType instanceof LArray ) 
	    {
	      curType = ( (LArray) curType).getElementType();
	    } else {
	      curType = ( (LPointer) curType).getPointedToType() ;
	    }
	    // visit current child to verify that index expr Type = int 
	    curChild.jjtAccept( this,null);
	      
	  }
 
	}
	else if( curChild instanceof ASTDotAccessSuffix) {
	  curType= (LType) curChild.jjtAccept( this, curType);
	  if ( curType == null )
	    ok = false ;
	}
	else if( curChild instanceof ASTRefAccessSuffix) {

	  curType= ( LType ) curChild.jjtAccept( this, curType);
	  if ( curType == null )
	    ok = false ;   
	}
	else if ( curChild instanceof ASTPostDeIncrement ) {
	  
	  if( curType instanceof LConstructedType ) {
	    errorMessage(node,"Error: can not de/increment expression");
	    ok = false ;
	  }	  
	}
      }
      // set the recovered Type as the Node Type
      if( ok )
	node.setType( curType);

    } else if( child instanceof ASTQualifiedType ) { 
      node.setType ( ((ASTQualifiedType)child).getType() ) ;
    } else if ( child instanceof ASTFullyScopedType ){
      node.setType ( ((ASTFullyScopedType)child).getType() ) ;
    } else if ( child instanceof ASTBuiltinTypeSpec ) {
       node.setType ( ((ASTBuiltinTypeSpec)child).getType() ) ;
    }
    else {
      errorMessage(node,"Illegal use of postfixed expression");
      ok = false ;
    }
    return data;
  }

  /**
   * Visits a dot access suffix node ASTDotAccessSuffix.
   *
   * <p>Node created when accessing class object member(filed or method).
   *  Matchs class field/method access expansion: -> "." IdExpr[ArgumentList]
   * <p><b>Algorithm:</b><ul>
   * <li> Make sure @param data is a class or struct type.
   * <li> Set data as the current scope.
   * <li> If node has one child IdExpr 
   *  <ul>
   *  <li> Set <a href="localIDLookup"> local research flag</a> to true. This  
   *  is to tell child IdExpr that symbol should be searched locally.
   *  <li> Visit child then reset local research flag.
   *  <li> return type is the child type.
   *  </ul>
   * <li> Else ( children number = 2) =>
   * <ul>
   *  <li> Set current scope to <a href="postfixContextScope"> current postfix
   *  expression scope</a>. This is to ensure that ArgumentList variables are
   *  those of the postfix expression context.
   *  <li> Construct a comparator and enrich it visiting ArgumentList child.
   *  <li> Restore back scope to <code>data</code>.
   *  <li> Look if there's any method, verifying that comparator and having 
   *  the 1st child name, in the 1 st child scope. 
   *  <li> The return type is the recovered function return type.   
   * </ul>
   * <li> Restore former scope.
   * <li> return the recovered return type.
   * </ul>
   * @param node node to visit.
   * @param data .Current type of postfix expression to update..
   * @return : the updated type. 
   */
  public Object visit(ASTDotAccessSuffix node,Object data)
                                          throws ParseException {
    boolean ok = true ;				    
    if(! (data instanceof LType) )
       return null;

    if(! (data instanceof LClass)) 
    {
      errorMessage(node,"Error: Invalid class member access");
      return null ;
    }
    LType returnType = null ;
    Scope formerScope = node.stm.getCurrentScope();
    int nbChildren = node.jjtGetNumChildren();

    node.stm.setCurrentScope( (Scope) data);

    if ( nbChildren == 1) { // case of "." IdExpr
      // visit the child.
      // set local search flag to true to tell IdExpr that symbol should be 
      // set locally.
      localIDLookup = true ;
      if ( debug )
	System.out.println("debug typecheck DotAccessSuffix: local ID search");
      node.jjtGetChild(0).jjtAccept( this ,null ) ;
      // flag back to false.
      localIDLookup = false ;
      returnType = ( (Typed) node.jjtGetChild(0)).getType(); 
    }
    else  { // case of 2 children

      ScopedName idSN = ((ASTIdExpr)node.jjtGetChild(0)).getScopedName();
      Scope idScope = idSN.Scope ;
      String idName = idSN.Name;
      LCallableAssignComparator fs = new LCallableAssignComparator();
      // set back current scope to current post fix expression scope.
      // This is done to look for argument list variables in the call scope.
      SimpleNode.stm.setCurrentScope( postfixContextScope ) ;
      // Update Comparator with ArgumentList parameter types
      fs = (LCallableAssignComparator) node.jjtGetChild(1).jjtAccept(this,fs);
      // set back current scope to local research scope.
      node.stm.setCurrentScope( (Scope) data);
      // make local search of method in the work scope.
      LCallable idfct = node.stm.getMethod( idName, fs,null, null , true );

      if(idfct == null) 
      {
	errorMessage(node,"Method "+ idName +" not found");
	ok = false ;
      }
      else {
	if ( debug)
	System.out.println("debug TypeCheck DotAccessSuffix: found function:"+
			   idfct.getName());
	// decorate node
	if ( idfct != null ) 
	  decorateASTName((ASTIdExpr)node.jjtGetChild(0),idfct);

	returnType = idfct.getReturnType();
        if (nld_required) {
          try {
            node.setProperty("NLDScope", idfct);
          } catch(PropertyAccessException e) {}
        }
      }
    }

    node.stm.setCurrentScope( formerScope );
    if ( ok )
      return returnType;
    else
      return null ;
  }    
      
  /**
   * Visits an dot access suffix node ASTRefAccessSuffix.
   *
   * <p>Node created when pointed object member(filed or method) access.
   * Matchs pointed field/method access expansion:-> "->" IdExpr[ArgumentList]
   * <p><b>Algorithm:</b><ul>
   * <li>Make sure data is a pointer : ORG.opengroup.adl.symboltable.LPointer 
   * <li> Make sure pointed type is a class or struct type.
   * <li> Set pointed type as the current scope.
   *  <ul>
   *  <li> If node has one child IdExpr => Visit child and make return type 
   *  the child type.
   *  <li> Else ( children number = 2)=> Construct a comparator  and enrich it 
   *  visiting ArgumentList child. Look if there's any method, verifying that
   *  comparator and having the 1st child name, in the 1 st child scope. 
   *  The return type is the recovered function return type.   
   *   </ul>
   * <li> Restore former scope.
   * <li> return the recovered return type.
   * </ul>
   * @param node node to visit.
   * @param data .Current type of postfix expression to update..
   * @return : the updated type. 
   */
  public Object visit(ASTRefAccessSuffix node,Object data)
                                          throws ParseException {

   if( (data ==null) || !(data instanceof LType)  )
       return null ;
   boolean ok = true ;
   LType returnType = null ;
   int nbChildren = node.jjtGetNumChildren();

   if(debug) {
     System.out.println(" in ASTRefAccess, input Argument data: "+
			((LType)data).getName());
   }

   // make sure current type is a pointer.
   if(! (data instanceof LPointer))
   { 
     errorMessage(node,"Invalid pointer class field access");
     return null  ;
   }

   LType  pointedType = ( (LPointer)data).getPointedToType();

   // pointed type should be a class or a struct.
   if(!(pointedType instanceof LClass ) )
   { 
     errorMessage(node,"Error : Invalid class member access"); 
     ok = false ;
   }
   // save current scope.
   Scope formerScope = node.stm.getCurrentScope();
   // we should get placed in the local scope 
   Scope workScope = (Scope) pointedType;


   if ( ok && nbChildren == 1) { // case of "->" IdExpr
     // work scope is the recovered class / struct.
     node.stm.setCurrentScope( workScope);
     // visit first child. Set local search flag.
     localIDLookup = true ;
     node.jjtGetChild(0).jjtAccept(this , data) ;
     returnType = ( (Typed) node.jjtGetChild(0)).getType();
     // local search back to false.
     localIDLookup = false ;
     if( debug) {
       System.out.println(" In RefAccess.field, returnType= "+ 
			  returnType+ " and pointedType =" +
			  pointedType.getName());
     }

   } else if ( ok ) { // case of "->" IdExpr ArgumentList

     ScopedName idSN = ((ASTIdExpr)node.jjtGetChild(0)).getScopedName();
     Scope idScope = idSN.Scope ;
     String idName = idSN.Name;
     // set back current scope to current post fix expression scope.
     // This is done to look for argument list variables in the call scope.
     SimpleNode.stm.setCurrentScope( postfixContextScope ) ;
     LCallableAssignComparator fs = new LCallableAssignComparator();
     // while visiting ArgumentList The sscope should be set to formerScope
     fs = (LCallableAssignComparator) node.jjtGetChild(1).jjtAccept(this,fs);
     // back to work scope where method should be searched locally.
     node.stm.setCurrentScope( workScope);
     // look locally the method
     LCallable idfct = node.stm.getMethod( idName, fs,null, null, true );

     if(idfct == null) {
       errorMessage(node,"Method "+ idName +" not found");
       ok = false ;
     } else {
       returnType = idfct.getReturnType();
       // decorate node
       decorateASTName((ASTIdExpr)node.jjtGetChild(0),idfct);
       if (nld_required) {
         try {
           node.setProperty("NLDScope", idfct);
         } catch(PropertyAccessException e) {}
       }
     }

     if( ok && debug) {
       System.out.println(" In RefAccess.Method, returnType= "+ 
			  returnType.getName()+ " and pointedType =" +
			  pointedType.getName());
   
       System.out.println("debug TypeCheck: RefAccessSuffix found function:"+
			  idfct);
     }
   }

   // restore former scope.
   node.stm.setCurrentScope( formerScope);
   // return the recovered type.
   if ( ok )
     return returnType;
   else
     return null ;
  }


  /**
   * Visits an array suffix ASTArraySuffix.
   *
   * <p><b>Algorithm:</b><ul>
   *  Node has only one child. Verify the child type is an integer.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTArraySuffix node,Object data) throws ParseException {
    
    LType childType = ( (Typed) node.jjtGetChild(0)).getType();
    if( ! isAssignable (intType,childType) )
       errorMessage(node,"Can not convert array index type "+childType+
		    " to int");

    return data;
  }
       
    

  /**
   * Visits a constructor definition node ASTCtorDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * Only compoundStatement is visited. No verification of constructor
   * initialization is done !
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTCtorDefinition node,Object data)
  throws ParseException {

    int nbchild = node.jjtGetNumChildren() ;
    if ( node.jjtGetChild( nbchild - 1) instanceof ASTCompoundStatement )
      node.jjtGetChild( nbchild -1).jjtAccept ( this , data ) ;

    return data ;
  }

  /**
   * Visits a adl new exression node ASTTDDADLnewExpr.
   * <p> Visit ordered only by PrimaryExpr node visit.
   * <p><b>Algorithm:</b><ul>
   * <li> Make sure <code>data</code> is an LCallableEqualsComparator.
   * <li> Make sure IdExpr is the name of a type and that type is either an
   * LClass.
   * Look for a constructor of that class/struct satisfying the comparator.
   * Type is the class/struct type.
   * @param node node to visit.
   * @param data . the TDDADLnewExpr argument list comparator
   * @return the recovered class/struct whose constrcutor was called..  
   */ 
  public Object visit (ASTTDDADLnewExpr node,Object data) throws ParseException
  {
    if ( data == null || !( data instanceof LCallableEqualsComparator))
      return null ;
    ScopedName sn = node.getScopedName() ;
    LType type = node.stm.getType(sn.Name,sn.Scope) ;
    if ( type != null && ( type instanceof LClass ) ) {
      LCallable method = node.stm.getMethod( ((LClass)type).getName(),
			      (LCallableEqualsComparator)data ,
			      null,
			      ((LClass)type).getScope());
     if ( method == null)
       errorMessage(node,"Illegal use of ADL_new: can not find specified"
			+" constructor in class "+type.getName() );
     else
       // decorate id expression node
       decorateASTName((ASTIdExpr)node.jjtGetChild(0),method);
    } else {
      errorMessage(node,"Illegal use of ADL_new expression.");
    }
    return type ;
  }


  /**
   * Visits a primary expression ASTPrimaryExpr.
   *
   * <p> Node exit  in 2 cases : the expansion -> IdExpr ArgumentList
   *                         and the expansion -> TDDADLnewExpr ArgumentList
   * <br> In both cases second child is an argument list.
   * => Case of a function or constrctor  call with or without parameters.
   * <p> This visit verifies called function is defined in the current scope
   * <p><b>Algorithm:</b><ul> 
   *  <li> Create an LCallableAssignComparator fs 
   * <li> Visit ArgumentList child to updates fs with argument types.
   * <li> If 1st child is TDDADLnewExpr, make his visit having as an argument
   * child comparator. Visit should return a type that is saved as the node 
   * type. End node visit.
   * <li> Else : 1st child is always IdExpr. <ul>
   *  <li> gets fucntion name from the 1st child IdExpression.
   *  <li> Verifies a Method with that function name and that comparator exits
   *  <li> See if recovered function is an ADL inline and set flag if yes.
   *  <li> If in test directive scope, make sure found method is a test method.
   *  <li> Returns the found function Return Type when recovered function is 
   *  not a constructor of the class under test.
   *  <li> If no method was found, look for a constructor. In which case, type
   *  is the constructor LClass.
   *  </ul>
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return the recovered function ( LCallable).  
   */
 public Object visit (ASTPrimaryExpr node,Object data) throws ParseException
 { 

    LType fctType = null ;
    boolean ok = true ;
    ScopedName primeExprSN ;
    // build a comparator.
    LCallableAssignComparator fs = new LCallableAssignComparator();
    //updated comparator (with added parameters) = result of 2nd child visit.
    fs = (LCallableAssignComparator) node.jjtGetChild(1).jjtAccept(this,fs);
    // get primary expressions soped name in both fisrt child cases
    if ( node.jjtGetChild(0) instanceof ASTTDDADLnewExpr ) {
      // visit child 0 and pass him as an argument the argumentlist 2nd child
      data = node.jjtGetChild(0).jjtAccept( this , fs ) ;
      // save result of the visit as the node type .
      if ( (data != null) && ( data instanceof LType))
	node.setType( (LType) data ) ;
      return data ;
    } 

    // now we can assume 1st child is IdExpr.
    primeExprSN = (( ASTIdExpr )node.jjtGetChild(0)).getScopedName();
    Scope primeScope = primeExprSN.Scope ;
    String fctName = primeExprSN.Name;
    //ask symbol table for a method with that comparator in appropriate scope.
    LCallable primefct = node.stm.getMethod( fctName, fs,null, primeScope);
    if ( debug ) {
      System.out.println("primeScope "+ primeScope );
    }
    if(primefct != null) {
      // decorate name node
      decorateASTName(( ASTIdExpr )node.jjtGetChild(0),primefct);
      if ((! inTestDirective) &&
          (primefct instanceof LMethod) &&
          (((LMethod)primefct).isTestMethod())) {
        errorMessage(node, "Error in call of " + primefct.getName() +
                     ": test methods can be used only in test directives");
      }
      // If the recovered method is an ADL Inline, mark that in the node.
      else if ( primefct instanceof ADLInline ) {
	node.isADLInline = true ;
	// see if recoverd symbol is a LMethod of the tested class
      } else if ( (primefct instanceof LMethod ) && (( primeScope == null ) || 
	     (( primeScope instanceof LClass) &&
	      (primeScope.equals( currentADLClass.getLClass()))) ) ) {
	node.isTestedClassMember = true ;
      }
      // now if fctName  is the tested adl class name => constructor => type
      // is the tested LClass.
      // else type is the tested function return type.
      if (currentADLClass!=null && fctName.endsWith(currentADLClass.getName()))
	fctType = currentADLClass.getLClass() ;
      else
	fctType = primefct.getReturnType();

      if (nld_required) {
        try {
          node.setProperty("NLDScope", primefct);
        } catch(PropertyAccessException e) {}
      }
    }
    // now look for a constructor
    else {
      if ( debug)
       System.out.println("debug Typecheck: no function "+fctName+" was found"+
			  ".. Looking for a constructor");
      primefct = node.stm.getConstructor(fctName,fs,null,primeScope);
      if ( primefct == null )
      {
	ok = false ;
	errorMessage(node,"Function "+ fctName +" not found");
      }
      else {
	// decorate name node
	decorateASTName(( ASTIdExpr )node.jjtGetChild(0),primefct);
	// return type is the constructor LClass.
	fctType = node.stm.getClass(fctName ,primeScope); 
      }
    }

    if( debug ) {
      System.out.println
        ("debug TypeCheck PrimaryExpr: found function:"+primefct);
    }

    if ( ok)
      node.setType( fctType);
    return fctType;
  }

 
  /**
   * Visits a function call argument list node ASTArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * Called when visiting PostfixExpr or PrimaryExpr.
   * <li> If data is not LCallableAssignComparator=> stop visit.
   * <li> Else loop on children to add parameters to comparator.
   *  <ul> 
   *  <li> Get "Typed" child parameter and add recoeverd type to comprator.
   *  </ul>
   * </ul>
   * @param node node to visit.
   * @param data .the LCallableAssignComparator to which parameters are added.
   * @return the modified comparator.   
   */
  public Object visit(ASTArgumentList node,Object data) throws ParseException {
    LType paramType;

    if ( !( data instanceof LCallableAssignComparator ) )
      return node.childrenAccept( this , data );

    int nbchild = node.jjtGetNumChildren();
 
    if ( nbchild ==0)
      return data;
    else {
      for(int i=0;i< nbchild;i++ ) {
	node.jjtGetChild(i).jjtAccept(this,null);
	paramType = ((Typed) node.jjtGetChild(i)).getType();
	if( debug ) {
	  System.out.println("ArgumentList : recovered Parameter Type "+
			     paramType.getName());
	}
	( (LCallableAssignComparator)data).addParameterType( paramType);
      }


    }

    return data;
  } 
	


  /**
   * Visits an ADL return statement node ASTReturn.
   * <p> node type is current ADL method return type.
   * <p><b>Algorithm:</b><ul>
   * <li> node type is <a href="#adlMethodReturnType"> adlMethodReturnType</a>
   * <li> Set recovered type as the ASTReturn node type.
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
   public Object visit(ASTReturn node,Object data) throws ParseException { 

     LType nodeType = this.adlMethodReturnType ;
     // save type in the node.
     node.setReturnType ( nodeType ) ;
     return data ;
   }


  /**
   * Visits an ADL try statement node ASTADLTryStmt.
   * Type checking of ADL Try Statement
   * <p><b>Algorithm:</b><ul>
   * Verify Type try Assertion group and catch Assertion groups are compatible
   * <li> first visit children.
   * <li> get synthesis type of catch and try blocks and make sure they are
   * compatible.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
 
  public Object visit(ASTADLTryStmt node,Object data) throws ParseException {
    // visit try block
    node.jjtGetChild(0).jjtAccept(this,data);
    // visit catch blocks
    Scope formerScope = SimpleNode.stm.getCurrentScope();
    try {
      Scope[] catchBlockScopes = (Scope[])node.getProperty("catchBlockScopes");
      for ( int i = 0; i < catchBlockScopes.length; i++)
      { 
	SimpleNode.stm.setCurrentScope(catchBlockScopes[i]);
	// catch block are nodes 2, 4...
	node.jjtGetChild(2*i + 2).jjtAccept(this,data);
	// restore former scope
	SimpleNode.stm.setCurrentScope(formerScope);
      }
    }
    catch (PropertyAccessException e) {}

    boolean ok = true  ;
    int nbchild = node.jjtGetNumChildren();
    LType tryType = (( ASTADLAssGrp) node.jjtGetChild(0)).getType();
    LType catchType ;

    for (int i=1; i < nbchild && ok ;i++) {
      if( node.jjtGetChild(i) instanceof ASTADLAssGrp) {
	catchType = ((ASTADLAssGrp) node.jjtGetChild(i)).getType() ;
	// check type assignability
	if ( tryType == null || !isAssignable(tryType, catchType ) )
	{
	  errorMessage(node,"Illegal adl try statement: Can not convert "+
		      ((ASTADLAssGrp) node.jjtGetChild(i)).getType()+" to "+
		       tryType );
	  ok = false ;
	}
      }
    }
 

    return data;
  }



 /**
   * Visits an ADL implication expression ASTADLImplExpr.
   *
   * <p><b>Algorithm:</b><ul>
   * Make sure both node children have bool types.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child. 
   */
  public Object visit(ASTADLImplExpr node,Object data)
    throws ParseException {

    // First visit children
    data = node.childrenAccept(this, data);
    // get their types.
    LType firstType  = ( (Typed ) node.jjtGetChild(0)).getType() ;
    LType secondType = ( (Typed ) node.jjtGetChild(1)).getType() ;
    // make sure types converible to bool.
    if (! CPPBuiltinType.convertibleToBool( firstType) ||
	! CPPBuiltinType.convertibleToBool( secondType ) )
      errorMessage(node,"Illegal use of "+ node.getOp()+": expression type "+
		   "can not be converted to bool");

    return data ;
  }


  //_____________________ Type checking TDD code ___________________________


  /**
  * Visits a TDD class declaration ASTTDDClass.
  * 
  * <p> Simply opens the scope of the TDD class.
  * @param node node to visit.
  * @param data .returned by the visited child : not used.
  * @return data returned by the visited child.
  */
 public Object visit(ASTTDDClassDecl node,Object data) throws ParseException {
   
   Scope formerScope = node.stm.getCurrentScope();
   node.stm.setCurrentScope(node.getScope() );
   //visit children
   data= node.childrenAccept(this, data);
   //restore former scope.
   node.stm.setCurrentScope( formerScope );

   return data;
 }

  /**
  * Visits a TDD field declaration ASTTDDFieldDeclaration.
  *
  * <p>Number of children is odd. Fisrt child is always DeclarationSpecifiers.
  * Next come pairs of Declarator(or DirectDeclarator) and Typed Expression.
  * <br> Type check is making sure that expression types assignable to 
  * declarator types.
  * <p><b>Algorithm:</b><ul>
  * <li> Get number of children and declarators vector.
  * <li> loop on initialization children and verify that their types assignable
  * to the related declared types.
  * </ul>
  * @param node node to visit.
  * @param data .returned by the visited child : not used.
  * @return data returned by the visited child.   
  */
 public Object visit(ASTTDDFieldDeclaration node,Object data)
 throws ParseException {

   // get children number.
   int childNum = node.jjtGetNumChildren() ;
   // get declarator types.
   boolean ok = true ;
   Vector types = node.getTypes() ;
   LType exprType, declaredType ;

   for ( int i= 0; i < (childNum-1)/2 && ok  ; i++ ) 
   {
     // visit expression child.
     node.jjtGetChild(2*i+2).jjtAccept( this, data ) ;
     // get its type
     exprType = ( (Typed) node.jjtGetChild(2*i+2)).getType() ;
     // declared type is the types vector i'th element.
     declaredType = (LType) types.elementAt(i) ;
     if ( ! isAssignable(declaredType, exprType ) )
     {
       errorMessage(node,"Type mismatch in TDD field declaration: "+exprType+
		    " can no be assigned to "+declaredType);
       ok = false ;
     }
   }

   return data ;
 }


 /**
   * Visits a TDD dataset domain node ASTTDDDatasetDomain.
   *
   * <p>Node has 2 children.
   * Verify initializer child type assignable to dataset type.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit second child(initialization expression child) and get its type
   * <li> get dataset domain type from node.
   * <li> Check if initializer type assignable to dataset type.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.   
   */
 public Object visit(ASTTDDDatasetDomain node,Object data)
  throws ParseException {
    if (node.jjtGetNumChildren() == 1) {
      // untyped parameter: must be a void dataset expression
      // Visit child.
      node.jjtGetChild(0).jjtAccept(this,null);
      // get intializer type.
      LType initType = ( (Typed) node.jjtGetChild(0)).getType();
 
      if(debug_tdd) {
	System.out.println("debug_tdd:Dataset anonymous domain initializer "+
			   "type= "+
			   ( (initType !=null)? initType.getName():""));
      }
      // check initilizer type is void
      if( (initType == null) || !(initType instanceof CPPBuiltinType) ||
	  !(((CPPBuiltinType)initType).isVoid()))
	errorMessage(node,"Error: untyped dataset domain definition is not a"+
		     " void dataset expression");
    }
    else {
      // Visit second child.
      node.jjtGetChild(1).jjtAccept(this,null);
      // get intializer type.
      LType initType = ( (Typed) node.jjtGetChild(1)).getType();
      // get dataset type saved within the node
      LType datasetType =  node.getDomainType();
 
      if(debug_tdd) {
	System.out.println("debug_tdd:Dataset domain type="+
			   ((datasetType!=null)? datasetType.getName():"")+ 
			   " and domain initializer type= "+
			   ( (initType !=null)? initType.getName():""));
      }
      // check initilizer type assignable to dataset type
      if( datasetType == null || !isAssignable(datasetType, initType )  )
	errorMessage(node,"Error: type mismatch in dataset domain definition");
    }
   return data;
 }



 /** Visits a TDD dataset declaration node ASTTDDDatasetDecl.
   * 
   * Node has exactly two children.
   * <p><b>Algorithm:</b><ul>
   * <li> Update current dataset name.
   * <li> Visit children the reset current dataset name.
   * <li>get declaration dataset type and left hand initialization. expression
   *  type. and check if erifies both types are compatible.
   * </ul>
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   */
 public Object visit(ASTTDDDatasetDecl node,Object data) throws ParseException
 {
   // update current dataset name.
   currentDatasetName = node.getDataset().getName() ;
   // visit children.
   Node initNode = node.jjtGetChild(1);
   initNode.jjtAccept( this,null);
   // reset current dataset name
   currentDatasetName = null ;
   LType declaredType =  node.getDatasetType();
   // make sure initialization type is assignable to declared dataset type
   if (((Typed) initNode).getType() == null) {
     errorMessage((SimpleNode) initNode,
		  "Type check error in initializer for dataset "
		  + currentDatasetName + ".");
   } else if (!isAssignable(declaredType,((Typed) initNode).getType()) ) {
      errorMessage(node,
		   "Incompatible type for dataset initializer: Can't convert "
		   + ((Typed) initNode).getType().getName()
		   + " to "
		   + declaredType.getName() + ".");
   }

   return data;

 }


 /** 
   * Visits a TDD dataset literal node ASTTDDDatasetLiteral.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit children.
   * <li>perform dataset children types synthesis and check compatiblility 
   * using<a href="#getChildrenTypesSynthesis">getChildrenTypesSynthesis()</a>
   * <li> Set node type accordingly.
   * </ul>
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   */
 public Object visit(ASTTDDDatasetLiteral node,Object data)
  throws ParseException {

   // visit children
   data = node.childrenAccept(this, data);
   int nbchild = node.jjtGetNumChildren();
   // get children synthesis type
   try {
     LType literalType = getChildrenTypesSynthesis(node,0,nbchild-1 );
     // set node type
     node.setType( literalType );
   } catch ( ParseException e) {
     errorMessage(node,"Type mismatch in dataset literal members: "+
		  e.getMessage());
   }

   return data;
 }

 /**
   * Visits a TDD test directive node ASTTDDTestDirective.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Open test directive scope.
   * <li> Visit children.
   * </ul>
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   */
  public Object visit(ASTTDDTestDirective node,Object data)
   throws ParseException {

     if ( debug_tdd)
       System.out.println("debug_tdd: starting visit of TDD Test Directive");
    // set flag in test directive to true
    inTestDirective = true ;
    // get initial scope
    Scope initial = SimpleNode.stm.getCurrentScope() ;
    // open test directive scope. 
    SimpleNode.stm.setCurrentScope( node.getScope() );
    // Visit children
    node.childrenAccept(this,data);
    // Exit scope
    SimpleNode.stm.setCurrentScope( initial ) ;
    inTestDirective = false ;
    
    return data ;
  }

 /**
   * Visits a TDD dataset member node ASTTDDDatasetMember.
   *
   * <p><b>Algorithm:</b><ul>
   * Type checked only when has 2 children.
   * <li> Visit both node children.
   * <li> Get first child type and second type.
   * <li> verify that both types are assignable to long.
   * Type checking of TDD Dataset Literal. Node has at least on child
   * The node ASTDatasetMember is craeted only when nb of children == 2
   * verifies type firstchild is compatible with the type of second child
   */
 public Object visit(ASTTDDDatasetMember node,Object data) 
  throws ParseException {

   // visit children
   data = node.childrenAccept(this, data);
   if ( node.jjtGetNumChildren() == 1)
     return data ;
   // get first and second chiild type.
   LType firstType = ( ( Typed) node.jjtGetChild(0)).getType();
   LType secondType = ( ( Typed) node.jjtGetChild(1)).getType();

   if ( ! isAssignable(CPPBuiltinType.longT, firstType ) ||
	! isAssignable(CPPBuiltinType.longT, secondType ) )
     errorMessage(node,"Error: can not convert dataset member types to long");

   return data;
 }


 /**
   * Visits TDD factory definition node ASTTDDFactoryDef.
   * <p><b>Algorithm:</b><ul>
   * <li> get node factory type and make sure it is pointer.
   * <li> open factory scope and set inFactoryScope flag.
   * <li> Visit CompoundStatement child ( child(2) )
   * <li> If number of children > 2 => relinquish statement exists =>Open node
   * relinquish scope and visit remaining children.
   * <li> Restore initial scope and set back inFactoryScope flag to false.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.  
   */

 public Object visit(ASTTDDFactoryDef node,Object data) throws ParseException
 {

   int nbChildren = node.jjtGetNumChildren();
   Node curChild;
   Node returnChild = null;
   currentFactoryType = node.getFactoryType();

   Scope formerScope = node.stm.getCurrentScope();

   Scope factoryScope = node.getFactoryScope();
   if( factoryScope != null) {
     node.stm.setCurrentScope( factoryScope);
   }

   inFactoryDefinitionScope = true;
   if ( debug_tdd )
     System.out.println("debug_tdd Typecheck Factory definition: Visit Block");
   // visit compond statement child
   node.jjtGetChild(2).jjtAccept( this ,data );
   //If relinquish exist, open relinquish scope & visit related children.
   if ( nbChildren > 3 ) {
     SimpleNode.stm.setCurrentScope( node.getRelinquishScope() ) ;
     if ( debug_tdd)
       System.out.println("debug_tdd Typecheck factory definition: visit "+
			  "relinquish block");
     node.childrenAccept( this, data , 3) ;
   }
   inFactoryDefinitionScope = false;
   node.stm.setCurrentScope( formerScope);

   return data;

 }



 /**
   * Visits a  return statement node ASTReturnStmt.
   * 
   * <p><b>Algorithm:</b><ul>
   * <li>  visit child when exists.
   * <li> When inFactoryDefinitionScope : Get child type and check its  
   * compatible with  <a href="#currentFactoryType"> current factory type</a>.
   * </ul>
   * @param node node to visit.
   * @param data .returned by the visited child : not used.
   * @return data returned by the visited child.  
   */

 public Object visit(ASTReturnStmt node,Object data) throws ParseException {

   if ( node.jjtGetNumChildren() == 0)
     return data ;
   if ( debug)
     System.out.println("debug Type check: Visiting return statement child");
   Node myChild = node.jjtGetChild(0);
   myChild.jjtAccept(this,null);
   if( inFactoryDefinitionScope) {
     LType rt = ((Typed) myChild).getType();
     if(currentFactoryType== null || !isAssignable(currentFactoryType,rt))
       errorMessage(node,"Can't convert return statement type "+rt+" to "+
		    "current factory return type "+currentFactoryType);
   }

   if ( debug)
     System.out.println("debug Type check: End visit of return statement");
   return data;
 }

/**
  * Visits a TDD dataset expression node ASTTDDDatasetExpr.
  * 
  * <p> Node is created only when it has at least 2 children.
  * <p><b>Algorithm:</b><ul>
  * <li> Visit children.
  * <li> Call <a href="#getChildrenTypesSynthesis">getChildrenTypesSynthesis()
  * to have syntheis type of all node children.
  * <li> Make sure children types compatible (recovered synthesis type !=null)
  * <li> Set node type accordingly.
  * </ul>
  * @param node node to visit.
  * @param data .returned by the visited child : not used.
  * @return data returned by the visited child.
  */

  public Object visit(ASTTDDDatasetExpr node,Object data)
   throws ParseException {
    LType curType;
    int nbChildren = node.jjtGetNumChildren();

    // visit children 
    data= node.childrenAccept(this, null);
    if ( debug_tdd )
      System.out.println("debug_tdd: Visiting TDD Expression");
    // get node children syntheis type.
    try {
      LType synType = getChildrenTypesSynthesis( node ,0, nbChildren-1 );
      // set node type with the recovered synthesis type.
      node.setType ( synType ) ;
    } catch ( ParseException e) {
      errorMessage(node,"Type mismatch in TDD dataset expression: "+
		   e.getMessage());
    }

    return data;
  }
 

/**
  * Visits a TDD dataset singleton TDDDatasetSingleton.
  *
  * <p> Node is created only when dataset singleton is an identifier.
  * <p><b>Algorithm:</b><ul>
  * <li>get dataset name and ask symboltable for a  LExpression with that name
  * <li> Make sure recovered expression is TDDDataset.
  * <li> Node type is the TDDDataset type.
  * </ul>
  * @param node the node to visit
  * @param data unused data.
  * @return received data object.
  */
  public Object visit(ASTTDDDatasetSingleton node,Object data) throws ParseException {

    // visit children 
    SimpleNode child = (SimpleNode)node.jjtGetChild(0);
    child.jjtAccept(this, null);

    node.setType(((Typed)child).getType());

    return data;
  }



/**
  * Visits a factory call node ASTTDDFactoryCall and checks if called factory
  * is defined.
  * 
  * <p><b>Algorithm:</b><ul>
  * <li> get the factory name and search scope from first child qualifiedId.
  * <li> Build an assign comparator.
  * <li> add remaining children (fatory arguments) types to factory comparator.
  * <li> Look in factory scope for a factory with that name and comparator.
  * <li> Node type is the recovered factory return type.
  * </ul>
  * @param node the node to visit
  * @param data unused data.
  * @return received data object.
  *
  * @see ORG.opengroup.adl.symboltable.LCallableAssignComparator
  */
  public Object visit(ASTTDDFactoryCall node,Object data)
   throws ParseException {

    LType paramType;
    // get simple factory name from child(0).
    String ftryName = ( (ASTQualifiedId) node.jjtGetChild(0)).getName() ;
    Scope ftryScope = ( (ASTQualifiedId) node.jjtGetChild(0)).getScope() ;
    // build the factory comparator.
    LCallableAssignComparator fs = new LCallableAssignComparator();   
    int nbChildren = node.jjtGetNumChildren();
    
    // add child  types to comparator. 
    for(int i=1;i< nbChildren;i++ ) {
      node.jjtGetChild(i).jjtAccept(this,null);
      paramType = ((Typed) node.jjtGetChild(i)).getType();
      fs.addParameterType( paramType);     
    }
    // look for factory in the symbol table.
    LCallable thefactory = node.stm.getMethod( ftryName, fs,null, ftryScope);

    if( debug_tdd) {
      System.out.println("debug_tdd: Found called Factory: "+ thefactory);
    }

    if( (thefactory == null) || !(thefactory instanceof TDDFactory))
    { 
      errorMessage(node,"Error: undefined factory  "+ ftryName);
    }
    else {
      // decorate AST name
      decorateASTName((ASTQualifiedId) node.jjtGetChild(0),thefactory);
      node.setSignature(thefactory);
      node.setType( thefactory.getReturnType() );
    }
    
    return data;
  }


/**
  * Visits a ASTNewExpr (allocation expression).
  * The visit is not complete.
  * @param node the node to visit
  * @param data unused data.
  * @return received data object.
  */
  public Object visit(ASTNewExpr node, Object data) throws ParseException {

    LType fctType = node.getType();
    if (fctType == null)
      errorMessage(node,"Error in \"new\" expression: type not recognized");
    boolean ok = true ;

    String fctName;
    if (fctType instanceof LPointer) {
      LType ptr = fctType;
      while (ptr instanceof LPointer)
        ptr = ((LPointer)ptr).getPointedToType();
      fctName = ptr.getName();
    }
    else {
      fctName = fctType.getName();
    }

    // build a comparator.
    LCallableAssignComparator fs = new LCallableAssignComparator();
    //updated comparator (with added parameters) = result of 2nd child visit.
    if (node.jjtGetChild(node.jjtGetNumChildren()-1) instanceof ASTArgumentList)
      fs = (LCallableAssignComparator) node.jjtGetChild(node.jjtGetNumChildren()-1)
        .jjtAccept(this, fs);

    //ask symbol table for a method with that comparator in appropriate scope.
    LCallable constr = node.stm.getConstructor(fctName, fs, null, null);
    if (constr == null) {
      ok = false ;
      errorMessage(node, "Constructor "+ fctName + fs.getParameterTypes() +" not found");
    }
    else {
      if (nld_required) {
        try {
          node.setProperty("NLDScope", constr);
        } catch(PropertyAccessException e) {}
      }
    }
    return fctType;
  }


  /**
   * Decorate ASTName with useful information.
   * This information comprise:<ul> 
   * <li> the enclosing structure (TDDClass, ADLClass...) of the expression,
   *      property name is "enclosingConstructName"
   * </ul>
   * @param astName ASTName node holding the expression name
   * @param symbol symbol corresponding to the expression
   */

  private void decorateASTName(SimpleNode astName, Symbol symbol) {
    // For TDD dataset and factories save symbol's enclosing structure name.
    if (debug && (astName != null) )
      System.out.println("Decorating " + astName + " : " +
			 ((astName instanceof ASTIdExpr) ?
			  ((ASTIdExpr)astName).getIdName() :
			  ((ASTQualifiedId)astName).getName()) +
			 ", symbol type : " +
			 symbol.getClass().getName());

    if (     ( symbol instanceof TDDFactory ) 
          || ( symbol instanceof TDDDataset ) 
	  || (   ( symbol instanceof LVariable )
	      && ( symbol.getEnclosingConstruct() instanceof TDDClass ))) {
      try {
	TDDClass tddClass = (TDDClass) symbol.getEnclosingConstruct();
	if (tddClass != null) {
	  astName.setProperty("enclosingConstructName", tddClass.getName());
	}
      } catch (PropertyAccessException e) { }
    }

    if (( symbol instanceof LVariable ) &&
	((CModifier)symbol.getModifier()).isStatic()) {
      try {
          astName.setProperty("isStatic", Boolean.TRUE);
      } catch (PropertyAccessException e) { }
    }
  }
  

  // ________________ Private Features ____________________________________

  /** Type boolean : bool */
  private LType boolType = (LType) (SimpleNode.stm).getSymbol("bool");

  /** the Type int */
  private LType intType = (LType) (SimpleNode.stm).getSymbol("int");

  /** boolean to tell if we are visiting the ADL_Expression of an ADL binding*/
  private boolean checkingBinding = false ;
  /** vector containing non initialized bindings that should be childre
    * Relevant only when <code>checkingBinding</code> is true.
    */
  private Vector bindingsToCheck = null ;

  /** boolean for debugging TDD code. true if display desired */
  private static final boolean debug_tdd = false ;

  /* boolean to tell if we are in the scope of a Factory definition
   * used to handle return statement type checking
   */
  private boolean inFactoryDefinitionScope = true; 

  /* LType telling the return Type of the current factory 
   * valid only when inFactoryDefinitionScope = true;
   */
  private LType currentFactoryType ;

  /** Cycle detector for dataset cycles detection */
  private CycleDetector cycleDetect = new CycleDetector() ;

  /** Name of the dataset currently being defined. */
  private String currentDatasetName = null;

  /**
   * current ADL method return type. Updated during ADLBhvDecl visit.
   */
  private LType adlMethodReturnType = null ;

  /* Private field telling the type of the "normal" expression
   * set to default ( boolean type ) when visiting ASTADLBhvDecl
   * updated to the user defined type when visiting ADLBhvClassification
   * stored as ASTNormalExpr node property when visited.
   */
  private LType NormalExprType = boolType;

  /* Private field telling the type of the "abnormal" expression
   * set to default ( boolean type ) when visiting ASTADLBhvDecl
   * updated to the user defined type when visiting ADLBhvClassification
   * stored as ASTAbnormalExpr node property when visited.
   */
  private LType AbNormalExprType = boolType;

  /* Private field storing the current ADLClass used while visiting ASTThis
   */
  private ADLClass currentADLClass= null;
  /** Boolean telling if the IdExpr should be searched in local scope only
    * Set to true while visiting IDExpr child of a Ref/DotAccessSuffix node.
    * Checked during visit of IdExpr to know how symbol search should be.
    */
  private boolean localIDLookup = false ;
  /** Handles current post fix context scope.
    * Updated while visiting ASTPostfixExpr node.
    * Used in visit  Dot/RefAccessSuffix nodes to search argument list in the
    * postfix expression context scope.
    */
  private Scope postfixContextScope ;

  /** Currently visited tested method. */
  private ADLMethod _visitedADLMethod = null;

  /** Handle on the NL engine associated with the compiler */
  private NLEngine nlengine;

  /** boolean for debugging. Set to true when information display required */
  private static final boolean debug = false  ;

}

// ***********************************************************************
// History Log:
// $Log: ADLCPPTypeCheckVisitor.java,v $
// Revision 1.36  1998/07/28 13:51:40  mallet
// Fixed type check of dataset declaration: made it similar to ADL/Java
// and ADL/C.
//
// Revision 1.35  1998/07/02 09:38:53  bareau
// Enhanced the error messages in the visit for NewExpr.
//
// Revision 1.34  1998/07/01 15:05:51  mallet
// Changed errorMessage() method to end type checking at the first error
// encountered.
//
// Revision 1.33  1998/06/30 18:19:47  bareau
// Performs a rudimentary type-check for "new" expressions, and prepare
// for NL unparsing.
//
// Revision 1.32  1998/06/29 16:33:35  mallet
// Replaced import of several classes in ORG.opengroup.adl.parsers by an
// import on demand (.*).
//
// Revision 1.31  1998/06/25 15:38:52  cinard
// Fixed bug : exception parameters declared in catch clauses now have
// their own scope.
//
// Revision 1.30  1998/06/16 13:09:40  cinard
// Added support for quantified assertions type check.
//
// Revision 1.29  1998/06/16 12:46:15  mallet
// Added annotation of IdExpr to distinguish static fields: new property
// "isStatic".
//
// Revision 1.28  1998/06/12 14:43:55  mallet
// Changed visit of an IdExpr to incorporate the case where the id
// expression refers to a dataset (as in a TDD_DatasetSingleton).
// Changed visit of TDD_DatasetSingleton: not always an ID.
// Fixed bug in decorateASTName: TDD fields were not decorated - the type
// of their symbol is LVariable.
//
// ***********************************************************************
