// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1996-1997 Sun Microsystems Inc.
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adlcpp/ParseUnit.java,v $
// $Revision: 1.7 $
// $Date: 1998/10/23 15:01:46 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL for C++ parser
// NAME:        ParseUnit.java
//
// HISTORY:
//      Cyril Inard d'argence, The Open Group - May 07, 1998: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adlcpp;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.*;
import ORG.opengroup.adl.compiler.ADLCompiler;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import ORG.opengroup.adl.nlengine.*;

/**
 * The ParseUnit class enable to monitor the parsing of an adl or tdd
 * source file. 
 * <P>In addition, this class provide static methods to create and/or find
 * ParseUnit objects.
 */

public class ParseUnit
{ 

  // -- Public static methods --

  /** 
   * Get or create a ParseUnit object for supplied source file.
   * @param path full path name of the source file.
   */

  public static ParseUnit getParseUnit(String path, NLEngine nlengine)
  {
     File   file          = new File(path);
     ParseUnit parseUnit = (ParseUnit)_parseUnits.get(file.getName());
     if ( parseUnit == null )
     {
        parseUnit = new ParseUnit(file, nlengine);
        _parseUnits.put(file.getName(),parseUnit);
     }
     return parseUnit;
  }

  // -- Public methods --

  /**
   * Test if the specified file can be parsed, that is at least exists
   * and is readable.
   */

  public boolean canParse()
  {
     return _file.canRead();
  }

  /**
   * Pure parse and type check a source file and construct an AST according to 
   * the rule defined in the ADL for C++ grammar.
   * This methods performs the two passes of the two-pass parsing:<ul>
   * <li>first pass is the AST construction: the AST is constructed and the
   *     symbol table is filled as symbols are defined,
   * <li>second pass is the type check, performed by the ADLTypeCheck visitor:
   *     AST nodes corresponding to typed expressions are decorated with
   *     type information.
   * </ul>
   * <P>Note referenced source files can also be pure parsed are (first pass 
   *    only) as part of this process.</P>
   */

  public Node parse()
  {
     SimpleNode ret = null;

     if ( debug )
     {
        System.out.println("debug: ParseUnit.parse(" + _path + ")");
     }

     // First pass : pure parse

     if ( ! isParsed() )
     {
        pureParse();
     }

     // Second pass : type check if necessary

     if ( ! hasErrors() && ! isTypeChecked() )
     {
        ret = typeCheck();
     }
     else if ( isTypeChecked() )
     {
        ret = _ast;
     }
     
     if ( debug )
     {
        System.out.println("debug: ParseUnit.parse(" + _path + ")"
                          + " return " + ret);
     }

     return ret;
  }

  /**
   * Pure parse a source file and construct an AST according to 
   * the rule defined in the ADL for C++ grammar.
   * This methods performs the first pass of the two-pass parsing.
   * @see ORG.opengroup.adl.adlc.ParseUnit#parse
   */
  
  public SimpleNode pureParse()
  {
     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParse(" + _path + ")");
     }

     // Parse only if it has not been done.

     if ( ! isParsed() )
     {

        // save symbol table context
        Object savedStmContext = SimpleNode.stm.getContext();
                     
        ADLCPPParser parser = null;

        try
        {

           DataInputStream input = new DataInputStream
                                       (new BufferedInputStream
					(ADLCompiler.getStream(_path)));
           
           parser = new ADLCPPParser(_path,
				     new ADLCPPParserTokenManager
				         (new ASCII_CharStream(input,1,1)));

	   // TBD - uncomment only for parser production debugging!
           // if ( debug ) 
           // {
	   // parser.enable_tracing();
           // }

           if (_nld_required)
             parser.setNLDrequired(_nl_engine);

           _ast = parser.TranslationUnit();

           _isParsed = true;

           // Update dependencies
           setDependencies(parser.getUsedSourceFiles());
        }
        catch (FileNotFoundException e)
        {
           // higly improbable since the file has normally been verified
           // in Adlt main
           _hasErrors = true;
           System.err.println(_path + ": File not found!");
        }
	catch (IOException ioe)
	{
	  _hasErrors = true;
	  System.err.println("Preprocessing error: " + ioe.getMessage());
	}
        catch ( ParseException e ) 
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Parse exception caught.";
           System.err.println(message);
        }
        catch ( TokenMgrError e ) 
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Lexical error encountered.";
           System.err.println(message);
        }
	catch ( Throwable e )
	{
	  System.out.println("Caught " + e.getClass() + e.getMessage());
	  if ( debug ) e.printStackTrace(System.out);
	}

        if ( debug )
        {
           // Dumping symbol table

           System.out.println("-- Dumping symbol table --");
           System.out.println(SimpleNode.stm);
        }

        // restore symbol table context
        SimpleNode.stm.setContext(savedStmContext);
     }

     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParse(" + _path + ")"
                           + " return " + _ast );
     }

     return _ast;
  }

  /**
   * Pure parse a C source file and construct an AST according to 
   * the rule defined in the ADL for Java grammar.
   * This methods performs the first pass of the two-pass parsing.
   * @see ORG.opengroup.adl.adlc.ParseUnit#parse
   */
  
  public SimpleNode pureParseCPPFile()
  {
     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParseCPPFile(" + _path + ")");
     }

     // Parse only if it has not been done.

     if ( ! isParsed() )
     {

        // save symbol table context
        Object savedStmContext = SimpleNode.stm.getContext();
                     
        ADLCPPParser parser = null;

        try
        {

           DataInputStream input = new DataInputStream
                                       (new BufferedInputStream
                                            (new FileInputStream(_path)));
           
           parser = new ADLCPPParser(_path,
				     new ADLCPPParserTokenManager
				         (new ASCII_CharStream(input,1,1)));

	   // TBD - uncomment only for parser production debugging!
           // if ( debug ) 
           // {
	   //   parser.enable_tracing();
           // }

           if (_nld_required)
	     parser.setNLDrequired(_nl_engine);

           _ast = parser.TranslationUnit();

           _isParsed = true;

           // Update dependencies
	   setDependencies(parser.getUsedSourceFiles());

        }
        catch (FileNotFoundException e)
        {
           // higly improbable since the file has normally been verified
           // in Adlt main
           _hasErrors = true;
           System.err.println(_path + ": File not found!");
        }
        catch ( ParseException e ) 
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Parse exception caught.";
           // TBD - check for line number 
           System.err.println(message);
        }

        if ( debug )
        {
           // Dumping symbol table

           System.out.println("-- Dumping symbol table --");
           System.out.println(SimpleNode.stm);
        }

        // restore symbol table context
        SimpleNode.stm.setContext(savedStmContext);
     }

     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParseCFile(" + _path + ")"
                           + " return " + _ast );
     }

     return _ast;
  }

  /** Is source file parsed? */

  public boolean isParsed() { return _isParsed; }

  /** Were errors encountered while parsing the file. */
  
  public boolean hasErrors()  { return _hasErrors; }
  
  // -- Private constructors --

  /**
   * Constructs a parse unit object. Useful to monitor parsing
   * of an ADL or TDD source file.
   * <P>Note this constructor is private. The normal way to create
   * a parse unit is simply to invoke static method getParseUnit().
   * </P>
   * @param file adl or tdd source file.
   */

  private ParseUnit(File file)
  {
     _file     = file;
     _path     = file.getPath();
     _fileName = file.getName();
  }

  private ParseUnit(File file, NLEngine nlengine)
  {
    _file     = file;
    _path     = file.getPath();
    _fileName = file.getName();
    _nld_required = nlengine != null;
    _nl_engine = nlengine;
  }

  // -- Private methods --

  /** Get AST for that source file. */
    
  private SimpleNode getAST() { return _ast; }

  /** Has source file been type checked. */

  private boolean isTypeChecked()
  {
     return _isTypeChecked;
  }

  /** Is source file currently being type checked. */

  private boolean isBeingTypeChecked()
  {
     return _isBeingTypeChecked;
  }

  /**
   * Perform type check of the source file.
   * <P>The type check is the second pass of the parsing. It is performed by 
   * the ADLTypeCheck visitor (AST nodes corresponding to typed expressions 
   * are decorated with type information).
   * @see ORG.opengroup.adl.adljava.ParseUnit#parse
   * @return  the AST obtained.
   */

  private SimpleNode typeCheck()
  {
     boolean ok = true;
     boolean completeTypeCheck = true;
     SimpleNode ret = null;

     if ( debug )
     {
        System.out.println("debug: ParseUnit.typeCheck(" + _path + ")");
     }

     // type check only if it has not yet been done (successfuly or not)

     if ( hasErrors() )
     {
        ok = false;
     }
     else 
     {
        if ( ! isParsed() )
        {
           pureParse();
        }

        if ( ! hasErrors() && ! isTypeChecked() )
        {
	   // mark current file as being type checked
	   _isBeingTypeChecked = true;

           ParseUnit[] usedFiles = getDependencies();

           for (int i = 0; i < usedFiles.length; i++)
           {
	      // First pure parse dependent files.
           
              if ( ! usedFiles[i].isParsed() )
              {
                 usedFiles[i].pureParse();
              }
               
	      // then type check if possible to be able to detect cycles:
	      // dependency paths are added during type check only!
	      if (! usedFiles[i].hasErrors()
		  && ! usedFiles[i].isTypeChecked()
		  && ! usedFiles[i].isBeingTypeChecked()) {
		usedFiles[i].typeCheck();
	      } else if (usedFiles[i].isBeingTypeChecked()) {
		// hmm. the used file is currently being type checked,
		// sounds like a circular dependency
		// the current type check cannot be complete until we've
		// also type checked the dependency the other way around
		// example: A uses B uses A,
		// typeCheck(A) -> typeCheck(B) -> isBeingTypeChecked(A)
		//    => typeCheck(B) is incomplete until
		// typeCheck(B) -> typeCheck(A)
		//    => returns status of typeCheck(A)
		//       if hasErrors(A) => typeCheck(B) = false
		//       else do type check B completely
		completeTypeCheck = false;
	      }

              if ( usedFiles[i].hasErrors() )
              {
                 ok = false;
		 System.err.println(_fileName
				    + ": error(s) in dependent file: "
				    + usedFiles[i]._fileName);
                 break;
              }

           }

           // Then type check

           if ( ok )
           {
              // save symbol table context
              Object savedStmContext = SimpleNode.stm.getContext();
              
              ADLCPPTypeCheckVisitor typeCheckVisitor = 
		                 new ADLCPPTypeCheckVisitor(_path, _nld_required);

              if (_nld_required)
                 typeCheckVisitor.setNLDrequired(_nl_engine);

              try
              {
                 Object stmContext = _ast.getProperty("stmContext");
                 SimpleNode.stm.setContext(stmContext);
              }
              catch ( PropertyAccessException e ) { }
              
              try
	      { 
		ok = ((Boolean)_ast.jjtAccept(typeCheckVisitor,null))
                     .booleanValue();
	      }
	      catch ( ParseException e )
	      {
		ok = false;
	      }

              
              // update state
              if ( ok)
              {
                 _isTypeChecked = completeTypeCheck;
                 ret = _ast;
              }
              else
              {
                 _hasErrors = true;
              }

              // restore symbol table context
              SimpleNode.stm.setContext(savedStmContext);
           }

	   _isBeingTypeChecked = false;
        }
     }

     if ( debug )
     {
        System.out.println("debug: ParseUnit.typeCheck(" + _path + ")"
                           + " return " + ret + " (has errors? "
			   + _hasErrors + ")");
     }

     return ret;
  }

  /** Set source file dependencies. */
  
  private void setDependencies(String[] depNames)
  {
     _fileDependencies = new ParseUnit[depNames.length];
     for ( int i = 0; i < depNames.length; i++ )
     {
        _fileDependencies[i] = getParseUnit(depNames[i], _nl_engine);
     }
  }
  
  /** Get source file dependencies */
  
  private ParseUnit[] getDependencies()
  {
     return _fileDependencies;
  }

  // -- Private attributes --

  /** File object associated to the file to parse. */
  private final File   _file;

  /** Last component of full path name. */
  private final String _fileName;

  /** Full path name */
  private final String _path;

  /** Is file pure parsed */
  private boolean      _isParsed      = false;

  /** Is file currently being type checked */
  private boolean _isBeingTypeChecked = false;

  /** Is file type checked (second pass). */
  private boolean      _isTypeChecked = false;

  /** NLD to generate. */
  private boolean _nld_required = true;

  /** NLEngine to feed with mapping. */
  private NLEngine _nl_engine = null;

  /** Does file contains errors. */
  private boolean      _hasErrors     = false;

  /** The AST resulting from the pure parse. */
  private SimpleNode   _ast           = null;

  /** Other parse unit referenced in this parse unit. */
  // TBD - improve
  private ParseUnit[]  _fileDependencies = new ParseUnit[0];

  /** Map names of source files onto ParseUnit objects. */
  private static final HashMap _parseUnits = new HashMap();

  /** Debug flag.*/
  private static final boolean debug = false;

  // -- Test --

  /**
   * Main to use for unit tests of the parser.
   * @param args command arguments that should be [-d] [-c] inputfile.
   */

  public static void main(String args[]) 
  {
     Node            topNode = null;
     String          filename;

     boolean         traced       = false;
     boolean         parsingCPPFile = false ;
     
     int ai;

     for (ai = 0; ai < args.length && args[ai].startsWith("-"); ai++)
     {
        // turn on debugging
        if      ( args[ai].equals("-d") ) { traced       = true; }
        else if ( args[ai].equals("-c") ) { parsingCPPFile = true; }
     }

     if (ai == (args.length-1))
     {
        System.out.println("Reading from file " + args[ai] + " . . .");
        filename = args[ai];
     }
     else
     {
        System.out.println
        ( "Usage: java ORG.opengroup.adl.parsers.adlcpp.ParseUnit"
        + " [-d][-c] inputfile");
        return;
     }

     ParseUnit pu = ParseUnit.getParseUnit(filename, null);
     if ( pu != null )
     {
        if ( parsingCPPFile )
        {
           topNode = pu.pureParseCPPFile();
        }
        else
        {
           topNode = pu.parse();
        }
     }

     if ( topNode != null )
     {
        if ( parsingCPPFile )
        {
           // Unparse file
           ORG.opengroup.adl.unparsers.cpp.CPPUnparserVisitor unparser =
              new ORG.opengroup.adl.unparsers.cpp.CPPUnparserVisitor("tool");
           unparser.setDirectory(".");
           try
	   {
	     unparser.unparse((ASTTranslationUnit)topNode);
	   }
	   catch ( ParseException e ) 
	   {
	     System.out.println("Error during unparsing.");
	   }
        }

        System.out.println("Program parsed successfully.");
     }
     else
     {
        System.out.println("Error encountered.");
     }

     // Dumping symbol table

     System.out.println("-- Dumping symbol table --");
     System.out.println(SimpleNode.stm);
        
  }

}

// ***********************************************************************
// History Log:
// $Log: ParseUnit.java,v $
// Revision 1.7  1998/10/23 15:01:46  mallet
// Bug fixed: put back extension filter preprocessing by calling
// ADLCompiler.getStream() instead of creating a
// FileInputStream. IOException is also caught to handle preprocessing
// errors.
//
// Revision 1.6  1998/07/28 15:56:04  mallet
// Changed ParseUnit behavior to be able to detect cyclic dependencies
// between files even if they are not compiled in the same compiler
// session: a dependency file is not only parsed but also type
// checked. This ensures that all dependency paths from dependent files
// are recorded when we start type checking the current file. When there
// is a circular dependency, the type check of a dependent file is
// flagged as incomplete and should therefore be done again if required.
//
// Revision 1.5  1998/06/08 15:52:36  cinard
// Bug fixed : lexical error (token manager errors) are not reported anymore
// as internal copiler errors.
//
// Revision 1.4  1998/06/08 14:50:45  cinard
// Uncommented call to setNLDRequired().
//
// Revision 1.3  1998/06/05 14:03:08  bareau
// Integrated support for NL parsing.
//
// Revision 1.2  1998/05/28 14:45:35  cinard
// Improved parser initialization.
//
// Revision 1.1  1998/05/27 12:05:03  cinard
// Added support for use clause.
//
// Revision 1.1  1998/05/11 12:03:55  cinard
// ParseUnit is part of the mechanism defined to parse multiple adl source file.
//
// ***********************************************************************
