// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1994-1997 Sun Microsystems Inc.
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adlcpp/SimpleNode.java,v $
// $Revision: 1.11 $
// $Date: 1998/06/29 16:29:23 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   AST
// NAME:        SimpleNode.java
//
// HISTORY:
//      Rob Duncan, Sun Microsystems Inc. - Feb 22, 1997: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adlcpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;

import java.io.PrintStream;

/**
 * The root class for the AST/C++ node hierarchy.
 */
public class SimpleNode extends ADLNode implements Node, Cloneable {

  /** <I>Static</I> symbol table manager */
  static public CPPSymbolTableMgr stm = new CPPSymbolTableMgr();

  /** <I>Static</I> symbol factory */
  static public CPPSymbolFactory sbf = (CPPSymbolFactory)
                                       stm.getSymbolFactory();

  /** The parent node of the current node */
  protected Node parent;

  /** The children nodes of the current node */
  protected Node[] children;

  /** The parser that builds the current node */
  protected ADLCPPParser parser;

  /** node file name */
  private String file ;

  /** Line number private field */
  private int lineNumber = 0;

  /** node current column */
  private int columnNumber = 0;

  /** node scope */
  private Scope scope;

  /**
    * Each node type must implement a one argument constructor to
    * create an empty node.  The new node's association table will be
    * empty, and its field properties will be set to their initial
    * values.  The node will be additive.  The node's jjtCreate() method
    * will call this constructor to build the node.
    *
    * @see #jjtCreate
    */
  public SimpleNode(int id) {
    super(id);
  }

  /**
    * Auxiliary constructor. Calls the main one and initializes
    * the "parser" field.
    */
  public SimpleNode(ADLCPPParser p, int id) {
    this(id);
    parser = p;
  }


  /**
    * Set the parameter node as the parent node of the current node
    */
  public void jjtSetParent(Node n) {
    parent = n;
  }


  /**
    * Return the current node's parent node
    */
  public Node jjtGetParent() { return parent; }


  /**
    * Add the 1st parameter node as <I>i</I>th child node of the current node
    * (with <I>i</I> the second parameter)
    */
  public void jjtAddChild(Node n, int i) {
    if (children == null) {
      children = new Node[i + 1];
    } else if (i >= children.length) {
      Node c[] = new Node[i + 1];
      System.arraycopy(children, 0, c, 0, children.length);
      children = c;
    }
    children[i] = n;
  }


  /**
    * Return the <I>i</I>th child node of the current node
    */
  public Node jjtGetChild(int i) {
    return children[i];
  }


  /**
    * Return the number of children of the current node
    */
  public int jjtGetNumChildren() {
    return (children == null) ? 0 : children.length;
  }

  /** 
   * Reset node children
   */
  public void resetChildren() {
    children = null ;
  }

  public void setScope(Scope s) {

    scope = s;
  }

  public Scope getScope() {

    return scope;
  }


  /**
   * Trims the node children after the first <code>n</code> children.
   *
   * @param n the number of children to keep. Should be smaller 
   * that the original children number.
   */
  public void trimChildren(int n) {
    if (children != null && n < children.length) {
      Node newchildren[] = new Node[n];
      for (int i = 0; i < n; i++) {
	newchildren[i] = children[i];
      }
      // children table is now the new constructed one.
      children = newchildren;
    }
  }

  /**
   * Accept a visitor.
   *
   * @param visitor the visitor
   * @param data some data passed as an argument for the visit
   */
  public Object jjtAccept(ADLCPPParserVisitor visitor, Object data)
    throws ParseException {
    return visitor.visit(this, data);
  }

  /**
   * Visits children nodes.
   *
   * @param visitor the visitor
   * @param data some data passed as an argument for the visit
   */
  public Object childrenAccept(ADLCPPParserVisitor visitor, Object data)
    throws ParseException {
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
        children[i].jjtAccept(visitor, data);
      }
    }
    return data;
  }

  /**
   * Visits children nodes starting from index.
   *
   * @param visitor the visitor
   * @param data some data passed as an argument for the visit
   */
   public Object childrenAccept(ADLCPPParserVisitor visitor,Object data, 
				int begin ) throws ParseException {
    if (children != null) {
      for (int i = begin; i < children.length; ++i) {
        children[i].jjtAccept(visitor, data);
      }
    }
    return data;
  }
   
  /**
    * Output node's id.
    *
    * You can override this method in subclasses of SimpleNode to
    * customize the way the node appears when the tree is dumped.  If
    * your output uses more than one line you should override
    * toString(String), otherwise overriding toString() is probably all
    * you need to do.
    */
  public String toString() { 
    return ADLCPPParserTreeConstants.jjtNodeName[id];
  }

  /**
    * Return a deep clone of the current node
    */
  public Object clone() throws CloneNotSupportedException {
    SimpleNode copy = (SimpleNode)super.clone();

    // create a new array of children and clone them
    if (children != null) {
      copy.children = new Node[children.length];
    }

    if (children != null) {
      SimpleNode childCopy;
      for (int i = 0; i < children.length; i++) {
        childCopy = (SimpleNode)((SimpleNode)children[i]).clone();
        copy.jjtAddChild(childCopy, i);
        childCopy.jjtSetParent(copy);
      }
    }

    return copy;
  }

  /**
    * Produce a textual representation of the nodes and properties of
    * the AST.  The level argument controls the amount of detail in the
    * output.
    */
  protected void dumpNode(PrintStream o, String label, String pad, int level)
  {
    super.dumpNode(o, label, pad, level);

    dumpNodeChildren(o, pad, level);
  }

  protected void dumpNodeChildren(PrintStream o, String pad, int level)
  {
    int n = 0;
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
	SimpleNode k = (SimpleNode)children[i];
	k.dumpNode(o, "" + n++, pad + " ", level);
      }
    }
  }

  /** Saves node line number */
  public void setLineNumber(int ln) { this.lineNumber = ln; }

  /** Saves node current file name */
  public void setCurrentFile ( String filename ) {
    this.file = filename ;
  }

  /** Return node line number */
  public int  getLineNumber()       { return this.lineNumber; }

  // Column number management
  /** Saves column number */
  public void setColumnNumber(int cn) { this.columnNumber = cn; }
  /** returns node column number */
  public int  getColumnNumber()       { return this.columnNumber; }


  /** Displays node error message */
  public  void errorMessage(String msg) throws ParseException {

    if (debug) {
      msg = msg + "\n Scope where error occured :" +
	stm.getCurrentScope();
    }

    errorMessage(getLineNumber(), msg);
  }

  /** Displays node error message specifying the error line number */
  public  void errorMessage(int line, String msg ) throws ParseException
  {
    line = line - 1; // off by one!
    if (debug) {
      msg = msg + " Current scope of error detection :" +
	stm.getCurrentScope();
    }

    throw new ParseException(file + ":" + line + ": " + msg + "\n");
  }


  /** debug flag */
  protected final boolean debug = false ;

}

// ***********************************************************************
// History Log:
// $Log: SimpleNode.java,v $
// Revision 1.11  1998/06/29 16:29:23  mallet
// Removed firstToken field: now in ADLNode.
//
// Revision 1.10  1998/06/05 13:59:55  bareau
// Moved set/get scope from ASTs to SimpleNode.
//
// Revision 1.9  1998/05/27 10:18:36  cinard
// Fixed a bug in trimChidren() method.
//
// Revision 1.8  1998/04/28 12:38:05  cinard
// Updated code to take into account replacement of CSymbolTableMgr,
// CSymbolFactory and CBuiltinType by CPPSymbolTableMgr, CPPSymbolFactory
// and CPPBuiltinType.
// Minor changes in CPPSymbolFactory interface : modifier in newPointer()
// added.
//
// Revision 1.7  1998/04/01 16:38:01  mallet
// Moved common code out into parsers.ADLNode.
//
// ***********************************************************************
