// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adljava/ADLTypeCheckHelper.java,v $
// $Revision: 1.7 $
// $Date: 1998/04/28 12:42:24 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL for Java parser
// NAME:        ADLTypeCheckHelper.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group. - Dec 01, 1997: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adljava;

import java.util.Stack;
import java.util.Enumeration;
import java.util.Vector;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.j.*;

/**
 * Helper class for type check visitor.
 */

public class ADLTypeCheckHelper
{
  private static final boolean debug = false;

  /** 
   * Check all children of a node are assignable to some type.
   * If this is true the method returns the synthetized type.
   * @param parent the parent node of the typed nodes in the AST for
   *               which the check is to be performed.
   * @return the synthetized types if types are compatible, null otherwise.
   */

  static public LType checkTypesAreAssignable(SimpleNode parent,LType type)
  throws BadTypeException
  {
     int childCount = parent.jjtGetNumChildren();
     LType[] tmp = new LType[childCount];
     for (int i = 0; i < childCount; i++)
     {
        TypedNode child = (TypedNode)parent.jjtGetChild(i);
        if ( ! type.assignable(child.getType()) )
        {
           throw new BadTypeException
                     (child,
                      "Cannot convert " + child.getTypeName() 
                      + " to " + type.getName()
                     );
        }
        else
        {
           tmp[i] = child.getType();
        }
     }

     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = 
                 (SimpleNode)parent.jjtGetChild(errorTypeIndex );
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }

  /** 
   * Check all children of a node are of the same type kind.
   * If this is true the method returns the synthetized type.
   * @param parent the parent node of the typed nodes in the AST for
   *               which the check is to be performed.
   * @return the synthetized types if types are compatible, null otherwise.
   */

  static public LType checkTypesAreSameKind(SimpleNode parent,Integer kind)
  throws BadTypeException
  {
     int childCount = parent.jjtGetNumChildren();
     LType[] tmp = new LType[childCount];
     for (int i = 0; i < childCount; i++)
     {
        TypedNode child = (TypedNode)parent.jjtGetChild(i);
        if ( ! kind.equals(child.getTypeKind()) )
        {
           throw new BadTypeException(child,
                                      "Can't convert " + child.getTypeName()
                                      + " to " + kindToString(kind));
        }
        else
        {
           tmp[i] = child.getType();
        }
     }
     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = 
                 (SimpleNode)parent.jjtGetChild(errorTypeIndex );
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }

  /** 
   * Check all children of a node are of the same type kind.
   * If this is true the method returns the synthetized type.
   * @param parent the parent node of the typed nodes in the AST for
   *               which the check is to be performed.
   * @return the synthetized types if types are compatible, null otherwise.
   */

  static public LType checkTypesAreSameKind(SimpleNode parent,Integer kind,
                                            int first,int last)
  throws BadTypeException
  {
     LType[] tmp = new LType[last - first + 1];
     TypedNode child = null;
     for (int i = first; i <= last; i++)
     {
        child = (TypedNode)parent.jjtGetChild(i);
        if ( ! kind.equals(child.getTypeKind()) )
        {
           throw new BadTypeException(child,"");
        }
        else
        {
           tmp[i - first] = child.getType();
        }
     }
     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = 
                (SimpleNode)parent.jjtGetChild(first + errorTypeIndex );
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }

  /** 
   * Synthetize types (with types being the children of provided an AST node).
   * Note children must be instance of TypedNode.
   * @see ADLTypeCheckHelper#synthetizeTypes(LType[])
   * @param parent the parent node of the typed nodes in the AST for
   *               which the synthesis is to be performed.
   * @return the synthetized types if types are compatible, null otherwise.
   */
  static public LType synthetizeTypes(SimpleNode parent)
  throws BadTypeException
  {
     int childCount = parent.jjtGetNumChildren();
     LType[] tmp = new LType[childCount];
     for (int i = 0; i < childCount; i++)
     {
        SimpleNode child = (SimpleNode)parent.jjtGetChild(i);
        if ( child instanceof TypedNode )
        {
           if ( debug ) 
              System.out.println
              ("debug: ADLTypeCheckHelper.synthetizeTypes() - "
               + " child class " + child.getClass() + " of type "
               + ((TypedNode)child).getType() );

           tmp[i] = ((TypedNode)child).getType();
           if ( tmp[i] == null )
           {
              throw new BadTypeException(child,"No type for expression.");
           }
        }
        else
        {
           throw new Error("trying to get the type of a non typed node: " +
                           child.getClass());
        }
     }
     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = 
                 (SimpleNode)parent.jjtGetChild(errorTypeIndex );
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }


  /** 
   * Synthetize types, with types being a subset of an AST node's children.
   * Note children must be instance of TypedNode and must exist.
   * @see ADLTypeCheckHelper#synthetizeTypes(LType[])
   * @param parent the parent node of the typed nodes in the AST for
   *               which the synthesis is to be performed.
   * @param first  index of the first child typed node to consider
   * @param last   index of the last child typed node to consider 
   * @return the synthetized types if types are compatible, null otherwise.
   */
  static public LType synthetizeTypes(SimpleNode parent, int first, int last)
  throws BadTypeException
  {
     LType[] tmp = new LType[last - first + 1];
     for (int i = first; i <= last; i++)
     {
        SimpleNode child = (SimpleNode)parent.jjtGetChild(i);
        if ( child != null && child instanceof TypedNode )
        {
           tmp[i - first] = ((TypedNode)child).getType();
           if ( tmp[i - first] == null )
           {
              throw new BadTypeException(child,"No type for expression.");
           }
        }
        else
        {
           String msg =    ( child == null )
                         ? ( "provided node has no child with index " + i )
                         : ( "trying to get the type of a non typed node." ) ;
           throw new Error(msg);
        }
     }
     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = 
                 (SimpleNode)parent.jjtGetChild(first + errorTypeIndex );
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }
  /** 
   * Synthetize types.
   * @see ADLTypeCheckHelper#synthetizeTypes(LType[])
   * @param typedNodes array containing the typed nodes to synthetize.
   * @return the synthetized types if types are compatible, null otherwise.
   */
  static public LType synthetizeTypes(TypedNode[] typedNodes)
  throws BadTypeException
  {
     LType[] tmp = new LType[typedNodes.length];
     for (int i = 0; i < typedNodes.length; i++ )
     {
        if ( typedNodes[i] != null )
        {
           tmp[i] = typedNodes[i].getType();
        }
        if ( tmp[i] == null )
        {
           throw new BadTypeException(typedNodes[i],"No type for expression.");
        }

     }
     LType ret = synthetizeTypes(tmp);
     if ( ret == null )
     {
        SimpleNode errorNode = null; 
        if ( errorTypeIndex >= 0 )
        {
           errorNode = typedNodes[errorTypeIndex];
        }
        throw new BadTypeException(errorNode,errorMessage);
     }
     return ret;
  }

  static private String kindToString(Integer kind)
  {
     String s = null;
     if      ( JTypeKind.BOOLEAN_TYPE.equals(kind) )    s = "boolean";
     else if ( JTypeKind.ARITHMETIC_TYPE.equals(kind) ) s = "arithmetic";
     else if ( JTypeKind.REFERENCE_TYPE.equals(kind) )  s = "reference";
     return s;
  }

  static int errorTypeIndex = -1;
  static String errorMessage = "";

  /** 
   * Synthetize types (types are provided as an array of types).
   * That is return the mininum type among the list of provided types 
   * considering the Java language types partial ordering.
   * @param types  array of types to be considered for synthesis.
   * @return the synthetized type if types are compatible, null otherwise.
   */
  static private LType synthetizeTypes(LType[] types)
  {
     int kind;

     // Reset error index and error message
     errorTypeIndex = -1;
     errorMessage = "";

     // Get kind of the provided types.
     if ( types.length > 0 ) 
     {
        kind = types[0].getTypeKind().intValue();
     }
     else
     {
        return null;
     }

     // Check all types are non nulls and of the same kind
     for ( int i = 0; i < types.length; i++ )
     {
        if ( types[i].getTypeKind().intValue() != kind )
        {
           errorMessage =   "Can't convert " + types[i].getName() + " to "
                          + types[0].getName();
           errorTypeIndex = i;
           return null;
        }
     }

     // Compute synthesis
     if ( kind == JTypeKind.BOOLEAN_TYPE.intValue() )
     {
        return JTypes.booleanT;
     }
     else if ( kind == JTypeKind.ARITHMETIC_TYPE.intValue() )
     {
        boolean byteDetected        = false;
        boolean charOrShortDetected = false;
        boolean intOrLessDetected   = false;

        LType minType = types[0];
        for ( int i = 1; i < types.length; i++ )
        {
           if ( types[i].assignable(minType) )
           {
              minType = types[i];
           }
           if      ( types[i] == JTypes.byteT ) 
           {
              byteDetected = true;
           }
           else if ( types[i] == JTypes.charT || types[i] == JTypes.shortT )
           {
              charOrShortDetected = true;
           }
           else
           {
               intOrLessDetected = true;
           }
        }
        // Since the order on types is partial, in case only char, short and
        // byte were detected, synthetize to int
        if ( ( ! intOrLessDetected ) && byteDetected && charOrShortDetected ) 
        {
           minType = JTypes.intT;
        }
        return minType;
     }
     else if ( kind == JTypeKind.REFERENCE_TYPE.intValue() )
     {
        // Sounds reasonable to synthetise reference types to any of the types
        return types[0];
     }
     else
     {
        return null;
     }
  }
}
