// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adljava/ADLTypeCheckVisitor.java,v $
// $Revision: 1.50 $
// $Date: 1998/07/28 13:52:26 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL for Java parser
// NAME:        ADLTypeCheckVisitor.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group. - Dec 01, 1997: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adljava;

import java.util.Stack;
import java.util.Enumeration;
import java.util.Vector;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.j.*;
import ORG.opengroup.adl.nlengine.*;

/** 
 * This class defines a visitor which can be applied to an AST
 * generated from the ADLJavaParser grammar. It performs several
 * type checks over.
 *
 * <p>The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and performs type
 * checks for nodes where this is relevant.
 *
 * <p>A helper class ADLTypeCheckHelper is used by this visitor for basic 
 * type check actions such as type synthesis.
 *
 * <p>The type checking starts by accepting this visitor (accept method) from
 * the top node of the AST, that is the ASTADL_CompilationUnit.
 *
 * @see ADLTypeCheckVisitor#visit(ASTADL_CompilationUnit,Object)
 * @see ADLTypeCheckHelper
 */

public class ADLTypeCheckVisitor implements ADLJavaParserVisitor
{

  /**
   * Constructs the type check visitor.
   *
   * @param fileName name of the source file from which the AST was 
   *                 constructed. Used for error reporting only.
   */
  public ADLTypeCheckVisitor(String fileName)
  {
     // Reset symbol table scope
     SimpleNode.stm.setCurrentScope(SimpleNode.stm.getUnitScope());
     scopeStack.push(SimpleNode.stm.getUnitScope());

     this.fileName = fileName;
     int bnd1 = fileName.lastIndexOf('/') + 1;
     int bnd2 = fileName.lastIndexOf('.');
     sourceName = fileName.substring(bnd1, bnd2);
  }


  // -- Inner class --
  
  /**
   * Stores useful information during the process of finding the
   * type of primary expressions.
   */

  class SearchTypeInfo
  {
    /** 
     * Constructs search type information. 
     * @param type type to stored.
     * @param scopedName name to stored.
     * @param scope scope to be stored.
     * @param searchConstructor flag indicating wether next operation to 
     *                          search is a constructor or not.
     */
    SearchTypeInfo
    (LType type,String scopedName,Scope scope,boolean searchConstructor)
    {
       _type              = type;
       _scopedName        = scopedName;
       _scope = ( type instanceof Scope ? (Scope)type : scope );
       _searchConstructor = searchConstructor;
//        _astName           = null;
       _astLastName       = null;
    }

    /**
     * Returns the ASTName (if there is one) coming first in the primary
     * expression. This is useful when information is to be attached to
     * that ASTName while the PrimarySuffixes are processed.
     */
//     ASTName getASTName()                { return _astName; }
     SimpleNode getASTLastName()                { return _astLastName; }

    /**
     * Saves a reference to the ASTName that may come in first position
     * in the processed primary expression.
     * @param astName the astName to save.
     */
//     void    setASTName(ASTName astName) 
//     {
//        // name can be set only once
//        if ( _astName == null ) { _astName = astName; }
//     }
    void    setASTLastName(SimpleNode astPrefix) 
    {
       // name can be set only once
      //if ( _astLastName == null ) { _astLastName = astPrefix; }
      _astLastName = astPrefix;
    }

    /**
     * Get scoped name.
     */
    String getScopedName()              { return _scopedName; }

    /**
     * Set search type information scoped name. The effect is to
     * update type.
     * @param s scoped name.
     */
    void setScopedName(String s)
    {
       _type = null; // reset type
       _scopedName = s; 
    }
    
    /**
     * Get the type held in search type information object.
     */
    LType getType()
    {
       return _type;
    }

    /**
     * Set search type information type. The effect is to update
     * type, scope and scopedName fields.
     * @param type new type.
     */
    void setType(LType type)
    {
       _type = type;
       _scope = ( type instanceof Scope ? (Scope)type : null );
       _scopedName = "";
       _searchConstructor = false;
    }

    /**
     * Return current primary expression scope.
     */
    Scope getScope()
    {
       if ( _type instanceof Scope ) 
       { 
          return _type.getScope();
       }
       else 
       {
          return _scope;
       }
    }

    /**
     * Set scope. Side effect is to update the type and scopedName field.
     */
    void setScope(Scope scope)
    {
       _scope = scope;
       _type  = ( ( scope instanceof LType ) ? (LType)scope : null );
       _scopedName = null;
    }

    /**
     * Return true if search type info indicates a constructor is searched,
     * false otherwise.
     */
    boolean getSearchConstructor()
    {
       return _searchConstructor;
    }

    /**
     * Indicates a constructor is searched.
     */
    void setSearchConstructor()
    {
       _searchConstructor = true;
    }

    /** Type found while parsing the primary expression. */
    private LType   _type               = null;

    /** Current scoped name found while parsing the primary expression. */
    private String  _scopedName         = null;

    /**
     * Current scope where remaining suffixes are to be searched while
     * processing the primary expression.
     */
    private Scope   _scope              = null;

    /**
     * Flag indicating wether the next operation to search is a constructor
     * or not.
     */
    private boolean _searchConstructor  = false;

    /**
     * Reference that should be set to the first ASTName of the Primary
     * expression in case there is one.
     */
//     private ASTName _astName    = null;
    private SimpleNode _astLastName    = null;

    /**
      * Flag indicating that the astNameKind property has been set.
      */
    boolean _astNameKindPropertySet = false;
  }


  public void setNLDrequired(NLEngine nle) {

    nld_required = true;
    nlengine = nle;
  }


  // -- Visit methods --

  /**
   * Visit SimpleNode.
   * <p> Just called in case acceptor not unimplemented in subclass.
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(SimpleNode node,Object data) {
    System.out.println( node +
		       ": acceptor not unimplemented in subclass?");
    node.childrenAccept(this,data);
    return data;
  }

  /**
    * Visit ASTCompilationUnit.
    * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */
  public Object visit(ASTCompilationUnit node,Object data) {
    return data;
  }

  /**
    * Visit ASTPackageDeclaration.
    * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */
  public Object visit(ASTPackageDeclaration node,Object data) {
    return data;
  }

  /**
    * Visit ASTTypeDeclaration.
    * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.

  public Object visit(ASTTypeDeclaration node,Object data) {
    return data;
  }
    */

  /**
   * Visit ASTADL_CompilationUnit.
   * <p> This is the entry point for the type check visitor.
   * <p><b>Algorithm : </b><ul> 
   * <li> visit children 
   * <li> set names of classes from unnamed package as a property
   * </ul>
   * @param node node to visit. Should be the top node of an AST obtained
   *             with the ADL for Java parser.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADL_CompilationUnit node,Object data) {
    node.childrenAccept(this,data);
    try 
    {
       node.setProperty("classesFromUnnamedPackage",
                        SimpleNode.stm.getClassesFromUnamedPackage());
    }
    catch ( PropertyAccessException e ) { }
    return new Boolean(typeCheckOk);
  }

  /**
   * Visit ASTImportDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTImportDeclaration node,Object data) {
    return data;
  }


  /**
   * Visit ASTTDD_UseDeclaration.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTTDD_UseDeclaration node,Object data) {
    return data;
  }

  /**
   * Visit ASTADL_ClassDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> if necessary check adl superclass is accessible
   * <li> push class scope on top of scope stack
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADL_ClassDeclaration node,Object data)
  {
    // If necessary check adl superclass is accessible
    String superClassName = null;
    try 
    {
       superClassName = (String)node.getProperty("extends"); 
    }
    catch ( PropertyAccessException e ) {}
    if ( superClassName != null )
    {
       _adlSuperClass = SimpleNode.stm.getADLClass(superClassName,null);
       if ( _adlSuperClass == null )
       {
          errorMessage
             ("Cannot find adl class " + superClassName
              + ": adl superclass of tested class must be reachable!",
              node
             );
       }
    }
    scopeStack.push(node.getScope());  // Enter scope
    node.childrenAccept(this,data);
    scopeStack.pop();                  // Exit scope
    return data;
  }

  /**
   * Visit ASTADL_InlineDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set inline declaration flag
   * <li> push inline scope (the inline itself) on top of scope stack
   * <li> visit children 
   * <li> check inline assertion group is compatible with inline return type
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */
  public Object visit(ASTADL_InlineDeclaration node,Object data)
  {
     this.inADLInlineScope = true;
     LCallable operation = (LCallable)node.getScope();
     scopeStack.push(operation);        // Enter operation scope
     node.childrenAccept(this,data);
     // Check third children ( ADL_AssertionGroup) type is 
     // compatible with inline return type
     TypedNode assertionGroup = (TypedNode)node.jjtGetChild(2);
     if ( ! operation.getReturnType().assignable(assertionGroup.getType()) )
     {
        errorMessage
        (  "Incompatible type for assertion group in inline. Can't convert "
           + assertionGroup.getTypeName()
           + " to " 
           + operation.getReturnType().getName(),
           assertionGroup
        ); 
     }
     scopeStack.pop();                  // Exit scope
     this.inADLInlineScope = false;
     return data;
  }

  /**
   * Visit ASTADL_BehaviorDeclaration.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push behavior declaration (ADL method) scope on top of scope stack
   * <li> save ADL method return type for further reference in the children
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorDeclaration node,Object data)
  {
     _visitedADLMethod = (ADLMethod)node.getScope();
     this._visitedOpeReturnType = _visitedADLMethod.getReturnType();
     scopeStack.push(_visitedADLMethod); // Enter scope
     node.childrenAccept(this,data);
     scopeStack.pop();                   // Exit scope
     this._visitedOpeReturnType = null;
     _visitedADLMethod = null;
     return data;
  }

  /**
   * Visit ASTADL_Prologue.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set prolog scope flag
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Prologue node,Object data) 
  {
     this.inPrologueScope = true;
     node.childrenAccept(this,data);
     this.inPrologueScope = false;
     return data;
  }

  /**
   * Visit ASTADL_Epilogue.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set epilog scope flag
   * <li> push epilog scope on top of scope stack
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Epilogue node,Object data)
  {
     this.inEpilogueScope = true;
     scopeStack.push(node.getScope());  // Enter scope
     node.childrenAccept(this,data);
     scopeStack.pop();                  // Exit scope
     this.inEpilogueScope = false;
     return data;
  }

  /**
   * Visit ASTADL_BehaviorSpecification.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> set adl method flag
   * <li> visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorSpecification node,Object data)
  {
     this.inADLMethodScope = true;
     node.childrenAccept(this,data);
     this.inADLMethodScope = false;
     return data;
  }

  /**
   * Visit ASTADL_NormalBehavior.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_NormalBehavior node,Object data) 
  {
    node.childrenAccept(this,data);
    if (nld_required)
      node.setScope((Scope)scopeStack.peek());
    return data;
  }

  /**
   * Visit ASTADL_AbNormalBehavior.
   * <p><b>
   * <p> <b> Algorithm : </b> </b><ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_AbNormalBehavior node,Object data) 
  {
    node.childrenAccept(this,data);
    if (nld_required)
      node.setScope((Scope)scopeStack.peek());
    return data;
  }

  /**
   * Visit ASTADL_BehaviorClassification.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> just visit children 
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_BehaviorClassification node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visit ASTADL_AssertionGroup.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> push epilog scope on top of scope stack
   * <li> visit children 
   * <li> get all ADL_Statement in assertion group, in case there is several
   *      ADL_Statement check their types are boolean,
   * <li> set assertion group type to that of enclosed ADL_Statement
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_AssertionGroup node,Object data) 
  {
    scopeStack.push(node.getScope());  // Enter scope
    node.childrenAccept(this,data);
    // Find index of first and last ADL_Statement 
    int numChildren = node.jjtGetNumChildren();
    int firstADLStatement = numChildren; // initialize past last index
    int lastADLStatement  = numChildren - 1;
    for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
    {
       if ( node.jjtGetChild(i) instanceof ASTADL_Statement )
       {
          firstADLStatement = i;
          break;
       }
    }
    int numStatement = lastADLStatement - firstADLStatement + 1;
    if ( numStatement > 1 )
    {
       try 
       {
          LType assertionGroupType = ADLTypeCheckHelper.checkTypesAreSameKind
                                     (node,
                                      JTypeKind.BOOLEAN_TYPE,
                                      firstADLStatement,
                                      lastADLStatement);
          node.setType(assertionGroupType); 
       } 
       catch ( BadTypeException e )
       {
          errorMessage(  "Assertion group with multiple assertions"
                       + " shall contain only boolean assertion."
                       +  e.getMessage()
                       ,  e.getErrorNode() );
       }
    }
    else if ( numStatement == 1 )
    {
       node.setType
       (((TypedNode)node.jjtGetChild(firstADLStatement)).getType());
    }

    if ( debug )
    {
       System.out.println
       ("debug: type of ADLAssertionGroup =  " + node.getTypeName());
    }

    scopeStack.pop();                  // Exit scope
    return data;
  }

  /**
    * Visit ASTADL_Binding.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> push binding scope on top of scope stack
    * <li> visit last children (ASTExpression)
    * <li> check binding type and initializer expression are compatible
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_Binding node,Object data) 
  {
     scopeStack.push(node.getScope());  // Enter scope
     // Only the last child (ASTExpression) is to be visited
     TypedNode expressionNode = (TypedNode)node.jjtGetChild
                                           (node.jjtGetNumChildren() - 1);
     expressionNode.jjtAccept(this,VISITING_EXPRESSION);
     try
     {
        LType bindingType = (LType)node.getProperty("bindingType");
        if (      bindingType == null 
             || ! bindingType.assignable(expressionNode.getType()) 
           )
        {
           errorMessage
           (  "Incompatible type for expression in binding. Can't convert "
              + expressionNode.getTypeName()
              + " to " 
              + bindingType.getName(),
              node
           );

        }
     } 
     catch ( PropertyAccessException e ) { }
     scopeStack.pop();                  // Exit scope
     return data;
  }

  /**
    * Visit ASTADL_Statement.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> set type to that of child
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_Statement node,Object data) 
  {
    node.childrenAccept(this,data);
    // Set type to that of child
    node.setType(((TypedNode)node.jjtGetChild(0)).getType());
    return data;
  }

  /**
    * Visit ASTADL_TryStatement.
    * Check catch clause assertion group types are compatibles with try 
    * block type.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> get typed nodes to use for type check
    * <li> check types are compatible and perform synthesis
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_TryStatement node,Object data)
  {
    node.childrenAccept(this,data);

    // Get typed nodes to use for type check :
    // . ADLAssertionGroup in try block that is child number 0
    // . ADLAssertionGroups in each catch clause block, that is, children 
    //   number 2, 4...

    Vector typedNodes = new Vector();
    typedNodes.addElement(node.jjtGetChild(0));
    for (int i = 2; i < node.jjtGetNumChildren(); i = i+2)
       typedNodes.addElement(node.jjtGetChild(i));
    TypedNode[] _typedNodes = new TypedNode[typedNodes.size()];
    typedNodes.copyInto(_typedNodes);

    // Synthetize types

    try
    {
       LType type = ADLTypeCheckHelper.synthetizeTypes(_typedNodes);
       node.setType(type);
    }
    catch ( BadTypeException e )
    {
       errorMessage
       ("Incompatible type for try catch block. " + e.getMessage(),
        e.getErrorNode()
       );
    }
    return data;
  }

  /**
    * Visit ASTADL_IfStatement.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> set type to boolean
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */
  
  public Object visit(ASTADL_IfStatement node,Object data) 
  {
//     node.childrenAccept(this,data);
//     // Set type to boolean
//     node.setType(JTypes.booleanT);

     // visit first children (indicate it can be an expression)
     TypedNode condNode = (TypedNode)node.jjtGetChild(0);
     condNode.jjtAccept(this,VISITING_EXPRESSION);

     // check first child is boolean
     if (! JTypes.booleanT.equals(condNode.getType())) {
       errorMessage("Incompatible type for if. Can't convert " +
           condNode.getTypeName() +  " to boolean.", condNode);
     }
     
     // visit "then" child
     TypedNode thenNode = (TypedNode)node.jjtGetChild(1);
     thenNode.jjtAccept(this, data);
     LType thenType = thenNode.getType();

     if (node.jjtGetNumChildren() == 2) {
       // no "else" child: "then" child must be boolean
       if (! JTypes.booleanT.equals(thenType)) {
         errorMessage("Incompatible type for if. If statement with no " +
                      "\"else\" must be boolean", thenNode);
       }
       node.setType(thenType);
     }
     else {
       // "then" child and "else" children must have the same type
       TypedNode elseNode = (TypedNode)node.jjtGetChild(2);
       elseNode.jjtAccept(this, data);
       TypedNode[] branches = { thenNode, elseNode };
       try {
         node.setType(ADLTypeCheckHelper.synthetizeTypes(branches));
       }
       catch(BadTypeException e) {
         errorMessage("Incompatible type for if statement. " + e.getMessage(),
                      e.getErrorNode());
       }
     }

     if (nld_required)
       node.setScope((Scope)scopeStack.peek());
     return data;
  }

  /**
    * Visit ASTADL_Assertion.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> visit node's last child (ADL_Expression or ADL_QuantifiedAssertion) 
    * <li> set node type to that of last child
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_Assertion node,Object data) 
  {
     // Only the last child (ADL_Expression or ADL_QuantifiedAssertion) is
     // to be visited

     TypedNode expressionNode = (TypedNode)node.jjtGetChild
                                           (node.jjtGetNumChildren() - 1);
     expressionNode.jjtAccept(this,VISITING_EXPRESSION);
     LType assertionType = expressionNode.getType();

     // Set node type to assertion type
     node.setType(assertionType);
     if (nld_required)
       node.setScope((Scope)scopeStack.peek());
     return data;
  }

  /**
   * Visit ASTADL_Label.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Label node,Object data)
  {
    return data;
  }

  /**
   * Visit ASTADL_Tags.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Tags node,Object data) 
  {
    return data;
  }
  
  /**
    * Visit ASTADL_QuantifiedAssertion.
    * <p> <b> Algorithm : </b> <ul> 
    * <li> push quantified assertion scope on scope stack
    * <li> set type to boolean
    * </ul>
    * @param node node to visit.
    * @param data arbitrary data received from the caller.
    */

  public Object visit(ASTADL_QuantifiedAssertion node,Object data) 
  {
     scopeStack.push(node.getScope());  // Enter scope
     node.childrenAccept(this,data);
     node.setType(JTypes.booleanT);     // Set type to boolean
     if (nld_required)
       node.setScope((Scope)scopeStack.peek());
     scopeStack.pop();                  // Exit scope
     return data;
  }

  /**
   * Visit ASTADL_Quantifier.
   * <p> <b> Algorithm : </b> <ul> <li> not relevant for type check. </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTADL_Quantifier node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }
  
  /**
   * Visits an ASTTDD_ADLnewExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the search type information object
   * <li>get first child node, it contains the name of the tested constructor
   * <li>update search type information object
   * </ul>
   *
   * @param node the node to visit
   * @param data contains a SearchTypeInformation object
   * @return updated SearchTypeInformation object
   *
   */
  public Object visit(ASTTDD_ADLnewExpression node, Object data) 
  {
     // get the search type information object
     SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
     // get first child node, it contains the name of the tested constructor
     SimpleNode childNode = (SimpleNode)node.jjtGetChild(0);
     // update search type information object
     searchTypeInfo.setScopedName(childNode.getName());
     searchTypeInfo.setSearchConstructor();
     return searchTypeInfo;
  }


  /**
   * Visits an ASTTDD_ADLExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the search type information object
   * <li>visit children (indicate children can be expressions)
   * <li>check first child kind is reference
   * <li>update search type information
   * <li>update node type
   * </ul>
   * Note during this processing, the "isClass" property is set.
   *
   * @param node the node to visit
   * @param data contains a SearchTypeInformation object
   * @return updated SearchTypeInformation object
   *
   */

  public Object visit(ASTTDD_ADLExpression node, Object data) 
  {

     // get the search type information object.
     SearchTypeInfo searchTypeInfo = (SearchTypeInfo)data ;

     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     TypedNode childExpr = (TypedNode)node.jjtGetChild(0);
     
     // check first child kind is reference
     if (    childExpr.getTypeKind() != JTypeKind.REFERENCE_TYPE 
          && ! ( childExpr.getType() instanceof LClass )
        )
     {
        errorMessage
        (  "Expression in ADL expression shall be a reference to an object "
         + "or a class. "
         + " Impossible to convert "
         + childExpr.getTypeName() 
         + " to a reference or to a class." ,
         childExpr
        );
     }
     else
     {
        // update search type information
        searchTypeInfo.setType(childExpr.getType());
        
        // update node type
        node.setType(childExpr.getType());

        // update node "isClass" property
        try
        {
           node.setProperty("isClass",childExpr.getProperty("isClass"));
        }
        catch ( PropertyAccessException e ) { }
     }
     return data;
  }

  /**
   * Visits an ASTADL_ImplExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check 0th and 2nd childrens (impl. op skipped) are of boolean type
   * <li>set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_ImplExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check 0th and 2nd children (impl. operator skipped) are of boolean type
     TypedNode erroneousNode = null;
     if (! JTypes.booleanT.equals(((TypedNode)node.jjtGetChild(0)).getType()) )
     {
        erroneousNode = (TypedNode)node.jjtGetChild(0); 
     }
     if (! JTypes.booleanT.equals(((TypedNode)node.jjtGetChild(2)).getType()) )
     {     
        erroneousNode = (TypedNode)node.jjtGetChild(2); 
     }
     if ( erroneousNode != null )
     {
        String operator = "";
        try {
           operator = 
                (String)((SimpleNode)node.jjtGetChild(1)).getProperty("op");
        }
        catch (PropertyAccessException e) { }
        errorMessage
        (  "Incompatible type for " + operator
         + ". Can't convert " + erroneousNode.getTypeName() + " to boolean.",
         erroneousNode
        );
     }
     else
     {
        // set expression type to boolean
        node.setType(JTypes.booleanT);
     }
     if ( debug )
        System.out.println("debug: (ADLImplExpression) - type "
                           + "(" + node.getType() + ") " 
                           + ((TypedNode)node.jjtGetChild(0)).getTypeName() 
                           + " " 
                           + ((TypedNode)node.jjtGetChild(2)).getTypeName()
                          );
     return data;
  }

  /**
   * Visits an ASTADL_ImplOp.
   * <p><b>Algorithm:</b> nothing specific to do here.
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_ImplOp node,Object data) 
  {
    return data;
  }

  /**
   * Visits an ASTConditionalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check first child is boolean 
   * <li>synthetize expression type using the second and third expressions
   *     nodes in conditional expression
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTConditionalExpression node,Object data) 
  {
     boolean ok = true;

     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check first child is boolean 
     if ( ((TypedNode)node.jjtGetChild(0)).getType() != JTypes.booleanT )
     {
        errorMessage
        ("Incompatible type for ?:. Can't convert " +
         ((TypedNode)node.jjtGetChild(0)).getTypeName() + " to boolean.",
         (SimpleNode)node.jjtGetChild(0)
        );
        ok = false;
     }

     // synthetize expression type using the second and third expressions
     // nodes in conditional expression
     if ( ok )
     {
        try
        {
           LType type = ADLTypeCheckHelper.synthetizeTypes(node,1,2);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for ?:. " + e.getMessage(), e.getErrorNode());
        }
     }

     if ( debug )
        System.out.println("debug: (ConditionalExpression) - type "
                           + "(" + node.getTypeName() + ") " 
                           + ((TypedNode)node.jjtGetChild(0)).getTypeName() 
                           + " ? "
                           + ((TypedNode)node.jjtGetChild(1)).getTypeName()
                           + " : "
                           + ((TypedNode)node.jjtGetChild(2)).getTypeName()
                          );
    return data;
  }

  /**
   * Visits an ASTConditionalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions are boolean 
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTConditionalOrExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expressions are boolean 
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.BOOLEAN_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for ||. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visits an ASTConditionalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all child expressions are boolean 
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTConditionalAndExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all child expresions are boolean
     // and set expression type accordingly
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.BOOLEAN_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for &&. " + e.getMessage(), e.getErrorNode());
     }

     return data;
  }

  /**
   * Visits an ASTInclusiveOrExpression .
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>expression children types shall be all booleans or all compatible 
   *     with long. So we just check all children types are compatible with
   *     the type of the first child (either boolean or compatible with long).
   * <li>in case types are ok, just perform type synthesis
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTInclusiveOrExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // expression children types shall be all booleans or all compatible 
     // with long. So we just check all children types are compatible with
     // the type of the first child (either boolean or compatible with long).

     LType type = null;
     TypedNode firstChild = (TypedNode)node.jjtGetChild(0);
     if ( JTypes.booleanT.assignable(firstChild.getType()) )
     {
        type = JTypes.booleanT;
     }
     else if ( JTypes.longT.assignable(firstChild.getType()) )
     {
        type = JTypes.longT;
     }
     else
     {
        errorMessage
        ("Incompatible type for |. Can't convert " 
         + firstChild.getTypeName()
         + " to arithmetic or boolean.",
         firstChild
        );
     }

     // types are ok, just perform type synthesis (synthesis is a consequence 
     // of the call to checkTypesAreAssignable() method)
     if ( type != null )
     {
        try 
        {
           // check compatibility and update type
           type = ADLTypeCheckHelper.checkTypesAreAssignable(node,type);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for |. " + e.getMessage(), e.getErrorNode());
        }
     }

     if ( debug )
     {
        System.out.print("debug: (InclusiveOrExpression) - type "
                          + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           ("| " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }

     return data;
  }
 
  /**
   * Visits an ASTExclusiveOrExpression .
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>expression children types shall be all booleans or all compatible 
   *     with long. So we just check all children types are compatible with
   *     the type of the first child (either boolean or compatible with long).
   * <li>in case types are ok, just perform type synthesis
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTExclusiveOrExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     // expression children types shall be all booleans or all compatible 
     // with long. So we just check all children types are compatible with
     // the type of the first child (either boolean or compatible with long).

     LType type = null;
     TypedNode firstChild = (TypedNode)node.jjtGetChild(0);
     if ( JTypes.booleanT.assignable(firstChild.getType()) )
     {
        type = JTypes.booleanT;
     }
     else if ( JTypes.longT.assignable(firstChild.getType()) )
     {
        type = JTypes.longT;
     }
     else
     {
        errorMessage
        ("Incompatible type for ^. Can't convert " 
         + firstChild.getTypeName()
         + " to arithmetic or boolean.",
         firstChild
        );
     }

     // types are ok, just perform type synthesis (synthesis is a consequence 
     // of the call to checkTypesAreAssignable() method)

     if ( type != null )
     {
        try 
        {
           // check compatibility and update type
           type = ADLTypeCheckHelper.checkTypesAreAssignable(node,type);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for ^. " + e.getMessage(), e.getErrorNode());
        }
     }

     if ( debug )
     {
        System.out.print("debug: (ExclusiveOrExpression) - type "
                          + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           ("| " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     
     return data;
  }

  /**
   * Visits an ASTAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>expression children types shall be all booleans or all compatible 
   *     with long. So we just check all children types are compatible with
   *     the type of the first child (either boolean or compatible with long).
   * <li>in case types are ok, just perform type synthesis
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTAndExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // expression children types shall be all booleans or all compatible 
     // with long. So we just check all children types are compatible with
     // the type of the first child (either boolean or compatible with long).

     LType type = null;
     TypedNode firstChild = (TypedNode)node.jjtGetChild(0);
     if ( JTypes.booleanT.assignable(firstChild.getType()) )
     {
        type = JTypes.booleanT;
     }
     else if ( JTypes.longT.assignable(firstChild.getType()) )
     {
        type = JTypes.longT;
     }
     else
     {
        errorMessage
        ("Incompatible type for &. Can't convert " 
         + firstChild.getTypeName()
         + " to arithmetic or boolean.",
         firstChild
        );
     }

     // types are ok, just perform type synthesis (synthesis is a consequence 
     // of the call to checkTypesAreAssignable() method)

     if ( type != null )
     {
        try 
        {
           // check compatibility and update type
           type = ADLTypeCheckHelper.checkTypesAreAssignable(node,type);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for &. " + e.getMessage(), e.getErrorNode());
        }
     }

     if ( debug )
     {
        System.out.print("debug: (AndExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           ("| " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     
     return data;
  }

  /**
   * Visits an ASTEqualityExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>perform type synthesis from children types
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTEqualityExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     // perform type synthesis from children types
     try 
     {
        LType type = ADLTypeCheckHelper.synthetizeTypes(node);
        node.setType(JTypes.booleanT);
     } 
     catch ( BadTypeException e )
     {
        errorMessage("Incompatible type for == or !=. "    
                     + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.print("debug: (EqualityExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }

     return data;
  }

  /**
   * Visits an ASTInstanceOfExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check first child kind is reference
   * <li>if it is set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTInstanceOfExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check first child kind is reference
     // and if it is set expression kind to boolean
     if ( ((TypedNode)node.jjtGetChild(0)).getTypeKind() 
          != JTypeKind.REFERENCE_TYPE )
     {
        errorMessage
        ("Impossible for "
         + ((TypedNode)node.jjtGetChild(0)).getTypeName() 
         + " to be instance of " 
         + ((SimpleNode)node.jjtGetChild(1)).getName() + ". ",
         (TypedNode)node.jjtGetChild(0)
        );
     }
     else
     {
        node.setType(JTypes.booleanT);
     }
     if ( debug )
     {
        System.out.print("debug: (InstanceofExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        System.out.print
        ( ((TypedNode)node.jjtGetChild(0)).getTypeName()+ " instanceof " 
         +((SimpleNode)node.jjtGetChild(1)).getName() );
       System.out.println("");
    }
    return data;
  }

  /**
   * Visits an ASTRelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all children expressions are arithmetic
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTRelationalExpression node,Object data) 
  {
     node.childrenAccept(this,VISITING_EXPRESSION);

     // Check all children expressions are arithmetic
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(JTypes.booleanT);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for relational operator. " + e.getMessage(),
         e.getErrorNode());
     }

     if ( debug )
     {
        System.out.print("debug: (RelationalExpression) - type "
                        + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }

  /**
   * Visits an ASTShiftExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all children expressions are compatible with long
   * <li>in case types are ok, set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTShiftExpression node,Object data) 
  {
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check all children expressions are compatible with long
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreAssignable
                                        (node,JTypes.longT);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for shift operator. " + e.getMessage(),
         e.getErrorNode());
     }

     if ( debug )
     {
        System.out.print("debug: (ShiftExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getType());

        System.out.println("");
     }

     return data;
  }

  /**
   * Visits an ASTAdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>determine if there is a string among children,
   * <li>if there is a string, check there are only + operators 
   *     (+ may be overloaded for String), and check there are no two 
   *     consecutive int. If checks are ok, expression type is set to String.
   * <li>otherwise, there is no string, normal +/- operator processing is 
   *     performed: just check types are only arithmetic types  and update 
   *     expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTAdditiveExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     boolean ok = true;

     // determine if there is a string among children
     boolean stringHere             = false;
     int     stringIndex            = -1;

     for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
     {
        TypedNode typedNode = (TypedNode)node.jjtGetChild(i);
        LType     type      = typedNode.getType();
        if ( JTypes._String.equals(type) )
        {
           stringHere = true;
           stringIndex = i;
           break;
        }
     }

     if ( stringHere )
     {
        // there is a string, determine if there are only + operators 
        // ( in which case + is overloaded for String)
        try 
        {
           String[] operators = (String[])node.getProperty("operators");
           for ( int i = 0; i < operators.length; i++ )
              if ( ! operators[i].equals("+") )
              {
                 errorMessage
                 ("Incompatible type for " + operators[i]
                  + ". Can't convert " + JTypes._String.getName() 
                  + " to arithmetic.",
                  (SimpleNode)node.jjtGetChild(stringIndex)
                 );
                 ok = false;
                 break;
              }
        }
        catch ( PropertyAccessException e ) { }

        if ( ok )
        {
           // check there are no two consecutive int
           int consecutiveInt = 0;
           for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
           {
              TypedNode typedNode = (TypedNode)node.jjtGetChild(i);
              LType     type      = typedNode.getType();
              if (     type != null
                   &&  JTypeKind.ARITHMETIC_TYPE.equals(type.getTypeKind())
                 )
              {
                 consecutiveInt++;
                 if ( consecutiveInt == 2 ) 
                 { 
                    errorMessage
                    ("Incompatible type for + or -. Can't convert " 
                     + JTypes._String.getName() + " to int.",
                    (SimpleNode)node.jjtGetChild(stringIndex)
                    );
                    ok = false;
                    break;
                 }
              }
              else
              {
                 consecutiveInt = 0;
              }
           }
        }
        // if ok set expression type to string
        if ( ok ) {  node.setType(JTypes._String); }
     }
     else // no String
     {
        // No string => only arithmetic allowed.
        try 
        {
           // check compatibility and update type
           LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                        (node,JTypeKind.ARITHMETIC_TYPE);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for + or -. " + e.getMessage(),
            e.getErrorNode()
           );
        }
     }

     if ( debug )
     {
        System.out.print("debug: (AdditiveExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     } 

     return data;
  }

  /**
   * Visits an ASTMultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTMultiplicativeExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)   
     node.childrenAccept(this,VISITING_EXPRESSION);

     // only arithmetic allowed.
     try 
     {
        // Check compatibility
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                     (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for * or %. " + e.getMessage(),
         e.getErrorNode()
        );
     }

     if ( debug )
     {
        System.out.print("debug: (MultiplicativeExpression) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }

  /**
   * Visits an ASTUnaryPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check children types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTUnaryPlusMinus node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check children types are all arithmetic
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
       String op = "";
       try { op = (String)node.getProperty("op"); }
       catch (PropertyAccessException exc) { }
        errorMessage
        ("Incompatible type for " + op + ". " 
         + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.println("debug: (UnaryPlusMinus) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visits an ASTPreIncrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check children types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTPreIncrementExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check children types are only arithmetic
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for ++. " 
         + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.println("debug: (PreIncrementExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visits an ASTPreDecrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check children types are only arithmetic 
   * <li>update expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTPreDecrementExpression node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     
     // check children types are only arithmetic
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for --. " 
         + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.println("debug: (PreDecrementExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }
     
     return data;
  }

  /**
   * Visits an ASTUnaryNotPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>if op is "~" check children are all arithmetic
   * <li>else if op is "!" check children types are all boolean
   * <li>and perform type synthesis accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTUnaryNotPlusMinus node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     String op = "";
     try { op = (String)node.getProperty("op"); }
     catch (PropertyAccessException e) { }
     // if op is "~" check children are all arithmetic
     // else if op is "!" check children types are all boolean
     // and perform type synthesis accordingly
     if ( op.equals("~") )
     {
        try 
        {
           LType type = ADLTypeCheckHelper.checkTypesAreAssignable
                                           (node,JTypes.longT);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for " + op + ". " 
            + e.getMessage(), e.getErrorNode());
        }
     }
     else // operator == "!"
     {
        try 
        {
           LType type = ADLTypeCheckHelper.checkTypesAreAssignable
                                           (node,JTypes.booleanT);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage
           ("Incompatible type for " + op + ". " 
            + e.getMessage(), e.getErrorNode());
        }
     }

     if ( debug )
     {
        System.out.println("debug: (UnaryNotPlusMinus) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visits an ASTADL_CallStateExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>increase nested call state scope level
   * <li>visit children (indicate children can be expressions) 
   * <li>update expression type according to the child type
   * <li>decrease nested call state scope level
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_CallStateExpression node,Object data) 
  {
     // increase nested call state scope level
     this.callStateScopeLevel++;

     // visit children (indicate children can be expressions) 
     node.childrenAccept(this,VISITING_EXPRESSION);

     // update expression type according to the child type
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());

     if ( debug )
     {
        System.out.print("debug: (ADL_CallStateExpression) - type "
                           + "(" + node.getTypeName() + ") ");
        System.out.print
        (" " + ((TypedNode)node.jjtGetChild(0)).getTypeName());

        System.out.println("");
     }

     // decrease nested call state scope level
     this.callStateScopeLevel--;

     return data;
  }

   /**
   * Visits an ASTCastExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child holding expression to cast
   * <li>get cast type
   * <li>check that both cast type and type of expression to cast are not null
   * <li>check that either cast type is a reference or cast type may be
   *     assigned to the type of the expression to cast.
   * <li>if these checks are ok, set expression type to cast type.
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
 public Object visit(ASTCastExpression node,Object data) 
  {
     // visit child holding expression to cast
     TypedNode origTypeNode = ( node.jjtGetNumChildren() == 1 
                                ? (TypedNode)node.jjtGetChild(0)
                                : (TypedNode)node.jjtGetChild(1)
                              );

     origTypeNode.jjtAccept(this,VISITING_EXPRESSION);

     // get cast type
     LType  castType     = null;
     try
     {
        castType = (LType)node.getProperty("type");
     } catch ( PropertyAccessException e ) { }

     String castTypeName =   (castType != null) 
                           ? castType.getName() 
                           : "<no type>" ;
    
     // get type of expression to cast
     LType origType = origTypeNode.getType();

     // perform some checks : 
     // 1) both cast type and type of expression to cast are not null
     // and 2) either cast type is a reference or cast type may be
     // assigned to the type of the expression to cast.
     // if these checks are ok, set expression type to cast type.

     if ( castType == null )
     {
        errorMessage("Cannot cast to unknown type : " 
                     + castTypeName,node);
     }
     else if ( origType == null )
     {
        errorMessage("Cannot cast from unknown type : " 
                     + origTypeNode.getTypeName(),origTypeNode);
     }
     else if ( ! castType.getTypeKind().equals(origType.getTypeKind()) )
     {
       errorMessage("Invalid cast from " + origType.getName() 
                    + " to " +  castTypeName,
                     node);
     }
     else if (    JTypeKind.REFERENCE_TYPE.equals(castType.getTypeKind() )
                && ! origType.assignable(castType)
             )
     {
        errorMessage("Invalid cast from " + origType.getName() 
                     + " to " +  castTypeName,
                     node);
     }
     else
     {
        node.setType(castType);
     }

     if ( debug )
     {
        System.out.print("debug: (CastExpression) - type "
                          + "(" + castTypeName + ") ");
        System.out.print
        (" " + origTypeNode.getTypeName());

        System.out.println("");
     }
     
     return data;
  }

  /**
   * Visits an ASTNormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Normal node,Object data) 
  {
    node.setType(JTypes.booleanT);
    return data;
  }

  /**
   * Visits an ASTAbnormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Abnormal node,Object data) 
  {
    node.setType(JTypes.booleanT);
    return data;
  }

  /**
   * Visits an ASTADL_Return.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get return type of the tested method
   * <li>check return expression is allowed here
   * <li>update expression type according to return type
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Return node,Object data) 
  {
    // get return type of the tested method
    LExpression e = SimpleNode.stm.getExpression
                    ("return",(Scope)scopeStack.peek());
    if ( e == null || e.getType() == null )
      {
        throw new Error("Cannot find return type of the tested method.");
      }
    // check return expression is allowed here
    checkSymbolIsAllowed(e,node);

    // update expression type according to return type
    node.setType(e.getType());
    
    // process search type info if any
    if ( data instanceof SearchTypeInfo )
    {
      // expression is part of a PrimaryExpression,
      // just add useful information (type of the return expression)
      // to the search type information
       SearchTypeInfo searchTypeInfo = (SearchTypeInfo)data;
       searchTypeInfo.setType(node.getType());
       searchTypeInfo.setASTLastName(node);
    }

    if ( debug )
      {
        System.out.println("debug: (Return) - type "
                           + "(" + node.getTypeName() + ") ");
      }
 
    return data;
  }

  /**
   * Visits an ASTNormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Unchanged node,Object data) 
  {
     // visit children (indicate children can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // set expression type to boolean
     node.setType(JTypes.booleanT);

     return data;
  }

  /**
   * Visits an ASTPrimaryExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children applying a special processing : primary expression 
   *     is constructed with a PrimaryPrefix
   *     and zero or more PrimarySuffix. A search type information object
   *     is passed in between childs. Each child being in charge
   *     of updating the search type information object.
   * <li>process resulting search type information as detailed below :
   * <li>if search type information  contains a type set primary expression 
   *     type to that type
   * <li>if search type information contains no type, get expression
   *     corresponding to that information
   * <li>or if the expression is not found, search for a class
   * <li>update primary expression type accordingly
   * </ul>
   *
   * Note during this processing, the "isClass" property is set.
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTPrimaryExpression node,Object data) 
  {
     // Primary expression is constructed with a PrimaryPrefix
     // and zero or more PrimarySuffix. A search type information object
     // is passed in between childs. Each child being in charge
     // of updating the search type information object.
     SearchTypeInfo searchTypeInfo = new SearchTypeInfo
                                         (  null,
                                            "",
                                            (Scope)scopeStack.peek(),
                                            false
                                         );
     int numChild = node.jjtGetNumChildren();
     for (int i = 0; i < numChild ; i++)
     { 
        searchTypeInfo = (SearchTypeInfo) 
                         node.jjtGetChild(i).jjtAccept(this,searchTypeInfo);
     }
     // Process resulting search type information :
     // (1) if search type information  contains a type set primary expression 
     // type to that type
     // (2) if search type information contains no type, get expression
     //     corresponding to that information
     // (3) or if the expression is not found, search for a class
     // (4) update primary expression type accordingly
       
     if ( searchTypeInfo.getType() != null )
     {
        // If search type information  contains a type set primary expression 
        // type to that type
        node.setType(searchTypeInfo.getType());
        try 
        {
           // Update "isClass" property in node. It is set to false
           // since expression is not a class.
           node.setProperty("isClass",new Boolean(false)); 
        }
        catch ( PropertyAccessException e ) { }
     }
     else
     {
        // If search type information contains no type, get expression
        // corresponding to that information and update type accordingly
        LExpression lExpr = SimpleNode.stm.getExpression
                            (searchTypeInfo.getScopedName(),
                             searchTypeInfo.getScope());
        if ( lExpr != null )
        {
           checkSymbolIsAllowed(lExpr,node);
           node.setType(lExpr.getType());
//            decorateASTName(searchTypeInfo.getASTName(),lExpr);
           decorateASTName(searchTypeInfo.getASTLastName(),lExpr);
           try 
           {
              // Update "isClass" property in node. It is set to false
              // since expression is not a class.
              node.setProperty("isClass",new Boolean(false)); 
           }
           catch ( PropertyAccessException e ) { }
        }
        else
        {
           // If expression is not found, dearch for a class, and update type
           // accordingly
           LClass lClass = SimpleNode.stm.getClass
                           (searchTypeInfo.getScopedName(),
                            searchTypeInfo.getScope());
           if ( lClass != null )
           {
              checkSymbolIsAllowed(lClass,node);
              node.setType(lClass);
//               decorateASTName(searchTypeInfo.getASTName(),lClass);
              decorateASTName(searchTypeInfo.getASTLastName(),lClass);
              try 
              {
                 // Update "isClass" property in node. It is set to true
                 // since expression is a class.
                 node.setProperty("isClass",new Boolean(true)); 
              }
              catch ( PropertyAccessException e ) { }
           }
           else
           {
              errorMessage
              ("Undefined expression " + searchTypeInfo.getScopedName(),node);
           }
        }
     }
     if ( debug )
       System.out.println("debug: ADLTypeCheckVisitor(ASTPrimaryExpression) - "
                          + " found expression type = " 
                          + (   node.getType() != null 
                              ? node.getTypeName() : null ) );
     return data;
  }

  /**
   * Visits an ASTThisPrefix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>processing depends on the expression being part of a PrimaryExpression
   *     or being a stand alone expression :
   * <li>if expression is part of a PrimaryExpression,
   *     just add useful information to the search type information
   *     that is the "this" prefix to the scoped name
   * <li>if it is a stand alone expression is alone, just find and set its type
   * </ul>
   *
   * @param node the node to visit
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data sis updated during the visit,
   *             otherwise data is unused.
   * @return received data object, it has been updated if it is a 
   *         SearchTypeInfo object.
   *
   */
  public Object visit(ASTThisPrefix node,Object data) 
  {
    // processing depends on the expression being part of a PrimaryExpression
    // or being a stand alone expression
    SearchTypeInfo searchTypeInfo;
    if ( data instanceof SearchTypeInfo )
    {
       // expression is part of a PrimaryExpression,
       // just add useful information to the search type information
       // that is the "this" prefix to the scoped name.
       searchTypeInfo = (SearchTypeInfo)data;
       searchTypeInfo.setScopedName(node.getName());
       searchTypeInfo.setASTLastName(node);
       return searchTypeInfo;
    }
    else
    {
       // stand alone expression is alone, just find and set its type
       LExpression lExpression = SimpleNode.stm.getExpression
                                 (node.getName(),(Scope)scopeStack.peek());
       if ( lExpression != null ) 
       {
          checkSymbolIsAllowed(lExpression,node);
          node.setType(lExpression.getType());
       }
       else 
       { 
           errorMessage
           ("Undefined expression " + node.getName() ,node);
       }
       return data;
    }
  }

  /**
   * Visits an ASTSuperPrefix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>processing depends on the expression being part of a PrimaryExpression
   *     or being a stand alone expression :
   * <li>if expression is part of a PrimaryExpression,
   *     just add useful information to the search type information
   *     that is the "this" prefix to the scoped name
   * <li>if it is a stand alone expression is alone, just find and set its type
   * </ul>
   *
   * @param node the node to visit
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data sis updated during the visit,
   *             otherwise data is unused.
   * @return received data object, it has been updated if it is a 
   *         SearchTypeInfo object.
   *
   */
  public Object visit(ASTSuperPrefix node,Object data) 
  {
    // processing depends on the expression being part of a PrimaryExpression
    // or being a stand alone expression
    SearchTypeInfo searchTypeInfo;
    if ( data instanceof SearchTypeInfo )
    {
       // expression is part of a PrimaryExpression,
       // just add useful information to the search type information
       // that is the "this" prefix to the scoped name.
       searchTypeInfo = (SearchTypeInfo)data;
       searchTypeInfo.setScopedName(node.getName());
       searchTypeInfo.setASTLastName(node);
       return searchTypeInfo;
    }
    else
    {
       // stand alone expression, just find and set its type now.
       LExpression lExpression = SimpleNode.stm.getExpression
                                 (node.getName(),(Scope)scopeStack.peek());
       if ( lExpression != null ) 
       {
          checkSymbolIsAllowed(lExpression,node);
          node.setType(lExpression.getType());
       }
       else 
       { 
           errorMessage
           ("Undefined expression " + node.getName() ,node);
       }
       return data;
    }
  }


  /**
   * Visits an ASTADL_SuperSemantics.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>search in adl superclass for semantics
   * <li>if needed recursively search in adl superclasses (forcing
   *     parsing if necessary)
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */

  public Object visit(ASTADL_SuperSemantics node,Object data) 
  {
     // Check an adl super class contains tested method

     boolean isConstructor = _visitedADLMethod.isConstructorSemantics();
     ADLClass adlCalledSuperClass = _adlSuperClass;
     ADLMethod adlMethod = (ADLMethod)SimpleNode.stm.getADLMethod
                           (isConstructor ? adlCalledSuperClass.getName()
                                          : _visitedADLMethod.getName(),
                            _visitedADLMethod.getOperationComparator(),
                            adlCalledSuperClass,
                            true);

     while ( adlMethod == null && adlCalledSuperClass != null)
     {
        // Try to find superclass (if necessary parse adl file).

        String superClassName = adlCalledSuperClass.getSuperClassName();        

        if ( superClassName == null )
        {
           adlCalledSuperClass = null;
        }
        else
        {
           // First, check in symbol table

           adlCalledSuperClass = SimpleNode.stm.getADLClass(superClassName,null);

           if ( adlCalledSuperClass == null )
           {
              // Symbol table does not contain super class, force parsing.

              String fileName = SimpleNode.stm.getFileNameForADLClass
                                (superClassName);

              if ( fileName != null )
              {
                 ParseUnit parseUnit = ParseUnit.getParseUnit(fileName, nlengine);
                 if ( parseUnit != null )
                 {
                    parseUnit.pureParse();
                 }
              }

              // ... now symbol should be in symbol table.

              adlCalledSuperClass = SimpleNode.stm.getADLClass(superClassName,null);
           }

           // Search for the method.

           if ( adlCalledSuperClass != null )
           {
              adlMethod = (ADLMethod)SimpleNode.stm.getADLMethod
                          (isConstructor ? adlCalledSuperClass.getName()
                                         : _visitedADLMethod.getName(),
                           _visitedADLMethod.getOperationComparator(),
                           adlCalledSuperClass,
                           true);
           }
        }
     }
     
     if ( adlMethod == null )
     {
        errorMessage(  "Super.semantics cannot be used: no semantics defined"
                     + " for " 
                     + _visitedADLMethod.getSignature()
                     + " in adl super class " + _adlSuperClass.getName(),
                     node);
     }

     if ( isConstructor && adlMethod != null )
     {
        try { node.setProperty("definingADLClass",adlCalledSuperClass.getName()); }
        catch ( PropertyAccessException e ) { }
     }

     return data;
  }
  
  /**
   * Visits an ASTParenthExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child expression
   * <li>set type to that of child
   * <li>if expression is part of a PrimaryExpression,
   *     just add useful information to the search type information object
   *     received : that is the type of the child
   * <li>if it is a stand alone expression is alone, just find and set its type
   * </ul>
   *
   * @param node the node to visit
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data sis updated during the visit,
   *             otherwise data is unused.
   * @return received data object, that has been updated if it is a 
   *         SearchTypeInfo object.
   *
   */
  public Object visit(ASTParenthExpression node,Object data) 
  {
     // visit child expression
     node.childrenAccept(this,data);

     if (debug) 
        System.out.println
        ("debug: ADLTypeCheckVisitor(ASTParenthExpression) - child class is " 
         + node.jjtGetChild(0).getClass()
        );

     // type set to type of expression in parenthesis.
     node.setType(((TypedNode)node.jjtGetChild(0)).getType());

     // in case data is instance of search type info
     // update search type information
     if ( data instanceof SearchTypeInfo )
     {
        SearchTypeInfo sti = (SearchTypeInfo)data;
        //        sti.setASTLastName(node);
        sti.setType(node.getType());
     }

     return data;
  }

  /**
   * Visits an ASTBracketSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child expression
   * <li>get type of expression in bracket
   * <li>check expression inside brackets is of integer arithmetic type
   * <li>check expression before brackets is of array type
   * <li>update received search type information object
   * </ul>
   *
   * @param node the node to visit
   * @param data SearchTypeInfo object, updated during the visit.
   * @return updated data object.
   *
   */
  public Object visit(ASTBracketSuffix node,Object data) 
  {
     // visit child expressions
    node.childrenAccept(this,VISITING_EXPRESSION);

    // get type of expression in bracket
    TypedNode nodeInBracket = (TypedNode)node.jjtGetChild(0);

    // check expression inside brackets is of integer arithmetic type
    if ( ! JTypeKind.ARITHMETIC_TYPE.equals(nodeInBracket.getTypeKind()) )
    {
       errorMessage(  "Incompatible type for []. Cannot convert "
                    + nodeInBracket.getTypeName()
                    + " to integer",node);
    }
    else if ( ! JTypes.longT.assignable(nodeInBracket.getType()) )
    {
       errorMessage(    "Incompatible type for []. Explicit cast needed "
                      + " to convert "
                      + nodeInBracket.getTypeName()
                      + " to integer",node);
    } 

    // check expression before brackets is of array type
    SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
    LExpression lExpression = 
                SimpleNode.stm.getExpression(searchTypeInfo.getScopedName(),
                                             searchTypeInfo.getScope()
                                            );
    if ( lExpression != null )
    {
       checkSymbolIsAllowed(lExpression,node);
       LType lType = lExpression.getType();
       if ( lType instanceof JArray) {
//           decorateASTName(searchTypeInfo.getASTName(),lExpression);
          decorateASTName(searchTypeInfo.getASTLastName(),lExpression);
          searchTypeInfo.setType(((JArray)lType).getElementType());
       }
       else
       {
          errorMessage(   "Expression " + searchTypeInfo.getScopedName()
                       + " is not an array : cannot apply brackets.",node);
       }
    }
    else
    {
       errorMessage
       ("Undefined expression " + searchTypeInfo.getScopedName() ,node);
    }
    return searchTypeInfo;
  }

  /**
   * Visits an ASTDotSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get SearchTypeInfo object
   * <li>update scoped name in SearchTypeInfo object
   * </ul>
   *
   * @param node the node to visit
   * @param data SearchTypeInfo object, updated during the visit.
   * @return updated data object.
   *
   */
  public Object visit(ASTDotSuffix node,Object data) 
  {
    // update scoped name
    SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
    String newName = searchTypeInfo.getScopedName();
    if ( newName == null )
    {
       newName = "";
    }
    else if ( newName.length() > 0 )
    {
       newName += "." ;
    }
    newName += node.getName();
    searchTypeInfo.setScopedName(newName);
    searchTypeInfo.setASTLastName(node);
    return data;
  }

  /**
   * Visits an ASTIntegerLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTIntegerLiteral node,Object data) 
  {
     // Just set type
    try {
      node.setType( ((String)node.getProperty("typeName")).equals("int") ?
                    JTypes.intT : JTypes.longT );
    }
    catch (PropertyAccessException e) {}
    return data;
  }

  /**
   * Visits an ASTFloatLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTFloatLiteral node,Object data) 
  {
     // Just set type
    try {
      node.setType( ((String)node.getProperty("typeName")).equals("float") ?
                    JTypes.floatT : JTypes.doubleT );
    }
    catch (PropertyAccessException e) {}
    return data;
  }

  /**
   * Visits an ASTCharacterLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTCharacterLiteral node,Object data) 
  {
     // Just set type
     node.setType(JTypes.charT); 
     return data;
  }

  /**
   * Visits an ASTStringLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>update search type info if required
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTStringLiteral node,Object data) 
  {
     // Update search type info if required
     if ( data instanceof SearchTypeInfo )
     {
        ((SearchTypeInfo)data).setType(JTypes._String);
     }
     // Set set type
     node.setType(JTypes._String);
     return data;
  }

  /**
   * Visits an ASTBooleanLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTBooleanLiteral node,Object data) 
  {
     // Just set type
     node.setType(JTypes.booleanT); 
     return data;
  }

  /**
   * Visits an ASTNullLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set type to that of the literal
   * </ul>
   *
   * @param node the node to visit
   * @param unused data 
   * @return received data
   *
   */
  public Object visit(ASTNullLiteral node,Object data) 
  {
     // Just set to null type
     node.setType(JTypes._null);
     return data;
  }

  /**
   * Visits an ASTArguments.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children. If ASTArguments is part of a PrimaryExpression 
   *     (data instanceof SearchTypeInfo) then indicates children may be
   *      expressions
   * <li>if ASTArguments is part of a PrimaryExpression, visited arguments
   *     together with preceding scoped names are a call to an operation
   *     (method, constructor...). Processing is as follows :
   * <li>retrieve useful information from received SearchTypeInfo object
   * <li>constructs an operation comparator from the arguments in
   *     the visited expression
   * <li>if operation is a constructor, retrieve the corresponding
   *     class (useful to set the search scope for the constructor)
   * <li>search for the operation
   * <li>update search type info object accordingly (type set to
   *     constructor's class for a constructor, or to operation
   *     return type for other operations
   * </ul>
   *
   * @param node the node to visit
   * @param data SearchTypeInfo object, updated during the visit.
   * @return updated data object.
   *
   */

  public Object visit(ASTArguments node,Object data) 
  {
     // visit children. If ASTArguments is part of a PrimaryExpression 
     // (data instanceof SearchTypeInfo) then indicate children may be 
     // expressions
     if ( data instanceof SearchTypeInfo )
     {
        node.childrenAccept(this,VISITING_EXPRESSION);
     }
     else
     {
        node.childrenAccept(this,data);
     }

     // if ASTArguments is part of a PrimaryExpression, visited arguments
     // together with preceding scoped names are a call to an operation
     // (method, constructor...)
     // Processing is as follows :
     // (1) retrieve useful information from received SearchTypeInfo object
     // (2) constructs an operation comparator from the arguments in
     //     the visited expression
     // (3) if searched operation is a constructor, first, retrieve the 
     //     constructor's class and search for the constructor in that class
     //     scopecorresponding
     // (4) if searched operation is not a constructor, perform a normal
     //     search
     // (5) update search type info object accordingly (type set to
     //     constructor's class for a constructor, or to operation
     //     return type for other operations

     if ( data instanceof SearchTypeInfo )
     {
        SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
        String scopedName = searchTypeInfo.getScopedName();
        Scope  scope      = searchTypeInfo.getScope();
        boolean searchConstructor = searchTypeInfo.getSearchConstructor();
        LCallableAssignComparator comp = new LCallableAssignComparator();

        // in case an argument list is present - get argument types
        // to construct operation comparator
        if ( node.jjtGetNumChildren() == 1 )
        { 
           Node argumentList = node.jjtGetChild(0);
           for (int i = 0; i < argumentList.jjtGetNumChildren() ; i++)
           {
              Node argExpression = argumentList.jjtGetChild(i);
              if ( argExpression instanceof TypedNode )
              {
                 LType lType = ((TypedNode)argExpression).getType();
                 if ( debug )
                    System.out.println
                    ("debug: ADLTypeCheckVisitor(ASTArguments) - "
                     + " adding param type = " 
                     + ( lType != null ? lType.getName() : null )
                    );
                 comp.addParameterType(lType);
              }
              else
              {
                 throw new Error
                       ("Child in Arguments is not a subclass of TypedNode");
              }
           }
        }
        if ( debug )
          System.out.println("debug: ADLTypeCheckVisitor(ASTArguments) - "
                             + " search operation = " + scopedName);
        
        LClass lClass = null;
        LCallable ope = null;
        // either search for a constructor or for a normal method
        if ( searchConstructor )
        {
           // get class
           lClass = SimpleNode.stm.getClass(scopedName,scope);
           // get the last identifier in scoped name
           String name = scopedName.substring(scopedName.lastIndexOf(".") + 1);
           // and search for the constructor in class scope
           ope = SimpleNode.stm.getMethod(name,comp,null,(Scope)lClass);
        }
        else
        {
           // search for the operation
           // TBC - update this when direct lookup of operation in
           //       ADLClass will be setup
           ope = SimpleNode.stm.getMethod(scopedName,comp,null,scope);
        }
        if ( ope != null )
        {
           checkSymbolIsAllowed(ope,node);
//            decorateASTName(searchTypeInfo.getASTName(),ope);
           decorateASTName(searchTypeInfo.getASTLastName(),ope);
           
           // update search type info object accordingly (type set to
           // constructor's class for a constructor, or to operation
           // return type for other operations
           if ( searchConstructor )
           {
              searchTypeInfo.setType(lClass);
           }
           else
           {
              searchTypeInfo.setType(ope.getReturnType());
           }
        }
        else
        {
           errorMessage
           ("Public operation " + scopedName + " not found." ,node);
        }
        if ( debug )
          System.out.println("debug: ADLTypeCheckVisitor(ASTArguments) - "
                          + " found operation = " + ope);

        return searchTypeInfo ;
     }
     else
     {
        // Visit is finished - just return data
        return data;
     }
  }

  /**
   * Visits an ASTArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children expressions.
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTArgumentList node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTAllocationExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get allocated type name 
   * <li>get allocated type
   * <li>check type is not null
   * <li>depending on second child kind, perform the following :
   * <li>if it's an ASTArrayDimension, just set type to array
   * <li>if it's an ASTArguments : expression is a constructor call - check a 
   *     constructor with that signature exists, and set type to the class
   *     of the constructor
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTAllocationExpression node,Object data) 
  {
     // get type name 
    String typeName = null;
    try { typeName = (String)node.getProperty("typeName"); }
    catch(PropertyAccessException e) {}
     
     // get the type
     LType lType = SimpleNode.stm.getType(typeName,(Scope)scopeStack.peek());

     // check type is not null
     if ( lType == null )
     {
        // type was not found. Report error.
        errorMessage
        (    "Cannot create new instance of " + typeName + ". " 
           + "Class " + typeName + " not found."
           ,(SimpleNode)node.jjtGetChild(0)
        );
     }
     else
     {

        // depending on second child kind, perform the following :
        // 1. if it's an ASTArrayDimension, just set type to array
        // 2. if it's an arguments : expression is a constructor call - check 
        //    a constructor with that signature exists, and set type to 
        //    the found type
        SimpleNode second = (SimpleNode)node.jjtGetChild(node.jjtGetNumChildren()-1);
        if ( second instanceof ASTArrayDimensions )
        {
           // it's an ASTArrayDimension, get the array dimension and
           // set expression type accordingly
           try 
           {
              int arrayDim = 0;
              arrayDim = ((Integer)second.getProperty("arrayLevel"))
                         .intValue();
              while ( arrayDim-- > 0 ) 
              { lType = SimpleNode.sbf.newArray(lType); }
              node.setType(lType);
           }
           catch ( PropertyAccessException e ) { }
        }
        else if ( second instanceof ASTArguments )
        {
           // if it's an arguments : expression is a constructor call
           
           if ( ! ( lType instanceof LClass ) )
           {
              errorMessage
              ("Invalid argument type " + typeName + " for new.",node);
           }
           else
           {
              // check the constructor exists, this is achieved by initializing
              // a search type information object and visiting second child
              // (arguments) with this serch type information object
              SearchTypeInfo searchTypeInfo = new SearchTypeInfo
                                                  (  null,
                                                     typeName,
                                                     (Scope)lType,
                                                     true
                                                  );
              searchTypeInfo.setASTLastName(node);
              searchTypeInfo = (SearchTypeInfo) 
                               second.jjtAccept(this,searchTypeInfo);

              // process search type information
              if ( searchTypeInfo.getType() == null )
              {
                 // constructor was not found. Report error.
                 errorMessage
                 (    "No public constructor with the supplied signature found"
                    + " for class " + typeName 
                    + ". This may be because of a wrong number of arguments "
                    + ", incompatible types or a non-public access.",second
                 );
              }
              else
              {
                 // set expression type to that of the constructor class
                 node.setType(lType);
              }
           }
        }
     }
     if ( debug )
        System.out.println("debug: (AllocationExpression) - "
                          + " type = " + node.getTypeName());

     if ( data instanceof SearchTypeInfo )
     {
        ((SearchTypeInfo)data).setASTLastName(node);
     }
     return data;
  }

  /**
   * Visits an ASTArrayDimensions.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTArrayDimensions node,Object data) 
  {
    node.childrenAccept(this,VISITING_EXPRESSION);
    return data;
  }

  /**
   * Visits an ASTADL_ThrownExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>since type was checked during first pass (parsing), only
   *     set expression type to boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_ThrownExpression node,Object data) 
  {
     node.setType(JTypes.booleanT);
     return data;
  }

  /**
   * Visits an ASTADL_DomainList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_DomainList node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTADL_Domain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTADL_Domain node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTVariableDeclaratorId.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTVariableDeclaratorId node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTMethodDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTMethodDeclarator node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTFormalParameters.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTFormalParameters node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTFormalParameter.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTFormalParameter node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTType.
   *
   * <p><b>Algorithm:</b> nothing specific to do here.
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTType node,Object data) 
  {
    return data;
  }

  /**
   * Visits an ASTResultType.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTResultType node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTName.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>processing depends on the name being part of a PrimeryExpression
   *     or being a stand alone expression or dataset expression
   * <li>if name is part of a PrimaryExpression, just update the name
   *     in the received search type information object
   * <li>if name is stand alone and is an expression, find the expression
   *     and set node's type accordingly
   * <li>if name is stand alone and is a dataset expression, find the 
   *     expression and set node's type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data data that may be a SearchTypeInfo object,
   *             if it is it means the expression is part of a 
   *             primary expression and data sis updated during the visit,
   *             otherwise data is unused.
   * @return received data object, that has been updated if it is a 
   *         SearchTypeInfo object.
   *
   */
  public Object visit(ASTName node,Object data) 
  {
     // processing depends on the name being part of a PrimeryExpression
     // or being a stand alone expression or dataset expression
     if ( data instanceof SearchTypeInfo )
     {
        // name is part of a PrimaryExpression just update the name in the 
        // received search type information object
        SearchTypeInfo searchTypeInfo =(SearchTypeInfo)data ;
        searchTypeInfo.setScopedName(node.getName());
        // store AST name in SearchTypeInfo for later decoration
//         searchTypeInfo.setASTName(node);
        searchTypeInfo.setASTLastName(node);
        return data;
     }
     else if ( data instanceof Integer && data.equals(VISITING_EXPRESSION) )
     {
        // stand alone expression name, just find it and set expression type
        // accordingly
        LExpression lExpression = SimpleNode.stm.getExpression
                                  (node.getName(),(Scope)scopeStack.peek());
        if ( lExpression != null ) 
        {
           checkSymbolIsAllowed(lExpression,node);
           if ( debug )
              System.out.println
              ("debug: ADLTypeCheckVisitor(ASTName) - "
               + " setting type = " + ( lExpression.getType() != null 
                                        ? lExpression.getType().getName() 
                                        : null
                                      )
              );
           node.setType(lExpression.getType());
           decorateASTName(node,lExpression);
        }
        else 
        { 
           errorMessage
           ("Undefined expression " + node.getName() ,node);
        }
        return null;
     }
     else if ( data instanceof Integer && data.equals(VISITING_DATASET_EXPR) )
     {
        // dataset expression, just find it and set expression type
        // accordingly
        LExpression lExpression = SimpleNode.stm.getExpression
                                  (node.getName(),(Scope)scopeStack.peek());
        // directly check symbol is allowed (no need to call for 
        // checkSymbolIsAllowed)
        if ( lExpression == null )
        {
           errorMessage
           ("Undefined dataset " + node.getName() + ". " ,node);
        }
        else
        {
           decorateASTName(node,lExpression);
           node.setType(lExpression.getType());
        }
        return null;
     }
     
     return data;
  }

  /**
   * Visits an ASTNameList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTNameList node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTTDD_DatasetDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get dataset type and name form node properties
   * <li>indicate a dataset is being defined
   * <li>visit dataset initializer
   * <li>mark end of dataset definition
   * <li>check for assignability of initializer to defined dataset
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetDeclaration node,Object data) 
  {
     // get dataset type and name form node properties
     LType datasetType = null;
     String datasetName = null;
     try
     {
        datasetType = (LType)node.getProperty("datasetType");
        datasetName = (String)node.getProperty("datasetName");
     }
     catch ( PropertyAccessException e ) { }

     // indicate a dataset is being defined
     currentDataset = node;
     currentDatasetName = datasetName;

     // visit dataset initializer
     TypedNode initNode = (TypedNode)node.jjtGetChild(1);
     initNode.jjtAccept(this,VISITING_EXPRESSION);

     // mark end of dataset definition
     currentDataset = null;
     currentDatasetName = null;
        
     // check for assignability of initializer to defined dataset
     if ( initNode.getType() == null )
     {
        errorMessage("Type check error in initializer for dataset"
		     + datasetName + ".",
		     (SimpleNode)node.jjtGetChild(0));
     }
     else if ( ! datasetType.assignable(initNode.getType()) )
     { 
        errorMessage
        (  "Incompatible type for dataset initializer. Can't convert "
         + initNode.getTypeName()
         + " to " 
         + datasetType.getName(),initNode
        );
     }
     return data;
  }

  /**
   * Visits an ASTTDD_FactoryDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit factory body
   * <li>if necessary visit factory relinquish clause
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_FactoryDeclaration node,Object data) 
  {
     // visit factory body
     node.jjtGetChild(0).jjtAccept(this,data);

     // if necessary visit factory relinquish clause
     if ( node.jjtGetNumChildren() > 1 )
     {
        scopeStack.push(node.getScope());  // enter relinquish clause scope
        for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
        {   
           ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
        }
        scopeStack.pop();                  // exit relinquish clause scope
     }

     return data;
  }

  /**
   * Visits an ASTTDD_TestDirective.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>indicate test directive visit
   * <li>set test directive scope
   * <li>visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_TestDirective node,Object data) 
  {
     this.inTestDirective = true;
     scopeStack.push(node.getScope());  // enter scope 
     node.childrenAccept(this,data);
     scopeStack.pop();                  // exit scope
     this.inTestDirective = false;
     return data;
  }

  /**
   * Visits an ASTTDD_DatasetDomain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>type check is performed only if the node's got an initializer :
   * <li>get dataset type
   * <li>visit initializer
   * <li>get initializer type
   * <li>check initializer can be assigned to the dataset
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetDomain node,Object data) 
  {
     // perform type check only if the node's got an initializer
    if ( node.jjtGetNumChildren() > 1 )
      {
        // there's an initializer, get it's type
        LType datasetType = null;
        try
        {
           datasetType = (LType)node.getProperty("datasetType");
        }
        catch ( PropertyAccessException e ) { }


        // visit initializer
        TypedNode initNode = (TypedNode)node.jjtGetChild(1);
        initNode.jjtAccept(this,VISITING_EXPRESSION);
        
        if ( debug )
        {
           System.out.println("debug: DatasetDomain - initType = " 
                              + initNode.getType());
        }

        // check initializer can be assigned to the dataset
        if ( initNode.getType() == null )
        {
           errorMessage(  "Dataset " 
                        + ((SimpleNode)node.jjtGetChild(0).jjtGetChild(1)).getName() 
                        + " could not be initialized.", initNode  
                       );
        }
        else if ( ! datasetType.assignable(initNode.getType()) )
        { 
           errorMessage
           (  "Incompatible type for dataset initializer. Can't convert "
            + initNode.getTypeName()
            + " to " 
            + datasetType.getName(),initNode
           );
        }
      }
    else {
      node.jjtGetChild(0).jjtAccept(this,VISITING_EXPRESSION);
    }
     return data;
  }

  /**
   * Visits an ASTTDD_DatasetLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform children types synthesis
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetLiteral node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);
     // perform children types synthesis and set expression type accordingly
     try
     {
        LType type = ADLTypeCheckHelper.synthetizeTypes(node);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage("Incompatible types in dataset literal. "
                     + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.print("debug: (DatasetLiteral) - type "
                           + "(" + node.getType() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());

        System.out.println("");
     }
     return data;
  }

 /**
  * Visits an ASTTDD_FactoryCall.
  *
  * <p><b>Algorithm:</b><ul>
  * <li>visit chilren (except child 0 which is a name)
  * <li>get factory name :  child 0
  * <li>get factory arguments : child 1 and following
  * <li>get the factory symbol and set node type to the factory's returned 
  *     type
  * </ul>
  *
  * @param node the node to visit
  * @param data unused data.
  * @return received data object.
  *
  */
  public Object visit(ASTTDD_FactoryCall node,Object data) 
  {
    // visit chilren (except child 0 which is a name)
    for ( int i = 1; i < node.jjtGetNumChildren(); i++)
    {
       ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
    }

    // get factory name :  child 0
    String factoryName = ((SimpleNode)node.jjtGetChild(0)).getName();
    LCallableAssignComparator comp = new LCallableAssignComparator();

    // get factory arguments : child 1 and following
    for ( int i = 1; i < node.jjtGetNumChildren(); i++ ) 
    {
       TypedNode pNode = (TypedNode)node.jjtGetChild(i);
       comp.addParameterType(pNode.getType());
    }

    // get the factory symbol and set node type to the factory's returned type
    TDDFactory factory = SimpleNode.stm.getTDDFactory
                         (factoryName,comp,(Scope)scopeStack.peek());
    if ( factory == null )
    {
        errorMessage("Undefined factory " + factoryName + ".",node);
    }
    else
    {
       // set node type to return type of returned factory
       node.setType(factory.getReturnType());
       // save ASTName kind
       decorateASTName((ASTName)node.jjtGetChild(0),factory);
    }

    return data;
  }

 /**
  * Visits an ASTTDD_DatasetSingleton.
  *
  * <p><b>Algorithm:</b><ul>
  * <li>visit chilren (except child 0 which is a name)
  * <li>set expression type according to first child type
  * <li>if a dataset is currently being defined and this dataset singleton
  *     is a name, check for circular dependency into dataset definition,
  *     that is indicate new dependency (using the CycleDetector) and check
  *     it does not detects a cycle among dataset dependencies
  * </ul>
  *
  * @param node the node to visit
  * @param data unused data.
  * @return received data object.
  *
  */
  public Object visit(ASTTDD_DatasetSingleton node,Object data) 
  {
     // visit child (indicate a dataset expression is visited)
     node.childrenAccept(this,VISITING_DATASET_EXPR);
     
     // set expression type according to child type
     TypedNode child = (TypedNode)node.jjtGetChild(0);
     node.setType( child.getType() );

     // if a dataset is currently being defined and this dataset singleton
     // is a name, check for circular dependency into dataset definition:
     if (    (currentDataset != null)
          && (child instanceof ASTName)
        )
     {
        // indicate new dependency (using the CycleDetector) and check
        // it does not detects a cycle among dataset dependencies
        try 
        {
           String childName = (String)child.getProperty("name");
           String childTDDClassName = (String)child.getProperty("enclosingConstructName");
           cycleDetector.addDependency(sourceName + "." + currentDatasetName,
                                       childTDDClassName + "." + childName);
        }
        catch (PropertyAccessException e) {}
        catch (CycleException e)
        {
           Enumeration cycle = e.getCycle().elements();
           String s = "";
           while ( cycle.hasMoreElements() )
           { s += cycle.nextElement() + " "; }
           errorMessage(  "Cycle detected for dataset " + currentDatasetName
                          + ": " + s,
                          currentDataset
                       );
        }
     }
     return data;
  }

  /**
   * Visits an ASTTDD_SingleDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_SingleDeclarator node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTTDD_DatasetMember.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children (indicating they can be expresions)
   * <li>in case there's a range, check boundaries are integer arithmetics
   * <li>set member type to that obtained from child type synthesis
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetMember node,Object data) 
  {
     // just visit children (indicating they can be expresions)
     node.childrenAccept(this,VISITING_EXPRESSION);
    
     // in case there's a range, check boundaries are integer arithmetics
     if ( node.jjtGetNumChildren() > 1 ) 
     {
        try
        { 
           LType type = ADLTypeCheckHelper.checkTypesAreAssignable
                        (node,JTypes.longT);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage("Only integer types allowed in range. " 
                        + e.getMessage(), e.getErrorNode());
        }
     }
     else
     {
        node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     }
     if ( debug )
     {
        System.out.print("debug: (DatasetMember) - type "
                          + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }

  /**
   * Visits an ASTTDD_DatasetExpr.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform children types synthesis
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetExpr node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // perform children types synthesis
     // and set expression type accordingly
     if ( node.jjtGetNumChildren() > 1 )
     {
        // several children
        try
        { 
           LType type = ADLTypeCheckHelper.synthetizeTypes(node);
           node.setType(type);
        }
        catch ( BadTypeException e )
        {
           errorMessage("Incompatible type for dataset concatenation. " 
                        +  e.getMessage(), e.getErrorNode());
        }
     }
     else
     {
        // one child
        node.setType(((TypedNode)node.jjtGetChild(0)).getType());
     }

     if ( debug )
     {
        System.out.print("debug: (DatasetExpr) - type "
                         + "(" + node.getTypeName() + ") ");
        for (int i = 0; i < node.jjtGetNumChildren(); i++)
           System.out.print
           (" " + ((TypedNode)node.jjtGetChild(i)).getTypeName());
        System.out.println("");
     }
     return data;
  }
    
  /**
   * Visits an ASTTDD_DatasetConcatenationExpr.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child (indicate it can be an expression)
   * <li>set expression type according to child type
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_DatasetConcatenationExpr node,Object data) 
  {
     node.childrenAccept(this,VISITING_EXPRESSION);
     node.setType( ((TypedNode)node.jjtGetChild(0)).getType() );
     if ( debug )
     {
        System.out.print("debug: (DatasetConcatenationExpr) - type "
                         + "(" + node.getTypeName() + ") ");
        System.out.println("");
     }

     return data;
  }

  /**
   * Visits an ASTTDD_ClassDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>enter class scope
   * <li>visit children
   * <li>exit class scope
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_ClassDeclaration node,Object data) 
  {
     scopeStack.push(node.getScope());  // Enter scope
     node.childrenAccept(this,data);
     scopeStack.pop();                  // Exit scope
     return data;
  }

  /**
   * Visits an ASTTDD_ClassBodyDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_ClassBodyDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTTDD_FieldDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>retrieve field types property
   * <li>visit each initializer and check its type is assignable
   *     to the declared field
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTDD_FieldDeclaration node,Object data) 
  {
     LType[] fieldTypes = null;

     // retrieve field types property
     try { fieldTypes = (LType[])node.getProperty("fieldTypes"); }
     catch ( PropertyAccessException e ) {}

     // visit initializers with the field types
     for ( int i = 2, j = 0; i < node.jjtGetNumChildren(); i = i + 2, j++ )
     {
        node.jjtGetChild(i).jjtAccept(this,fieldTypes[j]);
     }
     
     return data;
  }

  /**
   * Visits an ASTClassDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTClassDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTClassBodyDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTClassBodyDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTInterfaceDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTInterfaceDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTInterfaceMemberDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTInterfaceMemberDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTFieldDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTFieldDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTStaticInitializer.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTStaticInitializer node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTConstructorDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTConstructorDeclaration node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTMethodDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get method return type from method symbol
   * <li>set currently visited operation return type
   * <li>enter operation scope
   * <li>visit children except last one (method body)
   * <li>indicate test method body is being visited and visit it
   * <li>exit operation scope
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTMethodDeclaration node,Object data) 
  {
     // get method return type from method symbol
     LCallable operation = (LCallable)node.getScope();

     // set currently visited operation return type
     this._visitedOpeReturnType = operation.getReturnType();

     scopeStack.push(operation);     // enter scope

     // visit children except last one ( method body )
     for ( int i = 1; i < node.jjtGetNumChildren() - 1 ; i++ )
     {
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);
     }
     
     // visit test method body
     // state a test method body is being visited

     this.inTestMethod = true;
     SimpleNode body = (SimpleNode)node.jjtGetChild
                                        (node.jjtGetNumChildren() - 1);
     body.jjtAccept(this,data);
     this.inTestMethod = false;


     scopeStack.pop();               // exit scope


     this._visitedOpeReturnType = null;

     return data;
  }

  /**
   * Visits an ASTStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

   /**
   * Visits an ASTVariableDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get variable scope
   * <li>enter variable scope, note this scope will be closed only when the 
   *     enclosing block is left
   * <li>visit variable initializer if there is one (while visiting 
   *     assignability of initalizer to variable is checked)
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTVariableDeclarator node,Object data) 
  {
     // get variable scope
     Scope variableScope = node.getScope();
     if ( variableScope != null ) 
     { 
        // enter variable scope, note this scope will be closed
        // only when the enclosing block is left.
        scopeStack.push(variableScope);
     }
     // visit variable initializer if there is one ((while visiting 
     // assignability of initalizer to variable is checked)
     if ( node.jjtGetNumChildren() > 1 )
     {
        LType type = null;
        try
        {
           // visit variable initalizer with the variable type
           // retrieved from a node property
           type = (LType)node.getProperty("variableType");
           ((SimpleNode)node.jjtGetChild(1)).jjtAccept(this,type);
        }
        catch ( PropertyAccessException e ) { }
     }
     return data;
  }

  /**
   * Visits an ASTBlock.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get block scope
   * <li>enter block scope
   * <li>visit children
   * <li>exit block scope ( sevaral variable declaration scopes may have
   *     been opened and shall be exited - that is popped from the scope
   *     stack)
   * <li>
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTBlock node,Object data) 
  {
     Scope blockScope = node.getScope();  // get block scope
     scopeStack.push(blockScope);    // enter block scope
     node.childrenAccept(this,data); // visit children

     // exit block scope ( sevaral variable declaration scopes may have
     // been opened and shall be exited - that is popped from the scope
     // stack)
     Scope popedScope;
     do { popedScope = (Scope)scopeStack.pop(); }
     while ( popedScope != blockScope );
     return data;
  }

  /**
   * Visits an ASTExplicitConstructorInvocation.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTExplicitConstructorInvocation node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTBlockStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTBlockStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTLabeledStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTLabeledStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTEmptyStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTEmptyStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTSwitchStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>perform type check : first expression shall be compatible with int
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTSwitchStatement node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // perform type check : first expression shall be compatible with int.
     LType   type = ((TypedNode)node.jjtGetChild(0)).getType();
     Integer kind = type.getTypeKind();
     if ( ! JTypeKind.ARITHMETIC_TYPE.equals(kind) )
     {
        errorMessage
        (  "Incompatible type for switch. Can't convert "
         + ((TypedNode)node.jjtGetChild(0)).getTypeName()
         + " to int.",
         (SimpleNode)node.jjtGetChild(0)
        );
     }
     else if ( ! JTypes.intT.assignable(type) )
     {
        errorMessage
        (  "Incompatible type for switch. Explicit cast needed to convert "
           + ((TypedNode)node.jjtGetChild(0)).getTypeName() +  " to int.",
         (SimpleNode)node.jjtGetChild(0)
        );
     }
     return data;
  }

  /**
   * Visits an ASTSwitchLabel.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if a label expression is present: visit it
   * <li>and check expression is compatible with int
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTSwitchLabel node,Object data) 
  {
    // if a label expression is present : visit
    if ( node.jjtGetNumChildren() > 0 )
    {
       node.childrenAccept(this,VISITING_EXPRESSION);
       
       // and check expression is compatible with int
       LType   type = ((TypedNode)node.jjtGetChild(0)).getType();
       Integer kind = type.getTypeKind();
       if ( ! JTypeKind.ARITHMETIC_TYPE.equals(kind) )
       {
          errorMessage
          (  "Incompatible type for case. Can't convert "
             + ((TypedNode)node.jjtGetChild(0)).getTypeName()
             + " to int.",
             (SimpleNode)node.jjtGetChild(0)
          );
       }
       else if ( ! JTypes.intT.assignable(type) )
       {
          errorMessage
          (  "Incompatible type for case. Explicit cast needed to convert "
           + ((TypedNode)node.jjtGetChild(0)).getTypeName() +  " to int.",
           (SimpleNode)node.jjtGetChild(0)
          );
       }
    }   
    return data;
  }

  /**
   * Visits an ASTIfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first children (indicate it can be an expression)
   * <li>visit other children
   * <li>check first child is a boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTIfStatement node,Object data) 
  {
     // visit first children (indicate it can be an expression)
     TypedNode firstNode = (TypedNode)node.jjtGetChild(0);
     ((SimpleNode)firstNode).jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children
     for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);

     // check first child is boolean
     if ( ! JTypes.booleanT.equals(firstNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for if. Can't convert " 
           + firstNode.getTypeName() +  " to boolean.",
           firstNode
          );
     }
     return data;
  }

  /**
   * Visits an ASTWhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first children (indicate it can be an expression)
   * <li>visit other children
   * <li>check first child is a boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTWhileStatement node,Object data) 
  {
     // visit first children (indicate it can be an expression)
     TypedNode firstNode = (TypedNode)node.jjtGetChild(0);
     firstNode.jjtAccept(this,VISITING_EXPRESSION);
     
     // visit other children
     for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,data);

     // check first child is boolean
     if ( ! JTypes.booleanT.equals(firstNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for while. Can't convert " 
           + firstNode.getTypeName() +  " to boolean.",
           firstNode
          );
     }
     return data;
  }

  /**
   * Visits an ASTDoStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child
   * <li>visit last child (indicate it can be an expression)
   * <li>check last child is a boolean
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTDoStatement node,Object data) 
  {
     // visit first child 
     ((SimpleNode)node.jjtGetChild(0)).jjtAccept(this,data);

     // visit last child (indicate it can be an expression) 
     TypedNode lastNode = (TypedNode)node.jjtGetChild(1);
     lastNode.jjtAccept(this,VISITING_EXPRESSION);
     
     // check last child is boolean
     if ( ! JTypes.booleanT.equals(lastNode.getType()) )
     {
          errorMessage
          (  "Incompatible type for do. Can't convert " 
           + lastNode.getTypeName() +  " to boolean.",
           lastNode
          );
     }
     return data;
  }

  /**
   * Visits an ASTForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>enter "for" statement scope
   * <li>visit children (except but for ASTForInit and ASTForUpdate indicate
   *     child can be an expression)
   * <li>exit "for" statement scope
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTForStatement node,Object data) 
  {
     scopeStack.push(node.getScope());  // enter "for" statement scope

     // visit children (except but for ASTForInit and ASTForUpdate indicate
     // child can be an expression)
     SimpleNode child;
     for ( int i = 0; i < node.jjtGetNumChildren(); i++ )
     {
        child = (SimpleNode)node.jjtGetChild(i);
        if ( child instanceof ASTForInit || child instanceof ASTForUpdate )
           child.jjtAccept(this,data);
        else
           child.jjtAccept(this,VISITING_EXPRESSION);
     }

     scopeStack.pop();                  // exit "for" statement scope

     return data;
  }

  /**
   * Visits an ASTForInit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTForInit node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTBreakStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTBreakStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTContinueStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTContinueStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTReturnStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>check if the statement contains an expression
   * <li>if there's an expression, visit the expression, and save its type
   * <li>if it's a stand alone return, save return type is void
   * <li>check saved type is compatible with return type
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTReturnStatement node,Object data) 
  {
     // check if the statement contains an expression
     LType   type;
     if ( node.jjtGetNumChildren() > 0 )
     {
        // there's an expression, visit the expression, and save its type 
        node.childrenAccept(this,VISITING_EXPRESSION);
        type = ((TypedNode)(node.jjtGetChild(0))).getType();  
     }
     else
     {
        // it's a stand alone return, save return type is void
        type = JTypes.voidT;
     }

     // check type is compatible with return type
     if ( ! (    JTypes.voidT.equals(this._visitedOpeReturnType) 
              && JTypes.voidT.equals(type) 
            )
          && ( ! this._visitedOpeReturnType.assignable(type) ) 
        )
     {
        errorMessage
        (  "Incompatible type for return. Can't convert "
         + this._visitedOpeReturnType.getName()
         + " to " 
         + ( type == null ? "<no type>" : type.getName() ),
         node
        );
     }

     return data;
  }

  /**
   * Visits an ASTThrowStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child (indicate it can be an expression)
   * <li>get child type
   * <li>check child type is compatible with Throwable
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTThrowStatement node,Object data) 
  {
     // visit child (indicate it can be an expression)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // get child type
     LType type = ((TypedNode)node.jjtGetChild(0)).getType();  

     // check child type is compatible with Throwable
     if ( ! JTypes._Throwable.assignable(type) )  
     {
        errorMessage
        (  "Can't throw " + ((TypedNode)node.jjtGetChild(0)).getTypeName()
           + "; it must be a subclass of class java.lang.Throwable.",
           node
        );
     }
     return data;
  }

  /**
   * Visits an ASTForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (only first child can be an expression)
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTSynchronizedStatement node,Object data) 
  {
     // visit children (only first child can be an expression)
     for ( int i = 0; i < node.jjtGetNumChildren(); i++)
     {
        ((SimpleNode)node.jjtGetChild(i)).jjtAccept(this,
                                                 (  i == 0 
                                                  ? VISITING_EXPRESSION
                                                  : data )
                                                 );
     }
     return data;
  }

  /**
   * Visits an ASTTryStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children 
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTTryStatement node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit child
   * <li>set expression type to child's type
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTExpression node,Object data) 
  {
    node.childrenAccept(this,VISITING_EXPRESSION);
    node.setType(((TypedNode)node.jjtGetChild(0)).getType());
    return data;
  }

  /**
   * Visits an ASTVariableInitializer.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>get type of the variable to initalize (contained in data)
   * <li>check initializeer type can be assigned to variable
   * </ul>
   *
   * @param node the node to visit
   * @param data type of the variable to initialize
   * @return null.
   *
   */
  public Object visit(ASTVariableInitializer node,Object data) 
  {
     // visit children (indicate they can be expressions)
    node.childrenAccept(this,VISITING_EXPRESSION);

    // get type of the variable to initalize (contained in data)
    LType type             = (LType) data;

    // check initializeer type can be assigned to variable
    TypedNode initTypeNode = (TypedNode)node.jjtGetChild(0);

    LType initType         = initTypeNode.getType();
    if ( ! type.assignable(initType) )
    {
       errorMessage
       (  "Incompatible type for variable initializer. Can't convert "
        + initTypeNode.getTypeName()
        + " to " 
        + type.getName(),
        node
       );
    } 
    return null;
  }

  /**
   * Visits an ASTLocalVariableDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>enter variable scope in case a new scope was defined for this variable
   * <li>visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTLocalVariableDeclaration node,Object data) 
  {
     // enter variable scope in case a new scope was defined for this variable
     Scope variableScope = node.getScope();
     if ( variableScope != null ) 
     { 
        // new scope defined, enter variable scope.
        // Note this scope will be closed only when the enclosing block is left
        scopeStack.push(variableScope);
     }
     // visit children
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visits an ASTAssignment.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate they can be expressions)
   * <li>get left and right operand
   * <li>get operator from node property "op"
   * <li>perform type check and type synthesis, as follows:
   * <li>if operator += is overrided for String : any type can be on right hand
   *     side
   * <li>for all operators : right hand side operand shall be assignable to
   *     left hand side operand
   * <li>for operator = : assignability is enough
   * <li>for operators *= /= %= += -= : type are required to be arithmetic
   *     beside being assignable
   * <li>for operators << >> >>> : type are required to be integer arithmetic
   *     beside being assignable
   * <li>for operators &= |= ^= : type are required to be either all
   *     arithmetics or all booleans beside being assignable
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTAssignment node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     boolean ok = true;

     // get left and right operand
     TypedNode left  = (TypedNode)node.jjtGetChild(0);
     TypedNode right = (TypedNode)node.jjtGetChild(1);

     // get operator

     String op = "";
     try { op = (String)node.getProperty("op"); }
     catch (PropertyAccessException e) { }

     // perform type check and type synthesis
     if ( (left == null) || (left.getType() == null) )
     {
        errorMessage
        (  "Incompatible type for assignment. Left hand type unknown.",
           (TypedNode)node.jjtGetChild(0)
        );
     }
     else if ( (right == null) || (right.getType() == null) )
     {
        errorMessage
        (  "Incompatible type for assignment. Right hand type unknown.",
           (TypedNode)node.jjtGetChild(1)
        );
     }
     else if ( op.equals("+=") && JTypes._String.equals(left.getType()) )
     {
        // operator += is overrided for String : any type can be on right hand
        // side
        node.setType(left.getType());
     }
     else if ( ! left.getType().assignable( right.getType() ) )
     {
        // all operators : right hand side operand shall be assignable to
        // left hand side operand
        errorMessage
        (  "Incompatible type for " + op + ". Can't convert "
         + right.getTypeName()
         + " to " 
         + left.getTypeName(),
         node
        );
     }
     else
     {
        if ( op.equals("=") )
        {
           // operator = : assignability is enough
           node.setType(left.getType());
        }
        else if (   op.equals("*=")
                 || op.equals("/=")
                 || op.equals("%=")
                 || op.equals("+=")
                 || op.equals("-=")
                )
        {
           // operators *= /= %= += -= : type are required to be arithmetic
           // beside being assignable
           if (  left.getTypeKind()  != JTypeKind.ARITHMETIC_TYPE )
           {
              errorMessage
              ("Incompatible type for " + op 
               + ". Only arithmetic types allowed.",node);
           } 
           else node.setType(left.getType());
        }
        else if (    op.equals("<<=")
                  || op.equals(">>=")
                  || op.equals(">>>=")
                )
        {
           // operators << >> >>> : type are required to be integer arithmetic
           // beside being assignable
           if (     ! JTypes.longT.assignable(right.getType())
                ||  ! JTypes.longT.assignable(left.getType())
              )
           {
              errorMessage
              ("Incompatible type for " + op
               + ". Only integer types allowed.",node);
           } 
           else node.setType(left.getType());
        }
        else if (    op.equals("&=")
                  || op.equals("|=")
                  || op.equals("^=")
                )
        {
           // operators &= |= ^= : type are required to be either all 
           // arithmetics or all booleans beside being assignable
           if (     ( left.getTypeKind()  != JTypeKind.ARITHMETIC_TYPE )
                 || ( left.getTypeKind()  != JTypeKind.BOOLEAN_TYPE )
              )
           {
              errorMessage
              ("Incompatible type for " + op
               + ". Only arithmetic or boolean types allowed.",node);
           } 
           else node.setType(left.getType());
        } 
     }
     return data;
  }

  /**
   * Visits an ASTPostfixExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children (indicate children can be expressions)
   * <li>check all children are arithmetic
   * <li>set expression type accordingly
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTPostfixExpression node,Object data) 
  {
     // visit children (indicate they can be expressions)
     node.childrenAccept(this,VISITING_EXPRESSION);

     // check children are arithmetic
     try 
     {
        LType type = ADLTypeCheckHelper.checkTypesAreSameKind
                                        (node,JTypeKind.ARITHMETIC_TYPE);
        node.setType(type);
     }
     catch ( BadTypeException e )
     {
        errorMessage
        ("Incompatible type for postfix expression. " 
         + e.getMessage(), e.getErrorNode());
     }

     if ( debug )
     {
        System.out.println("debug: (PostFixExpression) - type "
                         + "(" + node.getTypeName() + ") ");
     }

     return data;
  }

  /**
   * Visits an ASTStatementExpressionList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTStatementExpressionList node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }

  /**
   * Visits an ASTForUpdate.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the node to visit
   * @param data unused data.
   * @return received data object.
   *
   */
  public Object visit(ASTForUpdate node,Object data) 
  {
    node.childrenAccept(this,data);
    return data;
  }


  // -- private --

  /**
   * Check an expression can be used at the place it is found.
   * @param expression   the expression for which the check is to be performed.
   * @param relatedNode  node to use to get line number in case an 
   *                     error is to be reported.
   * @return true if the expression can be used, false otherwise.
   */
  private boolean checkSymbolIsAllowed
                  (Symbol symbol, SimpleNode relatedNode)
  {
     if ( symbol instanceof ADLBinding )
     {
        if (    ( this.callStateScopeLevel > 0 )
             ||   this.inPrologueScope 
             || ( this.inEpilogueScope && ! this.inADLMethodScope )
           )
        {
           errorMessage("A binding cannot be used in the scope of a call-state"
                        + " expression, nor in a prologue, nor in a global "
                        + " epilogue.", relatedNode);
           return false;
        }
     }
     else if ( symbol instanceof ADLInline )
     {
        if (    this.inADLInlineScope
             || this.inPrologueScope 
             || this.inEpilogueScope
           )
        {
           errorMessage("An inline cannot be used in the scope of an other "
                        + " inline, nor in a prologue, nor in an epilogue.",
                       relatedNode);
           return false;
        }
     }
     else if ( symbol instanceof TDDDataset )
     {
        // Not only test directive is to be tested here because for the other
        // places where dataset expression are allowed the test is straight
        // forward since only ASTDatasetExpr are allowed.
        if ( ! this.inTestDirective )
        {
           errorMessage
           (  "Dataset expression " + symbol.getName() + " cannot be"
            + " used here. Dataset expression are allowed only in dataset "
            + " definition, in factory invocation or in test directive.",
            relatedNode
           );
           return false;
        }
     }
     else if ( symbol instanceof LMethod )
     {
        // Test methods can only be used in test directives
        if (    ((LMethod)symbol).isTestMethod() 
             && ! ( this.inTestDirective || this.inTestMethod )
           )
        {
           errorMessage
           (  "Test method " + symbol.getName() + " cannot be"
            + " used outside a test directives or another test methods.",
            relatedNode
           );
           return false;
        }
     }
     return true;
  }

  /**
   * Decorate ASTName with usefull information.
   * This information comprise:<ul> 
   * <li> the enclosing structure (TDDClass, ADLClass...) of the expression,
   *      property name is "enclosingConstructName"
   * <li> the class of the expression corresponding to the name. It enables
   *      to know wether an encountered name is that of a Java variable, an 
   *      ADLInline, a Java method... The property name is "astNameKind".
   * </ul>
   * @param astName ASTName node holding the expression name
   * @param symbol symbol corresponding to the expression
   */

  private void decorateASTName(SimpleNode astName, Symbol symbol)
  {
     // For TDD dataset and factories save symbol's enclosing structure name.

     if (   ( symbol instanceof TDDFactory ) 
	 || ( symbol instanceof TDDDataset ) 
         || (    ( symbol instanceof LAttribute ) 
	      && ( symbol.getEnclosingConstruct() instanceof TDDClass ) )
	)
     {
        try
        {
           TDDClass tddClass = (TDDClass) symbol.getEnclosingConstruct();
           if (tddClass != null)
             astName.setProperty("enclosingConstructName", tddClass.getName());
        }
        catch(PropertyAccessException e) { }
     }

     try
     {
        Class symbolClass = symbol.getClass();

       // the following prevent setting the property twice
       if ( astName != null)
       {
         boolean astNameKindPropertySet = false;
         try 
         {
           String unused = (String)astName.getProperty("astNameKind");
           astNameKindPropertySet = true;
         }
         catch(PropertyAccessException e) { }
         if ( ! astNameKindPropertySet )
         {
           // hook for range methods
           if (     astName.equals("ADL_short_range")
                 || astName.equals("ADL_int_range")
                 || astName.equals("ADL_long_range")
              )
           {
              // range method - property is objectACO
              astName.setProperty
              ("astNameKind","objectACO");
           }
           else
           {
              // normal behaviour
              // store last suffix of expression class
              String s = symbolClass.toString();
              astName.setProperty
              ("astNameKind",s.substring(s.lastIndexOf('.')+1));

              if (nld_required) {
                if (symbol instanceof Scope) {
                  astName.setProperty("NLDScope", symbol);
                }
              }
           }
         }
       }
     }
     catch ( PropertyAccessException e ) {}
  }

  /**
   * Display an error message.
   * @param msg the message to display.
   */

  private static void errorMessage(String msg)
  {
     errorMessage(msg,0);
  }

  /**
   * Display an error message.
   * @param msg the message to display.
   * @param erroneousNode node to use to get additional information 
   *                      (e.g. line number)
   */

  private static void errorMessage(String msg,SimpleNode erroneousNode)
  {
     int line = ( erroneousNode != null ) ? erroneousNode.getLineNumber() : 0;
     errorMessage(msg,line);
  }

  /**
   * Display an error message.
   * @param msg the message to display.
   * @param line number of the line where the error occured.
   */
  
  private static void errorMessage(String msg,int line)
  {
     System.out.println(fileName + ":" + line  + ": " + msg + "\n");
     if ( typeCheckOk) { typeCheckOk = false; }
  }

  /**
   * Display a debug message.
   * @param msg the message to display.
   * @param node node to use to get additional information (e.g. line number)
   */

  private static void debugMessage(String msg,SimpleNode node)
  {
     int line = ( node != null ) ? node.getLineNumber() : 0;
     System.out.println("debug:" + line  + ": " + msg + "\n");
  }

  /** Type check status. */
  private static boolean typeCheckOk = true;

  /** Name of the source file from which the AST was constructed. */
  private static String fileName = null;

  /** Name of the source file from which the AST was constructed,
      without path nor extension. */
  private static String sourceName = null;

  /** Constant indicating that an expression is being visited. */
  private static final Integer VISITING_EXPRESSION   = new Integer(0);

  /** Constant indicating that a dataset expression is being visited. */
  private static final Integer VISITING_DATASET_EXPR = new Integer(1);

  /** Adl superclass of the visited ADL class. */
  private ADLClass _adlSuperClass = null;

  /**  Return type of the operation (ADLMethod, Method, TDD Factory)
   *   currently being visited.
   */
  private LType _visitedOpeReturnType = null;

  /** Currently visited tested method. */
  private ADLMethod _visitedADLMethod = null;

  /** Stack containing the scopes where symbols shall be searched. */
  private Stack scopeStack = new Stack();

  /** Level of nesting inside call states scopes */
  static private int callStateScopeLevel = 0;

  /** Flag indicating if we're in a prologue scope */
  static private boolean inPrologueScope = false;

  /** Flag indicating if we're in an epilogue scope */
  static private boolean inEpilogueScope = false;

  /** Flag indicating if we're in an ADL method scope. */
  static private boolean inADLMethodScope = false;

  /** Flag indicating if we're in an ADL inline scope. */
  static private boolean inADLInlineScope = false;

  /** Flag indicating if we're in a TDD test directive. */
  static private boolean inTestDirective = false;

  /** Flag indicating if we're in a TDD test method. */
  static private boolean inTestMethod = false;

  /** Cycle detector used to detect cycles in dataset expressions. */
  private static CycleDetector cycleDetector = new CycleDetector();

  /** Dataset currently being defined. */
  private SimpleNode currentDataset = null;

  /** Name of the dataset currently being defined. */
  private String currentDatasetName = null;

  /** documentation to be generated */
  private boolean nld_required = false;

  /** Handle on the NL engine associated with the compiler */
  private NLEngine nlengine;

  /** Debug flag. */
  private static boolean debug = false;

}

// ***********************************************************************
// History Log:
// $Log: ADLTypeCheckVisitor.java,v $
// Revision 1.50  1998/07/28 13:52:26  mallet
// Fixed message in case of type check error in dataset initializer.
//
// Revision 1.49  1998/07/01 12:08:42  mallet
// Minor change to an error message.
//
// Revision 1.48  1998/06/30 10:13:11  bareau
// Added support for NL unparsing of allocation expressions (setASTLastName,
// in order that the node be correctly decorated).
//
// Revision 1.47  1998/06/12 13:19:02  bareau
// Added support for NL unparsing of quantified assertions.
//
// Revision 1.46  1998/06/11 15:41:42  bareau
// Bug fix: the ADL_IfStatement was not type-checked
// (first node is boolean, "then" node is boolean if no "else" node,
// same type for "then" and "else" nodes otherwise).
//
// Revision 1.45  1998/06/10 12:44:54  mallet
// Removed debug message.
//
// Revision 1.44  1998/06/10 09:05:36  cinard
// Fixed bug in type check : property "enclosingConstructName" is now set
// only for factories, datasets and tdd fields. It was also set for all class
// attributes!
//
// Revision 1.43  1998/06/10 08:07:09  cinard
// Fixed bug in type check : tdd field initializers were not visited,
// VariableDeclaratorId nodes were visited instead! The bug was introduced
// because VariableInitializer nodes are at indexes 2, 4 ... in
// TDD_FieldDeclaration while they are at index 1 in VariableDeclarator.
//
// Revision 1.42  1998/05/26 15:16:04  bareau
// Added property enclosingConstructName for constant names,
// so that external constants in TDD classes can be referenced.
//
// Revision 1.41  1998/05/15 13:59:37  bareau
// Added support for NLD unparsing.
//
// Revision 1.40  1998/04/28 12:44:45  cinard
// Fixed a bug when visiting array subscripts.
//
// Revision 1.39  1998/04/15 13:55:23  bareau
// Added support for NL processing. NLD annotations are tokenized
// and sent to a distinct NLD parser (hence there is no more NLD_*
// productions in the ADL/TDD grammars).
// A handle for the NL engine associated to the current compiler is
// provided.
//
// Revision 1.38  1998/03/31 12:35:37  bareau
// Bug fix: typing error in ASTADL_SuperSemantics (constructor case).
//
// Revision 1.37  1998/03/25 14:59:08  cinard
// Improved support for parsing multiple source file. The parsing of some
// source files can now be triggered while parsing or type checking
// a file.
//
// Revision 1.36  1998/03/19 17:36:27  bareau
// Made cycle detection more robust: now detects cross-referenced definitions
// "inter" tdd classes and not only "intra" tdd class.
//
// LIMITATION: uncorrect cross-referenced definitions involving at least
// two files are detected only when these files are simultaneously
// compiled (i.e. they are both on the command line).
//
// Revision 1.35  1998/03/19 09:22:42  cinard
// Classes from unnamed package names are now a property of the AST
// ("classesFromUnnamedPackage").
//
// Revision 1.34  1998/03/18 16:19:48  cinard
// Symbol table factory (sbf) is used instead of symbol table
// manager (stm) to create symbols.
// Implemented mechanism for "use" clause.
// Implemented checks for "extends" clause.
//
// Revision 1.33  1998/03/13 11:20:14  bareau
// Bug fixes:
// - in the short form syntax of bindings ("define <type> foo with <expr>;"),
// the type check was not performed.
// - long/double literals (0L, 1l, 3D...) were not correctly typed (the
// typeName was set but not the LType).
//
// Revision 1.32  1998/03/10 17:26:34  mallet
// Merged v1.1 with development trunk
//
// Revision 1.31.2.1  1998/03/05 17:44:30  cinard
// Improved error message (indicates now that public method or constructor
// was not found when relevant).
//
// Revision 1.31  1998/02/27 12:52:41  cinard
// Added a TBC notice for later (a check will have to be updated when multiple
// ADL source files may be checked when compiling a TDD files).
// Updated copyright notice.
//
// Revision 1.30  1998/02/24 16:34:27  cinard
// Added productions for CompilationUnit, PackageDeclaration.
// Visitors should implement visit methods for ASTCompilationUnit and
// ASTPackageDeclaration.
//
// Revision 1.29  1998/02/23 14:54:45  cinard
// Changed references to JTypeChecker fields to references to JTypeKind fields.
//
// Revision 1.28  1998/02/12 16:02:57  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.27  1998/02/11 12:50:12  bareau
// Moved "return" from ADL_BasicExpression to PrimaryPrefix.
//
// Revision 1.26  1998/02/10 14:06:57  cinard
// Fixed a bug in type check (type cast checks when types are unknown).
// Excluded NLD nodes from type check visit.
//
// Revision 1.25  1998/02/03 14:32:51  bareau
// Bug fix for void datasets: in TDDDatasetDomain, visit expression node
// even when there is no SingleDeclarator.
//
// Revision 1.24  1998/02/03 12:50:22  cinard
// New check : TDD test method calls allowed only in test directives or in
// 	    other test methods.
//
// Revision 1.23  1998/01/28 18:15:40  cinard
// Fixed minor bug on qualified constructor names.
//
// Revision 1.22  1998/01/26 15:11:53  bareau
// Fixed conditional check in saveASTNameKind().
//
// Revision 1.21  1998/01/26 14:16:12  cinard
// *** empty log message ***
//
// Revision 1.20  1998/01/26 14:08:20  cinard
// Fixed conditional check in saveASTNameKind().
//
// Revision 1.19  1998/01/26 12:56:04  cinard
// Fixed bug with astNameKind property.
//
// Revision 1.18  1998/01/22 14:22:12  cinard
// Updated and fixed comments.
//
// Revision 1.17  1998/01/22 11:45:30  cinard
// ASTName nodes containing ADL_*_range names have their astNameKind property
// set to objectACO.
//
// Revision 1.16  1998/01/19 14:29:23  cinard
// Improved SearchTypeInfo inner class design.
//
// Revision 1.15  1998/01/13 16:29:36  cinard
// Fixed bug when type cheking foo().bar() and likes expressions.
//
// Revision 1.14  1998/01/13 16:21:46  cinard
// Fixed a bug. Type check of expressions like "foo".toString() are now ok.
//
// Revision 1.13  1998/01/12 16:26:15  cinard
// Fixed bug : tdd fields names can now be used as dataset (singleton) within
// dataset definitions and factory calls.
//
// Revision 1.12  1998/01/12 16:01:17  cinard
// Fixed a type check bug with TDD_Field.
//
// Revision 1.11  1998/01/12 13:21:08  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.10  1998/01/08 16:25:30  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.9  1998/01/08 11:42:57  cinard
// Fixed a bug while type checking PrimaryExpression (type field in
// SearchTypeInformation was not set to null when a new DotSuffix was encountered).
// Added comments.
//
// Revision 1.8  1998/01/08 10:00:09  cinard
// Added comments.
// Fixed a bug with the "isClass" property of TDD_ADLexpression.
//
// Revision 1.7  1998/01/07 19:17:51  cinard
// Added comments.
// Modified type check for ASTPrimaryExpression and ASTTDD_ADLExpression.
//
// Revision 1.6  1998/01/07 10:59:13  cinard
// Fixed bug in type check for ADL_new expression.
//
// Revision 1.5  1998/01/06 15:52:14  cinard
// Added comments.
//
// Revision 1.4   1998/01/06 10:52:22   bareau
// Bug fix: scope problem for the ADL "return".
//
// Revision 1.3   1998/01/06 10:23:31   cinard
// Fixed problem with use of parenthesed expressions as primary prefix.
//
// Revision 1.2   1998/01/05 17:52:44   cinard
// Fixed type check for ADL try/catch blocks.
//
// Revision 1.1   1998/01/05 16:03:12   mallet
// Type Check visitor running after parsing.
//
// ***********************************************************************
