// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1996-1997 Sun Microsystems Inc.
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adljava/ParseUnit.java,v $
// $Revision: 1.8 $
// $Date: 1998/10/23 15:01:47 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADLfor Java parser
// NAME:        ParseUnit.java
//
// HISTORY:
//      Cyril Inard d'argence, The Open Group - Mar 20, 1998: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adljava;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.*;
import ORG.opengroup.adl.compiler.ADLCompiler;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.j.*;
import ORG.opengroup.adl.nlengine.*;

/**
 * The ParseUnit class enable to monitor the parsing of an adl or tdd
 * source file. 
 * <P>In addition, this class provide static methods to create and/or find
 * ParseUnit objects.
 */

public class ParseUnit
{ 

  // -- Public static methods --

  /** 
   * Get or create a ParseUnit object for supplied source file.
   * @param path full path name of the source file.
   */

  public static ParseUnit getParseUnit(String path, NLEngine nlengine)
  {
     File   file          = new File(path);
     ParseUnit parseUnit = (ParseUnit)_parseUnits.get(file.getName());
     if ( parseUnit == null )
     {
        parseUnit = new ParseUnit(file, nlengine);
        _parseUnits.put(file.getName(),parseUnit);
     }
     return parseUnit;
  }

  // -- Public methods --

  /**
   * Test if the specified file can be parsed, that is at least exists
   * and is readable.
   */

  public boolean canParse()
  {
     return _file.canRead();
  }

  /**
   * Pure parse and type check a source file and construct an AST according to 
   * the rule defined in the ADL for Java grammar.
   * This methods performs the two passes of the two-pass parsing:<ul>
   * <li>first pass is the AST construction: the AST is constructed and the
   *     symbol table is filled as symbols are defined,
   * <li>second pass is the type check, performed by the ADLTypeCheck visitor:
   *     AST nodes corresponding to typed expressions are decorated with
   *     type information.
   * </ul>
   * <P>Note referenced source files can also be pure parsed are (first pass 
   *    only) as part of this process.</P>
   */

  public Node parse()
  {
     SimpleNode ret = null;

     if ( debug )
     {
        System.out.println("debug: ParseUnit.parse(" + _path + ")");
     }

     // First pass : pure parse

     if ( ! isParsed() )
     {
        pureParse();
     }

     // Second pass : type check if necessary

     if ( ! hasErrors() && ! isTypeChecked() )
     {
        ret = typeCheck();
     }
     else if ( isTypeChecked() )
     {
        ret = _ast;
     }
     
     if ( debug )
     {
        System.out.println("debug: ParseUnit.parse(" + _path + ")"
                          + " return " + ret);
     }

     return ret;
  }

  /**
   * Pure parse a source file and construct an AST according to 
   * the rule defined in the ADL for Java grammar.
   * This methods performs the first pass of the two-pass parsing.
   * @see ORG.opengroup.adl.adljava.ParseUnit#parse
   */
  
  public SimpleNode pureParse()
  {
     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParse(" + _path + ")");
     }

     // Parse only if it has not been done.

     if ( ! isParsed() )
     {

        // save symbol table context
        Object savedStmContext = SimpleNode.stm.getContext();
                     
        ADLJavaParser parser = null;
        try
        {
           
           parser = new ADLJavaParser
                        (new BufferedInputStream
                            (ADLCompiler.getStream(_path)));

           parser._fileName = _path;
           
           if (_nld_required)
             parser.setNLDrequired(_nl_engine);

           parser.disable_tracing();
           
           SimpleNode.stm.reInit();  // reinitialize symbol table
           
           _ast = parser.ADL_CompilationUnit();

           _isParsed = true;

           // Update dependencies

           setDependencies(parser.getUsedSourceFiles());
        }
        catch (FileNotFoundException e)
        {
           // higly improbable since the file has normally been verified
           // in Adlt main
           _hasErrors = true;
           System.err.println(_path + ": File not found!");
        }
	catch (IOException ioe)
	{
	  _hasErrors = true;
	  System.err.println("Preprocessing error: " + ioe.getMessage());
	}
        catch ( ParseException e ) 
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Parse exception caught.";
           parser.errorMessage(message,parser.getToken(0).beginLine);
        }
        catch ( TokenMgrError e ) 
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Lexical error encountered.";
           parser.errorMessage(message,parser.getToken(0).beginLine);
        }
        catch ( JSymbolError e )
        {
           _hasErrors = true;
           String message = ( e.getMessage() != null )
                            ? e.getMessage() 
                            : "Java symbol error encountered.";
           parser.errorMessage(message,parser.getToken(0).beginLine);
        }

        // restore symbol table context
        SimpleNode.stm.setContext(savedStmContext);
     }

     if ( debug )
     {
        System.out.println("debug: ParseUnit.pureParse(" + _path + ")"
                           + " return " + _ast );
     }

     return _ast;
  }

  /** Is source file parsed? */

  public boolean isParsed() { return _isParsed; }

  /** Were errors encountered while parsing the file. */
  
  public boolean hasErrors()  { return _hasErrors; }
  
  // -- Private constructors --

  /**
   * Constructs a parse unit object. Useful to monitor parsing
   * of an ADL or TDD source file.
   * <P>Note this constructor is private. The normal way to create
   * a parse unit is simply to invoke static method getParseUnit().
   * </P>
   * @param file adl or tdd source file.
   */

  private ParseUnit(File file)
  {
     _file     = file;
     _path     = file.getPath();
     _fileName = file.getName();
  }

  private ParseUnit(File file, NLEngine nlengine)
  {
    _file     = file;
    _path     = file.getPath();
    _fileName = file.getName();
    _nld_required = nlengine != null;
    _nl_engine = nlengine;
  }

  // -- Private methods --

  /** Get AST for that source file. */
    
  private SimpleNode getAST() { return _ast; }

  /** Has source file been type checked. */

  private boolean isTypeChecked()
  {
     return _isTypeChecked;
  }

  /** Is source file currently being type checked. */

  private boolean isBeingTypeChecked()
  {
     return _isBeingTypeChecked;
  }

  /**
   * Perform type check of the source file.
   * <P>The type check is the second pass of the parsing. It is performed by 
   * the ADLTypeCheck visitor (AST nodes corresponding to typed expressions 
   * are decorated with type information).
   * @see ORG.opengroup.adl.adljava.ParseUnit#parse
   * @return  the AST obtained.
   */

  private SimpleNode typeCheck()
  {
     boolean ok = true;
     boolean completeTypeCheck = true;
     SimpleNode ret = null;

     if ( debug )
     {
        System.out.println("debug: ParseUnit.typeCheck(" + _path + ")");
     }

     // type check only if it has not yet been done (successfuly or not)

     if ( hasErrors() )
     {
        ok = false;
     }
     else 
     {
        if ( ! isParsed() )
        {
           pureParse();
        }

        if ( ! hasErrors() && ! isTypeChecked() )
        {
	   // mark current file as being type checked
	   _isBeingTypeChecked = true;

           ParseUnit[] usedFiles = getDependencies();

           for (int i = 0; i < usedFiles.length; i++)
           {
	      // First pure parse dependent files.
           
              if ( ! usedFiles[i].isParsed() )
              {
                 usedFiles[i].pureParse();
              }
               
	      // then type check if possible to be able to detect cycles:
	      // dependency paths are added during type check only!
	      if (! usedFiles[i].hasErrors()
		  && ! usedFiles[i].isTypeChecked()
		  && ! usedFiles[i].isBeingTypeChecked()) {
		usedFiles[i].typeCheck();
	      } else if (usedFiles[i].isBeingTypeChecked()) {
		// hmm. the used file is currently being type checked,
		// sounds like a circular dependency
		// the current type check cannot be complete until we've
		// also type checked the dependency the other way around
		// example: A uses B uses A,
		// typeCheck(A) -> typeCheck(B) -> isBeingTypeChecked(A)
		//    => typeCheck(B) is incomplete until
		// typeCheck(B) -> typeCheck(A)
		//    => returns status of typeCheck(A)
		//       if hasErrors(A) => typeCheck(B) = false
		//       else do type check B completely
		completeTypeCheck = false;
	      }

              if ( usedFiles[i].hasErrors() )
              {
                 ok = false;
		 System.err.println(_fileName
				    + ": error(s) in dependent file: "
				    + usedFiles[i]._fileName);
                 break;
              }

           }

           // Then type check

           if ( ok )
           {
              // save symbol table context
              Object savedStmContext = SimpleNode.stm.getContext();
              
              ADLTypeCheckVisitor typeCheckVisitor = new ADLTypeCheckVisitor
                                                         (_path);

              if (_nld_required)
                typeCheckVisitor.setNLDrequired(_nl_engine);

              try
              {
                 Object stmContext = _ast.getProperty("stmContext");
                 SimpleNode.stm.setContext(stmContext);
              }
              catch ( PropertyAccessException e ) { }
              
              ok = ((Boolean)_ast.jjtAccept(typeCheckVisitor,null))
                   .booleanValue();

              
              // update state
              if ( ok)
              {
                 _isTypeChecked = completeTypeCheck;
                 ret = _ast;
              }
              else
              {
                 _hasErrors = true;
              }

              // restore symbol table context
              SimpleNode.stm.setContext(savedStmContext);
           }

	   _isBeingTypeChecked = false;
        }
     }

     if ( debug )
     {
        System.out.println("debug: ParseUnit.typeCheck(" + _path + ")"
                           + " return " + ret + " (has errors? "
			   + _hasErrors + ")");
     }

     return ret;
  }

  /** Set source file dependencies. */
  
  private void setDependencies(String[] depNames)
  {
     _fileDependencies = new ParseUnit[depNames.length];
     for ( int i = 0; i < depNames.length; i++ )
     {
        _fileDependencies[i] = getParseUnit(depNames[i], _nl_engine);
     }
  }
  
  /** Get source file dependencies */
  
  private ParseUnit[] getDependencies()
  {
     return _fileDependencies;
  }

  // -- Private attributes --

  /** File object associated to the file to parse. */
  private final File   _file;

  /** Last component of full path name. */
  private final String _fileName;

  /** Full path name */
  private final String _path;

  /** Is file pure parsed */
  private boolean      _isParsed      = false;

  /** Is file currently being type checked */
  private boolean _isBeingTypeChecked = false;

  /** Is file type checked (second pass). */
  private boolean      _isTypeChecked = false;

  /** NLD to generate. */
  private boolean _nld_required = true;

  /** NLEngine to feed with mapping. */
  private NLEngine _nl_engine = null;

  /** Does file contains errors. */
  private boolean      _hasErrors     = false;

  /** The AST resulting from the pure parse. */
  private SimpleNode   _ast           = null;

  /** Other parse unit referenced in this parse unit. */
  private ParseUnit[]  _fileDependencies;

  /** Map names of source files onto ParseUnit objects. */
  private static final HashMap _parseUnits = new HashMap();

  /** Debug flag.*/
  private static final boolean debug = false;

  // -- Test --

  /**
   * Main to use for unit tests of the parser.
   * @param args command arguments that should be [-d] inputfile.
   */

  public static void main(String args[]) 
  {
     Node    topNode = null;
     boolean traced = false;
     String  filename;

     int ai = 0;
     if (ai < args.length && args[ai].equals("-d"))
     {
        // turn on debugging
        traced = true;
        ai++;
     }

     if (ai == (args.length-1))
     {
        System.out.println("Reading from file " + args[ai] + " . . .");
        filename = args[ai];
     }
     else
     {
        System.out.println
        ( "Usage: java ORG.opengroup.adl.parsers.adljava.ParseUnit"
        + " [-d] inputfile");
        return;
     }

     ParseUnit pu = ParseUnit.getParseUnit(filename, null);
     if ( pu != null )
     {
        topNode = pu.parse();
     }

     if ( topNode != null )
     {
        System.out.println("Program parsed successfully.");
     }
     else
     {
        System.out.println("Error encountered.");
     }

     // Dumping symbol table

     System.out.println("-- Dumping symbol table --");
     System.out.println(SimpleNode.stm);
        
     // if ( topNode != null )
     // {
     //    ADLDumpVisitor dumpVisitor = new ADLDumpVisitor();
     //    topNode.jjtAccept(dumpVisitor,null);
     // }  
  }

}

// ***********************************************************************
// History Log:
// $Log: ParseUnit.java,v $
// Revision 1.8  1998/10/23 15:01:47  mallet
// Bug fixed: put back extension filter preprocessing by calling
// ADLCompiler.getStream() instead of creating a
// FileInputStream. IOException is also caught to handle preprocessing
// errors.
//
// Revision 1.7  1998/07/28 15:56:05  mallet
// Changed ParseUnit behavior to be able to detect cyclic dependencies
// between files even if they are not compiled in the same compiler
// session: a dependency file is not only parsed but also type
// checked. This ensures that all dependency paths from dependent files
// are recorded when we start type checking the current file. When there
// is a circular dependency, the type check of a dependent file is
// flagged as incomplete and should therefore be done again if required.
//
// Revision 1.6  1998/06/15 12:02:21  cinard
// JSymbolError is now caught.
//
// Revision 1.5  1998/06/08 15:52:36  cinard
// Bug fixed : lexical error (token manager errors) are not reported anymore
// as internal copiler errors.
//
// Revision 1.4  1998/05/26 15:16:38  bareau
// Turned debugging off.
//
// Revision 1.3  1998/04/15 13:55:23  bareau
// Added support for NL processing. NLD annotations are tokenized
// and sent to a distinct NLD parser (hence there is no more NLD_*
// productions in the ADL/TDD grammars).
// A handle for the NL engine associated to the current compiler is
// provided.
//
// Revision 1.2  1998/03/25 15:15:54  cinard
// Fixed a bug in unit testing method.
//
// Revision 1.1  1998/03/25 14:59:09  cinard
// Improved support for parsing multiple source file. The parsing of some
// source files can now be triggered while parsing or type checking
// a file.
//
// ***********************************************************************
