// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/adljava/SimpleNode.java,v $
// $Revision: 1.10 $
// $Date: 1998/06/29 16:36:20 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL Java Parser
// NAME:        SimpleNode.java
//
// HISTORY:
//      Rob Duncan, Sun Microsystems Inc. - Feb 22, 1997: Created
//      Lionel Mallet, The Open Group RI - Nov 27, 1997: Merged ADLNode and SimpleNode
//
// ************************************************************************

package ORG.opengroup.adl.parsers.adljava;

import java.io.PrintStream;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.j.*;

/**
 * The root class for the AST/Java node hierarchy.
 */
public class SimpleNode extends ADLNode implements Node, Cloneable {

  /** <I>Static</I> symbol table manager */
  static public JSymbolTableMgr stm = new JSymbolTableMgr();

  /** <I>Static</I> symbol factory */
  static public SymbolFactory sbf = stm.getSymbolFactory();

  /** The parent node of the current node */
  protected Node parent;

  /** The children nodes of the current node */
  protected Node[] children;

  /** The parser that builds the current node */
  protected ADLJavaParser parser;

  /** line number in the source file that was at the origin of the
    * construction of the current node */
  private int lineNumber = 0;

  /** column number in the source file that was at the origin of the
    * construction of the current node */
  private int columnNumber = 0;

  /** Scope of the current node */
  private Scope  scope;

  /**
    * Each node type must implement a one argument constructor to
    * create an empty node.  The new node's association table will be
    * empty, and its field properties will be set to their initial
    * values.  The node will be additive.  The node's jjtCreate() method
    * will call this constructor to build the node.
    *
    * @see #jjtCreate
    */
  public SimpleNode(int i) {
    super(i);
  }


  /**
    * Auxiliary constructor. Calls the main one and initializes
    * the "parser" field.
    */
  public SimpleNode(ADLJavaParser p, int i) {
    this(i);
    parser = p;
  }

  /**
    * Set the parameter node as the parent node of the current node
    */
  public void jjtSetParent(Node n) {
    parent = n;
  }


  /**
    * Return the current node's parent node
    */
  public Node jjtGetParent() { return parent; }


  /**
    * Add the 1st parameter node as <I>i</I>th child node of the current node
    * (with <I>i</I> the second parameter)
    */
  public void jjtAddChild(Node n, int i) {
    if (children == null) {
      children = new Node[i + 1];
    } else if (i >= children.length) {
      Node c[] = new Node[i + 1];
      System.arraycopy(children, 0, c, 0, children.length);
      children = c;
    }
    children[i] = n;
  }


  /**
    * Return the <I>i</I>th child node of the current node
    */
  public Node jjtGetChild(int i) {
    return children[i];
  }


  /**
    * Return the number of children of the current node
    */
  public int jjtGetNumChildren() {
    return (children == null) ? 0 : children.length;
  }

  /**
    * Accept a visitor.
    * @param visitor the visitor
    * @param data some data passed as an argument for the visit
    */
  public Object jjtAccept(ADLJavaParserVisitor visitor, Object data) {
    return visitor.visit(this, data);
  }

  /**
    * Let all children nodes of the current node accept a visitor
    * @param visitor the visitor
    * @param data some data passed as an argument for the visit
    */
  public Object childrenAccept(ADLJavaParserVisitor visitor, Object data) {
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
        children[i].jjtAccept(visitor, data);
      }
    }
    return data;
  }


  /**
    * Output node's id.
    *
    * You can override this method in subclasses of SimpleNode to
    * customize the way the node appears when the tree is dumped.  If
    * your output uses more than one line you should override
    * toString(String), otherwise overriding toString() is probably all
    * you need to do.
    */
  public String toString() { 
    return ADLJavaParserTreeConstants.jjtNodeName[id];
  }


  /**
    * Return a deep clone of the current node
    */
  public Object clone() throws CloneNotSupportedException {
    SimpleNode copy = (SimpleNode)super.clone();

    // create a new array of children and clone them
    if (children != null) {
      copy.children = new Node[children.length];
    }

    if (children != null) {
      SimpleNode childCopy;
      for (int i = 0; i < children.length; i++) {
        childCopy = (SimpleNode)((SimpleNode)children[i]).clone();
        copy.jjtAddChild(childCopy, i);
        childCopy.jjtSetParent(copy);
      }
    }

    return copy;
  }


  /**
    * Produce a textual representation of the nodes and properties of
    * the AST.  The level argument controls the amount of detail in the
    * output.
    */
  protected void dumpNode(PrintStream o, String label, String pad, int level)
  {
    super.dumpNode(o, label, pad, level);

    dumpNodeChildren(o, pad, level);
  }

  protected void dumpNodeChildren(PrintStream o, String pad, int level)
  {
    int n = 0;
    if (children != null) {
      for (int i = 0; i < children.length; ++i) {
	SimpleNode k = (SimpleNode)children[i];
	k.dumpNode(o, "" + n++, pad + " ", level);
      }
    }
  }


  // Line number management

  /**
    * Set the line number of the current node.
    * @see lineNumber
    */
  public void setLineNumber(int ln) {
    this.lineNumber = ln;
  }


  /** Get the current node's line number.
    * @see lineNumber
    */
  public int  getLineNumber() {
    return this.lineNumber;
  }


  // Column number management

  /**
    * Set the column number of the current node.
    * @see columnNumber
    */
  public void setColumnNumber(int cn) {
    this.columnNumber = cn;
  }


  /** Get the current node's column number.
    * @see columnNumber
    */
  public int  getColumnNumber() {
    return this.columnNumber;
  }


  // Methods necessary for tree transformation (visitors).
  // To be completed (here or in the various AST nodes)


  /**
    * Set the current node's property "name"
    */
  public void setName(String name) {
    try {
      setProperty("name", name);
    }
    catch (PropertyAccessException e) {}
  }


  /**
    * Return the property "name" of the current node
    */
  public String getName() {
    String name;
    try {
      name = (String)getProperty("name");
    }
    catch (PropertyAccessException e) {
      name = toString() + "#" + serialNumber;
    }
    return name;
  }


  /**
    * Set the property "value" of the current node as a string
    */
  public void setValue(String value) {
    try {
      setProperty("value", value);
    }
    catch (PropertyAccessException e) {}
  }


  /**
    * Set the property "value" of the current node as a long
    */
  public void setValue(long value) {
     setValue(String.valueOf(value));
  }


  /**
    * Set the property "value" of the current node as a double
    */
  public void setValue(double value) {
     setValue(String.valueOf(value));
  }


  /**
    * Set the property "value" of the current node as a char
    */
  public void setValue(char value) {
     setValue(String.valueOf(value));
  }


  /**
    * Set the property "value" of the current node as a boolean
    */
  public void setValue(boolean value) {
     setValue(String.valueOf(value));
  }


  /**
    * Return the property "value" of the current node
    */
  public String getValue() {
    String value;
    try {
      value = (String)getProperty("value");
    }
    catch (PropertyAccessException e) {
      value = toString() + "#" + serialNumber;
    }
    return value;
  }


  /**
    * Set the scope o the current node
    */
  public void setScope(Scope s) {
    this.scope = s;
  }


  /**
    * Return the current node's scope
    */
  public Scope getScope() {
    return this.scope;
  }


  /**
    * Set the operator property of the current node
    * (arithmetic or logical expression)
    */
  public void setOp(String operator) {
    try {
      setProperty("op", operator);
    } catch(PropertyAccessException e) { }
  }


  /**
    * Return the current node's operator property
    */
  public String getOp() {
    String op;
    try {
      op = (String)getProperty("op");
    } catch(PropertyAccessException e) {
      op = "<undefined operator>";
    }
    return op;
  }


  /**
    * Set the vector property of operators of the current node
    * (arithmetic or logical expression)
    */
  public void setOps(String[] operators) {
    try {
      setProperty("operators", operators);
    } catch(PropertyAccessException e) {}
  }


  /**
    * Return the current node's operator vector property
    */
  public String getOps() {
    String[] ops;
    try {
      ops = (String[])getProperty("operators");
    } catch(PropertyAccessException e) {
      ops = new String[1];
      ops[0] = "<undefined operators>";
    }
    return ops[0];
  }


}

// ***********************************************************************
// History Log:
// $Log: SimpleNode.java,v $
// Revision 1.10  1998/06/29 16:36:20  mallet
// Removed firstToken field: now in ADLNode.
//
// Revision 1.9  1998/04/01 16:32:47  mallet
// Moved common code out into parsers.ADLNode.
//
// Revision 1.8  1998/03/18 16:19:31  cinard
// Symbol table factory (sbf) is used instead of symbol table
// manager (stm) to create symbols.
//
// Revision 1.7  1998/02/27 12:53:53  cinard
// Updated copyright notice.
//
// Revision 1.6  1998/02/12 16:02:59  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.5  1998/02/11 17:24:26  bareau
// Fixed a syntax error of the previous commit (non terminated comment).
//
// Revision 1.4  1998/02/11 17:11:51  bareau
// Javadoc'ed.
//
// Revision 1.3  1998/01/05 16:09:20  mallet
// Added methods to simplify translator code (get*).
// Removed useless method tokenString.
// Fixed cloneable interface implementation.
//
// ***********************************************************************
