// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1996-1997 Sun Microsystems Inc.
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ************************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/parsers/nld/NLDParseUnit.java,v $
// $Revision: 1.3 $
// $Date: 1998/06/10 15:39:02 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   NLD parser
// NAME:        NLDParseUnit.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Apr 1 (!), 1998: Created
//
// ************************************************************************

package ORG.opengroup.adl.parsers.nld;

import java.io.*;
import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.*;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.nlengine.*;

/**
 * The NLDParseUnit class enable to monitor the parsing of an nld
 * source file or an nld string given by the adl/tdd parser.
 * <P>In addition, this class provide static methods to create and/or find
 * NLDParseUnit objects.
 */

public class NLDParseUnit
{ 

  // -- Public static methods --

  /** 
   * Get or create a NLDParseUnit object for supplied source file.
   * @param path full path name of the source file.
   * @param nl the NL engine associated with the current compiler
   */

  public static NLDParseUnit getNLDParseUnit(String path, NLEngine nl)
  {
     File   file          = new File(path);
     NLDParseUnit parseUnit = (NLDParseUnit)_parseUnits.get(file.getName());
     if ( parseUnit == null )
     {
        parseUnit = new NLDParseUnit(file, nl);
        _parseUnits.put(file.getName(),parseUnit);
     }
     return parseUnit;
  }

  // -- Public methods --

  /**
   * Test if the specified file can be parsed, that is at least exists
   * and is readable.
   */

  public boolean canParse()
  {
     return _file.canRead();
  }


  /**
    * Parse a NLD token found in a ADL/TDD file, with no scope.
    */
  public void parse(String nld) {

    parse(nld, null, 1);
  }


  /**
    * Parse a NLD token found in a ADL/TDD file, with the scope
    * where it was defined.
    */
  public void parse(String nld, Scope scope, int lineNumber) {

    if (debug) {
      System.out.println("Parsing nld block @line " + lineNumber);
    }

    _nl_engine.setScope(scope);

    NLDParser parser = null;
    try {
      parser = new NLDParser(new StringReader(nld));
      parser.disable_tracing();
      parser.setNLEngine(_nl_engine);
      parser.setFileName(_fileName);
      parser.adjustBeginLine(lineNumber-1);

      parser.NLD_Annotation();
    }
    catch ( ParseException e ) {
      _hasErrors = true;
      String message = ( e.getMessage() != null )
        ? e.getMessage() 
        : "Parse exception caught.";
      parser.errorMessage(message, parser.getToken(0).beginLine, false);
    }
  }


  /**
   * Parse a NLD file.
   */
  
  public void parse()
  {
    if (debug) {
      System.out.println("debug: NLDParseUnit.parse(" + _path + ")");
    }

     NLDParser parser = null;
     try {
       parser = new NLDParser(new BufferedInputStream(new FileInputStream(_path)));
       parser.disable_tracing();
       parser.setFileName(_path);

       parser.NLD_Annotation();
     }
     catch (FileNotFoundException e) {
       // higly improbable since the file has normally been verified
       // in Adlt main
       _hasErrors = true;
       System.err.println(_path + ": File not found!");
     }
     catch ( ParseException e ) {
       _hasErrors = true;
       String message = ( e.getMessage() != null )
         ? e.getMessage() 
         : "Parse exception caught.";
       parser.errorMessage(message, parser.getToken(0).beginLine, false);
     }

     if ( debug )
     {
        System.out.println("debug: NLDParseUnit.pureParse(" + _path + ")");
     }
  }

  /** Were errors encountered while parsing the file. */
  
  public boolean hasErrors()  { return _hasErrors; }
  
  // -- Private constructors --


  /**
   * Constructs a parse unit object. Useful to monitor parsing
   * of an NLD source file.
   * <P>Note this constructor is private. The normal way to create
   * a parse unit is simply to invoke static method getNLDParseUnit().
   * </P>
   * @param file nld source file.
   */

  private NLDParseUnit(File file, NLEngine nlengine) {
    if (debug) {
      System.out.println("NLDParseUnit constructor for " + file.getName());
    }
    _file      = file;
    _path      = file.getPath();
    _fileName  = file.getName();
    _nl_engine = nlengine;
  }


  // -- Private attributes --

  /** The NL engine associated with the currewnt compiler */
  private NLEngine _nl_engine;

  /** File object associated to the file to parse. */
  private final File _file;

  /** Last component of full path name. */
  private final String _fileName;

  /** Full path name */
  private final String _path;

  /** Does file contains errors. */
  private boolean _hasErrors = false;

  /** Map names of source files onto NLDParseUnit objects. */
  private static final HashMap _parseUnits = new HashMap();

  /** Debug flag.*/
  private static final boolean debug = false;

}

// ***********************************************************************
// History Log:
// $Log: NLDParseUnit.java,v $
// Revision 1.3  1998/06/10 15:39:02  bareau
// Modification of the grammar: in a nld block, each mapping that
// begins by "." describes a component of the current scope (e.g.
// .amt = "blabla";
// in a nld block in the scope of a method with parameter amt is a
// mapping for this parameter.
// "." alone represents the current scope.
// A mapping that does not begin by "." describes an external
// component.
//
// Revision 1.2  1998/05/15 13:54:31  bareau
// Complete revision.
//
// Revision 1.1  1998/04/15 13:52:03  bareau
// Parser for NLD annotations. Early revision, with no action.
//
// ***********************************************************************
