//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1994-1996 Sun Microsystems Inc.
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/runtime/ADLConcatenationSet.java,v $
// $Revision: 1.12 $
// $Date
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL Java Runtime Environment
// NAME:	ADLConcatenationSet.java
// 	
// HISTORY:
// 	Matt Evans, Sun Microsystems Inc. - Dec 16, 1996: Created
//      Lionel Mallet, The Open Group RI  - Oct  9, 1997: Renamed
//      See Log at end of file
// 		
// ***********************************************************************

package ORG.opengroup.adl.runtime;

import java.util.*;

/**	
 *  This class represents a TDD dataset which is defined
 *  as a concatenation of other TDD datasets. Its number of values is the
 *  sum of the number of values of its members.
 */
public class ADLConcatenationSet extends ADLDataSet {

  /** vector of elements in the concatenation set */
  private Vector elementSet = new Vector();
  /** number of elements in the concatenation set (caching the vector
   * size) */
  private int numElements = 0;

  /** cache of index of provided element after a call to provide
   * @see #provide */
  private int providedSet = -1;
  /** cache of index of provided value in provided element after a call
   * to provide
   * @see #provide */
  private int providedIndex = -1;

  /**
   * Basic constructor for named datasets: initializes
   * <code>testContext</code> and label.
   *
   * @param tc the test context in which this dataset will be used
   * @param label the name of the dataset
   */
  public ADLConcatenationSet(ADLTest tc, String label) {
    testContext = tc;
    setLabel(label);
  }

  /**
   * Adds a dataset to the concatenation: the number of values of the
   * concatenation is incremented by the number of values in the added
   * dataset.
   *
   * @param e the dataset to add to the concatenation
   */
  public void addElement (ADLDataSet e) {
    elementSet.addElement(e);
    numElements++;
    numValues += e.getNumValues();
  }

  /**
   * Provides a value from the concatenation set.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>determine the dataset that should provide the value
   * <li>call <code>provide</code> on this dataset
   * </ul>
   *
   * Note: the index of the dataset concerned by the call to provide, and
   * the index of the value in this dataset is cached for subsequent call
   * to relinquish.
   *
   * @param index the ordinal number of the combination of parameters
   * @return      the object returned by the concerned dataset
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            of the range of potential instances of the concatenation
   *            set.
   * @exception ADLRuntimeException runtime exception that should be
   *            caught in <code>ADLTest.runTestInstance()</code>.
   * @see #providedSet
   * @see #providedIndex
   */
  public Object provide(int index)
    throws ArrayIndexOutOfBoundsException, ADLRuntimeException {

    // check instance number
    checkBounds(index);

    if (locateDataSet(index)) {
      ADLDataSet d = (ADLDataSet) elementSet.elementAt(providedSet);
      return d.provide(providedIndex);
    } else {
      // we should normally never execute this code
      return null;
    }
  }

  /**
   * Relinquishes a value provided by the concatenation set.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if both the index of the providing set and the index in this
   *     set is not -1, call the actual relinquish method on the
   *     dataset
   *     (from the implementation subclass)
   * </ul>
   *
   * Note: the concerned dataset and the index in this dataset are
   * taken from the local cache.
   *
   * @param index The ordinal number of the grid point to get
   * @return      true if the relinquish succeeded, or false if it did not
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            of the range of potential instances of the concatenation
   *            set.
   * @exception ADLRuntimeException runtime exception that should be
   *            caught in <code>ADLTest.runTestInstance()</code>.
   * @see #providedSet
   * @see #providedIndex
   */
  public boolean relinquish(int index)
    throws ArrayIndexOutOfBoundsException, ADLRuntimeException {
    // check instance number
    checkBounds(index);

    if (providedSet != -1 && providedIndex != -1) {
      ADLDataSet d = (ADLDataSet)elementSet.elementAt(providedSet);
      return d.relinquish(providedIndex);
    } else {
      return true;
    }
  }

  /**
   * Checks that a test data instance number is within bounds for
   * this dataset: the superclass check does the correct job.
   *
   * @param index the test data instance number to check
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            of the range of potential instances of the concatenation
   *            set.
   */
  protected void checkBounds(int instance)
    throws ArrayIndexOutOfBoundsException {
      // check to see if there are any elements
      if (numElements == 0) {
	throw new ArrayIndexOutOfBoundsException("ADLConcatenationSet does not contain any elements");
      }
      
      // check to see if the instance request is in bounds.
      super.checkBounds(instance);
  }

  /**
   * Locates the index of the dataset containing the value referenced by
   * the given instance number.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>reset current cache values
   * <li>for each element in the concatenation set:<ul>
   * <li>if the difference between the instance number and the number of
   *     values in the element is lower than 0, then:<ul>
   * <li>cache the element index
   * <li>cache the value index (current instance number)
   * </ul><li>else decrement the instance number by the number of values
   *     in the element
   * </ul></ul>
   *
   * @param instance the instance number (should be starting at 0)
   * @return a boolean indicating whether the searched value has been
   *         located or not
   * @exception ArrayIndexOutOfBoundsException the instance number is out
   *            of the range of potential instances of the concatenation
   *            set.
   */
  private boolean locateDataSet (int instance)
    throws ArrayIndexOutOfBoundsException {

    // reset current cache
    providedSet = -1;
    providedIndex = -1;

    // perform magic computation
    for (int j = 0; j < numElements; j++) {
      ADLDataSet d = (ADLDataSet)elementSet.elementAt(j);
      int num = instance - d.getNumValues();

      if (num < 0 && instance >= 0) {
	providedSet = j;
	providedIndex = instance;
	return true;
      } else {
	instance = num;
      }
    }

    // alas, we've got there. So we've not found an appropriate set
    return false;
  }

}

// ***********************************************************************
// History Log:
// $Log: ADLConcatenationSet.java,v $
// Revision 1.12  1998/08/06 14:09:37  mallet
// Moved Java runtime source tree in /afs/gr/users/mallet/Adl/adl2/src/ORG/opengroup/adl/runtime.
//
// Revision 1.11  1998/02/27 10:08:29  mallet
// Updated copyright notice.
//
// Revision 1.10  1998/02/12 16:04:32  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.9  1998/02/11 18:20:03  mallet
// Added ADLRuntimeException in throw clause of provide and relinquish.
//
// Revision 1.8  1998/01/08 14:40:42  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.7  1998/01/05 17:17:11  mallet
// Changed visibility of constructor: now public.
//
// Revision 1.6  1997/12/23 09:11:46  mallet
// Fixed javadoc comments.
//
// Revision 1.5  1997/11/26 13:59:45  mallet
// Fixed runtime package name: now ORG.opengroup.adl.runtime.
//
// Revision 1.4  1997/10/29 13:37:52  mallet
// ADLConcatenationSet now extends ADLDataSet: removed shared code.
// Changed protection of some attributes.
//
// Revision 1.3  1997/10/15 08:40:01  mallet
// Made class public.
// Added field testContext as ADLTest.
// Added constructor to initialize both testContext and label.
// Made "out of bound" message more detailed.
//
// Revision 1.2  1997/10/10 15:42:05  mallet
// Fixed small compilation bugs.
//
// Revision 1.1  1997/10/10 13:56:49  mallet
// First implementation of TDD datasets for Java runtime.
//
//
// ***********************************************************************
