//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1994-1996 Sun Microsystems Inc.
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/runtime/ADLFactorySet.java,v $
// $Revision: 1.18 $
// $Date: 1998/08/06 14:09:38 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL Java Runtime Environment
// NAME:	ADLFactorySet.java
// 	
// HISTORY:
// 	Matt Evans, Sun Microsystems Inc. - Dec 29, 1996: Created
//      Lionel Mallet, The Open Group RI  - Oct  9, 1997: Renamed
//      See Log at end of file
// 		
// ***********************************************************************

package ORG.opengroup.adl.runtime;

/**	
 * The ADLFactorySet class defines an abstract class implementing a TDD
 * dataset which is defined as a method call, providing dynamic values
 * as opposed to static values of ADLLiteralSet. For instance, even though
 * an ADLFactorySet might have a single value, it may as well return a 
 * different value for each call to its provide method.
 * <p>This class also provides access to TET reporting by delegating
 * to the test context.
 *
 * @see ORG.opengroup.adl.runtime.ADLTest
 */
public abstract class ADLFactorySet extends ADLDataSet
implements ADLResultCodes {

  /** the factory method name as defined in the TDD file */
  protected String factoryName = null;

  /** local cache of actual factory parameters after a call to provide */
  private Object [] factoryParameters;
  /** local cache of provided object after a call to provide */
  private Object providedObject = null;

  /** the list of parameters of the factory */
  private ADLParameterSet params = new ADLParameterSet();

  /**
   * Default constructor that just initializes the test context, the label
   * and the default number of values (1).
   *
   * @param tc the test in which the factory set is used
   * @param label the name of the factory set
   */
  protected ADLFactorySet(ADLTest tc, String label) {
    testContext = tc;
    setLabel(label);
    numValues = 1; // even when there is no parameter, the factory can
    // provide at least one value
  }

  /**
   * Adds a new formal parameter for the factory: just add to the list of
   * parameters and cache the number of values
   *
   * @param p the TDD dataset to use as parameter
   */
  public void addParameter (ADLDataSet p) {
    params.addElement(p);
    numValues = params.getNumValues();
  }

  /**
   * Returns the number of formal parameters.
   */
  public int getNumParameters() {
    return params.getNumElements();
  }

  /**
   * Provides an object built from the factory from a combination of
   * parameters determined by an ordinal number.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>verify the ordinal number requested
   * <li>call <code>provide</code> on the parameters
   * <li>report the value of each parameter
   * <li>call the actual provide method (from the implementation subclass)
   * <li>set the factory result, and, in case of failure,
   *     call <code>tdd_skip</code>
   * <li>report the value returned by the actual provide method
   * </ul>
   *
   * <p>Note: the provided parameters and provided object are actually
   * cached for subsequent call to relinquish.
   *
   * @param index the ordinal number of the combination of parameters
   * @return      the object built from the factory
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            the range of potential instances of the dataset.
   * @exception ADLRuntimeException tdd_end_case or tdd_end_test has been
   *            called in provide, propagate to
   *            <code>ADLTest.runTestInstance()</code>.
   */
  public Object provide(int index)
    throws ArrayIndexOutOfBoundsException, ADLRuntimeException {

    int factoryStatus;
    int numParameters;

    // check instance number
    checkBounds(index);

    numParameters = getNumParameters();

    // get the parameter values for the factory method		
    if (numParameters > 0) {
      factoryParameters = params.provide(index);
    }

    if (testContext.debugMode) {
      String info = "Provide for factory " + factoryName + "(";

      for (int x = 0; x < numParameters; x++) {
	info += factoryParameters[x].toString();
	if (x < numParameters - 1) {
	  info += ",";
	}
      }
      info += ")";
    
      testContext.infoline(info);
    }

    // call the provide method
    try {
      providedObject = factory_provide(factoryParameters);
    }
    catch (ADLRuntimeException e) {
      // propagate an ADLRuntimeException to ADLTest.runTestInstance
      throw e;
    }
    catch (Throwable t) {
      providedObject = null;
      testContext.infoline("Caught exception during provide of factory " +
			   factoryName + ": " + t.getClass().getName() +
			   "(" + t.getMessage() + ")");
    }

    String provide_info = "*NULL*";
    if (testContext.skipAssertions) {
      // if tdd_skip has been called in provide, status is SKIPPED
      factoryStatus = ADL_SKIPPED;
    } else if (providedObject != null) {
      // provide was successful
      factoryStatus = ADL_PASS;
      provide_info = providedObject.toString();
    } else {
      // provide probably failed
      factoryStatus = ADL_FAIL;
      testContext.tdd_skip();
    }

    testContext.factoryResult(factoryStatus, 
			      "\tfactory returned: " + provide_info);
    
    return providedObject;
  }
  
  /**
   * Actual implementation of the provide clause of the factory: defined
   * in subclasses.
   *
   * @param params the array of actual parameters to the factory.
   * @exception Throwable exception that occured in the provide clause
   *            of the factory
   */
  abstract protected Object factory_provide(Object [] params) throws Throwable;
  
  /**
   * Relinquishes the object built from a factory with a combination of
   * parameters determined by an ordinal number.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>verify the ordinal number requested
   * <li>if the provided object is not null, call the actual relinquish
   *     method, passing the actual paramaters of the factory and the
   *     provided object
   * <li>then call <code>relinquish</code> on the parameters
   * <li>set the factory relinquish result
   * <li>reset the cache of actual parameters and provided object
   * <li>report the result of the relinquish call
   * </ul>
   *
   * <p>Note: the actual parameters and provided object are taken from the
   * local cache, which is then reset.
   *
   * @param index the ordinal number of the grid point to get
   * @return      true if the relinquish succeeded, or false if it did not
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            the range of potential instances of the dataset.
   * @exception ADLRuntimeException tdd_end_case or tdd_end_test has been
   *            called in relinquish, propagate to
   *            <code>ADLTest.runTestInstance()</code>.
   */
  public boolean relinquish(int index)
    throws ArrayIndexOutOfBoundsException, ADLRuntimeException {

    boolean factoryStatus = true;

    // check instance number
    checkBounds(index);

    if (testContext.debugMode)
      testContext.infoline("Relinquish for factory " + factoryName);

    if (providedObject != null) {
      try {
	factoryStatus = 
	  factory_relinquish(factoryParameters, providedObject);
	if (getNumParameters() > 0) {
	  factoryStatus &= params.relinquish(index);
	}
      }
      catch (ADLRuntimeException e) {
	// propagate an ADLRuntimeException to ADLTest.runTestInstance
	throw e;
      }
      catch (Throwable t) {
	factoryStatus = false;
	testContext.infoline("Caught exception during relinquish of factory " +
			     factoryName + ": " + t.getClass().getName() +
			     "(" + t.getMessage() + ")");
      }

      providedObject = null;
    }

    factoryParameters = null;

    if (testContext.debugMode)
      testContext.infoline("\tfactory returned: " + factoryStatus);

    return factoryStatus;
  }
   
  /**
   * Actual implementation of the relinquish method.
   * 
   * <p>Note: this is currently empty instead of abstract because the
   * relinquish clause is optional.
   *
   * @param params the array of actual parameters passed to the
   *               corresponding provide method
   * @param obj    the object returned by the call to the provide method
   * @return a boolean status to indicate a failure (false)
   * @exception Throwable exception that occured in the relinquish clause
   *            of the factory
   */
  protected boolean factory_relinquish(Object [] params, Object obj) throws Throwable {
    return true;
  }

  // methods to be used in factory method (To Be Completed)

  /**
   * Writes an informative entry in the journal: forward to the
   * testContext.
   *
   * @param line	a String to write
   * @see ORG.opengroup.adl.runtime.ADLTest#infoline
   */
  public void infoline(String line) {
    testContext.infoline(line);
  }

  /**
   * Reports a request for ending the current test case and requests
   * remaining assertions to be skipped: forward to the testContext.
   *
   * @param msg the explanation message (can be null)
   * @exception ADLEndCaseException the current test case should be ended
   * @return true
   * @see ORG.opengroup.adl.runtime.ADLTest#tdd_end_case
   */
  public boolean tdd_end_case(String msg) throws ADLEndCaseException {
    return testContext.tdd_end_case(msg);
  }

  /**
   * Indicates a factory failure and requests remaining assertions to be
   * skipped: forward to the testContext.
   *
   * @return true
   * @see ORG.opengroup.adl.runtime.ADLTest#tdd_skip
   */
  public boolean tdd_skip() {
    return testContext.tdd_skip();
  }

  /**
   * Reports a request for ending the current test: forward to the
   * testContext.
   *
   * @exception ADLEndTestException the current test should be ended
   * @return true
   * @see ORG.opengroup.adl.runtime.ADLTest#tdd_end_test
   */
  public boolean tdd_end_test() throws ADLEndTestException {
    return testContext.tdd_end_test();
  }

  /**
   * Records a result from within a TDD test clause and an explanation
   * for that result: forward to the testContext.
   *
   * @param code the result code
   * @param msg	 the explanation to record
   * @return true
   * @see ORG.opengroup.adl.runtime.ADLTest#tdd_result
   * @see ORG.opengroup.adl.runtime.ADLResultCodes
   */
  public boolean tdd_result (int code, String msg) {
    return testContext.tdd_result(code, msg);
  }

}

// ***********************************************************************
// History Log:
// $Log: ADLFactorySet.java,v $
// Revision 1.18  1998/08/06 14:09:38  mallet
// Moved Java runtime source tree in /afs/gr/users/mallet/Adl/adl2/src/ORG/opengroup/adl/runtime.
//
// Revision 1.17  1998/05/28 15:04:44  mallet
// Modified a bit the code of relinquish to avoid processing the info
// string when not in "Long" report mode.
// Other minor cosmetic changes.
//
// Revision 1.16  1998/05/12 14:22:45  shane
// Added code to hide certain output unless debugMode is enabled.
// Also added debugMode to the testContext.
//
// Revision 1.15  1998/02/27 10:08:30  mallet
// Updated copyright notice.
//
// Revision 1.14  1998/02/12 16:04:34  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.13  1998/02/11 18:23:29  mallet
// Added handling of ADLRuntimeException in provide and relinquish to
// allow tdd_end_case and tdd_end_test to act as expected: exception is
// propagated.
// Added test so that factory status is not overriden when tdd_skip has
// been called.
//
// Revision 1.12  1998/01/22 09:50:25  mallet
// Removed tdd_end_case() method: only keep tdd_end_case(String).
//
// Revision 1.11  1998/01/21 16:56:46  mallet
// Updated tdd_end_case and tdd_end_test methods signature: now throw an ADLEndCaseException/ADLEndTestException.
//
// Revision 1.10  1998/01/08 14:40:43  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.9  1997/12/23 09:15:32  mallet
// Fixed javadoc comments.
// Changed factory_relinquish to provided a default implementation
// instead of being abstract: the default implementation just returns true.
// Removed 'throws' clause for tdd_* methods.
//
// Revision 1.9  1997/12/23 08:43:47  mallet
// Fixed javadoc comments.
// Changed factory_relinquish to provided a default implementation
// instead of being abstract: the default implementation just returns true.
//
// Revision 1.8  1997/11/26 13:59:46  mallet
// Fixed runtime package name: now ORG.opengroup.adl.runtime.
//
// Revision 1.7  1997/11/21 15:19:12  mallet
// Modified the way the factory reports on provide/relinquish calls to
// make it more similar to the C++ ADLFactorySet class reporting.
// Added an infoline reporting method so that it can be used in a factory
// body.
//
// Revision 1.6  1997/11/14 09:27:50  mallet
// Fixed the number of values returned by a factory without parameter.
// Added some javadoc comments.
//
// Revision 1.5  1997/10/29 13:40:08  mallet
// ADLFactorySet now extends ADLDataSet: removed shared code.
// ADLFactorySet also implements ADLResultCodes: fixed access to result
// codes.
//
// Revision 1.4  1997/10/15 08:44:00  mallet
// Added constructor to initialize both testContext and label.
// Removed unused method setTestContext.
// Fixed type of caught data (now Throwable), made emitted message more
// detailed (for both provide and relinquish methods).
// Fixed sugnature of abstract methods factory_provide and
// factory_relinquish.
// Made "out of bounds" message more detailed.
//
// Revision 1.3  1997/10/10 20:21:24  shane
// Fixed some typos to get it to compile.
// Changed the tdd_result method to take an int code instead of a String code.
//
// Revision 1.2  1997/10/10 15:42:06  mallet
// Fixed small compilation bugs.
//
// Revision 1.1  1997/10/10 13:56:50  mallet
// First implementation of TDD datasets for Java runtime.
//
//
// ***********************************************************************
