//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// Copyright (c) 1994-1996 Sun Microsystems Inc.
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/runtime/ADLParameterSet.java,v $
// $Revision: 1.16 $
// $Date: 1998/08/06 14:09:39 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL Java Runtime Environment
// NAME:	ADLParameterSet.java
// 	
// HISTORY:
// 	Matt Evans, Sun Microsystems Inc. - Dec 9, 1996: Created
//      Lionel Mallet, The Open Group RI  - Oct  9, 1997: Renamed
//      See Log at end of file
// 		
// ***********************************************************************


package ORG.opengroup.adl.runtime;

import java.util.*;

/**	
 * The ADLParameterSet Implements a list of dataset parameters for test
 * directive and factory parameters.
 *
 * @see ORG.opengroup.adl.runtime.ADLDataSet
 * @see ORG.opengroup.adl.runtime.ADLTest
 * @see ORG.opengroup.adl.runtime.ADLFactorySet
 */
public class ADLParameterSet {

  /** the list of dataset that makes the parameters */
  protected Vector elementSet = new Vector();
  /** the number of parameters */
  protected int numElements = 0;
  /** the number of different combinations of the parameters
   * (1 by default since even a an empty parameter list has one
   * combination) */
  protected int numValues = 1;

  /** local cache of the combination of index into the each parameter
   * after a call to provide */
  private int[] elementIndexes;

  /** local cache of the array of maximum number of instances for each
   * parameter */
  private int[] maxInstanceArray = null;

  /**
   * Returns the number of values the parameter list can yield, that is
   * the number of different combinations of the paramters.
   */
  public int getNumValues() {
    return numValues;
  }

  /**
   * Adds a new TDD dataset in the parameter list: the number of values
   * of the parameter list is multiplied by the number of values of the
   * added dataset.
   *
   * @param e the TDD dataset to add
   */
  public void addElement (ADLDataSet e) {

    elementSet.addElement(e);
    numElements++;
    numValues *= e.getNumValues();
  }

  /**
   * Returns the number of parameters (TDD datasets) in the parameter list.
   */
  public int getNumElements() {
    return numElements;
  }

  /**
   * Provides the set of data corresponding to a combination of the
   * parameters given by its ordinal number as an array of Object.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>verify the ordinal number of the combination
   * <li>compute the indexes for each of the parameters
   * <li>call <code>provide</code> on each parameter with its
   *     corresponding index
   * </ul>
   *
   * <p>Note: the indexes for each of the parameters is cached for future
   * call to relinquish.
   *
   * @param index the ordinal number of the combination to get
   * @return      an array of Object containing the set of data values
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            the range of potential combinations of the parameters.
   * @exception ADLRuntimeException runtime exception that should be
   *            caught in <code>ADLTest.runTestInstance()</code>.
   */
  public Object[] provide(int index) throws
    ArrayIndexOutOfBoundsException, ADLRuntimeException {

    checkBounds(index);

    elementIndexes = computeIndexes(index);

    Object[] ret = new Object[numElements];

    for (int x = 0; x < numElements; x++) {
      ADLDataSet data = (ADLDataSet) elementSet.elementAt(x);
      ret[x] = data.provide(elementIndexes[x]);
    }

    return ret;
  }

  /**
   * Relinquishes the set of data corresponding to a combination of the
   * parameters given by its ordinal number.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>verify the ordinal number of the combination
   * <li>call <code>relinquish</code> on each parameter with its
   *     corresponding index
   * </ul>
   *
   * <p>Note: the indexes for each of the parameters is taken from the
   * local cache.
   *
   * @param index the ordinal number of the grid point to get
   * @return      true if the relinquish succeeded, or false if it did not
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            the range of potential combinations of the parameters.
   * @exception ADLRuntimeException runtime exception that should be
   *            caught in <code>ADLTest.runTestInstance()</code>.
   */
  public boolean relinquish(int index) throws
    ArrayIndexOutOfBoundsException, ADLRuntimeException {

    checkBounds(index);

    boolean result = true;

    for (int x = 0; x < numElements; x++) {
      ADLDataSet data = (ADLDataSet)elementSet.elementAt(x);
      result &= data.relinquish(elementIndexes[x]);
    }
    return result;
  }

  /**
   * Returns the array of maximum number of instances per parameter.
   *
   * @return an array of int containing the maximum number of
   *         instances of each parameter
   */
  private int [] computeArrayofMaxInstances() {

    int [] elementInstanceValues = new int [numElements];

    for (int x = 0; x < numElements; x++) {
      ADLDataSet elem = (ADLDataSet) elementSet.elementAt(x);
      elementInstanceValues[x] = elem.getNumValues();
    }

    return elementInstanceValues;
  }

  /**
   * Computes the index, for each parameter, of the value spotted by a
   * combination of the parameters given by its ordinal number.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>verify the ordinal number of the combination
   * <li>if there is no parameter in the list, return null
   * <li>get the array of maximum instances for each parameter
   * <li>for each parameter, starting from the last parameter:<ul>
   * <li>the index in this paramater is the remaining of the integral
   *     division of the instance number and the maximum number of values
   *     of the parameter
   * <li>the instance number is integrally divided by the maximum number
   *     of values of the parameter
   * </ul><li>if finally the instance number is not equal to 0, there must
   *     have been an error
   * </ul>
   *
   * <p><b>Hypothesis:</b>the parameters vary from the rightmost first to
   * the leftmost last.
   *
   * <p>Note: the instance request should be starting at 0.
   *
   * @param instance the ordinal number of the combination of the
   *        parameters
   * @return an array of int containing the index, for each parameter,
   *         of the value spotted by the combination of the parameters,
   *         or null if the parameter list is empty
   */
  private int[] computeIndexes (int instance) {

    // check to see if the instance request is in bounds.
    checkBounds(instance);

    // check to see if there are any elements
    if (numElements == 0) {
      return null;
    }

    // compute the array of element max instance values (but only once)
    if (maxInstanceArray == null)
      maxInstanceArray =  computeArrayofMaxInstances();

    // allocate array of computed indexes.
    int[] ret = new int[numElements];

    // perform magic computation
    for (int j = numElements - 1; j >= 0; j--) {
      ret[j] = (instance % maxInstanceArray[j]);
      instance /= maxInstanceArray[j];

    }

    // final check
    if (instance != 0) {
      throw new ArrayIndexOutOfBoundsException("instance != 0");
    }
    return ret;
  }

  /**
   * Verifies that a given ordinal number is within the bounds of
   * the number of combinations of the parameters in the list.
   *
   * @param index the ordinal number to verify
   * @exception ArrayIndexOutOfBoundsException the index is out of
   *            the range of potential combinations of the parameters.
   */
  private void checkBounds(int index) throws
    ArrayIndexOutOfBoundsException {
    if (index < 0 || index >= numValues) {
      throw new ArrayIndexOutOfBoundsException("ADLParameterSet: index out of range (" + index + "/" + numValues + ")");
    }
  }
}

// ***********************************************************************
// History Log:
// $Log: ADLParameterSet.java,v $
// Revision 1.16  1998/08/06 14:09:39  mallet
// Moved Java runtime source tree in /afs/gr/users/mallet/Adl/adl2/src/ORG/opengroup/adl/runtime.
//
// Revision 1.15  1998/02/27 10:08:31  mallet
// Updated copyright notice.
//
// Revision 1.14  1998/02/12 16:04:34  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.13  1998/02/11 18:19:24  mallet
// Added ADLRuntimeException in throw clause of provide and relinquish.
//
// Revision 1.12  1998/01/08 14:38:53  mallet
// Updated and fixed javadoc comments.
// Added a local cache for array of max instance numbers to avoid
// computing this array for each call to provide.
//
// Revision 1.11  1997/12/23 09:11:47  mallet
// Fixed javadoc comments.
//
// Revision 1.10  1997/11/26 13:59:46  mallet
// Fixed runtime package name: now ORG.opengroup.adl.runtime.
//
// Revision 1.9  1997/11/14 09:27:15  mallet
// Fixed some comments on methods.
//
// Revision 1.8  1997/10/29 13:35:42  mallet
// Removed useless constructor ADLParameterSet(ADLDataSet[] elem).
//
// Revision 1.7  1997/10/15 08:50:36  mallet
// Fixed default value for numValues: when there is no parameter, the
// number of values if 1, not 0.
// Fixed the computation of numValues in addParameter accordingly.
// Added a private checkBounds method to replace redundant code.
// Made "out of bounds" message more detailed.
// Fixed bug in provide: was using a local variable elementIndexes
// instead of cache class field.
//
// Revision 1.6  1997/10/10 20:22:26  shane
// Fixed a typo in the comments.
//
// Revision 1.5  1997/10/10 15:42:06  mallet
// Fixed small compilation bugs.
//
// Revision 1.4  1997/10/10 15:03:58  mallet
// Fixed two small bugs (unknown variables).
//
// Revision 1.3  1997/10/10 13:56:50  mallet
// Class ADLTestProductSet has been renamed as ADLParameterSet.
// Completely revamped according to new TDD dataset classes design.
//
// Revision 1.2  1997/10/08 12:37:01  mallet
// Cosmetic changes to add standard footer and reformat header.
//
// ***********************************************************************
