//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/runtime/ADLTest.java,v $
// $Revision: 1.22 $
// $Date: 1998/08/06 14:09:40 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	ADL Java Runtime Environment
// NAME:	ADLTest.java
// 	
// HISTORY:
// 	Shane P. McCarron, The Open Group - Sep 20, 1997: Created
//      See Log at end of file
// 		
// ***********************************************************************

package ORG.opengroup.adl.runtime;

import java.util.*;
import java.io.*;

/**
 * The ADLTest class is an abstract class that acts as the primary
 * interface to the system for generated test cases. A generated test
 * case should extend this class, providing a main routine that calls
 * the provided default constructor. This constructor will in turn
 * activate the class and set up a TET test session by calling
 * <code>testCase.main()</code>.
 *
 * @see ORG.opengroup.adl.runtime.ADLTestCase#main
 */
public abstract class ADLTest implements ADLResultCodes {

  /** reference to the test session established when test is started */
  private TET.TestSession testSession;

  /** list of parameters for the test */
  private ADLParameterSet params = new ADLParameterSet();

  /** reference to the ADLTestCase linking this test to TET */
  static protected ADLTestCase testCase = new ADLTestCase();

  /** Tmp variable used in directives main method, to work-around a JET/TET bug. */
  static protected ADLTest mainADLTest;

  /**
   * boolean indicator allowing a factory or test directive code to
   * indicate whether an ic has to be skipped or not
   */
  public boolean skipAssertions = false;


  /**
   * debug mode indicator allowing to customize the level of reporting
   */
  public boolean debugMode = false;


  /**
   * Adds a parameter to the test data set for the test.
   *
   * This method should be called during test case startup by the main
   * routine of a subclass of this class - once for each test data
   * parameter.
   *
   * @param p the ADLDataSet to add
   */
  public void addParameter (ADLDataSet p) {
    params.addElement(p);
  }

  /**
   * Returns the total number of test data instances for this test
   * directive: forward call to <code>params</code>.
   */
  public int getNumInstances() {
    return params.getNumValues();
  }

  /**
   * Executes a specific test instance: this method is only called by
   * the ADLTestCase driver.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>record the current TET test session
   * <li>reset <code>skipAssertions</code> to false
   * <li>call <code>provide</code> on the parameters
   * <li>call <code>runInstance</code> passing the provided parameters
   * <li>finally call <code>relinquish</code> on the parameters
   * </ul>
   *
   * @param ts          the current TET test session
   * @param testIndex	the test number to execute
   * @see ORG.opengroup.adl.runtime.ADLParameterSet#provide
   * @see ORG.opengroup.adl.runtime.ADLParameterSet#relinquish
   * @see #runInstance
   */
  public void runTestInstance(TET.TestSession ts, int testIndex) {

    // Store the test session for use during this test
    testSession = ts;

    // be sure skipAssertions is set to default value before provide
    skipAssertions = false;

    // initialize debugMode (we'd better find a way to initialize it
    // *once* for all test instances)
    String debugString = getvar("ADL_RPT_DETAIL");
    debugMode = (debugString != null &&
		 !debugString.equalsIgnoreCase("SHORT"));

    try {
      // get the parameter values for the test method		
      Object[] parameters = params.provide(testIndex);

      // call the test directive
      runInstance(parameters);
    }
    catch (ADLEndCaseException e) {
      this.setblock();
      this.infoline(e.getMessage());
    }
    catch (ADLEndTestException e) {
      this.setblock();
      this.infoline(e.getMessage());
      // should also notify TET to end the test execution
    }
    catch (Throwable t) {
      if (debugMode) {
	this.setblock();
	this.infoline("Caught exception: " + t.getClass().getName() + 
		      "(" + t.getMessage() + ")");
      }
    }
    finally {
      this.setblock();
      try {
	// clean up after the test instance
	if (!params.relinquish(testIndex)) {
	  this.infoline("Relinquish failed!");
	}
      }
      catch (ADLEndTestException e) {
	this.infoline(e.getMessage());
	// should also notify TET to end the test execution
      }
      catch (Throwable t) {
	// catch all other exceptions but ignore them, just report
	this.infoline("Caught exception: " + t.getClass().getName() + 
		      "(" + t.getMessage() + ")");
      }
    }
  }

  /**
   * Abstract test directive: to be implemented in subclasses.
   *
   * @param params the parameters as an array of Object.
   */
  protected abstract void runInstance(Object[] params) throws Throwable;

  // Methods for working with the test harness

  /**
   * Writes an informative entry in the journal.
   *
   * @param line	a String to write
   */
  public void infoline(String line) {
    testSession.tet_infoline(line);
  }

  /**
   * Get a TET configuration variable value.
   *
   * @param name the name of the variable to fetch
   * @return	 a String representing the value of the variable, 
   *             or <code>null</code> if the variable is undefined.
   */
  public String getvar(String name) {
    return testSession.tet_getvar(name);
  }

  /**
   * Set the block ID for TET.
   *
   * <p>Note: this method should only be called by <code>assertionStart</code>
   * or ACO code.
   */
  public void setblock() {
    testSession.tet_setblock();
  }

  /**
   * Reports an intermediate result to TET.
   *
   * @param res	the result to record
   * @see ORG.opengroup.adl.runtime.ADLResultCodes
   */
  public void result(int res) {
    testSession.tet_result(res);
  }

  /**
   * Reports the beginning of the test of an assertion: set block
   * counter first to be able to distinguish from previous assertion.
   *
   * <p>Note: this method should only be called by ACO code.
   *
   * @param assertion	the text of the assertion being tested
   */
  public void assertionStart(String assertion) {
    if (debugMode) {
      this.setblock();
      this.infoline("Checking assertion: " + assertion);
    }
  }

  /**
   * Reports the result of an assertion.
   *
   * <p>Note: this method should only be called by ACO code.
   *
   * @param res	the result number to use
   * @see ORG.opengroup.adl.runtime.ADLResultCodes
   */
  public void assertionResult(int res) {
    if (debugMode)
	this.infoline("result is " + res);
    this.result(res);
  }

  /**
   * Records a result for a factory and an explanation for that result.
   *
   * <p>Note: this method should only be called by ACO code.
   *
   * @param res	the result code
   * @param msg	the explanation to record
   * @see ORG.opengroup.adl.runtime.ADLResultCodes
   */
  public void factoryResult(int res, String msg) {
    if (debugMode)
	this.infoline(msg);
    this.result(res);
  }

  /**
   * Records a result from within a TDD test clause and an explanation
   * for that result.
   *
   * @param res	the result code
   * @param msg	the explanation to record
   * @return true
   * @see ORG.opengroup.adl.runtime.ADLResultCodes
   */
  public boolean tdd_result(int res, String msg) {
    if (debugMode)
	this.infoline(msg);
    this.result(res);
    return true;
  }

  /**
   * Executes and reports a TDD level assertion.
   *
   * @param assertText the text of the assertion being tested
   * @param assertRes  the result of the assertion being tested
   * @return true
   */
  public boolean tdd_assert(String assertText, boolean assertRes) {
    if (debugMode) {
      this.setblock();
      this.infoline("Checking tdd assertion: " + assertText);
    }
    this.assertionResult(assertRes ? ADL_PASS : ADL_FAIL);
    return assertRes;
  }
    
  /**
   * Indicates a factory failure and requests remaining assertions to be
   * skipped.
   *
   * @return true
   */
  public boolean tdd_skip() {
    this.infoline("Skipping test case due to factory failure");
    this.result(ADL_SKIPPED);

    skipAssertions = true;

    return true;
  }

  /**
   * Reports a request to end the current test case.
   *
   * @param msg the explanation message (can be null)
   * @exception ADLEndCaseException the current test case should be ended
   * @return true
   */
  public boolean tdd_end_case(String msg) throws ADLEndCaseException {
    // skip remaining assertions (just in case the thrown exception is
    // catched by the user code)
    skipAssertions = true;

    throw new ADLEndCaseException(msg);
//     return true;
  }

  /**
   * Reports a request to end the current test.
   *
   * @exception ADLEndTestException the current test should be ended
   * @return true
   */
  public boolean tdd_end_test() throws ADLEndTestException {
    // skip remaining assertions (just in case the thrown exception is
    // catched by the user code)
    skipAssertions = true;

    throw new ADLEndTestException();
//     return true;
  }


}

// ***********************************************************************
// History Log:
// $Log: ADLTest.java,v $
// Revision 1.22  1998/08/06 14:09:40  mallet
// Moved Java runtime source tree in /afs/gr/users/mallet/Adl/adl2/src/ORG/opengroup/adl/runtime.
//
// Revision 1.21  1998/05/28 15:01:35  mallet
// Added initialization of debugMode indicator through getvar.
// Added some setblock() calls to have correct block numbers during
// relinquish or tdd_assert for instance.
// Added some "if (debigMode)" selectors to reduce the size of the
// journal file when not in "Long" report mode.
//
// Revision 1.20  1998/05/12 14:22:46  shane
// Added code to hide certain output unless debugMode is enabled.
// Also added debugMode to the testContext.
//
// Revision 1.19  1998/05/06 11:48:03  mallet
// Fix bug in tdd_assert: assertion text was never reported to the user.
//
// Revision 1.18  1998/03/02 17:36:44  bareau
// Added a tmp static variable "mainADLTest" used in directives "main" method,
// to work-around a JET/TET bug.
//
// Revision 1.17  1998/02/27 10:08:32  mallet
// Updated copyright notice.
//
// Revision 1.16  1998/02/12 16:04:35  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.15  1998/02/11 18:17:47  mallet
// Changed tdd_skip to set result to ADL_SKIPPED: was ADL_UNINITIATED.
// Updated runTestInstance so that exceptions raised in provide and
// relinquish are caught.
//
// Revision 1.14  1998/01/22 09:49:41  mallet
// Changed signature of runInstance method: now throws Throwable.
// Added processing of thrown exceptions in runTestInstance method to
// handle "end_case" and "end_test" and relieve generated code from this
// processing.
// Updated tdd_end_case and tdd_end_test to avoid redundant reporting.
//
// Revision 1.13  1998/01/21 17:04:57  mallet
// Removed unreachable code.
//
// Revision 1.12  1998/01/21 16:55:30  mallet
// Fixed bug in tdd_assert: used to return true instead of assertion
// result.
// Implemented tdd_end_case and tdd_end_test so that they throw an
// exception: currently only tdd_end_case works since nothing is done at
// the test driver level for tdd_end_test. It is currently equivalent to
// tdd_end_case. Implementing tdd_end_test may require some modifications
// in JET/TET.
//
// Revision 1.11  1998/01/08 14:40:45  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.10  1997/12/23 09:11:48  mallet
// Fixed javadoc comments.
//
// Revision 1.9  1997/11/26 13:59:47  mallet
// Fixed runtime package name: now ORG.opengroup.adl.runtime.
//
// Revision 1.8  1997/11/21 15:17:00  mallet
// Modified scope of the TET.TestSession and ADLParameterSet members:
// made them private.
//
// Revision 1.7  1997/11/14 09:24:36  mallet
// Added a static ADLTestCase member: this simplifies a little bit the
// generated code.
//
// Revision 1.6  1997/10/29 13:46:35  mallet
// Moved out result codes into ADLResultCodes interface: ADLTest
// implements this interface to get access to these (fixed their use as
// well).
//
// Revision 1.5  1997/10/15 08:55:37  mallet
// Made all status code fields public.
// Removed the public abstract method main: should be static but Java
// does not allow static abstract methods.
// Added tdd_assert method for TDD level assertions.
// Added setting of skipAssertions field in tdd_skip.
// Moved setblock call before "Checking assertion..." message in
// assertionStart.
//
// Revision 1.4  1997/10/10 20:24:57  shane
// Added tdd_ methods needed by ADLFactory...
// Added abstract main.
// Changed to use new ADLParameterData class.
// Generally made it compile.
//
// Revision 1.3  1997/10/08 12:34:12  mallet
// Interim version to be completed.
// Cosmetic changes to add standard header and footer.
// Defined result codes.
// Defined abstract method theTestDirective.
// Added boolean to indicate the ACO that the ic is skipped.
// Fixed bug in method result (cannot call itself!).
// Added method assertionStart to report on beginning of assertion check.
// Removed setblock call in method assertionResult.
//
// ***********************************************************************
