// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/ADLClass.java,v $ 
// $Revision: 1.21 $
// $Date: 1998/06/22 14:12:38 $ 
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        ADLClass.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import com.objectspace.jgl.UnaryPredicate;
import java.util.Vector;
import java.util.Enumeration;

/**
 * ADL Class symbol.
 * ADL superclasses have a special handling mechanism. Since getting 
 * information about an adl extended class is heavy (it requires
 * parsing the adl superclass source file) adl superclasses are parsed
 * only when required (e.g. to check for adl inherited semantics).
 * In the meantime, only the names of the adl superclasses are
 * stored in  the adl class symbol (method addSuperClassName()).
 * When an adl superclass is parsed, then the symbol is updated
 * through a call to addSuperClass() method.
 * In the future, a closer collaboration in between adl class 
 * symbol and the parser could be set up to avoid this two pass
 * mechanism (setting names then classes) but this would add a 
 * dependency in between symbols and the parsers which is not very
 * clean.
 */

public class  ADLClass extends LConstructedTypeImpl 
                       implements FileScope
{

   // -- Constructors --

   /**
    * Constructs an ADL class to supply semantics for the target language
    * class provided.
    * @param testedClass       the target language class for which semantics 
    *                          is to be defined.
    * @param adlSuperClassName name of the adl superclass
    */

   public ADLClass(LClass          testedClass,
                   String          adlSuperClassName) 
   {
      super
      ( ( ( testedClass != null ) ? testedClass.getName() : "adl" ),
       new LModifier(0x0000),
       SymbolId.ADL_CLASS
      );

      _testedClass       = testedClass;

      if ( _testedClass != null )
	_fullName = _testedClass.getFullName();

      if ( adlSuperClassName != null && adlSuperClassName.length() > 0 )
	_adlSuperClassNames.addElement(adlSuperClassName);

   }


  // -- Public methods --

  /**
   * Get a symbol in the scope, extended version.
   * <p>When using this version of getSymbol, supplementary parameters are
   * provided that shall help finding the object.
   * <pre>
   * The symbol shall satisfy some criteria defined by the provided 
   * predicate object. This predicate object help finding the symbol in 
   * case several symbols are likely to match the supplied name 
   * (e.g. overloaded methods).
   * </pre>
   * <pre>
   * Credentials shall be supplied for those cases where some symbols
   * hava non-public visibility. Retrieving them may require the
   * use of credentials.
   * </pre>
   * @param name The name of the expected symbol.
   * @param predicate An arbitrary predicate object.
   * @param credentials Credentials of the scope.
   * @return The found symbol.
   */

  public Symbol getSymbol
         (String name,UnaryPredicate predicate,Object credentials)
  {
     // Check that the requested symbol is not the tested class itself
     // or keyword this.
     if ( name.equals(getName()) )
     {
        // It is, return it if verifies the predicate.
        if ( predicate == null || predicate.execute(_testedClass) )
        {
           return _testedClass;
        }
     }

     // Search in class fields

     Symbol s = super.getSymbol(name,predicate,credentials);

     // TBD Set credentials

     // Check target language class in case no symbol was found.

     if ( s == null && _testedClass != null )
     {
        s = _testedClass.getSymbol(name,predicate,credentials);
     }
     
     // Return it.
     return s;
 
  }
 
   /**
    * Set the target language class for which semantics is to be supplied.
    * @param testedClass The target language class for which semantics is to
    *                    be defined.
    */

   public void setLClass(LClass testedClass) 
   {
      setName(testedClass.getName());
      _testedClass = testedClass;
      _fullName = testedClass.getFullName();
   }

   /**
    * Get the target language class for which semantics is to be supplied.
    * @return The target language class for which semantics is to be 
    *         defined.
    */

   public LClass getLClass() 
   {
      return _testedClass;
   }

  /**
   * Has adl class a superclass with supplied name?
   */

  public boolean hasSuperClass(String name)
  {
    return _adlSuperClassNames.contains(name);
  }

   /**
    * Add an adl superclass name.
    * @param adlSuperClassName Name.
    */

   public void addSuperClassName(String adlSuperClassName) 
   {
      _adlSuperClassNames.addElement(adlSuperClassName);
   }

   /**
    * Get the name of the ADL superclass of this ADL class.
    * @return Name.
    */

   public String getSuperClassName() 
   {
      return   ( _adlSuperClassNames.size() > 0 ) 
             ? (String)_adlSuperClassNames.elementAt(0) 
             : null ;
   }

   /**
    * Get the names of the ADL superclasses of this ADL class.
    * @return Name.
    */

   public String[] getSuperClassNames() 
   {
     String tmp[] = new String[_adlSuperClassNames.size()];
     _adlSuperClassNames.copyInto(tmp);
     return tmp;
   }

  /**
   * Add a superclass for this ADL class (for inheritance of semantics)
   * @param adlSuperClass Superclass for this ADL class.
   */

   public void addSuperClass(ADLClass adlSuperClass) 
   {
      _adlSuperClasses.addElement(adlSuperClass);
   }
  
   /**
    * Get the superclass for this ADL class (for inheritance of semantics)
    * @return Superclass for this ADL class.
    */

   public ADLClass getSuperClass() 
   {
     return   ( _adlSuperClasses.size() > 0 ) 
            ? (ADLClass)_adlSuperClasses.elementAt(0) 
            : null ;
   }

   /**
    * Get the superclasses for this ADL class (for inheritance of semantics)
    * @return Superclass for this ADL class.
    */

   public ADLClass[] getSuperClasses() 
   {
     ADLClass[] tmp = new ADLClass[_adlSuperClasses.size()];
     _adlSuperClasses.copyInto(tmp);
     return tmp;
   }


   /**
    * Check the class defined by this symbol is a subclass of the
    * class defined by the supplied symbol.
    * @param t Type symbol.
    * @return True if it is, false otherwise.
    */

   boolean subclassOf(LType t)
   {
     boolean isSubclass = false;
     Enumeration e = _adlSuperClasses.elements();
     while ( e.hasMoreElements() )
     {
       ADLClass adlSuperClass = (ADLClass)e.nextElement();
       isSubclass =  adlSuperClass.equals(t) || adlSuperClass.subclassOf(t);
       if ( isSubclass ) break;
     }
     return isSubclass;
   }

  // -- Symbol interface implementation --

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
     return SymbolSelector.ADL_CLASS_SELECT;
  }

  /**
    * Get the scope name in NL format.
    * Here, this is in fact the associated LClass scope name.
    */
  public Vector getNLName() {

    Scope parent = ( getLClass() != null ) 
                 ? getLClass().getParent()
                 : null ;
    Vector ret = (parent == null) ? new Vector() : parent.getNLName();
    ret.addElement(getName());
    return ret;
  }
  
  /**
   *  Get fully qualified name.
   */

  public String getFullName() 
  {
     return ( _fullName != null ) ? _fullName : getName() ;
  }
  
  // -- Protected methods --

  /**
   * Returns a header for the description of the symbol.
   */
  
  protected String header()
  {
     String s = "";
     
     for ( int i = 0; i < _adlSuperClassNames.size(); i++)
     {
       if ( i == 0 ) s += " inherits ";
       else          s += ", "; 
       s += (String) _adlSuperClassNames.elementAt(i);
     }

     // Stringified class
     
     return   "adl class " + getFullName() + s;
  }
  
  // -- Private attributes --
  
  /** Target language class for which this ADL class is defined.  */

  private LClass _testedClass = null; 

  /** Name of the ADL class being extended */
  
  private Vector _adlSuperClassNames = new Vector();
  
  /** ADL class being extended */

  private Vector _adlSuperClasses = new Vector();

  /** Fully qualified name. */

  private String _fullName = null;
}

// ***********************************************************************
// History Log:
// $Log: ADLClass.java,v $
// Revision 1.21  1998/06/22 14:12:38  cinard
// Improved support for fully qualified name.
//
// Revision 1.20  1998/06/17 16:04:42  cinard
// Added getFullName() method.
//
// Revision 1.19  1998/06/16 12:24:24  cinard
// Allow parent attribute to be null.
//
// Revision 1.18  1998/06/08 13:42:17  cinard
// Replaced  calls to obsolete method LClass.setSuperClass() with
// calls to LClass.addSuperClass().
//
// Revision 1.17  1998/05/27 09:44:18  cinard
// Added support for FileScope interface.
//
// Revision 1.16  1998/05/15 13:25:35  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.15  1998/04/28 11:48:59  cinard
// Allow tested class to be null. The ADLClass symbol can hence be used
// to represent C ADL annotated declarations.
//
// Revision 1.14  1998/03/25 14:49:56  cinard
// Updated ADLClass constructor and newADLClass() method (now adl
// super class name is required).
//
// Revision 1.13  1998/03/18 15:49:52  cinard
// New strategy for symbol selection (a single SymbolSelector class
// is now used). Symbol classes updated acordingly.
//
// Revision 1.12  1998/02/27 12:48:20  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.11  1998/02/23 15:08:49  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.10  1998/02/12 16:03:00  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.9  1998/01/23 16:49:35  cinard
// Improved symbol lookup for "this" & <className> symbols in both LClass and
// ADLClass symbols.
//
// Revision 1.8  1998/01/19 14:36:57  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.7  1998/01/12 13:18:59  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.6  1997/11/26 17:12:53  cinard
// Changed package name.
//
// Revision 1.5  1997/11/14 09:58:53  cinard
// LModule scope now accepts duplicate symbols.
// SymbolTableMgr has a getADLClass() method.
//
// Revision 1.4  1997/11/13 08:20:00  cinard
// Added get LClass and get super ADLClass methods.
//
// Revision 1.3  1997/11/12 13:43:09  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.2  1997/11/03 12:59:55  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.1  1997/10/27 11:18:01  cinard
// Added language independant symbols to the symboltable package.
//
//
// ***********************************************************************
