// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/LBlockScope.java,v $
// $Revision: 1.15 $
// $Date: 1998/05/15 13:25:36 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        LBlockScope.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.UnaryPredicate;


/**
 *  LBlockScope.
 */

public class LBlockScope implements Scope
{
   // -- Protected constructors --

   /**
    * Constructs block scope.
    * @param parent parent scope.
    */
   protected LBlockScope(Scope parent)
   {
      this.parent     = parent;
      symbols         = new Hashtable();
      anonymousScopes = new Vector();
   }

   // -- Public Methods --

   /**
    * Opens an anonymous scope being a child of this one.
    * @return the anonymous scope.
    */

   public Scope openScope()
   {
      Scope scope = new LBlockScope(this);
      anonymousScopes.addElement(scope);
      return scope;
   }

   /**
    * Opens a scope associated with a symbol in the current scope.
    * @param s symbol associated with the scope.
    * @return the opened scope.
    */

   public Scope openScope(Symbol s)
   {
      // Check the symbol exists in that scope
      if ( ! symbols.contains(s) )
      { 
         throw new Error("Symbol " + s.getName() + " not in that scope.");
      }
      else if ( ! ( s instanceof Scope ) )
      {
         throw new Error("Symbol " + s.getName() + " has no scope.");
      }
      else
      {
         // State symbol is defined
         s.setDefined();

         // Set scope parent.
         ((Scope)s).setParent(this);
         return (Scope)s;
      }
   }

   /**
    * Set the parent scope.
    * @param p the parent scope.
    */

   public void setParent(Scope p) { parent = p; }

   /**
    * Put a symbol in the scope.
    * @param symbol The symbol to put into the scope.
    */

   public Scope getParent() { return parent; }

   /**
    * Put a symbol in the scope.
    * @param symbol The symbol to put into the scope.
    */

   public void putSymbol(Symbol symbol)
   {
      // Set default modifier
      symbol.getModifier().set(defaultModifier);

      symbols.put(symbol.getName(),symbol);
   } 

  /**
   * Get a symbol in the scope, extended version.
   * When using this version of getSymbol, supplementary parameters are
   * provided that shall help finding the object.
   * <p>
   * The symbol shall satisfy some criteria defined by the provided 
   * predicate object. This predicate object help finding the symbol in 
   * case several symbols are likely to match the supplied name 
   * (e.g. overloaded methods).
   * </p>
   * <p>
   * Credentials shall be supplied for those cases where some symbols
   * hava non-public visibility. Retrieving them may require the
   * use of credentials.
   * </p>
   * @param name The name of the expected symbol.
   * @param predicate An arbitrary predicate object.
   * @param credentials Credentials of the scope.
   * @return The found symbol.
   */

   public Symbol getSymbol
                 (String name,UnaryPredicate predicate,Object credentials)
   {
      Symbol s = (Symbol)symbols.get(name);
      if ( ( s != null ) && ( predicate != null ) && ! predicate.execute(s) )
      {
         s = null;
      }
      return s;
   }

   /**
    * Get a symbol in the scope.
    * @param name The name of the expected symbol.
    * @return The found symbol.
    */

   public Symbol getSymbol(String name)
   {
      return getSymbol(name,null,null);
   }

  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   * @param defaultModifier default modifer (bit mask).
   */

  public void setDefaultModifier(int defaultModifier) 
  {
     this.defaultModifier = defaultModifier;
  }

   // -- to string methods --

   /**
    * Returns a description of the block, indented version. 
    * @param indent indentation string.
    */

   public String toString(String indent)
   {
      String s =   indent + "Block" + "\n"
                 + indent + "{"  + "\n" ;
      Enumeration se = symbols.elements();
      while ( se.hasMoreElements() )
      {
         s += ((Symbol)se.nextElement()).toString(indent+"  ");
      }
      Enumeration ae = anonymousScopes.elements();
      while ( ae.hasMoreElements() )
      {
         s += ((Scope)ae.nextElement()).toString(indent+"  ");
      }
      s += indent + "}\n";
      return s;
   }

  /**
    * Get the scope name in NL format (vector of strings).
    * Here, this is in fact the parent scope name.
    */
  public Vector getNLName() {

    Scope parent = getParent();
    return (parent == null) ? new Vector() : parent.getNLName();
  }

   /**
    * Returns a description of the block, not indented. 
    */

   public String toString()
   {
      return toString("");
   }

  // -- Private --

  /** Anonymous scopes opened in that block scope. */
  private Vector    anonymousScopes;

  /** Symbols contained in that scope. */
  private Hashtable symbols;

  /** Parent scope.  */
  Scope parent;

  /** Default modifier */
  private int defaultModifier = 0x0000;

  /** Debug flag */
  private final static boolean debug = false;
}

// ***********************************************************************
// History Log:
// $Log: LBlockScope.java,v $
// Revision 1.15  1998/05/15 13:25:36  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.14  1998/04/28 12:00:12  cinard
// Fixed a bug in getSymbol() method : found symbols were not checked
// against the supplied search predicate.
//
// Revision 1.13  1998/04/15 14:02:58  bareau
// Added a method getNLName to convert scope to string, for NL processing.
//
// Revision 1.12  1998/02/27 12:48:24  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.11  1998/02/23 15:08:51  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.10  1998/02/12 16:03:03  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.9  1998/01/19 14:36:59  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.8  1998/01/12 13:19:02  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.7  1997/12/08 12:58:40  cinard
// Added a new type check mechanism.
// Fixed some bugs.
//
// Revision 1.6  1997/11/26 17:12:55  cinard
// Changed package name.
//
// Revision 1.5  1997/11/12 13:43:11  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.4  1997/11/04 16:04:20  cinard
// Made some modification for C and C++ parsers.
//
// Revision 1.3  1997/11/04 09:04:28  cinard
// Scope.getSymbol() does not throw AmbiguousRequestException anymore.
// symboltable.java package name replaced with symbol_table.j (problems
// with javadoc). Note that symboltable/CVS/* should be updated manually
// (replacing java with j where necessary).
//
// Revision 1.2  1997/11/03 12:59:56  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.1  1997/10/27 11:18:02  cinard
// Added language independant symbols to the symboltable package.
//
//
// ***********************************************************************
