// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/LCallable.java,v $
// $Revision: 1.24 $
// $Date: 1998/06/25 12:16:47 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        LCallable.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;

import com.objectspace.jgl.UnaryPredicate;

/**
 * Superclass for all callable and likes (functions, methods, constructors).
 */

public abstract class LCallable implements Symbol, Scope
{

  // -- Constructors --

  /**
   * Constructs myself to be a callable symbol (method, functions and likes).
   * @param name the callable name.
   * @param modifier modifier flag
   * @param symbolId symbol's identifier 
   */
  protected LCallable(String name,LModifier modifier, long symbolId)
  {
     this.name       = name;
     this.modifier   = modifier;
     this._symbolId  = symbolId;

     parameters      = new Vector();
     exceptions      = new Vector();
     symbols         = new Hashtable();
     anonymousScopes = new Vector();
  }

  // -- Public methods --
  
  // -- Symbol interface implementation --
  
  /** 
   * Symbol's identifier.
   */
  public long getSymbolId()
  {
     return _symbolId;
  }

  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */
  
  public boolean isA(long symbolId) 
  { 
     return ( symbolId & _symbolId ) != 0; 
  }


  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */
 
  public String getName() { return name; }
  
  /**
   * Get the symbol's full name.
   */
 
  public String getFullName() { return getSignature(); }

  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */

  public void   setName(String n) { name = n; }

  /** 
   * Tells if the callable symbol's got an access to a scope.
   * This is never the case.
   * @return Always false.
   */

  public boolean hasScopeAccess() { return false; }

  /**
   * Get the scope to which this symbol has access. It is always null
   * for a callable symbol.
   * @return Never returns, error always thrown.
   */

  public Scope   getScope()
  {
     throw new Error("LCallable.getScope() should not be called."); 
  }

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
      return getOperationComparator(getSymbolId());
  }

  /**
   * Get a comparator enabling to select a callable with the same 
   * signature than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getOperationComparator()
  {
     return getOperationComparator(SymbolId.OPERATION_MASK);
  }

  /**
   * Get a comparator enabling to select a callable with the same 
   * signature than that of the symbol against which this method
   * is called.
   * @param symbolIdMask mask for selecting symbols.
   * @return the comparator.
   */

  public UnaryPredicate getOperationComparator(long symbolIdMask)
  {
      LCallableEqualsComparator comparator 
                                = new LCallableEqualsComparator
                                      (SymbolId.OPERATION_MASK & symbolIdMask);

      comparator.setReturnType(this.getReturnType());
      
      Enumeration e = parameters.elements();
      while ( e.hasMoreElements() )
      {
         LType t = ((LParameter)e.nextElement()).getType();
         comparator.addParameterType(t);
      }

      if ( lastParameterIsEllipsis() )
      {
         comparator.setLastParameterIsEllipsis();
      }

      Enumeration f = exceptions.elements();
      while ( f.hasMoreElements() )
      {
         comparator.addException((LType)f.nextElement());
      }

      return comparator;
  }

  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise. A callable is defined as its scope is opened.
   */

  public boolean isDefined()
  {
     return isDefined;
  }

  /**
   * Indicates that the symbol is both declared and defined.
   */

  public void setDefined()
  {
     isDefined = true;
  }

  /**
   * Get the symbol's modifier.
   * @return symbol's modifier.
   */

  public LModifier getModifier() { return modifier; }

  /**
   * Set the symbol's modifier.
   * @param modifier symbol's modifier.
   */

  public void setModifier(LModifier modifier) { this.modifier = modifier; }


  // --  Scope Interface Implementation --

  /**
   * Opens an anonymous scope into this scope.
   * Anonymous scopes are those scopes not directly related to symbols,
   * an example of such an anonymous scope the scope defined by a simple
   * block of code.
   * @return The new anonymous scope opened.
   */

  public Scope openScope()
  {
     Scope newBlock = new  LBlockScope(this);
     anonymousScopes.addElement(newBlock);
     return newBlock;
  }

   /**
    * Opens a scope associated with a symbol.
    * The provided symbol must owns its own scope (its hasScope() method
    * shall return true).
    * @param s The symbol for which the scope is to be opend.
    * @return The new scope opened.
    */

  public Scope openScope(Symbol s)
  {
     if ( ! ( s instanceof Scope ) )
     {
        throw new Error("Symbol " + s.getName() + " has no scope.");
     }
     else
     {
        // State symbol is defined
        s.setDefined();

        // Set scope parent.
        ((Scope)s).setParent(this);

        return (Scope)s;
     }
  }

  /**
   * Set the parent scope.
   * @param p the parent scope.
   */
    
  public void setParent(Scope p) { parent = p; }

  /**
   * Get the parent scope.
   * @return The parent scope.
   */

  public Scope getParent()       { return parent; }

  /**
   * Put a symbol in the scope.
   * @param symbol The symbol to put into the scope.
   */

  public void putSymbol(Symbol symbol)
  {
     // Set default modifier
     symbol.getModifier().set(defaultModifier);

     symbols.put(symbol.getName(),symbol);
  } 

  /**
   * Get a symbol in the scope, extended version.
   * <p>When using this version of getSymbol, supplementary parameters are
   * provided that shall help finding the object.
   * <pre>
   * The symbol shall satisfy some criteria defined by the provided 
   * predicate object. This predicate object help finding the symbol in 
   * case several symbols are likely to match the supplied name 
   * (e.g. overloaded methods).
   * </pre>
   * <pre>
   * Credentials shall be supplied for those cases where some symbols
   * have non-public visibility. Retrieving them may require the
   * use of credentials.
   * </pre>
   * @param name The name of the expected symbol.
   * @param predicate A predicate object.
   * @param credentials Credentials of the scope.
   * @return The found symbol.
   */


  public Symbol getSymbol
                (String name, UnaryPredicate predicate, Object credentials)
  {
     if ( debug ) 
     {
        System.out.println
        ("debug: LCallable.getSymbol(" 
         + name + ", " 
         + predicate + ")"
        ); 
     }
     Symbol s = (Symbol) symbols.get(name);
     if ( debug ) 
     {
        System.out.println ("debug: LCallable.getSymbol() - found first =" + s); 
     }
     if ( s != null && ( predicate == null || predicate.execute(s) ) )
     {
       // Symbol satisfies the predicate
       return s;
     }
     else
     {
        s = null;
        // Symbol is not suitable, check parameters. 
        Enumeration e = parameters.elements();
        while ( e.hasMoreElements() )
        {
           Symbol p = (Symbol)e.nextElement();
           if (    name.equals(p.getName())  
                && ( predicate == null || predicate.execute(p) )
              )
           {
              s = p;
              break;
           }
        }
     }
     if ( debug ) 
     {
        System.out.println
        ("debug: LCallable.getSymbol(" 
         + name + ", " 
         + predicate + ") return = " + s 
        ); 
     }
     return s;
  }

  /**
   * Get a symbol in the scope.
   * @param name The name of the expected symbol.
   * @return The found symbol.
   */

  public Symbol getSymbol(String name) 
  {
     return getSymbol(name,null,null);
  }

  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   */

  public void setDefaultModifier(int defaultModifier) 
  {
     this.defaultModifier = defaultModifier;
  }

  // -- specific  methods --

  /**
   * Set the return type of the callable.
   * @param type Return type of the callable.
   */

  public void setReturnType(LType type)
  {
     this.returnType = type;
  }

  /**
   * Get the return type of the callable.
   * @return Return type of the callable.
   */

  public LType getReturnType()
  {
     return returnType;
  }

  /**
   * Add a parameter (an argument) to the callable symbol.
   * @param parameter The parameter to add.
   */

  public void addParameter(LParameter parameter)
  {
     parameters.addElement(parameter);
  }

  /**
   * Indicate the operation has an ellipsis as its last parameter.
   */
  public void setLastParameterIsEllipsis()
  {
     _lastParameterIsEllipsis = true;
  }

  /**
   * Is last parameter an ellipsis?
   */
  public boolean lastParameterIsEllipsis()
  {
     return _lastParameterIsEllipsis;
  }


  /**
   * Add a parameter type to the callable symbol.
   * @param type  The parameter type to add.
   */

  public void addParameterType(LType type)
  {
     LParameter parameter = new LParameter("",null,type);
     parameters.addElement(parameter);
  }

  /**
   * Add a parameter name to the callable symbol.
   * <p>First parameter whose name is zero length's string has its name
   * set to name.
   * @param String  The parameter name to add.
   */

  public void addParameterName(LType type)
  {
     for (int i = 0; i < parameters.size(); i++)
     {
        LParameter parameter = (LParameter) parameters.elementAt(i);
        if ( parameter.getName().equals("") )
        {
           parameter.setName(name);
           return;
        }
     }
     throw new Error("No parameter's name missing.");
  }

  /**
   * Returns operation parameters.
   */
  public LParameter[] getParameters()
  {
     LParameter[] tmp = new LParameter[parameters.size()];
     parameters.copyInto(tmp);
     return tmp;
  }

  /**
   * Add an exception to the callable symbol.
   * @param exc The exception to add.
   */

  public void addException(LType exc)
  {
     exceptions.addElement(exc);
  }

  /**
   * Does this operation already declares it throws the supplied exception? 
   */

  public boolean throwsException(LType exception)
  {
    return exceptions.contains(exception);
  }

  /**
   * Returns operation exceptions.
   */
  public LType[] getExceptions()
  {
     LType[] tmp = new LType[exceptions.size()];
     exceptions.copyInto(tmp);
     return tmp;
  }

  /**
   * Compare this object and an other object for equality.
   * @param obj the reference object with which to compare.
   * @return True if this object is the same as the argument; false otherwise. 
   */

  public boolean equals(Object obj)
  {
     if ( ! ( obj instanceof LCallable ) )
     {
        return false;
     }

     LCallable lc = (LCallable) obj;

     LParameter[] myParams    = this.getParameters();
     LParameter[] otherParams = lc.getParameters();

     // Caution : return type may be null in either callable.
     if  (   ( returnType == null && lc.getReturnType() != null )
          || ( returnType != null && ! returnType.equals(lc.getReturnType()) )
         )
     {
        return false; 
     }

     // Note : parameters are never null.
     if ( myParams.length != otherParams.length ) 
     {
        return false;
     }
     else
     {
        for (int i = 0; i < parameters.size(); i++)
        {
           if ( ! myParams[i].equals(otherParams[i]) )
           {
              return false;
           }
        }
     }

     if ( lastParameterIsEllipsis() != lc.lastParameterIsEllipsis() )
     {
        return false;
     }

     LType[] myExcs    = this.getExceptions();
     LType[] otherExcs = lc.getExceptions();

     // Note : exceptions are never null.
     if ( myExcs.length != otherExcs.length ) 
     {
        return false;
     }
     else
     {
        for (int i = 0; i <myExcs.length ; i++)
        {
           if ( ! myExcs[i].equals(otherExcs[i]) )
           {
              return false;
           }
        }
        return true;
     }
  }

  /**
   * Returns a string description of the callable signature.
   * @return Description of the callable signature.
   */

  public String getSignature()
  {
     // Parameters

     String sParameters = "(";
     Enumeration e = parameters.elements();
     if ( e.hasMoreElements() )
     {
        LParameter p = (LParameter)e.nextElement();
        sParameters += ( p.getType() != null ? p.getType().getName() 
                                             : "<no type>" )
                     + " " +  p.getName();
     }
     while ( e.hasMoreElements() )
     {
        LParameter p = (LParameter)e.nextElement();
        sParameters += ", " 
                     + ( p.getType() != null ? p.getType().getName() 
                                             : "<no type>" ) 
                     + " " +  p.getName();
     }
     if ( lastParameterIsEllipsis() )
     {
        if ( ! sParameters.equals("(") ) { sParameters += ", "; }
        sParameters += "...";
     }
     sParameters += ")";

     // Modifier

     String sModf = ( modifier != null ? modifier.toString() : "" );
     if ( sModf.length() > 0 ) { sModf += " "; }

     // Return type
     String sReturnType = (  getReturnType() != null 
                           ? getReturnType().getName() + " "  : " " );

     // Exceptions

     String sExceptions = "";
     Enumeration f = exceptions.elements();
     if ( f.hasMoreElements() )
     {
        LType excp = (LType)f.nextElement();
        sExceptions += " throws " + excp.getName();
     }
     while ( f.hasMoreElements() )
     {
        LType excp = (LType)f.nextElement();
        sExceptions += ", " + excp.getName();
     }

     // Symbols and anonymous scopes 

     return    header() + ( header().length() > 0 ? " " : "" ) 
             + sModf
             + sReturnType
             + getName() + sParameters
             + sExceptions;

  }

  /**
   * Returns a string vector description of the callable signature for NL purpose
   * (no return type, no parameter name, no exception)
   * @return Description of the callable signature.
   */

  public Vector getNLName()
  {
    Scope parent = getParent();
    Vector ret =  (parent == null) ? new Vector() : parent.getNLName();
    ret.addElement(name);
    ret.addElement(getNLSignature());
    return ret;
  }

  public String getNLSignature() {

    String sParameters = "(";
    Enumeration e = parameters.elements();
    if ( e.hasMoreElements() )
     {
        LParameter p = (LParameter)e.nextElement();
        sParameters += p.getType().getName();
     }
    while ( e.hasMoreElements() )
      {
        LParameter p = (LParameter)e.nextElement();
        sParameters += ", " + p.getType().getName();
      }
    sParameters += ")";

    return sParameters;
  }


  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...)
   */

  public Symbol getEnclosingConstruct()
  {
     Scope parent = getParent();
     while ( parent != null )
     {
        if (    ( parent instanceof Symbol )
             && ((Symbol)parent).isA(SymbolId.CONSTRUCTED_MASK) )
        {
           return (LConstructedType)parent;
        }
        parent = parent.getParent();
     }
     return null;
  }

 
  /**
   * Returns a string description of the callable, indented version.
   * @param indent Indentation.
   * @return Description of the callable.
   */

  public String toString(String indent)
  {
     String s =  indent + getSignature();

     // Symbols and anonymous scopes 

     String body = "";
     Enumeration se = symbols.elements();
     while ( se.hasMoreElements() )
     {
         body += ((Symbol)se.nextElement()).toString(indent+"  ");
     }
     Enumeration ae = anonymousScopes.elements();
     while ( ae.hasMoreElements() )
     {
        body += ((Scope)ae.nextElement()).toString(indent+"  ");
     }
     s += (   body.length() > 0 
            ? "\n" + indent + "{\n" + body + indent + "}\n"
            : " {}\n"
          );
     return s;
  }


  /**
   * Returns a string description of the callable, not indented version.
   * @return Description of the callable.
   */

  public String toString()
  {
     return getName();
  }


  // -- Protected methods --

  /**
   * Returns a header for the string describing the callable. 
   * Should be overriden in concrete subclasses.
   */ 

  abstract protected String header();

  // -- Private feature --

  /** Name of the callable. */
  private String    name;

  /** Symbol's identifier */
  private final long _symbolId;

  /** Modifier flag of the expression */
  private LModifier modifier = new LModifier(0x0000);
  
  /** Return type of the callable. */
  private LType     returnType;

  /** Parameters of the callable. */
  private Vector    parameters;

  /** Exceptions of the callable. */
  private Vector    exceptions;

  /** Parent scope of the callable */
  private Scope     parent;

  /** Anonymous scopes contained in the callable. */
  private Vector    anonymousScopes;

  /** Symbols contained in the callable. */
  private Hashtable symbols;

  /** Defined flag */
  private boolean isDefined = false;

  /** Default modifier */
  private int defaultModifier;

  /** Is last parameter an ellipsis? */
  private boolean _lastParameterIsEllipsis = false;

  /** Debug flag. */
  private static final boolean debug = false;
}

// ***********************************************************************
// History Log:
// $Log: LCallable.java,v $
// Revision 1.24  1998/06/25 12:16:47  cinard
// Added throwException() method.
//
// Revision 1.23  1998/06/10 15:29:04  bareau
// Moved "isTestMethod" from LCallable to LMethod (and fixed it: runtime
// methods like infoline, tdd_assert... are _not_ test methods).
// Added an "isRuntimeMethod" flag in LMethod, and set it in symbol
// factories for runtime methods.
//
// Revision 1.22  1998/05/15 13:25:36  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.21  1998/05/11 09:36:11  cinard
// Allow to put scoped symbols in a LCallable's scope.
//
// Revision 1.20  1998/04/28 11:58:52  cinard
// Implemented support for ellipsis in parameters.
//
// Revision 1.19  1998/04/15 14:02:59  bareau
// Added a method getNLName to convert scope to string, for NL processing.
//
// Revision 1.18  1998/03/18 15:48:02  cinard
// Removal of ambiguities on operation invocation.
//
// Revision 1.17  1998/03/10 17:26:35  mallet
// Merged v1.1 with development trunk
//
// Revision 1.16.2.1  1998/03/09 16:00:16  cinard
// Modified toString() method so that only the name of the symbol is
// returned.
//
// Revision 1.16  1998/02/27 12:48:25  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.15  1998/02/23 15:08:52  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.14  1998/02/12 16:03:04  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.13  1998/02/11 10:00:34  cinard
// Moved isTestMethod() method upper in the inheritance tree.
//
// Revision 1.12  1998/01/19 14:36:59  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.11  1998/01/12 13:19:03  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.10  1998/01/06 14:56:56  cinard
// Changed exception type in method signatures to LType (it was LClass before,
// which is not suitable for C++).
//
// Revision 1.9  1997/12/19 21:16:35  cinard
// Added support for exceptions in throws clauses of methods.
// LCallable and LCallableEqualsComparator now contains a addException()
// method.
//
// Revision 1.8  1997/12/08 12:58:41  cinard
// Added a new type check mechanism.
// Fixed some bugs.
//
// Revision 1.7  1997/11/26 17:12:56  cinard
// Changed package name.
//
// Revision 1.6  1997/11/13 13:24:39  cinard
// Improved callable display.
//
// Revision 1.5  1997/11/12 13:43:12  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.4  1997/11/04 16:04:21  cinard
// Made some modification for C and C++ parsers.
//
// Revision 1.3  1997/11/04 09:04:28  cinard
// Scope.getSymbol() does not throw AmbiguousRequestException anymore.
// symboltable.java package name replaced with symbol_table.j (problems
// with javadoc). Note that symboltable/CVS/* should be updated manually
// (replacing java with j where necessary).
//
// Revision 1.2  1997/11/03 12:59:57  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.1  1997/10/27 11:18:03  cinard
// Added language independant symbols to the symboltable package.
//
//
// ***********************************************************************
