// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/LExpression.java,v $
// $Revision: 1.15 $
// $Date: 1998/04/28 12:07:28 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        LExpression.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import com.objectspace.jgl.UnaryPredicate;

/**
 *  Superclass for expression like symbols (variables, parameters, attributes).
 */

abstract public class LExpression implements Symbol
{
  // -- Constructors --

  /** 
   * Constructs me to be a LExpression (all attributes set).
   * @param name expression's name.
   * @param modifier expression's modifier flag.
   * @param type expression's type.
   * @param symbolId symbol's identifier
   */

  protected  LExpression(String    name,
                         LModifier modifier,
                         LType     type,
                         long       symbolId)
  {
     this.name  = name;
     this.modifier = modifier;
     this.type  = type;
     this._symbolId = symbolId;
  }

  // -- Symbol interface implementation --

  /** 
   * Symbol's identifier.
   */
  public long getSymbolId()
  {
     return _symbolId;
  }

  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */

  public boolean isA(long symbolId) 
  { 
     return ( symbolId & _symbolId ) != 0; 
  }

  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */

  public String getName() { return name; }

  /**
   * Get the symbol's full name.
   */
 
  public String getFullName() { return getName(); }


  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */

  public void   setName(String n) { name = n; }

  /**
   * Check if symbol has direct or indirect access to a scope.
   * For an expression scope access granted if expression's type 
   * has scope access (i.e. can be applied . or :: operators).
   * @return True if symbol has access to a scope, false otherwise.
   */

  public boolean hasScopeAccess()
  {
     return type instanceof Scope;
  }

  /**
   * Get the scope the symbol has access to.
   * @return The scope that can be accessed to.
   */

  public Scope   getScope()
  {
     return type.getScope();
  }
  

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
     return new SymbolSelector(getSymbolId());
  }

  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise. An expression declaration implies its
   * definition.
   */

  public boolean isDefined()
  {
     return true;
  }

  /**
   * Indicates that the symbol is both declared and defined.
   */

  public void setDefined() 
  { /* Expressions are always defined when declared */ }


  /**
   * Get the symbol's modifier flag.
   * @return The symbol's modifier flag.
   */

  public LModifier getModifier() { return modifier; }

  /**
   * Set the symbol's modifier flag.
   * @param modifier  The symbol's modifier flag.
   */

  public void setModifier(LModifier modifier) 
  { 
     this.modifier = modifier; 
  }

  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...)
   */

  public Symbol getEnclosingConstruct()
  {
     if ( related instanceof LConstructedType )
     {
        return (LConstructedType)related;
     }
     else
     {
        return null;
     }
  }

  // -- Specific methods --
  
  /**
   * Can a value of type t be assigned to a variable of this type.
   * @param  t The type for which the assignability is checked.
   * @return True if it can, false otherwise.
   */

  public boolean assignable(Symbol t) 
  {
     return (this.getType()).assignable(t);
  }

  /** Returns the expression's type. */   
  public LType getType() { return type; }

  /** 
   * Set the expression's type.
   * @param type expression's type.
   */   
  public void setType(LType type) { this.type = type; }

  /**
   * Returns a header for the description of the expression.
   * Simplest header is the expression kind (variable, ...).
   * Default implementation returns runtime class name, but this 
   * method should always be overridded in concrete subclasses.
   */
 
  public String header() { return this.getClass().toString() ; }

  /**
   * Returns a description of the symbol, indented.
   * @param indent Indentation string.
   * @return Description of the symbol.
   */

  public String toString(String indent)
  {
     // Modifier

     String sModf = getModifier().toString();
     if ( sModf.length() > 0 ) { sModf += " "; }

     return      indent + header() + ( header().length() > 0 ? " " : "" )
                        + sModf 
                        + getType().mixNameAndType(getName())
                        + "\n";
  }

  /**
   * Tells if expression was defined in LClass scope.
   * @return true if related parent scope is an LClass.
   */
  public boolean isLClassAttribute()
  {
    return ( related != null ) && ( related instanceof LClass ) ;
  }


  /**
   * Returns a description of the symbol, not indented.
   * @return Description of the symbol.
   */

  public String toString()
  {
     return getName();
  }

  /**
   * Saved scope where LExpression was defined.
   */
  public void setRelated( Object parentScope )
  {
    related = parentScope  ;
  }

  /**
   * Returns the parent related scope .
   * @result parent related scope.
   */
  private Object getRelated()
  {
    return related ;
  }

  /**
   * Set defining node of the symbol.
   */
  public void setDefiningNode( Object definingNode )
  {
    this.definingNode = definingNode;
  }

  /**
   * Returns the defining node of the expression.
   */

  public Object getDefiningNode()
  {
    return this.definingNode ;
  }

  // -- Private feature --

  /** Related scope where symbol was defined. */
  private Object related ;

  /** Node defining the symbol. */
  private Object definingNode ;

  /** Type of the expression. */
  private LType   type;

  /** Modifier flag of the expression */
  private LModifier modifier;
  
  /** Name of the expression */
  private String  name;  

  /** Symbol's identifier */
  private final long _symbolId;
}

// ***********************************************************************
// History Log:
// $Log: LExpression.java,v $
// Revision 1.15  1998/04/28 12:07:28  cinard
// Expression's type can be null (useful when an expression definition
// is complex and the type is not available at the beginning of the
// expression's parsing).
// Added a defining node field and accessor, where the node defining
// the expression can be hidden by the parser. Note - this mechanism
// could be generalized to all the symbols.
//
// Revision 1.14  1998/03/18 15:49:54  cinard
// New strategy for symbol selection (a single SymbolSelector class
// is now used). Symbol classes updated acordingly.
//
// Revision 1.13  1998/03/10 17:26:36  mallet
// Merged v1.1 with development trunk
//
// Revision 1.12.2.1  1998/03/09 16:00:17  cinard
// Modified toString() method so that only the name of the symbol is
// returned.
//
// Revision 1.12  1998/02/27 12:48:29  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.11  1998/02/23 15:08:56  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.10  1998/02/12 16:03:07  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.9  1998/02/06 15:21:16  wassel
// Added methods setRelated and getRelated returning the scope where symbol is defined
//
// Revision 1.8  1998/01/19 14:37:01  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.7  1998/01/12 13:19:05  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.6  1997/12/08 12:58:43  cinard
// Added a new type check mechanism.
// Fixed some bugs.
//
// Revision 1.5  1997/11/26 17:12:59  cinard
// Changed package name.
//
// Revision 1.4  1997/11/12 13:43:16  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.3  1997/11/04 16:04:22  cinard
// Made some modification for C and C++ parsers.
//
// Revision 1.2  1997/11/03 12:59:59  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.1  1997/10/27 11:18:04  cinard
// Added language independant symbols to the symboltable package.
//
//
// ***********************************************************************
