// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/LModule.java,v $
// $Revision: 1.18 $
// $Date: 1998/06/03 07:34:34 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        LModule.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Nov 06,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Enumeration;
import java.util.Vector;
import com.objectspace.jgl.HashMap;
import com.objectspace.jgl.HashSet;
import com.objectspace.jgl.UnaryPredicate;
import com.objectspace.jgl.util.ConditionalEnumeration;

/**
 *  Abstract superclass for compilation unit, namespace, package.
 */

public class LModule implements FileScope , Symbol
{

  /**
   * Constructs the module.
   * @param name name of the module.
   * @param parent parent scope.
   * @param symbolId symbol's identifier  
   */

   public LModule(String name,Scope parent,long symbolId)
   {
      this.parent = parent;
      this.name   = name;
      this._symbolId = symbolId;
      symbols = new HashMap(true);
   }

  // -- Symbol interface implementation --

  /** 
   * Symbol's identifier.
   */
  public long getSymbolId()
  {
     return _symbolId;
  }


  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */
  public boolean isA(long symbolId) 
  { 
     return ( symbolId & _symbolId ) != 0; 
  }

  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */

  public String getName() { return name; }

  /**
   * Get the symbol's name in NL scope format.
   * Just an empty vector for compilation unit and name spaces;
   * this method is overloaded only in LPackage.
   * @see ORG.opengroup.adl.symboltable.LPackage#getNLName
   * @return The symbol's name.
   */

  public Vector getNLName() {

    return new Vector();
  }

  /**
   * Get the symbol's full name.
   */
 
  public String getFullName() { return getName(); }


  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */

  public void   setName(String name) { this.name = name; }

  /** 
   * Tells if the type has got an access to a scope.
   * @return Always true.
   */

  public boolean hasScopeAccess()
  {
     return true;
  }

  /**
   * Get the scope to which this symbol has access.
   * @return The scope.
   */

  public Scope getScope()
  {
     return this;
  }

  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...)
   */

  public Symbol getEnclosingConstruct()
  {
     Scope parent = getParent();
     while ( parent != null )
     {
        if ( parent instanceof LModule )
        {
           return (LModule)parent;
        }
        parent = parent.getParent();
     }
     return null;
  }

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
     return new SymbolSelector(getSymbolId());
  }

  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise. A namespace .
   */

  public boolean isDefined()
  {
     return true;
  }

  /**
   * Indicates that the symbol is both declared and defined.
   */

  public void setDefined()
  {
     // Namespace is always defined.
  }

  /**
   * Get the symbol's modifier.
   * @return symbol's modifier.
   */

  public LModifier getModifier() { return modifier; }

  /**
   * Set the symbol's modifier.
   * @param modifier symbol's modifier.
   */

  public void setModifier(LModifier modifier) { this.modifier = modifier; }


   // -- Scope interface implementation --

   /**
    * Opens an anonymous scope into this scope.
    * Anonymous scopes are those scopes not directly related to symbols,
    * an example of such an anonymous scope the scope defined by a simple
    * block of code.
    * @return The new anonymous scope opened.
    */

   public Scope openScope()
   {
     Scope newBlock = new  LBlockScope(this);
     anonymousScopes.addElement(newBlock);
     return newBlock;
   }

   /**
    * Opens a scope associated with a symbol.
    * The provided symbol must owns its own scope (its hasScope() method
    * shall return true).
    * @param s The symbol for which the scope is to be opend.
    * @return The new scope opened.
    */

   public Scope openScope(Symbol s)
   {
      // Check the symbol exists in that scope
      if ( symbols.countValues(s) == 0 )
      { 
         throw new Error("Symbol " + s.getName() + " not in that scope.");
      }
      else if ( ! ( s instanceof Scope ) )
      {
         throw new Error("Symbol " + s.getName() + " has no scope.");
      }
      else
      {
         // State symbol is defined
         s.setDefined();

         // Set scope parent.
         ((Scope)s).setParent(this);

         return (Scope)s;
      }
   }

   /**
    * Set the parent scope.
    * @param p the parent scope.
    */

   public void setParent(Scope p) { parent = p; }

   /**
    * Get the parent scope.
    * @return The parent scope.
    */

   public Scope getParent() { return parent; }

   /**
    * Put a symbol in the scope.
    * @param symbol The symbol to put into the scope.
    */

   public void putSymbol(Symbol symbol)
   {
      // Set default modifier
      symbol.getModifier().set(defaultModifier);

      symbols.add(symbol.getName(),symbol);
   } 

   /**
    * Get a symbol in this scope.
    * Symbol is searched in this scope and if necessary in superclasses scopes.
    */

   public Symbol getSymbol
                 (String name,UnaryPredicate predicate,Object credentials)
   { 
     HashSet alreadyVisistedModules = new HashSet();
     return getSymbol(name,predicate,credentials,alreadyVisistedModules);
   }


   /**
    * Get a symbol in the scope.
    * @param name The name of the expected symbol.
    * @return The found symbol.
    */

   public Symbol getSymbol(String name) 
   {
      return getSymbol(name,null,null);
   }
  
  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   */

  public void setDefaultModifier(int defaultModifier) 
  {
     this.defaultModifier = defaultModifier;
  }

  /**
   * Returns a description of the symbol, indented.
   * @param indent Indentation string.
   * @return Description of the scope.
   */

  public String toString(String indent)
  {
     String s;

     s =   indent + header() + " " + getName() + "\n"
         + indent + "{"  + "\n" ;
     // Put included modules before other symbols
     Enumeration includedModules = _includedModules.elements();
     while ( includedModules.hasMoreElements() )
     {
       s += indent + "  include " 
	 + ((FileScope)(includedModules.nextElement())).getFileName() + "\n";
     }
     // Put nested modules before other symbols.
     Enumeration e = symbols.elements();
     while ( e.hasMoreElements() )
     {
        Symbol symbol = (Symbol)e.nextElement();
        if ( symbol instanceof LModule )
        {
           s += symbol.toString(indent+"  ");
        }
     }
     Enumeration f = symbols.elements();
     while ( f.hasMoreElements() )
     {
        Symbol symbol = (Symbol)f.nextElement();
        if ( ! ( symbol instanceof LModule ) )
        {
           s += symbol.toString(indent+"  ");
        }
     }
     s += indent + "}\n";
     return s;

  }

  /**
   * Returns a description of the symbol, not indented.
   * @param indent Indentation string.
   * @return Description of the scope.
   */

  public String toString()
  {
     return getName();
  }

  // -- module include mechanism support (e.g. c and c++ #include)
  
  /**
   * Get the name of the underlying file.
   */

  public String getFileName() 
  { 
     return getName();
  }

  /**
   * Add an included file scope.
   * @return The parent scope.
   */

  public void addIncludedFile(FileScope module) 
  { 
    if ( debug ) System.out.println
		 ("debug : LModule<" + getFileName() + ">.addIncludedFile(" 
                 + module.getFileName() + ")");
    _includedModules.addElement(module);
  }

  /**
   * Indicate parsing is completed for this file, and recursively
   * for all included files.
   */

  public void setParsingCompleted()
  {
     // for this file

     _isParsingCompleted = true;

     // for included files
     
     Enumeration e = symbols.elements();
     while ( e.hasMoreElements() )
     {
        Object s = e.nextElement();
        if ( s instanceof LModule )
        {
           LModule module = (LModule)s;
           if ( ! module.isParsingCompleted() )
           {
              module.setParsingCompleted();
           }
        }
     }

  }

  /**
   * Is parsing completed?
   */

  public boolean isParsingCompleted()
  {
     return _isParsingCompleted;
  }

  // -- Protected --

  /**
   * Returns a header for the string describing the callable.
   * Should be overriden in concrete subclasses.
   */

  protected String header()
  {
     return this.getClass().toString();
  }

  // -- private --

  /**
   * Get a symbol in this scope.
   * Symbol is searched in this scope and if necessary in superclasses scopes.
   * @param alreadyVisistedModules holds the references of already 
   *                               interrogated modules (to avoid infinite 
   *                               loop in requests).
   */

  private Symbol getSymbol
                 (String         name,
		  UnaryPredicate predicate,
		  Object         credentials,
		  HashSet        alreadyVisistedModules)
  { 
     Symbol s = null;

     // visit the module in case it has not already been visited
     if ( alreadyVisistedModules.count(this) == 0 ) {
       // indicate this scope has been visited
       alreadyVisistedModules.add(this);

       // Check that the requested symbol is not the constructed type itself
       // or keyword this.
       if ( name.equals(getName()) ) {
	 // It is, return it if verifies the predicate.
	 if ( predicate != null && predicate.execute(this) ) {
	   return this;
	 }
       }

       Enumeration selectedSymbols = symbols.values(name);
      
       // Apply predicate

       if ( predicate != null ) {
	 selectedSymbols = new ConditionalEnumeration
	                       (selectedSymbols,predicate);
       }

       /**
	* Depending on the number of selected symbols, the strategy is as
	* follows :
	* 1) if no symbols were selected, return null,
	* 2) if exatly one symbol was selected, return it,
	* 3) if more than one symbol were selected throw an Error
	*    (AmbiguousRequestException)
	*/

       if ( ! selectedSymbols.hasMoreElements() ) { 
	 s = null;
       }
       else {
	 s = (Symbol)selectedSymbols.nextElement();
	 
	 if ( selectedSymbols.hasMoreElements() ) {
	   throw new Error 
	             ( "AmbiguousRequestException: "
                     + "Several symbols satisfies your request :" 
                     + s + " and " + selectedSymbols.nextElement()
                     ); 
	 }
       }

       // recursively search in included modules
       if ( s == null ) {
	 Enumeration e = _includedModules.elements();
	 while ( e.hasMoreElements() ) {
	   LModule included = (LModule)e.nextElement();
	   s = included.getSymbol(name,
				  predicate,
				  credentials,
				  alreadyVisistedModules);
	   if ( s != null ) { break; }
	 }
       }
     }
     return s;
   }

  // -- Private attributes --

  /** Anonymous scopes contained in the LModule. */
  private Vector    anonymousScopes = new Vector();

  /** Name of the type.  */
  String name = "";

  /** Symbol's identifier */
  private final long _symbolId;

  /** Modifier */
  private LModifier modifier = new LModifier(0x0000);

  /** Hashtable to store the type's symbols.  */
  private HashMap symbols;

  /** Parent scope */
  Scope parent;

  /** Default modifier */
  private int defaultModifier = 0x0000;

  /** Is parsing for this module completed? */
  private boolean _isParsingCompleted = false;

  /** Included modules */
  private Vector _includedModules = new Vector();
 
  /** Debug flag. */

  private final static boolean debug = false;
}

// ***********************************************************************
// History Log:
// $Log: LModule.java,v $
// Revision 1.18  1998/06/03 07:34:34  bareau
// Moved the method getNLName's behavior from LModule to LPackage.
//
// Revision 1.17  1998/05/27 09:44:19  cinard
// Added support for FileScope interface.
//
// Revision 1.16  1998/05/15 13:25:38  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.15  1998/05/11 11:43:49  cinard
// Added support for allowing a module to include some others. Module
// inclusion semantic is that of the #include mechanism in C and C++.
//
// Revision 1.14  1998/04/15 14:02:59  bareau
// Added a method getNLName to convert scope to string, for NL processing.
//
// Revision 1.13  1998/03/18 15:49:55  cinard
// New strategy for symbol selection (a single SymbolSelector class
// is now used). Symbol classes updated acordingly.
//
// Revision 1.12  1998/03/10 17:26:37  mallet
// Merged v1.1 with development trunk
//
// Revision 1.11.2.1  1998/03/09 16:00:17  cinard
// Modified toString() method so that only the name of the symbol is
// returned.
//
// Revision 1.11  1998/02/27 12:48:31  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.10  1998/02/23 15:08:57  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.9  1998/02/12 16:03:09  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.8  1998/02/11 10:43:54  wassel
// Opening an anonymous scope in LModule is now possible.
//
// Revision 1.7  1998/02/02 16:11:03  wassel
// Fixed a bug in putAlias method.
//
// Revision 1.6  1998/01/19 14:37:02  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.5  1998/01/12 13:19:07  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.4  1997/12/08 12:58:43  cinard
// Added a new type check mechanism.
// Fixed some bugs.
//
// Revision 1.3  1997/11/26 17:13:00  cinard
// Changed package name.
//
// Revision 1.2  1997/11/14 09:58:54  cinard
// LModule scope now accepts duplicate symbols.
// SymbolTableMgr has a getADLClass() method.
//
// Revision 1.1  1997/11/12 13:43:18  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// ***********************************************************************
