// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/LTypeImpl.java,v $
// $Revision: 1.4 $
// $Date: 1998/03/18 15:49:58 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        LTypeImpl.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Feb 17,  1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Hashtable;
import com.objectspace.jgl.UnaryPredicate;

/**
 * Superclass for target language types. 
 */

public abstract class LTypeImpl implements LType
{
  /**
   * Default implementation for the constructor of the abstract class.
   * This constructor may be called in concrete subclasses' constructors.
   * @param name The name of the type.
   * @param modifier The modifier.
   * @param symbolId symbol's identifier
   */

  protected LTypeImpl(String name,LModifier modifier,long symbolId)
  {
     this.name         = name;
     this.modifier     = modifier;
     this._symbolId     = symbolId;
  }

  /**
   * Is this equals to supplied o.
   */
  public boolean equals(Object o)
  {
     return    o instanceof LType 
            && super.equals(((LType)o).getRealType());
  }
  
  /**
   * Returns synthesis type of the @param t and type for which method applies.
   * <p><b>Algorithm:</b><ul>
   * <li> If type <code>t</code> is null return null.
   * <li> Else if <code>t</code> assignable to <code>this</code> return <code>
   * this</code> type.
   * Else if <code>this</code> assignable to <code>t</code> return <code>t
   * </code>
   * <li> Else return null.
   * </ul>
   * @return is the smallest type assignable to both types,
   *         is null is no common assignable type found.
   */ 
  public LType synthesisType( LType t) {

     LType returnType ;
     if ( t  == null ) {
        return null;
     }
     else {
        if ( this.assignable(t) ) {
           returnType = this;
        }
        else if( t.assignable( this) ) {
           returnType = t;
        }
        else {
           returnType = null;
        }
     }
     return returnType;
  }

  // -- Symbol interface implementation --

  /** 
   * Symbol's identifier.
   */
  public long getSymbolId()
  {
     return _symbolId;
  }

  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */

  public boolean isA(long symbolId) 
  { 
     return ( symbolId & _symbolId ) != 0; 
  }

  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */
   
  public String getName() { return name; }

  /**
   * Get the symbol's full name.
   */
  
  public String getFullName() { return getName(); }

  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */

  public void setName(String n) { name = n; }

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
     return new SymbolSelector(getSymbolId());
  }

  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise. Default is to return true, constructed types
   * should override this method.
   */

  public boolean isDefined()
  {
     return true;
  }

  /**
   * Indicates that the symbol is both declared and defined.
   * Default is to do nothing, constructed types should override this
   * method.
   */

  public void setDefined()
  {
     // Default is to do nothing.
  }

  /**
   * Get the symbol's modifier.
   * @return symbol's modifier.
   */

  public LModifier getModifier() { return modifier; }

  /**
   * Set the symbol's modifier.
   * @param modifier symbol's modifier.
   */

  public void setModifier(LModifier modifier) { this.modifier = modifier; }

  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...). Default implementation is to return null,
   * constructed types shall override this.
   */

  public Symbol getEnclosingConstruct()
  {
     return null;
  }

  // -- LType specific methods --

  /**
   * Can a value of type t be assigned to a variable of this type.
   * @param  t The type for which the assignability is checked.
   * @return True if it can, false otherwise.
   */

  public boolean assignable(Symbol t)
  {
     return this.equals(t); // default behavior that should be overrided.
  }

  /**
   * Returns true if type is void. Default is to return false.
   */

  public boolean isVoid()
  {
     return false;
  }

  /**
   * Returns a string describing me.
   */
  public String toString() { return getName(); }

  /**
   * Returns a string for the declaration of an expression of this
   * type. Useful for mixed type and variable name (e.g. pointer
   * to function). Default behaviour is to return "type name".
   * @param name expression's name.
   */

  public String mixNameAndType(String name)
  {
     return getName() + " " + name;
  }

  // -- Package methods --


  /** 
   *  Get the type kind. Useful for languages that defines several
   *  type kinds (e.g. in Java boolean, arithmetic and references).
   *  @return an integer indicating the kind of the type. 
   */

  public Integer getTypeKind() { return new Integer(0); } 

  /**
   * Real type. That is the 'ultimate' real (i.e. not typedef'ed type ) type 
   * for which this type is defined. Default is to return this. 
   */

  public LType getRealType() 
  {
     return this;
  } 


  /**
   * Make up new typedef for this type.
   * @param name typedef name.
   */

  public LType newTypedef(String name)
  {
     // default should be overrided
     throw new Error("No typedef defined for " + this.getClass());
  }

  // -- Protected fields --

  /** Modifier */

  protected LModifier modifier = new LModifier(0x0000);

  // -- Private features --
  
  /** Symbol's name. */
  private String name = null;

  /** Symbol's identifier */
  private final long _symbolId;

}

// ***********************************************************************
// History Log:
// $Log: LTypeImpl.java,v $
// Revision 1.4  1998/03/18 15:49:58  cinard
// New strategy for symbol selection (a single SymbolSelector class
// is now used). Symbol classes updated acordingly.
//
// Revision 1.3  1998/03/10 17:26:38  mallet
// Merged v1.1 with development trunk
//
// Revision 1.2.2.1  1998/03/09 16:00:18  cinard
// Modified toString() method so that only the name of the symbol is
// returned.
//
// Revision 1.2  1998/02/27 12:48:41  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.1  1998/02/23 15:09:05  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// ***********************************************************************
