// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/Scope.java,v $
// $Revision: 1.12 $
// $Date: 1998/05/15 13:25:39 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        Scope.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Vector;
import com.objectspace.jgl.UnaryPredicate;

/**
 * Scope defines the interface necessary to implement a scope. Scopes
 * are units of visibility for a set of symbols.
 */

public interface Scope
{
   /**
    * Opens an anonymous scope into this scope.
    * Anonymous scopes are those scopes not directly related to symbols,
    * an example of such an anonymous scope the scope defined by a simple
    * block of code.
    * @return The new anonymous scope opened.
    */

   Scope openScope();

   /**
    * Opens a scope associated with a symbol.
    * The provided symbol must owns its own scope (its hasScope() method
    * shall return true).
    * @param s The symbol for which the scope is to be opend.
    * @return The new scope opened.
    */

   Scope openScope(Symbol s);

   /**
    * Set the parent scope.
    * @param p the parent scope.
    */

   void setParent(Scope p);

   /**
    * Get the parent scope.
    * @return The parent scope.
    */

   Scope getParent(); 

   /**
    * Put a symbol in the scope.
    * @param symbol The symbol to put into the scope.
    */

   void putSymbol(Symbol symbol);

   /**
    * Get a symbol in the scope.
    * @param name The name of the expected symbol.
    * @return The found symbol.
    */

   Symbol getSymbol(String name) ;

  /**
   * Get a symbol in the scope, extended version.
   * When using this version of getSymbol, supplementary parameters are
   * provided that shall help finding the object.
   * <p>
   * The symbol shall satisfy some criteria defined by the provided 
   * predicate object. This predicate object help finding the symbol in 
   * case several symbols are likely to match the supplied name 
   * (e.g. overloaded methods).
   * </p>
   * <p>
   * Credentials shall be supplied for those cases where some symbols
   * hava non-public visibility. Retrieving them may require the
   * use of credentials.
   * </p>
   * @param name The name of the expected symbol.
   * @param predicate An arbitrary predicate object.
   * @param credentials Credentials of the scope.
   * @return The found symbol.
   */

  Symbol getSymbol(
		     String name,
		     UnaryPredicate predicate,
		     Object credentials
		   );

  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   * @param defaultModifier default modifer (bit mask).
   */

  public void setDefaultModifier(int defaultModifier);

  /**
    * Get the name of the scope in NL format (vector of strings).
    * @param the scope name
    */
  public Vector getNLName();

  /**
   * Returns a description of the scope, indented.
   * @param indent Indentation string.
   * @return Description of the scope.
   */

  public String toString(String indent);

  /**
   * Returns a description of the scope, not indented.
   * @return Description of the scope.
   */

  public String toString();

}

// ***********************************************************************
// History Log:
// $Log: Scope.java,v $
// Revision 1.12  1998/05/15 13:25:39  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.11  1998/02/27 12:48:43  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.10  1998/02/23 15:09:07  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.9  1998/02/12 16:03:19  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.8  1998/01/19 14:37:04  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.7  1998/01/12 13:19:13  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.6  1997/11/26 17:13:05  cinard
// Changed package name.
//
// Revision 1.5  1997/11/04 16:04:24  cinard
// Made some modification for C and C++ parsers.
//
// Revision 1.4  1997/11/04 09:04:30  cinard
// Scope.getSymbol() does not throw AmbiguousRequestException anymore.
// symboltable.java package name replaced with symbol_table.j (problems
// with javadoc). Note that symboltable/CVS/* should be updated manually
// (replacing java with j where necessary).
//
// Revision 1.3  1997/11/03 13:00:02  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.2  1997/10/27 11:18:08  cinard
// Added language independant symbols to the symboltable package.
//
// Revision 1.1  1997/10/13 15:01:19  cinard
// Generic classes used to manage symbols.
// Should be used by all ADL parsers.
//
// ***********************************************************************
