// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/Symbol.java,v $
// $Revision: 1.12 $
// $Date: 1998/03/18 15:48:02 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        Symbol.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import com.objectspace.jgl.UnaryPredicate;

/**
 * A symbol is any identifier that may be encountered while parsing an
 * ADL source files, amongst symbols are identifiers for types, classes,
 * methods or attributes. The main attribute of a symbol is its name.
 * @author Cyril Inard d'Argence
 */

public interface Symbol
{

  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */
  public String getName();

  /**
   * Get the symbol's full name (e.g. fully qualified name).
   */
  public String getFullName();

  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */
  public void   setName(String n);

  /**
   * Check if symbol has direct or indirect access to a scope.
   * Direct access means the symbol owns a scope (e.g. package, class)
   * Indirect access means the symbol does not own a scope but is 
   * related to a scope (e.g. a reference to an object)
   * Simply speaking any symbol on which the scope operator (. or ::)
   * can be applied has access to a scope.
   * @return True if symbol has direct or indirect access to a scope,
   *         false otherwise.
   */
  public boolean hasScopeAccess();

  /**
   * Get the scope the symbol has access to.
   * @return The scope that can be accessed to.
   */
  public Scope   getScope();

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */
  public UnaryPredicate getComparator();


  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise.
   */
  public boolean isDefined();

  /**
   * Indicates that the symbol is both declared and defined.
   */
  public void setDefined();

  /**
   * Get the symbol's modifier.
   * @return symbol's modifier.
   */
  public LModifier getModifier();

  /**
   * Set the symbol's modifier.
   * @param modifier symbol's modifier.
   */
  public void setModifier(LModifier modifier);

  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */
  public boolean isA(long symbolId);

  /** 
   * Symbol's identifier.
   */
  public long getSymbolId();

  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...)
   */
  public Symbol getEnclosingConstruct();

  /**
   * Returns a description of the symbol, indented.
   * @param indent Indentation string.
   * @return Description of the symbol.
   */
  public String toString(String indent);

  /**
   * Returns a description of the symbol, not indented.
   * @return Description of the scope.
   */

  public String toString();

}

// ***********************************************************************
// History Log:
// $Log: Symbol.java,v $
// Revision 1.12  1998/03/18 15:48:02  cinard
// Removal of ambiguities on operation invocation.
//
// Revision 1.11  1998/02/27 12:48:44  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.10  1998/02/23 15:09:08  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.9  1998/02/12 16:03:19  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.8  1998/01/19 14:37:05  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.7  1998/01/12 13:19:14  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.6  1997/11/26 17:13:05  cinard
// Changed package name.
//
// Revision 1.5  1997/11/12 13:43:23  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.4  1997/11/04 16:04:25  cinard
// Made some modification for C and C++ parsers.
//
// Revision 1.3  1997/11/03 13:00:03  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.2  1997/10/27 11:18:08  cinard
// Added language independant symbols to the symboltable package.
//
// Revision 1.1  1997/10/13 15:01:19  cinard
// Generic classes used to manage symbols.
// Should be used by all ADL parsers.
//
// ***********************************************************************
