// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/SymbolTableMgr.java,v $
// $Revision: 1.33 $
// $Date: 1998/07/28 13:19:17 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable
// NAME:        SymbolTableMgr.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import com.objectspace.jgl.UnaryPredicate;
import com.objectspace.jgl.predicates.UnaryAnd;
import com.objectspace.jgl.predicates.UnaryNot;


/**
 * Abstract symbol table manager that shall be used to derive concrete 
 * symbol table manager. Most of the methods have default implementation.
 */

abstract public class SymbolTableMgr 
{
  // -- Inner classes --

  /**
   * Helper class used to save information about a symbol research. This
   * information will help to search the symbol again in case it was not
   * found during the first research, and then call the symbol listener
   * back.
   * Note this class is part of the symbol listener mechanism 
   * implementation.
   */
  public class SymbolListenerInfo
  {
    /**
     * Constructs a symbol listener information.
     * @param symbolListener symbol listener interested in the symbol.
     * @param scopedName fully scoped name of the symbol that was searched.
     * @param scopeOverride scope in which the search shall occur.
     * @param predicate predicate used to filter suitable symbols.
     */
    SymbolListenerInfo
    (
       SymbolListener symbolListener,
       String       scopedName,
       Scope        scopeOverride,
       UnaryPredicate       predicate
    )
    {
       this.symbolListener = symbolListener;
       this.scopedName   = scopedName;
       this.scopeOverride  = scopeOverride;
       this.predicate     = predicate;
    }
   
    /** Symbol listener interested in the symbol. */

    public SymbolListener symbolListener;

    /** Fully scoped name of the symbol that was searched. */

    public String       scopedName;

    /** Scope in which the search shall occur. */

    public Scope        scopeOverride;

    /** Predicate used to filter suitable symbols. */

    public UnaryPredicate       predicate;
  }

  // -- Public Constructors --

  /**
   * Constructs symbol table manager.
   * <P>Unit scope is set to the Compilation Unit scope.</P>
   */

  public SymbolTableMgr()
  {
     this.unitScope  = new LCompilationUnit(null);
     setCurrentScope(this.unitScope);
     symbolListenerInfos = new Vector();
  }

  // -- Public Methods --

  /**
   * Reinitialize the symbol table. Should be used when a new compilation
   * unit is parsed.
   */
  public void reInit()
  {
     _parsedTDDClass = null;
     _usedTDDClasses = new Vector();
     _usedADLClasses = new Vector();
  }

  /**
   * Reinitialize the symbol table, and set symbol table context.
   * Should be used when a new compilation unit is type checked.
   * @param stmContext context information obtained from a former parsing.
   */

  public void setContext(Object stmContext)
  {
     if ( debug )
     {
        System.out.println
        ("debug : SymbolTableMgr.setContext(" + stmContext + ")");
     }
     if ( stmContext instanceof ADLCompilationUnitInfo )
     {
        ADLCompilationUnitInfo adlInfo = (ADLCompilationUnitInfo)stmContext;
        _usedADLClasses = adlInfo.getUsedADLClasses();
        _usedTDDClasses = adlInfo.getUsedTDDClasses();
     }
  }

  // -- Scope interface implementation --

  /**
   * Opens an anonymous scope into the current scope.
   * Current scope is set to the new scope.
   * Anonymous scopes are those scopes not directly related to symbols,
   * an example of such an anonymous scope the scope defined by a simple
   * block of code.
   * @return the opened anonymous scope.
   */

  public Scope openScope()
  {
     setCurrentScope(currentScope.openScope());
     return currentScope;
  };

  /**
   * Opens a scope associated with a symbol in the current scope.
   * Current scope is set to the new scope.
   * @param s The symbol for which the scope is to be opend.
   * @return the opened anonymous scope.
   */

  public Scope openScope(Symbol s)
  {
     setCurrentScope(currentScope.openScope(s));

     // In case symbol is a TDD class set the list of used
     // tdd classes.
     if ( s instanceof TDDClass )
     {
        _parsedTDDClass = (TDDClass)s;
        _parsedTDDClass.setUsedTDDClasses(_usedTDDClasses);
     }

     return getCurrentScope();
  };

  /**
   * Closes the current scope.
   * Current scope becomes the current scope's parent.
   */
  public void closeScope()
  {
     setCurrentScope(currentScope.getParent());
  }; 


  /**
   * Get a symbol, simple version.
   * <p>Should be called to search for symbols whose names were classified
   * ambiguous.
   * <p>The symbol table manager tries to classify the supplied name and to 
   * return an adhoc symbol for it.
   * @param scopedName The scoped name of the expected symbol.
   * @return The found symbol.
   */
  public Symbol getSymbol(String scopedName)
  {
     return getSymbol(scopedName,null,null,null);
  }

  /**
   * Get a symbol, extended version.
   * <p>Should be called to search for symbols whose names were classified 
   * ambiguous.
   * <p>The symbol table manager tries to classify the supplied name and to 
   * return an adhoc symbol for it.
   * <pre>
   * Scope override shall be supplied to override the default behaviour,
   * which is to search in the current scope. This enable incremental
   * search for complex expressions like foo.bar[].boo.baa().zip.
   * </pre>
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

  public Symbol getSymbol(String scopedName, 
                          UnaryPredicate predicate,
                          SymbolListener symbolListener,
                          Scope scopeOverride)
  {
     return getSymbol(scopedName,predicate,symbolListener,scopeOverride,false);
  }

  /**
   * Get a symbol, extended version.
   * <p>Should be called to search for symbols whose names were classified 
   * ambiguous.
   * <p>The symbol table manager tries to classify the supplied name and to 
   * return an adhoc symbol for it.
   * <pre>
   * Scope override shall be supplied to override the default behaviour,
   * which is to search in the current scope. This enable incremental
   * search for complex expressions like foo.bar[].boo.baa().zip.
   * </pre>
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  abstract public Symbol getSymbol
                         (
                            String scopedName, 
                            UnaryPredicate predicate,
                            SymbolListener symbolListener,
                            Scope scopeOverride,
                            boolean localSearch
                         );


  /**
   *  Returns compilation unit top level scope.
   */

  public Scope getCompilationUnit()
  {
     return unitScope;
  }

  /**
   * Get a module.
   */

  public LModule getModule(String name)
  {
      LModule ret  = (LModule)unitScope.getSymbol
                              (name,
                               SymbolSelector.MODULE_SELECT,
                               null);
      return ret;
  }

  /**
   * Indicate parsing is completed for all modules.
   */

  public void indicateParsingIsCompleted()
  {
     if ( unitScope instanceof LModule )
     {
        ((LModule)unitScope).setParsingCompleted();
     }
  }

  
  /**
   * Get a package symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

  public LPackage getPackage(String scopedName,
                             Scope scopeOverride)
  {
     return getPackage(scopedName,scopeOverride,false);
  }

  /**
   * Get a package symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LPackage getPackage(String scopedName,
                             Scope scopeOverride,
                             boolean localSearch)
  {
     return (LPackage) getSymbol
                       (
                        scopedName,
                        SymbolSelector.PACKAGE_SELECT,
                        null,
                        scopeOverride,
                        localSearch
                       );
  }


  /**
   * Get a type symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

   public LType getType(String scopedName,Scope scopeOverride)
   {
      return getType(scopedName,scopeOverride,false);
   }

  /**
   * Get a type symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

   public LType getType(String scopedName,
                        Scope scopeOverride,
                        boolean localSearch)
   {
      if ( debug ) 
      {
         System.out.println("debug: SymbolTableMgr.getType("
                            + scopedName + "," 
                            + scopeOverride + ","
                            + localSearch + ")");
      }
      LType ret = (LType) getSymbol
                          (        
                           scopedName,
                           SymbolSelector.TYPE_SELECT,
                           null,
                           scopeOverride
                          );
      if ( debug ) 
      {
         System.out.println("debug: SymbolTableMgr.getType("
                            + scopedName
                            + "," + scopeOverride 
                            + "," + localSearch + ")"
                            + " return " + ret );
      }
      return ret;
   }

  /**
   * Get a class symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @return The found symbol.
   */

   public LClass getClass(String scopedName)
   {
      return getClass(scopedName,null,false);
   }

  /**
   * Get a class symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

   public LClass getClass(String scopedName,Scope scopeOverride)
   {
      return getClass(scopedName,
                      scopeOverride,
                      false);
   }

  /**
   * Get a class symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */
  
   public LClass getClass(String scopedName,
                          Scope scopeOverride,
                          boolean localSearch)
   {
      return (LClass) getSymbol
                      (
                       scopedName,
                       SymbolSelector.CLASS_SELECT,
                       null,
                       scopeOverride,
                       localSearch
                      );
   }

  /**
   * Get an ADL class symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

   public ADLClass getADLClass(String scopedName,Scope scopeOverride)
   {
      return (ADLClass) getSymbol
                        (
                         scopedName,
                         SymbolSelector.ADL_CLASS_SELECT,
                         null,
                         scopeOverride
                        );
   }

  /**
   * Get a TDD class symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @return The found symbol.
   */

   public TDDClass getTDDClass(String scopedName)
   {
      if ( debug )
      {
         System.out.println
         ("debug: SymbolTableMgr.getTDDClass("+scopedName+")");
      }
      // Search in unitScope
      TDDClass ret = (TDDClass)getSymbol(scopedName,
                                         SymbolSelector.TDD_CLASS_SELECT,
                                         null,
                                         getUnitScope(),
                                         true);
      if ( debug )
      {
         System.out.println
         ("debug: SymbolTableMgr.getTDDClass("+scopedName+") return "+ret);
      }
      return ret;
   }

  /**
   * Get a struct symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

  public LClass getStruct(String scopedName,Scope scopeOverride)
  {
     return getStruct(scopedName,scopeOverride,false);
  }

  /**
   * Get a struct symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LClass getStruct(String scopedName,
                          Scope scopeOverride,
                          boolean localSearch)
  {
     return (LClass) getSymbol
                      (
                       scopedName,
                       SymbolSelector.STRUCT_SELECT,
                       null,
                       scopeOverride,
                       localSearch
                      );
  }

  /**
   * Get a union symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

  public LUnion getUnion(String scopedName,Scope scopeOverride)
  {
     return getUnion(scopedName,scopeOverride,false);
  }

  /**
   * Get a union symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LUnion getUnion(String scopedName,
                         Scope scopeOverride,
                         boolean localSearch)
  {
     return (LUnion) getSymbol
                     (
                      scopedName,
                      SymbolSelector.UNION_SELECT,
                      null,
                      scopeOverride,
                      localSearch
                     );
  }

  /**
   * Get an expression symbol (local variable, field, parameter).
   * Should be called to search for names that were classified as expressions.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @return The found symbol.
   */

  public LExpression getExpression(String scopedName, Scope scopeOverride)
  {  
     return getExpression(scopedName, scopeOverride, false);
  }

  /**
   * Get an expression symbol (local variable, field, parameter).
   * Should be called to search for names that were classified as expressions.
   * @param scopedName   The scoped name of the expected symbol.
   * @param scopeOverride  Scope in which the research shall take place,
   *                       if null is supplied current scope is assumed.
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LExpression getExpression(String scopedName,
                                   Scope scopeOverride,
                                   boolean localSearch)
  {  
     return (LExpression) getSymbol
                          (
                             scopedName,
                             SymbolSelector.EXPRESSION_SELECT,
                             null,
                             scopeOverride,
                             localSearch
                          );
  }

  /**
   * Get a function or method.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @return The found symbol.
   */

  public LCallable getMethod
                   (
                    String scopedName,
                    UnaryPredicate predicate,
                    SymbolListener symbolListener,
                    Scope scopeOverride
                   )
  {
    return getMethod
           (scopedName,predicate,symbolListener,scopeOverride,false);
  }

  /**
   * Get a function or method.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LCallable getMethod
                   (
                    String scopedName,
                    UnaryPredicate predicate,
                    SymbolListener symbolListener,
                    Scope scopeOverride,
                    boolean localSearch
                   )
  {
 
     LCallable ret = null;

     if ( predicate instanceof SymbolSelector )
     {
        ((SymbolSelector)predicate).setSymbolIdMask
                                    (      SymbolId.OPERATION_MASK
                                     & ( ~ SymbolId.ADL_METHOD) );
        ret = (LCallable)getSymbol(scopedName,
                                   predicate,
                                   symbolListener,
                                   scopeOverride,
                                   localSearch);
     }

     return ret ;
  }

  /**
   * Get a constructor symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @return The found symbol.
   */

   public LCallable getConstructor
                    (
                       String scopedName,
                       UnaryPredicate predicate,
                       SymbolListener symbolListener,
                       Scope scopeOverride
                    )
  {
     return getConstructor
            (scopedName,predicate,symbolListener,scopeOverride,false);
  }

  /**
   * Get a constructor symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate     An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param symbolListener Reference to the symbol listener that shall
   *                       be called back later in case the symbol
   *                       is not found during the call (may be null).
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public LCallable getConstructor
                   (
                    String scopedName,
                    UnaryPredicate predicate,
                    SymbolListener symbolListener,
                    Scope scopeOverride,
                    boolean localSearch
                   )
  {
     LCallable ret = null;

     // First, get search scope (should be a class symbol)
     LClass lClass = null;
     if ( localSearch )
     {
        if ( scopeOverride instanceof LClass )
        {
           lClass = (LClass) scopeOverride;
        }
     }
     else
     {
        lClass =  getClass(scopedName,scopeOverride);
     }

     // Insulate constructor name
      
     // Split the scoped name -> scopedName [ :: | . ] name 
     int index              = scopedName.lastIndexOf(".");
     int scopeOverrideIndex = scopedName.lastIndexOf("::");
     String name;
     
     int maxIndex = Math.max(index,scopeOverrideIndex);
     if ( index == maxIndex )
     {   
        name       = scopedName.substring(index + 1);
     }
     else // scopeOverrideIndex == maxIndex
     {
        name       = scopedName.substring(scopeOverrideIndex + 2);
     }

     if ( predicate instanceof SymbolSelector )
     {
        ((SymbolSelector)predicate).setSymbolIdMask( SymbolId.METHOD
                                                   | SymbolId.CONSTRUCTOR);
        ret =  (LCallable)getMethod
               (name,predicate,symbolListener,(Scope)lClass,true);
     }

     return ret;

  }

  /**
   * Get an ADL method symbol. 
   * @param scopedName The scoped name of the expected symbol.
   * @param predicate  An arbitrary predicate object defining some
   *                   criteria that the searched symbol shall
   *                   satisfy (may be null).
   * @return The found symbol.
   */

  public LCallable getADLMethod
                   (
                    String scopedName,
                    UnaryPredicate predicate,
                    Scope scopeOverride
                   )
  {
     return getADLMethod(scopedName,predicate,scopeOverride,false);
  }

  /**
   * Get an ADL method symbol. 
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate    An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param localSearch  should be true in case the search has to
   *                     take place in current scope only.
   * @return The found symbol.
   */

  public LCallable getADLMethod
                   (
                    String scopedName,
                    UnaryPredicate predicate,
                    Scope scopeOverride,
                    boolean localSearch
                   )
  {
     ADLMethod ret = null;

     if ( predicate instanceof SymbolSelector )
     {
        ((SymbolSelector)predicate).setSymbolIdMask(SymbolId.ADL_METHOD);
        ret = (ADLMethod)getSymbol(scopedName,
                                   predicate,
                                   null,
                                   scopeOverride,
                                   localSearch);
     }

     return ret;
  }

  /**
   * Get a TDD factory symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate    An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @return The found symbol.
   */

  public TDDFactory getTDDFactory
                    (
                     String scopedName,
                     UnaryPredicate predicate,
                     Scope scopeOverride
                    )
  {
     return getTDDFactory(scopedName,predicate,scopeOverride,false);
  }

  /**
   * Get a TDD factory symbol.
   * @param scopedName   The scoped name of the expected symbol.
   * @param predicate    An arbitrary predicate object defining some
   *                     criteria that the searched symbol shall
   *                     satisfy (may be null).
   * @param localSearch    should be true in case the search has to
   *                       take place in current scope only.
   * @return The found symbol.
   */

  public TDDFactory getTDDFactory
                    (
                     String scopedName,
                     UnaryPredicate predicate,
                     Scope scopeOverride,
                     boolean localSearch 
                    )
  {
     TDDFactory ret = null;

     if ( predicate instanceof SymbolSelector )
     {
        ((SymbolSelector)predicate).setSymbolIdMask(SymbolId.TDD_FACTORY);

        ret = (TDDFactory)getSymbol(scopedName,
                                    predicate,
                                    null,
                                    scopeOverride,
                                    localSearch);
     }

     return ret;
  }

  // -- Put methods --

  /**
   * Put a symbol in the symbol table, simple version.
   * The symbol is put in the current scope.
   * In case the symbol is already in the scope (because of a earlier
   * declaration) the methods returns the already declared symbol, 
   * otherwise, it put the supplied symbol in the current scope and 
   * returns it.
   * @param symbol The symbol to put into the scope.
   * @return the symbol that was actually found or put in the symbol
   *         table.
   * @exception ORG.opengroup.adl.symboltable.SymbolAlreadyDefinedException 
   *            in case symbol is already defined.
   */

  public Symbol putSymbol(Symbol symbol)
     throws SymbolAlreadyDefinedException
  {
     return putSymbol(symbol,null);
  }

  /**
   * Put a symbol in the symbol table, extended version.
   * The symbol is put in the supplied scope or in current scope in case no 
   * scope was supplied.
   * In case the symbol is already in the scope (because of a earlier
   * declaration) the methods returns the already declared symbol, 
   * otherwise, it put the supplied symbol in the current scope and 
   * returns it.
   * @param symbol The symbol to put into the scope.
   * @param scopeOverride optionnal scope to take into account instead
   *                      of current scope.
   * @return the symbol that was actually found or put in the symbol
   *         table.
   * @exception ORG.opengroup.adl.symboltable.SymbolAlreadyDefinedException 
   *            in case symbol is already defined.
   */

  public Symbol putSymbol(Symbol symbol,Scope scopeOverride)
     throws SymbolAlreadyDefinedException
  {
     if ( scopeOverride == null ) { scopeOverride = getCurrentScope(); }

     // Scope where the symbol is to be put
     Scope putScope = (symbol instanceof LModule ? unitScope : scopeOverride );

     // Check if a symbol with the same declaration exists in the 
     // current scope

     String         name       = symbol.getName();
     UnaryPredicate comparator = symbol.getComparator();

     Symbol existingSymbol = putScope.getSymbol
                             (symbol.getName(),comparator,null);

     if ( existingSymbol != null )
     {
        // Symbol exists, check this symbol is not defined.
        if ( existingSymbol.isDefined() )
        {
           // It is defined, throw an exception.
           throw  new SymbolAlreadyDefinedException
                      (symbol + " already defined");
        }
        else
        {
           // Symbol is not yet defined, return the symbol already
           // in scope.
           return existingSymbol;
        }
     }
     else
     {
        // Symbol is not yet declared, put it in the scope and return it.
        putScope.putSymbol(symbol);
        return symbol;
     }
     
  }

  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now up to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   * @param defaultModifier default modifier (bit mask)
   */
  public void setDefaultModifier(int defaultModifier)
  {
     currentScope.setDefaultModifier(defaultModifier);
  }
  
  /**
   * Set current scope where symbols should be searched by default.
   * @param scope current scope.
   */

  public void setCurrentScope(Scope scope)
  {
    if ( debug ) System.out.println("debug: SymbolTableMgr.setCurrentScope(" 
                                   + scope + ")");
     currentScope = scope;
  }

  /**
   * Returns current scope.
   */
  public Scope getCurrentScope()
  {
     return currentScope;
  }

  /**
   * Get a builtin type symbol.
   * @param name   The name of the builtin type.
   * @return The found symbol.
   * @exception ORG.opengroup.adl.symboltable.BadSymbolNameException
   *            in case supplied name is not that of a builtin type.
   */

  abstract public LType getBuiltinType(String name)
     throws BadSymbolNameException;

  /**
   * Checks if a builtin type name is valid or not.
   * @param name The name to check.
   * @return true if it is valid, false otherwise.
   */

  abstract public boolean isValidBuiltinTypeName(String name);

  // -- Check scope methods --

  /**
   * Return true if the current scope is inside a scope defined
   * by a symbol of the supplied class, false otherwise.
   * <p>Example :
   * <pre>
   * SymbolTableMgr stm = new SymbolTableMgr();
   * ...
   * if ( stm.isInScope(LCallableClass.class) )
   * {
   *    throw new ParseException("Pertinent error message!");
   * }
   * </pre>
   * @param scopeClass class of the scope for which the check is performed.
   */

   public boolean isInScope(Class scopeClass)
   {
      Scope scope = currentScope;
      while ( scope != null )
      {
         if ( scopeClass.isInstance(scope) ) { return true; }
         scope = scope.getParent();
      }
      return false;
   }

  // -- ADL file check --

  /**
   * Returns full path name of the source file defining an ADL class if
   * this file exists and can be reached from incpath.
   * @param name name of the adl class.
   * @return full path name if file is found, null otherwise.
   */
 
  public String getFileNameForADLClass(String name)
  {
     return _adlFileChecker.getFileNameForADLClass(name);
  }

  // -- TDD file check --

  /**
    * Returns full path name of the source file defining a TDD class if
    * this file exists and can be reached from incpath.
    * @param name name of the tdd class.
    * @return full path name if file is found, null otherwise.
    */
 
  public String getFileNameForTDDClass(String name)
  {
     if ( debug )
     {
        System.out.println
        ("debug : SymbolTableMgr.getFileNameForTDDClass("+name+")");
     }
     String ret = _adlFileChecker.getFileNameForTDDClass(name);
     if ( debug )
     {
        System.out.println
        ("debug : SymbolTableMgr.getFileNameForTDDClass("+name+") return " 
        + ret);
     }
     return ret;
  }

  // -- Specific methods --

  /**
   * Get compilation unit scope.
   */

  public Scope getUnitScope()
  {
     return unitScope;
  }

  /**
   * Tells symbol table manager that parsing is finished and that 
   * symbol listeners should be called back.
   */

  public void finishParsing()
  {
     Enumeration e = symbolListenerInfos.elements();
   
     while ( e.hasMoreElements() )
     {
        SymbolListenerInfo symbolListenerInfo =
                           (SymbolListenerInfo)e.nextElement();

        // Try to get the symbol now
        
        Symbol s = getSymbol
                   (
                    symbolListenerInfo.scopedName,
                    symbolListenerInfo.predicate,
                    null,
                    symbolListenerInfo.scopeOverride
                   );
        
        // Whatever is retrieved send it back to the symbol listener

        symbolListenerInfo.symbolListener.symbolFound(s);
     }
     
     symbolListenerInfos.removeAllElements();
  }

  /**
   * Add a used TDD class (in case current parsed class is a TDD class).
   */

  public void addUsedTDDClass(TDDClass tddClass)
  {
     if ( debug ) 
     {
        System.out.println
        ("debug: SymbolTableMgr.addUsedTDDClass(" + tddClass.getName() + ").");
     }
     _usedTDDClasses.addElement(tddClass);
  }

  /**
   * Add a used ADL class (in case current parsed class is a TDD class).
   */

  public void addUsedADLClass(ADLClass adlClass)
  {
     if ( debug ) 
     {
        System.out.println
        ("debug: SymbolTableMgr.addUsedADLClass(" + adlClass.getName() + ").");
     }
     _usedADLClasses.addElement(adlClass);
  }

  /** Return an a factory for symbols. */

  abstract public SymbolFactory getSymbolFactory();


  // -- toString method --

  /**
   * Returns a description of the symbol table. Symbols are displayed
   * in a tree structure.
   */
  public String toString()
  {
     return unitScope.toString("   ");
  }

  // -- protected methods --

  /** 
   * Returns a symbol listener information object.
   * @param symbolListener symbol listener interested in the symbol.
   * @param scopedName fully scoped name of the symbol that was searched.
   * @param scopeOverride scope in which the search shall occur.
   * @param predicate predicate used to filter suitable symbols.
   * @return newsymbol listener information object. 
   */

  protected Object newSymbolListenerInfo
                   (
                      SymbolListener symbolListener,
                      String scopedName, 
                      Scope currentScope,
                      UnaryPredicate predicate
                   )
  {
      return new SymbolListenerInfo 
                 ( symbolListener, scopedName, currentScope, predicate);
  }

  /**
   * Stores symbol listener information so that client can be called back
   * after the symbol is found.
   * @param sli symbol listener information object.
   */

  protected void storeSymbolListenerInfo(Object sli)
  {
     symbolListenerInfos.addElement(sli);
  }

  // -- protected attributes --
  
  /** Global scope for the symbol table manager. */
  protected Scope unitScope;

  /** Current scope of the symbol table manager. */
  protected Scope currentScope;

  /** Stores information for the SymbolListener objects. */
  protected Vector symbolListenerInfos;


  /** 
   * Get used TDD Classes.
   */
  protected Vector getUsedTDDClasses()
  {
     return _usedTDDClasses;
  }

  /** 
   * Get used ADL Classes.
   */
  public Vector getUsedADLClasses()
  {
     return _usedADLClasses;
  }

  // -- private --

  /** ADL file checker. */
  private ADLFileChecker _adlFileChecker = new ADLFileChecker();

  /** Used tdd classes. */
  private Vector _usedTDDClasses = new Vector();

  /** Used tdd classes. */
  private Vector _usedADLClasses = new Vector();

  /** Currently parsed TDDClass. */
  private TDDClass _parsedTDDClass = null;

  /** Debug flag. */
  private static final boolean debug = false;

}

// ***********************************************************************
// History Log:
// $Log: SymbolTableMgr.java,v $
// Revision 1.33  1998/07/28 13:19:17  mallet
// Added management of list of used ADL classes.
//
// Revision 1.32  1998/05/27 09:56:59  cinard
// Made some clean up in code and improved debug messages.
//
// Revision 1.31  1998/05/11 11:48:45  cinard
// Added getModule() method.
// Added indicateParsingIsCompleted() method.
// These methods are part of the module incusion mechanism (this mechanism
// helps when handling #included files).
//
// Revision 1.30  1998/04/28 12:19:58  cinard
// Improved debug messages in getSymbol().
//
// Revision 1.29  1998/03/25 14:51:53  cinard
// Changed some method visibility.
//
// Revision 1.28  1998/03/19 16:42:56  cinard
// Fixed a bug in SymbolTablMgr.getMethod() function with CPP.
//
// Revision 1.27  1998/03/19 15:35:05  cinard
// Simplified ADL file check mechanism.
//
// Revision 1.26  1998/03/18 15:51:56  cinard
// Symbol factory is now available via getFactory().
// Implemented support for symbol table context (mainly
// used to keep context information when parsing and
// typecheck are separated).
// Updated to new Symbol selection mechanism.
//
// Revision 1.25  1998/03/02 15:32:28  cinard
// Replaced LStruct with LClass (struct symbols behave just like classes).
// Factory and find methods (newStruct() and getStruct() remains with
// their signatures updated).
//
// Revision 1.24  1998/02/27 12:48:45  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
// Updated copyright notice.
//
// Revision 1.23  1998/02/26 14:32:30  wassel
// getMethod returns an LCallable that is not an ADLMethod. This is to avoid conflict between LMethods and ADLMethods having the same predicate.
// To get a reference of an ADLmethod, getADLMethod should be used.
//
// Revision 1.22  1998/02/25 11:05:43  cinard
// Introduced a new flag (localSearch) in getSymbol() and likes methods. A local
// search is restricted to the current (or supplied) scope, that is if the
// searched symbol is not found in current or supplied scope, the research
// does not continue in the parent scopes.
//
// Revision 1.21  1998/02/23 15:09:08  cinard
// Modified symbols inheritance tree. Now LType and its subclasses are
// interfaces. New L...Impl classes provide implementation for these
// interfaces. New L...Typedef classes provide implementations for
// typedefs (C and C++).
//
// Typedefs are created using the SymbolTableManager.newTypedef()
// method.
//
// Revision 1.20  1998/02/19 13:15:41  wassel
// Fixed bug in getMethod: casting found symbol as an LCallable should be done only when symbol is instance of LCallable.
//
// Revision 1.19  1998/02/12 16:03:21  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.18  1998/02/10 13:10:20  cinard
// Reinitialize method added so that the symbol table can be reinitialized
// (useful when new compilation unit are parsed).
//
// Revision 1.17  1998/02/05 10:55:57  wassel
// Added factory methods.
//
// Revision 1.16  1998/01/19 14:37:05  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.15  1998/01/12 13:19:15  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.14  1998/01/07 11:01:06  cinard
// Fixed a bug in get constructor method.
//
// Revision 1.13  1997/12/16 09:22:27  cinard
// Added getTDDFactoryMethod().
//
// Revision 1.12  1997/12/08 12:58:46  cinard
// Added a new type check mechanism.
// Fixed some bugs.
//
// Revision 1.11  1997/11/26 17:13:06  cinard
// Changed package name.
//
// Revision 1.10  1997/11/19 15:48:59  cinard
// Corrected some minor bugs.
//
// Revision 1.9  1997/11/14 09:58:55  cinard
// LModule scope now accepts duplicate symbols.
// SymbolTableMgr has a getADLClass() method.
//
// Revision 1.8  1997/11/13 15:41:01  cinard
// Improved ADLInline management. Added SymbolTableMgr.newADLInline() method.
//
// Revision 1.7  1997/11/13 13:27:41  cinard
// Added a scope override parameter to putSymbol method.
// Both Symbol putSymbol(Symbol s, Scope scopeOverride) and
// Symbol putSymbol(Symbol s) can be used now.
//
// Revision 1.6  1997/11/13 09:43:31  cinard
// Improved Builtin type management. Notably symboltable.c.CSymbolTableMgr.
// getBuiltinType() handles all forms of builtin types names.
//
// Revision 1.5  1997/11/12 13:43:23  cinard
// Corrected some bugs.
// Replaced LCallableSearcher class with LCallableAssignComparator and
// LCallableEqualsComparator classes.
// Added LModuleComparator and LTypeComparator classes (useful for
// Java names classification).
//
// Revision 1.4  1997/11/04 16:03:25  cinard
// Made modification for C and C++ parsers, notably :
// o handling scoped names beginning with ::
// o added getClass(), getStruct() and getUnion() methods
// o modified putSymbol() to allow separate declaration and definition
//   of symbols,
// o enable default access type setting.
//
// Revision 1.3  1997/11/04 09:04:31  cinard
// Scope.getSymbol() does not throw AmbiguousRequestException anymore.
// symboltable.java package name replaced with symbol_table.j (problems
// with javadoc). Note that symboltable/CVS/* should be updated manually
// (replacing java with j where necessary).
//
// Revision 1.2  1997/11/03 13:00:03  cinard
// Improved scope management : duplicates in LClass scope allowed.
// Added PoinerToFunction class.
//
// Revision 1.1  1997/10/13 15:01:20  cinard
// Generic classes used to manage symbols.
// Should be used by all ADL parsers.
//
// ***********************************************************************
