// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/j/JClassProxy.java,v $ 
// $Revision: 1.14 $ 
// $Date: 1998/05/27 09:24:54 $ 
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable.j
// NAME:        JClassProxy.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 31,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable.j;

import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.jbcfile.ClassFile;

import java.io.FileNotFoundException;
import java.util.Vector;
import java.util.Enumeration;

import com.objectspace.jgl.UnaryPredicate;
import com.objectspace.jgl.HashSet;


/**
 * Java Proxy Class symbol. 
 * <p> Proxy classes are used to avoid loading a whole class bytecode
 * before class information is really needed. Only at these points is
 * the class symbol completely loaded.
 */

public final class JClassProxy implements LClass
{
  // -- Constructors --

  /**
   * Constructs a class proxy.
   * @param fullName fully qualified name of the class 
   * @param jsl java symbol loader to use to load the class.
   * @param related arbitrary related object (e.g. bytecode class file)
   */
   
  public JClassProxy(String        fullName,
                     JSymbolLoader jsl,
                     Object        related)
  {
     _name            = fullName.substring(fullName.lastIndexOf(".")+1);
     _fullName        = fullName;
     _loaded          = false;
     _symbolLoader    = jsl;
     _related         = related;
  }

  // -- Public methods --

  // -- Object --

  /**
   * Is this equals to supplied o.
   */

  public boolean equals(Object o)
  {
     // Optimization : if proxy is not loaded just check this equals o
     
     if ( ! _loaded ) { return super.equals(o); }
     return _class.equals(o);

  }

  // -- Symbol interface implementation --

  /**
   * Get the symbol's name.
   * @return The symbol's name.
   */

  public String getName() { return _name; }


  /**
    * Get the scope name in NL format (vector of strings).
    * Here, this is in fact the associated class scope name
    */
  public Vector getNLName() 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getNLName();
  }

  /**
   * Get the symbol's full name.
   */
  
  public String getFullName() { return _fullName; }

  /**
   * Set the symbol's name.
   * @parma n The symbol's name.
   */

  public void   setName(String n) 
  {
     // TBD - think about removing this method
     // _name = n;
  }

  /**
   * Check if symbol has direct or indirect access to a scope.
   * Direct access means the symbol owns a scope (e.g. package, class)
   * Indirect access means the symbol does not own a scope but is 
   * related to a scope (e.g. a reference to an object)
   * Simply speaking any symbol on which the scope operator (. or ::)
   * can be applied has access to a scope.
   * @return True if symbol has direct or indirect access to a scope,
   *         false otherwise.
   */

  public boolean hasScopeAccess() 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.hasScopeAccess();
  }

  /**
   * Get the scope the symbol has access to.
   * @return The scope that can be accessed to.
   */

  public Scope   getScope()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getScope();
  }

  /**
   * Get symbol's enclosing construct (e.g. a class, a TDD class,
   * an ADL class...)
   */

  public Symbol getEnclosingConstruct()
  {
     Scope parent = _parent;
     while ( parent != null )
     {
        if ( parent instanceof LModule )
        {
           return (LModule)parent;
        }
        parent = parent.getParent();
     }
     return null;
  }

  /**
   * Get a comparator enabling to select a symbol with the same
   * runtime class than that of the symbol against which this method
   * is called.
   * @return the comparator.
   */

  public UnaryPredicate getComparator()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getComparator();
  }

  /**
   * Returns true if the symbol is both declared and defined,
   * false otherwise.
   */

  public boolean isDefined() 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.isDefined();
  }

  /**
   * Indicates that the symbol is both declared and defined.
   */

  public void setDefined()
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.setDefined(); 
  } 

  /**
   * Get the symbol's modifier.
   * @return symbol's modifier.
   */

  public LModifier getModifier() 
  {
     if  ( ! _loaded ) { return _modifier; }
     else return _class.getModifier();
  }

  /**
   * Set the symbol's modifier.
   * @param modifier symbol's modifier.
   */

  public void setModifier(LModifier modifier) 
  {
     if ( ! _loaded ) { _modifier = modifier; }
     else { _class.setModifier(modifier); }
  }

  /**
   * Does the symbol's identifier is equals to id? Note a bit mask
   * containing the symbol's id bit is ok.
   * @param symbolId a symbol identifier.
   */

  public boolean isA(long symbolId) 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.isA(symbolId);
  }

  /** 
   * Symbol's identifier.
   */

  public long getSymbolId() 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getSymbolId();
  }

  /**
   * Returns a description of the symbol, indented.
   * @param indent Indentation string.
   * @return Description of the symbol.
   */

  public String toString(String indent) 
  {
     return   indent + "<proxy " + _fullName + ": loaded = " + _loaded + ">\n"
            + ( _loaded ? _class.toString(indent) : "" );
  }

  /**
   * Returns a description of the symbol, not indented.
   * @return Description of the symbol.
   */

  public String toString() 
  {
     return   "<proxy " + _fullName + ": loaded = " + _loaded + ">"  
            + ( _loaded ? " class = " + _class.getFullName() : "" );
  }

  // -- LType interface implementation --
  
  /**
   * Make up new typedef for this type.
   * @param name typedef name.
   */

  public LType newTypedef(String name)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.newTypedef(name);
  }

  /**
   * Returns synthesis type of the param t and type for which method applies.
   * @return is the smallest type assignable to both types,
   *         is null is no common assignable type found.
   */ 

  public LType synthesisType(LType t) 
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.synthesisType(t);
  }

  /**
   * Can a value of type t be assigned to a variable of this type.
   * @param  t The type for which the assignability is checked.
   * @return True if it can, false otherwise.
   */

  public boolean assignable(Symbol t)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.assignable(t);
  }

  /**
   * Returns true if type is void. Default is to return false.
   */

  public boolean isVoid()
  {
     // No need to load bytecode.
     return false;
  }


  /**
   * Returns a string for the declaration of an expression of this
   * type. Useful for mixed type and variable name (e.g. pointer
   * to function). Default behaviour is to return "type name".
   * @param name expression's name.
   */

  public String mixNameAndType(String name)
  {
     // Behave like JClass. No need to load bytecode.
     return _name + " " + name;
  }


  /** 
   *  Get the type kind. Useful for languages that defines several
   *  type kinds (e.g. in Java boolean, arithmetic and references).
   *  @return an integer indicating the kind of the type. 
   */

  public Integer getTypeKind()
  {
     // No need to load bytecode.
     return JTypeKind.REFERENCE_TYPE;
  }

  /**
   * Real type. That is the 'ultimate' real (i.e. not typedef'ed type ) type 
   * for which this type is defined. Default is to return this. 
   */

  public LType getRealType()
  {
     // No typedef in Java.
     return this;
  }

  // -- Scope interface implementation --

  /**
   * Opens an anonymous scope into this scope.
   * Anonymous scopes are those scopes not directly related to symbols,
   * an example of such an anonymous scope the scope defined by a simple
   * block of code.
   * @return The new anonymous scope opened.
   */

  public Scope openScope()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.openScope();
  }

  /**
   * Opens a scope associated with a symbol.
   * The provided symbol must owns its own scope (its hasScope() method
   * shall return true).
   * @param s The symbol for which the scope is to be opend.
   * @return The new scope opened.
   */

  public Scope openScope(Symbol s)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.openScope(s);
  }


  /**
   * Set the parent scope.
   * @param p the parent scope.
   */

  public void setParent(Scope p)
  {
     _parent = p;
  }


  /**
   * Get the parent scope.
   * @return The parent scope.
   */

  public Scope getParent()
  {
     return _parent;
     //if ( ! _loaded ) { loadProxy(); }
     //return _class.getParent();
  }


  /**
   * Put a symbol in the scope.
   * @param symbol The symbol to put into the scope.
   */

  public void putSymbol(Symbol symbol)
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.putSymbol(symbol);
  }

  /**
   * Get a symbol in the scope, extended version.
   * When using this version of getSymbol, supplementary parameters are
   * provided that shall help finding the object.
   * <pre>
   * The symbol shall satisfy some criteria defined by the provided 
   * predicate object. This predicate object help finding the symbol in 
   * case several symbols are likely to match the supplied name 
   * (e.g. overloaded methods).
   * </pre>
   * <pre>
   * Credentials shall be supplied for those cases where some symbols
   * hava non-public visibility. Retrieving them may require the
   * use of credentials.
   * </pre>
   * @param name The name of the expected symbol.
   * @param predicate An arbitrary predicate object.
   * @param credentials Credentials of the scope.
   * @return The found symbol.
   */

  public Symbol getSymbol
                (String name,UnaryPredicate predicate,Object credentials)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getSymbol(name,predicate,credentials);
  }


  /**
   * Get a symbol in the scope.
   * @param name The name of the expected symbol.
   * @return The found symbol.
   */

  public Symbol getSymbol(String name)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getSymbol(name);
  }

  
  /**
   * Set default modifier (int bit mask).  All symbols that
   * will be put in the scope from now to the next setDefaultModifier()
   * call will have their modifier modified (bitwise and with default
   * modifier).
   */

  public void setDefaultModifier(int defaultModifier) 
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.setDefaultModifier(defaultModifier);
  }

  // -- Constructed type specific methods --

  /**
   * Add an additional lookup scope. Useful to add additional scopes where
   * symbols can be searched (e.g. ADL global prologue).
   * @param lookupScope additional lookup scope.
   */

  public void addLookupScope(Scope scope)
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.addLookupScope(scope);
  }

  // -- LClass interface --

  /** 
   * Add a superclass.
   * @param superClass The super class to add.
   */

  public void addSuperClass(LClass superClass)
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.addSuperClass(superClass);
  }

  /** 
   * Add implemented interfaces.
   * @param lInterface The interface to add.
   */

  public void addInterface(LClass lInterface)
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.addInterface(lInterface);
  }


  /** 
   * Add a friend class.
   * @param aFriend friend class to add.
   */

  public void addFriend(LClass aFriend)
  {
     if ( ! _loaded ) { loadProxy(); }
     _class.addFriend(aFriend);
  }

  /**
   * Returns the superclasses of this class.
   * @return Array containing this class superclasses.
   */

  public LClass[] getSuperClasses()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getSuperClasses();
  }

  /**
   * Returns the number of superclasses of this class.
   * @return number of superclasses.
   */

  public int getSuperClassesCount()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getSuperClassesCount();
  }


  /**
   * Returns the interfaces this class implements.
   * @return Array containing this class interfaces
   */

  public LClass[] getInterfaces()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getInterfaces();
  }


  /**
   * Check the class defined by this symbol is a subclass of the
   * class defined by the supplied symbol.
   * @param jType Type symbol.
   * @return True if it is, false otherwise.
   */

  public boolean subclassOf(LType t)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.subclassOf(t);
  }

  /**
   * Returns names of all methods in this class or its superclasses.
   */

  public HashSet getMethodNames()
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getMethodNames();
  }

  /**
   * Get methods satisfying predicates in this class and all its 
   * superclasses.
   * @param name The name of the expected symbol.
   * @param predicate An arbitrary predicate object.
   * @return The methods.
   */

  public Vector getMethods(String name, UnaryPredicate predicate)
  {
     if ( ! _loaded ) { loadProxy(); }
     return _class.getMethods(name,predicate);
  }


  /** Check if class is loaded. */

  boolean isLoaded()  { return _loaded; }
  
  // -- File scope interface implementation --

  /**
   * Get the name of the underlying file.
   */

  public String getFileName() 
  { 
     throw new Error("Should not be called.");
  }


  /**
   * Add an included file scope. Unused.
   * @return The parent scope.
   */

  public void addIncludedFile(FileScope module) 
  { 
     throw new Error("Should not be called.");
  }

  /**
   * Indicate parsing is completed for this file, and recursively
   * for all included files.
   */

  public void setParsingCompleted()
  {
     throw new Error("Should not be called.");
  }

  /**
   * Is parsing completed?
   */

  public boolean isParsingCompleted()
  {
     throw new Error("Should not be called.");
  }

  // -- Private methods --

  /** Finish loading class symbol. */

  private void loadProxy()
  {
     _class = _symbolLoader.loadClass
              (_fullName,
               _related instanceof ClassFile ? (ClassFile)_related : null);

     if ( _class == null )
     {
        throw new Error("Could not load bytecode for " + _fullName); 
     }

     _loaded  = true;
     _related = null;
     _class.setParent(_parent);
  }

  // -- Private fields --

  /** Symbol loader */
  private final JSymbolLoader _symbolLoader;

  /** Name of the class */
  private final String _name;

  /** Loaded bytecode flag */
  private boolean _loaded = false;

  /** Full scoped name for the class */
  private final String _fullName;

  /** Loaded class */
  private JClass _class = null;

  /** Modifier */
  private LModifier _modifier = new JModifier(JModifier.PUBLIC);

  /** Related object. */
  Object _related = null;

  /** Parent scope */
  Scope _parent = null;

  /** Debug flag */
  private static final boolean debug = false;
}

// ***********************************************************************
// History Log:
// $Log: JClassProxy.java,v $
// Revision 1.14  1998/05/27 09:24:54  cinard
// Implementation of FileScope interface added.
//
// Revision 1.13  1998/05/15 13:25:39  bareau
// Added support for NL unparsing (mainly added a method getNLName() in Scope).
//
// Revision 1.12  1998/03/18 16:04:58  cinard
// Added fullName() method.
//
// Revision 1.11  1998/03/10 17:26:40  mallet
// Merged v1.1 with development trunk
//
// Revision 1.10.2.1  1998/03/06 19:27:18  cinard
// Set full class name in LClass.
//
// Revision 1.10  1998/02/27 12:38:44  cinard
// Updated copyright notice.
//
// Revision 1.9  1998/02/23 14:29:23  cinard
// True Proxy design pattern implemented for JClassProxy.
// Improved type compatibility checks (complex type checking mechanism
// has been replaced by assignable() methods in subclasses of LType).
// JTypeKind.java now defines type kind constants that were formerly
// in JTypeChecker.java (update your client code!).
//
// Revision 1.8  1998/02/12 16:03:32  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.7  1998/01/22 11:42:32  cinard
// Fixed a bug in symbols equality checks.
//
// Revision 1.6  1998/01/19 14:37:08  cinard
// Changed COM.objectspace.jgl to com.objectspace.jgl (update from JGL 3.0.0
// to 3.1.0).
//
// Revision 1.5  1998/01/12 13:19:20  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.4  1997/12/08 12:53:58  cinard
// Added type check mechanism.
// Fixed some byte code import bugs.
//
// Revision 1.3  1997/11/27 09:04:08  cinard
// Debug flag off!!
//
// Revision 1.2  1997/11/26 17:13:12  cinard
// Changed package name.
//
// Revision 1.1  1997/11/12 13:38:27  cinard
// Added classes to enable importing bytecode to populate the
// symbol table.
//
// ***********************************************************************
