// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/symboltable/j/JSymbolLoader.java,v $
// $Revision: 1.16 $
// $Date: 1998/06/15 11:52:56 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   symboltable.j
// NAME:        JSymbolLoader.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Oct 09,  1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.symboltable.j;

import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Vector;
import java.io.*;
import java.lang.reflect.*;

import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.jbcfile.*;

/**
 * Motivation : populate the symbol table when import, use or extend 
 * clauses are encountered. 
 *
 * Loads Java, ADL, and TDD symbols as they are imported (import, use, 
 * extends,implements clauses) in the parsed file. The loaded symbols 
 * are discovered via CLASSPATH and/or INCPATH, relevant checks are
 * performed by the SymbolLoader, and finally symbols are put in the
 * symbol table.
 *
 * The SymbolLoader collaborates with the SymbolTableMgr so that the
 * loaded symbols are put in the symbol table as part of the loading 
 * process. 
 */

public class JSymbolLoader 
{

   //-- Constructors --

   /**
    * Constructs the symbol loader to load symbols into the supplied
    * compilation unit scope (ADL or TDD)
    * @param stm The java symbol table manager to populate.
    */

   public JSymbolLoader(JSymbolTableMgr stm)
   {
      this.stm = stm;
      this.sbf = stm.getSymbolFactory();
      
      // Initialize class path (user incpath first, then classpath)

      String adlClassPath = System.getProperty("adl.incpath","")
                          + File.pathSeparator 
                          + System.getProperty("java.class.path","");

      classImporter = new ClassImporter(adlClassPath);
   }

   // -- Methods --

   /**
    * Import a package.
    * Since imported symbols are defined elsewhere they are constructed
    * and made available as part of the import mechanism.
    * @param name Package scoped name.
    * @return Symbol for the imported package 
    * @exception ORG.opengroup.adl.symboltable.j.PackageNotAccessibleException
    *            when package cannot be imported.
    */

   LPackage importPackage(String name)
   throws PackageNotAccessibleException
   {
      // Recursively import parent package
      if ( name.indexOf(".") > 0 )
      {
         LPackage unused = importPackage
                           (name.substring(0,name.lastIndexOf(".")));
      }
         
      // Check package is not yet imported
      LPackage pack = (LPackage) packages.get(name);
      if ( pack != null ) { return pack; }

      // Checks that the package is accessible from classpath
      if ( ! classImporter.canImportPackage(name) )
      {
          throw new PackageNotAccessibleException
                    ("package " + name + " not found.");
      }


      // Put package in the right scope
      if ( debug) { System.out.println("debug: JSymbolLoader.importPackage() -"
                                       + "name = <" + name + ">");
                  } 
      pack = sbf.newPackage(name,null);
      try
      {
         LPackage unused = (LPackage) stm.putSymbol(pack);
      }
      catch (SymbolAlreadyDefinedException e) { throw new Error(e.toString());}
      
      // Add package to imported packages
      packages.put(name,pack);

      return pack;
   }

   /**
    * Get a package.
    * @param name Package scoped name.
    * @return Symbol for the found package 
    */

   LPackage getPackage(String name)
   {
      try
      {
         return importPackage(name); 
      }
      catch ( PackageNotAccessibleException e ) { return null; }
   }

  /**
   * Check class is accessible.
   * @param scopedName Class scoped name.
   * @return true if class is accessible, false otherwise.
   */
  boolean canImportClass(String scopedName)
  {
     // Side effect is to create a proxy, which is the same
     // amount of work that just checking the class file exists,
     // and which will give better response time for future 
     // requests.

     return ( getClass(scopedName) != null );
  }
  
  /**
   * Get a class symbol.
   * <p>Since imported symbols are defined elsewhere they are constructed
   * and made available as part of the import mechanism.
   * <p> <b> Algorithm: </b> <ul>
   * <li>import class file 
   * <li>update scopedName (add package part it case it was not supplied)
   * <li>check if this class file has not been imported yet
   * <li>if it was not imported yet, import it and process classes referenced 
   *     in the imported bytecode :
   * <li>create proxy symbols for all classes related to the imported classes
   * <li>complete the imported class construction
   * <li>put imported class and constructed proxies in the symbol table
   * </ul>
   * @param scopedName Class scoped name.
   * @return Symbol for the imported class.
   *         (or transparently a proxy for this symbol)
   * @exception java.lang.io.FileNotFoundException
   *            when byte code file was not found.
   */

  LClass getClass(String scopedName)
  {
     LClass proxy = getProxy(scopedName);

     if ( proxy != null )
     {
        return proxy;
     }

     // Import class file.

     ClassFile classFile = classImporter.importClass(scopedName);

     if ( classFile != null )
     {
        proxy = getOrCreateProxy(scopedName,classFile);
        proxy.setModifier(new JModifier(classFile.getAccessFlags()));

        // Put the class in its right location
        // A named package or the default package (compilation unit scope)

        String packageName = (   scopedName.lastIndexOf(".") == -1
                               ? ""
                               : scopedName.substring
                                 (0,scopedName.lastIndexOf("."))
                             );

        LModule  module = null;
           
        if ( packageName.length() > 0 )
        {	
           try
           { 
              module = importPackage(packageName);
           }
           catch ( PackageNotAccessibleException e )
           {
              throw new Error(e.toString());
           }
        }
        else
        {
           module = (LModule)stm.getCompilationUnit();
        }
        
        proxy.setParent(module);
        module.putSymbol(proxy);
     }
     return proxy;
  }

  /**
   * Actually load class information for bytecode.
   * @param fullName fully qualified name of the class symbol to load.
   * @param classFile ClassFile for that class.
   * @return the loaded JClass.
   */

  JClass loadClass(String fullName,ClassFile classFile)
  {

     if ( debug )
        System.out.println
        ("debug: JSymbolLoader.loadClass(" + fullName + ",cf)");

     // Import class file if necessary.
     
     if ( classFile == null ) 
     {
        classFile = classImporter.importClass(fullName);
        if ( classFile == null )
        {
	   throw new JSymbolError
                     ("Could not find bytecode for class " 
                      + fullName 
                      + ". Check your CLASSPATH!");
        }
     }

     LClass lClass = sbf.newClass(fullName,0x0000);

     constructClassFromByteCode(lClass,classFile);

     if ( debug )
        System.out.println
        ("debug: JSymbolLoader.loadClass(" + fullName + ",cf)."
          + " loaded = " + ( lClass != null ? lClass.getFullName() : null )
          + " .Return");

     return (JClass)lClass;     
  }

  // -- Private methods --

  /**
   * Construct a class symbol from a class file object (directly 
   * constructed from bytecode).
   * <p> <b> Algorithm: </b> <ul>
   * <li>constructs modifiers
   * <li>constructs superclass
   * <li>constructs interfaces
   * <li>constructs fields
   * <li>constructs methods
   * <li>set loaded flag in proxy
   * </ul>
   * @param jClass class symbol to construct.
   * @param cf class file to use for the construction.
   */

  private void constructClassFromByteCode(LClass jClass,ClassFile cf)
  {
     // Get modifier

     jClass.setModifier(new JModifier(cf.getAccessFlags()));
     
     // Get super class
     
     String className = cf.getSuperClassName();
     if ( className != null )
     {
        jClass.addSuperClass(getOrCreateProxy(className));
     }

     // Get interfaces
     
     String[] interfaceNames = cf.getInterfaceNames();
     for (int i = 0; i < interfaceNames.length; i++)
     {
        jClass.addInterface(getOrCreateProxy(interfaceNames[i]));
     }
     
     // Get fields
     fieldInfo[] fi = cf.getFields();
     for (int i = 0; i < cf.getFieldsCount(); i++)
     {
        int modifier = fi[i].getAccessFlag();
        // Only add public fields
        if ( ( modifier & JModifier.PUBLIC ) != 0 )
        { 
           String fieldName   = fi[i].getName();
           BCType bcType      = fi[i].getType();
           LExpression att = sbf.newAttribute
                             (fieldName,modifier,bcTypeToJType(bcType));
           att.setRelated(jClass);
           jClass.putSymbol (att);
        }
     }

     // Get methods
     methodInfo[] mi = cf.getMethods();
     for (int i = 0; i < cf.getMethodsCount(); i++)
     {
        int modifier = mi[i].getAccessFlag();
        // Only add public methods
        if ( ( modifier & JModifier.PUBLIC ) != 0 )
        {
           String methodName = mi[i].getName();
           // In case it's a constructor, replace <init> witn class name.
           if ( methodName.equals("<init>") )
           {
              methodName = jClass.getName();
           }
           BCType  bcType = mi[i].getReturnType();
           LMethod jm     = sbf.newMethod(methodName,modifier);
           jm.setReturnType(bcTypeToJType(bcType));
           BCType[] bcTypes = mi[i].getParameters();
           for (int j = 0; j < bcTypes.length; j++)
           {
              jm.addParameter
              (sbf.newParameter("_",0x0000,bcTypeToJType(bcTypes[j])));
           }
           String[] exceptionNames = mi[i].getExceptions();
           for (int k = 0; k < exceptionNames.length; k++)
           {
              jm.addException
              (getOrCreateProxy(exceptionNames[k]));
           }

           jm.setParent(jClass);
           jClass.putSymbol(jm);
        }
     }
  }

  /**
   * Get an existing proxy.
   */

  private LClass getProxy(String name)
  {
     return (LClass)classes.get(name);
  }

  /**
   * Get an existing proxy or create a new one for a class.
   * @param name name of the class.
   * @param cf optional ClassFile for the proxy.
   */

  private LClass getOrCreateProxy(String name,ClassFile cf)
  {
     if ( debug )
        System.out.println
        ("   debug: JSymbolLoader.getOrCreateProxy(" + name + ")"); 
                             
     LClass proxy = (LClass)classes.get(name);
     if ( proxy == null )
     { 
        proxy = new JClassProxy(name,
                                this,
                                cf);
        classes.put(name,proxy);
        if ( debug )
            System.out.println
           ("   debug: JSymbolLoader.getProxy(" + name + "). Created proxy."); 
     }
     return proxy;     
  }

  /**
   * Get an existing proxy or create a new one for a class.
   * @param name name of the class.
   */
  private LClass getOrCreateProxy(String name)
  {
     return getOrCreateProxy(name,null);
  }

  /** 
   * Get a type symbol from the bytecode name of a Java type.
   * @param bcType type name found in bytecode.
   * @return type symbol.
   */

  private LType bcTypeToJType(BCType bcType)
  {
     LType  ret = null;

     if ( bcType.isBuiltin() )
     {
        // Builtin type
        ret = JBuiltinType.getBuiltinType(bcType.getName());
     }
     else
     {
        // It's a class
        ret = getOrCreateProxy(bcType.getName()); 
     }
     
     int arrayLevel = bcType.getArrayLevel();
     while ( arrayLevel > 0 )
     {
        ret = sbf.newArray(ret);
        arrayLevel--;
     }
     
     return ret;
  }

  /**
   * Use clause.
   * Make the named ADL file available for use.
   * @param name Class scoped name.
   * @return Symbol for the imported class.
   *         (or transparently a proxy for this symbol)
   */
  void useClause(String name)
  {
     // Checks that the adl file is accessible from incpath
  }
   
  // -- Private stuff --

  /** Symbol table manager. */

  private JSymbolTableMgr stm;

  /** Symbol table factory. */
  private SymbolFactory sbf;

  /** Class importer. */

  private ClassImporter classImporter;

  /** Imported classes. */

  private Hashtable classes  = new Hashtable();

  /** Imported packages. */

  private Hashtable packages = new Hashtable();

  /** Debug flag. */

  private static final boolean debug = false;

}

// ***********************************************************************
// History Log:
// $Log: JSymbolLoader.java,v $
// Revision 1.16  1998/06/15 11:52:56  cinard
// Fixed bug : symbol loader now throws a JSymbolError in case a java
// class file cannot be found.
//
// Revision 1.15  1998/03/18 16:05:27  cinard
// Force setting of class attributes and methods parent
// reference.
//
// Revision 1.14  1998/02/27 12:37:28  cinard
// Updated processing for the ClassFile objects.
// Updated copyright notice.
//
// Revision 1.13  1998/02/23 15:19:00  cinard
// Fixed comments.
//
// Revision 1.12  1998/02/23 14:29:23  cinard
// True Proxy design pattern implemented for JClassProxy.
// Improved type compatibility checks (complex type checking mechanism
// has been replaced by assignable() methods in subclasses of LType).
// JTypeKind.java now defines type kind constants that were formerly
// in JTypeChecker.java (update your client code!).
//
// Revision 1.11  1998/02/12 16:03:34  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.10  1998/01/20 12:56:42  cinard
// Modified bytecode loading : only public fields and methods of classes are
// loaded.
//
// Revision 1.9  1998/01/12 13:19:21  cinard
// Updated and fixed javadoc comments.
//
// Revision 1.8  1997/12/19 21:15:19  cinard
// Added support for exctracting exception (methods throw clauses) from
// the bytecode.
//
// Revision 1.7  1997/12/08 12:53:59  cinard
// Added type check mechanism.
// Fixed some byte code import bugs.
//
// Revision 1.6  1997/11/27 16:10:18  cinard
// Improve error reporting for class and package import mechanism.
//
// Revision 1.5  1997/11/27 09:04:08  cinard
// Debug flag off!!
//
// Revision 1.4  1997/11/27 08:28:59  cinard
// Fixed some bugs in Makefile.
//
// Revision 1.3  1997/11/26 17:13:12  cinard
// Changed package name.
//
// Revision 1.2  1997/11/19 15:42:14  cinard
// Improved import on demand and single type import behavior.
//
// Revision 1.1  1997/11/12 13:38:28  cinard
// Added classes to enable importing bytecode to populate the
// symbol table.
//
//
// ***********************************************************************
