// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/tools/FilterProcess.java,v $
// $Revision: 1.5 $
// $Date: 1998/04/15 16:24:15 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   tools
// NAME:        FilterProcess.java
//
// HISTORY:
//      Cyril Inard d'Argence, TOG - Mar 20,  1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.tools;

import java.io.*;

/**
 * Filter process. Enable to filter data through a command.
 * 
 * <P> There are two way supply data to the filter process :<ul>
 * <li>either by supplying a file or an input stream in the filter process
 * constructor (in this case, data read on the input stream are automatically
 * written on the filter process output stream and filtered),
 * <li>or, in case no file or input stream were supplied, by getting
 * the filter process output stream and directly write on it.
 * </ul>
 * In either cases, filtered data are available via the filter process 
 * input stream.
 */

public final class FilterProcess implements Runnable
{

  /**
    * Constructs the filter process to filter data.
    * <P> Note data shall be supplied through the FilterProcess
    * output stream.
    * @param filterCommand command for the filter process.
    */
 
 public FilterProcess(String filterCommand)
  throws IOException
  {
     _fp = Runtime.getRuntime().exec(filterCommand);
     _in = null;
     _processInput  = _fp.getInputStream();
     _processOutput = _fp.getOutputStream();
  }

  /**
    * Constructs the filter process to filter data.
    * <P> Note data shall be supplied through the FilterProcess
    * output stream.
    * @param filterCommand command for the filter process.
    * @param env environment variables for the command
    */
 
 public FilterProcess(String filterCommand, String[] env)
  throws IOException
  {
     _fp = Runtime.getRuntime().exec(filterCommand, env);
     _in = null;
     _processInput  = _fp.getInputStream();
     _processOutput = _fp.getOutputStream();
  }

  /** 
   * Constructs the filter process to filter a file.
   * @param filterCommand command for the filter process.
   * @param fileName name of the file that is to be filtered.
   */

  public FilterProcess(String filterCommand, String fileName)
  throws IOException
  {
     _fp = Runtime.getRuntime().exec(filterCommand);
     _in = new DataInputStream(new FileInputStream(fileName));
     _processInput  = _fp.getInputStream();
     _processOutput = _fp.getOutputStream();
     new Thread(this).start();
  }

  /** 
   * Constructs the filter process to filter an input stream.
   * @param in input stream that is to be filtered.
   */

  public FilterProcess(String filterCommand, InputStream in)
  throws IOException 
  {
     _fp = Runtime.getRuntime().exec(filterCommand);
     _in = in;
     _processInput = _fp.getInputStream();
     _processOutput = _fp.getOutputStream();
     new Thread(this).start();
  }

  /**
   * Get filter process output stream.
   */

  public OutputStream getOutputStream()
  {
    // Return output stream in case filter process has not yet been fed. 
    if ( _in != null )
    {
      throw new Error
	        ( "Filter process output stream not available: filter is fed"
                + " directly from the input stream supplied in the"
                + " constructor.");
    }
    return _processOutput;
  }

  /**
   * Get filter process input stream.
   */

  public InputStream getInputStream()
  {
    return _processInput;
  }

  /**
   * Do filter input in case filter is applied to an input stream
   * that was supplied within the constructor's parameters.
   */

  public void run() 
  {
     try
     {
        byte buffer[] = new byte[512];
        int byteRead  = _in.read(buffer);
              
        while ( byteRead != -1)
        {
           _processOutput.write(buffer, 0, byteRead);
           byteRead  = _in.read(buffer);
        }

	// Close the process output
        _processOutput.close();

	// Allow garbage collection and prevent re-use of the closed
	// stream.
	_processOutput = null;
     }
     catch ( IOException e )
     {
        System.err.println("Error while filtering : " + e.getMessage()); 
     }
  }


  public void close() throws IOException {

    _processOutput.close();
    _processOutput = null;
    _processInput.close();
    _fp.destroy();
  }

  
  // -- private --
  
  /** Filter process */
  private final Process      _fp;

  /** Input stream that will be piped into the filter process. */
  private final InputStream  _in;

  /** Input stream connected to the filter process. */
  private final InputStream  _processInput;

  /** Output stream connected to the filter process. */
  private OutputStream  _processOutput;

  // -- unit testings --

  /**
   * Unit tests.
   * Usage : command file, or command (std input assumed) or nothing
   * cat <std input assumed>.
   */
  
  public static void main(String[] args)
  throws IOException
  {
     FilterProcess fs;
     if ( args.length > 1 )
     {
        fs = new FilterProcess(args[0],args[1]);
     }
     else if ( args.length == 1 )
     {
        fs = new FilterProcess(args[0],System.in);
     }
     else
     {
        fs = new FilterProcess("cat",System.in);
     }

     byte buffer[] = new byte[512];
     InputStream is = fs.getInputStream();
     int byteRead  = is.read(buffer);
     while ( byteRead != -1)
     {
        System.out.write(buffer, 0, byteRead);
        byteRead  = is.read(buffer);
     }
  }

}

// ***********************************************************************
// History Log:
// $Log: FilterProcess.java,v $
// Revision 1.5  1998/04/15 16:24:15  bareau
// Restaured the FilterProcess(String) constructor, used by the C++ compiler.
//
// Revision 1.4  1998/04/15 14:00:31  bareau
// Added a close method for clean termination, and environment variables
// to launch the command.
//
// Revision 1.3  1998/03/31 10:19:54  wassel
// Added FilterProcessing implementing preprocessing files input streams.
//
// ***********************************************************************
