// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1997-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions to use, copy, modify and distribute this software are
// governed by the terms and conditions set forth in the file COPYRIGHT,
// located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/ADLCTransformVisitor.java,v $
// $Revision: 1.5 $
// $Date: 1998/07/28 16:08:41 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.c
// NAME:        ADLCTransformVisitor
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Jun 15, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.*;


/**
 * Transformation visitor.  Must translate an ADL AST into a C AST.
 * <P> For an overview of the algorithms used by this transformation,
 * see the ADL2.0 Translation System Design Specification, Chapter 5
 * (Code Generation).
 * <p>The translating starts by calling the <a href="#translate">translate()</a>
 * method passing the top of AST to translate as parameter.
 * @see CGeneric
 * @see ADLCParserVisitor
 */

public class ADLCTransformVisitor extends CGeneric {
  
  class BoolContext {

    private String tmp;
    private boolean b;

    BoolContext(String t, boolean bb) {
      tmp = t;
      b = bb;
    }

    String getString() {
      return tmp;
    }

    boolean getBool() {
      return b;
    }
  }

  /* handles */

  /** Stack of BoolContexts for lazy evaluation of logical operations */
  private Stack boolContextStack;

  /* Flag for debug mode */
  private final static boolean debug = false;

  /* handles */
  
  /** Handle for the global prolog block */
  private ASTCompoundStatement globalProlog;

  /** Handle for the global epilog block */
  private ASTCompoundStatement globalEpilog;

  /** Handle for the local prolog block */
  private ASTCompoundStatement localProlog;

  /** Handle for the local epilog block */
  private ASTCompoundStatement localEpilog;

  /** Handle for inline declarations (vector) */
  private Vector inlineDeclaration;

  /** Handle for the declaration of temporaries for
   * call-state expressions (vector) */
  private Vector preDeclaration;
  
  /** Handle for the declaration of temporaries for
    * call-state expressions (vector) */
  private Vector saveBlock;

  /** Handle for "pre-assertions" : statements that precede assertions
    * (local var declaration for instance). */
  private Vector preAssertion;

  /** Tags of the current assertion (vector) */
  private String[] assTags;

  /** Handle on the declaration of the inline that corresponds
    * to the currently encountered inline call */
  private ASTADL_InlineDeclaration currentInline;

  /** Handle on the parameters of the declaration of the
    * currently encountered inline call */
  private LExpression[] inlineParams;

  /** Handle on the actual parameters of the inline call that is
    * currently being inlined */
  private ASTArgumentList inlineActualParams;

  /** Auxiliary class for code generation */
  private ACOauxiliary ACO;

  /** Index for call-state tmp variables */
  private int savIndex;

  /** Index for preAssertions */ 
  private int preAssIndex;

  /** Flags set when visiting specific nodes. */
  private boolean
  isInAssertion, isInCallState, isInQuantified, isInBinding;

  /** Flag set when the currently visited expression can be decomposed
    * for the TET debug mode.
    * Must always be equal to 
    * <PRE>isInAssertion && 
    *   !(isInCallState || isInQuantified || isInBinding)</PRE>
    */
  private boolean isDecomposable;

  /** Flag set when visiting the substituted inline body
    * while transforming an inline call */
  private boolean inlining;

  private int nbv = 1;


  /**
    * Generates statement for a debug mode message
    * @param msg the string of the message to be generated.
    * @param id the temporary variable declared for preAssertion.
    * @return a statement node that is a postfix expression.
    * @see #preAssertion
    */
  private ASTStatement makeDebugModeMsg(String msg, String id, LType type) {
    
    // ADL_Printf(<msg>, ADL_Print_<type>(<id>));

    String printType = CTypeAux.getPrintType(type);

    ASTPostfixExpression printArg = new PostfixExprBuilder
      ( printType,
        new ArgumentListBuilder(makeName(id))
        .generatedAST())
      .generatedAST();

    ASTStatement createdNode = new StatementBuilder
      ( new PostfixExprBuilder
	( "ADL_printf",
	  new ArgumentListBuilder(makeString(msg))
          .addArgument(printArg)
          .generatedAST()))
      .generatedAST();

    return createdNode;
  }


  /**
    * Generates a declaration of a variable, either as 
    * a preDeclaration (call-state case) or as a preAssertion.
    * <BR>In the preAssertion case, the string parameter (that represents
    * the expression) is stored as a property of the ASTDeclaration.
    * @param saveblk "is a call-state".
    * @param type type of the temporary variable to generate.
    * @param theString token string of the rhs expression, for a preAssertion.
    * @return the name of the generated temporary variable.
    * @see #preDeclaration
    * @see #preAssertion
    */
  private String makeSaveDecl(boolean saveblk, LType type, String theString) {

    if (debug)
      System.out.println("\tmakeSaveDecl : " + saveblk + " " + type.getName() + " " + theString);

    String lhs = (saveblk ?
                  ("__ADL_savTmp_" + (savIndex++)) :
                  ("__ADL_tmp_" + (preAssIndex++)));

    // TBR: <type> <lhs> = <default value>;
    // <type> <lhs>;

    ASTDeclaration VarDecl = 
      new DeclarationBuilder
      ( type.getName(), //createdType, 
	new DirectDeclaratorBuilder(lhs).generatedAST()
	).generatedAST();

    if (saveblk) {
      preDeclaration.addElement(VarDecl);
    }
    else {
      try { VarDecl.setProperty("tokenString",theString); }
      catch(PropertyAccessException e) {}
      preAssertion.addElement(VarDecl);
    }

    return lhs;
  }


  /**
    * Generates a temporary declaration of a variable, initialized with 
    * a passed expression, as a saveBlock (call-state case).
    * 
    * @param type type of the tmp var to generate.
    * @param expr initialization expression.
    * @return the string name of the generated tmp variable.
    * @see #saveBlock
    * @see #preAssertion
    */
  private String makeSaveBlock(LType type, Node expr) {
    
    if (debug)
      System.out.println("\tmakeSaveBlock : " + type +
                         " " + expr);

    // <type> <lhs> = <expr>;

    String lhs = makeSaveDecl(true, type, "");
    ASTStatement stmt = new StatementBuilder
      ( new AssignmentBuilder(lhs, "=", expr))
      .generatedAST();

    saveBlock.addElement(stmt);
    return lhs;
  }


  /**
    * Generates a temporary declaration of a variable, initialized with 
    * a passed expression, as a preAssertion.
    * <BR>The string parameter (that represents the expression) 
    * is stored as a property of the ASTDeclaration.
    * @param type type of the tmp var to generate.
    * @param expr initialization expression.
    * @param theString token string of the rhs expression, for a preAssertion.
    * @return the name of the generated tmp variable.
    * @see #preAssertion
    */
  private String makePreAssertion(LType type, Node expr, String theString) {

    if (debug)
      System.out.println("\tmakePreAssertion : " +
			 (type == null ? "<no type>" : type.getName()) +
                         " " + expr + " " + theString);

    String lhs = "__ADL_tmp_" + (preAssIndex++);
    ASTDeclaration varDecl;

    // <lhs> = <expr>;

    ASTDeclarationSpecifiers varDeclType = makeType(type.getName());
    if (boolContextStack.empty()) {
      varDecl = new DeclarationBuilder
        ( varDeclType,
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder(lhs, expr))
          .generatedAST())
        .generatedAST();
    }
    else {
      // when there is 1 boolContext <"foo", bool> in the stack, the expression
      // is guarded by a conditional expression:
      // foo ? <expr> : <default value> // if bool is true ("true branch")
      // foo ? <default value> : <expr> // otherwise ("else branch")
      // the <default value> is the expression's type default value (0 for numeric, etc)

      // when several boolContext are stacked, conditional expressions
      // must be nested (from the bottom to the top of the stack)

      ASTParentheizedExpression parenth = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      ASTConditionalExpression cond = new
        ASTConditionalExpression(ADLCParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(parenth, cond, 0);
      for (Enumeration enum = boolContextStack.elements(); enum.hasMoreElements(); ) {
        BoolContext bc = (BoolContext)enum.nextElement();
        parentChild(cond, makeName(bc.getString()), 0);
        parentChild(cond,
                    // default value for un-evaluated branch
                    CTypeAux.getDefaultValue(type, false),
                    (bc.getBool() ? 2 : 1));
        if (enum.hasMoreElements()) {
          ASTParentheizedExpression parenth2 = new
            ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
          ASTConditionalExpression cond2 = new
            ASTConditionalExpression(ADLCParserTreeConstants.JJTCONDITIONALEXPRESSION);
          parentChild(parenth2, cond2, 0);
          parentChild(cond, parenth2, (bc.getBool() ? 1 : 2));
          cond = cond2;
        }
        else {
          parentChild(cond, expr, (bc.getBool() ? 1 : 2));
        }
      }
      varDecl = new DeclarationBuilder
        ( varDeclType,
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder(lhs, parenth))
          .generatedAST())
        .generatedAST();
    }

    try {
      varDecl.setProperty("tokenString", theString);
      varDecl.setProperty("exprType", type);
    }
    catch (PropertyAccessException e) {}
    
    preAssertion.addElement(varDecl);

    if (debug) {
      System.out.println("\tAdding the preAssertion:");
      ((SimpleNode)varDecl).dump(System.out, 100);
      System.out.println("\tmakePreAssertion returns: " + lhs);
    }
    return lhs;
  }

  
  /**
    * Visit pattern for a typed node.
    * <P><B>Expression decomposition algorithm</B>
    * <BR><I>(This algorithm is used for all expression nodes)</I>
    * <P><UL><LI>Visit the children (sub-expressions).
    * <LI><B>IF</B> the expression is to be decomposed (i.e. it is contained in
    * an assertion expression and it is neither inside a call-state expression,
    * a try/catch or a quantified assertion), call makeSaveAssignment to
    * generate a local variable declaration, i.e. to generate the declaration
    * of a local temporary variable that will be defined by the type and the
    * value of the expression.
    * <LI><B>ELSE</B> nothing to do: just return the visited node. </UL>
    * <P><I>For a more detailed overview of the expression declaration algorithm,
    * see the ADL2.0 Translation System Design Specification, Chapter 5
    * (Code Generation), paragraph 5.2.1.3.</I>
    * @param node the ASTConditionalExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  private Object decompose(Node node, Object data) {

    if (debug) {
      System.out.println("Decomposing node AST"+node+" ...");
      ((SimpleNode)node).dump(System.out, 100);
    }

    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, null), i);
    }
    if (isDecomposable) {
      return makeName
	(makePreAssertion(((TypedNode)node).getType(),
			  node,
			  ((SimpleNode) node).getExpressionImage()));
    }
    else {
      return node;
    }
  }


  /** 
   * Launch the visitor on each child of the current node
   * @param n the current node
   */
  private void childrenAccept(Node n) {

    if (debug) System.out.println("\tchildrenAccept: "+ n);

    for (int i = 0; i <  n.jjtGetNumChildren(); i++) {
      parentChild(n, (Node)n.jjtGetChild(i).jjtAccept(this, null), i);
    }
  }


  /**
   * Starts translating an AST.
   * Initializes needed handles.
   * @param ast the root node of the AST ADL to translate.
   * @exception ParseException thrown by parser.
   * @return the root node of the transformed tree
   * @see #visit(ORG.opengroup.adl.parsers.adlc.ASTTranslationUnit, java.lang.Object)
   */
  public Node translate(Node ast) {
  
    globalProlog = null;
    globalEpilog = null;
    localProlog = null;
    localEpilog = null;
    inlineDeclaration = new Vector();

    return (Node)ast.jjtAccept(this, null);
  }

  /** 
   * Constructor.
   * <BR>Creates an ACOauxiliary that will be associated with the 
   * currently transformed adl class.
   * @see ACOauxiliary
   */
  public ADLCTransformVisitor() {
    
    if (debug) System.out.println("ADLCTransformVisitor Constructor");
    ACO = new ACOauxiliary(this);
  }


  /**
   * Visit pattern for a node of type ASTADL_TranslationUnit
   *
   * <P><B>Algorithm:</B>
   * creates a node ASTTranslationUnit with children:<UL>
   * <LI> the external declaration children of the input node
   * <LI> the visit to the node's ASTADL_Class child
   * </UL>
   * Sets the name of the TranslationUnit to the moduleName.
   * @param node the ASTADL_TranslationUnit to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node (ASTTranslationUnit: the root of the
   * transformed AST).
   */
  public Object visit(ASTADL_TranslationUnit node, Object data) {

    if (debug) {
      System.out.println("\n\t************************************");
      System.out.println("\t*  TRANSLATING AST/ADL to AST/C  *");
      System.out.println("\t************************************\n");
      System.out.println("Visiting AST" + node + " ...");
    }

    // save former stm context
    Object savedStmContext = SimpleNode.stm.getContext();

    // set stm context for this ast
    try {
      Object stmContext = node.getProperty("stmContext");
      SimpleNode.stm.setContext(stmContext);
    }
    catch ( PropertyAccessException e ) { }

    Node child; ASTIncludeFileList child0;
    int insertChild = 0;
    ASTTranslationUnit createdNode = new 
      ASTTranslationUnit(ADLCParserTreeConstants.JJTTRANSLATIONUNIT);
    try { createdNode.setProperty("fileExtension", "c"); }
    catch(PropertyAccessException e) {}
    int nbChildren = node.jjtGetNumChildren();

    if (node.jjtGetChild(0) instanceof ASTIncludeFileList) {
      child0 = (ASTIncludeFileList)node.jjtGetChild(0);
    }
    else {
      child0 = new ASTIncludeFileList(ADLCParserTreeConstants.JJTINCLUDEFILELIST);
    }
    ASTIncludeFileDeclaration inc1 =
      new ASTIncludeFileDeclaration(ADLCParserTreeConstants.JJTINCLUDEFILEDECLARATION);
    inc1.setName("\"ADLTest.h\"");

    parentChild(child0, inc1, child0.jjtGetNumChildren());      
    parentChild(createdNode, child0, insertChild++);
    
    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);

      if (child instanceof ASTADL_AnnotatedDeclaration) {
	for (Enumeration e=((Vector)(child.jjtAccept(this,data))).elements();
	     e.hasMoreElements();) {
	  parentChild(createdNode, (Node)e.nextElement(), insertChild++);
	}
      }
    }
    createdNode.setName(ACO.getModuleName()+"ACO");

    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
    }

    // restore former stm context
    SimpleNode.stm.setContext(savedStmContext);

    return createdNode;
  }


  /**
   * Visit pattern for a node of type ASTADL_InlineDeclaration
   * <BR>Should not be visited!
   * <BR>For the processing of inline calls, see the algorithm
   * for primary expressions:
   * @see #visit(ORG.opengroup.parsers.adlc.ASTPrimaryExpression,java.lang.Object)
   * @param node the ASTADL_InlineDeclaration to visit.
   * @param data unused.
   * @return an error
   */
  public Object visit(ASTADL_InlineDeclaration node, Object data)  {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    // not visited: cf ASTADL_Class
    return new InternalError();
  }


  /**OK
    * Visit pattern for a node of type ASTADL_AnnotatedDeclaration
    * <P><B>Algorithm:</B><UL>
    * <LI>Stores information locally (possible inline, global prolog/epilog)
    * or into the associated ACOauxiliary (class name, possible superclass).
    * <LI> Visits the ASTADL_BehaviorDeclaration child.
    * <LI> Creates an ASTClassSpecifier node with 2 children:<OL>
    * <LI> the inherited class name in ASTBaseSpecifier node
    * <LI> and an ASTClassMemberList node parent of:<OL>
    * <LI> the class constructor
    * <LI> the declarations of tested functions.
    * </OL></UL>
    * <LI> Return a vector with the visited ASTADL_BehaviorDeclaration child
    * and the ASTClassSpecifier.
    * </OL></UL>
    * @param node the ASTADL_Class to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return a vector (ClassDeclaration).
    */
  public Object visit(ASTADL_AnnotatedDeclaration node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child, child0;
    Vector fctDeclaration = new Vector();

    int nbChildren = node.jjtGetNumChildren();

    int insertChild = 0;

    ACO.setModuleName(node.getName());

    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);

      if (child instanceof ASTADL_Prolog) {
        globalProlog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADL_Epilog) {
        globalEpilog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADL_InlineDeclaration) {
        inlineDeclaration.addElement(child);
      }
    }
    // now the loop for behavior declarations
    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_BehaviorDeclaration) {
	fctDeclaration.addElement((Node)child.jjtAccept(this, data));
      }
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + fctDeclaration);
    return fctDeclaration;
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorDeclaration
    * <P><B>Algorithm:</B><UL>
    * <LI>Creates a node of type ASTFunctionDefinition, when testing a function
    * that corresponds to the current behavior description.
    * <LI>Stores local information into the associated ACOauxiliary (function
    * name, return type, parameters)
    * or locally (local prolog/epilog).
    * <LI> Visits the ASTADL_BehaviorSpecification child.
    * </UL>
    * @param node the ASTADL_BehaviorDeclaration to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node (ASTFunctionDefinition) or (ASTCtorDefinition).
    */
  public Object visit(ASTADL_BehaviorDeclaration node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    // Initialize local prolog and epilog
    localProlog = null;
    localEpilog = null;

    ASTFunctionDefinition  createdNode = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);
    Node child = null;
    int nbChildren = node.jjtGetNumChildren();

    preDeclaration = new Vector();
    saveBlock = new Vector();

    ADLMethod currentFct = (ADLMethod)node.getScope();
    
    int insertChild = 0;
    ACO.initializeFct();
    // Store function return type
    ACO.setFctReturnType(currentFct.getReturnType().getName());
    ACO.setFctTypeDefaultValue(currentFct.getReturnType());
    // Store Function name in ACOauxiliary
    ACO.setFctName(currentFct.getName());
    // Store Function parameters
    ACO.setFctParams(currentFct.getParameters());      
    
    if (debug) System.out.println("\nTesting the function " + ACO.getFctName());

    child = node.jjtGetChild(0);
    if (child instanceof ASTDeclarationSpecifiers) {
      parentChild(createdNode, child, insertChild++);
      ACO.setFctDeclSpec((ASTDeclarationSpecifiers)child);
      child = node.jjtGetChild(1);
    }

    // function declarator
    ACO.setFctDeclTor((ASTFunctionDeclarator)child);
    parentChild(createdNode, child, insertChild++);

    // setFctName to "ADL_<moduleName>ACO_<fctName>"
    for (int i = 0; i < createdNode.jjtGetNumChildren(); i++) {
      child = createdNode.jjtGetChild(i);
      if (child instanceof ASTFunctionDeclarator) {
        do { child = child.jjtGetChild(child.jjtGetNumChildren()-1); }
        while (! (child instanceof ASTFunctionDirectDeclarator));
      }
      if (child instanceof ASTFunctionDirectDeclarator)
        ((SimpleNode)child.jjtGetChild(0)).setName("ADL_" + ACO.getModuleName() +
                                                   "ACO_" + ACO.getFctName());
    }

    // Visiting bhvSpec adds function body to createdNode
    int tmp = 0;
    for (int i=1; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_BehaviorSpecification) {
	tmp = i;
      }	
      else if (child instanceof ASTADL_Prolog) {
	localProlog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADL_Epilog) {
	localEpilog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
    } 
    parentChild(createdNode,
                (Node)node.jjtGetChild(tmp).jjtAccept(this, data),
                insertChild++);

    savIndex = 0;
    if (debug) {
      System.out.println("\nReturn Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }
  
  
  /**
    * Visit pattern for a node of type ASTADL_Prolog
    * <BR>Should not be visited!
    * @param node the ASTADL_Prolog to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return an error
    */
  public Object visit(ASTADL_Prolog node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");
    // not visited
    return new InternalError();
  }


  /**
    * Visit pattern for a node of type ASTADL_Epilog
    * <BR>Should not be visited!
    * @param node the ASTADL_Epilog to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return an error
    */
  public Object visit(ASTADL_Epilog node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");
    // not visited
    return new InternalError();
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorSpecification
    * <P><B>Algorithm: </B>generates all statements of the body of the
    * currently visited behavior description:<OL>
    * <LI>initialization statements: ACO.makeACO_Initialization(hasInheritedSemantics)
    * <LI>global prolog statements
    * <LI>local prolog statements
    * <LI>callstate temporary vairables declarations and initializations
    * (statements of the handle "saveBlock", created by the visit to the
    * ASTADL_AssertionGroup child)
    * <LI>call to the tested function: ACO.makeACO_CallBlock()
    * <LI>statements from the visit to the ASTADL_AssertionGroup child
    * <LI>local epilog statements
    * <LI>global epilog statements
    * <LI>finalization statements: ACO.makeACO_Final()
    * </OL>
    * @param node the ASTADL_BehaviorSpecification to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the created node (ASTCompoundStatement).
    * @see ACOauxiliary#makeACO_Initialization
    * @see ACOauxiliary#makeACO_CallBlock
    * @see ACOauxiliary#makeACO_Final
    */
  public Object visit(ASTADL_BehaviorSpecification node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child, newChild = null;
    ASTCompoundStatement createdNode =
      new ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    int nbChildren = node.jjtGetNumChildren();

    int insertChild = 0;
    child = node.jjtGetChild(0);
    if (child instanceof ASTADL_BehaviorClassification) {
      child.jjtAccept(this, data); // only side-effect: saving (ab)normalSpec
      child = node.jjtGetChild(1); // ADL_AssertionGroup
    }

    Vector ACO_Initialization = ACO.makeACO_Initialization();
    int nbc = ACO_Initialization.size();
    for (int i = 0; i < nbc; i++) {
      newChild = (Node)ACO_Initialization.elementAt(i);
      parentChild(createdNode, newChild, insertChild);
      insertChild += 1;
    }
    
    if (globalProlog != null) {
      nbc = globalProlog.jjtGetNumChildren();
      ASTCompoundStatement GPclone = null;
      try {
	GPclone = (ASTCompoundStatement)(globalProlog.clone());
      }
      catch(CloneNotSupportedException e) {}
      for (int i = 0; i < nbc; i++) {
        parentChild( createdNode, 
		     (Node)GPclone.jjtGetChild(i).jjtAccept(this, data),
		     insertChild);
        insertChild += 1;
      }
    }

    if (localProlog != null) {
      nbc = localProlog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localProlog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    // need to visit the child assertion group to know how many
    // preDeclaration are to be inserted

    ASTCompoundStatement visitedAssGrp =
      (ASTCompoundStatement)child.jjtAccept(this, data);
    Vector ACO_CallBlock = ACO.makeACO_CallBlock();

    nbc = preDeclaration.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)preDeclaration.elementAt(i), insertChild+i);
    }
    insertChild += nbc;


    ASTIfStatement ifStSkipAss = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    parentChild(createdNode, ifStSkipAss, insertChild++);
    ASTUnaryOperatorExpression notSkip = 
      new ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
    try { notSkip.setProperty("operator", "!"); }
    catch(PropertyAccessException e) {}

    parentChild(notSkip, makeName("skipAssertions"), 0);
    parentChild(ifStSkipAss, notSkip, 0);

    ASTCompoundStatement thenBlk =
      new ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    parentChild(ifStSkipAss, thenBlk, 1);

    nbc = saveBlock.size();
    if (debug) System.out.println("Size of saveBlock: " + nbc);
    for (int i = 0; i < nbc; i++) {
      parentChild(thenBlk, (Node)saveBlock.elementAt(i), i);
    }

    nbc = ACO_CallBlock.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_CallBlock.elementAt(i), insertChild+i);
    }
    insertChild += nbc;

    nbc = visitedAssGrp.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, visitedAssGrp.jjtGetChild(i), insertChild+i);
    }
    insertChild += nbc;

    if (localEpilog != null) {
      nbc = localEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localEpilog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    if (globalEpilog != null) {
      nbc = globalEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        try {
	  newChild = (Node)((SimpleNode)globalEpilog.jjtGetChild(i)).clone();
        }
        catch(CloneNotSupportedException e) {}
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    Vector ACO_Final = ACO.makeACO_Final();
    nbc = ACO_Final.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_Final.elementAt(i), insertChild); 
      insertChild += 1;
    }

    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorClassification
    * <BR>stores normal/abnormal specifications in the associated ACOauxiliary.
    * @param node the ASTADL_BehaviorClassification to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @returns null, that will not be used by the caller 
    */
  public Object visit(ASTADL_BehaviorClassification node, Object data) { 

    Node child;
    int nbc = node.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_NormalBehavior) {
        ACO.setNormalSpec(child.jjtGetChild(0));
      }
      else {
        ACO.setAbnormalSpec(child.jjtGetChild(0));
      }
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": null");
    return null;
  }

  /**
    * Visit pattern for a node of type ASTADL_AssertionGroup
    * <P><B>Algorithm: </B>
    * <UL><LI><B>IF</B> this is a nested assertion group: the assertion group
    * is considered as an expression.
    * <UL><LI><B>IF</B> it contains more than one assertion
    * <BR>Returns an expression: the result of the visit to this child assertion.
    * <LI><B>ELSE</B> Returns an expression: the "AND" of the results of the 
    * visits to these child assertions (if the input node is considered as a
    * decomposable expression, makePreAssertion is called and the visit
    * returns the name node of the variable that was initialized by the input
    * node value).</UL>
    * <LI><B>ELSE</B> the created node is an ASTCompoundStatement that contains
    * <I>either</I> the return node of the visit to the grand-child when 
    * the latter is an ASTADL_IfStatement <I>or</I> the
    * statements of the ASTCompoundStatement returned by the visit to an 
    * ASTADL_Assertion grand-child.
    * @param node the ASTADL_AssertionGroup to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #makePreAssertion(ORG.opengroup.adl.symboltable.LType,
    *                        ORG.opengroup.adl.parsers.adlc.Node,
    *                        java.lang.String)
    */
  public Object visit(ASTADL_AssertionGroup node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int nbChildren;
    String tmpStr = "", tmpName;
    if (isDecomposable) {
      tmpStr = node.getExpressionImage();
    }      
    
    if (isInAssertion || isInBinding) { // nested
      nbChildren = node.jjtGetNumChildren();
     // group of boolean expressions => conjunction
      if (nbChildren > 1) {
	ASTLogicalAndExpression createdNode = new 
	  ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);
	nbChildren = node.jjtGetNumChildren();
	for (int i = 0; i < nbChildren; i++) {
          parentChild( createdNode, 
                       (Node)node.jjtGetChild(i).jjtAccept(this, data),
                       i);
	}
	if (! (isInCallState || isInQuantified)) {
	  tmpName = makePreAssertion(node.getType(), createdNode, tmpStr);
	  if (debug)
	    System.out.println("Return Visiting AST" + node + ": Name");
	  return makeName(tmpName);
	}
	else {
	  if (debug)
	    System.out.println("Return Visiting AST" + node + ": " + createdNode);
	  return createdNode;
	}
      }
      else {
	return (Node)node.jjtGetChild(0).jjtAccept(this, data);
      }
    }
    else { // not nested : "real" assertion group
      ASTCompoundStatement createdNode =
        new ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
      Node visitedAss,child1, child2;
      int insertChild = 0;
      for (int i = 0; i < node.jjtGetNumChildren(); i++) {
        child1 = node.jjtGetChild(i);
        if (child1 instanceof ASTADL_Binding) {
	  for (Enumeration e = ((Vector)child1.jjtAccept(this, data)).elements(); 
             e.hasMoreElements(); ) {
	    parentChild(createdNode, (Node)e.nextElement(), insertChild++);
	  }
	}
        else { // child1 instanceof ASTADL_Statement
          child1 = child1.jjtGetChild(0);
          if (child1 instanceof ASTADL_IfStatement) {
            parentChild(createdNode, (Node)child1.jjtAccept(this, data), insertChild++);
          }
          else { // child1 : ADL_Assertion  
            visitedAss = (Node)child1.jjtAccept(this, data);
            for (int j = 0; j < visitedAss.jjtGetNumChildren(); j++) {
              parentChild(createdNode, visitedAss.jjtGetChild(j), insertChild++);
            }
          }
        }
      }
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      }
      return createdNode;
    }
  }

  
  /**
    * Visit pattern for a node of type ASTADL_Binding
    * <BR>generates a vector of variable declaration that
    * corresponds to the binding.
    * @param node the ASTADL_Binding to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the vector of transformed node.
    */
  public Object visit(ASTADL_Binding node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Vector translatedBinding = new Vector();
    Node createdNode = null;
    isInBinding = true;
    isDecomposable = false;
    // first child is ASTDeclaration or ASTADL_NamedParamList
    Node decls = node.jjtGetChild(0); 
    // second child is an expression
    Node expr = node.jjtGetChild(1);

    if (decls instanceof ASTDeclaration) {
      // named parameter = declSpec + decltor
      ASTDeclaration params = (ASTDeclaration)decls;
      // declaration == decl specifiers + Init declarator
      // Binding initdeclarator == (decltor/direct decltor)+expression 
      createdNode = new DeclarationBuilder
	( (ASTDeclarationSpecifiers)params.jjtGetChild(0),
	  new InitDeclaratorBuilder
	  ( params.jjtGetChild(1),
            (Node)expr.jjtAccept(this, data)).generatedAST())
        .generatedAST();
    }
    else { // child(0) is ASTADL_NamedParamList
      for (int i = 0; i < decls.jjtGetNumChildren(); i++) {
	translatedBinding.addElement(decls.jjtGetChild(i));
      }
      LExpression init = null;
      try { init = (LExpression)node.getProperty("bindingInit"); }
      catch(PropertyAccessException e) {}
      if (init == null) {
	createdNode = new StatementBuilder
	  ((Node)expr.jjtAccept(this, data), true).generatedAST();
      }
      else {
	createdNode = new StatementBuilder
	  ( new AssignmentBuilder
	    ( init.getName(),
	      "=",
	      (Node)expr.jjtAccept(this, data)))
          .generatedAST();
      }
    }
    isInBinding = false;
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    translatedBinding.addElement(createdNode);
    return translatedBinding;
  }

  /**
    * Visit pattern for a node of type ASTADL_Statement.
    * <BR>No special processing: just retransmits to the child node.
    * @param node the ASTADL_Statement to visit.
    * @param data unused.
    * @return the transformed node.
    */
  public Object visit(ASTADL_Statement node, Object data) { 

    if (debug) System.out.println("ASTADL_Statement");

    return (Node)node.jjtGetChild(0).jjtAccept(this, data);
  }


  /**
    * Visit pattern for a node of type ASTADL_IfStatement
    * <P><B>Algorithm: </B><UL><LI><B>IF</B> inside an assertion,
    * it is considered as an expression and will generate an "?:"
    * conditional expression by visiting the child nodes. If there
    * is no else child node, the else branch of the generated conditional
    * expression will be the boolean constant true.
    * <LI><B>ELSE</B> generates a C if statement by visiting the
    * child nodes (condition, then node, else node if present).
    * </UL>
    * @param node the ASTADL_IfStatement to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node (ASTConditionalExpression / ASTIfStatement).
    */
  public Object visit(ASTADL_IfStatement node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child; 

    if (isInAssertion) { // nested
      ASTParentheizedExpression parenth = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      ASTConditionalExpression condNode = new
        ASTConditionalExpression(ADLCParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(parenth, condNode, 0);

      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        parentChild(condNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
      }
      if (nbc == 2) {
        child = makeConstant("true");
        parentChild(condNode, child, 2);
      }

      if (isDecomposable) {
        return makeName(makePreAssertion(node.getType(),
						 parenth,
						 node.getExpressionImage()));
      }
      else {
        return parenth;
      }
    }
    else { // not nested
      ASTIfStatement createdNode = new
        ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
      
      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
// 	if (i == 2) { // case of "else"
// 	  ASTStatement elseNode = new
// 	    ASTStatement(ADLCParserTreeConstants.JJTSTATEMENT);
//           try { elseNode.setProperty("endsWithSemiColon", Boolean.FALSE); }
//           catch(PropertyAccessException e) {}

// 	  parentChild(elseNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), 0);
// 	  parentChild(createdNode, elseNode, i);
// 	}
// 	else { 
          parentChild(createdNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
// 	}
      }
      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    }
  }


  /**
   * Visit pattern for a node of type ASTADL_Assertion
   * <P><B>Algorithm</B><UL>
   * <LI><B>IF</B> nested: the assertion node is in fact considered as a
   * boolean expression: just visit the child.
   * <LI>Generates the code that corresponds to an assertion;
   * i.e. for assertion "get_active_accounts()== @get_active_accounts()+1" :
   <PRE>
   |  testContext->assertionStart("get_active_accounts() == @get_active_accounts() + 1");
   |  if (!skipAssertions) {
   |    int __ADL_tmp_1 = bankObject->get_active_accounts();
   |    int __ADL_tmp_2 = __ADL_savTmp_1;
   |    int __ADL_tmp_3 = __ADL_tmp_2 + 1;
   |
   |    tddRes = (__ADL_tmp_1 == __ADL_tmp_3);
   |
   |    if (debugMode) {
   |      // report on each sub-expression
   |      testContext->infoline(adlstr.reset() << "get_active_accounts()\t-> " << __ADL_tmp_1);
   |      testContext->infoline(adlstr.reset() << "@get_active_accounts()\t-> " << __ADL_tmp_2);
   |      testContext->infoline(adlstr.reset() << "@get_active_accounts() + 1\t-> " << __ADL_tmp_3);
   |      testContext->infoline(adlstr.reset() << "get_active_accounts() == @get_active_accounts() + 1\t-> " << tddRes);
   |    }
   |  testContext->assertionResult((tddRes ? ADL_PASS : ADL_FAIL));
   |  }
   |  else {
   |     testContext->assertionResult(ADL_UNEVALUATED);
   |  }
   </PRE>
   * The variable declarations __ADL_tmp_i are "preAssertion"
   * obtained by the expression decomposition algorithm.
   * </UL>
   * @see #visit(ORG.opengroup.adl.parsers.adlc.ASTConditionalExpression,java.lang.Object)
   * @param node the ASTADL_Assertion to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the transformed node (ASTCompoundStatement).
   */
  public Object visit(ASTADL_Assertion node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node visitedAss = null, child = node.jjtGetChild(0);
    int nbChildren = node.jjtGetNumChildren(), insertChild = 0;
    String tmpStr = "", assLabel = null;
    boolean labelledAss = false, taggedAss = false, quantif = false;
    
    if (debug) {
      System.out.println("\t\tisInAssertion = "+isInAssertion);
      System.out.println("\t\tisInBinding   = "+isInBinding);
    }
    if (isInAssertion || isInBinding) {
      for(int i = 0; i < nbChildren; i++) {
	child = node.jjtGetChild(i);
	if (child instanceof ASTADL_Labels) {
          // $$$$ WARNING labels not used in nested assertions
          return null;
	}
	else if (child instanceof ASTADL_Tags) {
          // $$$$ WARNING tags not used in nested assertions
          return null;
	}
        else if (child instanceof ASTADL_QuantifiedAssertion) {
          // $$$$ ERROR no quantified assertion in nested assertion groups
          return null;
        }
        else { // expression of the assertion : to visit
          return (Node)child.jjtAccept(this, data);
        }
      }
    }
    preAssertion = new Vector();
    boolContextStack = new Stack();

    ASTCompoundStatement createdNode = new 
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    
    for(int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_Labels) {
        labelledAss = true;
        assLabel = (String)child.jjtAccept(this, data);
        if (debug) System.out.println("LABEL : " + assLabel);
      }
      else if (child instanceof ASTADL_Tags) {
        taggedAss = true;
	assTags = (String[])child.jjtAccept(this, data);
      }
      else if (child instanceof ASTADL_QuantifiedAssertion) {
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInQuantified || isInBinding);
        visitedAss = (Node)child.jjtAccept(this, data);
        isInAssertion = false;
	isDecomposable = false;
        quantif = true;
	tmpStr = ((SimpleNode) child).getExpressionImage();
      }
      else { // expression of the assertion
        // visit the expression: decomposition algorithm
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInQuantified || isInBinding);
        visitedAss = (Node)child.jjtAccept(this, data);
        isInAssertion = false; isDecomposable = false;
	tmpStr = ((SimpleNode) child).getExpressionImage();
      }
    }

    int nbp = preAssertion.size();

    if (labelledAss) {
      tmpStr = "[ " + assLabel + " ] " + tmpStr;
    }
    
    // ADL_assertionStart(<tmpStr>)
    
    parentChild(createdNode, 
                new StatementBuilder
                ( new PostfixExprBuilder
                  ( "ADL_assertionStart",
                    new ArgumentListBuilder(makeString(tmpStr))
                    .generatedAST()))
                .generatedAST(),
                0);

    // if tagged assertion:
    // ADL_infoline("Tagged assertion (Untestable)";)

    if (taggedAss) {
      parentChild(createdNode, 
                  new StatementBuilder
                  ( new PostfixExprBuilder
                    ( "ADL_infoline",
                      new ArgumentListBuilder(makeString("Tagged assertion (Untestable)"))
                      .generatedAST()))
                  .generatedAST(),
                  1);
    }
    else {
      // if (!skipAssertions)
      
      ASTIfStatement ifStSkipAss = new
        ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
      
      ASTUnaryOperatorExpression condition = new
        ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
      try { condition.setProperty("operator", "!"); }
      catch(PropertyAccessException e) {}
      ASTName skipAss = makeName("skipAssertions");
      parentChild(condition, skipAss, 0);
      
      parentChild(ifStSkipAss, condition, 0);
      
      ASTCompoundStatement thenBlk = new
        ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
      parentChild(ifStSkipAss, thenBlk, 1);
      

      if (debug)
	System.out.println("\t\tIs quantified assertion ? -> "+quantif);
      if (quantif) { // quantified assertion:
        // store the statements returned by the visit of the quantified assertion
        // (this visit returns an ASTCompoundStatement)
        for (int i = 0; i < visitedAss.jjtGetNumChildren(); i++) {
          parentChild(thenBlk, (Node)visitedAss.jjtGetChild(i), i);
        }
        insertChild = visitedAss.jjtGetNumChildren();
      }
      else {
	for (int i = 0; i < nbp; i++) {
	  if (debug) {
	    System.out.println("\nDumping preAssertion "+i); 
	    ((SimpleNode)(Node)preAssertion.elementAt(i)).dump(System.out, 100);
	  }
	  parentChild(thenBlk, (Node)preAssertion.elementAt(i), i);
	}
	insertChild = nbp;
	
	// tddRes = last tmp
	parentChild(thenBlk,
		    new StatementBuilder
		    ( new AssignmentBuilder("tddRes", "=", visitedAss))
                    .generatedAST(),
		    insertChild);
	insertChild += 1;
      }

      // if (debugMode) {
      if (!quantif) {
        // No debug mode for quantified assertions
	ASTIfStatement ifDebug = new
	  ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
	parentChild(ifDebug, makeName("ADL_debugMode"), 0);
	parentChild(thenBlk, ifDebug, insertChild);
	insertChild += 1;
	
	ASTCompoundStatement thenBlk2 = new
          ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
	parentChild(ifDebug, thenBlk2, 1);
	
	// infoline of sub-expressions
	
	Node tmpName; ASTDeclaration vardecl;
        ASTStatement assTmpValue;
	String declName; LType declType = null;
	int insertChild2 = 0, nbChild = 0;
	
	for (int i = 0; i < nbp; i++) {
	  vardecl = (ASTDeclaration)preAssertion.elementAt(i);
          String msg;
          try {
            msg = vardecl.getProperty("tokenString") + "\\t-> %s";
            declType = (LType)vardecl.getProperty("exprType");
          }
          catch (PropertyAccessException e) {
            msg = "unknown string";
          }

          declName = ((SimpleNode)vardecl.jjtGetChild(1).jjtGetChild(0)
                       .jjtGetChild(0).jjtGetChild(0).jjtGetChild(0)).getName();

          if ((declType instanceof LReference) || (declType instanceof LConstructedType)) {
            assTmpValue = makeDebugModeMsg(msg, "&" + declName, declType);
          }
          else {
            assTmpValue = makeDebugModeMsg(msg, declName, declType);
          }

          parentChild(thenBlk2, assTmpValue, insertChild2);
          insertChild2 += 1;
	}
      }

      // ADL_assertionResult((tddRes ? ADL_PASS : ADL_FAIL));

      ASTConditionalExpression passFail = new
        ASTConditionalExpression(ADLCParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(passFail, makeName("tddRes"), 0);
      parentChild(passFail, makeName("ADL_PASS"), 1);
      parentChild(passFail, makeName("ADL_FAIL"), 2);
      parentChild(thenBlk,
		  new StatementBuilder
		  ( new PostfixExprBuilder
		    ( "ADL_assertionResult",
                      new ArgumentListBuilder(passFail).generatedAST()))
                  .generatedAST(),
                  insertChild++);
      
      // ADL_assertionResult((tddRes ? ADL_PASS : ADL_FAIL));
      // else ADL_assertionResult(ADL_UNEVALUATED);
      ASTStatement elseBlk = new StatementBuilder
        ( new PostfixExprBuilder
          ( "ADL_assertionResult",
            new ArgumentListBuilder(makeName("ADL_UNEVALUATED"))
            .generatedAST()))
        .generatedAST();
      parentChild(ifStSkipAss, elseBlk, 2);
      
      parentChild(createdNode, ifStSkipAss, 1);
    }
    
    preAssertion = null;
    preAssIndex = 0;
    
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      //      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }
  

  /**
   * Visit pattern for a node of type ASTADL_Labels
   * @param node the ASTADL_Labels to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the name of the label (String)
   */
  public Object visit(ASTADL_Labels node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": " + node.getName());
    }
    // not visited
    return node.getName();
  }


  /**
   * Visit pattern for a node of type ASTADL_Tags
   * @param node the ASTADL_Tags to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return an error
   */
  public Object visit(ASTADL_Tags node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    String[] assTags;
    try {
      assTags = (String[])node.getProperty("taglist");
    }
    catch(PropertyAccessException e) {
      assTags = new String[1];
      assTags[0] = "<no tag>";
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + assTags);
    return assTags;
  }


  /**
   * Visit pattern for a node of type ASTADL_QuantifiedAssertion.
   * <P>Generates the code that corresponds to a quantified assertion.
   * for instance for assertion
   * <BR>"forall (long i : ADL_long_range(1,10)) { Stock_balance(i) >= 0; }" :
<PRE>
|  testContext->assertionStart("forall(long i : ADL_long_range(1,10)){Stock_balance(i)>=0;}");
|  if (!skipAssertions) {
|     tddRes = true;
|     for (long i = 1; tddRes && (i <= 10); i++)
|       tddRes = StockBrokerObject->Stock_balance(i) >= 0;
|     testContext->assertionResult(( tddRes ? ADL_PASS : ADL_FAIL ));
|  }
|  else {
|     testContext->assertionResult(ADL_UNEVALUATED);
|  }
</PRE>
   * <P><B>Notes</B><UL>
   * <LI>a quantified assertion is considered as non-decomposable
   * <LI>in case of "forall", the code generated for tddRes is "tddRes = true;"
   * <LI>in case of "exists", the code generated for tddRes is "tddRes = false;
   * <LI>if there are several enumeration domains, the generated code is an
   * imbrication of "for" loops.  </UL>
   * @param node the ASTADL_QuantifiedAssertion to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the created node (ASTCompoundStatement).
   */
  public Object visit(ASTADL_QuantifiedAssertion node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTCompoundStatement createdNode = new ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);

    boolean isForall = ((ASTADL_Quantifier)node.jjtGetChild(0)).getName().equals("forall");
    int insertChild = 0;

    // tddRes = <isForall>;
    if (isForall) {
      parentChild(createdNode,
                  new StatementBuilder
                  ( new AssignmentBuilder
                    ("tddRes", "=", makeBoolean("true")))
                  .generatedAST(),
                  insertChild++);
    }
    else {
      parentChild(createdNode,
                  new StatementBuilder
                  ( new AssignmentBuilder
                    ("tddRes", "=", makeBoolean("false")))
                  .generatedAST(),
                  insertChild++);
    }

    int nbLoops = node.jjtGetChild(1).jjtGetNumChildren();
    ASTADL_NamedParam[] domDecl = new ASTADL_NamedParam[nbLoops];
    Node[] domExpr = new Node[nbLoops];

    for (int i = 0; i < nbLoops; i++) {
      domDecl[i] = (ASTADL_NamedParam)node.jjtGetChild(1).jjtGetChild(i).jjtGetChild(0);
      domExpr[i] = node.jjtGetChild(1).jjtGetChild(i).jjtGetChild(1);
    }

    String iterator = "";
    Node infBorn = null, supBorn = null;
    ASTForStatement forNode = new
      ASTForStatement(ADLCParserTreeConstants.JJTFORSTATEMENT);

    Node last;

    ASTDeclaration forDecl;
    ASTLogicalAndExpression andExpr;
    ASTPostfixExpression incExpr;

    parentChild(createdNode, forNode, insertChild++);
    for (int i = 0; i < nbLoops; i++) {
      try {
	iterator = ((LExpression)domDecl[i].getProperty("expression")).getName();
      }
      catch (PropertyAccessException e) {}
	
      try {
	infBorn = domExpr[i].jjtGetChild(1).jjtGetChild(0);
	supBorn = domExpr[i].jjtGetChild(1).jjtGetChild(1);
      }
      catch (NullPointerException e) {
	// second child of ADLDomain must be PostfixExpression having 2 children:
	// Name and ArgumentList having 2 expression children.
	System.err.println("Error : Invalid adl domain in quantified assertion");
      }
      // long i
      forDecl = new DeclarationBuilder
	( domDecl[i].jjtGetChild(0), // Declaration specifiers
	  new InitDeclaratorBuilder
	  ( domDecl[i].jjtGetChild(1), // Declarator or Direct declarator
	    infBorn).generatedAST()
	  ).generatedAST();
      // tddRes && i<supBorn
      andExpr = new
        ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);
      if (isForall) {
        parentChild(andExpr, makeName("tddRes"), 0);
      }
      else { // exists
	ASTUnaryOperatorExpression condition = new
	  ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
	try { condition.setProperty("operator", "!"); }
        catch(PropertyAccessException e) {}
	parentChild(condition, makeName("tddRes"), 0);
	parentChild(andExpr, condition, 0);
      }
      ASTRelationalExpression relExpr = new
        ASTRelationalExpression(ADLCParserTreeConstants.JJTRELATIONALEXPRESSION);
      try { relExpr.setProperty("operator", "<="); }
      catch(PropertyAccessException e) {}
      parentChild(relExpr, makeName(iterator), 0);
      parentChild(relExpr, supBorn, 1);
      ASTParentheizedExpression parenth = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      parentChild(parenth, relExpr, 0);

      parentChild(andExpr, parenth, 1);

      // i++
      ASTPostDeIncrement increment = new
        ASTPostDeIncrement(ADLCParserTreeConstants.JJTPOSTDEINCREMENT);
      try { increment.setProperty("operator", "++"); }
      catch(PropertyAccessException e) {}
      incExpr = new 
	PostfixExprBuilder(iterator,increment).generatedAST();

      parentChild(forNode, forDecl, 0);
      parentChild(forNode, andExpr, 1);
      parentChild(forNode, incExpr, 2);
//       String forKind[] = {"Declaration","Expression", "Expression"}; 
//       forNode.setKind(forKind);

      last = forNode;

      if (i == nbLoops-1) {
        // visit the assertion-expression
        isInQuantified = true; isDecomposable = false;
        Node assGrp = (Node)node.jjtGetChild(2).jjtAccept(this, data);    
        isInQuantified = false;
        isDecomposable = isInAssertion && ! (isInCallState || isInBinding);

        // tddRes = assGrp
        parentChild(last,
                    new StatementBuilder
                    ( new AssignmentBuilder
                      ( "tddRes", "=", assGrp))
                    .generatedAST(),
                    3);
      }
      else {
        forNode = new
          ASTForStatement(ADLCParserTreeConstants.JJTFORSTATEMENT);
        parentChild(last, forNode, 3);
      }
    }
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }


  /**
   * Visit pattern for a node of type ASTADL_ImplExpression
   * <P><B>Algorithm: </B><UL><LI>visits the 2 operand expressions (returning
   * ret_exp1, ret_exp2).
   * <LI><B>CASE</B> operator :<UL>
   * <LI><B>==?</B>return !(ret_exp1) || (ret_exp2)
   * <LI><B>?==</B>return  (ret_exp2) || !(ret_exp1)
   * <LI><B>?=?</B>return ret_exp1 == ret_exp2
   * <LI><B>?:?</B>return (ret_exp1) ? "abnormal" : !(ret_exp2 && "abnormal")
   * </UL></UL>
   * @param node the ASTADL_ImplExpression to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the created node.
   */
  public Object visit(ASTADL_ImplExpression node, Object data) { 
    if (debug) {
      System.out.println("\nVisiting AST" + node + " ...");
    }

    Node createdNode;
    Node ch1 = node.jjtGetChild(0),
      ch2 = node.jjtGetChild(2);
    String op = null;
    try { op = (String)((SimpleNode)node.jjtGetChild(1)).getProperty("op"); }
    catch(PropertyAccessException e) {}
    String tmpStr = "", tmpName;

    if (isDecomposable) {
      tmpStr = node.getExpressionImage();
    }      
     
    if (op.equals("==>") || 
        op.equals("<==")) {
      if (op.equals("<==")) {  // exchange (exp1) and (exp2)
        Node tmp = ch1;
        ch1 = ch2;
        ch2 = ch1;
      }

      // e1 ==> e2 changed into !(e1) || (e2)
      createdNode = new
        ASTLogicalOrExpression(ADLCParserTreeConstants.JJTLOGICALOREXPRESSION);

      parentChild(createdNode, 
		  new NotBuilder(ch1,
				 !(ch1 instanceof ASTParentheizedExpression)
				 ).generatedAST(),
		  0);
      parentChild(createdNode, ch2, 1);
      createdNode = (Node)createdNode.jjtAccept(this, data);
    }

    else if (op.equals("<=>")) {
      // e1 <=> e2 changed into (e1) == (e2)
      String[] opseq = { "==" };
      createdNode = new
        ASTEqualityExpression(ADLCParserTreeConstants.JJTEQUALITYEXPRESSION);
      try { ((SimpleNode)createdNode).setProperty("operators", opseq); }
      catch(PropertyAccessException e) {}

      parentChild(createdNode, ch1, 0);
      parentChild(createdNode, ch2, 1);
    }

    else {// e1 <:> e2 changed into (!(e1) || "abnormal") && (!"abnormal" || !(e2) || (e1))
      createdNode = new
        ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);

      ASTParentheizedExpression subnode1 = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      ASTLogicalOrExpression snode1 = new
        ASTLogicalOrExpression(ADLCParserTreeConstants.JJTLOGICALOREXPRESSION);
      parentChild(createdNode, subnode1, 0);
      parentChild(subnode1, snode1, 0);
      ch1 = (Node)ch1.jjtAccept(this, data);
      parentChild(snode1,
		  new NotBuilder(ch1,
				 !(ch1 instanceof ASTParentheizedExpression)
				 ).generatedAST(),
		  0);
      parentChild(snode1, makeName("abnormal"), 1);

      ASTParentheizedExpression child2 = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      parentChild(createdNode, child2, 1);

      ASTLogicalOrExpression subnode2 = new
        ASTLogicalOrExpression(ADLCParserTreeConstants.JJTLOGICALOREXPRESSION);
      parentChild(child2, subnode2, 0);
      Node p1bis = null;
      try {
        p1bis = (Node)((SimpleNode)ch1).clone();
      } catch(CloneNotSupportedException e) {}

      // lazy evaluation: 2nd child must be evaluated only when abnormal is true
      boolContextStack.push(new BoolContext("abnormal", true));
      ch2 = (Node)ch2.jjtAccept(this, data);
      boolContextStack.pop();

      parentChild(subnode2,
                  new NotBuilder(makeName("abnormal"), false).generatedAST(),
                  0);
      parentChild(subnode2, new NotBuilder(ch2, false).generatedAST(), 1);
      parentChild(subnode2, p1bis, 2);
    }

    if (isDecomposable) {
      if (debug)
	System.out.println("Return Visiting AST" + node + ": Name");
      return makeName(makePreAssertion(node.getType(),
                                         createdNode, tmpStr));
    }
    else {
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }
      return createdNode;
    }    
  }


  /**
    * Visit pattern for a node of type ASTConditionalExpression: a call to decompose(node,data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConditionalExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConditionalExpression node, Object data) { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    // lazy evaluation:
    // b ? e1 : e2
    // --> t1 = visit(b);
    // t2 = t1 ? visit(e1) : <default value>;
    // t3 = t1 ? <default value> : visit(e2);
    // t4 = t1 ? t2 : t3

    parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
    // ATTENTION: we suppose that the boolean expression is ALWAYS decomposed,
    int ind = preAssIndex-1;
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, true));
    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    boolContextStack.pop();
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, false));
    parentChild(node, (Node)node.jjtGetChild(2).jjtAccept(this, data), 2);
    boolContextStack.pop();

    if (isDecomposable) {
      return makeName(makePreAssertion(node.getType(),
					       node,
					       node.getExpressionImage()));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTLogicalAndExpression
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTLogicalAndExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTLogicalAndExpression node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...: visit number "+nbv++);
    }
    int nbChildren = node.jjtGetNumChildren();

    if (nbChildren > 2) {
      // to simplify lazy evaluation mechanism, we use associativity:
      // e1 || e2 || e3... is first changed in e1 || (e2 || e3...)
      ASTLogicalAndExpression newAnd = new
        ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);
      newAnd.copySourceTokens(node);
      parentChild(newAnd, node.jjtGetChild(0), 0);

      ASTLogicalAndExpression newSubAnd = new
        ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);

      // copy source tokens and skip first child tokens
      newSubAnd.copySourceTokens(node);
      newSubAnd.firstToken = ((SimpleNode)node.jjtGetChild(1)).firstToken;

      parentChild(newAnd, newSubAnd, 1);

      for (int i = 1; i < nbChildren; i++) {
        parentChild(newSubAnd, node.jjtGetChild(i), i-1);
      }

      return newAnd.jjtAccept(this, data);
    }
    else {
      // lazy evaluation:
      // e1 || e2
      // --> t1 = visit(e1);
      // t2 = t1 ? false : visit(e2)
      // t3 = t1 || t2
      parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
      boolContextStack.push
        ( new BoolContext("__ADL_tmp_" + (preAssIndex-1), true));
      
      parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
      boolContextStack.pop();

      if (isDecomposable) {
        return makeName(makePreAssertion(node.getType(),
						 node,
						 node.getExpressionImage()));
      }
      else {
        return node;
      }
    }
  }

  /**
    * Visit pattern for a node of type ASTLogicalOrExpression
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTLogicalOrExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTLogicalOrExpression node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...: visit number "+nbv++);
    }
    int nbChildren = node.jjtGetNumChildren();

    if (nbChildren > 2) {
      // to simplify lazy evaluation mechanism, we use associativity:
      // e1 || e2 || e3... is first changed in e1 || (e2 || e3...)
      ASTLogicalOrExpression newOr = new
        ASTLogicalOrExpression(ADLCParserTreeConstants.JJTLOGICALOREXPRESSION);
      newOr.copySourceTokens(node);
      parentChild(newOr, node.jjtGetChild(0), 0);

      ASTLogicalOrExpression newSubOr = new
        ASTLogicalOrExpression(ADLCParserTreeConstants.JJTLOGICALOREXPRESSION);

      // copy source tokens and skip first child tokens
      newSubOr.copySourceTokens(node);
      newSubOr.firstToken = ((SimpleNode)node.jjtGetChild(1)).firstToken;

      parentChild(newOr, newSubOr, 1);

      for (int i = 1; i < nbChildren; i++) {
        parentChild(newSubOr, node.jjtGetChild(i), i-1);
      }

      return newOr.jjtAccept(this, data);
    }
    else {
      // lazy evaluation:
      // e1 || e2
      // --> t1 = visit(e1);
      // t2 = t1 ? false : visit(e2)
      // t3 = t1 || t2
      parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
      boolContextStack.push
        ( new BoolContext("__ADL_tmp_" + (preAssIndex-1), true));
      
      parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
      boolContextStack.pop();

      if (isDecomposable) {
        return makeName(makePreAssertion(node.getType(),
						 node,
						 node.getExpressionImage()));
      }
      else {
        return node;
      }
    }
  }

  /**
    * Visit pattern for a node of type ASTExpression
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTExpression node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...: visit number "+nbv++);
    }
    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTInclusiveOrExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTInclusiveOrExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTInclusiveOrExpression node, Object data) { 

    if (debug) System.out.println("ASTInclusiveOrExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTExclusiveOrExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTExclusiveOrExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTExclusiveOrExpression node, Object data) { 

    if (debug) System.out.println("ASTExclusiveOrExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTAndExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTAndExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTAndExpression node, Object data) { 

    if (debug) System.out.println("ASTAndExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTEqualityExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTEqualityExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTEqualityExpression node, Object data) { 

    if (debug) System.out.println("ASTEqualityExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTRelationalExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTRelationalExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTRelationalExpression node, Object data) { 

    if (debug) System.out.println("ASTRelationalExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTShiftExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTShiftExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTShiftExpression node, Object data) { 

    if (debug) System.out.println("ASTShiftExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTAdditiveExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTAdditiveExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTAdditiveExpression node, Object data) { 

    if (debug) System.out.println("ASTAdditiveExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTMultiplicativeExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTMultiplicativeExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTMultiplicativeExpression node, Object data) { 

    if (debug) System.out.println("ASTMultiplicativeExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTUnaryOperatorExpression
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTUnaryOperatorExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #visit(ORG.opengroup.parsers.adlc.ASTConditionalExpression)
    */
  public Object visit(ASTUnaryOperatorExpression node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

      return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTCastExpression.
    * <BR>Just visit the expression child (nothing to do for
    * the type child).
    * @param node the ASTCastExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  public Object visit(ASTCastExpression node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int lastnode = node.jjtGetNumChildren()-1;
    parentChild(node, (Node)node.jjtGetChild(lastnode).jjtAccept(this, data), lastnode);

    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + node);
    return node;
  }


  /**
    * Visit pattern for a node of type ASTADL_CallStateExpression
    *
    * <P><B>Algorithm: </B><OL>
    * <LI>call makeSaveBlock to create a "saveBlock", i.e. a declaration
    * of a temporary variable that will be initialized with the type
    * and value of this node child expression before the call to the tested
    * function.
    * <LI>call "makePreAssertion" to create a "preAssertion", i.e. a
    * declaration of a temporary variable that will be initialized with
    * the type and value of the precedently created variable.
    * <BR>This second variable is used as a sub-expression of the
    * decomposition algorithm.
    * </OL>
    * <PRE>
    <type of @expr> __ADL_savTmp_i = <@expr>;
   // ...
   // <call to tested function>
   // ...
   // assertion:
   <type of @expr> __ADL_tmp_j = __ADL_savTmp_i;
      </PRE>
    * @param node the ASTADL_CallStateExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #makeSaveBlock
    * @see #makePreAssertion(ORG.opengroup.adl.symboltable.LType,
    *                        ORG.opengroup.adl.parsers.adlc.Node,
    *                        java.lang.String)
    */
  public Object visit(ASTADL_CallStateExpression node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    //First child is UnaryExpression
    Node child = node.jjtGetChild(0); 
    if (isInCallState) {
      return child.jjtAccept(this, data);
    }
    ASTName createdNode;
    String exprType = ((TypedNode)child).getType().getName();

    isInCallState = true;
    isDecomposable = false;
    createdNode = makeName
      (makeSaveBlock(((TypedNode)child).getType(), (Node)child.jjtAccept(this, null)) );
    isInCallState = false;
    isDecomposable = isInAssertion && ! (isInQuantified || isInBinding);
    if (isDecomposable) {
      return makeName( makePreAssertion(((TypedNode)child).getType(),
						createdNode,
						node.getExpressionImage()));
    }
    else {
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }
      return createdNode;
    }
  }

  /**
   * Visit pattern for a node of type ASTADL_Unchanged.
   * <P><B>Algorithm:</B><UL>
   * <LI>clone the node children (argument list of the "unchanged" expression).
    * <LI>copy the source tokens (that represents the 
    * code of the expression) on the cloned arguments.
   * <LI><B>IF</B> there is <I>one</I> argument "arg", create a node "(@arg ==
   * clone_arg)" and visit its 2 children nodes.
   * <LI><B>ELSE</B> (there are <I>n</I> arguments
   * "arg<SUB>1</SUB>"..."arg<SUB>n</SUB>"): create a node
   * "((@arg<SUB>1</SUB> == clone_arg<SUB>1</SUB>) && ... &&
   * (@arg<SUB>n</SUB> == clone_arg<SUB>n</SUB>))" and visit all
   * "@arg<SUB>i</SUB>" and "clone_arg<SUB>i</SUB>" child nodes. </UL>
   * @param node the ASTADL_Unchanged to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the created node.
   */
  public Object visit(ASTADL_Unchanged node, Object data) {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    String tokenStr = null;

    Node argList = node.jjtGetChild(0);
    int nbexpr = argList.jjtGetNumChildren();

    if (nbexpr == 0) {
      // unchanged() ---> true
      ASTConstant createdNode = makeBoolean("true");
      createdNode.copySourceTokens(node);

      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    }

    Node argCopy = null;

    try {
      argCopy = (Node)((SimpleNode)argList).clone();
    } catch (CloneNotSupportedException e) {}

    if (nbexpr == 1) {
      // unchanged(expr) ---> (expr == @expr)
      
      Node expr1 = argList.jjtGetChild(0);
      Node expr2 = argCopy.jjtGetChild(0);

      ASTParentheizedExpression createdNode = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      ASTEqualityExpression eq = new
        ASTEqualityExpression(ADLCParserTreeConstants.JJTEQUALITYEXPRESSION);
      String[] opseq = { "==" };
      try { eq.setProperty("operators", opseq); }
      catch(PropertyAccessException e) {}
      eq.copySourceTokens(node);
      eq.setType(CBuiltinType.intT);

      ASTADL_CallStateExpression callstatexpr = new
        ASTADL_CallStateExpression(ADLCParserTreeConstants.JJTADL_CALLSTATEEXPRESSION);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens((SimpleNode) expr1);
      callstatexpr.prependToken(new Token("@"));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(createdNode, (Node)eq.jjtAccept(this, data), 0);
      
      if (debug) {
	System.out.println("\nReturn Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }

      return createdNode;
    }

    // unchanged(expr1, ..., exprn) ---> ((expr1 == @expr1) && ... && (exprn == @exprn))
    ASTParentheizedExpression createdNode = new
      ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
    ASTLogicalAndExpression andNode = new
      ASTLogicalAndExpression(ADLCParserTreeConstants.JJTLOGICALANDEXPRESSION);
    andNode.copySourceTokens(node);
    andNode.setType(CBuiltinType.intT);

    SimpleNode expr1, expr2; ASTParentheizedExpression parenth;
    ASTEqualityExpression eq; ASTADL_CallStateExpression callstatexpr;
    String[] opseq = { "==" };

    for (int i = 0; i < nbexpr; i++) {
      expr1 = (SimpleNode)argList.jjtGetChild(i);
      expr2 = (SimpleNode)argCopy.jjtGetChild(i);

      parenth = new
        ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      eq = new
        ASTEqualityExpression(ADLCParserTreeConstants.JJTEQUALITYEXPRESSION);
      try { eq.setProperty("operators", opseq); }
      catch(PropertyAccessException e) {}
      eq.setType(CBuiltinType.intT);

      callstatexpr = new
        ASTADL_CallStateExpression(ADLCParserTreeConstants.JJTADL_CALLSTATEEXPRESSION);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens(expr1);
      callstatexpr.prependToken(new Token("@"));
      eq.copySourceTokens(callstatexpr);
      eq.prependToken(new Token(expr1.getExpressionImage() + " == "));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(parenth, eq, 0);
      parentChild(andNode, parenth, i);
    }

    andNode.copySourceTokens(node);
    parentChild(createdNode, (Node)andNode.jjtAccept(this, data), 0);

    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
}

  /**
   * Visit pattern for a node of type ASTADL_Normal
   * @param node the ASTADL_Normal to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the ASTName node "normal".
   */
  public Object visit(ASTADL_Normal node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": Name");
    }
    ASTName n = makeName("normal");
    n.setType(CBuiltinType.intT);
    n.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(n.getType(),
					       n,
					       n.getExpressionImage()));
    }
    else {
      return n;
    }
  }


  /**
   * Visit pattern for a node of type ASTADL_Abnormal
   * @param node the ASTADL_Abnormal to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the ASTName node "normal".
   */
  public Object visit(ASTADL_Abnormal node, Object data) { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": Name");
    }
    ASTName abn = makeName("abnormal");
    abn.setType(CBuiltinType.intT);
    abn.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(abn.getType(),
					       abn,
					       abn.getExpressionImage()));
    }
    else {
      return abn;
    }
  }

  /**
    * Visit pattern for a node of type ASTADL_Return 
    * @param node the ASTADL_Return to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the ASTName node "__ADL_return".
    */
  public Object visit(ASTADL_Return node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    if (inlining && isInCallState) {
//       throw new ParseException("\"return\" in call-state expression via an inline");
    }
    ASTName tmpExpr = makeName("__ADL_return");
    tmpExpr.setType(node.getType());
    tmpExpr.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(node.getType(),
					       tmpExpr,
					       tmpExpr.getExpressionImage()));
    }
    else {
      return tmpExpr;
    }
  }


  /**
    * Visit pattern for a node of type ASTString.
    *
    * @param node the ASTString to visit.
    * @param data unused.
    * @return the node.
    */
  public Object visit(ASTString node, Object data) { 

    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + node);

    return node;
  }

  
  /**
   * Visit pattern for a node of type ASTParentheizedExpression
   * @param node the ASTParentheizedExpression to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the node.
   */
  public Object visit(ASTParentheizedExpression node, Object data) { 
    childrenAccept(node);
    return node;
  }
  
  
  /**
   * Visit pattern for a node of type ASTArgumentList
   * @param node the ASTArgumentList to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the node.
   */
  public Object visit(ASTArgumentList node, Object data) { 
    childrenAccept(node);
    return node;
  }
  
  
  /**
   * Visit pattern for a node of type ASTPostfixExpression
   * @param node the ASTPostfixExpression to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
//   public Object visit(ASTPostfixExpression node, Object data) { 
//     if (debug) {
//       System.out.println("Visiting AST" + node + " ...");
//     }
   
//     int insertChild = 0;
//     Node child0 = node.jjtGetChild(0), child;

//     ASTPostfixExpression createdNode = new
//       ASTPostfixExpression(ADLCParserTreeConstants.JJTPOSTFIXEXPRESSION);

//     // Case of child0 != ASTSimpleTypeSpecifier
//     if (!((child0 instanceof ASTBuiltinTypeSpec) || (child0 instanceof ASTQualifiedType))) {
//       Node tmp = (Node)child0.jjtAccept(this,data);
//       if (debug) {
// 	System.out.println("\tTmp node: " + tmp);
//       }
//       if (tmp instanceof ASTPostfixExpression) {
// 	for( int i = 0; i < tmp.jjtGetNumChildren(); i++) {
// 	  parentChild(createdNode, tmp.jjtGetChild(i), insertChild++);
// 	}
//       }
//       else {
// 	parentChild(createdNode, tmp, insertChild++);
//       }
//     }

//     for (int i = 1; i < node.jjtGetNumChildren(); i++) {
//       child = node.jjtGetChild(i);
//       if (child instanceof ASTPostDeIncrement) 
// 	parentChild(createdNode, child, insertChild++);
//       else
// 	parentChild(createdNode, (Node)child.jjtAccept(this, data), insertChild++);
//     }
//     if (debug) {
//       System.out.println("Return Visiting AST" + node + ": " + createdNode);
//       ((SimpleNode)createdNode).dump(System.out, 100);
//     }

//     if (isDecomposable) {
//       return makeName(makePreAssertion(node.getType(), createdNode, node.getExpressionImage()));
//     }
//     else {
//       return createdNode;
//     }
// //    return createdNode;
// }


  /**
   * Visit pattern for a node of type ASTRefAccessSuffix
   * @param node the ASTRefAccessSuffix to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTRefAccessSuffix node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTRefAccessSuffix createdNode = new
      ASTRefAccessSuffix(ADLCParserTreeConstants.JJTREFACCESSSUFFIX);
    parentChild(createdNode, node.jjtGetChild(0), 0);
    if (node.jjtGetNumChildren() > 1) {
      ASTArgumentList arguments = (ASTArgumentList)node.jjtGetChild(1);
      parentChild(createdNode, arguments, 1);
      for (int i = 0; i<arguments.jjtGetNumChildren(); i++)
	parentChild(arguments, (Node)arguments.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return createdNode;
}


  /**
   * Visit pattern for a node of type ASTDotAccessSuffix
   * @param node the ASTDotAccessSuffix to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTDotAccessSuffix node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTDotAccessSuffix createdNode = new DotAccessSuffixBuilder
      (node.jjtGetChild(0)).generatedAST();
    if (node.jjtGetNumChildren() > 1) {
      ASTArgumentList arguments = (ASTArgumentList)node.jjtGetChild(1);
      parentChild(createdNode, arguments, 1);
      for (int i = 0; i<arguments.jjtGetNumChildren(); i++)
	parentChild(arguments, (Node)arguments.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return createdNode;
}


  /**
   * Visit pattern for a node of type ASTPostfixExpression
   * the node is created when it has 2 chidren: Name + ArgumentList
   *
   * <P><B>Algorithm</B>
   * <UL>
   *   <LI><B>IF</B> the first child is an inline name:
   *     <BR>replace the first child by a clone of the body of the
   *     corresponding inline and re-visit the node with flag
   *     "inlining" true.
   *   <LI><B>ELSE</B><I>(not a call to an inline macro)</I>
   *     <UL>
   *       <LI><B>IF</B> "inlining"
   *       <LI><B>ELSE</B>
   *         <BR>if this is a qualified name then it is not modified,
   *         otherwise it refers to the "real object" and must be
   *         qualified: "<name>" is replaced by
   *         "<adlclass name>Object-><name>"
   *     </UL>
   * </UL>
   * @param node the ASTPrimaryExpression to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTPostfixExpression node, Object data) { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int nbChildren = node.jjtGetNumChildren();
    SimpleNode child1 = (SimpleNode)node.jjtGetChild(0);
    String tmpStr = "";

    if (node.jjtGetChild(0) instanceof ASTName) {
      ASTArgumentList child2 = (ASTArgumentList)node.jjtGetChild(1);
      boolean isInline;
      String child1Name = child1.getName();
      Node createdNode = null;

      try { isInline = ((String)child1.getProperty("astNameKind")).equals("ADLInline"); }
      catch(PropertyAccessException e) { isInline = false; }

      if (isInline) {
        inlineParams = null;
        for (Enumeration e = inlineDeclaration.elements(); 
             e.hasMoreElements(); ) {
          currentInline = (ASTADL_InlineDeclaration)e.nextElement();
          ADLInline currentADLInline = (ADLInline)currentInline.getScope();
          if ((currentADLInline.getName()).equals(child1Name)) {
            // Inline having at list 1 parameter
            if (currentADLInline.getParameters().length > 0) {
              inlineParams = currentADLInline.getParameters();
              if (debug) System.out.println("\t1Inline Params is AST"+inlineParams);
              break;
            }// Inline without parameters --> empty argument list
            else
              inlineParams = new LExpression[0];
          }
        }
        if (inlineParams == null) {
          if (debug) System.out.println("inline not found!!!");
        }
        inlineActualParams = child2;
        ASTADL_AssertionGroup inlineBody = null;
        try {
          inlineBody = (ASTADL_AssertionGroup)
            ((SimpleNode)currentInline.jjtGetChild(2)).clone();
        } catch (CloneNotSupportedException e) {}

	// reach the statement in the inline body assertion group
	SimpleNode statement = (SimpleNode)inlineBody
	  .jjtGetChild(0) // ADL_Statement
	  .jjtGetChild(0); // ADL_IfStatement or ADL_Expression
	if (statement instanceof ASTADL_Assertion) {
	  // ADL_Assertion : reach the ADL_Expression node
	  statement = (SimpleNode)statement
	    .jjtGetChild(statement.jjtGetNumChildren() - 1);
	} else {
	  // ADL_IfStatement : we have the right node in hand
	}

	// the next statement has no side-effect on the Token chain of
	// the original inline body node since it has first been cloned.
	statement.prependToken(new Token("[inline call: " +
				       node.getExpressionImage() + "] "));

        if (node.jjtGetNumChildren() == 2) {
          inlining = true;
          createdNode = (Node)inlineBody.jjtAccept(this, data);
          if (debug)
            System.out.println("Return Visiting AST" + node + ": " + createdNode);
          inlining = false;
          return createdNode;
        }
        else {
          inlining = true;
          PostfixExprBuilder primBuild = new
            PostfixExprBuilder((Node)inlineBody.jjtAccept(this, data));
          for (int i = 2; i < node.jjtGetNumChildren(); i++) {
            primBuild.addSuffix((Node)node.jjtGetChild(i).jjtAccept(this, data));
          }
          inlining = false;

          ASTPostfixExpression prim = primBuild.generatedAST();
          prim.setType(node.getType());
	  prim.copySourceTokens(node);
          return prim;// visitedPrim;
        }
      }
      else { // child1 instanceof ASTName && ! isInline
        if (inlining) {
          String firstName = child1Name;
          int param = -1;
          for (int i = inlineParams.length-1; i >= 0; i--) {
            if (firstName.equals(inlineParams[i].getName())) {
              param = i;
              break;
            }
          }
          if (param != -1) { // formal parameter
            Node actual = null;
            try { // case of many expressions
              if (inlineActualParams.jjtGetChild(0) instanceof ASTExpression)
                actual = (Node)
                  ((SimpleNode)inlineActualParams.jjtGetChild(0).jjtGetChild(param))
                  .clone();
              else // case of one expression
                actual = (Node)
                  ((SimpleNode)inlineActualParams.jjtGetChild(param))
                  .clone();
            }
            catch (CloneNotSupportedException e) {}

            //  inlineActualParam <Refsuffix: aName> <aSuffix>

            ASTParentheizedExpression parenth = new
              ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
            parentChild(parenth, actual, 0);
            createdNode = new PostfixExprBuilder
              ( parenth,
                new RefAccessSuffixBuilder
                ( child1Name,
                  child2)
                .generatedAST())
              .generatedAST();
            ((ASTPostfixExpression)createdNode).setType(node.getType());

	    ((SimpleNode) createdNode).copySourceTokens(node);

            for (int i = 1; i < node.jjtGetNumChildren(); i++) {
              parentChild(createdNode, node.jjtGetChild(i), i);
            }

            return createdNode.jjtAccept(this, data);
          }
        }

      }
    }
    else { // !(child1 instanceof ASTName)
      parentChild(node, (Node)child1.jjtAccept(this, data), 0);
    }

    for (int i = 1 ; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
    }

    if (isDecomposable) {
      if (debug) {
	System.out.println("\tIn node "+node+":");
	System.out.println("\t\tnode type = "+node.getType());
      }
      return makeName(makePreAssertion(node.getType(),
					       node,
					       node.getExpressionImage()));
    }
    else
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + node);
	((SimpleNode)node).dump(System.out, 100);
      }
      return node;
  }
  
  
  /**
    * Visit pattern for a node of type ASTAssignmentExpression
    * @param node the ASTAssignmentExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  public Object visit(ASTAssignmentExpression node, Object data) {
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return node;
  }


  /**
    * Visit pattern for a node of type ASTConstant
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConstant to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConstant node, Object data) { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTConstantExpression
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConstantExpression to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConstantExpression node, Object data) { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    return decompose(node, data);
  }


  //TBR
  /**
   * Visit pattern for a node of type ASTName
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> "inlining" and the name is the name of a parameter of
    * the inline that is currently being inlined, then create a node that is
    * a clone of the corresponding actual parameter, visit it and <B>return</B>
    * it.
    * <LI>If the expression is decomposable, call makePreAssertion, else
    * return the node.
    * </UL>
    * @param node the ASTName to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the created node.
   */
  public Object visit(ASTName node, Object data) { 

    if (debug) System.out.println("Visiting ASTName: " + node.getName());

    if (inlining && isInCallState) {
      String astNameKind = null;
      try {
        astNameKind = (String)node.getProperty("astNameKind");
      }
      catch(PropertyAccessException e) { }
      if (astNameKind.equals("ADLBinding")) {
        System.err.println(node.getLineNumber() + " Binding in a call-state via an inline");
      }
    }

    String tmpName = node.getName();
    if (inlining) {
      if (debug) System.out.println("\tNode AST" + inlineParams + " has " +
                                    inlineParams.length +" children.");
      int param = -1;
      for (int i = inlineParams.length-1; i >= 0; i--) {
        if (tmpName.equals(inlineParams[i].getName())) {
          param = i;
          break;
        }	  
      }

      if (debug) {
	System.out.println("\tParameter number: "+param);
	System.out.println("\tinlineActualParams: "+inlineActualParams);
	((SimpleNode)inlineActualParams).dump(System.out, 100);
      }
      if (param != -1) {
        Node createdNode = null;
	try { // case of many expressions
	  if (inlineActualParams.jjtGetChild(0) instanceof ASTExpression)
	    createdNode = (Node)
	      ((SimpleNode)inlineActualParams.jjtGetChild(0).jjtGetChild(param))
	      .clone();
	  else // case of one expression
	    createdNode = (Node)
	      ((SimpleNode)inlineActualParams.jjtGetChild(param))
	      .clone();
	}
	catch (CloneNotSupportedException e) {}
	if (debug) 
	  System.out.println("Return Visiting AST" + node + ": "+createdNode);
        return createdNode;
      }
    }
    
    if (isDecomposable) {
      if (debug) {
	System.out.println("\tisDecomposable: " + isDecomposable);
	System.out.println("\tNode type: " + node.getType());
      }
      Node tmpExpr = makeName
	( makePreAssertion(node.getType(), node, tmpName));
      if (debug) 
	System.out.println("Return Visiting AST" + node + ": "+ tmpExpr);
      return tmpExpr;
    }
    if (debug) 
      System.out.println("Return Visiting AST" + node + ": "+ node);
    return node;
  }

}

// ***********************************************************************
// History Log:
// $Log: ADLCTransformVisitor.java,v $
// Revision 1.5  1998/07/28 16:08:41  mallet
// Replaced makeIdExpression calls with makeName.
//
// Revision 1.4  1998/07/02 11:58:07  mallet
// Added visit of ADL_Statement to remove it from resulting AST.
// Fixed processing of PrimaryExpression in case of an inline call: was
// incorrectly skipping the ADL_AssertionGroup and its ADL_Statement
// children to reach the final statement in the inline body.
// Fixed visit of IdExpression when replacing an inline actual parameter:
// was incorrectly getting the nth actual parameter when there was more
// than one parameter.
//
// Revision 1.3  1998/06/30 16:05:41  mallet
// Added one parameter to call of CTypeAux.getDefaultValue: always
// false since we call this method for "lazy" members of a logical and|or.
//
// Revision 1.2  1998/06/30 08:44:48  mallet
// Turned debug mode off for delivery.
// Fixed some typos in AST node names.
// Replaced access to expressionImage property by call to
// getExpressionImage().
// Replaced setting of expressionImage property on newly created nodes by
// use of copySourceTokens() and other token chain related methods.
// Fixed bugs in processing of ASTADL_Unchanged: was only visiting the
// children of the equivalent equality expression instead of the equality
// itself.
// Fixed bug in visit of ASTPrimaryExpression for inline calls: was
// incorrectly getting the inline body expression.
// Changed debug mode reporting for inline calls: now also reporting the
// initial inline call expression.
//
// Revision 1.1  1998/06/23 13:39:43  bareau
// First version of the ADL/C translator (ADL part only).
//
// Revision 1.1  1998/06/23 13:27:10  bareau
// First version of the ADL/C translator (ADL part only).
//
//
// ***********************************************************************



