// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CBuilder.java,v $
// $Revision: 1.3 $
// $Date: 1998/10/14 08:12:48 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.c
// NAME:        CBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Jun 24, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.c.CModifier;
import java.util.Vector;
import java.util.Hashtable;

/**
  * Abstract class, ancestor of the builder classes for all TDD
  * declarations (datasets, factories, directives).
  * <P>Builds code for the class <I>declaration</I>: name of the class,
  * inherited classes, declaration of constructor and functions.
  * @author Cyrille Bareau
  * @see CDatasetBuilder
  * @see CDirectiveBuilder
  * @see CFactoryBuilder
  */

public abstract class CBuilder {

  /** Handle on the root of the generated C AST */
  public ASTTranslationUnit compilUnit;

  /** Name of the feature declaration in the source tdd class */
  protected String featureName;

  /** Name of the source tdd class */
  protected String unitName; 

  /** Name of the generated C class */
  protected String className;

  /** Index of anonymous directives */
  protected int anonymousIndex = 1;

  /** Handle on the generated declaration AST */
  protected ASTDeclaration declaration;

  private Hashtable includes;

  private boolean isDataset = false;
  private boolean isDirective = false;
  private boolean relinquish = false;

  static private boolean debug = false;


  public String getClassName() {
    return className;
  }

  public String getFeatureName() {
    return featureName;
  }

  /**
   * Builds the qualified name of the file that implements the feature
   * (dataset or factory) described by the parameters.
   * @param prefix "F_" for a factory, "S_" for a dataset.
   * @param TDDClassName the name of the TDD class defining the feature
   * @param name the feature name, qualified if imported.
   * @return the built file name
   */
  public static String makeQualifiedName(String prefix,
                                         String TDDClassName,
                                         String featureName) {

    return prefix + TDDClassName + "_" + featureName;
  }

  /**
    * Auxiliary constructor. Just initializates relinquish (last parameter)
    * before calling the main one.
    */
  public CBuilder (SimpleNode node, String prefix, String unit,
		     boolean b, ASTIncludeFileList incList) {
    this(node, prefix, unit, incList);
    relinquish = b;
  }


  /**
    * Main constructor.
    * @param node the current feature declaration node
    * @param unit the name of the source tdd class
    */
  public CBuilder (SimpleNode node, String prefix, String unit,
		     ASTIncludeFileList incList) {
    super();

    if (debug)
      System.out.println("CBuilder constructor " + node + " " + prefix + " " + unit);

    unitName = unit;
    int nbChildren = 0;

    compilUnit = new
      ASTTranslationUnit(ADLCParserTreeConstants.JJTTRANSLATIONUNIT);

    includes = new Hashtable();
    ASTIncludeFileList includeList;
    try {
      includeList = (ASTIncludeFileList)incList.clone();
    } catch (CloneNotSupportedException e) {
      includeList = incList;
    }
    ADLCGenericVisitor.parentChild(compilUnit, includeList, 0);
    for (int i = 0; i < incList.jjtGetNumChildren(); i++) {
      ASTIncludeFileDeclaration incl =
	(ASTIncludeFileDeclaration)incList.jjtGetChild(i);
      String incFileName = incl.getName();
      if (!includes.contains(incFileName))
	includes.put(incFileName, incFileName);
    }
    
    if (node instanceof ASTTDD_DatasetDeclaration) {
      if ((node.jjtGetChild(0).jjtGetNumChildren() < 2) ||
          (node.jjtGetChild(0).jjtGetChild(1) instanceof ASTInitDeclaratorList) ||
          (node.jjtGetChild(0).jjtGetChild(1) instanceof ASTInitDeclarator))
        CTDDTransformVisitor.errorMessage("Incorrect dataset type",
                                            ((SimpleNode)node.jjtGetChild(0)).getLineNumber());
      Node decl = node.jjtGetChild(0).jjtGetChild(1);
      while (decl.jjtGetNumChildren() > 1)
        decl = decl.jjtGetChild(1);
      featureName = ((SimpleNode)decl.jjtGetChild(0).jjtGetChild(0)).getName();
      isDataset = true;
    }
    else if (node instanceof ASTTDD_TestDirective) {
      // the parser names each directive
      featureName = ((ASTTDD_TestDirective)node).getName();
      //addInclude(unit + "TDD.c");
      isDirective = true;
    }
    else if (node instanceof ASTTDD_FactoryDefinition) {
      Node fctDecl = node.jjtGetChild(1);
      while (fctDecl.jjtGetNumChildren() > 1)
        fctDecl = fctDecl.jjtGetChild(1);
      featureName = ((SimpleNode)fctDecl.jjtGetChild(0).jjtGetChild(0)).getName();
      addInclude(unit + "_CST.h");
    }
    else {
      // error
    }
    className = prefix + unitName + "_" + featureName;
    compilUnit.setName(className);
    try { compilUnit.setProperty("fileExtension", "c"); }
    catch(PropertyAccessException e) {}

    if (debug)
      System.out.println("className : " + className + " _ unitName : " +
                         unitName + " _ featureName " + featureName);
  }


  /**
    * Generic build method.
    * Creates the tree that is common to all builder-generated C files,
    * namely the root ASTTranslationUnit and some of its children:
    * class declaration (ASTDeclaration), name of the feature, name of
    * the inherited classes, declaration of constructor and functions.
    * @param parentName name of the inherited feature class (ADLLiteralSet, etc.)
    * @see #compilUnit
    * @see #className
    */
  protected void build(String parentName) {

    if (debug)
      System.out.println("CBuilder build");

    if (parentName.startsWith("F_") || isDirective)
      addInclude(parentName + ".c");
    else
      addInclude(parentName + ".h");
  }


  public void addInclude(String inclF) {

    if (debug)
      System.out.println("CBuilder addInclude " + inclF);

    String inclFile = "\"" + inclF + "\"";

    if (! (includes.contains(inclFile))) {
      includes.put(inclFile, inclFile);
      ASTIncludeFileDeclaration incl = new
        ASTIncludeFileDeclaration(ADLCParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      incl.setName(inclFile);
      ADLCGenericVisitor.parentChild(compilUnit.jjtGetChild(0),
                                     incl,
                                     compilUnit.jjtGetChild(0).jjtGetNumChildren());
    }
  }

}

// ***********************************************************************
// History Log:
// $Log: CBuilder.java,v $
// Revision 1.3  1998/10/14 08:12:48  mallet
// Bugs fixed: pr/155, pr/156: removed useless constructor for anonymous
// test directives: the parser now names each directive.
//
// Revision 1.2  1998/07/28 16:12:36  mallet
// Added method getFeatureName().
// Changed name of constant file from <unit>_CST.c to <unit>_CST.h.
// Fixed bug in constructor: a test directive can have no name, in which
// case the featureName can be null.
//
// Revision 1.1  1998/07/03 10:07:32  bareau
// Current (incomplete) version of the C/TDD translator.
//
//
// ***********************************************************************
