// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CDatasetBuilder.java,v $
// $Revision: 1.3 $
// $Date: 1998/08/05 16:42:07 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.c
// NAME:        CDatasetBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Feb 3, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.c.CModifier;
import ORG.opengroup.adl.symboltable.LType;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Creates a Java compilation unit AST that corresponds to a dataset
  * declaration.
  * <BR>Uses CDatasetExprVisitor; is used by CTDDTransformVisitor.
  * <P>Example: definition of dataset TEST in tdd class bankTest:
  * <BR>"dataset bank TEST = make_bank(DEPOSITS, BANKSIZE + DEPOSITS,
  * INT_VALUES);"
  * @author Cyrille Bareau
  * @see CDatasetExprVisitor
  * @see CTDDTransformVisitor
  */

public class CDatasetBuilder extends CBuilder {


  /** The prefix used to give names to the components of the dataset
    * expression of the dataset declaration. */
  private String strStackInit = "set";

  /** The name of the method used to add components of the dataset expression.
    * (addElement for dataset concatenation, addParameter for factory call,
    * setValues for literal dataset). */
  private String addMethod;

  /** The name of the class the current dataset inherits from
      (set by method findType). */
  private String classType;

  private LType expectedType;

  static private boolean debug = false;


  /**
    * Single constructor.
    * <BR>Initializes the unit name (that begins with "S_")
    * @param node the dataset declaration node
    * @param unit source unit name
    */
  public CDatasetBuilder (ASTTDD_DatasetDeclaration node, String unit,
			    ASTIncludeFileList incList) {
    super(node, "S_", unit, incList);
    build(node);
  }


  /**
    * The main method. Gets the dataset type and calls makeInitializer.
    *
    * @param node the root node
    */
  private void build(ASTTDD_DatasetDeclaration node) {

    if (debug)
      System.out.println("CDatasetBuilder build");

    addInclude("ADLDataSet.h");
    addInclude(unitName + "_CST.h");

    ASTDeclarationSpecifiers declSpecifier =
      (ASTDeclarationSpecifiers)node.jjtGetChild(0).jjtGetChild(0);
    expectedType = declSpecifier.getType();

    if (declSpecifier.getModifier() != 0x0000)
      CTDDTransformVisitor.warningMessage("Type modifier(s) for dataset types will be ignored",
					  node.getLineNumber());

    ADLCGenericVisitor.parentChild(compilUnit,
                                   makeInitializer(node.jjtGetChild(1)),
                                   compilUnit.jjtGetNumChildren());
  } 


  /**
    * Creates the AST constructor of the dataset class.
    * <BR>Initializes a CDatasetExprVisitor and launches it on the passed
    * dataset expression.
    * Example for dataset TEST:
<PRE>
| 1|  static ADLDataSet *S_bankTest_BA1() {
| 2|    ADLDataSet *set = NULL;
| 3|    ADLDataSet *s1 = NULL, *s2 = NULL;
| 4|
| 5|    set = F_bankTest_make_account();
| 6|    if (set) {
| 7|        set->label= ADL__newString("BA1");
| 8|        if (s1 = S_bankTest_ACCOUNTS()) {
| 9|            ADL_FactoryAddParameter(set, s1);
|10|        }
|11|        if (s2 = S_bankTest_BALANCES()) {
|12|            ADL_FactoryAddParameter(set, s2);
|13|        }
|14|    }
|15|    if ((s1 == NULL) || (s2 == NULL)) {
|16|        set == NULL;
|17|    }
|18|    return(set);
|19|}
</PRE>
    * <BR>In this example, lines 2 to 17 are generated by the DatasetExpressionVisitor
    * run on the dataset expression. The return statement (line 18) is
    * generated by makeInitializer.
    * @param dsExpr the dataset expression of the dataset declaration
    * @return the ASTConstructorDeclaration node root of the constructor tree
    */
  private ASTExternalDeclaration makeInitializer(Node dsExpr) {

    if (debug)
      System.out.println("CDatasetBuilder makeInitializer");

     ASTExternalDeclaration createdNode = new
       ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);

    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( ADLCGenericVisitor.makeType
				     ("ADLDataSet")),
                                   0);
    
    ASTFunctionDeclarator fctDeclTor = new FunctionDeclaratorBuilder
      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
        className)
      .generatedAST();
    ADLCGenericVisitor.parentChild(funcdef, fctDeclTor, 1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(funcdef, body, 2);
    int nbChildren = 0;


    // ADLDataSet* set = NULL;
    ADLCGenericVisitor.parentChild
      ( body,
        new DeclarationBuilder
        ( ADLCGenericVisitor.makeType("ADLDataSet"),
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder
            ( new DeclaratorBuilder
              ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                "set").generatedAST(),
              "NULL"))
          .generatedAST())
        .generatedAST(),
        nbChildren++);

    CDatasetExprVisitor dsExprVisitor =
      new CDatasetExprVisitor(strStackInit, "tc", false, this);
    dsExprVisitor.setExpectedType(expectedType);
    dsExpr.jjtAccept(dsExprVisitor, null);

    for (Enumeration e = dsExprVisitor.getDeclarationStatements().elements(); e.hasMoreElements(); ) {
      ADLCGenericVisitor.parentChild(body,
				     (SimpleNode)e.nextElement(),
				     nbChildren++);
    }

    Vector datasetComp = (Vector)dsExprVisitor.getDatasetComponents().peek();
    int nbComp = ((Integer)dsExprVisitor.getNbComponents().peek()).intValue();

    for (Enumeration e = datasetComp.elements(); e.hasMoreElements(); ) {
      ADLCGenericVisitor.parentChild(body, (SimpleNode)e.nextElement(), nbChildren++);
    }

    SimpleNode stmt = new ASTStatement(ADLCParserTreeConstants.JJTSTATEMENT);
    ADLCGenericVisitor.parentChild(body, stmt, nbChildren++);
    SimpleNode returnStmt = new ASTReturnStatement(ADLCParserTreeConstants.JJTRETURNSTATEMENT);
    ADLCGenericVisitor.parentChild(stmt, returnStmt, 0);
    ADLCGenericVisitor.parentChild(returnStmt,
				   ADLCGenericVisitor.makeName("set"),
				   0);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: CDatasetBuilder.java,v $
// Revision 1.3  1998/08/05 16:42:07  mallet
// Fixed returned type for all declared objects from, e.g.,
// FunctionDefinition to ExternalDeclaration.
// Made all generated functions and constants static.
//
// Revision 1.2  1998/07/30 08:40:42  mallet
// Complete and fully operational version of the TDD -> C translator.
//
// Revision 1.1  1998/07/03 10:07:33  bareau
// Current (incomplete) version of the C/TDD translator.
//
// 
// ***********************************************************************
