// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CDatasetExprVisitor.java,v $
// $Revision: 1.2 $
// $Date: 1998/07/30 08:40:42 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.c
// NAME:        CDatasetExprVisitor
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Jun 26, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Dataset expression visitor.
  * Generates a vector of statements that correspond to the declaration
  * and initialization of a dataset expression (see examples in CDatasetBuilder
  * and CDirectiveBuilder).
  * <BR>Is used by CDatasetBuilder and by CDirectiveBuilder.
  * <P> For an overview of the algorithms used by this transformation, see the
  * ADL2.0 Translation System Design Specification, Chapter 5 (Code Generation),
  * paragraph 5.2.2.6.
  * @author Cyrille Bareau
  * @see CDatasetBuilder
  * @see CDirectiveBuilder
  */

public class CDatasetExprVisitor extends ADLCGenericVisitor {

  /** A stack of vector of statements. These statements correspond to the
    * declaration and initialization of the components of the current dataset
    * expression (these components are the arguments of a factory call or the
    * operands of a dataset concatenation). */
  private Stack datasetComponents = new Stack();

  /** Number of components of the current dataset expression. */
  private Stack nbComponents = new Stack();

  /** A vector of the declaration statements that will have to be emitted
   * for all the created components. */
  private Vector declarationStatements = new Vector();

  /** A stack of strings used to generate the identifiers for the
    * dataset expression components. */
  private Stack strStack = new Stack();

  /** String for the testContext. <BR>"tc" from a dataset builder,
    * "this" from a directive builder. */
  private String testContext;

  /** Flag for anonymous dataset expressions. Anonymous dataset expressions
    * are arguments of a test directive and components of a dataset
    * expression. */
  private boolean anonymous;

  private CBuilder builderCaller;

  private LType expectedType = null;

  private static final boolean debug = false;


  /**
    * Single constructor.
    * @param strStackInit The string that initializes strStack.
    * @param tc The string that initializes testContext.
    */
  public CDatasetExprVisitor (String strStackInit, String tc,
                                    boolean anon, CBuilder caller) {
    super();
    testContext = tc;
    strStack.push(strStackInit);
    datasetComponents.push(new Vector());
    nbComponents.push(new Integer(0));
    anonymous = anon;
    builderCaller = caller;

    if (debug)
      System.out.println("*** new CDatasetExprVisitor: " + anonymous);
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetExpression.
    * <P><UL><LI><B>If</B> the current node has several children,
    * then this is a dataset concatenation: call makeConcatComponents.
    * <LI><B>ELSE</B> just re-transmits the visitor to the child node.</UL>
    * @see #makeConcatComponents
    * @param node the ASTTDD_DatasetExpression node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetExpression node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetExpression (" +
			 node.jjtGetNumChildren() + ")");

    if (node.jjtGetNumChildren() > 1) {
      makeConcatComponents(node, data);
    } else {
      node.jjtGetChild(0).jjtAccept(this, data);
    }

    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetSingleton.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the name is not a dataset name, then it is a constant
    * and must be considered as a singleton literal: call makeSingletonLiteral.
    * <LI><B>ELSE</B><UL>
    * <LI>Get the qualified name of the class that implements the dataset.
    * <LI>Generate the AST for the variable declaration:
<PRE>
|  &lt;dataset qualified name> &lt;getComponentName()> =
|         new &lt;dataset qualified name>(&lt;testContext>);
</PRE>
    * <LI>Add this AST to the vector that is on top of stack datasetComponents.
    * <LI>Add "+1" to the integer on top of stack nbComponents.
    * </UL></UL>
    * @see CBuilder#makeQualifiedName(String, String)
    * @see #makeConcatComponents
    * @see #makeSingletonLiteral
    * @param node the ASTTDD_DatasetSingleton node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetSingleton node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetSingleton");

    SimpleNode child = (SimpleNode)node.jjtGetChild(0);
    if (child instanceof ASTName) {
      // Fix up ASTName by visiting it
      child.jjtAccept(this, data);

      // Can be a dataset or a constant
      boolean isDS;
      try {
        isDS = ((String)child.getProperty("astNameKind"))
          .equals("TDDDataset");
      }
      catch(PropertyAccessException e) { isDS = false; }

      if (! isDS) { // field name
	makeSingletonLiteral((TypedNode)child, data);
      } else {
        // ADLDataSet* <getComponentName> = NULL;
	String datasetName = ((ASTName)child).getName();
	String componentName;

	if (anonymous) {
	  componentName = getComponentName();
	  ASTDeclaration decl = new DeclarationBuilder
	    ( "ADLDataSet",
	      new InitDeclaratorListBuilder
	      ( new InitDeclaratorBuilder
		( new DeclaratorBuilder
		  ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		    componentName)
		  .generatedAST(),
		  makeName("NULL")))
	      .generatedAST())
	    .generatedAST();
	  declarationStatements.addElement(decl);
	} else {
	  componentName = peekName();
	}

        // <componentName> = <datasetName>();
        ASTStatement assign = new StatementBuilder
          ( new AssignmentBuilder
            ( componentName, "=",
              new PostfixExprBuilder
              ( datasetName,
                new ArgumentListBuilder().generatedAST())
              .generatedAST()))
          .generatedAST();
	
	((Vector)datasetComponents.peek()).addElement(assign);
	int nbComp = ((Integer)nbComponents.pop()).intValue();
	nbComponents.push(new Integer(nbComp+1));
      }
    } else {
      makeSingletonLiteral((TypedNode)child, data);
    }
    return null;
  }


  /**
   * Visit pattern for a node of type Name.
   * <P><B>Algorithm</B><UL>
   * <li>Gets defining tdd class name
   * <LI><B>IF</B> the name is not a dataset name, then it is a
   *    constant:<ul>
   * <li>Sets name to
   *    <samp>&lt;defining tdd class name>_CST.&lt;constant name></samp>
   * </ul><LI><B>ELSE</B><UL>
   * <LI>Gets the qualified name of the dataset as
   *    <samp>S_&lt;defining tdd class name>_&lt;dataset name></samp>
   * <LI>Sets name to this qualified name
   * </UL></UL>
   * @see CBuilder#makeQualifiedName(String, String, String)
   * @param node the Name node to visit
   * @param data unused
   * @return nothing
   */
  public Object visit(ASTName node, Object data) {
    if (debug)
      System.out.println("visiting Name : " + node.getName());

    boolean isDS;
    try {
      isDS = ((String)node.getProperty("astNameKind"))
	.equals("TDDDataset");
    }
    catch (PropertyAccessException e) { isDS = false; }
    
    String TDDClassName = null;
    String datasetName;
    try {
      TDDClassName = (String)node.getProperty("enclosingConstructName");
    }
    catch (PropertyAccessException e) { }
    
    if (! isDS) { // field name
      if (TDDClassName != null && !TDDClassName.equals("")) {
	// TDD field name
	datasetName = TDDClassName + "_" + node.getName();
	builderCaller.addInclude(TDDClassName + "_CST.h");
	node.setName(datasetName);
      }
    } else {
      // general case: dataset name, not as root expression
      datasetName = CBuilder.makeQualifiedName("S_",
					       TDDClassName,
					       node.getName());
      builderCaller.addInclude(datasetName + ".c");
      node.setName(datasetName);
    }

    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetLiteral.
    * <P><B>Algorithm</B><OL>
    * <LI>Creates an AST for the variable declaration:
<PRE>
|  ADLDataSet *&lt;getComponentName()> = NULL;
|  &lt;getComponentName()> = ADL_LiteralSetInit(&lt;datasetName>, &lt;type>, NULL, 0);
|  if (&lt;getComponentName()>) { ... }
</PRE>
    * <LI>Pushes component name
    * <li>Adds created node to vector of components on top of datasetCompoents
    * <li>Visits children
    * </OL>
    * @param node the ASTTDD_DatasetLiteral node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetLiteral node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetLiteral");

    LType typeElts = node.getType();
    LType expType = (data != null) ? ((LType)data) : expectedType;
    String datasetName =
      (anonymous ? "anonymous" : builderCaller.getFeatureName());

    builderCaller.addInclude("ADLLiteralSet.h");

    if (anonymous) {
      // ADLDataSet * <getComponentName> = NULL;
      String componentName = pushName(getComponentName());

      ASTDeclaration decl = new DeclarationBuilder
	( "ADLDataSet",
	  new InitDeclaratorListBuilder
	  ( new InitDeclaratorBuilder
	    ( new DeclaratorBuilder
	      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		componentName)
	      .generatedAST(),
	      makeName("NULL")))
	  .generatedAST())
	.generatedAST();
      declarationStatements.addElement(decl);
    }

    // <type> <peekName()>_temp;
    // declaration statement (Note: we may not need it - it depends on
    // TDD_DatasetMember children - but having it here allows a unique
    // declaration)
    String temporaryName = peekName() + "_temp";
    ASTDeclaration tempDecl = new DeclarationBuilder
      (expType.getFullName(),
       new InitDeclaratorListBuilder
       ( new InitDeclaratorBuilder
	 ( new DeclaratorBuilder
	   ( temporaryName)
	   .generatedAST()))
       .generatedAST())
      .generatedAST();

    // <peekName()> =
    // ADL_LiteralSetInit(<datasetName>, <getLiteralType(type)>, NULL, 0);
    ArgumentListBuilder argsBuilder =
      new ArgumentListBuilder(makeString(datasetName))
      .addArgument(makeName(CTypeAux.getLiteralType(typeElts)));
    ASTArgumentList args = argsBuilder.addArgument(makeName("NULL"))
      .addArgument(makeConstant("0"))
      .generatedAST();

    ASTStatement assign = new StatementBuilder
      ( new AssignmentBuilder
        ( peekName(), "=",
          new PostfixExprBuilder("ADL_LiteralSetInit", args)
          .generatedAST()))
      .generatedAST();

    declarationStatements.addElement(tempDecl);
    ((Vector)datasetComponents.peek()).addElement(assign);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));

    // if (<peekName()>) { ... }
    ASTIfStatement ifStmt = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    parentChild(ifStmt, makeName(peekName()), 0);
    
    ASTCompoundStatement thenBlk = new 
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    parentChild(ifStmt, thenBlk, 1);

    // visit children
    datasetComponents.push(new Vector());
    if (typeElts.equals(expType)) {
      node.childrenAccept(this, data);
    }
    else {
      node.childrenAccept(this, expType);
    }

    // attach all parameter statements
    int childNumber = 0;
    Vector paramComponents = (Vector)datasetComponents.pop();

    // add components resulting from visit of parameter to the list 
    // of components of the dataset expression
    for (Enumeration e = paramComponents.elements(); 
	 e.hasMoreElements(); ) {
      parentChild(thenBlk, (SimpleNode)e.nextElement(), childNumber++);
    }
    
    // add if statement to the list of components for this dataset
    ((Vector)datasetComponents.peek()).addElement(ifStmt);

    if (anonymous)
      popName();

    return null;
  }
  

  /**
    * Visit pattern for a node of type ASTTDD_DatasetMember.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the current node has only one child: not a range:<ul>
    * <ul>Generate an AST for statement:
<pre>
|  &lt;ADL_LiteralAddValue(&lt;peekName()>, &&lt;peekName()>_temp);
</pre>
    * </ul><LI><B>ELSE</B> <I>(the current node has two children: range)</I>:<ul>
    * <li>Generate ASTs for statements:
<PRE>
|  ADL_LiteralAddRangeValues(&lt;peekName()>, (long)&lt;visit child1>, (long)&lt;visit child2>);
</PRE>
    * </UL>
    * <li>Created statement is added to the vector on top of stack
    * datasetComponents.
    * </ul>
    * @param node the ASTTDD_DatasetMember node to visit
    * @param data null <I>or</I> the string that represents the Java wrapper
    * type of the literal type if it is primitive.
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetMember node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetMember " +
			 node.jjtGetNumChildren());

    ASTStatement stmt1, stmt2;
    if (node.jjtGetNumChildren() == 1) { // simple case: no range
      // <peekName()>_temp = <node.child1>;
      // ADL_LiteralAddValue(<peekName>, &temp);

      SimpleNode arg = (SimpleNode)node.jjtGetChild(0);

      // Fix up Name children by visiting arg
      arg.jjtAccept(this, data);

      Node ch;
      LType actualType = ((TypedNode)arg).getType();

      if ((data != null) && ! (((LType)data).equals(actualType))) {
        ASTCastExpression castexpr = new
          ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
        ASTTypeName type = new
          ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
        parentChild(type, makeType(((LType)data).getFullName()), 0);
        parentChild(castexpr, type, 0);
        parentChild(castexpr, arg, 1);
        ch = castexpr;
      }
      else {
        ch = arg;
      }
      
      // declaration statement
      String temporaryName = peekName() + "_temp";

      // temporary assignment statement
      stmt1 = new StatementBuilder
	(new AssignmentBuilder(makeName(temporaryName), "=", ch))
	.generatedAST();

      // add value statement
      ASTUnaryOperatorExpression ref = 
	new ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
      try { ref.setProperty("operator", "&"); }
      catch(PropertyAccessException e) {}
      parentChild(ref, makeName(temporaryName), 0);

      stmt2 = new StatementBuilder
        (new PostfixExprBuilder("ADL_LiteralAddValue",
				new ArgumentListBuilder(makeName(peekName()))
				.addArgument(ref)
				.generatedAST())
         .generatedAST(),
	 true)
        .generatedAST();

      ((Vector)datasetComponents.peek()).addElement(stmt1);
      ((Vector)datasetComponents.peek()).addElement(stmt2);
    }
    else { // range -> long values
      // ADL_LiteralAddRangeValues(<peekName()>, (long)<node.child1>, (long)<node.child2>);
      SimpleNode arg1 = (SimpleNode)node.jjtGetChild(0);
      SimpleNode arg2 = (SimpleNode)node.jjtGetChild(1);
      Node ch1, ch2;

      // Fix up IdExpr children by visiting args
      arg1.jjtAccept(this, data);
      arg2.jjtAccept(this, data);

      ch1 = new ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
      ASTTypeName type = new
	ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
      parentChild(type, makeType("long"), 0);
      parentChild(ch1, type, 0);
      parentChild(ch1, arg1, 1);

      ch2 = new ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
      type = new ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
      parentChild(type, makeType("long"), 0);
      parentChild(ch2, type, 0);
      parentChild(ch2, arg2, 1);

      stmt1 = new StatementBuilder
        (new PostfixExprBuilder("ADL_LiteralAddRangeValues",
           new ArgumentListBuilder(makeName(peekName()))
           .addArgument(ch1)
           .addArgument(ch2)
           .generatedAST())
         .generatedAST(),
	 true)
        .generatedAST();

      ((Vector)datasetComponents.peek()).addElement(stmt1);
    }

    if (debug)
      System.out.println("end ASTDatasetMember");

    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_FactoryCall.
    * <UL>
    * <li>Adds a statement to declare an ADLDataSet instance:
<PRE>
|  ADLDataSet *&lt;getComponentName()> = NULL;
</PRE>
    * <li>Adds a statement to initialize the declared set:
<PRE>
|  &lt;getComponentName()> = &lt;factoryName>();
|  if (&lt;getComponentName()>) {
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushed a new Integer(0) on nbComponents
    * <li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Append each of the components in the then block
    * <LI>Add a statement to add the child element to the factory parameters:
<PRE>
|  if (&lt;peekName()>_p&lt;i>)
|     ADL_FactoryAddParameter(&lt;peekName()>, &lt;peekName()>_p&lt;i>);
</PRE>
    * </UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul>
    *
    * @param node the ASTTDD_FactoryCall node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_FactoryCall node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_FactoryCall");

    String datasetName =
      (anonymous ? "anonymous" : builderCaller.getFeatureName());
    String TDDClassName = null;
    try {
      TDDClassName = (String)node.getProperty("enclosingConstructName");
    }
    catch(PropertyAccessException e) { }

    String factName =
      CBuilder.makeQualifiedName("F_", TDDClassName, node.getName());

    builderCaller.addInclude(factName + ".c");
    
    if (anonymous) {
      // ADLDataSet *<componentName> = NULL;
      String componentName = pushName(getComponentName());
    
      ASTDeclaration declSet = new DeclarationBuilder
	( "ADLDataSet",
	  new InitDeclaratorListBuilder
	  ( new InitDeclaratorBuilder
	    ( new DeclaratorBuilder
	      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		componentName)
	      .generatedAST(),
	      makeName("NULL")))
	  .generatedAST())
	.generatedAST();
      declarationStatements.addElement(declSet);
    }
    
    // <peekName()> = <factName>();
    ASTStatement assign = new StatementBuilder
      ( new AssignmentBuilder
	( peekName(), "=",
	  new PostfixExprBuilder(factName,
				 new ArgumentListBuilder
				 ( makeString(datasetName))
				 .generatedAST())
	  .generatedAST()))
      .generatedAST();

    // if (<peekName()>) { ... }
    ASTIfStatement initStmt = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    parentChild(initStmt, makeName(peekName()), 0);
    
    ASTCompoundStatement thenBlk = new 
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    parentChild(initStmt, thenBlk, 1);
    
    ((Vector)datasetComponents.peek()).addElement(assign);
    ((Vector)datasetComponents.peek()).addElement(initStmt);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));
    nbComponents.push(new Integer(0));
    int childNumber = 0;

    // visit parameters of factory call
    boolean myAnonymous = anonymous;
    int numParameters = node.jjtGetNumChildren() - 1;

    anonymous = true;
    for (int i = 0; i <= numParameters; i++) {
      if (debug)
	System.out.println("ASTTDD_FactoryCall: " + factName +
			   " : parameter " + i);
    
      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, data);

      Vector paramComponents = (Vector)datasetComponents.pop();

      // add components resulting from visit of parameter to the list 
      // of components of the dataset expression
      for (Enumeration e = paramComponents.elements(); 
	   e.hasMoreElements(); ) {
	parentChild(thenBlk, (SimpleNode)e.nextElement(), childNumber++);
      }
	
      // if (<peekName()>_p<i>)
      //    ADL_FactoryAddParameter(<peekName()>, <peekName()>_p<i>);
      ASTIfStatement ifst = 
	new ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
      parentChild(ifst,
		  makeName(peekName() + "_p" + (i + 1)),
		  0);
      parentChild
        (ifst,
	 new StatementBuilder
	 ( new PostfixExprBuilder
	   ( makeName("ADL_FactoryAddParameter"),
	     new ArgumentListBuilder
	     ( makeName(peekName()))
	     .addArgument( makeName(peekName() + "_p" + (i + 1)))
	     .generatedAST()))
	 .generatedAST(),
	 1);
      parentChild(thenBlk, ifst, childNumber++);
    }

    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }

    return null;
  }


  /**
    * Creates the code for a singleton literal dataset.
<PRE>
|  ADLDataSet *&lt;componentName> = NULL;
|  &lt;literal type> &lt;componentName>_temp;
|  &lt;componentName>_temp = &lt;literal expression>
|  &lt;componentName> = ADLLiteralSetInit(&lt;datasetName>, &lt;getLiteralType(type)>, NULL, 0);
|  &lt;componentName>_temp = &lt;value>;    
|  ADL_LiteralAddValue(&lt;componentName>, &&lt;componentName>_temp);
</PRE>
    */
  private void makeSingletonLiteral(TypedNode node, Object data) {

    if (debug)
      System.out.println("makeSingletonLiteral: " + node);

    ASTStatement stmt1, stmt2;
    Node ch;
    String componentName = getComponentName();
    String datasetName =
      (anonymous ? "anonymous" : builderCaller.getFeatureName());
    LType actualType = node.getType();
    LType typ = (data == null) ? expectedType : ((LType)data);

    builderCaller.addInclude("ADLLiteralSet.h");

    if (anonymous) {
      // ADLDataSet * <componentName> = NULL;
      ASTDeclaration decl = new DeclarationBuilder
	( "ADLDataSet",
	  new InitDeclaratorListBuilder
	  ( new InitDeclaratorBuilder
	    ( new DeclaratorBuilder
	      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		componentName)
	      .generatedAST(),
	      makeName("NULL")))
	  .generatedAST())
	.generatedAST();
      declarationStatements.addElement(decl);
    }

    // temporary declaration statement
    // <type> <componentName>_temp;
    String temporaryName = componentName + "_temp";
    ASTDeclaration tempDecl = new DeclarationBuilder
      (typ.getFullName(),
       new InitDeclaratorListBuilder
       ( new InitDeclaratorBuilder
	 ( new DeclaratorBuilder
	   ( temporaryName)
	   .generatedAST()))
       .generatedAST())
      .generatedAST();
    declarationStatements.addElement(tempDecl);

    // <componentName> =
    // ADL_LiteralSetInit(<datasetName>, <getLiteralType(type)>, NULL, 0);
    ArgumentListBuilder argsBuilder =
      new ArgumentListBuilder(makeString(datasetName))
      .addArgument(makeName(CTypeAux.getLiteralType(typ)));
    ASTArgumentList args = argsBuilder.addArgument(makeName("NULL"))
      .addArgument(makeConstant("0"))
      .generatedAST();

    ASTStatement assign = new StatementBuilder
      ( new AssignmentBuilder
        ( componentName, "=",
          new PostfixExprBuilder("ADL_LiteralSetInit", args)
          .generatedAST()))
      .generatedAST();

    if (! typ.equals(actualType)) {
      ASTCastExpression castexpr = new
	ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
      ASTTypeName type = new
	ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
      parentChild(type, makeType(typ.getFullName()), 0);
      parentChild(castexpr, type, 0);
      parentChild(castexpr, node, 1);
      ch = castexpr;
    }
    else {
      ch = node;
    }
    
    // temporary assignment statement
    stmt1 = new StatementBuilder
      (new AssignmentBuilder(makeName(temporaryName), "=", ch))
      .generatedAST();
    
    // add value statement
    ASTUnaryOperatorExpression ref = 
      new ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
    try { ref.setProperty("operator", "&"); }
    catch(PropertyAccessException e) {}
    parentChild(ref, makeName(temporaryName), 0);
    
    stmt2 = new StatementBuilder
      (new PostfixExprBuilder("ADL_LiteralAddValue",
			      new ArgumentListBuilder
			      ( makeName(componentName))
			      .addArgument(ref)
			      .generatedAST())
       .generatedAST(),
       true)
      .generatedAST();
    
    ((Vector)datasetComponents.peek()).addElement(assign);
    ((Vector)datasetComponents.peek()).addElement(stmt1);
    ((Vector)datasetComponents.peek()).addElement(stmt2);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));

  }


  /**
    * Create the code for a dataset concatenation expression.
    * <UL><LI><B>IF</B> anonymous expression
<PRE>
|  ADLConcatenationSet &#60;getComponentName()&#62; = new
|     ADLConcatenationSet(&#60;testContext&#62;, null);
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushes a new Integer(0) on nbComponents
    * <li>Keeps value of <samp>anonymous</samp> and set it to <samp>true</samp>
    * </ul><li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Copies each of the components in this vector into vector on top
    *     of stack
    * <LI><B>IF</B> anonymous expression,<UL>
    * <LI>add statements to add the child element to the dataset:
<PRE>
|  &#60;peekName()&#62;->addElement(&#60;peekName()&#62;_p&#60;i&#62;);
</PRE>
    * </UL></UL>
    * <li>Restores previous value of <samp>anonymous</samp>
    * <li><B>IF</B> anonymous expression,<UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul></ul>
    * @param node the ASTTDD_DatasetExpression node to visit
    * @param data unused
    */
  private void makeConcatComponents(ASTTDD_DatasetExpression node, Object data) {

    if (debug)
      System.out.println("makeConcatComponents: " + node);

    String datasetName =
      (anonymous ? "anonymous" : builderCaller.getFeatureName());
    builderCaller.addInclude("ADLConcatenationSet.h");

    if (anonymous) {
      if (debug)
        System.out.println("makeConcatComponents: anonymous");

      // ADLDataSet *<getComponentName()> = NULL;
      String componentName = pushName(getComponentName());

      ASTDeclaration decl = new DeclarationBuilder
	( "ADLDataSet",
	  new InitDeclaratorListBuilder
	  ( new InitDeclaratorBuilder
	    ( new DeclaratorBuilder
	      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		componentName)
	      .generatedAST(),
	      makeName("NULL")))
	  .generatedAST())
	.generatedAST();
      declarationStatements.addElement(decl);
    }

    // <componentName> = ADLConcatenationSetInit(<datasetName>);
    ASTStatement assign = new StatementBuilder
      ( new AssignmentBuilder
	( peekName(), "=",
	  new PostfixExprBuilder("ADL_ConcatenationSetInit",
				 new ArgumentListBuilder
				 (makeString(datasetName))
				 .generatedAST())
	  .generatedAST()))
      .generatedAST();

    // if (<peekName()>) { ... }
    ASTIfStatement initStmt = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    parentChild(initStmt, makeName(peekName()), 0);
    
    ASTCompoundStatement thenBlk = new 
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    parentChild(initStmt, thenBlk, 1);
    
    ((Vector)datasetComponents.peek()).addElement(assign);
    ((Vector)datasetComponents.peek()).addElement(initStmt);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));
    nbComponents.push(new Integer(0));
    int childNumber = 0;

    // visit elements of concatenation
    boolean myAnonymous = anonymous;

    anonymous = true;
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      if (debug)
	System.out.println("makeConcatComponents: element " + (i + 1));

      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, data);

      Vector elemComponents = (Vector)datasetComponents.pop();
      
      // add components resulting from visit of element to the list 
      // of components of the dataset expression
      Vector thetop = (Vector)datasetComponents.peek();
      for (Enumeration e = elemComponents.elements(); 
	   e.hasMoreElements(); ) {
	parentChild(thenBlk, (SimpleNode)e.nextElement(), childNumber++);
      }

      // if (<peekName()>_p<i>)
      //   ADL_ConcatenationAddElement(<peekName()>, <peekName()>_p<i+1>);
      ASTIfStatement ifst = 
	new ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
      parentChild(ifst,
		  makeName(peekName() + "_p" + (i + 1)),
		  0);
      ASTStatement stmt = new StatementBuilder
	(new PostfixExprBuilder
	 ( "ADL_ConcatenationAddElement",
	   new ArgumentListBuilder
	   ( makeName(peekName()) )
	   .addArgument(makeName(peekName() + "_p" + (i + 1)))
	   .generatedAST()))
	.generatedAST();
      parentChild(ifst, stmt, 1);
      parentChild(thenBlk, ifst, childNumber++);
    }
    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }
  }

  /**
   * Set expected type
   */
  public void setExpectedType(LType type) {
    expectedType = type;
  }


  /**
    * @return declarationStatements
    */
  public Vector getDeclarationStatements() {
    return declarationStatements;
  }


  /**
    * @return datasetComponents
    */
  public Stack getDatasetComponents() {
    return datasetComponents;
  }


  /**
    * @return nbComponents
    */
  public Stack getNbComponents() {
    return nbComponents;
  }

  private String getComponentName() {
    return (String)strStack.peek() +
      "_p" +
      (((Integer)nbComponents.peek()).intValue() + 1);
  }

  private String pushName(String levelName) {
    strStack.push(levelName);
    return levelName;
  }

  private String popName() {
    return (String)strStack.pop();
  }

  private String peekName() {
    return (String)strStack.peek();
  }
}

// ***********************************************************************
// History Log:
// $Log: CDatasetExprVisitor.java,v $
// Revision 1.2  1998/07/30 08:40:42  mallet
// Complete and fully operational version of the TDD -> C translator.
//
// Revision 1.1  1998/07/03 10:07:33  bareau
// Current (incomplete) version of the C/TDD translator.
//
//
// ***********************************************************************
