// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CDirectiveBuilder.java,v $
// $Revision: 1.3 $
// $Date: 1998/10/14 08:14:23 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.c
// NAME:        CDirectiveBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Jun 25, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.c.CModifier;
import ORG.opengroup.adl.symboltable.*;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Creates a Java compilation unit AST that corresponds to a directive
  * declaration.
  * <BR>Uses CDatasetExprVisitor; is used by CTDDTransformVisitor.
  * <P>Example: definition of directive "mydir" in tdd class bankTest:
  * <BR>"mydir : test (bank b = B_COMBINE,int d = DEPOSITS,int w = WITHDRAWS+DEPOSITS)
  * test_bank (b,d,w);"
  * @author Cyrille Bareau
  * @see CDatasetExprVisitor
  * @see CTDDTransformVisitor
  */

public class CDirectiveBuilder extends CBuilder {

  /** Handle on the TDD single declarators used in domains in the directive
    * parameters */
  private Vector singleDecls;

  /** Names of the ACOs to generate. */
  private Vector acoNames = new Vector();
  

  public Vector getACONames() {
    return acoNames;
  }

  static private boolean debug = false;


  /**
    * Single constructor.
    * <BR>Initializes the unit name (that begins with "D_")
    * @param node the directive declaration node
    * @param unit source unit name
    */
  public CDirectiveBuilder (ASTTDD_TestDirective node, String unit,
                            ASTIncludeFileList incList)
    {
    super(node, "D_", unit, incList);
    singleDecls = new Vector();
    build(node);
  }


  /**
    * Performs some initializations (call to inherited CBuilder.build) and
    * launches the methods that generate the code.
    * @see CBuilder#build(String)
    */
  private void build(ASTTDD_TestDirective node) {

    if (debug)
      System.out.println("CDirectiveBuilder build");

    build(unitName + "TDD");

    int nbChildren = compilUnit.jjtGetNumChildren();

    ADLCGenericVisitor.parentChild(compilUnit, makeInitializer(node), nbChildren++);
    ADLCGenericVisitor.parentChild(compilUnit, makeRunInstance(node), nbChildren++);
  }


  /**
    * Creates the initializer of the generated directive.
    * Example for directive mydir:
<PRE>
| 1|  void ADL_initialize() {
| 2|     ADLDataSet *elt_p1 = NULL;
| 3|     ADLDataSet *elt_p2 = NULL;
| 4|     ADLDataSet *elt_p3 = NULL;
| 5|     elt_p1 = S_bankTest_B_COMBINE();
| 6|     elt_p2 = S_bankTest_DEPOSITS();
| 7|     elt_p3 = S_bankTest_WITHDRAWS();
| 8|     if (elt_p1) ADL_addParameter(elt_p1);
| 9|     if (elt_p2) ADL_addParameter(elt_p2);
|10|     if (elt_p3) ADL_addParameter(elt_p3);
|11|  }
</PRE>
    * The transformation of each directive parameter dataset expression is
    * performed by calling a CDatasetExprVisitor (in the example, lines
    * 2 to 7).
    * @see CDatasetExprVisitor
    * @param testDir the test directive
    * @return The root of the constructor (ASTConstructorDeclaration).
    */
  private ASTFunctionDefinition makeInitializer(ASTTDD_TestDirective testDir) {

    if (debug)
      System.out.println("CDirectiveBuilder makeInitializer");

    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);

    ADLCGenericVisitor.parentChild(createdNode,
                                   ADLCGenericVisitor.makeBuiltinType("void"),
                                   0);

    ADLCGenericVisitor.parentChild(createdNode,
                                   new FunctionDeclaratorBuilder("ADL_initialize")
                                   .generatedAST(),
                                   1);


    String strStackInit = "elt";
    ASTTDD_DatasetDomain domain;
    CDatasetExprVisitor exprVisitor =
      new CDatasetExprVisitor(strStackInit, "this", true, this);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(createdNode, body, 2);
    int nbChildren = 0; LType expectedType;

    // visit all test directive parameters to gather the statements 
    // that create them
    for (int i = 0; i < testDir.jjtGetNumChildren()-1; i++) {
      domain = (ASTTDD_DatasetDomain)testDir.jjtGetChild(i);
      if (domain.jjtGetNumChildren() > 1) {
        expectedType = ((ASTDeclarationSpecifiers)domain.jjtGetChild(0).jjtGetChild(0))
          .getType();
        singleDecls.addElement(domain.jjtGetChild(0));

        if (((ASTDeclarationSpecifiers)domain.jjtGetChild(0).jjtGetChild(0)).getModifier() != 0x0000)
          CTDDTransformVisitor.warningMessage("Type modifier(s) for tdd directive parameters will be ignored",
                                              domain.getLineNumber());
      }
      else { // void dataset
        expectedType = null;
        singleDecls.addElement(null);
      }

      exprVisitor.setExpectedType(expectedType);

      domain.jjtGetChild(domain.jjtGetNumChildren()-1).jjtAccept(exprVisitor, null);

    }

    // attach all the declarations for the directive parameters
    for (Enumeration e = exprVisitor.getDeclarationStatements().elements();
	 e.hasMoreElements(); ) {
        ADLCGenericVisitor.parentChild(body,
                                       (SimpleNode)e.nextElement(),
                                       nbChildren++);
    }
      
    // now attach all the statements that create the directive
    // parameters (components of the DatasetExpressionVisitor) to the current
    // block
    for (Enumeration e = ((Vector)exprVisitor.getDatasetComponents().peek()).elements();
           e.hasMoreElements(); ) {
        ADLCGenericVisitor.parentChild(body,
                                       (SimpleNode)e.nextElement(),
                                       nbChildren++);
    }

    // and also add the "ADL_addParameter" statements for the directive
    // parameters
    ASTIfStatement ifst;
    for (int i = 0; i < testDir.jjtGetNumChildren()-1; i++) {
      ifst = new ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
      ADLCGenericVisitor.parentChild
        (ifst,
         ADLCGenericVisitor.makeName(strStackInit + "_p" + (i+1)),
         0);
      ADLCGenericVisitor.parentChild
        (ifst,
         new StatementBuilder
         ( new PostfixExprBuilder
           ( "ADL_addParameter",
             new ArgumentListBuilder
             ( ADLCGenericVisitor.makeName(strStackInit + "_p" + (i+1)))
             .generatedAST()))
         .generatedAST(),
         1);
      ADLCGenericVisitor.parentChild(body, ifst, nbChildren++);
    }
    return createdNode;
  }

   
  /**
    * Creates the "ADL_runInstance" function of the generated directive.
    * Example for directive mydir:
<PRE>
|  void ADL_runInstance(void **params) {
|     bank *b = (bank *)params[0];
|     int d = *((int *)params[1]);
|     int w = *((int *)params[2]);
|     ADL_printf("Parameter bank *b = %s", ADL_Print_Pointer(b));
|     ADL_printf("Parameter int d = %s", ADL_Print_Int(d));
|     ADL_printf("Parameter int w = %s", ADL_Print_Int(w));
|     &#60;directive body&362;
|  }
</PRE>
    * <B>Note:</B> the directive body is visited by an RemoveADLVisitor
    * to transform the ADL/ADL_new expressions it contains.
    * @see RemoveADLVisitor
    * @return The root of the constructor (ASTConstructorDeclaration).
    */
  private ASTFunctionDefinition makeRunInstance(ASTTDD_TestDirective testDir)
    {

    if (debug)
      System.out.println("CDirectiveBuilder makeRunInstance");

    // protected void ADL_runInstance(void** paramValues)
    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCGenericVisitor.parentChild(createdNode,
                                   ADLCGenericVisitor.makeBuiltinType("void"),
                                   0);

    ADLCGenericVisitor.parentChild
      ( createdNode,
        new FunctionDeclaratorBuilder
        ( "ADL_runInstance",
          new ParameterDeclarationListBuilder
          ( new ParameterDeclarationBuilder
            ( ADLCGenericVisitor.makeBuiltinType("void"),
              new DeclaratorBuilder
              ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                new DirectDeclaratorBuilder("paramValues").generatedAST())
              .generatedAST()))
          .generatedAST())
        .generatedAST(),
        1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(createdNode, body, 2);

    int nbChildren = 0;
    ASTADL_NamedParam single; ASTInitDeclarator initDecl;
    ASTDeclarationSpecifiers typ1; Node declTor = null, declTorBis;
    ASTAbstractDeclarator abstr, abstrBis; ASTCastExpression castExpr;
    ASTTypeName typeName; ASTArraySuffix suffix; ASTUnaryOperatorExpression unary;
    ASTString str; ASTConstant index; boolean isPrimitiveType = false;
    SimpleNode id = null;
    ASTDotAccessSuffix reset;
    String paramName = "", paramDecl = "";
    Vector printStatements = new Vector();

    for (int indexSD = 0; indexSD < singleDecls.size(); indexSD++) {
      single = (ASTADL_NamedParam)singleDecls.elementAt(indexSD);
      if (single != null) { // not void dataset parameter

        typ1 = (ASTDeclarationSpecifiers)single.jjtGetChild(0);
        declTor = single.jjtGetChild(1); // ASTDeclarator
	paramDecl = single.getExpressionImage();

        if (declTor.jjtGetChild(0) instanceof ASTDirectDeclarator) { // non-pointer type
          isPrimitiveType = CTypeAux.isPrimitiveType(typ1);
	  id = (ASTDirectDeclarator)declTor.jjtGetChild(0);; // id is an ASTDirectDeclarator
        }
        else { // pointer type
          isPrimitiveType = false;
        }


        if (isPrimitiveType) {
          // <typ1> <id> = *((<typ1> *)params[<i>]);
          unary = new
            ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
          try { unary.setProperty("operator", "*"); }
          catch(PropertyAccessException e) {}

          initDecl = new InitDeclaratorBuilder(declTor, unary)
	    .generatedAST();

          ASTParentheizedExpression par = new
            ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);

          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
          ADLCGenericVisitor.parentChild(par, castExpr, 0);
          ADLCGenericVisitor.parentChild(unary, par, 0);

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, typ1, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          ADLCGenericVisitor.parentChild(abstr,
                                         new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                                         0);

          // primary expression
          suffix = new
            ASTArraySuffix(ADLCParserTreeConstants.JJTARRAYSUFFIX);
          index = new
            ASTConstant(ADLCParserTreeConstants.JJTCONSTANT);
          index.setName(String.valueOf(indexSD));
          ADLCGenericVisitor.parentChild(suffix, index, 0);

          ADLCGenericVisitor.parentChild(castExpr,
                                         new PostfixExprBuilder
                                         ( ADLCGenericVisitor.makeName("paramValues"),
                                           suffix)
                                         .generatedAST(),
                                         1);

          ADLCGenericVisitor.parentChild(body,
                                         new DeclarationBuilder(typ1, initDecl)
                                         .generatedAST(),
                                         nbChildren++);
        }
        else {
          // <typ1> <id> = (<typ1>)params[<i>];

          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);

          initDecl = new InitDeclaratorBuilder(declTor, castExpr)
            .generatedAST();

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, typ1, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          declTorBis = declTor;
          while (declTorBis.jjtGetChild(0) instanceof ASTPtrOperator) {
            ADLCGenericVisitor.parentChild(abstr,
					   declTorBis.jjtGetChild(0),
					   0);
            abstrBis = new
              ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	    abstrBis.setName("");
            ADLCGenericVisitor.parentChild(abstr, abstrBis, 1);
            abstr = abstrBis;
            declTorBis = declTorBis.jjtGetChild(1);
          }
	  // declTorBis.jjtGetChild(0) is now a DirectDeclarator
          id = (ASTDirectDeclarator)declTorBis.jjtGetChild(0);

          // primary expression
          suffix = new
            ASTArraySuffix(ADLCParserTreeConstants.JJTARRAYSUFFIX);
          index = new
            ASTConstant(ADLCParserTreeConstants.JJTCONSTANT);
          index.setName(String.valueOf(indexSD));
          ADLCGenericVisitor.parentChild(suffix, index, 0);

          ADLCGenericVisitor.parentChild(castExpr,
                                        new PostfixExprBuilder
                                        ( ADLCGenericVisitor.makeName("paramValues"), suffix)
                                        .generatedAST(),
                                        1);

          ADLCGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(typ1, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }


	// get parameter name from the DirectDeclarator node
	id = (SimpleNode)id.jjtGetChild(0); // Name or Declarator
	while (!(id instanceof ASTName)) {
	  while (!(id instanceof ASTDirectDeclarator))
	    id = (SimpleNode)id.jjtGetChild(id.jjtGetNumChildren() - 1);
	  id = (SimpleNode)id.jjtGetChild(0);
	}
	paramName = id.getName();

        // ADL_printf("Parameter int w = %s", ADL_Print_<type>(w));
        str = ADLCGenericVisitor.makeString("Parameter " + paramDecl +
					    " = %s");

        ASTPostfixExpression printArg = new PostfixExprBuilder
          ( CTypeAux.getPrintType(typ1.getType()),
            new ArgumentListBuilder
	    ( ADLCGenericVisitor.makeName(paramName))
            .generatedAST())
          .generatedAST();

	printStatements.addElement
	  ( new StatementBuilder
            ( new PostfixExprBuilder
              ( "ADL_printf",
                new ArgumentListBuilder(str)
                .addArgument(printArg)
                .generatedAST()))
            .generatedAST());

      }
    }

    for (Enumeration e = printStatements.elements(); e.hasMoreElements();)
      ADLCGenericVisitor.parentChild(body,
				     (Node)e.nextElement(),
				     nbChildren++);

    // user code:

    // visit the user code to change ADL expressions
    RemoveADLVisitor v = new RemoveADLVisitor();
    Node userStmt = 
      (Node)testDir.jjtGetChild(testDir.jjtGetNumChildren() - 1)
      .jjtAccept(v, null);
    Vector inclACO = v.getACONames();
    for (Enumeration e = inclACO.elements(); e.hasMoreElements(); ) {
      String acoName = (String)e.nextElement();
      if ( ! acoNames.contains(acoName) ) {
	acoNames.addElement(acoName);
      }
      addInclude(acoName + ".c");
    }

    if (userStmt instanceof ASTCompoundStatement) {
      for (int i = 0; i < userStmt.jjtGetNumChildren(); i++) {
        ADLCGenericVisitor.parentChild(body, userStmt.jjtGetChild(i), nbChildren++);
      }
    }
    else {
      ADLCGenericVisitor.parentChild(body, userStmt, nbChildren);
    }

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: CDirectiveBuilder.java,v $
// Revision 1.3  1998/10/14 08:14:23  mallet
// Bugs fixed: pr/155, pr/156: removed useless index of anonymous
// directive as argument to constructor.
//
// Revision 1.2  1998/07/30 08:40:43  mallet
// Complete and fully operational version of the TDD -> C translator.
//
// Revision 1.1  1998/07/03 10:07:33  bareau
// Current (incomplete) version of the C/TDD translator.
//
//
// ***********************************************************************
