// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CFactoryBuilder.java,v $
// $Revision: 1.3 $
// $Date: 1998/08/05 16:42:07 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.c
// NAME:        CFactoryBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Jun 25, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.Vector;
import java.util.Enumeration;


/**
  * Class that creates the C translation unit AST that corresponds to a
  * factory declaration.
  * <BR>Is used by CTDDTransformVisitor.
  * <P>Example: factory "make_bank" defined in tdd class "bankTest":
<PRE>
  factory account make_account(account account_number, int initDeposit) {
    return create(account_number, initDeposit);
  }
</PRE>
  * @author Cyrille Bareau
  * @see CTDDTransformVisitor
  */

public class CFactoryBuilder extends CBuilder {

  /** The return type declaration specifiers of the factory */
  private ASTDeclarationSpecifiers factDeclSpec;

  /** The name of the factory return type */
  private String factTypeName;

  /** The declarator of the factory */
  private ASTFunctionDeclarator factDeclTor;

  /** The declarator of the factory */
  private ASTFunctionDirectDeclarator factDirDeclTor;

  /** The formal parameters of the factory */
  private ASTParameterDeclarationList factParams;

  /** The body of the factory */
  private ASTCompoundStatement factBlk;

  /** The body of the (optional) relinquish block */
  private ASTCompoundStatement relBlock = null;

  /** The parameter of the relinquish block (for non void factories) */
  private ASTParameterDeclaration relParam = null;

  /** Flag "the factory is void: no returned object" */
  private boolean voidFactory;

  /** Flag "no relinquish clause in the factory declaration */
  private boolean noRelinquish = true;

  static final private boolean debug = false;


  /**
    * Single constructor.
    * @param node The factory declaration AST node
    * @param imports The import declarations
    * @param unit The name of the adlclass file
    */
  public CFactoryBuilder(ASTTDD_FactoryDefinition node, String unit,
			   ASTIncludeFileList incList) {

    super(node, "F_", unit, incList);
    build(node);
  }


  /**
    * Initializes the data and launches the methods code generation.
    */
  private void build(ASTTDD_FactoryDefinition node) {

    if (debug)
      System.out.println("CFactoryBuilder build");

    build("ADLFactorySet");

    factDeclSpec = (ASTDeclarationSpecifiers)node.jjtGetChild(0);
    
    if (factDeclSpec.getModifier() != 0x0000)
      CTDDTransformVisitor.warningMessage("Type modifier(s) for factory return type will be ignored",
                                            factDeclSpec.getLineNumber());

    factBlk = (ASTCompoundStatement)node.jjtGetChild(2);
    factTypeName = factDeclSpec.getType().getName();
    voidFactory = factDeclSpec.getType().equals(CBuiltinType.voidT);

    factDeclTor = (ASTFunctionDeclarator)node.jjtGetChild(1);
    Node declTor = factDeclTor;
    while (! (declTor.jjtGetChild(0) instanceof ASTFunctionDirectDeclarator)) {
      factTypeName += "*";
      voidFactory = false;
      declTor = declTor.jjtGetChild(1);
    }
    factDirDeclTor = (ASTFunctionDirectDeclarator)declTor.jjtGetChild(0);

    if (factDirDeclTor.jjtGetNumChildren() > 1) {
      factParams = (ASTParameterDeclarationList)factDirDeclTor.jjtGetChild(1).jjtGetChild(0);
    }
    else factParams = null;

    if (factParams != null) {
      for (int i = factParams.jjtGetNumChildren()-1; i >= 0; i--) {
        if (((ASTDeclarationSpecifiers)factParams.jjtGetChild(i).jjtGetChild(0))
            .getModifier() != 0x0000)
          CTDDTransformVisitor.warningMessage
            ("Type modifier(s) for factory parameters will be ignored",
             node.getLineNumber());
      }
    }


    if (node.jjtGetNumChildren() > 3) {
      noRelinquish = false;
      if (node.jjtGetNumChildren() > 4) {
        relParam = (ASTParameterDeclaration)node.jjtGetChild(3);
        relBlock = (ASTCompoundStatement)node.jjtGetChild(4);
      }
      else {
        relBlock = (ASTCompoundStatement)node.jjtGetChild(3);
      }
    }
    
    int nbChildren = compilUnit.jjtGetNumChildren();

    ADLCGenericVisitor.parentChild(compilUnit,
				   makeProvide(node),
				   nbChildren++);
    ADLCGenericVisitor.parentChild(compilUnit,
				   makeFactoryProvide(),
				   nbChildren++);
    if (! noRelinquish) {
      ADLCGenericVisitor.parentChild(compilUnit,
				     makeRelinquish(),
				     nbChildren++);
      ADLCGenericVisitor.parentChild(compilUnit,
				     makeFactoryRelinquish(),
				     nbChildren++);
    }
    ADLCGenericVisitor.parentChild(compilUnit,
				   makeInitializer(),
				   nbChildren++);
  }


  /**
    * Creates the factory constructor.
    * Example for factory make_bank:
<PRE>
|  static ADLDataSet *F_bankTest_make_bank(char *label) {
|    ADLDataSet * set = NULL;
|    if ( set = ADL_FactorySetInit("make_bank") ) {
|       set->label = ADL__new_string(label);
|       set->mode.f.factoryProvide = F_bankTest_make_bank_Provide;
|       set->mode.f.factoryRelinquish = F_bankTest_make_bank_Relinquish;
|    }
|    return set;
|  }
</PRE>
    * @return the root of the constructor (ASTConstructorDeclaration).
    */
  private ASTExternalDeclaration makeInitializer() {

    if (debug)
      System.out.println("CFactoryBuilder makeInitializer");

     ASTExternalDeclaration createdNode = new
       ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);

    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( ADLCGenericVisitor.makeType
				     ("ADLDataSet")),
                                   0);

    ADLCGenericVisitor.parentChild
      (funcdef,
       new FunctionDeclaratorBuilder
       ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
         className,
         new ParameterDeclarationListBuilder
         ( new ParameterDeclarationBuilder
           ( ADLCGenericVisitor.makeBuiltinType("char"),
             new DeclaratorBuilder
             ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
               "label")
             .generatedAST()))
         .generatedAST())
       .generatedAST(),
       1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(funcdef, body, 2);


    // ADLDataSet* set = NULL;
    ADLCGenericVisitor.parentChild
      ( body,
        new DeclarationBuilder
        ( ADLCGenericVisitor.makeType("ADLDataSet"),
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder
            ( new DeclaratorBuilder
              ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                "set").generatedAST(),
              "NULL"))
          .generatedAST())
        .generatedAST(),
        0);


    // if (set = ADL_FactorySetInit(<featurename>))
    ASTIfStatement ifSt = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    ADLCGenericVisitor.parentChild(body, ifSt, 1);

    ASTPostfixExpression initExpr = new PostfixExprBuilder
      ( "ADL_FactorySetInit",
        new ArgumentListBuilder(ADLCGenericVisitor.makeString(featureName))
        .generatedAST())
      .generatedAST();

    ADLCGenericVisitor.parentChild(ifSt,
                                   new AssignmentBuilder("set", "=", initExpr)
                                   .generatedAST(),
                                   0);

    ASTCompoundStatement thenSt = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(ifSt, thenSt, 1);
    int childNumber = 0;

    // set->label = ADL__newString(label);
    ASTPostfixExpression setExpr = new
      ASTPostfixExpression(ADLCParserTreeConstants.JJTPOSTFIXEXPRESSION);
    ADLCGenericVisitor.parentChild(setExpr,
                                   ADLCGenericVisitor.makeName("set"),
                                   0);
    ASTRefAccessSuffix suff1 = new
      ASTRefAccessSuffix(ADLCParserTreeConstants.JJTREFACCESSSUFFIX);
    suff1.setName("label");
    ADLCGenericVisitor.parentChild(suff1,
				   ADLCGenericVisitor.makeName("label"),
				   0);
    ADLCGenericVisitor.parentChild(setExpr, suff1, 1);

    ADLCGenericVisitor.parentChild
      ( thenSt,
        new StatementBuilder
        ( new AssignmentBuilder
	  ( setExpr, "=",
	    new PostfixExprBuilder("ADL__new_string",
				   new ArgumentListBuilder
				   ( ADLCGenericVisitor.makeName("label"))
				   .generatedAST())
	    .generatedAST()))
        .generatedAST(),
        childNumber++);

    // set->mode.f.factoryProvide = <className>_Provide
    setExpr = new
      ASTPostfixExpression(ADLCParserTreeConstants.JJTPOSTFIXEXPRESSION);
    ADLCGenericVisitor.parentChild(setExpr,
                                   ADLCGenericVisitor.makeName("set"),
                                   0);
    suff1 = new
      ASTRefAccessSuffix(ADLCParserTreeConstants.JJTREFACCESSSUFFIX);
    suff1.setName("mode");
    ADLCGenericVisitor.parentChild(setExpr, suff1, 1);

    ASTDotAccessSuffix suff2 = new
      ASTDotAccessSuffix(ADLCParserTreeConstants.JJTDOTACCESSSUFFIX);
    suff2.setName("f");
    ADLCGenericVisitor.parentChild(setExpr, suff2, 2);

    ASTDotAccessSuffix suff3 = new
      ASTDotAccessSuffix(ADLCParserTreeConstants.JJTDOTACCESSSUFFIX);
    suff3.setName("factoryProvide");
    ADLCGenericVisitor.parentChild(setExpr, suff3, 3);

    ADLCGenericVisitor.parentChild
      ( thenSt,
        new StatementBuilder
        ( new AssignmentBuilder(setExpr, "=",
                                ADLCGenericVisitor.makeName(className + "_Provide")))
        .generatedAST(),
        childNumber++);

    // set->mode.f.factoryRelinquish = <className>_Relinquish | NULL
    setExpr = new
      ASTPostfixExpression(ADLCParserTreeConstants.JJTPOSTFIXEXPRESSION);
    ADLCGenericVisitor.parentChild(setExpr,
				   ADLCGenericVisitor.makeName("set"),
				   0);
    suff1 = new
      ASTRefAccessSuffix(ADLCParserTreeConstants.JJTREFACCESSSUFFIX);
    suff1.setName("mode");
    ADLCGenericVisitor.parentChild(setExpr, suff1, 1);
    
    suff2 = new
      ASTDotAccessSuffix(ADLCParserTreeConstants.JJTDOTACCESSSUFFIX);
    suff2.setName("f");
    ADLCGenericVisitor.parentChild(setExpr, suff2, 2);
    
    suff3 = new
      ASTDotAccessSuffix(ADLCParserTreeConstants.JJTDOTACCESSSUFFIX);
    suff3.setName("factoryRelinquish");
    ADLCGenericVisitor.parentChild(setExpr, suff3, 3);
    
    if (! noRelinquish) {
      ADLCGenericVisitor.parentChild
        ( thenSt,
          new StatementBuilder
          ( new AssignmentBuilder(setExpr, "=",
                                  ADLCGenericVisitor.makeName(className + "_Relinquish")))
          .generatedAST(),
          childNumber++);
    } else {
      ADLCGenericVisitor.parentChild
        ( thenSt,
          new StatementBuilder
          ( new AssignmentBuilder(setExpr, "=",
                                  ADLCGenericVisitor.makeName("NULL")))
          .generatedAST(),
          childNumber++);
    }

    // return(set);
    ASTReturnStatement ret = new
      ASTReturnStatement(ADLCParserTreeConstants.JJTRETURNSTATEMENT);
    ADLCGenericVisitor.parentChild(ret,
                                   ADLCGenericVisitor.makeName("set"),
                                   0);
    ADLCGenericVisitor.parentChild(body, ret, 2);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }


  /**
    * Creates the method "factory_provide".
    * Example for factory make_account:
<PRE>
|  static void *F_bankTest_make_account_Provide(void **p) {
|      void *ret = NULL;
|      account val;
|  
|      account elt0;
|      int elt1;
|  
|      elt0 = *((account *) p[0]);
|      elt1 = *((int *) p[1]);
|      if (ADL_debugMode) {
|          ADL_printf("\taccount account_number = %s", ADL_Print_Int(elt0));
|          ADL_printf("\tint initDeposit = %s", ADL_Print_Int(elt1));
|      }
|  
|      if (ret = ADL__malloc(sizeof(account))) {
|          val = provide_make_account(elt0, elt1);
|          memcpy(ret, &val, sizeof(account));
|      }
|      return ret;
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTExternalDeclaration makeFactoryProvide() {

    if (debug)
      System.out.println("CFactoryBuilder makeFactoryProvide");

     ASTExternalDeclaration createdNode = new
       ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( ADLCGenericVisitor.makeBuiltinType
				     ("void")),
                                   0);
    ADLCGenericVisitor.parentChild
      (funcdef,
       new FunctionDeclaratorBuilder
       ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
         className + "_Provide",
          new ParameterDeclarationListBuilder
          ( new ParameterDeclarationBuilder
            ( ADLCGenericVisitor.makeBuiltinType("void"),
              new DeclaratorBuilder
              ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
                new DirectDeclaratorBuilder("params").generatedAST())
              .generatedAST()))
          .generatedAST())
        .generatedAST(),
        1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(funcdef, body, 2);
    int nbChildren = 0;

    // void* ret = NULL;
    ADLCGenericVisitor.parentChild
      ( body,
        new DeclarationBuilder
        ( ADLCGenericVisitor.makeBuiltinType("void"),
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder
	    ( new DeclaratorBuilder
	      ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
		"ret").generatedAST(),
	      "NULL"))
          .generatedAST())
	.generatedAST(),
        nbChildren++);

    // <factReturnType> val;
    ADLCGenericVisitor.parentChild
      ( body,
        new DeclarationBuilder(factTypeName,
                               ADLCGenericVisitor.makeName("val"))
        .generatedAST(),
        nbChildren++);

    
    ASTDeclarationSpecifiers ptype; String pname; ASTArraySuffix brack;
    ASTCastExpression castExpr; ASTPostfixExpression prim;
    ASTParentheizedExpression parenth; boolean isPrimitiveType = false;
    ASTConstant arrlit; Node declTor = null, declTorBis, param;
    ASTInitDeclarator initDecl; ASTTypeName typeName;
    ASTAbstractDeclarator abstr, abstrBis; ASTUnaryOperatorExpression unary;
    ASTDotAccessSuffix reset;
    ASTString str; int nbp = 0; String paramDecl = "";
    Vector printStatements = new Vector();

    if (factParams != null) {
      ASTParameterDeclarationList paramList;
      try {
        paramList = (ASTParameterDeclarationList)factParams.clone();
      } catch (CloneNotSupportedException e) { paramList = null; }

      nbp = paramList.jjtGetNumChildren();
      for (int i = 0; i < nbp; i++) {
        pname = "elt" + i; // the local parameter name
        param = paramList.jjtGetChild(i); // the ith ParameterDeclaration
        ptype = (ASTDeclarationSpecifiers)param.jjtGetChild(0);
	paramDecl = ((SimpleNode)param).getExpressionImage();
	declTor = param.jjtGetChild(1); // the ith parameter Declarator

        if (declTor.jjtGetChild(0) instanceof ASTDirectDeclarator) {
	  // non pointer type
          if (declTor.jjtGetChild(0).jjtGetNumChildren() > 1)
            CTDDTransformVisitor.errorMessage("No suffix for factory parameters",
					      ((SimpleNode)param).getLineNumber());
          isPrimitiveType = CTypeAux.isPrimitiveType(ptype);
         }
        else {
	  // pointer type
          if (declTor instanceof ASTAbstractDeclarator) {
            CTDDTransformVisitor.errorMessage("no abstract declarator in the parameters of a factory",
					      ((SimpleNode)param).getLineNumber());
          }
          isPrimitiveType = false;
        }

        arrlit = new
          ASTConstant(ADLCParserTreeConstants.JJTCONSTANT);
        arrlit.setName(String.valueOf(i));
        brack = new
          ASTArraySuffix(ADLCParserTreeConstants.JJTARRAYSUFFIX);
        ADLCGenericVisitor.parentChild(brack, arrlit, 0);
        prim = new PostfixExprBuilder
          ( ADLCGenericVisitor.makeName("params"),
            brack).generatedAST();

        if (isPrimitiveType) {
          // <ptype> elt<i> = *((<ptype> *)params[<i>]);
          unary = new
            ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
          try { unary.setProperty("operator", "*"); }
          catch(PropertyAccessException e) {}

          initDecl = new InitDeclaratorBuilder(pname, unary)
            .generatedAST();

          ASTParentheizedExpression par = new
            ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
          ADLCGenericVisitor.parentChild(unary, par, 0);

          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
          ADLCGenericVisitor.parentChild(par, castExpr, 0);

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          ADLCGenericVisitor.parentChild
            (abstr,
             new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
             0);

          ADLCGenericVisitor.parentChild(castExpr, prim, 1);

          ADLCGenericVisitor.parentChild(body,
                                         new DeclarationBuilder(ptype, initDecl)
                                         .generatedAST(),
                                         nbChildren++);
        }
        else { // pointer type
          // <ptype> elt<i> = (<ptype>)params[<i>];
          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          declTorBis = declTor;
          while (declTorBis.jjtGetChild(0) instanceof ASTPtrOperator) {
            ADLCGenericVisitor.parentChild(abstr,
					   declTorBis.jjtGetChild(0),
					   0);
            abstrBis = new
              ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	    abstrBis.setName("");
            ADLCGenericVisitor.parentChild(abstr, abstrBis, 1);
            abstr = abstrBis;
            declTorBis = declTorBis.jjtGetChild(1);
          }
	  // declTorBis.jjtGetChild(0) is now a DirectDeclarator
          ADLCGenericVisitor.parentChild(declTorBis.jjtGetChild(0),
                                         ADLCGenericVisitor.makeName(pname),
                                         0);

          ADLCGenericVisitor.parentChild(castExpr, prim, 1);

          initDecl = new InitDeclaratorBuilder(pname, castExpr)
            .generatedAST();

          ADLCGenericVisitor.parentChild(body,
                                         new DeclarationBuilder(ptype, initDecl)
                                         .generatedAST(),
                                         nbChildren++);
        }
    

        // ADL_printf("\t<ptype> <name> = %s", ADL_Print_<ptype>(elt_i));
        // ADL_printf("\tParameter int w = %s", ADL_Print_<type>(w));

        str = ADLCGenericVisitor.makeString("\\t" + paramDecl + " = %s");

        ASTPostfixExpression printArg = new PostfixExprBuilder
          ( CTypeAux.getPrintType(ptype.getType()),
            new ArgumentListBuilder(ADLCGenericVisitor.makeName(pname))
            .generatedAST())
          .generatedAST();

	printStatements.addElement(
            new StatementBuilder
            ( new PostfixExprBuilder
              ( "ADL_printf",
                new ArgumentListBuilder(str)
                .addArgument(printArg)
                .generatedAST()))
            .generatedAST());

      }
    }

    // if (ADL_debugMode)
    ASTIfStatement ifDebugMode = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    ADLCGenericVisitor.parentChild(ifDebugMode,
				   ADLCGenericVisitor.makeName("ADL_debugMode"),
				   0);
    ADLCGenericVisitor.parentChild(body, ifDebugMode, nbChildren++);
    ASTCompoundStatement thenBlk = new 
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(ifDebugMode, thenBlk, 1);

    int thenChildren = 0;
    for (Enumeration e = printStatements.elements(); e.hasMoreElements();)
      ADLCGenericVisitor.parentChild(thenBlk,
				     (Node)e.nextElement(),
				     thenChildren++);

    // if (ret = ADL__malloc(sizeof(<facttype>))) {

    ASTIfStatement ifRet = new
      ASTIfStatement(ADLCParserTreeConstants.JJTIFSTATEMENT);
    ADLCGenericVisitor.parentChild(body, ifRet, nbChildren++);

    ADLCGenericVisitor.parentChild
      ( ifRet,
        new AssignmentBuilder
        ( "ret", "=",
          new PostfixExprBuilder
          ( "ADL__malloc",
            new ArgumentListBuilder
            ( new PostfixExprBuilder
              ( "sizeof",
                new ArgumentListBuilder
                ( ADLCGenericVisitor.makeName(factTypeName))
                .generatedAST())
              .generatedAST())
            .generatedAST())
          .generatedAST())
        .generatedAST(),
        0);

    ASTCompoundStatement retGrp = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(ifRet, retGrp, 1);


    // val = provide_<featureName>(elt<0>, elt<1>, ...);
    // or, if void factory:
    // provide_<featureName>(elt<0>, elt<1>, ...);

    ArgumentListBuilder argb = new ArgumentListBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(ADLCGenericVisitor.makeName("elt" + i));
    }
    ASTPostfixExpression last = new PostfixExprBuilder
      ( ADLCGenericVisitor.makeName("provide_" + featureName),
        argb.generatedAST())
      .generatedAST();

    if (voidFactory) {
      ADLCGenericVisitor.parentChild
        ( retGrp,
          new StatementBuilder(last, true).generatedAST(),
          0);
    }
    else {
      ADLCGenericVisitor.parentChild
        ( retGrp,
          new StatementBuilder
          ( new AssignmentBuilder("val", "=", last))
          .generatedAST(),
          0);

      // memcpy(ret, &val, sizeof(<ptype>));

      ADLCGenericVisitor.parentChild
        ( retGrp,
          new StatementBuilder
          ( new PostfixExprBuilder
            ( "memcpy",
              new ArgumentListBuilder(ADLCGenericVisitor.makeName("ret"))
              .addArgument(ADLCGenericVisitor.makeName("&val"))
              .addArgument(new PostfixExprBuilder
                           ( "sizeof",
                             new ArgumentListBuilder(ADLCGenericVisitor.makeName(factTypeName))
                             .generatedAST())
                           .generatedAST())
              .generatedAST())
            .generatedAST(),
            true)
          .generatedAST(),
          1);
    }

    // return ret;

    ASTReturnStatement ret = new
      ASTReturnStatement(ADLCParserTreeConstants.JJTRETURNSTATEMENT);
    ADLCGenericVisitor.parentChild(body, ret, nbChildren++);
    ADLCGenericVisitor.parentChild(ret,
                                   ADLCGenericVisitor.makeName("ret"),
                                   0);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }


  /**
    * Creates the method "provide_&#60;factory name&#62;".
    * Example for factory make_bank:
<PRE>
|  static bank provide_make_bank(int maxAccts, int numAcctsActive, int initDeposit) {
|     &#60;factory body&#62;
|  }
</PRE>
    * <B>Note:</B> if the factory returns a primitive type, one must create a
    * FactoryVisitor to replace returned values in the factory body by values
    * of the corresponding Java wrapper class.
    * @see FactoryVisitor
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTExternalDeclaration makeProvide (ASTTDD_FactoryDefinition node) {

    if (debug)
      System.out.println("CFactoryBuilder makeProvide");

     ASTExternalDeclaration createdNode = new
       ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);

    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( (ASTDeclarationSpecifiers)node
				     .jjtGetChild(0)),
                                   0);

    ADLCGenericVisitor.parentChild(funcdef,
                                   node.jjtGetChild(2),
                                   2);

    try {
      ADLCGenericVisitor.parentChild(funcdef,
                                     (Node)((SimpleNode)node.jjtGetChild(1)).clone(),
                                     1);
    } catch(CloneNotSupportedException e) {}

    Node decl = funcdef.jjtGetChild(1);
    while(! (decl instanceof ASTFunctionDirectDeclarator))
      decl = decl.jjtGetChild(decl.jjtGetNumChildren()-1);
    ((SimpleNode)decl.jjtGetChild(0)).setName("provide_" + featureName);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }


  /**
    * Creates the method "factory_relinquish".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  static int F_bankTest_make_account_Relinquish(void **p, void *obj) {
|      int status = 1;
|      account elt0 = *((account *) p[0]);
|      int elt1 = *((int *) p[1]);
|
|      status = relinquish_make_account(elt0, elt1, *((account *)obj));
|      return status;
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTExternalDeclaration makeFactoryRelinquish() {

    if (debug)
      System.out.println("CFactoryBuilder makeFactoryRelinquish");

    ASTExternalDeclaration createdNode = new
      ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( ADLCGenericVisitor.makeBuiltinType
				     ("int")),
                                   0);
    ADLCGenericVisitor.parentChild
      (funcdef,
       new FunctionDeclaratorBuilder
       ( className + "_Relinquish",
         new ParameterDeclarationListBuilder
         ( new ParameterDeclarationBuilder
           ( ADLCGenericVisitor.makeBuiltinType("void"),
             new DeclaratorBuilder
             ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
               new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
               new DirectDeclaratorBuilder("params").generatedAST())
             .generatedAST()))
         .addParameter
         ( new ParameterDeclarationBuilder
           ( ADLCGenericVisitor.makeBuiltinType("void"),
             new DeclaratorBuilder
             ( new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
               "obj")
             .generatedAST()))
         .generatedAST())
       .generatedAST(),
       1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(funcdef, body, 2);
    int nbChildren = 0;

    // int status = 1;

    ADLCGenericVisitor.parentChild
      ( body,
        new DeclarationBuilder
        ( ADLCGenericVisitor.makeBuiltinType("int"),
          new InitDeclaratorListBuilder
          ( new InitDeclaratorBuilder
            ( new DeclaratorBuilder("status")
              .generatedAST(),
              "1"))
          .generatedAST())
        .generatedAST(),
        nbChildren++);
    
    
    ASTDeclarationSpecifiers ptype; String pname; ASTArraySuffix brack;
    ASTCastExpression castExpr; ASTPostfixExpression prim;
    ASTParentheizedExpression parenth; boolean isPrimitiveType;
    ASTConstant arrlit; ASTUnaryOperatorExpression unary;
    ASTInitDeclarator initDecl; ASTTypeName typeName;
    ASTAbstractDeclarator abstr, abstrBis;
    Node declTor, declTorBis, param;
    int nbp = 0;

    if (factParams != null) {
      nbp = factParams.jjtGetNumChildren();
      for (int i = 0; i < nbp; i++) {
        pname = "elt" + i; // the local parameter name
        param = factParams.jjtGetChild(i); // the ith ParameterDeclaration
        ptype = (ASTDeclarationSpecifiers)param.jjtGetChild(0);
	declTor = param.jjtGetChild(1); // the ith parameter Declarator

        if (declTor.jjtGetChild(0) instanceof ASTDirectDeclarator) {
	  // non pointer type
          isPrimitiveType = CTypeAux.isPrimitiveType(ptype);
         }
        else {
	  // pointer type
          isPrimitiveType = false;
        }

        arrlit = new
          ASTConstant(ADLCParserTreeConstants.JJTCONSTANT);
        arrlit.setName(String.valueOf(i));
        brack = new
          ASTArraySuffix(ADLCParserTreeConstants.JJTARRAYSUFFIX);
        ADLCGenericVisitor.parentChild(brack, arrlit, 0);
        prim = new PostfixExprBuilder
          ( ADLCGenericVisitor.makeName("params"),
            brack).generatedAST();

        if (isPrimitiveType) {
          // <ptype> elt<i> = *((<ptype> *)params[<i>]);

          unary = new
            ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
          try { unary.setProperty("operator", "*"); }
          catch (PropertyAccessException e) {}

          initDecl = new InitDeclaratorBuilder(pname, unary)
            .generatedAST();

          ASTParentheizedExpression par = new
            ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
          ADLCGenericVisitor.parentChild(unary, par, 0);

          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
          ADLCGenericVisitor.parentChild(par, castExpr, 0);

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          ADLCGenericVisitor.parentChild(abstr,
                                           ADLCGenericVisitor.makePtrOperator("PointerTo"),
                                           0);

          ADLCGenericVisitor.parentChild(castExpr, prim, 1);

          ADLCGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(ptype, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }
        else { // object type
          // <ptype> <id> = (<ptype>)params[<i>];

          castExpr = new
            ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);

          initDecl = new InitDeclaratorBuilder(pname, castExpr)
            .generatedAST();

          typeName = new
            ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
          ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	  abstr.setName("");
          ADLCGenericVisitor.parentChild(typeName, abstr, 1);

          declTorBis = declTor;
          while (declTorBis.jjtGetChild(0) instanceof ASTPtrOperator) {
            ADLCGenericVisitor.parentChild(abstr,
					   declTorBis.jjtGetChild(0),
					   0);
            abstrBis = new
              ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
	    abstrBis.setName("");
            ADLCGenericVisitor.parentChild(abstr, abstrBis, 1);
            abstr = abstrBis;
            declTorBis = declTorBis.jjtGetChild(1);
          }
	  // declTorBis.jjtGetChild(0) is now a DirectDeclarator
          ADLCGenericVisitor.parentChild(declTorBis.jjtGetChild(0),
                                         ADLCGenericVisitor.makeName(pname),
                                         0);

          ADLCGenericVisitor.parentChild(castExpr,
                                         new PostfixExprBuilder
                                         ( ADLCGenericVisitor.makeName("params"), prim)
                                         .generatedAST(),
                                         1);

          ADLCGenericVisitor.parentChild(body,
                                         new DeclarationBuilder(ptype, initDecl)
                                         .generatedAST(),
                                         nbChildren++);
        }
      }
    }

    // status = relinquish_<featureName>(elt0, elt1, ..., *((<factReturnType>*)obj));
    // or, if void factory:
    // status = relinquish_<featureName>(elt0, elt1, ...);

    ArgumentListBuilder argb = new ArgumentListBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(ADLCGenericVisitor.makeName("elt" + i));
    }
    if (! voidFactory) {
      unary = new
	ASTUnaryOperatorExpression(ADLCParserTreeConstants.JJTUNARYOPERATOREXPRESSION);
      try { unary.setProperty("operator", "*"); }
      catch(PropertyAccessException e) {}
      ASTParentheizedExpression par = new
	ASTParentheizedExpression(ADLCParserTreeConstants.JJTPARENTHEIZEDEXPRESSION);
      ADLCGenericVisitor.parentChild(unary, par, 0);

      castExpr = new
	ASTCastExpression(ADLCParserTreeConstants.JJTCASTEXPRESSION);
      typeName = new
	ASTTypeName(ADLCParserTreeConstants.JJTTYPENAME);
      ADLCGenericVisitor.parentChild(par, castExpr, 0);
      ADLCGenericVisitor.parentChild(castExpr, typeName, 0);
      ADLCGenericVisitor.parentChild(typeName,
				     ADLCGenericVisitor.makeType(factTypeName),
				     0);
      abstr = new
	ASTAbstractDeclarator(ADLCParserTreeConstants.JJTABSTRACTDECLARATOR);
      abstr.setName("");
      ADLCGenericVisitor.parentChild(typeName, abstr, 1);
      ADLCGenericVisitor.parentChild
	(abstr,
	 new ASTPtrOperator(ADLCParserTreeConstants.JJTPTROPERATOR),
	 0);
      
      ADLCGenericVisitor.parentChild(castExpr,
				     ADLCGenericVisitor.makeName("obj"),
				     1);
      argb.addArgument(unary);
    }
    
    ADLCGenericVisitor.parentChild
      ( body,
	new StatementBuilder
	( new AssignmentBuilder
	  ( "status", "=",
	    new PostfixExprBuilder
	    ( "relinquish_" + featureName,
	      argb.generatedAST())
	    .generatedAST()))
	.generatedAST(),
	nbChildren++);

    ASTReturnStatement ret = new
      ASTReturnStatement(ADLCParserTreeConstants.JJTRETURNSTATEMENT);
    ADLCGenericVisitor.parentChild(ret,
                                   ADLCGenericVisitor.makeName("status"),
                                   0);
    ADLCGenericVisitor.parentChild(body, ret, nbChildren++);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }


  /**
    * Creates the method "relinquish_&#60;factory name&#62;".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  static int relinquish_make_bank(int maxAccts, int numAcctsActive,
|                                       int initDeposit, bank b) {
|     &#60;relinquish body&#62;
|     return 1;
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTExternalDeclaration makeRelinquish() {

    if (debug)
      System.out.println("CFactoryBuilder makeRelinquish");

    ASTParameterDeclarationList paramsList = null;
    try { paramsList = (ASTParameterDeclarationList)factParams.clone(); }
    catch(CloneNotSupportedException e){}

    if (relParam != null) {
      ADLCGenericVisitor.parentChild(paramsList,
                                     relParam,
                                     paramsList.jjtGetNumChildren());
    }

    ASTExternalDeclaration createdNode = new
      ASTExternalDeclaration(ADLCParserTreeConstants.JJTEXTERNALDECLARATION);
    ASTFunctionDefinition funcdef = new
      ASTFunctionDefinition(ADLCParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCGenericVisitor.parentChild(funcdef,
                                   CTypeAux.makeStatic
				   ( ADLCGenericVisitor.makeBuiltinType
				     ("int")),
                                   0);
    ADLCGenericVisitor.parentChild
      (funcdef,
       new FunctionDeclaratorBuilder
       ( "relinquish_" + featureName,
         paramsList)
       .generatedAST(),
       1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCGenericVisitor.parentChild(funcdef, body, 2);

    // nothing to modify in the factory body
    ADLCGenericVisitor.parentChild(body, relBlock, 0);

    // return 1

    ASTConstant truelit = new
      ASTConstant(ADLCParserTreeConstants.JJTCONSTANT);
    truelit.setName("1");

    ASTReturnStatement ret = new
      ASTReturnStatement(ADLCParserTreeConstants.JJTRETURNSTATEMENT);
    ADLCGenericVisitor.parentChild(ret, truelit, 0);
    ADLCGenericVisitor.parentChild(body, ret, 1);

    ADLCGenericVisitor.parentChild(createdNode, funcdef, 0);

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: CFactoryBuilder.java,v $
// Revision 1.3  1998/08/05 16:42:07  mallet
// Fixed returned type for all declared objects from, e.g.,
// FunctionDefinition to ExternalDeclaration.
// Made all generated functions and constants static.
//
// Revision 1.2  1998/07/30 08:40:44  mallet
// Complete and fully operational version of the TDD -> C translator.
//
// Revision 1.1  1998/07/03 10:07:34  bareau
// Current (incomplete) version of the C/TDD translator.
//
//
// ***********************************************************************
