// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/c/CTypeAux.java,v $
// $Revision: 1.5 $
// $Date: 1998/08/05 16:41:03 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.c
// NAME:        CTypeAux
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Feb 6, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.c;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;

import java.util.StringTokenizer;

/**
 * Auxiliary class that gathers static methods to deal with types.
 * @author Cyrille Bareau
 */

public final class CTypeAux {

  /**
    * Returns "true" if the type parameter is a primitive type
    * (boolean or numeric).
    */
  public static boolean isPrimitiveType(ASTDeclarationSpecifiers type) {

    return type.getType().getRealType().isA(SymbolId.BUILTIN_TYPE);
  }


  /**
    * Returns "true" if the function parameter has no return type.
    */
  public static boolean isVoid(ASTDeclarationSpecifiers declSpec, Node declTor) {

    if (declTor.jjtGetChild(0) instanceof ASTPtrOperator) { // return pointer
      return false;
    }
    else return isVoid(declSpec);
  }


  /**
    * Returns "true" if the type parameter is the void type.
    */
  public static boolean isVoid(ASTDeclarationSpecifiers type) {

    if (type.jjtGetChild(0) instanceof ASTBuiltinTypeSpecifier)
      return ((SimpleNode)type.jjtGetChild(0)).getName().equals("void");
    else return false;
  }


  /**
    * Returns the default value for the parameter type
    * (0 for builtin types, null for objects, etc.)
    */
  public static Node getDefaultValue(LType type, boolean staticInitializer) {

    if (type.isA(SymbolId.STRUCT | SymbolId.UNION))
      throw new Error("Cannot handle struct/union in such expression!");
    else if (type.isA(SymbolId.ARRAY)) {
      if (staticInitializer) {
	ASTInitializer initArray = new
	  ASTInitializer(ADLCParserTreeConstants.JJTINITIALIZER);
	ADLCGenericVisitor.parentChild(initArray,
				       ADLCGenericVisitor.makeConstant("0"),
				       0);
	return initArray;
      }
      else
	throw new Error("Cannot handle array in such expression!");
    }
    else
      return ADLCGenericVisitor.makeConstant("0");
  }


  public static String getPrintType(LType type) {

    // by using getRealType here we shortcut any user defined type
    type = type.getRealType();
    if (CBuiltinType.isValidBuiltinTypeName(type.getName())) {
      String ret = "ADL_Print_";
      StringTokenizer st = new StringTokenizer(type.getName());
      String tok;
      while (st.hasMoreTokens()) {
        tok = st.nextToken();
        ret += Character.toUpperCase(tok.charAt(0)) +
               tok.substring(1);
      }
      return ret;
    }
    if ((type instanceof LPointer) &&
        ((CModifier)type.getModifier()).isConstant() &&
        ((LPointer)type).getPointedToType().equals(CBuiltinType.charT)) {
      return "ADL_Print_CharP";
    }
    return "ADL_Print_Pointer";
  }

  public static String getLiteralType(LType type) {

    // by using getRealType here we shortcut any user defined type
    type = type.getRealType();
    if (type instanceof LPointer)
      return "voidT";
    else {
      String typeName = type.getName();
      if (CBuiltinType.isValidBuiltinTypeName(typeName)) {
	StringTokenizer st = new StringTokenizer(typeName);
	String ret = st.nextToken();
	while (st.hasMoreTokens()) {
	  ret += "_" + st.nextToken();
	}
	ret += "T";
	return ret;
      }
      else
	return "intT";
    }
  }

  public static ASTDeclarationSpecifiers makeStatic(ASTDeclarationSpecifiers node) {
    // offset all children by one
    for (int j = node.jjtGetNumChildren(); j > 0; j--) {
      node.jjtAddChild(node.jjtGetChild(j - 1), j);
    }

    // attach a StorageStructSpecifier as first child
    ASTStorageStructSpecifier storage = new
      ASTStorageStructSpecifier(ADLCParserTreeConstants.JJTSTORAGESTRUCTSPECIFIER);
    storage.setName("static");
    ADLCGenericVisitor.parentChild(node, storage, 0);

    return node;
  }
}

// ***********************************************************************
// History Log:
// $Log: CTypeAux.java,v $
// Revision 1.5  1998/08/05 16:41:03  mallet
// Added a new method makeStatic() to make a DeclarationSpecifiers
// contain a StorageStructSpecifier indicating that the declared object
// is static.
//
// Revision 1.4  1998/07/28 16:05:53  mallet
// Fixed method isBuiltinType: use the information in the symbol
// table rather than testing the class of the type of the node. Also use
// the real type to skip any typedef.
// Changed method getPrintType to use the real type instead of the type
// (skipping typedefs - this would have to be changed if we implement the
// user defined print functions).
// Added a new method getLiteralType to determine the appropriate
// ADL_LiteralType for an ADLLiteralSet (int -> intT, short -> shortT, ...).
//
// Revision 1.3  1998/06/30 16:04:22  mallet
// Fixed bug in method getDefaultValue(): now handles more cases to
// determine the appropriate initialization value. A new argument has
// been added to indicate whether the returned value appears in a static
// initializer or in an expression. This makes a difference for array for
// instance since they cannot be initialized to {0} when this value is
// used in a conditional expression.
//
// Revision 1.2  1998/06/30 08:51:51  mallet
// Fixed bug in method getDefaultValue(): return 0 for all types except
// struct and union. Those types cannot currently appear in assertions
// (except through pointers or to access a field) because they cannot be
// initialized and printed (if required, that is when they appear in a
// member of a logical 'and' or 'or').
//
// Revision 1.1  1998/06/23 13:39:45  bareau
// First version of the ADL/C translator (ADL part only).
//
// Revision 1.1  1998/06/23 13:27:11  bareau
// First version of the ADL/C translator (ADL part only).
//
// Revision 1.6  1998/04/28 12:39:26  cinard
// Updated code to take into account replacement of CSymbolTableMgr,
// CSymbolFactory and CBuiltinType by CSymbolTableMgr, CPPSymbolFactory
// and CBuiltinType.
//
// Revision 1.5  1998/03/16 10:14:46  rami
// Added getDefaultValue(LType) returning node for default value of a type.
//
// Revision 1.4  1998/02/27 13:07:31  bareau
// Updated copyright notice.
//
// Revision 1.3  1998/02/26 17:50:00  bareau
// ADLT2 TDD translator, first release. Version OK for the bank example.
//
// Revision 1.2  1998/02/12 16:03:43  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/12 13:48:22  bareau
// Classes used in the TDD translation. Added because of code freeze: this is a
// pre-pre-version (it does not even compile yet).
//
// ***********************************************************************
