// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1997-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions to use, copy, modify and distribute this software are
// governed by the terms and conditions set forth in the file COPYRIGHT,
// located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/cpp/ADLCPPTransformVisitor.java,v $
// $Revision: 1.32 $
// $Date: 1998/10/15 18:14:10 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.cpp
// NAME:        ADLCPPTransformVisitor
//
// HISTORY:
//      Rami Khemakhem, TOG-RI - Dec 08, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.cpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.*;


/**
 * Transformation visitor.  Must translate an ADL AST into a C++ AST.
 * <P> For an overview of the algorithms used by this transformation,
 * see the ADL2.0 Translation System Design Specification, Chapter 5
 * (Code Generation).
 * <p>The translating starts by calling the <a href="#translate">translate()</a>
 * method passing the top of AST to translate as parameter.
 * @see CPPGeneric
 * @see ADLCPPParserVisitor
 */

public class ADLCPPTransformVisitor extends CPPGeneric {
  
  class BoolContext {

    private String tmp;
    private boolean b;

    BoolContext(String t, boolean bb) {
      tmp = t;
      b = bb;
    }

    String getString() {
      return tmp;
    }

    boolean getBool() {
      return b;
    }
  }

  /* handles */

  /** Stack of BoolContexts for lazy evaluation of logical operations */
  private Stack boolContextStack;

  /* Flag for debug mode */
  private final static boolean debug = false;

  /* handles */
  
  /** Handle for the global prolog block */
  private ASTCompoundStatement globalProlog;

  /** Handle for the global epilog block */
  private ASTCompoundStatement globalEpilog;

  /** Handle for the local prolog block */
  private ASTCompoundStatement localProlog;

  /** Handle for the local epilog block */
  private ASTCompoundStatement localEpilog;

  /** Handle for inline declarations (vector) */
  private Vector inlineDeclaration;

  /** Handle for tested functions/constructors (vector).
   * Used to add function/constructor definitions to the 
   * transfomed ASTADLClass node.
   */
  private Vector bhvDeclaration;
  
  /** Handle for the declaration of temporaries for
   * call-state expressions (vector) */
  private Vector preDeclaration;
  
  /** Handle for the declaration of temporaries for
    * call-state expressions (vector) */
  private Vector saveBlock;

  /** Handle for "pre-assertions" : statements that precede assertions
    * (local var declaration for instance). */
  private Vector preAssertion;

  /** Handle for super classes
    * (used to add super classes include files ). */
  private Vector Classes;

  /** Tags of the current assertion (vector) */
  private String[] assTags;

  /** Handle on the declaration of the inline that corresponds
    * to the currently encountered inline call */
  private ASTADLInlineDecl currentInline;

  /** Handle on the parameters of the declaration of the
    * currently encountered inline call */
  private Node inlineParams;

  /** Handle on the actual parameters of the inline call that is
    * currently being inlined */
  private ASTArgumentList inlineActualParams;

  /** Auxiliary class for code generation */
  private ACOauxiliary ACO;

  /** Index for call-state tmp variables */
  private int savIndex;

  /** Index for preAssertions */ 
  private int preAssIndex;

  /** Flags set when visiting specific nodes. */
  private boolean
  isInAssertion,isInCallState,isInTry,isInConstructor,isInQuantified,isInBinding;

  /** Flag set when the currently visited expression can be decomposed
    * for the TET debug mode.
    * Must always be equal to 
    * <PRE>isInAssertion && 
    *   !(isInCallState || isInTry || isInQuantified || isInBinding)</PRE>
    */
  private boolean isDecomposable;

  /** Flag set when visiting the substituted inline body
    * while transforming an inline call */
  private boolean inlining;

  private int nbv = 1;


  /**
    * Generates statement for a debug mode message
    * @param msg the string of the message to be generated.
    * @param id the temporary variable declared for preAssertion.
    * @return a statement node that is a postfix expression.
    * @see #preAssertion
    */
  private ASTStatement makeDebugModeMsg(String msg, Node id) {
    
    ASTBinaryArithExpr shift;
    ASTDotAccessSuffix reset;
    ASTString str;

    shift = new
      ASTBinaryArithExpr(ADLCPPParserTreeConstants.JJTBINARYARITHEXPR);
    shift.addOp("<<");shift.addOp("<<");
    reset = new DotAccessSuffixBuilder
      ("reset", new ArgumentListBuilder().generatedAST()).generatedAST();
    parentChild( shift,
		 new PostfixExprBuilder( makeIdExpr("adlstr"),
					 reset
					 ).generatedAST(),
		 0);

    parentChild(shift, makeString(msg), 1);
    parentChild(shift, id, 2);
    
    ASTStatement createdNode = new StatementBuilder
      ( new PostfixExprBuilder
	( "testContext",
	  new RefAccessSuffixBuilder
	  ("infoline", new ArgumentListBuilder(shift).generatedAST()
	    ).generatedAST()
	  ).generatedAST()
	).generatedAST();

    return createdNode;
  }


  /**
    * Generates a declaration of a variable, either as 
    * a preDeclaration (call-state case) or as a preAssertion.
    * <BR>In the preAssertion case, the string parameter (that represents
    * the expression) is stored as a property of the ASTDeclaration.
    * @param saveblk "is a call-state".
    * @param type type of the temporary variable to generate.
    * @param theString token string of the rhs expression, for a preAssertion.
    * @return the name of the generated temporary variable.
    * @see #preDeclaration
    * @see #preAssertion
    */
  private String makeSaveDecl(boolean saveblk, LType type, String theString) {

    if (debug)
      System.out.println("\tmakeSaveDecl : " + saveblk + " " + type.getName() + " " + theString);

    String lhs = (saveblk ?
                  ("__ADL_savTmp_" + (savIndex++)) :
                  ("__ADL_tmp_" + (preAssIndex++)));

    // TBR: <type> <lhs> = <default value>;
    // <type> <lhs>;

    ASTQualifiedType createdType = makeQualifiedType(type.getFullName());
    createdType.setQualifiedType(type);
    ASTDeclaration VarDecl = 
      new DeclarationBuilder
      ( createdType, 
	new DirectDeclaratorBuilder( makeQualifiedId(lhs)).generatedAST()
	).generatedAST();

    if (saveblk) {
      preDeclaration.addElement(VarDecl);
    }
    else {
      try {
        VarDecl.setProperty("tokenString",theString);
      }
      catch(PropertyAccessException e) {}
      preAssertion.addElement(VarDecl);
    }
    return lhs;
  }


  /**
    * Generates a temporary declaration of a variable, initialized with 
    * a passed expression, as a saveBlock (call-state case).
    * 
    * @param type type of the tmp var to generate.
    * @param expr initialization expression.
    * @return the string name of the generated tmp variable.
    * @see #saveBlock
    * @see #preAssertion
    */
  private String makeSaveBlock(LType type, Node expr) {
    
    if (debug)
      System.out.println("\tmakeSaveBlock : " + type +
                         " " + expr);

    // <type> <lhs> = <expr>;

    String lhs = makeSaveDecl(true, type, "");
    ASTStatement stmt = new StatementBuilder
      ( new AssignmentBuilder(lhs, "=", expr).generatedAST() 
	).generatedAST();
    saveBlock.addElement(stmt);
    return lhs;
  }


  /**
    * Generates a temporary declaration of a variable, initialized with 
    * a passed expression, as a preAssertion.
    * <BR>The string parameter (that represents the expression) 
    * is stored as a property of the ASTDeclaration.
    * @param type type of the tmp var to generate.
    * @param expr initialization expression.
    * @param theString token string of the rhs expression, for a preAssertion.
    * @return the name of the generated tmp variable.
    * @see #preAssertion
    */
  private String makePreAssertion(LType type, Node expr, String theString) {

    if (debug)
      System.out.println("\tmakePreAssertion : " + type.getName() +
                         " " + expr + " " + theString);

    String lhs="__ADL_tmp_" + (preAssIndex++);
    ASTDeclaration varDecl;

    // <type> <lhs> = <expr>;

    ASTQualifiedType varDeclType = makeQualifiedType(type.getFullName());
    varDeclType.setQualifiedType(type);
    if (boolContextStack.empty()) {
      varDecl = new DeclarationBuilder
	( varDeclType,
	  new InitDeclaratorBuilder(lhs, expr).generatedAST()
	  ).generatedAST();
    }
    else {
      // when there is 1 boolContext <"foo", bool> in the stack, the expression
      // is guarded by a conditional expression:
      // foo ? <expr> : <default value> // if bool is true ("true branch")
      // foo ? <default value> : <expr> // otherwise ("else branch")
      // the <default value> is the expression's type default value (0 for numeric, etc)

      // when several boolContext are stacked, conditional expressions
      // must be nested (from the bottom to the top of the stack)

      ASTParenthExpr parenth = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      ASTCondExpr cond = new
        ASTCondExpr(ADLCPPParserTreeConstants.JJTCONDEXPR);
      parentChild(parenth, cond, 0);
      for (Enumeration enum = boolContextStack.elements(); enum.hasMoreElements(); ) {
        BoolContext bc = (BoolContext)enum.nextElement();
        parentChild(cond, makeIdExpr(bc.getString()), 0);
        parentChild(cond,
                    // default value for un-evaluated branch
                    CPPTypeAux.getDefaultValue(type, false),
                    (bc.getBool() ? 2 : 1));
        if (enum.hasMoreElements()) {
          ASTParenthExpr parenth2 = new
            ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
          ASTCondExpr cond2 = new
            ASTCondExpr(ADLCPPParserTreeConstants.JJTCONDEXPR);
          parentChild(parenth2, cond2, 0);
          parentChild(cond, parenth2, (bc.getBool() ? 1 : 2));
          cond = cond2;
        }
        else {
          parentChild(cond, expr, (bc.getBool() ? 1 : 2));
        }
      }
      varDecl = new DeclarationBuilder
	( varDeclType,
	  new InitDeclaratorBuilder(lhs, parenth).generatedAST()
	  ).generatedAST();
    }

    try {
      varDecl.setProperty("tokenString", theString);
    }
    catch(PropertyAccessException e) {}
    
    preAssertion.addElement(varDecl);

    if (debug) {
      System.out.println("\tAdding the preAssertion:");
      ((SimpleNode)varDecl).dump(System.out, 100);
      System.out.println("\tmakePreAssertion returns: " + lhs);
    }
    return lhs;
  }

  
  /**
   * Generates a try/catch block that corresponds to an ADLtryStmt.
   * The try (resp. catch) block is composed of an assignment of a variable
   * passed as parameter, initialized by the expression found in the
   * ADLtryStmt's try (resp. catch) block.
   * <BR>Creates either a preDeclaration (call-state case) or a preAssertion.
   * @param saveblk "is a call-state"
   * @param node ADL try/catch initialization expression
   * @param lhs the name of the variable to be assigned to
   * @return the string name of the generated tmp variable.
   * @see #saveBlock
   * @see #preAssertion
   */
  private String makeTryStatement(boolean saveblk, ASTADLTryStmt node, String lhs) throws ParseException {
    
    if (debug)
      System.out.println("makeTryStatement : " + saveblk + " " + node + " " + lhs);

    // <node.getType> <tmpName>;
    String tmpName = makeSaveDecl(saveblk, node.getType(), lhs);

    // try { <tmpName> = <node.try_expr>; }
    // ( catch(<node.catch>) { <tmpName> = <node.catch_expr>; } )+

    ASTTryNode tryStmt = new
      ASTTryNode(ADLCPPParserTreeConstants.JJTTRYNODE);

    parentChild( tryStmt,
		 new CompoundStatementBuilder
		 ( new StatementBuilder
		   ( new AssignmentBuilder
		     ( tmpName,
		       "=",
		       (Node)node.jjtGetChild(0).jjtAccept(this, null)
		       ).generatedAST()
		     ).generatedAST()
		   ).generatedAST(),
		 0);

    ASTExceptionDeclaration exc;
    ASTCatchNode catchNode;
    for (int i = 1; i < node.jjtGetNumChildren()-1; i += 2) {
      // Node = ASTExceptionDeclaration, e.g. "catch(...)" 

      catchNode = new
	ASTCatchNode(ADLCPPParserTreeConstants.JJTCATCHNODE);
      parentChild(tryStmt, catchNode, (i+1)/2);

      exc = (ASTExceptionDeclaration)node.jjtGetChild(i);
      parentChild(catchNode, exc, 0);

      parentChild( catchNode, 
		   new CompoundStatementBuilder
		   ( new StatementBuilder
		     ( new AssignmentBuilder
		       ( tmpName, 
			 "=",
			 (Node)node.jjtGetChild(i+1).jjtAccept(this, null)
			 ).generatedAST()
		       ).generatedAST()
		     ).generatedAST(),
		   1);
    }
    if (saveblk) {
      saveBlock.addElement(tryStmt);
    }
    else {
      preAssertion.addElement(tryStmt);
    }
    return tmpName;
  }


  /**
   * Creates a try/catch block to trap thrown exceptions during the
   * evaluation of call-state expressions or assertions.
   *
   * The created block for a call-state block looks like:
<PRE>
|  try {
|    // call-state expressions block
|  } catch (...) {
|    testContext->infoline("Unexpected exception caught in a call-state expression within assertion");
|    testContext->result(ADL_UNRESOLVED);
|    skipAssertions = true;
|    testContext->skipAssertions = true;
|  }
</PRE>
    *
    * @param triedBlock the block of statement to be included in the
    *                   try/catch block
    * @param inCallState whether this exception catch block is intended
    *                    for a call-state expressions block or not
    * @return the try/catch block
    **/
  public ASTTryNode makeExceptionCatchBlock(ASTCompoundStatement triedBlock,
					    boolean inCallState) {
    String message, resultMethod;
    if (inCallState) {
      message = "in a call-state expression within assertion";
      resultMethod = "result";
    } else {
      message = "in assertion";
      resultMethod = "assertionResult";
    }

    // the try statement
    ASTTryNode tryStmt = new
      ASTTryNode(ADLCPPParserTreeConstants.JJTTRYNODE);

    parentChild(tryStmt, triedBlock, 0);

    ASTCatchNode catchFinalBlk = new ASTCatchNode(ADLCPPParserTreeConstants.JJTCATCHNODE);
    parentChild(tryStmt, catchFinalBlk, 1);

    ASTExceptionDeclaration finalExc = new ASTExceptionDeclaration(ADLCPPParserTreeConstants.JJTEXCEPTIONDECLARATION);
    parentChild(catchFinalBlk, finalExc, 0);

    // catch block
    ASTCompoundStatement catchBlock = new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    parentChild(catchFinalBlk, catchBlock, 1);

    // testContext->infoline("Unexpected exception caught  " + <message>);
    // testContext-><resultMethod>(ADL_UNRESOLVED);
    ASTStatement stmt = new StatementBuilder
      ( new PostfixExprBuilder
	( ADLCPPGenericVisitor.makeIdExpr("testContext"),
	  new RefAccessSuffixBuilder
	  ( "infoline",
	    new ArgumentListBuilder
	    ( ADLCPPGenericVisitor.makeString
	      ("Unexpected exception caught " + message)
	      ).generatedAST()
	    ).generatedAST()
	  ).generatedAST()
	).generatedAST();
    parentChild(catchBlock, stmt, 0);
    
    stmt = new StatementBuilder
      ( new PostfixExprBuilder
	( ADLCPPGenericVisitor.makeIdExpr("testContext"),
	  new RefAccessSuffixBuilder
	  ( resultMethod,
	    new ArgumentListBuilder
	    (ADLCPPGenericVisitor.makeIdExpr("ADL_UNRESOLVED")
	      ).generatedAST()
	    ).generatedAST()
	  ).generatedAST()
	).generatedAST();
    parentChild(catchBlock, stmt, 1);

    if (inCallState) {
      // skipAssertions = true;
      // testContext.skipAssertions = true;
      stmt = new StatementBuilder
	( new AssignmentBuilder
	  ( "skipAssertions", 
	    "=", 
	    ADLCPPGenericVisitor.makeBoolean("true")
	    ).generatedAST()
	  ).generatedAST();
      parentChild(catchBlock, stmt, 2);

      stmt = new StatementBuilder
	( new AssignmentBuilder
	  ( new PostfixExprBuilder
	    ( ADLCPPGenericVisitor.makeIdExpr("testContext"),
	      new RefAccessSuffixBuilder("skipAssertions").generatedAST()
	      ).generatedAST(),
	    "=", 
	    ADLCPPGenericVisitor.makeBoolean("true")
	    ).generatedAST()
	  ).generatedAST();
      parentChild(catchBlock, stmt, 3);
    }
    
    return tryStmt;
  }


  /**
    * Visit pattern for a typed node.
    * <P><B>Expression decomposition algorithm</B>
    * <BR><I>(This algorithm is used for all expression nodes)</I>
    * <P><UL><LI>Visit the children (sub-expressions).
    * <LI><B>IF</B> the expression is to be decomposed (i.e. it is contained in
    * an assertion expression and it is neither inside a call-state expression,
    * a try/catch or a quantified assertion), call makeSaveAssignment to
    * generate a local variable declaration, i.e. to generate the declaration
    * of a local temporary variable that will be defined by the type and the
    * value of the expression.
    * <LI><B>ELSE</B> nothing to do: just return the visited node. </UL>
    * <P><I>For a more detailed overview of the expression declaration algorithm,
    * see the ADL2.0 Translation System Design Specification, Chapter 5
    * (Code Generation), paragraph 5.2.1.3.</I>
    * @param node the ASTCondExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  private Object decompose(Node node, Object data) throws ParseException {

    if (debug) {
      System.out.println("Decomposing node AST"+node+" ...");
      ((SimpleNode)node).dump(System.out, 100);
    }

    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, null), i);
    }
    if (isDecomposable) {
      return makeIdExpr
	(makePreAssertion(((Typed)node).getType(),
			  node,
			  ((ADLNode) node).getExpressionImage()));
    }
    else {
      return node;
    }
  }


  /** 
   * Launch the visitor on each child of the current node
   * @param n the current node
   */
  private void childrenAccept(Node n) {

    if (debug) System.out.println("\tchildrenAccept: "+ n);

    for (int i = 0; i <  n.jjtGetNumChildren(); i++) {
      try {
      parentChild(n, (Node)n.jjtGetChild(i).jjtAccept(this, null), i);
      }
      catch(ParseException e) {}
    }
  }


  /**
   * Starts translating an AST.
   * Initializes needed handles.
   * @param ast the root node of the AST ADL to translate.
   * @exception ParseException thrown by parser.
   * @return the root node of the transformed tree
   * @see #visit(ORG.opengroup.adl.parsers.adlcpp.ASTTranslationUnit, java.lang.Object)
   */
  public Node translate(Node ast) throws ParseException {
  
    globalProlog = null;
    globalEpilog = null;
    localProlog = null;
    localEpilog = null;
    inlineDeclaration = new Vector();
    bhvDeclaration = new Vector();
    Classes = new Vector();

    return (Node)ast.jjtAccept(this, null);
  }

  /** 
   * Constructor.
   * <BR>Creates an ACOauxiliary that will be associated with the 
   * currently transformed adl class.
   * @see ACOauxiliary
   */
  public ADLCPPTransformVisitor() {
    
    if (debug) System.out.println("ADLCPPTransformVisitor Constructor");
    ACO = new ACOauxiliary(this);
  }


  /**
   * Visit pattern for a node of type ASTTranslationUnit
   *
   * <P><B>Algorithm:</B>
   * creates a node ASTTranslationUnit with children:<UL>
   * <LI> the external declaration children of the input node
   * <LI> the visit to the node's ASTADLClass child
   * </UL>
   * Sets the name of the TranslationUnit to the AdlClassName.
   * @param node the ASTTranslationUnit to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node (ASTTranslationUnit: the root of the
   * transformed AST).
   */
  public Object visit(ASTTranslationUnit node, Object data) throws ParseException{

    if (debug) {
      System.out.println("\n\t************************************");
      System.out.println("\t*  TRANSLATING AST/ADL to AST/C++  *");
      System.out.println("\t************************************\n");
      System.out.println("Visiting AST" + node + " ...");
    }

    // save former stm context
    Object savedStmContext = SimpleNode.stm.getContext();

    // set stm context for this ast
    try {
      Object stmContext = node.getProperty("stmContext");
      SimpleNode.stm.setContext(stmContext);
    }
    catch ( PropertyAccessException e ) { }

    Node child,child0;
    int insertChild = 0;
    ASTTranslationUnit createdNode = new 
      ASTTranslationUnit(ADLCPPParserTreeConstants.JJTTRANSLATIONUNIT);
    createdNode.setNodeExtension("cpp");
    int nbChildren = node.jjtGetNumChildren();

    child0 = node.jjtGetChild(0);
    int insertInc = child0.jjtGetNumChildren();
    if (child0 instanceof ASTIncludeFileList) {
      ASTIncludeFileDeclaration inc1 = new
	ASTIncludeFileDeclaration
	(ADLCPPParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      inc1.setFileName("\"ADLResultCodes.h\"");
      parentChild(child0, inc1, insertInc++);
      
      ASTIncludeFileDeclaration inc2 = new
	ASTIncludeFileDeclaration
	(ADLCPPParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      inc2.setFileName("\"ADLString.hh\"");
      parentChild(child0, inc2, insertInc++);
      
      ASTIncludeFileDeclaration inc3 = new
	ASTIncludeFileDeclaration
	(ADLCPPParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      inc3.setFileName("\"ADLObjectACO.hh\"");
      parentChild(child0, inc3, insertInc++);

      parentChild(createdNode, child0, insertChild++);
    }
    
    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);

      if (child instanceof ASTADLClass) {
	for (Enumeration e=((Vector)(child.jjtAccept(this,data))).elements();
	     e.hasMoreElements();) {
	  parentChild(createdNode, (Node)e.nextElement(), insertChild++);
	}
      }
    }
    createdNode.setName(ACO.getAdlClassName()+"ACO");

    for (int i=0; i < Classes.size(); i++) {

      ASTIncludeFileDeclaration incFile = new
	ASTIncludeFileDeclaration
	(ADLCPPParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      incFile.setFileName("\"" +
			  ((LClass)Classes.elementAt(i)).getName() + 
			  "ACO.cpp\"");
      parentChild(child0, incFile, insertInc++);
    } 
   
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
    }

    // restore former stm context
    SimpleNode.stm.setContext(savedStmContext);

    return createdNode;
  }


  /**
   * Visit pattern for a node of type ASTADLInlineDecl
   * <BR>Should not be visited!
   * <BR>For the processing of inline calls, see the algorithm
   * for primary expressions:
   * @see #visit(ORG.opengroup.parsers.adlcpp.ASTPrimaryExpr,java.lang.Object)
   * @param node the ASTADLInlineDecl to visit.
   * @param data unused.
   * @return an error
   */
  public Object visit(ASTADLInlineDecl node, Object data)  {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    // not visited: cf ASTADLClass
    return new InternalError();
  }


  /**OK
    * Visit pattern for a node of type ASTADLClass
    * <P><B>Algorithm:</B><UL>
    * <LI>Stores information locally (possible inline, global prolog/epilog)
    * or into the associated ACOauxiliary (class name, possible superclass).
    * <LI> Visits the ASTADLBhvDecl child.
    * <LI> Creates an ASTClassSpecifier node with 2 children:<OL>
    * <LI> the inherited class name in ASTBaseSpecifier node
    * <LI> and an ASTClassMemberList node parent of:<OL>
    * <LI> the class constructor
    * <LI> the declarations of tested functions/constructors.
    * </OL></UL>
    * <LI> Return a vector with the visited ASTADLBhvDecl child and the ASTClassSpecifier.
    * </OL></UL>
    * @param node the ASTADLClass to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return a vector (ClassDeclaration).
    */
  public Object visit(ASTADLClass node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child, child0;
    Vector ClassDeclaration = new Vector();

    ASTDeclaration createdNode = new
      ASTDeclaration(ADLCPPParserTreeConstants.JJTDECLARATION);
    ASTDeclarationSpecifiers declNode = new
      ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
    ASTClassSpecifier classNode = new
      ASTClassSpecifier(ADLCPPParserTreeConstants.JJTCLASSSPECIFIER);
    parentChild(createdNode, declNode, 0);
    parentChild(declNode, classNode, 0);

    int nbChildren = node.jjtGetNumChildren();

    ACO.setAdlClassName(node.getName());
    classNode.setName(node.getName() + "ACO");
    classNode.setClassKind("class");

    int insertChild = 2;
    child0 = node.jjtGetChild(0);
    if (child0 instanceof ASTADLSuperClass) {
      Classes = ((ASTADLSuperClass)child0).getSuperClasses();
      for (int i=0; i < Classes.size(); i++) {
	ACO.setSuperClasses(((LClass)Classes.elementAt(i)).getName()+"ACO");
      }
      // Create an ASTBaseClause node if there is more than 1 superClass
      if (Classes.size() > 1) {
	ASTBaseClause superClasses = new
	  ASTBaseClause(ADLCPPParserTreeConstants.JJTBASECLAUSE);
	for (int i=0; i < Classes.size(); i++) {
	  ASTBaseSpecifier superClass = new
	    ASTBaseSpecifier(ADLCPPParserTreeConstants.JJTBASESPECIFIER);
	  superClass.setName(((LClass)Classes.elementAt(i)).getName() + "ACO");
	  superClass.setVirtualParent("true");
	  parentChild(superClasses,superClass,i);
	}
	parentChild(classNode, superClasses, 0);
      }
      else {
	ASTBaseSpecifier superClasses = new
	  ASTBaseSpecifier(ADLCPPParserTreeConstants.JJTBASESPECIFIER);
	superClasses.setName(((LClass)Classes.elementAt(0)).getName() + "ACO");
	superClasses.setVirtualParent("true");
	parentChild(classNode, superClasses, 0);
      }
    }
    else {
      ASTBaseSpecifier superClasses = new
	ASTBaseSpecifier(ADLCPPParserTreeConstants.JJTBASESPECIFIER);
      superClasses.setName("ADLObjectACO"); // By default
      parentChild(classNode, superClasses, 0);
    }

    ASTClassMemberList ClassMembers = new
      ASTClassMemberList(ADLCPPParserTreeConstants.JJTCLASSMEMBERLIST);
    ASTAccessSpecifier accessClass = new
      ASTAccessSpecifier(ADLCPPParserTreeConstants.JJTACCESSSPECIFIER);
    accessClass.setAccessSpecifierName("public");
    parentChild(ClassMembers, accessClass, 0);
    parentChild(ClassMembers, ACO.makeACO_Constructor(), 1);

    parentChild(classNode, ClassMembers, 1);

    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);

      if (child instanceof ASTADLProlog) {
        globalProlog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADLEpilog) {
        globalEpilog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADLInlineDecl) {
        inlineDeclaration.addElement(child);
      }
    }
    // now the loop for behavior declarations
    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADLBhvDecl) {
	ClassDeclaration.addElement((Node)child.jjtAccept(this, data));
      }
    }
    for (Enumeration e=bhvDeclaration.elements();
	 e.hasMoreElements();) {
      parentChild(ClassMembers, (Node)e.nextElement(), insertChild++);
    }

    ClassDeclaration.insertElementAt(createdNode,0);
    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + ClassDeclaration);
    return ClassDeclaration;
  }


  /**
    * Visit pattern for a node of type ASTADLBhvDecl
    * <P><B>Algorithm:</B><UL>
    * <LI>Creates a node of type ASTFunctionDefinition, when testing a function
    * that corresponds to the current behavior description.
    * <LI>Creates a node of type ASTCtorDefinition, when testing a constructor
    * that corresponds to the current behavior description.
    * <LI>Stores local information into the associated ACOauxiliary (function
    * name, return type, parameters, isConstructor flag)
    * or locally (local prolog/epilog). Sets the flag isInConstructor.
    * <LI> Visits the ASTADLBhvSpec child.
    * </UL>
    * @param node the ASTADLBhvDecl to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node (ASTFunctionDefinition) or (ASTCtorDefinition).
    */
  public Object visit(ASTADLBhvDecl node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    // Initialize local prolog and epilog
    localProlog = null;
    localEpilog = null;

    ASTFunctionDefinition  createdNode = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);
    Node child = null;
    int nbChildren = node.jjtGetNumChildren();

    preDeclaration = new Vector();
    saveBlock = new Vector();
    
    int insertChild = 0;
    String AdlClassName = ACO.getAdlClassName();
    ACO.initializeFct();
    // Store function return type
    ACO.setFctReturnType(node.ResultFct.getReturnType().getName());
    ACO.setFctTypeDefaultValue(node.ResultFct.getReturnType());
    // Store Function name in ACOauxiliary
    ACO.setFctName(node.ResultFct.getName());
    // Store Function parameters
    ACO.setFctParams(node.ResultFct.getParameters());      
    
    child = node.jjtGetChild(0);
    if (child instanceof ASTCtorDeclarator) { // constructor
      if (debug) 
	System.out.println("\nTesting the constructor "+ACO.getFctName());
      isInConstructor = true;
      ACO.setIsConstructor(isInConstructor);
      ACO.setFctReturnType(AdlClassName+"*");
      ACO.setFctTypeDefaultValue(node.sbf.newPointer(node.ResultFct.getReturnType(),0x00));
    // <AdlClassName> * new_<AdlClassName>(...
      ASTDeclarationSpecifiers declSpec = new
	ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
      parentChild(declSpec, makeQualifiedType(AdlClassName), 0);
      parentChild(createdNode, declSpec, insertChild++);
      
      ASTFunctionDeclarator decl = new
	ASTFunctionDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDECLARATOR);
      ASTPtrOperator ptr = new
	ASTPtrOperator(ADLCPPParserTreeConstants.JJTPTROPERATOR);
      ptr.setPtrkind("PointerTo");
      parentChild(decl, ptr, 0);
      ASTFunctionDirectDeclarator directDecl = new
	ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);
      parentChild(decl, directDecl, 1);

      int declChild = 0;
      parentChild(createdNode, decl, insertChild++);
      for (int i = 0; i < child.jjtGetNumChildren(); i++)
	parentChild(directDecl, child.jjtGetChild(i), declChild++);
      for (int j=1; j < nbChildren; j++) {
	child = node.jjtGetChild(j);
	if (child instanceof ASTExceptionSpec) {
	  ACO.setFctThrowsList(node.ResultFct.getExceptions()); 
	  parentChild(directDecl, child, declChild++);
	}
      }
      decl.setFctName("new_"+AdlClassName);
      ACO.setFctPointedType(createdNode);
    } // not constructor
    else { //Testing a function
      if (debug) System.out.println("\nTesting the function "+ACO.getFctName());
      isInConstructor = false;
      ACO.setIsConstructor(isInConstructor);

      if (child instanceof ASTDeclarationSpecifiers) {
        parentChild(createdNode, child, insertChild++);
        // Construct an ASTDeclaration node to store function result type
        ASTDeclaration declFctType = new
          DeclarationBuilder( (ASTDeclarationSpecifiers)child, 
                              node.jjtGetChild(1) ).generatedAST();
        ACO.setFctPointedType(declFctType);
        child = node.jjtGetChild(1);
      }
      if (child instanceof ASTFunctionDeclarator) {
        parentChild(createdNode, child, insertChild++);
      }
      else if (child instanceof ASTFunctionDirectDeclarator) {
        parentChild(createdNode, child, insertChild++);
      }

      ACO.setFctThrowsList(node.ResultFct.getExceptions()); 
    }

    // adding function/constructor declarations without compound statement 
    // to bhvDeclaration: children of class declaration.
    try {
      bhvDeclaration.addElement(((SimpleNode)createdNode).clone());
    } 
    catch(CloneNotSupportedException e) {}

    // Adding scopeOverride to function names:
    // setFctName to <AdlClassName>+"ACO" + fctName
    for (int i=0; i < createdNode.jjtGetNumChildren(); i++) {
      child = createdNode.jjtGetChild(i);
      // Modify fct name in QualifiedId
      if (child instanceof ASTFunctionDeclarator) {
        do { child = child.jjtGetChild(1); }
        while (! (child instanceof ASTFunctionDirectDeclarator));
      }
      if (child instanceof ASTFunctionDirectDeclarator) {
        parentChild(child,
                    makeQualifiedId(AdlClassName + "ACO::" +
                                    ((ASTFunctionDirectDeclarator)child).getFctName().Name),
                    0);
      }
    }

    // Visiting bhvSpec adds function/constructor body to createdNode
    int tmp = 0;
    for (int i=1; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADLBhvSpec) {
	tmp = i;
      }	
      else if (child instanceof ASTADLProlog) {
	localProlog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADLEpilog) {
	localEpilog = (ASTCompoundStatement)child.jjtGetChild(0);
      }
    } 
    parentChild(createdNode, (Node)node.jjtGetChild(tmp).jjtAccept(this, data), insertChild++);

    if (debug) System.out.println("\tisInConstructor = " + isInConstructor);
    savIndex = 0;
    isInConstructor = false;
    if (debug) {
      System.out.println("\nReturn Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }
  
  
  /**
    * Visit pattern for a node of type ASTADLProlog
    * <BR>Should not be visited!
    * @param node the ASTADLProlog to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return an error
    */
  public Object visit(ASTADLProlog node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");
    // not visited
    return new InternalError();
  }


  /**
    * Visit pattern for a node of type ASTADLEpilog
    * <BR>Should not be visited!
    * @param node the ASTADLEpilog to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return an error
    */
  public Object visit(ASTADLEpilog node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");
    // not visited
    return new InternalError();
  }


  /**
    * Visit pattern for a node of type ASTADLBhvSpec
    * <P><B>Algorithm: </B>generates all statements of the body of the
    * currently visited behavior description:<OL>
    * <LI>initialization statements: ACO.makeACO_Initialization(hasInheritedSemantics)
    * <LI>global prolog statements
    * <LI>local prolog statements
    * <LI>callstate temporary vairables declarations and initializations
    * (statements of the handle "saveBlock", created by the visit to the
    * ASTADLAssGrp child)
    * <LI>call to the tested function: ACO.makeACO_CallBlock()
    * <LI>statements from the visit to the ASTADLAssGrp child
    * <LI>local epilog statements
    * <LI>global epilog statements
    * <LI>finalization statements: ACO.makeACO_Final()
    * </OL>
    * @param node the ASTADLBhvSpec to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the created node (ASTCompoundStatement).
    * @see ACOauxiliary#makeACO_Initialization
    * @see ACOauxiliary#makeACO_CallBlock
    * @see ACOauxiliary#makeACO_Final
    */
  public Object visit(ASTADLBhvSpec node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child, newChild = null;
    ASTCompoundStatement createdNode = new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    int nbChildren = node.jjtGetNumChildren();

    int insertChild = 0;
    child = node.jjtGetChild(0);
    if (child instanceof ASTADLBhvClassification) {
      child.jjtAccept(this, data); // only side-effect: saving (ab)normalSpec
      child = node.jjtGetChild(1); // ADLAssGrp
    }

    // child : ADLAssGrp (last child)
    boolean hasInheritedSemantics = false;
    for (int i=0; i<child.jjtGetNumChildren() && !hasInheritedSemantics; i++) {
      if (child.jjtGetChild(i) instanceof ASTADLInhSemantics) {
	hasInheritedSemantics = true;
	ACO.setSuperSemantics(((ASTADLInhSemantics)child.jjtGetChild(i)).getSuperClassName());
      }
    }

    Vector ACO_Initialization = ACO.makeACO_Initialization(hasInheritedSemantics);
    int nbc = ACO_Initialization.size();
    for (int i = 0; i < nbc; i++) {
      newChild = (Node)ACO_Initialization.elementAt(i);
      parentChild(createdNode, newChild, insertChild);
      insertChild += 1;
    }
    
    if ((globalProlog != null) && !isInConstructor) {
      nbc = globalProlog.jjtGetNumChildren();
      ASTCompoundStatement GPclone = null;
      try {
	GPclone = (ASTCompoundStatement)(globalProlog.clone());
      }
      catch(CloneNotSupportedException e) {}
      for (int i = 0; i < nbc; i++) {
        parentChild( createdNode, 
		     (Node)GPclone.jjtGetChild(i).jjtAccept(this, data),
		     insertChild);
        insertChild += 1;
      }
    }

    if (localProlog != null) {
      nbc = localProlog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localProlog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    // need to visit the child assertion group to know how many
    // preDeclaration are to be inserted

    ASTCompoundStatement visitedAssGrp = (ASTCompoundStatement)child.jjtAccept(this, data);
    Vector ACO_CallBlock = ACO.makeACO_CallBlock();

    nbc = preDeclaration.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)preDeclaration.elementAt(i), insertChild+i);
    }
    insertChild += nbc;


    if (!isInConstructor) {
      nbc = saveBlock.size();
      if (debug) System.out.println("Size of saveBlock: " + nbc);
      if (nbc > 0) {
	ASTIfNode ifStSkipAss = new
	  ASTIfNode(ADLCPPParserTreeConstants.JJTIFNODE);
	parentChild(createdNode, ifStSkipAss, insertChild++);
	ASTUnaryExpr notSkip = 
	  new ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
	notSkip.setOp("!");
	
	parentChild(notSkip, makeIdExpr("skipAssertions"), 0);
	parentChild(ifStSkipAss, notSkip, 0);
	
	ASTCompoundStatement thenBlk =
	  new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
	
	ASTTryNode triedBlk = makeExceptionCatchBlock(thenBlk, true);
	parentChild(ifStSkipAss, triedBlk, 1);
	
	for (int i = 0; i < nbc; i++) {
	  parentChild(thenBlk, (Node)saveBlock.elementAt(i), i);
	}
      }
    }

    nbc = ACO_CallBlock.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_CallBlock.elementAt(i), insertChild+i);
    }
    insertChild += nbc;

    nbc = visitedAssGrp.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, visitedAssGrp.jjtGetChild(i), insertChild+i);
    }
    insertChild += nbc;

    if (localEpilog != null) {
      nbc = localEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localEpilog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    if ((globalEpilog != null ) && !isInConstructor) {
      nbc = globalEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        try {
	  newChild = (Node)((SimpleNode)globalEpilog.jjtGetChild(i)).clone();
        }
        catch(CloneNotSupportedException e) {}
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    Vector ACO_Final = ACO.makeACO_Final();
    nbc = ACO_Final.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_Final.elementAt(i), insertChild); 
      insertChild += 1;
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADLBhvClassification
    * <BR>stores normal/abnormal specifications in the associated ACOauxiliary.
    * @param node the ASTADLBhvClassification to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @returns null, that will not be used by the caller 
    */
  public Object visit(ASTADLBhvClassification node, Object data) throws ParseException { 

    Node child;
    int nbc = node.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      child = node.jjtGetChild(i);
      if ((node.getFirstChildKind()).equals("NORMAL")) {
        ACO.setNormalSpec(child);
      }
      else {
        ACO.setAbnormalSpec(child);
      }
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": null");
    return null;
  }

  /**
    * Visit pattern for a node of type ASTADLAssGrp
    * <P><B>Algorithm: </B>
    * <UL><LI><B>IF</B> this is a nested assertion group: the assertion group
    * is considered as an expression.
    * <UL><LI><B>IF</B> it contains more than one assertion
    * <BR>Returns an expression: the result of the visit to this child assertion.
    * <LI><B>ELSE</B> Returns an expression: the "AND" of the results of the 
    * visits to these child assertions (if the input node is considered as a
    * decomposable expression, makePreAssertion is called and the visit
    * returns the name node of the variable that was initialized by the input
    * node value).</UL>
    * <LI><B>ELSE</B> the created node is an ASTCompoundStatement that contains
    * <I>either</I> the return node of the visit to the grand-child when 
    * the latter is an ASTADLTryStmt or an ASTADLIfStmt <I>or</I> the
    * statements of the ASTCompoundStatement returned by the visit to an 
    * ASTADLAssertion grand-child.
    * @param node the ASTADLAssGrp to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #makePreAssertion(ORG.opengroup.adl.symboltable.LType,
    *                        ORG.opengroup.adl.parsers.adlcpp.Node,
    *                        java.lang.String)
    */
  public Object visit(ASTADLAssGrp node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int nbChildren;
    String tmpStr = "", tmpName;
    if (isDecomposable) {
      tmpStr = node.getExpressionImage();
    }      
    
    if (isInAssertion || isInBinding) { // nested
      nbChildren = node.jjtGetNumChildren();
     // group of boolean expressions => conjunction
      if (nbChildren > 1) {
	ASTBinaryLogicalExpr createdNode = new 
	  ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
	createdNode.addOp("&&");
	nbChildren = node.jjtGetNumChildren();
	for (int i = 0; i < nbChildren; i++) {
          parentChild( createdNode, 
                       (Node)node.jjtGetChild(i).jjtAccept(this, data),
                       i);
	}
	if (! (isInCallState || isInTry || isInQuantified)) {
	  tmpName = makePreAssertion(node.getType(), createdNode, tmpStr);
	  if (debug)
	    System.out.println("Return Visiting AST" + node + ": IdExpr");
	  return makeIdExpr(tmpName);
	}
	else {
	  if (debug)
	    System.out.println("Return Visiting AST" + node + ": " + createdNode);
	  return createdNode;
	}
      }
      else {
	return (Node)node.jjtGetChild(0).jjtAccept(this, data);
      }
    }
    else { // not nested : "real" assertion group
      ASTCompoundStatement createdNode =
        new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
      Node visitedAss,child1, child2;
      int insertChild = 0;
      for (int i = 0; i < node.jjtGetNumChildren(); i++) {
        child1 = node.jjtGetChild(i);
        if (child1 instanceof ASTADLBinding) {
	  for (Enumeration e = ((Vector)child1.jjtAccept(this, data)).elements(); 
             e.hasMoreElements(); ) {
	    parentChild(createdNode, (Node)e.nextElement(), insertChild++);
	  }
	}
        else if (child1 instanceof ASTADLInhSemantics) {
          // already tested: see ASTADLBhvSpecification
        }
        else { // child1 instanceof ASTADLStatement
          if ((child1 instanceof ASTADLIfStmt) || (child1 instanceof ASTADLTryStmt)) {
            parentChild(createdNode, (Node)child1.jjtAccept(this, data), insertChild++);
          }
          else { // child1 : ADLAssertion  
            visitedAss = (Node)child1.jjtAccept(this, data);
            for (int j = 0; j < visitedAss.jjtGetNumChildren(); j++) {
              parentChild(createdNode, visitedAss.jjtGetChild(j), insertChild++);
            }
          }
        }
      }
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      }
      return createdNode;
    }
  }

  
  /**
    * Visit pattern for a node of type ASTADLBinding
    * <BR>generates a vector of variable declaration that
    * corresponds to the binding.
    * @param node the ASTADLBinding to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the vector of transformed node.
    */
  public Object visit(ASTADLBinding node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Vector translatedBinding = new Vector();
    isInBinding = true;
    isDecomposable = false;
    // first child is ASTDeclaration or ASTADLNamedParamList
    Node decls = node.jjtGetChild(0); 
    // second child is an expression
    Node expr = node.jjtGetChild(1);

    if ((expr instanceof ASTADLAssGrp) &&
        (expr.jjtGetChild(0) instanceof ASTADLTryStmt)) {

      if (decls instanceof ASTDeclaration) {
        translatedBinding.addElement(decls);
      }
      else { // ASTADLNamedParamList
        for (int i = 0; i < decls.jjtGetNumChildren(); i++) {
          translatedBinding.addElement(decls.jjtGetChild(i));
        }
      }

      ASTTryNode createdNode = new
        ASTTryNode(ADLCPPParserTreeConstants.JJTTRYNODE);

      ASTADLTryStmt tryexpr =
        (ASTADLTryStmt)expr.jjtGetChild(0);

      ASTDeclaration params;

      if (node.getInitialized() == null) {
        node.errorMessage("Cannot translate bindings with a \"try\" expression");
      }
      String bindingName = node.getInitialized().getName();
      String bindingType = node.getInitialized().getType().getName();

      // try { <bindingName> = <tryexpr.tryblk>; }
      // ( catch(<tryexpr.catch>) { <bindingName> = <tryexpr.catchblk>; } )+

      ASTCompoundStatement blk =
        new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
      parentChild(blk,
                  new StatementBuilder
                  ( new AssignmentBuilder
                    ( bindingName, "=",
                      (Node)tryexpr.jjtGetChild(0).jjtAccept(this, null))
                    .generatedAST())
                  .generatedAST(),
                  0);
      parentChild(createdNode, blk, 0);

      ASTCatchNode catchNode;
      for (int i = 1; i < tryexpr.jjtGetNumChildren()-1; i += 2) {
        // formal parameter, e.g. "catch(Exc e)"

        catchNode = new ASTCatchNode(ADLCPPParserTreeConstants.JJTCATCHNODE);
        parentChild(createdNode, catchNode, (i+1)/2);
        parentChild(catchNode, tryexpr.jjtGetChild(i).jjtGetChild(0), 0);

        blk = new
          ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
        parentChild(blk,
                    new StatementBuilder
                    ( new AssignmentBuilder
                      ( bindingName, "=",
                        (Node)tryexpr.jjtGetChild(i+1).jjtAccept(this, null))
                      .generatedAST())
                    .generatedAST(),
                    0);
        parentChild(catchNode, blk, 1);
      }
      isInBinding = false;

      translatedBinding.addElement(createdNode);
      return translatedBinding;
    }
    else {  // not a try
      Node createdNode = null;
      if (decls instanceof ASTDeclaration) {
        // named parameter = declSpec + decltor
        // declaration == decl specifiers + Init declarator
        // Binding initdeclarator == (decltor/direct decltor)+expression 
        createdNode = new DeclarationBuilder
          ( (ASTDeclarationSpecifiers)decls.jjtGetChild(0),
            new InitDeclaratorBuilder
            (decls.jjtGetChild(1),
             (Node)expr.jjtAccept(this, data)).generatedAST()
            ).generatedAST();
      }
      else { // child(0) is ASTADLNamedParamList
        for (int i = 0; i < decls.jjtGetNumChildren(); i++) {
          translatedBinding.addElement(decls.jjtGetChild(i));
        }
        if (node.getInitialized() == null) {
          createdNode = new StatementBuilder
            ((Node)expr.jjtAccept(this, data)).generatedAST();
        }
        else {
          createdNode = new StatementBuilder
            ( new AssignmentBuilder
              ( node.getInitialized().getName(),
                "=",
                (Node)expr.jjtAccept(this, data)
                ).generatedAST()
              ).generatedAST();
        }
      }
      isInBinding = false;
      if (debug) {
        System.out.println("Return Visiting AST" + node + ": " + createdNode);
        ((SimpleNode)createdNode).dump(System.out, 100);
      }
      translatedBinding.addElement(createdNode);
      return translatedBinding;
    }
  }

  /**
    * Visit pattern for a node of type ASTADLIfStmt
    * <P><B>Algorithm: </B><UL><LI><B>IF</B> inside an assertion,
    * it is considered as an expression and will generate an "?:"
    * conditional expression by visiting the child nodes. If there
    * is no else child node, the else branch of the generated conditional
    * expression will be the boolean constant true.
    * <LI><B>ELSE</B> generates a C++ if statement by visiting the
    * child nodes (condition, then node, else node if present).
    * </UL>
    * @param node the ASTADLIfStmt to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node (ASTCondExpr / ASTIfStmt).
    */
  public Object visit(ASTADLIfStmt node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node child; 

    if (isInAssertion) { // nested
      ASTParenthExpr parenth = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      ASTCondExpr condNode = new
        ASTCondExpr(ADLCPPParserTreeConstants.JJTCONDEXPR);
      parentChild(parenth, condNode, 0);

      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        parentChild(condNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
      }
      if (nbc == 2) {
        child = new
	  ASTConstant(ADLCPPParserTreeConstants.JJTCONSTANT);
	((ASTConstant)child).setCstVal("true");
        parentChild(condNode, child, 2);
      }

      if (isDecomposable) {
        return makeIdExpr(makePreAssertion(node.getType(),
					   parenth,
					   node.getExpressionImage()));
      }
      else {
        return parenth;
      }
    }
    else { // not nested
      ASTIfNode createdNode = new
        ASTIfNode(ADLCPPParserTreeConstants.JJTIFNODE);
      
      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
	if (i == 2) { // case of "else"
	  ASTElseNode elseNode = new
	    ASTElseNode(ADLCPPParserTreeConstants.JJTELSENODE);
	  parentChild(elseNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), 0);
	  parentChild(createdNode, elseNode, i);
	}
	else { 
        parentChild(createdNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
	}
      }
      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    }
  }


  /**
   * Visit pattern for a node of type ASTADLTryStatement
   * <P><B>Algorithm: </B><UL><LI><B>IF</B> inside an assertion, it is
   * considered as an expression (but not a decomposable one):
   * <BR>call makeTryStatement and return the Name node of the variable that
   * has been initialized with the value of the input node.
   * <LI><B>ELSE</B> generates a C++ try/catch statement by visiting the
   * child nodes (try block and handlers( catch block(s) )).  </UL>
   * @param node the ASTADLTryStmt to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the created node (ASTIdExpr / ASTTryNode).
   * @see #makeTryStatement(java.lang.boolean,
   *                        ORG.opengroup.adl.parsers.ASTADLTryStmt,
   *                        java.lang.String)
   */
  public Object visit(ASTADLTryStmt node, Object data) throws ParseException { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");

    if (isInAssertion) { // nested
      String tmpStr = "", tmpName;
      if (! (isInCallState || isInQuantified)) {
	tmpStr = node.getExpressionImage();
      }

      isInTry = true; isDecomposable = false;
      tmpName = makeTryStatement(isInCallState, node, tmpStr);
      isInTry = false; isDecomposable = ! (isInCallState || isInQuantified);

      return makeIdExpr(tmpName);
    }
    else { // not nested
      ASTTryNode createdNode = new
	ASTTryNode(ADLCPPParserTreeConstants.JJTTRYNODE);

      // try assertion group
      parentChild(createdNode, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);

      int nbChildren = node.jjtGetNumChildren(), j = 1;
      for (int i = 1; i < nbChildren-1; i += 2) {
	ASTCatchNode catchNode = new
	  ASTCatchNode(ADLCPPParserTreeConstants.JJTCATCHNODE);
	// catch exception declaration
	parentChild(catchNode, node.jjtGetChild(i), 0);
	// catch assertion group
	parentChild(catchNode, (Node)node.jjtGetChild(i+1).jjtAccept(this, data), 1);
	parentChild(createdNode, catchNode, j++);
      }
      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    }
  }
    
    
  /**
   * Visit pattern for a node of type ASTADLAssertion
   * <P><B>Algorithm</B><UL>
   * <LI><B>IF</B> nested: the assertion node is in fact considered as a
   * boolean expression: just visit the child.
   * <LI>Generates the code that corresponds to an assertion;
   * i.e. for assertion "get_active_accounts()== @get_active_accounts()+1" :
   <PRE>
   |  testContext->assertionStart("get_active_accounts() == @get_active_accounts() + 1");
   |  if (!skipAssertions) {
   |    int __ADL_tmp_1 = bankObject->get_active_accounts();
   |    int __ADL_tmp_2 = __ADL_savTmp_1;
   |    int __ADL_tmp_3 = __ADL_tmp_2 + 1;
   |
   |    tddRes = (__ADL_tmp_1 == __ADL_tmp_3);
   |
   |    if (debugMode) {
   |      // report on each sub-expression
   |      testContext->infoline(adlstr.reset() << "get_active_accounts()\t-> " << __ADL_tmp_1);
   |      testContext->infoline(adlstr.reset() << "@get_active_accounts()\t-> " << __ADL_tmp_2);
   |      testContext->infoline(adlstr.reset() << "@get_active_accounts() + 1\t-> " << __ADL_tmp_3);
   |      testContext->infoline(adlstr.reset() << "get_active_accounts() == @get_active_accounts() + 1\t-> " << tddRes);
   |    }
   |  testContext->assertionResult((tddRes ? ADL_PASS : ADL_FAIL));
   |  }
   |  else {
   |     testContext->assertionResult(ADL_UNEVALUATED);
   |  }
   </PRE>
   * The variable declarations __ADL_tmp_i are "preAssertion"
   * obtained by the expression decomposition algorithm.
   * </UL>
   * @see #visit(ORG.opengroup.adl.parsers.adlcpp.ASTCondExpr,java.lang.Object)
   * @param node the ASTADLAssertion to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the transformed node (ASTCompoundStatement).
   */
  public Object visit(ASTADLAssertion node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node visitedAss = null, child = node.jjtGetChild(0);
    int nbChildren = node.jjtGetNumChildren(), insertChild = 0;
    String tmpStr = "", assLabel = null;
    boolean labelledAss = false, taggedAss = false, quantif = false;
    
    if (debug) {
      System.out.println("\t\tisInAssertion = "+isInAssertion);
      System.out.println("\t\tisInBinding   = "+isInBinding);
    }
    if (isInAssertion || isInBinding) {
      for(int i = 0; i < nbChildren; i++) {
	child = node.jjtGetChild(i);
	if (child instanceof ASTADLLabel) {
          // $$$$ WARNING labels not used in nested assertions
          return null;
	}
	else if (child instanceof ASTADLTags) {
          // $$$$ WARNING tags not used in nested assertions
          return null;
	}
        else if (child instanceof ASTADLQuantifiedAssertion) {
          // $$$$ ERROR no quantified assertion in nested assertion groups
          return null;
        }
        else { // expression of the assertion : to visit
          return (Node)child.jjtAccept(this, data);
        }
      }
    }
    preAssertion = new Vector();
    boolContextStack = new Stack();

    ASTCompoundStatement createdNode = new 
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    
    for(int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADLLabel) {
        labelledAss = true;
        assLabel = (String)child.jjtAccept(this, data);
        if (debug) System.out.println("LABEL : " + assLabel);
      }
      else if (child instanceof ASTADLTags) {
        taggedAss = true;
	assTags = (String[])child.jjtAccept(this, data);
      }
      else if (child instanceof ASTADLQuantifiedAssertion) {
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInTry || isInQuantified || isInBinding);
        visitedAss = (Node)child.jjtAccept(this, data);
        isInAssertion = false;
	isDecomposable = false;
        quantif = true;
	tmpStr = ((SimpleNode)child).getExpressionImage();
      }
      else { // expression of the assertion
        // visit the expression: decomposition algorithm
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInTry || isInQuantified || isInBinding);
        visitedAss = (Node)child.jjtAccept(this, data);
        isInAssertion = false; isDecomposable = false;
	tmpStr = ((SimpleNode)child).getExpressionImage();
      }
    }

    int nbp = preAssertion.size();

    if (labelledAss) {
      tmpStr = "[ " + assLabel + " ] " + tmpStr;
    }
    
    // testContext->assertionStart(<tmpStr>)
    
    ASTPostfixExpr assertionStart = new PostfixExprBuilder
      (	makeIdExpr("testContext"),
	new RefAccessSuffixBuilder
	( makeIdExpr("assertionStart"),
	  new ArgumentListBuilder(makeString(tmpStr)).generatedAST()
	  ).generatedAST()
	).generatedAST();
    parentChild(createdNode, 
                new StatementBuilder(assertionStart).generatedAST(),
                0);

    // if tagged assertion:
    // testContext->infoline("Tagged assertion (Untestable)";)

    if (taggedAss) {
      ASTPostfixExpr untest = new PostfixExprBuilder
	( makeIdExpr("testContext"),
	  new RefAccessSuffixBuilder
	  ( makeIdExpr("infoline"),
	    new ArgumentListBuilder(makeString("Tagged assertion (Untestable)")).generatedAST()
	    ).generatedAST()
	  ).generatedAST();
      parentChild(createdNode, 
                  new StatementBuilder(untest).generatedAST(),
                  1);
    }
    else {
      // if (!skipAssertions)
      
      ASTIfNode ifStSkipAss = new
        ASTIfNode(ADLCPPParserTreeConstants.JJTIFNODE);
      
      ASTUnaryExpr condition = new
        ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
      condition.setOp("!");
      ASTIdExpr skipAss = makeIdExpr("skipAssertions");
      parentChild(condition, skipAss, 0);
      
      parentChild(ifStSkipAss, condition, 0);
      
      ASTCompoundStatement triedBlk =
	new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
      ASTTryNode tryStmt = makeExceptionCatchBlock(triedBlk, false);
      parentChild(ifStSkipAss, tryStmt, 1);
      

      if (debug)
	System.out.println("\t\tIs quantified assertion ? -> "+quantif);
      if (quantif) { // quantified assertion:
        // store the statements returned by the visit of the quantified assertion
        // (this visit returns an ASTCompoundStatement)
        for (int i = 0; i < visitedAss.jjtGetNumChildren(); i++) {
          parentChild(triedBlk, (Node)visitedAss.jjtGetChild(i), i);
        }
        insertChild = visitedAss.jjtGetNumChildren();
      }
      else {
	for (int i = 0; i < nbp; i++) {
	  if (debug) {
	    System.out.println("\nDumping preAssertion "+i); 
	    ((SimpleNode)(Node)preAssertion.elementAt(i)).dump(System.out, 100);
	  }
	  parentChild(triedBlk, (Node)preAssertion.elementAt(i), i);
	}
	insertChild = nbp;
	
	// tddRes = last tmp
	parentChild(triedBlk,
		    new StatementBuilder
		    ( new AssignmentBuilder("tddRes", "=", visitedAss
					    ).generatedAST()
		      ).generatedAST(),
		    insertChild);
	insertChild += 1;
      }

      // if (debugMode) {
      if (!quantif) {
        // No debug mode for quantified assertions
	ASTIfNode ifDebug = new
	  ASTIfNode(ADLCPPParserTreeConstants.JJTIFNODE);
	parentChild(ifDebug, makeIdExpr("debugMode"), 0);
	parentChild(triedBlk, ifDebug, insertChild);
	insertChild += 1;
	
	ASTCompoundStatement thenBlk2 = new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
	parentChild(ifDebug, thenBlk2, 1);
	
	// infoline of sub-expressions
	
	Node preass;
	ASTDeclaration vardecl;
	ASTBinaryArithExpr arg;
	String decl = "__ADLTmp";
	int insertChild2 = 0;
	int nbChild = 0;
	
	for (int i = 0; i < nbp; i++) {
	  preass = (Node)preAssertion.elementAt(i);
	  if (preass instanceof ASTDeclaration) {
	    vardecl = (ASTDeclaration)preass;	    
	    String msg;
	    try {
	      msg = vardecl.getProperty("tokenString") + "\\t-> ";
	    }
	    catch (PropertyAccessException e) {
	      msg = "unknown string";
	    }
	    Node vardeclChild = vardecl.jjtGetChild(1);
	    if (vardeclChild instanceof ASTInitDeclarator) {
	      if (vardeclChild.jjtGetChild(0) instanceof ASTDeclarator)
		decl = ((ASTDeclarator)vardeclChild.jjtGetChild(0)).getNamedDeclarator(CPPBuiltinType.intT,0).getName();
	      else
		decl = ((ASTQualifiedId)((ASTDirectDeclarator)vardeclChild.jjtGetChild(0)).jjtGetChild(0)).getName();
	    }
	    else if (vardeclChild instanceof ASTDeclarator) {
	      decl = ((ASTDeclarator)vardeclChild).getNamedDeclarator(CPPBuiltinType.intT,0).getName();
	    }
	    else if (vardeclChild instanceof ASTDirectDeclarator) {
	      decl = ((ASTQualifiedId)vardeclChild.jjtGetChild(0)).getName();
	    }
	    
	    Node assTmpValue = null;
	    // type is in QualifiedType
	    LType declType = ((ASTQualifiedType)vardecl.jjtGetChild(0).jjtGetChild(0)).getQualifiedType();
	    if ((declType instanceof LReference) || (declType instanceof LConstructedType)) {
	      ASTUnaryExpr refNode = new
		ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
	      refNode.setOp("&");
	      parentChild(refNode, makeIdExpr(decl), 0);
	      assTmpValue = makeDebugModeMsg(msg, refNode);
	    }
	    else {
	      assTmpValue = makeDebugModeMsg(msg, makeIdExpr(decl));
	    }

	    parentChild( thenBlk2, assTmpValue, insertChild2);
	    insertChild2 += 1;
	  }
	}
      }

      // testContext->assertionResult((tddRes ? ADL_PASS : ADL_FAIL));

      ASTCondExpr passFail = new
        ASTCondExpr(ADLCPPParserTreeConstants.JJTCONDEXPR);
      parentChild(passFail, makeIdExpr("tddRes"), 0);
      parentChild(passFail, makeIdExpr("ADL_PASS"), 1);
      parentChild(passFail, makeIdExpr("ADL_FAIL"), 2);
      parentChild(triedBlk,
		  new StatementBuilder
		  ( new PostfixExprBuilder
		    ( makeIdExpr("testContext"),
		      new RefAccessSuffixBuilder
		      ( makeIdExpr("assertionResult"),
			new ArgumentListBuilder(passFail).generatedAST()
			).generatedAST()
		      ).generatedAST()
		    ).generatedAST(),
                  insertChild++);
      
      // testContext->assertionResult((tddRes ? ADL_PASS : ADL_FAIL));
      // else { testContext->assertionResult(ADL_UNEVALUATED); }
      ASTElseNode elseBlk = new ASTElseNode(ADLCPPParserTreeConstants.JJTELSENODE);
      parentChild(ifStSkipAss, elseBlk, 2);
      parentChild(elseBlk,
		  new StatementBuilder
		  ( new PostfixExprBuilder
		    ( makeIdExpr("testContext"),
		      new RefAccessSuffixBuilder
		      ( makeIdExpr("assertionResult"),
			new ArgumentListBuilder(makeIdExpr("ADL_UNEVALUATED")).generatedAST()
			).generatedAST()
		      ).generatedAST()
		    ).generatedAST(),
                  0);
      
      parentChild(createdNode, ifStSkipAss, 1);
    }
    
    preAssertion = null;
    preAssIndex = 0;
    
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      //      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }
  

  /**
   * Visit pattern for a node of type ASTADLLabel
   * @param node the ASTADLLabel to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the name of the label (String)
   */
  public Object visit(ASTADLLabel node, Object data) throws ParseException { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": " + node.getName());
    }
    // not visited
    return node.getName();
  }


  /**
   * Visit pattern for a node of type ASTADLTags
   * @param node the ASTADLTags to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return an error
   */
  public Object visit(ASTADLTags node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    String[] assTags;
    try {
      assTags = (String[])node.getProperty("taglist");
    }
    catch(PropertyAccessException e) {
      assTags = new String[1];
      assTags[0] = "<no tag>";
    }
    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + assTags);
    return assTags;
  }


  /**
   * Visit pattern for a node of type ASTADLQuantifiedAssertion.
   * <P>Generates the code that corresponds to a quantified assertion.
   * for instance for assertion
   * <BR>"forall (long i : ADL_long_range(1,10)) { Stock_balance(i) >= 0; }" :
<PRE>
|  testContext->assertionStart("forall(long i : ADL_long_range(1,10)){Stock_balance(i)>=0;}");
|  if (!skipAssertions) {
|     tddRes = true;
|     for (long i = 1; tddRes && (i <= 10); i++)
|       tddRes = StockBrokerObject->Stock_balance(i) >= 0;
|     testContext->assertionResult(( tddRes ? ADL_PASS : ADL_FAIL ));
|  }
|  else {
|     testContext->assertionResult(ADL_UNEVALUATED);
|  }
</PRE>
   * <P><B>Notes</B><UL>
   * <LI>a quantified assertion is considered as non-decomposable
   * <LI>in case of "forall", the code generated for tddRes is "tddRes = true;"
   * <LI>in case of "exists", the code generated for tddRes is "tddRes = false;
   * <LI>if there are several enumeration domains, the generated code is an
   * imbrication of "for" loops.  </UL>
   * @param node the ASTADLQuantifiedAssertion to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the created node (ASTCompoundStatement).
   */
  public Object visit(ASTADLQuantifiedAssertion node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTCompoundStatement createdNode = new ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);

    boolean isForall = node.getQuantifier().equals("forall");
    int insertChild = 0;

    // tddRes = <isForall>;
    if (isForall) {
      parentChild( createdNode, new StatementBuilder
		   ( new AssignmentBuilder
		     ("tddRes", "=", makeBoolean("true")).generatedAST()
		     ).generatedAST(),
		   insertChild++);
    }
    else {
      parentChild( createdNode, new StatementBuilder
		   ( new AssignmentBuilder
		     ("tddRes", "=", makeBoolean("false")).generatedAST()
		     ).generatedAST(),
		   insertChild++);
    }

    int nbLoops = node.jjtGetChild(0).jjtGetNumChildren();
    ASTDeclaration[] domDecl = new ASTDeclaration[nbLoops];
    Node[] domExpr = new Node[nbLoops];

    for (int i = 0; i < nbLoops; i++) {
      domDecl[i] = (ASTDeclaration)node.jjtGetChild(0).jjtGetChild(i).jjtGetChild(0);
      domExpr[i] = node.jjtGetChild(0).jjtGetChild(i).jjtGetChild(1);
    }

    String iterator = "";
    Node infBorn = null, supBorn = null;
    ASTForNode forNode = new
      ASTForNode(ADLCPPParserTreeConstants.JJTFORNODE);

    Node last;

    ASTDeclaration forDecl;
    ASTBinaryLogicalExpr andExpr;
    ASTPostfixExpr incExpr;

    parentChild(createdNode, forNode, insertChild++);
    for (int i = 0; i < nbLoops; i++) {
      try {
	iterator = domDecl[i].getExpressionDecl().getName();
      }
      catch (ParseException e) {}
	
      try {
	infBorn = domExpr[i].jjtGetChild(1).jjtGetChild(0);
	supBorn = domExpr[i].jjtGetChild(1).jjtGetChild(1);
      }
      catch (NullPointerException e) {
	// second child of ADLDomain must be PostfixExpr having 2 children:
	// IdExpr and ArgumentList having 2 expression children.
	((SimpleNode)domExpr[i]).errorMessage("Error : Invalid adl domain in quantified assertion");
      }
      // long i
      forDecl = new DeclarationBuilder
	( domDecl[i].jjtGetChild(0), // Declaration specifiers
	  new InitDeclaratorBuilder
	  ( domDecl[i].jjtGetChild(1), // Declarator or Direct declarator
	    infBorn).generatedAST()
	  ).generatedAST();
      // tddRes && i<supBorn
      andExpr = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      andExpr.addOp("&&");
      if (isForall) {
        parentChild(andExpr, makeIdExpr("tddRes"), 0);
      }
      else { // exists
	ASTUnaryExpr condition = new
	  ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
	condition.setOp("!");
	parentChild(condition, makeIdExpr("tddRes"), 0);
	parentChild(andExpr, condition, 0);
      }
      ASTBinaryLogicalExpr relExpr = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      relExpr.addOp("<=");
      parentChild(relExpr, makeIdExpr(iterator), 0);
      parentChild(relExpr, supBorn, 1);
      ASTParenthExpr parenth = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      parentChild(parenth, relExpr, 0);

      parentChild(andExpr, parenth, 1);

      // i++
      ASTPostDeIncrement increment = new
        ASTPostDeIncrement(ADLCPPParserTreeConstants.JJTPOSTDEINCREMENT);
      increment.setOp("++");
      incExpr = new 
	PostfixExprBuilder(iterator,increment).generatedAST();

      parentChild(forNode, forDecl, 0);
      parentChild(forNode, andExpr, 1);
      parentChild(forNode, incExpr, 2);
      String forKind[] = {"Declaration","Expression", "Expression"}; 
      forNode.setKind(forKind);

      last = forNode;

      if (i == nbLoops-1) {
        // visit the assertion-expression
        isInQuantified = true; isDecomposable = false;
        Node assGrp = (Node)node.jjtGetChild(1).jjtAccept(this, data);    
        isInQuantified = false;
        isDecomposable = isInAssertion && ! (isInCallState || isInTry || isInBinding);

        // tddRes = assGrp
        parentChild(last, new StatementBuilder
                    ( new AssignmentBuilder
                      ( "tddRes", "=", assGrp).generatedAST()
		      ).generatedAST(),
                    3);
      }
      else {
        forNode = new
          ASTForNode(ADLCPPParserTreeConstants.JJTFORNODE);
        parentChild(last, forNode, 3);
      }
    }
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
  }


  /**
   * Visit pattern for a node of type ASTADLImplExpr
   * <P><B>Algorithm: </B><UL><LI>visits the 2 operand expressions (returning
   * ret_exp1, ret_exp2).
   * <LI><B>CASE</B> operator :<UL>
   * <LI><B>==?</B>return !(ret_exp1) || (ret_exp2)
   * <LI><B>?==</B>return  (ret_exp2) || !(ret_exp1)
   * <LI><B>?=?</B>return ret_exp1 == ret_exp2
   * <LI><B>?:?</B>return (ret_exp1) ? "abnormal" : !(ret_exp2 && "abnormal")
   * </UL></UL>
   * @param node the ASTADLImplExpr to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the created node.
   */
  public Object visit(ASTADLImplExpr node, Object data) throws ParseException { 
    if (debug) {
      System.out.println("\nVisiting AST" + node + " ...");
    }

    Node createdNode;
    Node ch1 = node.jjtGetChild(0),
      ch2 = node.jjtGetChild(1);
    String op = node.getOp();
    String tmpStr = "", tmpName;

    if (isDecomposable) {
      tmpStr = node.getExpressionImage();
    }      
     
    if (op.equals("==>") || 
        op.equals("<==")) {
      if (op.equals("<==")) {  // exchange (exp1) and (exp2)
        Node tmp = ch1;
        ch1 = ch2;
        ch2 = ch1;
      }

      // e1 ==> e2 changed into !(e1) || (e2)
      createdNode = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      ((ASTBinaryLogicalExpr)createdNode).addOp("||");

      parentChild(createdNode, 
		  new NotBuilder(ch1,
				 !(ch1 instanceof ASTParenthExpr)
				 ).generatedAST(),
		  0);
      parentChild(createdNode, ch2, 1);
      createdNode = (Node)createdNode.jjtAccept(this, data);
    }

    else if (op.equals("<=>")) {
      // e1 <=> e2 changed into (e1) == (e2)
      createdNode = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      ((ASTBinaryLogicalExpr)createdNode).addOp("==");

      parentChild(createdNode, ch1, 0);
      parentChild(createdNode, ch2, 1);
    }

    else {// e1 <:> e2 changed into (!(e1) || "abnormal") && (!"abnormal" || !(e2) || (e1))
      createdNode = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      ((ASTBinaryLogicalExpr)createdNode).addOp("&&");

      ASTParenthExpr subnode1 = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      ASTBinaryLogicalExpr snode1 = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      snode1.addOp("||");
      parentChild(createdNode, subnode1, 0);
      parentChild(subnode1, snode1, 0);
      ch1 = (Node)ch1.jjtAccept(this, data);
      parentChild(snode1,
		  new NotBuilder(ch1,
				 !(ch1 instanceof ASTParenthExpr)
				 ).generatedAST(),
		  0);
      parentChild(snode1, makeIdExpr("abnormal"), 1);

      ASTParenthExpr child2 = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      parentChild(createdNode, child2, 1);

      ASTBinaryLogicalExpr subnode2 = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      subnode2.addOp("||");
      subnode2.addOp("||");
      parentChild(child2, subnode2, 0);
      Node p1bis = null;
      try {
        p1bis = (Node)((SimpleNode)ch1).clone();
      } catch(CloneNotSupportedException e) {}

      // lazy evaluation: 2nd child must be evaluated only when abnormal is true
      boolContextStack.push(new BoolContext("abnormal", true));
      ch2 = (Node)ch2.jjtAccept(this, data);
      boolContextStack.pop();

      parentChild(subnode2,
                  new NotBuilder(makeIdExpr("abnormal"), false).generatedAST(),
                  0);
      parentChild(subnode2, new NotBuilder(ch2, false).generatedAST(), 1);
      parentChild(subnode2, p1bis, 2);
    }

    if (isDecomposable) {
      if (debug)
	System.out.println("Return Visiting AST" + node + ": IdExpr");
      return makeIdExpr(makePreAssertion(node.getType(),
                                         createdNode, tmpStr));
    }
    else {
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }
      return createdNode;
    }    
  }


  /**
    * Visit pattern for a node of type ASTCondExpr: a call to decompose(node,data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTCondExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTCondExpr node, Object data) throws ParseException { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    // lazy evaluation:
    // b ? e1 : e2
    // --> t1 = visit(b);
    // t2 = t1 ? visit(e1) : <default value>;
    // t3 = t1 ? <default value> : visit(e2);
    // t4 = t1 ? t2 : t3

    parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
    // ATTENTION: we suppose that the boolean expression is ALWAYS decomposed,
    int ind = preAssIndex-1;
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, true));
    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    boolContextStack.pop();
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, false));
    parentChild(node, (Node)node.jjtGetChild(2).jjtAccept(this, data), 2);
    boolContextStack.pop();

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(node.getType(),
					 node,
					 node.getExpressionImage()));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTBinaryLogicalExpr
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTBinaryLogicalExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTBinaryLogicalExpr node, Object data) throws ParseException { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...: visit number "+nbv++);
    }
    int nbChildren = node.jjtGetNumChildren();
    String op1 = (String)node.getOperatorsVector().elementAt(0);

    if (op1.equals("||") || op1.equals("&&")) {
      if (nbChildren > 2) {
	// to simplify lazy evaluation mechanism, we use associativity:
	// e1 || e2 || e3... is first changed in e1 || (e2 || e3...)
	ASTBinaryLogicalExpr newOr = new
	  ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
	newOr.addOp(op1);
	newOr.copySourceTokens(node);
	parentChild(newOr, node.jjtGetChild(0), 0);

	ASTBinaryLogicalExpr newSubOr = new
	  ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);

	// copy source tokens and skip first child tokens
	newSubOr.copySourceTokens(node);
	newSubOr.firstToken = ((SimpleNode)node.jjtGetChild(1)).firstToken;

	parentChild(newOr, newSubOr, 1);

	for (int i = 1; i < nbChildren; i++) {
	  parentChild(newSubOr, node.jjtGetChild(i), i-1);
	  if (i < nbChildren-1) {
	    newSubOr.addOp((String)node.getOperatorsVector().elementAt(i));
	  }
	}

	return newOr.jjtAccept(this, data);
      }
      else {
	// lazy evaluation:
	// e1 || e2
	// --> t1 = visit(e1);
	// t2 = t1 ? false : visit(e2)
	// t3 = t1 || t2
	parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
	if (op1.equals("||"))
	  boolContextStack.push
	    ( new BoolContext("__ADL_tmp_" + (preAssIndex-1), false));
	else
	  boolContextStack.push
	    ( new BoolContext("__ADL_tmp_" + (preAssIndex-1), true));

	parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
	boolContextStack.pop();
	
	if (isDecomposable) {
	  return makeIdExpr(makePreAssertion(node.getType(),
					     node,
					     node.getExpressionImage()));
	}
	else {
	  return node;
	}
      }
    }
    else {
      return decompose(node, data);
    }
  }

  /**
    * Visit pattern for a node of type ASTExpr
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTExpr node, Object data) throws ParseException { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...: visit number "+nbv++);
    }
    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTBinaryArithExpr
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTBinaryArithExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #visit(ORG.opengroup.parsers.adlcpp.ASTCondExpr)
    */
  public Object visit(ASTBinaryArithExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    childrenAccept(node);
    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(node.getType(),
					 node,
					 node.getExpressionImage()));
    }
    else {
      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + node);
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTUnaryExpr
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTUnaryExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #visit(ORG.opengroup.parsers.adlcpp.ASTCondExpr)
    */
  public Object visit(ASTUnaryExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

      return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTCastExpr.
    * <BR>Just visit the expression child (nothing to do for
    * the type child).
    * @param node the ASTCastExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  public Object visit(ASTCastExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int lastnode = node.jjtGetNumChildren()-1;
    parentChild(node, (Node)node.jjtGetChild(lastnode).jjtAccept(this, data), lastnode);

    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + node);
    return node;
  }


  /**
    * Visit pattern for a node of type ASTNewExpr
    * @param node the ASTNewExpr to visit.
    * @param data unused.
    * @return the node.
    */
  public Object visit(ASTNewExpr node, Object data) throws ParseException { 

    for (int i = 1; i < node.jjtGetNumChildren(); i++)
      node.jjtGetChild(i).jjtAccept(this, data);

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(node.getType(),
                                         node,
                                         node.getExpressionImage()));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTADLCallStateExpr
    *
    * <P><B>Algorithm: </B><OL>
    * <LI>call makeSaveBlock to create a "saveBlock", i.e. a declaration
    * of a temporary variable that will be initialized with the type
    * and value of this node child expression before the call to the tested
    * function.
    * <LI>call "makePreAssertion" to create a "preAssertion", i.e. a
    * declaration of a temporary variable that will be initialized with
    * the type and value of the precedently created variable.
    * <BR>This second variable is used as a sub-expression of the
    * decomposition algorithm.
    * </OL>
    * <PRE>
    <type of @expr> __ADL_savTmp_i = <@expr>;
   // ...
   // <call to tested function>
   // ...
   // assertion:
   <type of @expr> __ADL_tmp_j = __ADL_savTmp_i;
      </PRE>
    * @param node the ASTADLCallStateExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #makeSaveBlock
    * @see #makePreAssertion(ORG.opengroup.adl.symboltable.LType,
    *                        ORG.opengroup.adl.parsers.adlcpp.Node,
    *                        java.lang.String)
    */
  public Object visit(ASTADLCallStateExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    //First child is UnaryExpression
    Node child = node.jjtGetChild(0); 
    if (isInCallState) {
      return child.jjtAccept(this, data);
    }
    ASTIdExpr createdNode;
    String exprType = ((Typed)child).getType().getName();

    isInCallState = true;
    isDecomposable = false;
    createdNode = makeIdExpr
      (makeSaveBlock(((Typed)child).getType(), (Node)child.jjtAccept(this, null)) );
    isInCallState = false;
    isDecomposable = isInAssertion && ! (isInTry || isInQuantified || isInBinding);
    if (isDecomposable) {
      return makeIdExpr( makePreAssertion(((Typed)child).getType(),
					  createdNode,
					  node.getExpressionImage()));
    }
    else {
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }
      return createdNode;
    }
  }

  /**
   * Visit pattern for a node of type ASTUnchangedExpr.
   * <P><B>Algorithm:</B><UL>
   * <LI>clone the node children (argument list of the "unchanged" expression).
    * <LI>copy the source tokens (that represents the 
    * code of the expression) on the cloned arguments.
   * <LI><B>IF</B> there is <I>one</I> argument "arg", create a node "(@arg ==
   * clone_arg)" and visit its 2 children nodes.
   * <LI><B>ELSE</B> (there are <I>n</I> arguments
   * "arg<SUB>1</SUB>"..."arg<SUB>n</SUB>"): create a node
   * "((@arg<SUB>1</SUB> == clone_arg<SUB>1</SUB>) && ... &&
   * (@arg<SUB>n</SUB> == clone_arg<SUB>n</SUB>))" and visit all
   * "@arg<SUB>i</SUB>" and "clone_arg<SUB>i</SUB>" child nodes. </UL>
   * @param node the ASTUnchangedExpr to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the created node.
   */
  public Object visit(ASTUnchangedExpr node, Object data) throws ParseException {

    if (debug) System.out.println("Visiting AST" + node + " ...");

    String tokenStr = null;

    Node argList = node.jjtGetChild(0);
    int nbexpr = argList.jjtGetNumChildren();
    
    if (nbexpr == 0) {
      // unchanged() ---> true
      ASTConstant createdNode = makeBoolean("true");
      createdNode.copySourceTokens(node);

      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    }

    Node argCopy = null;
    try {
      argCopy = (Node)((SimpleNode)argList).clone();
    } catch (CloneNotSupportedException e) {}

    if (nbexpr == 1) {
      // unchanged(expr) ---> (expr == @expr)
      
      SimpleNode expr1 = (SimpleNode)argList.jjtGetChild(0);
      Node expr2 = argCopy.jjtGetChild(0);

      ASTParenthExpr createdNode = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      ASTBinaryLogicalExpr eq = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      eq.addOp("==");
      eq.copySourceTokens(node);

      ASTADLCallStateExpr callstatexpr = new
        ASTADLCallStateExpr(ADLCPPParserTreeConstants.JJTADLCALLSTATEEXPR);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens(expr1);
      callstatexpr.prependToken(new Token("@"));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(createdNode, (Node)eq.jjtAccept(this, data), 0);
      
      if (debug) {
	System.out.println("\nReturn Visiting AST" + node + ": " + createdNode);
	((SimpleNode)createdNode).dump(System.out, 100);
      }

      return createdNode;
    }

    System.out.println("unchanged(expr1, expr2, ..., exprn) case");
    // unchanged(expr1, ..., exprn) ---> ((expr1 == @expr1) && ... && (exprn == @exprn))
    ASTParenthExpr createdNode = new
      ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
    ASTBinaryLogicalExpr andNode = new
      ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
    andNode.addOp("&&");

    SimpleNode expr1, expr2; ASTParenthExpr parenth;
    ASTBinaryLogicalExpr eq; ASTADLCallStateExpr callstatexpr;

    for (int i = 0; i < nbexpr; i++) {
      System.out.println("looping on element " + i + " of unchanged");

      expr1 = (SimpleNode) argList.jjtGetChild(i);
      expr2 = (SimpleNode) argCopy.jjtGetChild(i);

      parenth = new
        ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
      eq = new
        ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
      eq.addOp("==");

      callstatexpr = new
        ASTADLCallStateExpr(ADLCPPParserTreeConstants.JJTADLCALLSTATEEXPR);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens(expr1);
      callstatexpr.prependToken(new Token("@"));
      eq.copySourceTokens(callstatexpr);
      eq.prependToken(new Token(expr1.getExpressionImage() + " == "));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(parenth, eq, 0);
      parentChild(andNode, parenth, i);
    }

    andNode.copySourceTokens(node);
    parentChild(createdNode, (Node)andNode.jjtAccept(this, data), 0);

    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }
    return createdNode;
}

  /**
   * Visit pattern for a node of type ASTNormalExpr
   * @param node the ASTNormalExpr to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the ASTIdExpr node "normal".
   */
  public Object visit(ASTNormalExpr node, Object data) throws ParseException { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": IdExpr");
    }
    ASTIdExpr n = makeIdExpr("normal");
    n.setType(CPPBuiltinType.boolT);
    n.copySourceTokens(node);

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(n.getType(),
					 n,
					 n.getExpressionImage()));
    }
    else {
      return n;
    }
  }


  /**
   * Visit pattern for a node of type ASTAbnormalExpr
   * @param node the ASTAbnormalExpr to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the ASTIdExpr node "normal".
   */
  public Object visit(ASTAbnormalExpr node, Object data) throws ParseException { 

    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": IdExpr");
    }
    ASTIdExpr abn = makeIdExpr("abnormal");
    abn.setType(CPPBuiltinType.boolT);
    abn.copySourceTokens(node);

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(abn.getType(),
					 abn,
					 abn.getExpressionImage()));
    }
    else {
      return abn;
    }
  }

  /**
    * Visit pattern for a node of type ASTReturn 
    * @param node the ASTReturn to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the ASTIdExpr node "__ADL_return".
    */
  public Object visit(ASTReturn node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    if (inlining && isInCallState) {
      throw new ParseException("\"return\" in call-state expression via an inline");
    }
    ASTIdExpr tmpExpr = makeIdExpr("__ADL_return");
    tmpExpr.setType(node.getType());
    tmpExpr.copySourceTokens(node);

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(node.getType(),
					 tmpExpr,
					 node.getExpressionImage()));
    }
    else {
      return tmpExpr;
    }
  }


  /**
    * Visit pattern for a node of type ASTString.
    *
    * @param node the ASTString to visit.
    * @param data unused.
    * @return the node.
    */
  public Object visit(ASTString node, Object data) { 

    if (debug)
      System.out.println("Return Visiting AST" + node + ": " + node);
    return node;
  }

  
  /**
   * Visit pattern for a node of type ASTParenthExpr
   * @param node the ASTParenthExpr to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the node.
   */
  public Object visit(ASTParenthExpr node, Object data) throws ParseException { 
    childrenAccept(node);
    return node;
  }
  
  
  /**
   * Visit pattern for a node of type ASTArgumentList
   * @param node the ASTArgumentList to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the node.
   */
  public Object visit(ASTArgumentList node, Object data) throws ParseException { 
    childrenAccept(node);
    return node;
  }
  
  
  /**
   * Visit pattern for a node of type ASTPostfixExpr
   * @param node the ASTPostfixExpr to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTPostfixExpr node, Object data) throws ParseException { 
    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
    }
   
    int insertChild = 0;
    Node child0 = node.jjtGetChild(0), child;

    ASTPostfixExpr createdNode = new
      ASTPostfixExpr(ADLCPPParserTreeConstants.JJTPOSTFIXEXPR);

    // Case of child0 != ASTSimpleTypeSpecifier
    if (!((child0 instanceof ASTBuiltinTypeSpec) || (child0 instanceof ASTQualifiedType))) {
      Node tmp = (Node)child0.jjtAccept(this,data);
      if (debug) {
	System.out.println("\tTmp node: " + tmp);
      }
      if (tmp instanceof ASTPostfixExpr) {
	for( int i = 0; i < tmp.jjtGetNumChildren(); i++) {
	  parentChild(createdNode, tmp.jjtGetChild(i), insertChild++);
	}
      }
      else {
	parentChild(createdNode, tmp, insertChild++);
      }
    }

    for (int i = 1; i < node.jjtGetNumChildren(); i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTPostDeIncrement) 
	parentChild(createdNode, child, insertChild++);
      else
	parentChild(createdNode, (Node)child.jjtAccept(this, data), insertChild++);
    }
    if (debug) {
      System.out.println("Return Visiting AST" + node + ": " + createdNode);
      ((SimpleNode)createdNode).dump(System.out, 100);
    }

    if (isDecomposable) {
      return makeIdExpr(makePreAssertion(node.getType(),
					 createdNode,
					 node.getExpressionImage()));
    }
    else {
      return createdNode;
    }
//    return createdNode;
}


  /**
   * Visit pattern for a node of type ASTRefAccessSuffix
   * @param node the ASTRefAccessSuffix to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTRefAccessSuffix node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTRefAccessSuffix createdNode = new
      ASTRefAccessSuffix(ADLCPPParserTreeConstants.JJTREFACCESSSUFFIX);
    parentChild(createdNode, node.jjtGetChild(0), 0);
    if (node.jjtGetNumChildren() > 1) {
      ASTArgumentList arguments = (ASTArgumentList)node.jjtGetChild(1);
      parentChild(createdNode, arguments, 1);
      for (int i = 0; i<arguments.jjtGetNumChildren(); i++)
	parentChild(arguments, (Node)arguments.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return createdNode;
}


  /**
   * Visit pattern for a node of type ASTDotAccessSuffix
   * @param node the ASTDotAccessSuffix to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTDotAccessSuffix node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    ASTDotAccessSuffix createdNode = new DotAccessSuffixBuilder
      (node.jjtGetChild(0)).generatedAST();
    if (node.jjtGetNumChildren() > 1) {
      ASTArgumentList arguments = (ASTArgumentList)node.jjtGetChild(1);
      parentChild(createdNode, arguments, 1);
      for (int i = 0; i<arguments.jjtGetNumChildren(); i++)
	parentChild(arguments, (Node)arguments.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return createdNode;
}


  /**
   * Visit pattern for a node of type ASTPrimaryExpr
   * the node is created when it has 2 chidren: IdExpr + ArgumentList
   * it is a call of an inline
   * the node created is: visit(currentInline.ADLAssGrp)
   *
   * <P><B>Algorithm</B>
   * <UL>
   *   <LI><B>IF</B> the first child is an inline name:
   *     <BR>replace the first child by a clone of the body of the
   *     corresponding inline and re-visit the node with flag
   *     "inlining" true.
   *   <LI><B>ELSE</B><I>(not a call to an inline macro)</I>
   *     <UL>
   *       <LI><B>IF</B> "inlining"
   *       <LI><B>ELSE</B>
   *         <BR>if this is a qualified name then it is not modified,
   *         otherwise it refers to the "real object" and must be
   *         qualified: "<name>" is replaced by
   *         "<adlclass name>Object-><name>" (or "<adlclass name>-><name>"
   *         in the case of a constructor)
   *     </UL>
   * </UL>
   * @param node the ASTPrimaryExpr to visit.
   * @param data unused.
    * @exception ParseException thrown by parser.
   * @return the transformed node ().
   */
  public Object visit(ASTPrimaryExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting AST" + node + " ...");

    int nbChildren = node.jjtGetNumChildren();
    ASTIdExpr child1 = (ASTIdExpr)node.jjtGetChild(0);
    ASTArgumentList child2 = (ASTArgumentList)node.jjtGetChild(1);
    boolean isInline;
    String child1Name = child1.getIdName();
    Node createdNode = null;

    isInline = node.isADLInline;
    if (isInline) {
      inlineParams = null;
      for (Enumeration e = inlineDeclaration.elements(); 
           e.hasMoreElements(); ) {
        currentInline = (ASTADLInlineDecl)e.nextElement();
        if ((currentInline.getADLInline().getName()).equals(child1Name)) {
	  if (currentInline.getADLInline().getParameters().length > 0) {
	    // Inline having at least 1 parameter
	    Node declarator = currentInline.jjtGetChild(1);

	    // search for the FunctionDirectDeclarator
	    while (!(declarator instanceof ASTFunctionDirectDeclarator)) {
	      declarator = declarator.jjtGetChild(1);
	    }
	      
	    // and get the formal paramater list (second child)
	    // note that we know for sure that second child is a
	    // ParameterList since getParameters().length > 0 (see above)
	    inlineParams = declarator.jjtGetChild(1);

	  } else { // Inline without parameters --> empty argument list
	    inlineParams = new ArgumentListBuilder().generatedAST();
	  }

	  // we found a matching inline, get out of loop
	  break;
	}
      }

      if (inlineParams == null) {
        if (debug) System.out.println("inline not found!!!");
      }
      inlineActualParams = child2;
      ASTADLAssGrp inlineBody = null;
      try {
	inlineBody = (ASTADLAssGrp)
	  ((SimpleNode)currentInline.jjtGetChild(2)).clone();
      } catch (CloneNotSupportedException e) {}

      // copy original source tokens in inline body expression node
      SimpleNode statement = (SimpleNode)inlineBody.jjtGetChild(0);
      if (statement instanceof ASTADLAssertion) {
	// ADLAssertion : reach the ADLExpression node
	statement = (SimpleNode)statement
	  .jjtGetChild(statement.jjtGetNumChildren() - 1);
      } else {
	// ADLIfStmt || ADLTryStmt : we have the right node in hand
      }
      // the next statement has no side-effect on the Token chain of
      // the original inline body node since it has first been cloned.
      statement.prependToken(new Token("[inline call: " +
				       node.getExpressionImage() + "] "));

      inlining = true;
      createdNode = (Node)inlineBody.jjtAccept(this, data);
      inlining = false;

      // copy source tokens from original node
      ((SimpleNode) createdNode).copySourceTokens(node);

      if (debug)
	System.out.println("Return Visiting AST" + node + ": " + createdNode);
      return createdNode;
    } // if (isInline)
    else {
      if (inlining) {
	int firstOverride = child1Name.indexOf("::");
	String firstName;
	if (firstOverride == -1)
	  firstName = child1Name;
	else
	  firstName = child1Name.substring(0, firstOverride);
	int param = -1;
	// Case if one parameter
	if (inlineParams instanceof ASTParameterDeclaration) {
	  if (firstName.equals(((ASTParameterDeclaration)inlineParams)
				.getFctParameter().getName()))
	    param = 1;
	}
	else { // More than one parameter
	  for (int i = inlineParams.jjtGetNumChildren()-1; i >= 0; i--) {
	    if (firstName.equals(((ASTParameterDeclaration)((ASTParameterDeclarationList)inlineParams).jjtGetChild(i)).getFctParameter().getName())) {
	      param = i;
	      break;
	    }
	  }
	}
	if (param != -1) { // formal parameter
	  Node actual = null;
	  try { // case of many expressions
	    if (inlineActualParams.jjtGetChild(0) instanceof ASTExpr)
	      actual = (Node)
		((SimpleNode)inlineActualParams.jjtGetChild(0).jjtGetChild(param))
		.clone();
	    else // case of one expression
	      actual = (Node)
		((SimpleNode)inlineActualParams.jjtGetChild(param))
		.clone();
	  }
	  catch (CloneNotSupportedException e) {}
	  
	  //	  parentChild(createdNode, actual, 0);
	  
	  if (firstOverride != -1) { // inlineParam::aName <aSuffix>
	    if (actual instanceof ASTIdExpr) {
	      // -->  inlineActualParam::aName <aSuffix>
	      ((ASTIdExpr)actual).setIdName(((ASTIdExpr)actual).getIdName() +
			     child1Name.substring(firstOverride));
	      createdNode = new
		ASTPrimaryExpr(ADLCPPParserTreeConstants.JJTPRIMARYEXPR);
	      parentChild(createdNode, actual, 0);
	      parentChild(createdNode, child2, 1);
	      ((ASTPrimaryExpr)createdNode).setType(node.getType());
	    }
	    else {
	      // -->  inlineActualParam <Refsuffix: aName> <aSuffix>

	      ASTParenthExpr parenth = new
		ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
	      parentChild(parenth, actual, 0);
	      createdNode = new PostfixExprBuilder
		( parenth,
		  new RefAccessSuffixBuilder
		  ( child1Name.substring(firstOverride+1),child2).generatedAST()
		  ).generatedAST();
	      ((ASTPostfixExpr)createdNode).setType(node.getType());
	    }
	  }

	  ((SimpleNode)createdNode).copySourceTokens(node);
	  
	  return createdNode.jjtAccept(this, data);
	}
      }

      if (node.isTestedClassMember) {
	if (child1.getIdName().indexOf("->") == -1) {
	  if (isInConstructor) {
	    child1.setIdName("__ADL_return->" + child1.getIdName());
	  }
	  else {
	    child1.setIdName(ACO.getAdlClassName() + "Object->" + child1.getIdName());
	  }
	}
      }
    }

    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    
    if (isDecomposable) {
      if (debug) {
	System.out.println("\tIn node "+node+":");
	System.out.println("\t\tnode type = "+node.getType());
      }
      return makeIdExpr(makePreAssertion(node.getType(),
					 node,
					 node.getExpressionImage()));
    }
    else
      if (debug) {
	System.out.println("Return Visiting AST" + node + ": " + node);
	((SimpleNode)node).dump(System.out, 100);
      }
      return node;
  }
  
  
  /* Visit pattern for a node of type ASTThis.
   *
    * @param node the ASTThis to visit.
    * @param data unused.
    * @return the transformed node ().
    */
  public Object visit(ASTThis node, Object data) { 
    
    if (debug) {
      System.out.println("Visiting AST" + node + " ...");
      System.out.println("Return Visiting AST" + node + ": IdEXpr");
    }
    if (isInConstructor)
      return makeIdExpr("__ADL_return");
    else
      return makeIdExpr(ACO.getAdlClassName() + "Object");
  }

  /**
   * Visit pattern for a node of type ASTADLThrownExpr
   * <BR>If the expression is decomposable, call makePreAssertion, else
   * return a true node.
   * @param node the ASTADLThrownExpr to visit.
   * @param data unused.
   * @exception ParseException thrown by parser.
   * @return the transformed node.
   */
  public Object visit(ASTADLThrownExpr node, Object data) throws ParseException { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");

    Node createdNode = null;

    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 0) { // Case of Thrown(...)
      createdNode = makeIdExpr("wasThrownException");
    }
    else if (nbChildren == 1) { // TypeNameList
      ASTTypeNameList nameList = (ASTTypeNameList)node.jjtGetChild(0);
      int nbExc = nameList.jjtGetNumChildren();
      if (nbExc == 1) { // case of thrown(Exc1)
	createdNode = makeIdExpr(ACO.findExc(((ASTTypeName)nameList.jjtGetChild(0)).getType()));
      }
      else { // case of thrown(Exc1, Exc2 ..)
	createdNode = new
	  ASTParenthExpr(ADLCPPParserTreeConstants.JJTPARENTHEXPR);
	ASTBinaryLogicalExpr orExpr = new
	  ASTBinaryLogicalExpr(ADLCPPParserTreeConstants.JJTBINARYLOGICALEXPR);
	parentChild(createdNode, orExpr, 0);
	ASTIdExpr subNode;
	for (int i = 0; i < nbExc; i++) {
	  subNode = makeIdExpr(ACO.findExc(((ASTTypeName)nameList.jjtGetChild(i)).getType()));
	  if (i != nbExc-1) orExpr.addOp("||");
	  parentChild(orExpr, subNode, i);
	}
      }
    }
    
    if (isDecomposable) {
//TBD Verify node.getType() returns "bool"
      createdNode = makeIdExpr(makePreAssertion(node.getType(),
						createdNode,
						node.getExpressionImage()));

     if (debug) System.out.println("Return visiting AST" + node + " : "+createdNode);
      return createdNode;
    }
    else {
      if (debug) System.out.println("Return visiting AST" + node + " : "+createdNode);
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTAssignmentExpr
    * @param node the ASTAssignmentExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    */
  public Object visit(ASTAssignmentExpr node, Object data) throws ParseException {
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
    }
    return node;
  }


  /**
    * Visit pattern for a node of type ASTConstant
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConstant to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConstant node, Object data) throws ParseException { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTConstantExpr
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConstantExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConstantExpr node, Object data) throws ParseException { 
    
    if (debug) System.out.println("Visiting AST" + node + " ...");
    
    return decompose(node, data);
  }


  //TBR
  /**
   * Visit pattern for a node of type ASTIdExpr
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> "inlining" and the name is the name of a parameter of
    * the inline that is currently being inlined, then create a node that is
    * a clone of the corresponding actual parameter, visit it and <B>return</B>
    * it.
    * <LI>If the expression is decomposable, call makePreAssertion, else
    * return the node.
    * </UL>
    * @param node the ASTIdExpr to visit.
    * @param data unused.
    * @exception ParseException thrown by parser.
    * @return the created node.
   */
  public Object visit(ASTIdExpr node, Object data) throws ParseException { 

    if (debug) System.out.println("Visiting ASTIdExpr: " + node.getIdName());

    if (node.getIdName().equals("adl_thrownException"))
      return node;

    //    if (inlining && isInCallState) {
    //      if (node.isADLBinding()) {
    //        throw new ParseException("Error: binding in a call-state via an inline");
    //      }
    //    }

    String tmpName = node.getIdName();
    if (inlining) {
      int nbChild = inlineParams.jjtGetNumChildren();
      if (debug) System.out.println("\tNode AST"+inlineParams+" has "+nbChild+" children.");
      int param = -1;
      if (inlineParams instanceof ASTParameterDeclarationList) {//children are ParameterDeclaration
	for (int i = inlineParams.jjtGetNumChildren()-1; i >= 0; i--) {
	  if (tmpName.equals(((ASTParameterDeclaration)inlineParams.jjtGetChild(i))
                             .getFctParameter().getName())) {
	    param = i;
	    break;
	  }	  
	}
      }
      else {// Only one child : Parameter Declaration
	if (tmpName.equals(((ASTParameterDeclaration)inlineParams).getFctParameter().getName()))
	    param = 0;
      }
      if (debug) {
	System.out.println("\tParameter number: "+param);
	System.out.println("\tinlineActualParams: "+inlineActualParams);
	((SimpleNode)inlineActualParams).dump(System.out, 100);
      }
      if (param != -1) {
        Node createdNode = null;
        try {
          inlining = false;
          createdNode = (Node)
	    ((SimpleNode) ((SimpleNode)inlineActualParams.jjtGetChild(param))
             .clone()).jjtAccept(this, data);
          inlining = true;
        }
        catch (CloneNotSupportedException e) {
          if (debug) e.printStackTrace(System.err);
        }
	if (debug) 
	  System.out.println("Return Visiting AST" + node + ": "+createdNode);
        return createdNode;
      }
    }
    
    boolean isStatic = false;
    try {
      isStatic = ((Boolean)node.getProperty("isStatic")).booleanValue();
    } catch (PropertyAccessException e) { }

    if (node.isTestedClassField()) {
      if (!isStatic) {
	if (isInConstructor) {
	  node.setIdName("__ADL_return->" + tmpName);
	}
	else {
	  node.setIdName(ACO.getAdlClassName() + "Object->" + tmpName);
	}
      } else {
	if (tmpName.indexOf("::") == -1) // this is not a FQN
	  node.setIdName(ACO.getAdlClassName() + "::" + tmpName);
      }
    }

    if (isDecomposable) {
      if (debug) {
	System.out.println("\tisDecomposable: " + isDecomposable);
	System.out.println("\tNode type: " + node.getType());
      }
      Node tmpExpr = makeIdExpr
	( makePreAssertion(node.getType(), node, tmpName));
      if (debug) 
	System.out.println("Return Visiting AST" + node + ": "+ tmpExpr);
      return tmpExpr;
    }
    if (debug) 
      System.out.println("Return Visiting AST" + node + ": "+ node);
    return node;
  }

}

// ***********************************************************************
// History Log:
// $Log: ADLCPPTransformVisitor.java,v $
// Revision 1.32  1998/10/15 18:14:10  mallet
// Bug fixed: pr/153: encapsulated call-state expressions and assertions
// within a try/catch statement to trap any exception raised during their
// evaluation.
//
// Revision 1.31  1998/07/28 16:00:59  mallet
// Fixed bug in visit of IdExpression for static fields: if the name of
// the field is not fully qualified, make it be.
//
// Revision 1.30  1998/06/30 18:21:44  bareau
// Added a visit method for ASTNewExpr node.
//
// Revision 1.29  1998/06/30 13:30:32  mallet
// Replaced access to expressionImage property by call to
// getExpressionImage().
// Added one parameter to call of CPPTypeAux.getDefaultValue: always
// false since we call this method for "lazy" members of a logical and|or.
// Replaced setting of expressionImage property on newly created nodes by
// use of copySourceTokens() and other token chain related methods.
// Fixed bugs in processing of "unchanged": was only visiting the
// children of the equivalent equality expression instead of the equality
// itself.
// Fixed bug in visit of ASTPrimaryExpression for inline calls: was
// incorrectly skipping the PtrOperator part of the FunctionDeclarator,
// and getting the inline body expression.
// Changed debug mode reporting for inline calls: now also reporting the
// initial inline call expression.
//
// Revision 1.28  1998/06/24 08:42:42  bareau
// Bug fix: the translation of a binding defined as a try expression
// is now correct.
//
// Revision 1.27  1998/06/23 08:12:43  bareau
// Bug fix: the visit of the actual argument of an inline call must be
// done with flag inlining false.
// Some simplifications in the code.
//
// Revision 1.26  1998/06/18 10:32:10  mallet
// Fixed incorrect use of type short name instead of full one (FQN).
// Made use of makeQualidifedType systematic.
//
// Revision 1.25  1998/06/16 15:17:32  bareau
// Bug fix: there was an error with inlines that return a constant expression.
//
// Revision 1.24  1998/06/16 12:48:45  mallet
// Fixed bug in generated code: static fields were treated as non-static,
// i.e., they were systematically accessed by dereferencing an instance
// (ADL_return-> or realObject->).
//
// Revision 1.23  1998/06/15 09:23:40  bareau
// Bug fix: wrong indices in visit of ADLQuantifiedAssertion (no created
// child for quantifier).
//
// Revision 1.22  1998/06/11 15:44:49  bareau
// Bug fix: the transformation for nested ADL_IfStatement was incorrect.
//
// Revision 1.21  1998/06/02 13:11:20  bareau
// Save the STM context at the beginning of the visit and restore it
// at the end.
//
// Revision 1.20  1998/04/28 14:56:59  cinard
// Updated call to CPPSymbolFactory.newPointer().
//
// Revision 1.19  1998/04/28 12:39:23  cinard
// Updated code to take into account replacement of CSymbolTableMgr,
// CSymbolFactory and CBuiltinType by CPPSymbolTableMgr, CPPSymbolFactory
// and CPPBuiltinType.
//
// Revision 1.18  1998/04/15 14:09:19  bareau
// Update according to the new NL mechanism: there are no NLD_* nodes.
//
// Revision 1.17  1998/04/01 16:49:23  mallet
// Replaced ADLNode by SimpleNode.
// Replaced NodeException by PropertyAccessException.
// Replaces AdditiveNodeException by PropertyAccessException.
// Added missing " around included file name.
//
// Revision 1.16  1998/03/31 12:41:01  bareau
// Bug fix: typing error.
//
// Revision 1.15  1998/03/26 15:45:22  rami
// Bug fix: an inline call may occur _before_ the corresponding
// inline definition.
//
// Revision 1.14  1998/03/20 11:26:58  rami
// Bug fix: Constant node must be decomposed fro lazy evaluation.
//
// Revision 1.13  1998/03/19 11:07:41  rami
// Implementation of lazy evaluation for logical expressions:
//         - Added inner class BoolContext for __ADL_tmp declarations.
//         - Used BoolContext in method makePreAssertion.
//         - Modified visit methods for BinaryLogicalExpr and ADLImplExpr.
//
// Revision 1.12  1998/03/18 16:56:29  cinard
// Changed stm.newPointer() to sbf.newPointer().
//
// Revision 1.11  1998/03/10 17:26:42  mallet
// Merged v1.1 with development trunk
//
// Revision 1.10.2.5  1998/03/09 19:56:37  rami
// Removed initializations from ADLCPPTransformVisitor constructor: yet done in translate method.
//
// Revision 1.10.2.4  1998/03/09 19:27:40  rami
// Fixed bug:missed initialization in translate method.
//
// Revision 1.10.2.3  1998/03/09 18:33:54  rami
// Fixed bug in transformation of expression with exception operator.
//
// Revision 1.10.2.2  1998/03/06 11:33:23  rami
// Improved transformation of ASTADLThrownExpr:
// 	Using findExc(LType) method to generate the string
// 	wasThrown_Exc<index number of exception>.
//
// Revision 1.10.2.1  1998/03/04 16:08:35  rami
// Fixed bug in local prolog and epilog.
// Removed scope override from function names in class declaration.
// Added function modifyFctName(Node, String).
//
// Revision 1.10  1998/02/27 15:46:03  rami
// Updated copyright notice.
// Bug fixed in ASTADLThronwExpr visit method.
//
// Revision 1.9  1998/02/27 08:39:43  rami
// SetKind to constructed ForNode telling unparser the kind of each child node.
//
// Revision 1.8  1998/02/26 15:25:35  rami
// ADLCPPTransformVisitor now extends CPPGeneric.
// Now all default visits to AST nodes visits children and return node.
//
// Revision 1.7  1998/02/24 10:28:26  rami
// Implemented and tested Bindings with multiple declaration behavior.
// Fixed debug mode messages. Added function makeDebugModeMsg(String, Node).
// Fixed call state expression in assertionStart message.
// Fixed Java doc comments.
//
// Revision 1.6  1998/02/17 10:57:20  mallet
// Merged v1.0 with development trunk
//
// Revision 1.5.2.3  1998/02/16 19:24:34  rami
// Fix bug: visit of ASTExpr must be decomposed and return a node.
//
// Revision 1.5.2.2  1998/02/16 12:27:18  rami
// Bug fix:
// 1/ add visit ASTParenthExpr to return the node; default visit returned data.
// 2/ add catchNode to the transfomed TryNode as a second child.
//
// Revision 1.5.2.1  1998/02/13 16:51:44  rami
// Removed useless error message.
//
// Revision 1.5  1998/02/12 16:03:39  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.4  1998/02/12 11:30:01  rami
// Version OK for examples.
// Debug flag set to false.
//
// Revision 1.3  1998/02/04 13:58:02  rami
// Moved the parentChild and make* methods to the generic visitor.
//
// Revision 1.2  1998/01/30 11:14:31  rami
// Added method translate(Node) to start the ADL translator.
// Added examples in comments.
// Fixed some bugs.
//
// ***********************************************************************



