// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/cpp/CPPBuilder.java,v $
// $Revision: 1.10 $
// $Date: 1998/10/14 08:12:49 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.cpp
// NAME:        CPPBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Feb 2, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.cpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.c.CModifier;
import java.util.Vector;
import java.util.Hashtable;

/**
  * Abstract class, ancestor of the builder classes for all TDD
  * declarations (datasets, factories, directives).
  * <P>Builds code for the class <I>declaration</I>: name of the class,
  * inherited classes, declaration of constructor and functions.
  * @author Cyrille Bareau
  * @see CPPDatasetBuilder
  * @see CPPDirectiveBuilder
  * @see CPPFactoryBuilder
  */

public abstract class CPPBuilder {

  /** Handle on the root of the generated C++ AST */
  public ASTTranslationUnit compilUnit;

  private ASTClassMemberList classMemberList;

  /** Name of the feature declaration in the source tdd class */
  protected String featureName;

  /** Name of the source tdd class */
  protected String unitName; 

  /** Name of the generated C++ class */
  protected String className;

  /** Handle on the generated declaration AST */
  protected ASTDeclaration declaration;

  private Hashtable includes;

  private boolean isDataset = false;
  private boolean isDirective = false;
  private boolean relinquish = false;

  static private boolean debug = false;


  public String getClassName() {
    return className;
  }

  /**
   * Builds the qualified name of the file that implements the feature
   * (dataset or factory) described by the parameters.
   * @param prefix "F_" for a factory, "S_" for a dataset.
   * @param TDDClassName the name of the TDD class defining the feature
   * @param name the feature name, qualified if imported.
   * @return the built file name
   */
  public static String makeQualifiedName(String prefix,
                                         String TDDClassName,
                                         String featureName) {

    return prefix + TDDClassName + "_" + featureName;
  }

  /**
    * Auxiliary constructor. Just initializates relinquish (last parameter)
    * before calling the main one.
    */
  public CPPBuilder (SimpleNode node, String prefix, String unit,
		     boolean b, ASTIncludeFileList incList) throws ParseException {
    this(node, prefix, unit, incList);
    relinquish = b;
  }


  /**
    * Main constructor.
    * @param node the current feature declaration node
    * @param unit the name of the source tdd class
    */
  public CPPBuilder (SimpleNode node, String prefix, String unit,
		     ASTIncludeFileList incList) throws ParseException {
    super();

    if (debug)
      System.out.println("CPPBuilder constructor " + node + " " + prefix + " " + unit);

    unitName = unit;
    int nbChildren = 0;

    compilUnit = new
      ASTTranslationUnit(ADLCPPParserTreeConstants.JJTTRANSLATIONUNIT);

    includes = new Hashtable();
    ASTIncludeFileList includeList;
    try {
      includeList = (ASTIncludeFileList)incList.clone();
    } catch (CloneNotSupportedException e) {
      includeList = incList;
    }
    ADLCPPGenericVisitor.parentChild(compilUnit, includeList, 0);
    for (int i = 0; i < incList.jjtGetNumChildren(); i++) {
      ASTIncludeFileDeclaration incl =
	(ASTIncludeFileDeclaration)incList.jjtGetChild(i);
      String incFileName = incl.getFileName();
      if (!includes.contains(incFileName))
	includes.put(incFileName, incFileName);
    }
    
    if (node instanceof ASTTDDDatasetDecl) {
      if ((node.jjtGetChild(0).jjtGetNumChildren() < 2) ||
          (node.jjtGetChild(0).jjtGetChild(1) instanceof ASTInitDeclaratorList) ||
          (node.jjtGetChild(0).jjtGetChild(1) instanceof ASTInitDeclarator))
        CPPTDDTransformVisitor.errorMessage("Incorrect dataset type",
                                            ((SimpleNode)node.jjtGetChild(0)).getLineNumber());
      Node decl = node.jjtGetChild(0).jjtGetChild(1);
      while (decl instanceof ASTDeclarator)
        decl = decl.jjtGetChild(1);
      if (decl.jjtGetNumChildren() > 1)
        CPPTDDTransformVisitor.errorMessage("No suffix for dataset declaration",
                                            ((SimpleNode)decl).getLineNumber());
      if (decl.jjtGetChild(0) instanceof ASTQualifiedId) {
        featureName = ((ASTQualifiedId)decl.jjtGetChild(0)).getName();
      }
      else CPPTDDTransformVisitor.errorMessage("No deletor for datasets",
                                               ((SimpleNode)decl.jjtGetChild(0)).getLineNumber());
      isDataset = true;
    }
    else if (node instanceof ASTTDDTestDirective) {
      // the parser names each directive
      featureName = ((ASTTDDTestDirective)node).getLabelName();
      //addInclude(unit + "TDD.cpp");
      isDirective = true;
    }
    else if (node instanceof ASTTDDFactoryDef) {
      Node fctDecl = node.jjtGetChild(1);
      while (!(fctDecl instanceof ASTFunctionDirectDeclarator)) {
        fctDecl = fctDecl.jjtGetChild(1);
      }
      featureName = ((ASTQualifiedId)fctDecl.jjtGetChild(0)).getName();
      addInclude(unit + "_CST.cpp");
    }
    else {
      // error
    }
    className = prefix + unitName + "_" + featureName;
    compilUnit.setName(className);
    compilUnit.setNodeExtension("cpp");

    if (debug)
      System.out.println("className : " + className + " _ unitName : " + unitName + " _ featureName " + featureName);
  }


  /**
    * Generic build method.
    * Creates the tree that is common to all builder-generated CPP files,
    * namely the root ASTTranslationUnit and some of its children:
    * class declaration (ASTDeclaration), name of the feature, name of
    * the inherited classes, declaration of constructor and functions.
    * @param parentName name of the inherited feature class (ADLLiteralSet, etc.)
    * @see #compilUnit
    * @see #className
    */
  protected void build(String parentName) {

    if (debug)
      System.out.println("CPPBuilder build");

    if (parentName.startsWith("F_") || isDirective)
      addInclude(parentName + ".cpp");
    else
      addInclude(parentName + ".hh");

    declaration = new
      ASTDeclaration(ADLCPPParserTreeConstants.JJTDECLARATION);
    ADLCPPGenericVisitor.parentChild(compilUnit, declaration, 1);

    ASTDeclarationSpecifiers declSpec = new
      ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
    ADLCPPGenericVisitor.parentChild(declaration, declSpec, 0);

    ASTClassSpecifier classSpec = new
      ASTClassSpecifier(ADLCPPParserTreeConstants.JJTCLASSSPECIFIER);
    ADLCPPGenericVisitor.parentChild(declSpec, classSpec, 0);
    classSpec.setClassKind("class");
    classSpec.setName(className);

    ASTBaseClause baseClause = new
      ASTBaseClause(ADLCPPParserTreeConstants.JJTBASECLAUSE);
    ADLCPPGenericVisitor.parentChild(classSpec, baseClause, 0);

    ASTBaseSpecifier baseSpec1 = new
      ASTBaseSpecifier(ADLCPPParserTreeConstants.JJTBASESPECIFIER);
    baseSpec1.setAccessSpec(CModifier.PUBLIC);
    baseSpec1.setName(parentName);
    ADLCPPGenericVisitor.parentChild(baseClause, baseSpec1, 0);

    if (parentName.startsWith("ADL")) {
      ASTBaseSpecifier baseSpec2 = new
        ASTBaseSpecifier(ADLCPPParserTreeConstants.JJTBASESPECIFIER);
      baseSpec2.setAccessSpec(CModifier.PUBLIC);
      baseSpec2.setName(unitName + "_CST");
      ADLCPPGenericVisitor.parentChild(baseClause, baseSpec2, 1);
    }

    classMemberList = new
      ASTClassMemberList(ADLCPPParserTreeConstants.JJTCLASSMEMBERLIST);
    ADLCPPGenericVisitor.parentChild(classSpec, classMemberList, 1);

    ASTAccessSpecifier member = new
      ASTAccessSpecifier(ADLCPPParserTreeConstants.JJTACCESSSPECIFIER);
    ADLCPPGenericVisitor.parentChild(classMemberList, member, 0);
    member.setAccessSpecifierName("public");
  }


  public void addPrivate() {

    if (debug)
      System.out.println("CPPBuilder addPrivate");

    ASTAccessSpecifier member = new
      ASTAccessSpecifier(ADLCPPParserTreeConstants.JJTACCESSSPECIFIER);
    ADLCPPGenericVisitor.parentChild(classMemberList, member, classMemberList.jjtGetNumChildren());
    member.setAccessSpecifierName("private");
  }


  public void addConstructorDecl(boolean isDirective) {

    if (debug)
      System.out.println("CPPBuilder addConstructorDecl " + isDirective);

    ASTCtorDefinition ctor = new
      ASTCtorDefinition(ADLCPPParserTreeConstants.JJTCTORDEFINITION);
    ADLCPPGenericVisitor.parentChild(classMemberList,
                                     ctor,
                                     classMemberList.jjtGetNumChildren());

    ASTCtorDeclarator ctorDecl = new
      ASTCtorDeclarator(ADLCPPParserTreeConstants.JJTCTORDECLARATOR);
    ADLCPPGenericVisitor.parentChild(ctor, ctorDecl, 0);

    ADLCPPGenericVisitor.parentChild(ctorDecl,
                                     ADLCPPGenericVisitor.makeQualifiedId(className),
                                     0);

    if (! isDirective) { // parameter: ADLTest *tc
      ASTParameterDeclaration paramDecl = new
        ParameterDeclarationBuilder
        ( ADLCPPGenericVisitor.makeQualifiedType("ADLTest"),
          new DeclaratorBuilder("PointerTo", "tc").generatedAST())
        .generatedAST();
      ADLCPPGenericVisitor.parentChild(ctorDecl, paramDecl, 1);
    }
  }


  public void addFunctionDecl(String type, boolean ptr, String methodName, boolean obj) {

    if (debug)
      System.out.println("CPPBuilder addFunctionDecl " + type + " " + methodName);

    // <type> [*]<name>(void **params[, void *obj]);

    ASTFunctionDefinition fct = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCPPGenericVisitor.parentChild(classMemberList, fct, classMemberList.jjtGetNumChildren());

    ADLCPPGenericVisitor.parentChild(fct,
                                     ADLCPPGenericVisitor.makeBuiltinType(type),
                                     0);

    ASTFunctionDirectDeclarator fctDirDeclTor = new
      ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);
    ADLCPPGenericVisitor.parentChild(fctDirDeclTor,
                                     ADLCPPGenericVisitor.makeQualifiedId(methodName),
                                     0);

    if (ptr) {
      ASTFunctionDeclarator fctDeclTor = new
        ASTFunctionDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDECLARATOR);
      ADLCPPGenericVisitor.parentChild(fct, fctDeclTor, 1);
      ADLCPPGenericVisitor.parentChild(fctDeclTor,
                                       ADLCPPGenericVisitor.makePtrOperator("PointerTo"),
                                       0);
      ADLCPPGenericVisitor.parentChild(fctDeclTor, fctDirDeclTor, 1);
    }
    else {
      ADLCPPGenericVisitor.parentChild(fct, fctDirDeclTor, 1);
    }

    // parameter of the fct: "void **params"

    ASTParameterDeclaration fctParamDecl = new
      ParameterDeclarationBuilder
      ( ADLCPPGenericVisitor.makeBuiltinType("void"),
        new DeclaratorBuilder("PointerTo", "PointerTo",
                              new DirectDeclaratorBuilder
                              ( "params").generatedAST())
        .generatedAST())
      .generatedAST();

    if (obj) {
    // parameter of the fct: "void **params, void *obj"
      ASTParameterDeclarationList fctParamsList = new
        ParameterDeclarationListBuilder(fctParamDecl)
        .addParameter
        ( new ParameterDeclarationBuilder
          ( ADLCPPGenericVisitor.makeBuiltinType("void"),
            new DeclaratorBuilder("PointerTo", "obj").generatedAST()))
        .generatedAST();

      ADLCPPGenericVisitor.parentChild(fctDirDeclTor, fctParamsList, 1);
    }
    else {
      ADLCPPGenericVisitor.parentChild(fctDirDeclTor, fctParamDecl, 1);
    }
  }


  public void addFunctionDecl(ASTFunctionDefinition fctDef) {

    if (debug)
      System.out.println("CPPBuilder addFunctionDecl: " + fctDef.jjtGetChild(1));

    ASTFunctionDefinition fctCopy = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);

    try {
      ADLCPPGenericVisitor.parentChild(fctCopy,
                                       (Node)((SimpleNode)fctDef.jjtGetChild(0)).clone(),
                                       0);
      ADLCPPGenericVisitor.parentChild(fctCopy,
                                       (Node)((SimpleNode)fctDef.jjtGetChild(1)).clone(),
                                       1);
    } catch (CloneNotSupportedException e) {}

    ADLCPPGenericVisitor.parentChild(classMemberList,
                                     fctCopy,
                                     classMemberList.jjtGetNumChildren());
  }


  public void addInclude(String inclF) {

    if (debug)
      System.out.println("CPPBuilder addInclude " + inclF);

    // waiting for Wassel to commit in development trunk
//     String inclFile = /*"\"" +*/ inclF /*+ "\""*/;
    String inclFile = "\"" + inclF + "\"";

    if (! (includes.contains(inclFile))) {
      includes.put(inclFile, inclFile);
      ASTIncludeFileDeclaration incl = new
        ASTIncludeFileDeclaration(ADLCPPParserTreeConstants.JJTINCLUDEFILEDECLARATION);
      incl.setFileName(inclFile);
      ADLCPPGenericVisitor.parentChild(compilUnit.jjtGetChild(0),
                                       incl,
                                       compilUnit.jjtGetChild(0).jjtGetNumChildren());
    }
  }

}

// ***********************************************************************
// History Log:
// $Log: CPPBuilder.java,v $
// Revision 1.10  1998/10/14 08:12:49  mallet
// Bugs fixed: pr/155, pr/156: removed useless constructor for anonymous
// test directives: the parser now names each directive.
//
// Revision 1.9  1998/06/30 13:12:21  mallet
// Replaced import of PropertyAccessException in
// ORG.opengroup.adl.parsers by an import on demand (.*).
//
// Revision 1.8  1998/06/12 15:14:50  mallet
// Moved method makeQualifiedName from NameAux to CPPBuilder.
// Fixed bug: list of include files in TDD file was not copied into
// dataset generated classes: added the IncludeFileList node from the TDD
// file as new argument to constructors.
//
// Revision 1.7  1998/04/01 16:49:51  mallet
// Replaced ADLNode by SimpleNode.
//
// Revision 1.6  1998/03/31 15:00:37  bareau
// Add "" to include file strings.
//
// Revision 1.5  1998/03/10 17:26:42  mallet
// Merged v1.1 with development trunk
//
// Revision 1.4.2.3  1998/03/09 10:28:36  bareau
// Cleaned up the code for error handling (wrote a static method
// "errorMessage" in CPPTDDTransformVisitor).
//
// Revision 1.4.2.2  1998/03/04 17:39:30  bareau
// Added new checks.
// Cleaned up the code.
//
// Revision 1.4.2.1  1998/03/03 10:59:54  bareau
// Set the debug flag to false.
// Bug fix in CPPDatasetExprVisitor: ASTTDDDatasetSingleton nodes were
// sent to the unparser.
//
// Revision 1.4  1998/02/27 13:07:28  bareau
// Updated copyright notice.
//
// Revision 1.3  1998/02/26 17:49:58  bareau
// ADLT2 TDD translator, first release. Version OK for the bank example.
//
// Revision 1.2  1998/02/12 16:03:40  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/12 13:48:20  bareau
// Classes used in the TDD translation. Added because of code freeze: this is a
// pre-pre-version (it does not even compile yet).
//
// ***********************************************************************
