// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/cpp/CPPDatasetBuilder.java,v $
// $Revision: 1.8 $
// $Date: 1998/06/30 13:12:22 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.cpp
// NAME:        CPPDatasetBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Feb 3, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.cpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.c.CModifier;
import ORG.opengroup.adl.symboltable.LType;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Creates a Java compilation unit AST that corresponds to a dataset
  * declaration.
  * <BR>Uses CPPDatasetExprVisitor; is used by CPPTDDTransformVisitor.
  * <P>Example: definition of dataset TEST in tdd class bankTest:
  * <BR>"dataset bank TEST = make_bank(DEPOSITS, BANKSIZE + DEPOSITS,
  * INT_VALUES);"
  * @author Cyrille Bareau
  * @see CPPDatasetExprVisitor
  * @see CPPTDDTransformVisitor
  */

public class CPPDatasetBuilder extends CPPBuilder {


  /** The prefix used to give names to the components of the dataset
    * expression of the dataset declaration. */
  private String strStackInit = "elt";

  /** The name of the method used to add components of the dataset expression.
    * (addElement for dataset concatenation, addParameter for factory call,
    * setValues for literal dataset). */
  private String addMethod;

  /** The name of the class the current dataset inherits from
      (set by method findType). */
  private String classType;

  private LType expectedType;

  static private boolean debug = false;


  /**
    * Finds the class type of the AST to build (ADLConcatenationSet,
    * ADLLiteralSet or a factory type).
    * <BR>Sets the addMethod field.
    * @param exp the dataset expression
    * @return the name (String) of the class type
    */
  private String findType(Node exp) {

    if (debug)
      System.out.println("CPPDatasetBuilder findType");

    if ((exp instanceof ASTTDDDatasetExpr) || (exp instanceof ASTTDDDatasetSingleton)) {
      addMethod = "addElement";
      addInclude(unitName + "_CST.cpp");
      return "ADLConcatenationSet";
    }
    if ((exp instanceof ASTTDDDatasetLiteral) || (exp instanceof ASTConstant)) {
      addMethod = "setValues";
      addInclude(unitName + "_CST.cpp");
      return "ADLLiteralSet";
    }
    if (exp instanceof ASTTDDFactoryCall) {
      addMethod = "addParameter";
      String TDDClassName = null;
      try {
        TDDClassName = ((String)((SimpleNode)exp.jjtGetChild(0))
                        .getProperty("enclosingConstructName"));
      }
      catch(PropertyAccessException e) { }
      return makeQualifiedName("F_",
			       TDDClassName,
			       ((ASTQualifiedId)exp.jjtGetChild(0)).getName());
    }
    // error
    return null;
  }


  /**
    * Single constructor.
    * <BR>Initializes the unit name (that begins with "S_")
    * @param node the dataset declaration node
    * @param unit source unit name
    */
  public CPPDatasetBuilder (ASTTDDDatasetDecl node, String unit,
			    ASTIncludeFileList incList) throws ParseException {
    super(node, "S_", unit, incList);
    build(node);
  }


  /**
    * The main method. Calls findType and the inherited CPPBuilder.build, then
    * calls makeConstructor.
    * @see CPPBuilder#build(String)
    * @param node the root node
    */
  private void build(ASTTDDDatasetDecl node) throws ParseException {

    if (debug)
      System.out.println("CPPDatasetBuilder build");

    Node dsExpr = node.jjtGetChild(1);
    classType = findType(dsExpr);

    expectedType = ((ASTDeclarationSpecifiers)node.jjtGetChild(0).jjtGetChild(0))
      .getType();

    if (((ASTDeclarationSpecifiers)node.jjtGetChild(0).jjtGetChild(0)).getModifier() !=
        0x0000)
      CPPTDDTransformVisitor.warningMessage("Type modifier(s) for dataset types will be ignored",
                                            node.getLineNumber());

    // call to inherited method build(String): see CPPBuilder
    build(classType);

    ADLCPPGenericVisitor.parentChild(compilUnit,
                                     makeConstructor(dsExpr),
                                     compilUnit.jjtGetNumChildren());
  } 


  /**
    * Creates the AST constructor of the dataset class.
    * <BR>Initializes a CPPDatasetExprVisitor and launches it on the passed
    * dataset expression.
    * Example for dataset TEST:
<PRE>
| 1|  S_bankTest_TEST::S_bankTest_TEST(ADLTest* tc) : F_bankTest_make_bank(tc, "TEST") {
| 2|     S_bankTest_DEPOSITS *elt_p1 = new S_bankTest_DEPOSITS(tc);
| 3|     ADLConcatenationSet *elt_p2 = new ADLConcatenationSet(tc, null);
| 4|     S_bankTest_BANKSIZE *elt_p2_p1 = new S_bankTest_BANKSIZE(tc);
| 5|     S_bankTest_DEPOSITS *elt_p2_p2 = new S_bankTest_DEPOSITS(tc);
| 6|     elt_p2->addElement(elt_p2_p1);
| 7|     elt_p2->addElement(elt_p2_p2);
| 8|     S_bankTest_INT_VALUES *elt_p3 = new S_bankTest_INT_VALUES(tc);
| 9|     addParameter(elt_p1);
|10|     addParameter(elt_p2);
|11|     addParameter(elt_p3);
|12|  }
</PRE>
    * The generated class S_bankTest_TEST inherits from F_bankTest_make_bank,
    * that implements the factory definition make_bank (F_bankTest_make_bank
    * inherits from ADLFactorySet).
    * <BR>In this example, lines 2 to 8 are generated by the DatasetExprVisitor
    * run on the dataset expression. The statements super(...) and addParameter
    * (lines 9-11) are generated by makeConstructor.
    * @param dsExpr the dataset expression of the dataset declaration
    * @return the ASTConstructorDeclaration node root of the constructor tree
    */
  private ASTCtorDefinition makeConstructor(Node dsExpr) throws ParseException {

    if (debug)
      System.out.println("CPPDatasetBuilder makeConstructor");

    addConstructorDecl(false);

    ASTCtorDefinition createdNode = new
      ASTCtorDefinition(ADLCPPParserTreeConstants.JJTCTORDEFINITION);

    ASTCtorDeclarator ctorDecl = new
      ASTCtorDeclarator(ADLCPPParserTreeConstants.JJTCTORDECLARATOR);
    ADLCPPGenericVisitor.parentChild(createdNode, ctorDecl, 0);

    ADLCPPGenericVisitor.parentChild(ctorDecl,
                                     ADLCPPGenericVisitor.makeQualifiedId(className + "::" + className),
                                     0);
    ADLCPPGenericVisitor.parentChild(ctorDecl,
                                     new ParameterDeclarationBuilder
                                     ( ADLCPPGenericVisitor.makeQualifiedType("ADLTest"),
                                       new DeclaratorBuilder("PointerTo", "tc").generatedAST())
                                     .generatedAST(),
                                     1);

    ASTCtorInitializer ctorInit = new
      ASTCtorInitializer(ADLCPPParserTreeConstants.JJTCTORINITIALIZER);
    ADLCPPGenericVisitor.parentChild(createdNode, ctorInit, 1);

    ASTSuperclassInit superclass = new
      ASTSuperclassInit(ADLCPPParserTreeConstants.JJTSUPERCLASSINIT);
    ADLCPPGenericVisitor.parentChild(ctorInit, superclass, 0);
    ADLCPPGenericVisitor.parentChild(superclass,
                                     ADLCPPGenericVisitor.makeQualifiedType(classType),
                                     0);
    ASTArgumentList argList = new
      ASTArgumentList(ADLCPPParserTreeConstants.JJTARGUMENTLIST);
    ADLCPPGenericVisitor.parentChild(superclass, argList, 1);
    ADLCPPGenericVisitor.parentChild(argList, 
                                  ADLCPPGenericVisitor.makeQualifiedId("tc"),
                                  0);

    ASTString str = new ASTString(ADLCPPParserTreeConstants.JJTSTRING);
    str.setName(featureName);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCPPGenericVisitor.parentChild(createdNode, body, 2);
    int nbChildren = 0;

    if (classType.startsWith("F_") || classType.equals("ADLLiteralSet")) {
      // 2nd superclass init "setLabel(<feature name>)"
      ASTStatement setLabelStmt = new
        ASTStatement(ADLCPPParserTreeConstants.JJTSTATEMENT);
      ADLCPPGenericVisitor.parentChild(body, setLabelStmt, nbChildren++);

      ASTPrimaryExpr setLabel = new
        ASTPrimaryExpr(ADLCPPParserTreeConstants.JJTPRIMARYEXPR);
      ADLCPPGenericVisitor.parentChild(setLabelStmt, setLabel, 0);
      ADLCPPGenericVisitor.parentChild(setLabel,
                                       ADLCPPGenericVisitor.makeIdExpr("setLabel"),
                                       0);
      ASTArgumentList argList2 = new
        ASTArgumentList(ADLCPPParserTreeConstants.JJTARGUMENTLIST);
      ADLCPPGenericVisitor.parentChild(setLabel, argList2, 1);
      ADLCPPGenericVisitor.parentChild(argList2, str, 0);
    }
    else {
      ADLCPPGenericVisitor.parentChild(argList, str, 1);
    }


    CPPDatasetExprVisitor dsExprVisitor =
      new CPPDatasetExprVisitor(strStackInit, "tc", false, this);
    dsExprVisitor.setExpectedType(expectedType);
    dsExpr.jjtAccept(dsExprVisitor, null);

    Vector datasetComp = (Vector)dsExprVisitor.getDatasetComponents().peek();
    int nbComp = ((Integer)dsExprVisitor.getNbComponents().peek()).intValue();

    for (Enumeration e = datasetComp.elements(); e.hasMoreElements(); ) {
      ADLCPPGenericVisitor.parentChild(body, (SimpleNode)e.nextElement(), nbChildren++);
    }

    String compName = strStackInit;

    for (int i = 1; i <= nbComp; i++) {
      compName = strStackInit + "_p" + i;

      ADLCPPGenericVisitor.parentChild(body, 
                                       new StatementBuilder
                                       ( new PostfixExprBuilder
                                         ( addMethod,
                                           new ArgumentListBuilder
                                           ( ADLCPPGenericVisitor.makeIdExpr(compName))
                                           .generatedAST())
                                         .generatedAST())
                                       .generatedAST(),
                                       nbChildren++);
    }

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: CPPDatasetBuilder.java,v $
// Revision 1.8  1998/06/30 13:12:22  mallet
// Replaced import of PropertyAccessException in
// ORG.opengroup.adl.parsers by an import on demand (.*).
//
// Revision 1.7  1998/06/12 15:42:58  mallet
// Simplified code due to changes in JDatasetExprVisitor.
// Added support for "use" clause.
//
// Revision 1.6  1998/03/11 14:54:11  bareau
// Added new warnings: type modifiers in tdd files (dataset definitions, factory/directive
// parameters) are ignored. New static method warningMessage in CPPTDDTransformVisitor.
// Improved support for casted expressions (there were useless double cast expressions
// in generated code).
//
// Revision 1.5  1998/03/10 17:26:43  mallet
// Merged v1.1 with development trunk
//
// Revision 1.4.2.2  1998/03/04 11:05:57  bareau
// The dataset and directive builders now send to the dataset expr visitor a
// string that represents the "expected" type of the dataset expression, so that
// this visitor generate an upcast if the actual type is different.
//
// Revision 1.4.2.1  1998/03/03 10:59:54  bareau
// Set the debug flag to false.
// Bug fix in CPPDatasetExprVisitor: ASTTDDDatasetSingleton nodes were
// sent to the unparser.
//
// Revision 1.4  1998/02/27 13:07:29  bareau
// Updated copyright notice.
//
// Revision 1.3  1998/02/26 17:49:58  bareau
// ADLT2 TDD translator, first release. Version OK for the bank example.
//
// Revision 1.2  1998/02/12 16:03:41  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/12 13:48:20  bareau
// Classes used in the TDD translation. Added because of code freeze: this is a
// pre-pre-version (it does not even compile yet).
//
// ***********************************************************************
