// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/cpp/CPPDatasetExprVisitor.java,v $
// $Revision: 1.9 $
// $Date: 1998/06/30 13:12:22 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.cpp
// NAME:        CPPDatasetExprVisitor
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Feb 10, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.cpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Dataset expression visitor.
  * Generates a vector of statements that correspond to the declaration
  * and initialization of a dataset expression (see examples in CPPDatasetBuilder
  * and CPPDirectiveBuilder).
  * <BR>Is used by CPPDatasetBuilder and by CPPDirectiveBuilder.
  * <P> For an overview of the algorithms used by this transformation, see the
  * ADL2.0 Translation System Design Specification, Chapter 5 (Code Generation),
  * paragraph 5.2.2.6.
  * @author Cyrille Bareau
  * @see CPPDatasetBuilder
  * @see CPPDirectiveBuilder
  */

public class CPPDatasetExprVisitor extends ADLCPPGenericVisitor {

  /** A stack of vector of statements. These statements correspond to the
    * declaration and initialization of the components of the current dataset
    * expression (these components are the arguments of a factory call or the
    * operands of a dataset concatenation). */
  private Stack datasetComponents;

  /** Number of components of the current dataset expression. */
  private Stack nbComponents;

  /** A stack of strings used to generate the identifiers for the
    * dataset expression components. */
  private Stack strStack;

  /** String for the testContext. <BR>"tc" from a dataset builder,
    * "this" from a directive builder. */
  private String testContext;

  /** Flag for anonymous dataset expressions. Anonymous dataset expressions
    * are arguments of a test directive and components of a dataset
    * expression. */
  private boolean anonymous;

  private CPPBuilder builderCaller;

  private LType expectedType = null;

  private static final boolean debug = false;


  /**
    * Single constructor.
    * @param strStackInit The string that initializes strStack.
    * @param tc The string that initializes testContext.
    */
  public CPPDatasetExprVisitor (String strStackInit, String tc, boolean anon, CPPBuilder caller) {
    super();
    testContext = tc;
    strStack = new Stack();
    strStack.push(strStackInit);
    datasetComponents = new Stack();
    datasetComponents.push(new Vector());
    nbComponents = new Stack();
    nbComponents.push(new Integer(0));
    anonymous = anon;
    builderCaller = caller;

    if (debug)
      System.out.println("*** new CPPDatasetExprVisitor: " + anonymous);
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetExpr.
    * <P><UL><LI><B>If</B> the current node has several children,
    * then this is a dataset concatenation: call makeConcatComponents.
    * <LI><B>ELSE</B> just re-transmits the visitor to the child node.</UL>
    * @see #makeConcatComponents
    * @param node the ASTTDD_DatasetExpr node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDDDatasetExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetExpr (" +
			 node.jjtGetNumChildren() + ")");

    if (node.jjtGetNumChildren() > 1) {
      makeConcatComponents(node, data);
    } else {
      node.jjtGetChild(0).jjtAccept(this, data);
    }
    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDDDatasetSingleton.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the name is not a dataset name, then it is a constant
    * and must be considered as a singleton literal: call makeSingletonLiteral.
    * <LI><B>ELSE</B><UL>
    * <LI>Get the qualified name of the class that implements the dataset.
    * <LI>Generate the AST for the variable declaration:
<PRE>
|  &lt;dataset qualified name> &lt;getComponentName()> =
|         new &lt;dataset qualified name>(&lt;testContext>);
</PRE>
    * <LI>Add this AST to the vector that is on top of stack datasetComponents.
    * <LI>Add "+1" to the integer on top of stack nbComponents.
    * </UL></UL>
    * @see CPPBuilder#makeQualifiedName(String, String)
    * @see #makeConcatComponents
    * @see #makeSingletonLiteral
    * @param node the ASTTDDDatasetSingleton node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDDDatasetSingleton node, Object data) throws ParseException {

    if (debug)
      System.out.println("visiting ASTTDDDatasetSingleton");

    SimpleNode child = (SimpleNode)node.jjtGetChild(0);
    if (child instanceof ASTIdExpr) {
      // Fix up ASTIdExpr by visiting it
      child.jjtAccept(this, data);

      // Can be a dataset or a constant
      boolean isDS = ((ASTIdExpr)child).isDataset;
      if (! isDS) { // field name
	makeSingletonLiteral((Typed)child, data);
      } else {
	// <datasetName>* <getComponentName()> = new <datasetName>(<testContext>);
	String datasetName = ((ASTIdExpr)child).getIdName();

	ASTDeclaration decl = new DeclarationBuilder
	  ( makeQualifiedType(datasetName),
	    new InitDeclaratorBuilder
	    ( new DeclaratorBuilder( "PointerTo", getComponentName()).generatedAST(),
	      new NewExprBuilder
	      ( datasetName,
		new ArgumentListBuilder
		( makeIdExpr(testContext)).generatedAST())
	      .generatedAST())
	    .generatedAST())
	  .generatedAST();
	
	((Vector)datasetComponents.peek()).addElement(decl/*createdNode*/);
	int nbComp = ((Integer)nbComponents.pop()).intValue();
	nbComponents.push(new Integer(nbComp+1));
      }
    } else {
      makeSingletonLiteral((Typed)child, data);
    }
    return null;
  }

    
  /**
    * Visit pattern for a node of type IdExpr.
    * <P><B>Algorithm</B><UL>
    * <li>Gets defining tdd class name
    * <LI><B>IF</B> the name is not a dataset name, then it is a
    *    constant:<ul>
    * <li>Sets name to
    *    <samp>&lt;defining tdd class name>_CST.&lt;constant name></samp>
    * </ul><LI><B>ELSE</B><UL>
    * <LI>Gets the qualified name of the dataset as
    *    <samp>S_&lt;defining tdd class name>_&lt;dataset name></samp>
    * <LI>Sets name to this qualified name
    * </UL></UL>
    * @see CPPBuilder#makeQualifiedName(String, String, String)
    * @param node the IdExpr node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTIdExpr node, Object data) throws ParseException {
    if (debug)
      System.out.println("visiting IdExpr : " + node.getIdName());

    boolean isDS = node.isDataset;
    String TDDClassName = null;
    String datasetName;
    try {
      TDDClassName = (String)node.getProperty("enclosingConstructName");
    }
    catch (PropertyAccessException e) { }
    
    if (! isDS) { // field name
      if (TDDClassName != null && !TDDClassName.equals("")) {
	// TDD field name
	datasetName = TDDClassName + "_CST::" + node.getIdName();
	
	builderCaller.addInclude(TDDClassName + "_CST.cpp");
	
	node.setIdName(datasetName);
      }
    } else {
      // general case: dataset name, not as root expression
      datasetName = CPPBuilder.makeQualifiedName("S_",
						 TDDClassName,
						 node.getIdName());
      
      builderCaller.addInclude(datasetName + ".cpp");

      node.setIdName(datasetName);
    }
    
    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDDDatasetLiteral.
    * <P><B>Algorithm</B><OL>
    * <LI>Creates an AST for the variable declaration:
<PRE>
|  ADLLiteralSet < &lt;getComponentName()> = new ADLLiteralSet(&lt;testContext>);
</PRE>
    * <LI>Pushes component name
    * <li>Adds created node to vector of components on top of datasetCompoents
    * <li>Visits children
    * </OL>
    * @param node the ASTTDD_DatasetLiteral node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDDDatasetLiteral node, Object data) throws ParseException {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetLiteral");

    // ADLLiteralSet * <getComponentName()> =
    //    new ADLLiteralSet(<testContext>, sizeof(<type of elts>));

    String componentName = getComponentName();
    pushName(componentName);

    LType typeElts = node.getType();
    LType expType = (data != null) ? ((LType)data) : expectedType;

    builderCaller.addInclude("ADLLiteralSet.hh");

    ASTDeclaration declSet = new DeclarationBuilder
      ( makeQualifiedType("ADLLiteralSet"),
        new InitDeclaratorBuilder
        ( new DeclaratorBuilder( "PointerTo", componentName).generatedAST(),
          new NewExprBuilder
          ( "ADLLiteralSet",
            new ArgumentListBuilder
            ( makeIdExpr(testContext))
            .addArgument(makeIdExpr("sizeof(" + expType.getName() + ")"))
            .generatedAST())
          .generatedAST())
        .generatedAST())
      .generatedAST();

    ((Vector)datasetComponents.peek()).addElement(declSet);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));

    if (typeElts.equals(expType)) {
      node.childrenAccept(this, data);
    }
    else {
      node.childrenAccept(this, expType);
    }

    popName();

    return null;
  }
  

  /**
    * Visit pattern for a node of type ASTTDD_DatasetMember.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the current node has only one child: not a range:<ul>
    * <ul>Generate an AST for statement:
<pre>
|  &lt;peekName()>->addValue(&lt;visit child>);
</pre>
    * </ul><LI><B>ELSE</B> <I>(the current node has two children: range)</I>:<ul>
    * <li>Generate ASTs for statements:
<PRE>
|  &lt;getComponentName()>->addRangeValues(&lt;visit child1>, &lt;visit child2>);
</PRE>
    * </UL>
    * <li>Created statement is added to the vector on top of stack
    * datasetComponents.
    * </ul>
    * @param node the ASTTDD_DatasetMember node to visit
    * @param data null <I>or</I> the string that represents the Java wrapper
    * type of the literal type if it is primitive.
    * @return nothing
    */
  public Object visit(ASTTDDDatasetMember node, Object data) throws ParseException {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetMember " +
			 node.jjtGetNumChildren());

    ASTStatement stmt;
    if (node.jjtGetNumChildren() == 1) { // simple case: no range
      // <peekName()>->addValue((<type elt>)<node.child1>);
      SimpleNode arg = (SimpleNode)node.jjtGetChild(0);

      // Fix up IdExpr children by visiting arg
      arg.jjtAccept(this, data);

      Node ch;
      LType actualType = ((Typed)arg).getType();

      if ((data != null) && ! (((LType)data).equals(actualType))) {
        ASTCastExpr castexpr = new ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);
        ASTTypeName type = new
          ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
        ASTDeclarationSpecifiers decl = new
          ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
        parentChild(decl, makeQualifiedType(((LType)data).getFullName()), 0);
        parentChild(type, decl, 0);
        parentChild(castexpr, type, 0);
        parentChild(castexpr, arg, 1);
        ch = castexpr;
      }
      else {
        ch = arg;
      }
      stmt = new StatementBuilder
        (new PrimaryExpressionBuilder
         ( makeIdExpr(peekName() + "->addValue"),
           new ArgumentListBuilder(ch)
           .generatedAST())
         .generatedAST())
        .generatedAST();

      ((Vector)datasetComponents.peek()).addElement(stmt);
    }
    else { // range -> integer values
      // <peekName()>->addRangeValues(<node.child1>, <node.child2>);
      SimpleNode arg1 = (SimpleNode)node.jjtGetChild(0);
      SimpleNode arg2 = (SimpleNode)node.jjtGetChild(1);
      Node ch1, ch2;

      // Fix up IdExpr children by visiting args
      arg1.jjtAccept(this, data);
      arg2.jjtAccept(this, data);

      if (data == null) {
        ch1 = arg1;
        ch2 = arg2;
      }
      else {
        LType actualType = ((Typed)arg1).getType();
        if (((LType)data).equals(actualType)) {
          ch1 = arg1;
        }
        else {
          ASTCastExpr castexpr = new ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);
          ASTTypeName type = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ASTDeclarationSpecifiers decl = new
            ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
          parentChild(decl, makeQualifiedType(((LType)data).getFullName()), 0);
          parentChild(type, decl, 0);
          parentChild(castexpr, type, 0);
          parentChild(castexpr, arg1, 1);
          ch1 = castexpr;
        }

        LType actualType2 = ((Typed)arg2).getType();
        if (((LType)data).equals(actualType2)) {
          ch2 = arg2;
        }
        else {
          ASTCastExpr castexpr = new ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);
          ASTTypeName type = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ASTDeclarationSpecifiers decl = new
            ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
          parentChild(decl, makeQualifiedType(((LType)data).getFullName()), 0);
          parentChild(type, decl, 0);
          parentChild(castexpr, type, 0);
          parentChild(castexpr, arg2, 1);
          ch2 = castexpr;
        }
      }
      stmt = new StatementBuilder
        (new PrimaryExpressionBuilder
         ( makeIdExpr(peekName() + "->addRangeValues"),
           new ArgumentListBuilder(ch1)
           .addArgument(ch2)
           .generatedAST())
         .generatedAST())
        .generatedAST();

      ((Vector)datasetComponents.peek()).addElement(stmt);
    }

    if (debug)
      System.out.println("end ASTDatasetMember");
    return null;
  }


  /**
    * Creates the code for a singleton literal dataset.
<PRE>
|  &#60;literal type&#62;[] &#60;componentName&#62;_a = new &#60;literal type&#62; [1];
|  &#60;componentName&#62;_a[0] = &#60;literal expression&#62;
|  ADLLiteralSet &#60;componentName&#62; = new
|     ADLLiteralSet(&#60;testContext&#62;, &#60;componentName&#62;_a);    
</PRE>
    */
  private void makeSingletonLiteral(Typed node, Object data) throws ParseException {

    if (debug)
      System.out.println("makeSingletonLiteral: " + node);

    // <type> <componentName>_a[] = { <node> };

    String componentName = getComponentName();
    LType typ = (data == null) ? expectedType : ((LType)data);

    ASTDeclarationSpecifiers declSpec = new
      ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
    parentChild(declSpec, makeQualifiedType(typ.getFullName()), 0);

    ASTDeclaratorSuffixes arraySuffix = new
      ASTDeclaratorSuffixes(ADLCPPParserTreeConstants.JJTDECLARATORSUFFIXES);
    arraySuffix.setArraySize(1);

    ASTDirectDeclarator declTor = new DirectDeclaratorBuilder
      ( componentName + "_a",
        arraySuffix)
      .generatedAST();

    ASTInitializer init = new
      ASTInitializer(ADLCPPParserTreeConstants.JJTINITIALIZER);
    parentChild(init, (Node)node, 0);

    ASTDeclaration declArray = new DeclarationBuilder
      ( declSpec,
        new InitDeclaratorBuilder(declTor, init).generatedAST())
      .generatedAST();


    ((Vector)datasetComponents.peek()).addElement(declArray);


    // ADLLiteralSet* <componentName> = new ADLLiteralSet(<testContext>, <componentName>_a, 1);

    builderCaller.addInclude("ADLLiteralSet.hh");

    ASTDeclaration declSet = new DeclarationBuilder
      ( makeQualifiedType("ADLLiteralSet"),
        new InitDeclaratorBuilder
        ( new DeclaratorBuilder( "PointerTo", componentName).generatedAST(),
          new NewExprBuilder
          ( "ADLLiteralSet",
            new ArgumentListBuilder
            ( makeIdExpr(testContext))
            .addArgument(makeIdExpr(componentName + "_a"))
            .addArgument(makeConstant("1"))
            .generatedAST())
          .generatedAST())
        .generatedAST())
      .generatedAST();


    ((Vector)datasetComponents.peek()).addElement(declSet);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));
  }


  /**
    * Create the code for a dataset concatenation expression.
    * <UL><LI><B>IF</B> anonymous expression
<PRE>
|  ADLConcatenationSet &#60;getComponentName()&#62; = new
|     ADLConcatenationSet(&#60;testContext&#62;, null);
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushes a new Integer(0) on nbComponents
    * <li>Keeps value of <samp>anonymous</samp> and set it to <samp>true</samp>
    * </ul><li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Copies each of the components in this vector into vector on top
    *     of stack
    * <LI><B>IF</B> anonymous expression,<UL>
    * <LI>add statements to add the child element to the dataset:
<PRE>
|  &#60;peekName()&#62;->addElement(&#60;peekName()&#62;_p&#60;i&#62;);
</PRE>
    * </UL></UL>
    * <li>Restores previous value of <samp>anonymous</samp>
    * <li><B>IF</B> anonymous expression,<UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul></ul>
    * @param node the ASTTDDDatasetExpr node to visit
    * @param data unused
    */
  private void makeConcatComponents(ASTTDDDatasetExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("makeConcatComponents: " + node);

    if (anonymous) {

      if (debug)
        System.out.println("makeConcatComponents: anonymous");

      // ADLConcatenationSet *<getComponentName()> = new ADLConcatenationSet(<testContext>) //, NULL);
      String componentName = pushName(getComponentName());

      builderCaller.addInclude("ADLConcatenationSet.hh");

      ASTDeclaration declSet = new DeclarationBuilder
        ( makeQualifiedType("ADLConcatenationSet"),
          new InitDeclaratorBuilder
          ( new DeclaratorBuilder( "PointerTo", componentName).generatedAST(),
            new NewExprBuilder
            ( "ADLConcatenationSet",
              new ArgumentListBuilder
              ( makeIdExpr(testContext))
              //              .addArgument(makeIdExpr("NULL"))
              .generatedAST())
            .generatedAST())
          .generatedAST())
        .generatedAST();


      ((Vector)datasetComponents.peek()).addElement(declSet);
      int nbComp = ((Integer)nbComponents.pop()).intValue();
      nbComponents.push(new Integer(nbComp+1));
      nbComponents.push(new Integer(0));
    }

    // visit elements of concatenation
    boolean myAnonymous = anonymous;

    anonymous = true;
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      if (debug)
	System.out.println("makeConcatComponents: element " + (i + 1));

      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, data);

      Vector elemComponents = (Vector)datasetComponents.pop();
      
      // add components resulting from visit of element to the list 
      // of components of the dataset expression
      Vector thetop = (Vector)datasetComponents.peek();
      for (Enumeration e = elemComponents.elements(); 
	   e.hasMoreElements(); ) {
	thetop.addElement((SimpleNode)e.nextElement());
      }

      if (myAnonymous) {
        // <peekName()>->addElement(<peekName()>_p<i+1>);
	ASTStatement stmt = new StatementBuilder
          (new PrimaryExpressionBuilder
           ( makeIdExpr(peekName() + "->addElement"),
             new ArgumentListBuilder
             ( makeIdExpr(peekName() + "_p" + (i + 1))).generatedAST())
           .generatedAST())
          .generatedAST();

        ((Vector)datasetComponents.peek()).addElement(stmt);
      }
    }
    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }
  }


  /**
    * Visit pattern for a node of type ASTTDDFactoryCall.
    * <UL><LI><B>IF</B> anonymous expression:<ul>
    * <li>Adds statement to create a factory set instance:
<PRE>
|  &#60;factory name&#62; &#60;getComponentName()&#62; = new
|     &#60;factory name&#62;(&#60;testContext&#62;, null);
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushed a new Integer(0) on nbComponents
    * <li>Keeps value of <samp>anonymous</samp> and set it to <samp>true</samp>
    * </ul><li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Copies each of the components in this vector into vector on top
    *     of stack
    * <LI><B>IF</B> anonymous expression,<UL>
    * <LI>add statements to add the child element to the dataset:
<PRE>
|  &#60;peekName()&#62;->addParameter(&#60;peekName()&#62;_p&#60;i&#62;);
</PRE>
    * </UL></UL>
    * <li>Restores previous value of <samp>anonymous</samp>
    * <li><B>IF</B> anonymous expression,<UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul></ul>
    *
    * @param node the ASTTDDFactoryCall node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDDFactoryCall node, Object data) throws ParseException {

    if (debug)
      System.out.println("visiting ASTTDDFactoryCall");

    String TDDClassName = null;
    try {
      TDDClassName = (String)((SimpleNode)node.jjtGetChild(0))
                      .getProperty("enclosingConstructName");
    }
    catch(PropertyAccessException e) { }

    String factName =
      CPPBuilder.makeQualifiedName("F_", TDDClassName,
				   ((ASTQualifiedId)node.jjtGetChild(0)).getName());

    if (anonymous) {
  
      if (debug)
        System.out.println("ASTTDDFactoryCall: " + factName +
			   " : anonymous");

      // <factName> *<getComponentName()> = new <factName>(<testContext>)//, NULL);
      String componentName = pushName(getComponentName());

      builderCaller.addInclude(factName + ".cpp");

      ASTDeclaration declSet = new DeclarationBuilder
        ( makeQualifiedType(factName),
          new InitDeclaratorBuilder
          ( new DeclaratorBuilder( "PointerTo", componentName).generatedAST(),
            new NewExprBuilder
            ( factName,
              new ArgumentListBuilder
              ( makeIdExpr(testContext))
              //              .addArgument(makeIdExpr("NULL"))
              .generatedAST())
            .generatedAST())
          .generatedAST())
        .generatedAST();


      ((Vector)datasetComponents.peek()).addElement(declSet);
      int nbComp = ((Integer)nbComponents.pop()).intValue();
      nbComponents.push(new Integer(nbComp+1));
      nbComponents.push(new Integer(0));
    }

    // visit parameters of factory call
    LParameter[] signParams = node.getSignature().getParameters();
    int numParameters = node.jjtGetNumChildren() - 1;
    boolean myAnonymous = anonymous;

    anonymous = true;

    for (int i = 1; i <= numParameters; i++) {
      if (debug)
	System.out.println("ASTTDD_FactoryCall: " + factName +
			   " : parameter " + i);
    
      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, signParams[i-1].getType());

      Vector paramComponents = (Vector)datasetComponents.pop();

      // add components resulting from visit of parameter to the list 
      // of components of the dataset expression
      Vector thetop = (Vector)datasetComponents.peek();
      for (Enumeration e = paramComponents.elements(); 
	   e.hasMoreElements(); ) {
	thetop.addElement((SimpleNode)e.nextElement());
      }
	
      if (myAnonymous) {
        // <peekName()>->addParameter(<peekName()>_p<i>);
	ASTStatement stmt = new StatementBuilder
          (new PrimaryExpressionBuilder
           ( makeIdExpr(peekName() + "->addParameter"),
             new ArgumentListBuilder
             ( makeIdExpr(peekName() + "_p" + i)).generatedAST())
           .generatedAST())
          .generatedAST();
        ((Vector)datasetComponents.peek()).addElement(stmt);
      }
    }
    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }

    return null;
  }


  /**
   * Set expected type
   */
  public void setExpectedType(LType type) {
    expectedType = type;
  }


  /**
    * @return datasetComponents
    */
  public Stack getDatasetComponents() {
    return datasetComponents;
  }


  /**
    * @return nbComponents
    */
  public Stack getNbComponents() {
    return nbComponents;
  }

  private String getComponentName() {
    return (String)strStack.peek() +
      "_p" +
      (((Integer)nbComponents.peek()).intValue() + 1);
  }

  private String pushName(String levelName) {
    strStack.push(levelName);
    return levelName;
  }

  private String popName() {
    return (String)strStack.pop();
  }

  private String peekName() {
    return (String)strStack.peek();
  }
}

// ***********************************************************************
// History Log:
// $Log: CPPDatasetExprVisitor.java,v $
// Revision 1.9  1998/06/30 13:12:22  mallet
// Replaced import of PropertyAccessException in
// ORG.opengroup.adl.parsers by an import on demand (.*).
//
// Revision 1.8  1998/06/18 10:32:10  mallet
// Fixed incorrect use of type short name instead of full one (FQN).
// Made use of makeQualidifedType systematic.
//
// Revision 1.7  1998/06/12 15:41:29  mallet
// Major revision of the code to fix reference to "use"d constants:
// 	- removed useless fields,
// 	- simplified code,
// 	- moved processing from visit of TDD_DatasetSingleton to visit
// 	  of IdExpr (according to change in the ADL/C++ grammar for
// 	  TDD_DatasetSingleton)
// 	- removed visit methods for Constant,
// 	- added case of "use"d constants in visit of IdExpr,
// 	- revised component name management: new methods
// 	  getComponentName(), peekName(), pushName(), popName(),
// 	- updated javadoc comments.
//
// Revision 1.6  1998/03/11 14:54:11  bareau
// Added new warnings: type modifiers in tdd files (dataset definitions, factory/directive
// parameters) are ignored. New static method warningMessage in CPPTDDTransformVisitor.
// Improved support for casted expressions (there were useless double cast expressions
// in generated code).
//
// Revision 1.5  1998/03/10 17:26:43  mallet
// Merged v1.1 with development trunk
//
// Revision 1.4.2.4  1998/03/06 14:35:29  bareau
// Set debug flag to false.
//
// Revision 1.4.2.3  1998/03/04 17:45:23  bareau
// When visiting a factory call, we now get the factory signature so that
// for each parameter the expected type is sent along with the visit to the
// actual parameter node (this is the expected type that serves as the basis
// for the dataset construction).
//
// Revision 1.4.2.2  1998/03/04 11:05:57  bareau
// The dataset and directive builders now send to the dataset expr visitor a
// string that represents the "expected" type of the dataset expression, so that
// this visitor generate an upcast if the actual type is different.
//
// Revision 1.4.2.1  1998/03/03 10:59:55  bareau
// Set the debug flag to false.
// Bug fix in CPPDatasetExprVisitor: ASTTDDDatasetSingleton nodes were
// sent to the unparser.
//
// Revision 1.4  1998/02/27 13:07:29  bareau
// Updated copyright notice.
//
// Revision 1.3  1998/02/26 17:49:59  bareau
// ADLT2 TDD translator, first release. Version OK for the bank example.
//
// Revision 1.2  1998/02/12 16:03:41  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/12 13:48:20  bareau
// Classes used in the TDD translation. Added because of code freeze: this is a
// pre-pre-version (it does not even compile yet).
//
// ***********************************************************************
