// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/cpp/CPPFactoryBuilder.java,v $
// $Revision: 1.12 $
// $Date: 1998/06/30 13:12:23 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.cpp
// NAME:        CPPFactoryBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Feb 6, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.cpp;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.Vector;


/**
  * Class that creates the C++ translation unit AST that corresponds to a
  * factory declaration.
  * <BR>Is used by CPPTDDTransformVisitor.
  * <P>Example: factory "make_bank" defined in tdd class "bankTest":
<PRE>
  factory bank make_bank (int maxAccts, int numAcctsActive, int initDeposit) {
    bank ret = new bank(maxAccts);
    for (int x=0;x < maxAccts && x < numAcctsActive;x++) {
      try { bankAccount ba = ret.open_account(initDeposit); }
      catch (bankException be) { tdd_end_case(); }
    }
    return ret;
  }
  // optional relinquish clause
  relinquish (bank b) {
    if (numAcctsActive > 0) {
      try { b.close_all_accounts(); }
      catch(Throwable e) {}
    }
  }
</PRE>
  * @author Cyrille Bareau
  * @see CPPTDDTransformVisitor
  */

public class CPPFactoryBuilder extends CPPBuilder {

  /** The return type declaration specifiers of the factory */
  private ASTDeclarationSpecifiers factDeclSpec;

  /** The name of the factory return type */
  private String factTypeName;

  /** The name of the factory */
  private String factName;

  /** The declarator of the factory */
  private ASTFunctionDeclarator factDeclTor;

  /** The declarator of the factory */
  private ASTFunctionDirectDeclarator factDirDeclTor;

  /** The formal parameters of the factory */
  private ASTParameterDeclarationList factParams;

  /** The "throws list" of the factory */
  private ASTExceptionSpec factThrowsList;

  /** The body of the factory */
  private ASTCompoundStatement factBlk;

  /** The body of the (optional) relinquish block */
  private ASTCompoundStatement relBlock = null;

  /** The parameter of the relinquish block (for non void factories) */
  private ASTParameterDeclaration relParam = null;

  /** Flag "the factory is void: no returned object" */
  private boolean voidFactory;

  /** Flag "no relinquish clause in the factory declaration */
  private boolean noRelinquish = true;

  static final private boolean debug = false;


  /**
    * Single constructor.
    * @param node The factory declaration AST node
    * @param imports The import declarations
    * @param unit The name of the adlclass file
    */
  public CPPFactoryBuilder(ASTTDDFactoryDef node, String unit,
			   ASTIncludeFileList incList) throws ParseException {

    super(node, "F_", unit, incList);
    build(node);
  }


  /**
    * Initializes the data and launches the methods code generation.
    */
  private void build(ASTTDDFactoryDef node) throws ParseException {

    if (debug)
      System.out.println("CPPFactoryBuilder build");

    build("ADLFactorySet");

    factDeclSpec = (ASTDeclarationSpecifiers)node.jjtGetChild(0);
    
    if (factDeclSpec.getModifier() != 0x0000)
      CPPTDDTransformVisitor.warningMessage("Type modifier(s) for factory return type will be ignored",
                                            factDeclSpec.getLineNumber());

    factBlk = (ASTCompoundStatement)node.jjtGetChild(2);
    factTypeName = factDeclSpec.getType().getName();

    factDeclTor = null;
    Node declTor = node.jjtGetChild(1);
    if (declTor instanceof ASTFunctionDirectDeclarator) {
      factDirDeclTor = (ASTFunctionDirectDeclarator)declTor;
    }
    else {
      factDeclTor = (ASTFunctionDeclarator)declTor;
      while (! (declTor instanceof ASTFunctionDirectDeclarator)) {
        factTypeName += ((ASTPtrOperator)declTor.jjtGetChild(0)).getPtrstring();
        declTor = declTor.jjtGetChild(1);
      }
      factDirDeclTor = (ASTFunctionDirectDeclarator)declTor;
    }

//     voidFactory = (! (factDeclTor.jjtGetChild(0) instanceof ASTPtrOperator)) &&
//       (factDeclSpec.jjtGetChild(0) instanceof ASTBuiltinTypeSpec) &&
//       (((ASTBuiltinTypeSpec)factDeclSpec.jjtGetChild(0)).getTypeName().equals("void"));

    voidFactory = (factDeclTor == null) && (factDeclSpec.getType().equals(CPPBuiltinType.voidT));

    factName = ((ASTQualifiedId)factDirDeclTor.jjtGetChild(0)).getName();
    if (factDirDeclTor.jjtGetChild(1) instanceof ASTParameterDeclarationList) {
      factParams = (ASTParameterDeclarationList)factDirDeclTor.jjtGetChild(1);
    }
    else if (factDirDeclTor.jjtGetChild(1) instanceof ASTParameterDeclaration) {
      factParams = new ParameterDeclarationListBuilder
        ( (ASTParameterDeclaration)factDirDeclTor.jjtGetChild(1)).generatedAST();
    }
    else factParams = null;

    if (factParams != null) {
      for (int i = factParams.jjtGetNumChildren()-1; i >= 0; i--) {
        if (((ASTDeclarationSpecifiers)factParams.jjtGetChild(i).jjtGetChild(0)).getModifier() != 0x0000)
          CPPTDDTransformVisitor.warningMessage("Type modifier(s) for factory parameters will be ignored",
                                                node.getLineNumber());
      }
    }


    if (node.jjtGetNumChildren() > 3) {
      noRelinquish = false;
      if (node.jjtGetNumChildren() > 4) {
        relParam = (ASTParameterDeclaration)node.jjtGetChild(3);
        relBlock = (ASTCompoundStatement)node.jjtGetChild(4);
      }
      else {
        relBlock = (ASTCompoundStatement)node.jjtGetChild(3);
      }
    }
    
    if (factDirDeclTor.jjtGetChild(factDirDeclTor.jjtGetNumChildren()-1)
        instanceof ASTExceptionSpec) {
      factThrowsList = (ASTExceptionSpec)factDirDeclTor.jjtGetChild(factDirDeclTor.jjtGetNumChildren()-1);
    }
   
    int nbChildren = compilUnit.jjtGetNumChildren();

    ADLCPPGenericVisitor.parentChild(compilUnit, makeConstructor(), nbChildren++);
    ADLCPPGenericVisitor.parentChild(compilUnit, makeFactoryProvide(), nbChildren++);
    if (! noRelinquish) {
      ADLCPPGenericVisitor.parentChild(compilUnit, makeFactoryRelinquish(), nbChildren++);
    }
    addPrivate();
    ADLCPPGenericVisitor.parentChild(compilUnit, makeProvide(), nbChildren++);
    if (! noRelinquish) {
      ADLCPPGenericVisitor.parentChild(compilUnit, makeRelinquish(), nbChildren++);
    }
  }


  /**
    * Creates the factory constructor.
    * Example for factory make_bank:
<PRE>
|  public  F_bankTest_make_bank(ADLTest tc, String label) {
|     super(tc, label);
|     factoryName = "make_bank";
|  }
</PRE>
    * @return the root of the constructor (ASTConstructorDeclaration).
    */
  private ASTCtorDefinition makeConstructor() {

    if (debug)
      System.out.println("CPPFactoryBuilder makeConstructor");

    addConstructorDecl(false);

    ASTCtorDefinition createdNode = new
      ASTCtorDefinition(ADLCPPParserTreeConstants.JJTCTORDEFINITION);

    ASTCtorDeclarator ctorDecl = new
      ASTCtorDeclarator(ADLCPPParserTreeConstants.JJTCTORDECLARATOR);
    ADLCPPGenericVisitor.parentChild(createdNode, ctorDecl, 0);

    ADLCPPGenericVisitor.parentChild(ctorDecl,
                                  ADLCPPGenericVisitor.makeQualifiedId(className + "::" + className),
                                  0);

    ADLCPPGenericVisitor.parentChild(ctorDecl,
                                     new ParameterDeclarationBuilder
                                     ( ADLCPPGenericVisitor.makeQualifiedType("ADLTest"),
                                       new DeclaratorBuilder("PointerTo", "tc").generatedAST())
                                     .generatedAST(),
                                     1);

    ASTCtorInitializer ctorInit = new
      ASTCtorInitializer(ADLCPPParserTreeConstants.JJTCTORINITIALIZER);
    ADLCPPGenericVisitor.parentChild(createdNode, ctorInit, 1);

    ASTSuperclassInit superclass = new
      ASTSuperclassInit(ADLCPPParserTreeConstants.JJTSUPERCLASSINIT);
    ADLCPPGenericVisitor.parentChild(ctorInit, superclass, 0);
    ADLCPPGenericVisitor.parentChild(superclass,
                                     ADLCPPGenericVisitor.makeQualifiedType("ADLFactorySet"),
                                     0);
    ASTArgumentList argList = new
      ASTArgumentList(ADLCPPParserTreeConstants.JJTARGUMENTLIST);
    ADLCPPGenericVisitor.parentChild(superclass, argList, 1);
    ADLCPPGenericVisitor.parentChild(argList, 
                                     ADLCPPGenericVisitor.makeQualifiedId("tc"),
                                     0);

    ASTString str = new ASTString(ADLCPPParserTreeConstants.JJTSTRING);
    str.setName(featureName);
    ADLCPPGenericVisitor.parentChild(argList, str, 1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCPPGenericVisitor.parentChild(createdNode, body, 2);

    return createdNode;
  }


  /**
    * Creates the method "factory_provide".
    * Example for factory make_bank:
<PRE>
|  public  Object factory_provide(Object[] p) throws Throwable {
|     int elt0 = ((Integer)p[0]).intValue();
|     int elt1 = ((Integer)p[1]).intValue();
|     int elt2 = ((Integer)p[2]).intValue();
|     return provide_make_bank(elt0, elt1, elt2);
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTFunctionDefinition makeFactoryProvide() throws ParseException {

    if (debug)
      System.out.println("CPPFactoryBuilder makeFactoryProvide");

    addFunctionDecl("void", true, "factory_provide", false);

    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCPPGenericVisitor.parentChild(createdNode,
                                     ADLCPPGenericVisitor.makeQualifiedType("void *"),
                                     0);

    ASTFunctionDeclarator fctDecl = new
      ASTFunctionDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDECLARATOR);
    ADLCPPGenericVisitor.parentChild(createdNode, fctDecl, 1);

    ASTFunctionDirectDeclarator fctDirDecl = new
      ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);
    ADLCPPGenericVisitor.parentChild(fctDecl, fctDirDecl, 0);
    ADLCPPGenericVisitor.parentChild
      (fctDirDecl,
       ADLCPPGenericVisitor.makeQualifiedId(className + "::factory_provide"),
       0);
    ADLCPPGenericVisitor.parentChild
      (fctDirDecl,
       new ParameterDeclarationBuilder
       ( ADLCPPGenericVisitor.makeBuiltinType("void"),
         new DeclaratorBuilder("PointerTo", "PointerTo",
                               new DirectDeclaratorBuilder("params").generatedAST())
         .generatedAST())
       .generatedAST(),
       1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCPPGenericVisitor.parentChild(createdNode, body, 2);
    int nbChildren = 0;

    
    ASTDeclarationSpecifiers ptype; String pname; ASTArraySuffix brack;
    ASTCastExpr castExpr; ASTPostfixExpr prim;
    ASTParenthExpr parenth; boolean isPrimitiveType = false;
    ASTConstant arrlit; Node declTor = null, declTorBis, param;
    ASTInitDeclarator initDecl; ASTTypeName typeName;
    ASTAbstractDeclarator abstr, abstrBis; ASTUnaryExpr unary;
    ASTBinaryArithExpr shift; ASTDotAccessSuffix reset;
    ASTString str; int nbp = 0; String paramName = "";

    if (factParams != null) {
      ASTParameterDeclarationList paramList;
      try {
        paramList = (ASTParameterDeclarationList)factParams.clone();
      } catch (CloneNotSupportedException e) { paramList = null; }

      nbp = paramList.jjtGetNumChildren();
      for (int i = 0; i < nbp; i++) {
        ptype = (ASTDeclarationSpecifiers)paramList.jjtGetChild(i).jjtGetChild(0);
        pname = "elt" + i;

        if (paramList.jjtGetChild(i).jjtGetNumChildren() == 1)
          CPPTDDTransformVisitor.errorMessage("Parameters of a factory must be named",
                                       ((SimpleNode)paramList.jjtGetChild(i)).getLineNumber());
        if (paramList.jjtGetChild(i).jjtGetChild(1) instanceof ASTAssignmentExpr)
          CPPTDDTransformVisitor.errorMessage("Parameters of a factory must not be assigned",
                                       ((SimpleNode)paramList.jjtGetChild(i).jjtGetChild(1)).getLineNumber());

        param = paramList.jjtGetChild(i).jjtGetChild(1);
        if (param instanceof ASTDirectDeclarator) {
          if (param.jjtGetNumChildren() > 1)
            CPPTDDTransformVisitor.errorMessage("No suffix for factory parameters",
                                         ((SimpleNode)param).getLineNumber());
          if (param.jjtGetChild(0) instanceof ASTQualifiedId) {
            isPrimitiveType = ptype.jjtGetChild(0) instanceof ASTBuiltinTypeSpec;
            declTor = ADLCPPGenericVisitor.makeQualifiedId(pname);
            paramName = ((ASTQualifiedId)param.jjtGetChild(0)).getFullName();
          }
          else CPPTDDTransformVisitor.errorMessage("No deletor or nested types in factory parameters",
                                            ((SimpleNode)param.jjtGetChild(0)).getLineNumber());
        }
        else {
          if (param instanceof ASTAbstractDeclarator) {
            CPPTDDTransformVisitor.errorMessage("no abstract declarator in the parameters of a factory",
                                         ((SimpleNode)param).getLineNumber());
          }
          declTor = param;
          isPrimitiveType = false;
          paramName += ((ASTPtrOperator)param.jjtGetChild(0)).getPtrstring();
        }

        arrlit = new
          ASTConstant(ADLCPPParserTreeConstants.JJTCONSTANT);
        arrlit.setCstVal(String.valueOf(i));
        brack = new
          ASTArraySuffix(ADLCPPParserTreeConstants.JJTARRAYSUFFIX);
        ADLCPPGenericVisitor.parentChild(brack, arrlit, 0);
        prim = new PostfixExprBuilder
          ( ADLCPPGenericVisitor.makeIdExpr("params"),
            brack).generatedAST();

        if (isPrimitiveType) {
          // <ptype> elt<i> = *(<ptype> *)params[<i>];

          unary = new
            ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
          unary.setOp("*");

          initDecl = new InitDeclaratorBuilder(declTor, unary)
            .generatedAST();

          castExpr = new
            ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);
          ADLCPPGenericVisitor.parentChild(unary, castExpr, 0);

          typeName = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ADLCPPGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCPPGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
          ADLCPPGenericVisitor.parentChild(typeName, abstr, 1);

          ADLCPPGenericVisitor.parentChild(abstr,
                                           ADLCPPGenericVisitor.makePtrOperator("PointerTo"),
                                           0);

          ADLCPPGenericVisitor.parentChild(castExpr, prim, 1);

          ADLCPPGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(ptype, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }
        else { // object type
          // <ptype> <id> = (<ptype>)params[<i>];

          castExpr = new
            ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);

          typeName = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ADLCPPGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCPPGenericVisitor.parentChild(typeName, ptype, 0);

          declTorBis = declTor;
          abstr = new
            ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
          ADLCPPGenericVisitor.parentChild(typeName, abstr, 1);

          while (declTorBis instanceof ASTDeclarator) {
            ADLCPPGenericVisitor.parentChild(abstr, declTorBis.jjtGetChild(0), 0);
            abstrBis = new
              ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
            ADLCPPGenericVisitor.parentChild(abstr, abstrBis, 1);
            abstr = abstrBis;
            paramName += ((ASTPtrOperator)declTorBis.jjtGetChild(0)).getPtrstring();
            declTorBis = declTorBis.jjtGetChild(1);
          }
          if (declTorBis.jjtGetChild(0) instanceof ASTQualifiedId) {
            paramName += ((ASTQualifiedId)declTorBis.jjtGetChild(0)).getFullName();
            ADLCPPGenericVisitor.parentChild(declTorBis,
                                             ADLCPPGenericVisitor.makeQualifiedId(pname),
                                             0);
          }
          else CPPTDDTransformVisitor.errorMessage("No nested type in parameter",
                                            ((SimpleNode)declTorBis.jjtGetChild(0)).getLineNumber());

          ADLCPPGenericVisitor.parentChild(castExpr, prim, 1);

          initDecl = new InitDeclaratorBuilder(declTor, castExpr)
            .generatedAST();

          ADLCPPGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(ptype, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }
    

        // if (testContext->debugMode)
	//   testContext->infoline(adlstr.reset() << "\t<ptype> <name> = " << );
        shift = new
          ASTBinaryArithExpr(ADLCPPParserTreeConstants.JJTBINARYARITHEXPR);
        shift.addOp("<<"); shift.addOp("<<");

        reset = new
          ASTDotAccessSuffix(ADLCPPParserTreeConstants.JJTDOTACCESSSUFFIX);
        ADLCPPGenericVisitor.parentChild(reset,
                                         ADLCPPGenericVisitor.makeIdExpr("reset"),
                                         0);
        ADLCPPGenericVisitor.parentChild(reset,
                                         new ArgumentListBuilder().generatedAST(),
                                         1);

        ADLCPPGenericVisitor.parentChild(shift,
                                         new PostfixExprBuilder
                                         ( ADLCPPGenericVisitor.makeIdExpr("adlstr"),
                                           reset).generatedAST(),
                                         0);

        str = new
          ASTString(ADLCPPParserTreeConstants.JJTSTRING);

        str.setName("\\t " + ptype.getType().getName() + " " + paramName + " = ");
        ADLCPPGenericVisitor.parentChild(shift, str, 1);

        ADLCPPGenericVisitor.parentChild(shift,
                                         ADLCPPGenericVisitor.makeIdExpr(pname), // $$$$
                                         2);

	// the if statement
	ASTIfNode ifDebugMode = new
	  ASTIfNode(ADLCPPParserTreeConstants.JJTIFNODE);
	ADLCPPGenericVisitor.parentChild(ifDebugMode,
					 new PostfixExprBuilder
                                           ( "testContext",
					     new RefAccessSuffixBuilder
                                             ("debugMode")
					     .generatedAST())
					 .generatedAST(), 0);
	ADLCPPGenericVisitor.parentChild(ifDebugMode,
					 new StatementBuilder
                                         ( new PostfixExprBuilder
                                           ( "testContext",
                                             new RefAccessSuffixBuilder
                                             ("infoline", new ArgumentListBuilder(shift)
                                              .generatedAST())
                                             .generatedAST())
                                           .generatedAST())
                                         .generatedAST(),
                                         1);
        ADLCPPGenericVisitor.parentChild(body, ifDebugMode, nbChildren++);
      }
    }



    // return (void *)provide_<factName>(elt<0>, elt<1>, ...);
    // or, if void factory:
    // provide_<factName>(elt<0>, elt<1>, ...); return new Object();

    ArgumentListBuilder argb = new ArgumentListBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(ADLCPPGenericVisitor.makeIdExpr("elt" + i));
    }
    ASTPrimaryExpr last = new PrimaryExpressionBuilder
      ( ADLCPPGenericVisitor.makeIdExpr("provide_" + factName),
        argb.generatedAST())
      .generatedAST();

    if (voidFactory) {
      ADLCPPGenericVisitor.parentChild
        (body, new StatementBuilder(last).generatedAST(), nbChildren++);
    }
    else {
      ASTCastExpr voidStar = new
        ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);

      ASTTypeName voidName = new
        ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
      ADLCPPGenericVisitor.parentChild(voidStar, voidName, 0);

      ASTDeclarationSpecifiers voidSpec = new
        ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
      ADLCPPGenericVisitor.parentChild(voidSpec,
                                       ADLCPPGenericVisitor.makeIdExpr("void"),
                                       0);
      ADLCPPGenericVisitor.parentChild(voidName, voidSpec, 0);

      ASTAbstractDeclarator voidAbstr = new
        ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
      ADLCPPGenericVisitor.parentChild(voidName, voidAbstr, 1);

      ADLCPPGenericVisitor.parentChild(voidAbstr,
                                       ADLCPPGenericVisitor.makePtrOperator("PointerTo"),
                                       0);

      ASTReturnStmt ret = new
        ASTReturnStmt(ADLCPPParserTreeConstants.JJTRETURNSTMT);
      ADLCPPGenericVisitor.parentChild(ret, voidStar, 0);
      ADLCPPGenericVisitor.parentChild(voidStar, last, 1);
      ADLCPPGenericVisitor.parentChild(body, ret, nbChildren++);
    }

    return createdNode;
  }


  /**
    * Creates the method "provide_&#60;factory name&#62;".
    * Example for factory make_bank:
<PRE>
|  public bank provide_make_bank(int maxAccts, int numAcctsActive, int initDeposit)
|     throws Throwable {
|     &#60;factory body&#62;
|  }
</PRE>
    * <B>Note:</B> if the factory returns a primitive type, one must create a
    * FactoryVisitor to replace returned values in the factory body by values
    * of the corresponding Java wrapper class.
    * @see FactoryVisitor
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTFunctionDefinition makeProvide () throws ParseException {

    if (debug)
      System.out.println("CPPFactoryBuilder makeProvide");

    ASTDeclarationSpecifiers decl = null; ASTFunctionDeclarator declTor = null;
    ASTFunctionDirectDeclarator dirDeclTor = null;

    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);

    boolean primitiveType = !voidFactory &&
      (factDeclSpec.jjtGetChild(0) instanceof ASTBuiltinTypeSpec) && (factDeclTor == null);
    if (primitiveType) {
      decl = new
        ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
      ADLCPPGenericVisitor.parentChild(decl,
                                       ADLCPPGenericVisitor.makeBuiltinType("void"),
                                       0);
      declTor = new
        ASTFunctionDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDECLARATOR);
      dirDeclTor = new
        ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);
      ADLCPPGenericVisitor.parentChild(declTor, ADLCPPGenericVisitor.makePtrOperator("PointerTo"), 0);
      ADLCPPGenericVisitor.parentChild(declTor, dirDeclTor, 1);
      ADLCPPGenericVisitor.parentChild(dirDeclTor, factParams, 1);

      ADLCPPGenericVisitor.parentChild(createdNode, decl, 0);
      ADLCPPGenericVisitor.parentChild(createdNode, declTor, 1);
    }
    else {
      ADLCPPGenericVisitor.parentChild(createdNode, factDeclSpec, 0);
      if (factDeclTor != null) {
        try {
          declTor = (ASTFunctionDeclarator)factDeclTor.clone();
        } catch (CloneNotSupportedException e) {}
        Node tmp = declTor;
        while (tmp instanceof ASTFunctionDeclarator)
          tmp = tmp.jjtGetChild(1);
        dirDeclTor = (ASTFunctionDirectDeclarator)tmp;
        ADLCPPGenericVisitor.parentChild(createdNode, declTor, 1);
      }
      else {
        try {
          dirDeclTor = (ASTFunctionDirectDeclarator)factDirDeclTor.clone();
        } catch (CloneNotSupportedException e) {}
        ADLCPPGenericVisitor.parentChild(createdNode, dirDeclTor, 1);
      }
    }

    ADLCPPGenericVisitor.parentChild(dirDeclTor,
                                     ADLCPPGenericVisitor.makeQualifiedId
                                     ("provide_" + factName),
                                     0);

    addFunctionDecl(createdNode);

    ADLCPPGenericVisitor.parentChild(dirDeclTor,
                                     ADLCPPGenericVisitor.makeQualifiedId
                                     (className + "::provide_" + factName),
                                     0);

    if (primitiveType) {

      if (debug)
        System.out.println("CPPFactoryBuilder makeProvide 2: primitiveType");

      // visit factory body to modify return statements
      FactoryVisitor factV = new FactoryVisitor();
      Node visited = (Node)factBlk.jjtAccept(factV, null);
      ADLCPPGenericVisitor.parentChild(createdNode, visited, 2);
    }
    else { // nothing to modify in the factory body

      if (debug)
        System.out.println("CPPFactoryBuilder makeProvide 2: non primitive");
      ADLCPPGenericVisitor.parentChild(createdNode, factBlk, 2);
    }

    return createdNode;
  }


  /**
    * Creates the method "factory_relinquish".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  public  boolean factory_relinquish(Object[] p, Object obj) {
|     int elt0 = ((Integer)p[0]).intValue();
|     int elt1 = ((Integer)p[1]).intValue();
|     int elt2 = ((Integer)p[2]).intValue();
|     bank theObj = (bank)obj;
|     return relinquish_make_bank(elt0, elt1, elt2, theObj);
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTFunctionDefinition makeFactoryRelinquish() throws ParseException {

    if (debug)
      System.out.println("CPPFactoryBuilder makeFactoryRelinquish");

    addFunctionDecl("bool", false, "factory_relinquish", true);

    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);
    ADLCPPGenericVisitor.parentChild(createdNode,
                                     ADLCPPGenericVisitor.makeBuiltinType("bool"),
                                     0);

    ASTFunctionDeclarator fctDecl = new
      ASTFunctionDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDECLARATOR);
    ADLCPPGenericVisitor.parentChild(createdNode, fctDecl, 1);

    ASTFunctionDirectDeclarator fctDirDecl = new
      ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);
    ADLCPPGenericVisitor.parentChild(fctDecl, fctDirDecl, 0);
    ADLCPPGenericVisitor.parentChild
      (fctDirDecl,
       ADLCPPGenericVisitor.makeQualifiedId(className + "::factory_relinquish"),
       0);
    ADLCPPGenericVisitor.parentChild
      (fctDirDecl,
       new ParameterDeclarationListBuilder
       ( new ParameterDeclarationBuilder
         ( ADLCPPGenericVisitor.makeBuiltinType("void"),
           new DeclaratorBuilder("PointerTo", "PointerTo",
                                 new DirectDeclaratorBuilder("params").generatedAST())
           .generatedAST())
         .generatedAST())
       .addParameter
       ( new ParameterDeclarationBuilder
         ( ADLCPPGenericVisitor.makeBuiltinType("void"),
           new DeclaratorBuilder("PointerTo",
                                 new DirectDeclaratorBuilder("obj").generatedAST())
           .generatedAST())
         .generatedAST())
       .generatedAST(),
       1);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCPPGenericVisitor.parentChild(createdNode, body, 2);
    int nbChildren = 0;
    
    ASTDeclarationSpecifiers ptype; String pname; ASTArraySuffix brack;
    ASTCastExpr castExpr; ASTPostfixExpr prim;
    ASTParenthExpr parenth; boolean isPrimitiveType;
    ASTConstant arrlit; Node declTor; ASTUnaryExpr unary;
    ASTInitDeclarator initDecl; ASTTypeName typeName;
    ASTAbstractDeclarator abstr, abstrBis; Node declTorBis;
    int nbp = 0;

    if (factParams != null) {
      nbp = factParams.jjtGetNumChildren();
      for (int i = 0; i < nbp; i++) {
        ptype = (ASTDeclarationSpecifiers)factParams.jjtGetChild(i).jjtGetChild(0);
        pname = "elt" + i;

        if (factParams.jjtGetChild(i).jjtGetChild(1) instanceof ASTDirectDeclarator) {
          // primitive type
          isPrimitiveType = ptype.jjtGetChild(0) instanceof ASTBuiltinTypeSpec;
          declTor = new DirectDeclaratorBuilder(pname).generatedAST();
        }
        else {
          isPrimitiveType = false;
          declTor = factParams.jjtGetChild(i).jjtGetChild(1);
        }

        arrlit = new
          ASTConstant(ADLCPPParserTreeConstants.JJTCONSTANT);
        arrlit.setCstVal(String.valueOf(i));
        brack = new
          ASTArraySuffix(ADLCPPParserTreeConstants.JJTARRAYSUFFIX);
        ADLCPPGenericVisitor.parentChild(brack, arrlit, 0);
        prim = new PostfixExprBuilder
          ( ADLCPPGenericVisitor.makeIdExpr("params"),
            brack).generatedAST();

        if (isPrimitiveType) {
          // <ptype> elt<i> = *(<ptype> *)params[<i>];

          unary = new
            ASTUnaryExpr(ADLCPPParserTreeConstants.JJTUNARYEXPR);
          unary.setOp("*");

          initDecl = new InitDeclaratorBuilder(declTor, unary)
            .generatedAST();

          castExpr = new
            ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);
          ADLCPPGenericVisitor.parentChild(unary, castExpr, 0);

          typeName = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ADLCPPGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCPPGenericVisitor.parentChild(typeName, ptype, 0);

          abstr = new
            ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
          ADLCPPGenericVisitor.parentChild(typeName, abstr, 1);

          ADLCPPGenericVisitor.parentChild(abstr,
                                           ADLCPPGenericVisitor.makePtrOperator("PointerTo"),
                                           0);

          ADLCPPGenericVisitor.parentChild(castExpr, prim, 1);

          ADLCPPGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(ptype, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }
        else { // object type
          // <ptype> <id> = (<ptype>)params[<i>];

          castExpr = new
            ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);

          initDecl = new InitDeclaratorBuilder(declTor, castExpr)
            .generatedAST();

          typeName = new
            ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
          ADLCPPGenericVisitor.parentChild(castExpr, typeName, 0);
          ADLCPPGenericVisitor.parentChild(typeName, ptype, 0);

          declTorBis = declTor;
          abstr = new
            ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
          ADLCPPGenericVisitor.parentChild(typeName, abstr, 1);

          while (declTorBis instanceof ASTDeclarator) {
            ADLCPPGenericVisitor.parentChild(abstr, declTorBis.jjtGetChild(0), 0);
            abstrBis = new
              ASTAbstractDeclarator(ADLCPPParserTreeConstants.JJTABSTRACTDECLARATOR);
            ADLCPPGenericVisitor.parentChild(abstr, abstrBis, 1);
            abstr = abstrBis;
            declTorBis = declTorBis.jjtGetChild(1);
          }
          if (declTorBis.jjtGetChild(0) instanceof ASTQualifiedId) {
            ADLCPPGenericVisitor.parentChild(declTorBis,
                                             ADLCPPGenericVisitor.makeQualifiedId(pname),
                                             0);
          }
          else CPPTDDTransformVisitor.errorMessage("No deletor or nested types in factory parameters",
                                                   ((SimpleNode)declTorBis.jjtGetChild(0)).getLineNumber());

          ADLCPPGenericVisitor.parentChild(castExpr,
                                        new PostfixExprBuilder
                                        ( ADLCPPGenericVisitor.makeIdExpr("params"), prim)
                                        .generatedAST(),
                                        1);

          ADLCPPGenericVisitor.parentChild(body,
                                           new DeclarationBuilder(ptype, initDecl)
                                           .generatedAST(),
                                           nbChildren++);
        }
      }
    }


    // <factType> theObj = (<factType>)obj;

    if (! voidFactory) {

      castExpr = new
        ASTCastExpr(ADLCPPParserTreeConstants.JJTCASTEXPR);

      ASTInitDeclarator initDecl2 = new InitDeclaratorBuilder("theObj", castExpr)
        .generatedAST();

      // WARNING: it still not clear whether factTypeName is appropriate
      // or not to use as a QualifiedType name for nested class for
      // instance. Does it include the ScopeOverride name?
      typeName = new
        ASTTypeName(ADLCPPParserTreeConstants.JJTTYPENAME);
      ADLCPPGenericVisitor.parentChild(castExpr, typeName, 0);
      ADLCPPGenericVisitor.parentChild(typeName,
                                       ADLCPPGenericVisitor.makeQualifiedType(factTypeName),
                                       0);

      ADLCPPGenericVisitor.parentChild(castExpr,
                                       ADLCPPGenericVisitor.makeIdExpr("obj"),
                                       1);

      ADLCPPGenericVisitor.parentChild(body,
                                       new DeclarationBuilder
                                       ( ADLCPPGenericVisitor.makeQualifiedType(factTypeName),
                                         initDecl2)
                                       .generatedAST(),
                                       nbChildren++);
    }


    // return relinquish_<factName>(elt<0>, elt<1>, ..., theObj);
    // or, if void factory:
    // return provide_<factName>(elt<0>, elt<1>, ...);

    ArgumentListBuilder argb = new ArgumentListBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(ADLCPPGenericVisitor.makeIdExpr("elt" + i));
    }
    if (! voidFactory) {
      argb.addArgument(ADLCPPGenericVisitor.makeIdExpr("theObj"));
    }

    ASTPrimaryExpr last = new PrimaryExpressionBuilder
      ( ADLCPPGenericVisitor.makeIdExpr("relinquish_" + factName),
        argb.generatedAST())
      .generatedAST();

    ASTReturnStmt ret = new
      ASTReturnStmt(ADLCPPParserTreeConstants.JJTRETURNSTMT);
    ADLCPPGenericVisitor.parentChild(ret, last, 0);
    ADLCPPGenericVisitor.parentChild(body, ret, nbChildren++);

    return createdNode;
  }


  /**
    * Creates the method "relinquish_&#60;factory name&#62;".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  public  boolean relinquish_make_bank(int maxAccts, int numAcctsActive,
|                                       int initDeposit, bank b) {
|     &#60;relinquish body&#62;
|     return true;
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTFunctionDefinition makeRelinquish() {

    if (debug)
      System.out.println("CPPFactoryBuilder makeRelinquish");

    ASTFunctionDefinition createdNode = new
      ASTFunctionDefinition(ADLCPPParserTreeConstants.JJTFUNCTIONDEFINITION);

    ASTDeclarationSpecifiers decl = new
      ASTDeclarationSpecifiers(ADLCPPParserTreeConstants.JJTDECLARATIONSPECIFIERS);
    ADLCPPGenericVisitor.parentChild(decl,
                                     ADLCPPGenericVisitor.makeBuiltinType("bool"),
                                     0);
    ASTFunctionDirectDeclarator dirDeclTor = new
      ASTFunctionDirectDeclarator(ADLCPPParserTreeConstants.JJTFUNCTIONDIRECTDECLARATOR);

    ASTParameterDeclarationList paramsList = null;
    try {
      paramsList = (ASTParameterDeclarationList)factParams.clone();
    }
    catch(CloneNotSupportedException e){}

    if (relParam != null) {
      ADLCPPGenericVisitor.parentChild(paramsList,
                                       relParam,
                                       paramsList.jjtGetNumChildren());
    }
    ADLCPPGenericVisitor.parentChild(dirDeclTor, paramsList, 1);
    ADLCPPGenericVisitor.parentChild(createdNode, decl, 0);
    ADLCPPGenericVisitor.parentChild(createdNode, dirDeclTor, 1);

    ADLCPPGenericVisitor.parentChild(dirDeclTor,
                                     ADLCPPGenericVisitor.makeQualifiedId
                                     ("relinquish_" + factName),
                                     0);

    addFunctionDecl(createdNode);

    ADLCPPGenericVisitor.parentChild(dirDeclTor,
                                     ADLCPPGenericVisitor.makeQualifiedId
                                     (className + "::relinquish_" + factName),
                                     0);

    ASTCompoundStatement body = new
      ASTCompoundStatement(ADLCPPParserTreeConstants.JJTCOMPOUNDSTATEMENT);
    ADLCPPGenericVisitor.parentChild(createdNode, body, 2);

    // nothing to modify in the factory body
    ADLCPPGenericVisitor.parentChild(body, relBlock, 0);

    // return true

    ASTConstant truelit = new
      ASTConstant(ADLCPPParserTreeConstants.JJTCONSTANT);
    truelit.setCstVal(String.valueOf(true));

    ASTReturnStmt ret = new
      ASTReturnStmt(ADLCPPParserTreeConstants.JJTRETURNSTMT);
    ADLCPPGenericVisitor.parentChild(ret, truelit, 0);
    ADLCPPGenericVisitor.parentChild(body, ret, 1);

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: CPPFactoryBuilder.java,v $
// Revision 1.12  1998/06/30 13:12:23  mallet
// Replaced import of PropertyAccessException in
// ORG.opengroup.adl.parsers by an import on demand (.*).
//
// Revision 1.11  1998/06/18 10:32:11  mallet
// Fixed incorrect use of type short name instead of full one (FQN).
// Made use of makeQualidifedType systematic.
//
// Revision 1.10  1998/06/12 15:16:59  mallet
// Fixed bug: list of include files from the TDD file was not copied into
// generated factory classes: added new argument to constructor according
// to change in CPPBuilder.
//
// Revision 1.9  1998/06/05 08:34:40  mallet
// Added generation of "if (testContext->debugMode)" around each factory
// parameter display to reduce the size of the generated journal.
//
// Revision 1.8  1998/04/28 12:39:24  cinard
// Updated code to take into account replacement of CSymbolTableMgr,
// CSymbolFactory and CBuiltinType by CPPSymbolTableMgr, CPPSymbolFactory
// and CPPBuiltinType.
//
// Revision 1.7  1998/03/16 16:28:11  bareau
// Skipped the generation of a useless CompoundStatement in body of method
// "provide_<factory name>".
//
// Revision 1.6  1998/03/11 14:54:12  bareau
// Added new warnings: type modifiers in tdd files (dataset definitions, factory/directive
// parameters) are ignored. New static method warningMessage in CPPTDDTransformVisitor.
// Improved support for casted expressions (there were useless double cast expressions
// in generated code).
//
// Revision 1.5  1998/03/10 17:26:44  mallet
// Merged v1.1 with development trunk
//
// Revision 1.4.2.4  1998/03/09 15:09:56  bareau
// Bug fix: pointers to pointers to ... were incorrectly reproduced in the
// generated code.
//
// Revision 1.4.2.3  1998/03/09 10:28:37  bareau
// Cleaned up the code for error handling (wrote a static method
// "errorMessage" in CPPTDDTransformVisitor).
//
// Revision 1.4.2.2  1998/03/04 17:39:43  bareau
// Added new checks.
// Cleaned up the code.
//
// Revision 1.4.2.1  1998/03/03 10:59:56  bareau
// Set the debug flag to false.
// Bug fix in CPPDatasetExprVisitor: ASTTDDDatasetSingleton nodes were
// sent to the unparser.
//
// Revision 1.4  1998/02/27 13:07:30  bareau
// Updated copyright notice.
//
// Revision 1.3  1998/02/26 17:50:00  bareau
// ADLT2 TDD translator, first release. Version OK for the bank example.
//
// Revision 1.2  1998/02/12 16:03:42  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/12 13:48:21  bareau
// Classes used in the TDD translation. Added because of code freeze: this is a
// pre-pre-version (it does not even compile yet).
//
// ***********************************************************************
