// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JADLTransformVisitor.java,v $
// $Revision: 1.25 $
// $Date: 1998/10/15 18:14:09 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.j
// NAME:        JADLTransformVisitor
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import ORG.opengroup.adl.symboltable.j.JBuiltinType;
import ORG.opengroup.adl.symboltable.LType;
import java.util.*;

/**
  * Transformation visitor.  Must change an ADL AST into a Java AST.
  * <P> For an overview of the algorithms used by this transformation,
  * see the ADL2.0 Translation System Design Specification, Chapter 5
  * (Code Generation), paragraph 5.2.1.
  * @see JGenericVisitor
  * @see ADLJavaParserVisitor
  * @author Cyrille Bareau
  */

public class JADLTransformVisitor extends JGenericVisitor {


  class BoolContext {

    private String tmp;
    private boolean b;

    BoolContext(String t, boolean bb) {
      tmp = t;
      b = bb;
    }

    String getString() {
      return tmp;
    }

    boolean getBool() {
      return b;
    }
  }


  /* handles */

  /** Stack of BoolContexts for lazy evaluation of logical operations */
  private Stack boolContextStack;

  /** Handle for the global prolog block */
  private ASTBlock globalProlog;

  /** Handle for the global epilog block */
  private ASTBlock globalEpilog;

  /** Handle for the local prolog block */
  private ASTBlock localProlog;

  /** Handle for the local epilog block */
  private ASTBlock localEpilog;

  /** Handle for inline declarations (vector) */
  private Vector inlineDeclaration;

  /** Handle for the declaration of temporaries for
    * call-state expressions (vector) */
  private Vector preDeclaration;

  /** Handle for the initialization of temporaries for
    * call-state expressions (vector) */
  private Vector saveBlock;

  /** Handle for "pre-assertions" : block statements that precede assertions
    * (local var declaration for instance). */
  private Vector preAssertion;

  /** Tags of the current assertion */
  private String[] assTags;

  /** Handle on the declaration of the inline that corresponds
    * to the currently encountered inline call */
  private ASTADL_InlineDeclaration currentInline;

  /** Handle on the formal parameters of the declaration of the
    * currently encountered inline call */
  private ASTFormalParameters inlineFormalParams;

  /** Handle on the actual parameters of the inline call that is
    * currently being inlined */
  private ASTArgumentList inlineActualParams;

  /** Auxiliary class for code generation */
  private ACOauxiliary ACO;


  /** Index for call-state tmp variables */
  private int savIndex;

  /** Index for preAssertions */ 
  private int preAssIndex;

  /** Flag set when defining a constructor */
  private boolean isInConstructor;

  /** Flag set when visiting an assertion */
  private boolean isInAssertion;

  /** Flag set when visiting a binding */
  private boolean isInBinding;

  /** Flag set when visiting a call-state expression */
  private boolean isInCallState;

  /** Flag set when visiting an ADL try/catch statement */
  private boolean isInTry;

  /** Flag set when visiting a quantified assertion */
  private boolean isInQuantified;

  /** Flag set when the currently visited expression can be decomposed
    * for the TET debug mode.
    * Must always be equal to 
    * <PRE>isInAssertion &&
    *   !(isInCallState || isInTry || isInQuantified || isInBinding)</PRE>
    */
  private boolean isDecomposable;

  /** Flag set when visiting the substituted inline body,
    * while transforming an inline call */
  private boolean inlining;


  /** Flag set when there is a syntax error */
  private boolean syntaxError = false;


  private final static boolean debug = false;


  /**
    * Generates a local temporary variable declaration. The variable is
    * initialized with its type default value, either as a preDeclaration
    * (call-state case) or as a preAssertion (used in the try/catch case, where
    * declared values must be defaultly initialized before being given values
    * in the try/catch blocks).
    * <BR>In the preAssertion case, the string parameter (that represents the
    * expression) is stored as a property of the ASTLocalVariableDeclaration.
    * @param saveblk "is a call-state".
    * @param type type of the temporary variable to generate.
    * @param theString token string of the rhs expression, for a preAssertion.
    * @return the string name of the generated temporary variable.
    * @see #preDeclaration
    * @see #preAssertion
    */
  private String makeSaveDecl(boolean saveblk, LType type, String theString) {

    if (debug)
      System.out.println("makeSaveDecl : " + saveblk + " " + type + " " + theString);

    String lhs = (saveblk ?
                  ("__ADL_savTmp_" + (savIndex++)) :
                  ("__ADL_tmp_" + (preAssIndex++)));

    // <type> <lhs> = <default value>;

    ASTType typeNode = TypeAux.LType2ASTType(type);
    Node defaultValue = TypeAux.getDefaultValue(typeNode);

    ASTLocalVariableDeclaration locVarDecl = 
      new LocalVariableDeclarationBuilder(typeNode, lhs, defaultValue).generatedAST();
    ASTBlockStatement blkStmt =
      new BlockStatementBuilder(locVarDecl).generatedAST();

    if (saveblk) {
      preDeclaration.addElement(blkStmt);
    }
    else {
      try {
        locVarDecl.setProperty("tokenString", theString);
      }
      catch(PropertyAccessException e) {}
      preAssertion.addElement(blkStmt);
    }
    return lhs;
  }


  /**
    * Generates a local temporary variable declaration, initialized with the
    * passed expression, as a saveBlock.
    * @param type type of the temporary variable to generate.
    * @param expr initialization expression.
    * @return the string name of the generated tmp variable.
    * @see #saveBlock
    * @see #preAssertion
    */
  private String makeSaveBlock(LType type, Node expr) {

    if (debug)
      System.out.println("makeSaveBlock : " + type +
                         " " + expr);

    // <type> <lhs> = <expr>;

    String lhs = makeSaveDecl(true, type, "");
    ASTStatement blkStmt = new StatementBuilder
      ( new AssignmentBuilder(lhs, "=", expr) )
      .generatedAST();
    saveBlock.addElement(blkStmt);
    return lhs;
  }


  /**
    * Generates a local temporary variable declaration, initialized with the
    * passed expression, as a preAssertion.
    * <BR>The string parameter (that represents the code of the expression)
    * is stored as a property of the ASTLocalVariableDeclaration.
    * @param type type of the tmp variable to generate.
    * @param expr initialization expression.
    * @param theString token string of the rhs expression
    * @return the string name of the generated tmp variable.
    * @see #preAssertion
    */
  private String makePreAssertion(ASTType type, Node expr, String theString) {

    if (debug)
      System.out.println("makePreAssertion : " + type +
                         " " + expr + " " + theString);

    // <type> <lhs> = <expr>;

    String lhs = "__ADL_tmp_" + (preAssIndex++);
    ASTLocalVariableDeclaration locVarDecl;

    if (boolContextStack.empty()) {
      locVarDecl = new
        LocalVariableDeclarationBuilder(type, lhs, expr).generatedAST();
    }
    else {
      // when there is 1 boolContext <"foo", bool> in the stack, the expression
      // is guarded by a conditional expression:
      // foo ? <expr> : <default value> // if bool is true ("true branch")
      // foo ? : <default value> : <expr> // otherwise ("else branch")
      // the <default value> is the expression's type default value (0 for numeric, etc)

      // when several boolContext are stacked, conditional expressions
      // must be nested (from the bottom to the top of the stack)

      ASTParenthExpression parenth = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
      ASTConditionalExpression cond = new
        ASTConditionalExpression(ADLJavaParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(parenth, cond, 0);
      for (Enumeration enum = boolContextStack.elements(); enum.hasMoreElements(); ) {
        BoolContext bc = (BoolContext)enum.nextElement();
        parentChild(cond, makeName(bc.getString()), 0);
        parentChild(cond,
                    // default value for un-evaluated branch
                    TypeAux.getDefaultValue(type),
                    (bc.getBool() ? 2 : 1));
        if (enum.hasMoreElements()) {
          ASTParenthExpression parenth2 = new
            ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
          ASTConditionalExpression cond2 = new
            ASTConditionalExpression(ADLJavaParserTreeConstants.JJTCONDITIONALEXPRESSION);
          parentChild(parenth2, cond2, 0);
          parentChild(cond, parenth2, (bc.getBool() ? 1 : 2));
          cond = cond2;
        }
        else {
          parentChild(cond, expr, (bc.getBool() ? 1 : 2));
        }
      }
      locVarDecl = new
        LocalVariableDeclarationBuilder(type, lhs, parenth).generatedAST();
    }

    try {
      locVarDecl.setProperty("tokenString", theString);
    }
    catch(PropertyAccessException e) {}
    ASTBlockStatement blkStmt =
      new BlockStatementBuilder(locVarDecl).generatedAST();
    preAssertion.addElement(blkStmt);
    return lhs;
  }


  /**
    * Shorthand for makePreAssertion(ASTType, Node, String), with just the
    * expression node. The type and string parameters are synthetized from the
    * input expression node properties.
    * @see #makePreAssertion(ORG.opengroup.parsers.adljava.ASTType,
    *                        ORG.opengroup.parsers.adljava.Node,
    *                        java.lang.String)
    */
  private String makePreAssertion(TypedNode expr) {

    String theString = expr.getExpressionImage();
    LType ltype = expr.getType();

    return makePreAssertion(TypeAux.LType2ASTType(ltype), expr, theString);                     
  }


  /**
    * Generates a try/catch Java block that corresponds to an ADL_tryStatement.
    * The try (resp. catch) block is composed of an assignment of a variable
    * passed as parameter, initialized by the expression found in the
    * ADL_tryStatement's try (resp. catch) block.
    * <BR>Creates either a preDeclaration (call-state case) or a preAssertion.
    * @param saveblk "is a call-state"
    * @param node ADL try/catch initialization expression
    * @param lhs the name of the variable to be assigned to
    * @return the string name of the generated tmp variable.
    * @see #saveBlock
    * @see #preAssertion
    */
  private String makeTryStatement(boolean saveblk, ASTADL_TryStatement node, String lhs) {

    if (debug)
      System.out.println("makeTryStatement : " + saveblk + " " + node + " " + lhs);

    // <node.getType> <tmpName>;

    String tmpName = makeSaveDecl(saveblk, ((TypedNode)node).getType(), lhs);

    // try { <tmpName> = <node.try_expr>; }
    // ( catch(<node.catch>) { <tmpName> = <node.catch_expr>; } )+

    ASTTryStatement tryStmt = new
      ASTTryStatement(ADLJavaParserTreeConstants.JJTTRYSTATEMENT);

    ASTAssignment assig = new AssignmentBuilder
      (tmpName, "=", (Node)node.jjtGetChild(0).jjtAccept(this, null))
      .generatedAST();

    ASTStatement stmt = new
      ASTStatement(ADLJavaParserTreeConstants.JJTSTATEMENT);
    parentChild(stmt, assig, 0);

    ASTBlock blk = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    parentChild(blk, stmt, 0);
    parentChild(tryStmt, blk, 0);

    ASTFormalParameter exc;
    for (int i = 1; i < node.jjtGetNumChildren()-1; i += 2) {
      // formal parameter, e.g. "catch(Exc e)"

      exc = (ASTFormalParameter)node.jjtGetChild(i);
      parentChild(tryStmt, exc, i);

      assig = new AssignmentBuilder
        (tmpName, "=", (Node)node.jjtGetChild(i+1).jjtAccept(this, null))
        .generatedAST();

      stmt = new
        ASTStatement(ADLJavaParserTreeConstants.JJTSTATEMENT);
      parentChild(stmt, assig, 0);

      blk = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
      parentChild(blk, stmt, 0);
      parentChild(tryStmt, blk, i+1);
    }

    if (saveblk) {
      saveBlock.addElement(tryStmt);
    }
    else {
      preAssertion.addElement(tryStmt);
    }

    return tmpName;
  }

  /**
   * Creates a try/catch block to trap thrown exceptions during the
   * evaluation of call-state expressions or assertions.
   *
   * The created block for a call-state block looks like:
<PRE>
|  try {
|    // call-state expressions block
|  } catch (Throwable t) {
|    testContext.infoline("Unexpected exception caught (" +
|                         t.getClass().getName() +
|                         ") in a call-state expression within assertion");
|    testContext.result(ADL_UNRESOLVED);
|    skipAssertions = true;
|    testContext.skipAssertions = true;
|  }
</PRE>
    *
    * @param triedBlock the block of statement to be included in the
    *                   try/catch block
    * @param inCallState whether this exception catch block is intended
    *                    for a call-state expressions block or not
    * @return the try/catch block
    **/
  public ASTTryStatement makeExceptionCatchBlock(ASTBlock triedBlock,
						 boolean inCallState) {
    String message, resultMethod;
    if (inCallState) {
      message = "in a call-state expression within assertion";
      resultMethod = "result";
    } else {
      message = "in assertion";
      resultMethod = "assertionResult";
    }

    // the try statement
    ASTTryStatement tryStmt = new
      ASTTryStatement(ADLJavaParserTreeConstants.JJTTRYSTATEMENT);

    parentChild(tryStmt, triedBlock, 0);

    parentChild(tryStmt,
		new FormalParameterBuilder("Throwable", "t")
		.generatedAST(),
		1);

    // catch block
    ASTBlock catchBlock =
      new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    parentChild(tryStmt, catchBlock, 2);

    // testContext.infoline("Unexpected exception caught (" +
    //                      t.getClass().getName() +
    //                      ") " + <message>);
    // testContext.<resultMethod>(ADL_UNRESOLVED);
    ASTAdditiveExpression addStr = new
      ASTAdditiveExpression(ADLJavaParserTreeConstants.JJTADDITIVEEXPRESSION);
    String[] ops = { "+", "+" };
    addStr.setOps(ops);
    ASTStringLiteral str1 = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    str1.setValue("\"Unexpected exception caught (\"");
    parentChild(addStr, str1, 0);

    ASTDotSuffix getN = new ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
    getN.setName("getName");

    ASTPrimaryExpression str2 = new PrimaryExpressionBuilder
      ( JGenericVisitor.makeName("t.getClass"),
        new ArgumentsBuilder().generatedAST())
      .addSuffix(getN)
      .addSuffix(new ArgumentsBuilder().generatedAST())
      .generatedAST();
    parentChild(addStr, str2, 1);

    ASTStringLiteral str3 = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    String str3Value = "\") " + message + "\"";
    str3.setValue(str3Value);
    parentChild(addStr, str3, 2);

    ASTStringLiteral str = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    str.setValue(message);
    ASTStatement infolineStmt =
      new StatementBuilder
      ( new PrimaryExpressionBuilder
	( JGenericVisitor.makeName("testContext.infoline"),
	  new ArgumentsBuilder(addStr).generatedAST())
	.generatedAST()
	).generatedAST();
    parentChild(catchBlock, infolineStmt, 0);

    ASTStatement stmt = new StatementBuilder
      ( new PrimaryExpressionBuilder
          ( JGenericVisitor.makeName("testContext." + resultMethod),
            new ArgumentsBuilder
            ( JGenericVisitor.makeName("ADL_UNRESOLVED")).generatedAST())
	.generatedAST())
      .generatedAST();
    parentChild(catchBlock, stmt, 1);

    if (inCallState) {
      // skipAssertions = true;
      // testContext.skipAssertions = true;
      stmt = new StatementBuilder
	( new AssignmentBuilder( "skipAssertions", "=", ACO.trueLit() ))
	.generatedAST();
      parentChild(catchBlock, stmt, 2);
      stmt = new StatementBuilder
	( new AssignmentBuilder
	  ( "testContext.skipAssertions", "=", ACO.trueLit() ))
	.generatedAST();
      parentChild(catchBlock, stmt, 3);
    }
    
    return tryStmt;
  }


  /**
    * Visit pattern for a typed node.
    * <P><B>Expression decomposition algorithm</B>
    * <BR><I>(This algorithm is used for all expression nodes)</I>
    * <P><UL><LI>Visit the children (sub-expressions).
    * <LI><B>IF</B> the expression is to be decomposed (i.e. it is contained in
    * an assertion expression and it is neither inside a call-state expression,
    * a try/catch or a quantified assertion), call makeSaveAssignment to
    * generate a local variable declaration, i.e. to generate the declaration
    * of a local temporary variable that will be defined by the type and the
    * value of the expression.
    * <LI><B>ELSE</B> nothing to do: just return the visited node. </UL>
    * <P><I>For a more detailed overview of the expression declaration algorithm,
    * see the ADL2.0 Translation System Design Specification, Chapter 5
    * (Code Generation), paragraph 5.2.1.3.</I>
    * @param node the ASTConditionalExpression to visit.
    * @param data unused.
    * @return the transformed node.
    */
  private Object decompose(TypedNode node, Object data) {

    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, null), i);
    }
    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * States if the passed name node corresponds to the name of an element
    * (field or method) of the real tested object. The return is true if
    * the name is not qualified and it is neither a binding, a method
    * parameter or an element inherited from ADLObjectACO.
    */
  private boolean isRealObjectComponent(ASTName node) {

    boolean ret = node.getName().indexOf('.') == -1;
    if (ret) {
      String astNameKind = null;
      try {
        astNameKind = (String)node.getProperty("astNameKind");
      }
      catch(PropertyAccessException e) { ret = false; }
      ret = ret &&
        (! astNameKind.equals("LParameter")) &&
        (! astNameKind.equals("ADLBinding")) &&
        (! astNameKind.equals("LVariable")) &&
        (! astNameKind.equals("objectACO"));
    }
    return ret;
  }


  /**
    * Single constructor.
    * <BR>Creates an ACOauxiliary that will be associated with the
    * currently transformed adl class.
    * @see ACOauxiliary
    */
  public JADLTransformVisitor() {

    ACO = new ACOauxiliary(this);
  }


  /**
    * Entry point. Initializes needed handles and launches the visit.
    * @param topNode the root node of the tree to visit
    * @return the root node of the transformed tree
    */
  public Node translate(Node topNode) {

    globalProlog = null;
    globalEpilog = null;
    localProlog = null;
    localEpilog = null;
    inlineDeclaration = new Vector();

    return (Node)topNode.jjtAccept(this, null);
  }


  /**
    * Visit pattern for a node of type ASTADL_CompilationUnit.
    * 
    * <P><B>Algorithm:</B>
    * creates a node ASTADL_CompilationUnit with children:<UL>
    * <LI> the ASTImportDeclaration children of the input node
    * <LI> an ASTImportDeclaration for "ORG.opengroup.adl.runtime.*"
    * <LI> the visit to the node's ASTADL_ClassDeclaration child
    * </UL>
    * @param node the ASTADL_CompilationUnit to visit.
    * @param data unused.
    * @return the transformed node (ASTCompilationUnit: the root of the
    * transformed AST).
    */
  public Object visit(ASTADL_CompilationUnit node, Object data) {

    if (debug) System.out.println("ASTADL_CompilationUnit");

    Node child;
    int nbChildren = node.jjtGetNumChildren();
    ASTCompilationUnit createdNode = new 
      ASTCompilationUnit(ADLJavaParserTreeConstants.JJTCOMPILATIONUNIT);

    ASTPackageDeclaration pack = new
      ASTPackageDeclaration(ADLJavaParserTreeConstants.JJTPACKAGEDECLARATION);
    parentChild(pack, makeName("adltest"), 0);
    parentChild(createdNode, pack, 0);

    ASTImportDeclaration adlruntime = new
      ASTImportDeclaration(ADLJavaParserTreeConstants.JJTIMPORTDECLARATION);
    parentChild(adlruntime, makeName("ORG.opengroup.adl.runtime"), 0);
    parentChild(createdNode, adlruntime, 1);
    try {
      adlruntime.setProperty("onDemand", Boolean.TRUE);
    }
    catch (PropertyAccessException e) {}

    for (int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_ClassDeclaration) {
        // case of unnamed package
        String[] imports = null;
        try {
          imports = (String[])node.getProperty("classesFromUnnamedPackage");
        } catch(PropertyAccessException e) {}

        if (imports != null) {
          // The tested class is in an unnamed package: in the generated file(s),
          // any class (of this unnamed package) used by the tested class must be
          // explicitely imported
//           for (int j = nbChildren-1; j >= 0; j--) {
//             parentChild(node, node.jjtGetChild(j), j + imports.length);
//           }
          ASTImportDeclaration impNode;
          for (int j = imports.length-1; j >= 0; j--) {
            impNode = new
              ASTImportDeclaration(ADLJavaParserTreeConstants.JJTIMPORTDECLARATION);
            parentChild(impNode, JGenericVisitor.makeName(imports[j]), 0);
            parentChild(createdNode, impNode, i+j+2);
          }
        }

        parentChild(createdNode,
                    (Node)child.jjtAccept(this, data),
                    /*i+2*/createdNode.jjtGetNumChildren());
      }
      else if (child instanceof ASTImportDeclaration) {
        parentChild(createdNode, child, i+2);
      }
    }
    if (syntaxError) {
      return null;
    }
    else {
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_ClassDeclaration.
    * <P><B>Algorithm:</B><UL>
    * <LI>Stores information locally (casual inline declaration, global
    * prolog/epilog) or into the associated ACOauxiliary (class name,
    * casual superclass).
    * <LI> Visits the ASTADL_BehaviorDeclaration child.
    * <LI> Creates and returns an ASTClassDeclaration with 3 children:<OL>
    * <LI> the extended class name
    * <LI> the class constructor
    * <LI> the visited ASTADL_BehaviorDeclaration child.
    * </OL></UL>
    * @param node the ASTADL_ClassDeclaration to visit.
    * @param data unused.
    * @return the created node (ASTClassDeclaration).
    */
  public Object visit(ASTADL_ClassDeclaration node, Object data) { 

    if (debug) System.out.println("ASTADL_ClassDeclaration");

    Node child;
    ASTClassDeclaration createdNode = new
      ASTClassDeclaration(ADLJavaParserTreeConstants.JJTCLASSDECLARATION);
    int nbChildren = node.jjtGetNumChildren();

    // the FQN of the tested class is that of the current node
    String fullyQualifiedName = node.getName();
    // the "short" name of the tested class is the last component in FQN
    String shortName = fullyQualifiedName
      .substring(fullyQualifiedName.lastIndexOf(".") + 1);

    ACO.setAdlClassName(shortName);
    ACO.setImplementationClassName(fullyQualifiedName);

    // the name of the ACO class created: "short" name + ACO
    createdNode.setName(shortName + "ACO");

    int modif = JModifier.PUBLIC;
    try {
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}

    int insertChild = 2; // for MethodDeclarations (after Name and constructor)
    // a first loop mainly to detect all the inlines (an inline may be used
    // before it is defined)
    for (int i = 1; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTName) {
        ACO.setExtendedClass((ASTName)child);
      }
      else if (child instanceof ASTADL_Prologue) {
        globalProlog = (ASTBlock)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADL_Epilogue) {
        globalEpilog = (ASTBlock)child.jjtGetChild(0);
      }
      else if (child instanceof ASTADL_InlineDeclaration) {
        inlineDeclaration.addElement(child);
      }
    }
    // now the loop for behavior declarations
    for (int i = 1; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_BehaviorDeclaration) {
        parentChild(createdNode, (Node)child.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    parentChild(createdNode, ACO.getExtendedClass(), 0);

    parentChild(createdNode, ACO.makeACO_Constructor(), 1);

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorDeclaration.
    * <P><B>Algorithm:</B><UL>
    * <LI>Creates a node of type ASTMethodDeclaration, root of the method that
    * corresponds to the current behavior description.
    * <LI>Stores local information into the associated ACOauxiliary (method
    * name, return type, formal parameters, thrown exceptions, isConstructor
    * flag) or locally (local prolog/epilog). Sets the flag isInConstructor.
    * <LI> Visits the ASTADL_BehaviorSpecification child.
    * </UL>
    * @param node the ASTADL_BehaviorDeclaration to visit.
    * @param data unused.
    * @return the created node (ASTMethodDeclaration).
    */
  public Object visit(ASTADL_BehaviorDeclaration node, Object data) {

    if (debug) System.out.println("ASTADL_BehaviorDeclaration");

    Node child;
    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    int nbChildren = node.jjtGetNumChildren();

    preDeclaration = new Vector();
    saveBlock = new Vector();
    try {
      JModifier modif = new JModifier(((Integer)node.getProperty("modifier")).intValue());
      ACO.initializeMethod(modif.isStatic());
    }
    catch(PropertyAccessException e) {
      ACO.initializeMethod(false);
    }

    int j = 0, insertChild = 0;
    child = node.jjtGetChild(0);
    if (!(child instanceof ASTResultType)) { // constructor
      isInConstructor = true;
      parentChild(createdNode,  // method result type == adl class type
                  makeType(ACO.getImplementationClassName()),
                  insertChild++);
    }
    else {
      isInConstructor = false;
      j = 1;
      ACO.setMethodReturnType((ASTResultType)child);
      parentChild(createdNode, child, insertChild++);
    }
    ACO.setIsConstructor(isInConstructor);

    child = node.jjtGetChild(j++); // Name
    ACO.setMethodName(((ASTName)child).getName());

    child = node.jjtGetChild(j++); // Formal Parameters
    ACO.setMethodFormalParam((ASTFormalParameters)child);

    // create and insert MethodDeclarator node, using name and formalParam
    ASTMethodDeclarator subnode = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    subnode.setName(ACO.getMethodName());
    parentChild(subnode, child, 0);
    parentChild(createdNode, subnode, insertChild++);

    child = node.jjtGetChild(j++);
    if (child instanceof ASTNameList) {
      ACO.setMethodThrowsList((ASTNameList)child);
      parentChild(createdNode, ((ASTNameList)child), insertChild++);
      child = node.jjtGetChild(j++);
    }
    if (child instanceof ASTADL_Prologue) {
      localProlog = (ASTBlock)child.jjtGetChild(0);
      child = node.jjtGetChild(j++);
    }
    else {
      localProlog = null;
    }

    // child instanceof ASTADL_BehaviorSpecification
    // store first the local epilog
    if ((j < nbChildren) && (node.jjtGetChild(j) instanceof ASTADL_Epilogue)) {
      localEpilog = (ASTBlock)node.jjtGetChild(j).jjtGetChild(0);
    }
    else {
      localEpilog = null;
    }

    // visit the behavior spec
    parentChild(createdNode, (Node)child.jjtAccept(this, data), insertChild);

    savIndex = 0;
    isInConstructor = false;
    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_Prologue.
    * <BR>Should not be visited???
    * @param node the ASTADL_Prologue to visit.
    * @param data unused.
    * @return an error
    */
  public Object visit(ASTADL_Prologue node, Object data) { 

    if (debug) System.out.println("ASTADL_Prologue");

    // not visited
    internalError();
    return null;
  }


  /**
    * Visit pattern for a node of type ASTADL_Epilogue.
    * <BR>Should not be visited???
    * @param node the ASTADL_Epilogue to visit.
    * @param data unused.
    * @return an error
    */
  public Object visit(ASTADL_Epilogue node, Object data) { 

    if (debug) System.out.println("ASTADL_Epilogue");

    // not visited
    internalError();
    return null;
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorSpecification.
    * <P><B>Algorithm: </B>generates all statements of the body of the
    * currently visited behavior description:<OL>
    * <LI>initialization statements: ACO.makeACO_Initialization(hasSuperSemantics)
    * <LI>global prolog statements
    * <LI>local prolog statements
    * <LI>callstate temporary variables declarations and initializations
    * (statements of the handle "saveBlock", created by the visit to the
    * ASTADL_AssertionGroup child)
    * <LI>call to the tested method: ACO.makeACO_CallBlock()
    * <LI>statements that correspond to the evaluation of the assertions
    * (returned node from the visit to the ASTADL_AssertionGroup child)
    * <LI>local epilog statements
    * <LI>global epilog statements
    * <LI>finalization statements: ACO.makeACO_Final()
    * </OL>
    * @param node the ASTADL_BehaviorSpecification to visit.
    * @param data unused.
    * @return the created node (ASTBlock).
    * @see ACOauxiliary#makeACO_Initialization
    * @see ACOauxiliary#makeACO_CallBlock
    * @see ACOauxiliary#makeACO_Final
    */
  public Object visit(ASTADL_BehaviorSpecification node, Object data) {

    if (debug) System.out.println("ASTADL_BehaviorSpecification");

    Node child, newChild = null;
    ASTBlock createdNode = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    int nbChildren = node.jjtGetNumChildren();

    int insertChild = 0;
    child = node.jjtGetChild(0);
    if (child instanceof ASTADL_BehaviorClassification) {
      child.jjtAccept(this, data); // only side-effect: saving (ab)normalSpec

      child = node.jjtGetChild(1);
    }

    // child : ADL_AssertionGroup (last child)

//     boolean hasSuperSemantics = false;
//     for (int i = 0; i < child.jjtGetNumChildren() && !hasSuperSemantics; i++) {
//       hasSuperSemantics = child.jjtGetChild(i) instanceof ASTADL_SuperSemantics;
//     }

    boolean hasSuperSemantics = child.jjtGetNumChildren() > 0 &&
      child.jjtGetChild(0) instanceof ASTADL_SuperSemantics;

    Vector ACO_Initialization = ACO.makeACO_Initialization(hasSuperSemantics);
    int nbc = ACO_Initialization.size();
    for (int i = 0; i < nbc; i++) {
      newChild = (Node)ACO_Initialization.elementAt(i);
      parentChild(createdNode, newChild, insertChild);
      insertChild += 1;
    }
    

    if ((globalProlog != null) && !isInConstructor) {
      nbc = globalProlog.jjtGetNumChildren();
      ASTBlock GPclone = null;
      try { GPclone = (ASTBlock)(globalProlog.clone()); }
      catch(CloneNotSupportedException e) {}
      for (int i = 0; i < nbc; i++) {
        /*
        try {
          newChild = (Node)((SimpleNode)globalProlog.jjtGetChild(i)).clone();
        }
        catch(CloneNotSupportedException e) {}
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        */
        parentChild(createdNode,
                    (Node)GPclone.jjtGetChild(i).jjtAccept(this, data),
                    insertChild);
        insertChild += 1;
      }
    }

    if (localProlog != null) {
      nbc = localProlog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localProlog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    // need to visit the child assertion group to know how many
    // preDeclaration are to be inserted

    ASTBlock visitedAssGrp = (ASTBlock)child.jjtAccept(this, data);


    Vector ACO_CallBlock = ACO.makeACO_CallBlock();

    nbc = preDeclaration.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)preDeclaration.elementAt(i), insertChild+i);
    }
    insertChild += nbc;


    if (!isInConstructor) {
      // no @/unchanged in constructors, hence no saveBlock

      nbc = saveBlock.size();
      if (nbc > 0) {
	ASTIfStatement ifStSkipAss = new
	  ASTIfStatement(ADLJavaParserTreeConstants.JJTIFSTATEMENT);
	parentChild(createdNode, ifStSkipAss, insertChild++);
	
	parentChild(ifStSkipAss,
		    new NotBuilder(makeName("skipAssertions"), false)
		    .generatedAST(),
		    0);
	
	ASTBlock thenBlk =
	  new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);

	ASTTryStatement triedBlk = makeExceptionCatchBlock(thenBlk, true);
	parentChild(ifStSkipAss, triedBlk, 1);
	
	for (int i = 0; i < nbc; i++) {
	  parentChild(thenBlk, (Node)saveBlock.elementAt(i), i);
	}
      }
    }

    nbc = ACO_CallBlock.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_CallBlock.elementAt(i), insertChild+i);
    }
    insertChild += nbc;

    nbc = visitedAssGrp.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, visitedAssGrp.jjtGetChild(i), insertChild+i);
    }
    insertChild += nbc;


    if (localEpilog != null) {
      nbc = localEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        newChild = localEpilog.jjtGetChild(i);
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    if ((globalEpilog != null) && !isInConstructor) {
      nbc = globalEpilog.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        try {
          newChild = (Node)((SimpleNode)globalEpilog.jjtGetChild(i)).clone();
        }
        catch(CloneNotSupportedException e) {}
        parentChild(createdNode, (Node)newChild.jjtAccept(this, data), insertChild);
        insertChild += 1;
      }
    }

    Vector ACO_Final = ACO.makeACO_Final();
    nbc = ACO_Final.size();
    for (int i = 0; i < nbc; i++) {
      parentChild(createdNode, (Node)ACO_Final.elementAt(i), insertChild);
      insertChild += 1;
    }

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_BehaviorClassification.
    * <BR>stores normal/abnormal specifications in the associated ACOauxiliary.
    * @param node the ASTADL_BehaviorClassification to visit.
    * @param data unused.
    * @return nothing (in fact returns null, that will not be used by the caller)
    */
  public Object visit(ASTADL_BehaviorClassification node, Object data) { 

    if (debug) System.out.println("ASTADL_BehaviorClassification");

    Node child;
    int nbc = node.jjtGetNumChildren();
    for (int i = 0; i < nbc; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_NormalBehavior) {
        ACO.setNormalSpec((SimpleNode)child.jjtGetChild(0));
      }
      else {
        ACO.setAbnormalSpec((SimpleNode)child.jjtGetChild(0));
      }
    }
    return null;
  }


  /**
    * Visit pattern for a node of type ASTADL_AssertionGroup.
    * <P><B>Algorithm: </B>
    * <UL><LI><B>IF</B> this is a nested assertion group: the assertion group
    * is considered as an expression.
    * <UL><LI><B>IF</B> it contains more than one assertions
    * <BR>Returns an expression: the result of the visit to this child assertion.
    * <LI><B>ELSE</B> Returns an expression: the "AND" of the results of the
    * visits to these child assertions.
    * <LI>If the input node is considered as a decomposable expression,
    * makePreAssertion is called and the visit returns the name node of the
    * variable that was initialized by the input node value.</UL>
    * <LI><B>ELSE</B> the created node is an ASTBlock that contains
    * <I>either</I> the return node (an ASTBlockStatement) of the visit to the
    * grand-child when the latter is an ASTADL_TryStatement /
    * ASTADL_IfStatement, <I>or</I> the statements of the ASTBlock returned by
    * the visit to an ASTADL_Assertion grand-child.
    * @param node the ASTADL_AssertionGroup to visit.
    * @param data unused.
    * @return the created node (Node / ASTName / ASTConditionalAndExpression / ASTBlock)
    * @see #makePreAssertion(ORG.opengroup.adl.parsers.adljava.ASTType,
    *                        ORG.opengroup.adl.parsers.adljava.Node,
    *                        java.lang.String)
    */
  public Object visit(ASTADL_AssertionGroup node, Object data) {

    if (debug) System.out.println("ASTADL_AssertionGroup");

    int nbChildren;
    String tmpStr = "", tmpName;
    if (isDecomposable) {
        tmpStr = node.getExpressionImage();
    }      

    if (isInAssertion || isInBinding) { // nested
      nbChildren = node.jjtGetNumChildren();
      if ((nbChildren > 1) && (node.jjtGetChild(1) instanceof ASTADL_Statement)) {
        // group of boolean expressions => conjunction
        ASTConditionalAndExpression createdNode = new
          ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
        
        for (int i = 0; i < nbChildren; i++) {
          if (node.jjtGetChild(i) instanceof ASTADL_Statement) {
            parentChild(createdNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
          }
        }
        if (isDecomposable) {
          tmpName = makePreAssertion(makeType("boolean"), createdNode, tmpStr);
          return makeName(tmpName);
        }
        else {
          return createdNode;
        }
      }
      else { // nested assgroup of 1 assertion
        return (Node)node.jjtGetChild(0).jjtAccept(this, data);
      }
    }
    else { // not nested : "real" assertion group
      ASTBlock createdNode = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
      Node visitedAss, child1, child2;
      int insertChild = 0;
      for (int i = 0; i < node.jjtGetNumChildren(); i++) {
        child1 = node.jjtGetChild(i);
        if (child1 instanceof ASTADL_Binding) {
          Node bind = (Node)child1.jjtAccept(this, data);
          if (bind instanceof ASTBlock) {
            for (int j = 0; j < bind.jjtGetNumChildren(); j++) {
              parentChild(createdNode, bind.jjtGetChild(j), insertChild++);
            }
          }
          else {
            parentChild(createdNode, bind, insertChild++);
          }
        }
        else if (child1 instanceof ASTADL_SuperSemantics) {
          // already tested: see ASTADL_BehaviorSpecification
          if (isInConstructor) {
            try {
              ACO.setInheritedConstructor((String)((SimpleNode)child1).getProperty("definingADLClass"));
            } catch(PropertyAccessException e) {}
          }            
        }
        else { // child1 instanceof ASTADL_Statement
          child2 = child1.jjtGetChild(0);
          if ((child2 instanceof ASTADL_IfStatement) ||
              (child2 instanceof ASTADL_TryStatement)) {
            parentChild(createdNode, (Node)child2.jjtAccept(this, data), insertChild++);
          }
          else { // child2 : ADL_Assertion  -->  child2.jjtAccept : Block
            visitedAss = (Node)child2.jjtAccept(this, data);
            for (int j = 0; j < visitedAss.jjtGetNumChildren(); j++) {
              parentChild(createdNode, visitedAss.jjtGetChild(j), insertChild++);
            }
          }
        }
      }
      return createdNode;
    }
  }
    

  /**
    * Visit pattern for a node of type ASTADL_Assertion.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> nested: the assertion node is in fact considered as a
    * boolean expression: just visit the child.
    * <LI>Generate the AST for the code that corresponds to an assertion, i.e.
    * for instance for assertion
    * "get_active_accounts()==@get_active_accounts()+1" of adl class bank:
<PRE>
|  testContext.assertionStart("get_active_accounts()==@get_active_accounts()+1");
|  if (!skipAssertions) {
|     int __ADL_tmp_0 = bankObject.get_active_accounts();
|     int __ADL_tmp_1 = __ADL_savTmp_1;
|     int __ADL_tmp_2 = __ADL_tmp_1 + 1;
|     boolean __ADL_tmp_3 = __ADL_tmp_0 == __ADL_tmp_2;
|     tddRes = __ADL_tmp_3;
|     if (debugMode) {
|        testContext.infoline("get_active_accounts()\t-> " + __ADL_tmp_0);
|        testContext.infoline("@get_active_accounts()\t-> " + __ADL_tmp_1);
|        testContext.infoline("@get_active_accounts()+1\t-> " + __ADL_tmp_2);
|        testContext.infoline("get_active_accounts()==@get_active_accounts()+1\t-> "
|                               + __ADL_tmp_3);
|     }
|     testContext.assertionResult(( tddRes ? ADL_PASS : ADL_FAIL ));
|  }
|  else {
|     testContext.assertionResult(ADL_UNEVALUATED);
|  }
</PRE>
    * The local variable declarations __ADL_tmp_i are "preAssertion"
    * obtained by the expression decomposition algorithm.
    * </UL>
    * @see #visit(ORG.opengroup.parsers.adljava.ASTConditionalExpression,
    *             java.lang. Object)
    * @param node the ASTADL_Assertion to visit.
    * @param data unused.
    * @return the created node (ASTBlock).
    */
  public Object visit(ASTADL_Assertion node, Object data) { 

    if (debug) System.out.println("ASTADL_Assertion");

    Node visitedAss = null, child = node.jjtGetChild(0);
    int nbChildren = node.jjtGetNumChildren(), insertChild = 0;
    String tmpStr = "", assLabel = null;
    boolean labelledAss = false, taggedAss = false, quantif = false;
    
    if (isInAssertion || isInBinding) {
      // nested assertion: must be considered as an expression
      for(int i = 0; i < nbChildren; i++) {
        child = node.jjtGetChild(i);
        if (child instanceof ASTADL_Label) {
          // $$$$ WARNING labels not used in nested assertions
          return null;
        }
        else if (child instanceof ASTADL_Tags) {
          // $$$$ WARNING tags not used in nested assertions
          return null;
        }
        else if (child instanceof ASTADL_QuantifiedAssertion) {
          errorMessage("No quantified assertion in nested assertion group",
                       ((SimpleNode)child).getLineNumber());
        }
        else { // expression of the assertion : to visit
          return (Node)child.jjtAccept(this, data);
        }
      }
    }

    preAssertion = new Vector();
    boolContextStack = new Stack();

    ASTBlock createdNode = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    for(int i = 0; i < nbChildren; i++) {
      child = node.jjtGetChild(i);
      if (child instanceof ASTADL_Label) {
        labelledAss = true;
        assLabel = (String)child.jjtAccept(this, data);
      }
      else if (child instanceof ASTADL_Tags) {

        taggedAss = true;
        assTags = (String[])child.jjtAccept(this, data);
      }
      else if (child instanceof ASTADL_QuantifiedAssertion) {
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInTry || isInQuantified || isInBinding);

        visitedAss = (Node)child.jjtAccept(this, data);

        isInAssertion = false; isDecomposable = false;
        quantif = true;
	tmpStr = ((SimpleNode)child).getExpressionImage();
      }
      else { // expression of the assertion
        // visit the expression: decomposition algorithm
        isInAssertion = true;
        isDecomposable = ! (isInCallState || isInTry || isInQuantified || isInBinding);
        visitedAss = (Node)child.jjtAccept(this, data);
        isInAssertion = false; isDecomposable = false;
	tmpStr = ((SimpleNode)child).getExpressionImage();
      }
    }

    int nbp = preAssertion.size();

    if (labelledAss) {
      tmpStr = "\"[ " + assLabel + " ] " + tmpStr + "\"";
    }
    else {
      tmpStr = "\"" + tmpStr + "\"";
    }

    // testContext.assertionStart(<tmpStr>)

    ASTStringLiteral str = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    str.setValue(tmpStr);
    ASTPrimaryExpression assertionStart = new
      PrimaryExpressionBuilder(makeName("testContext.assertionStart"),
                               new ArgumentsBuilder(str).generatedAST())
      .generatedAST();
    parentChild(createdNode, 
                new StatementBuilder(assertionStart).generatedAST(),
                0);

    // if tagged assertion:
    // testContext.infoline("Tagged assertion (Untestable)";)

    if (taggedAss) {
      ASTStringLiteral tag = new
        ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
      tag.setValue("\"Tagged assertion (Untestable)\"");

      ASTPrimaryExpression untest = new
        PrimaryExpressionBuilder(makeName("testContext.infoline"),
                                 new ArgumentsBuilder(tag).generatedAST())
        .generatedAST();
      parentChild(createdNode, 
                  new StatementBuilder(untest).generatedAST(),
                  1);
    }
    else {
      // if (!skipAssertions)

      ASTIfStatement ifStSkipAss = new
        ASTIfStatement(ADLJavaParserTreeConstants.JJTIFSTATEMENT);

      parentChild(ifStSkipAss,
                  new NotBuilder(makeName("skipAssertions"), false)
                  .generatedAST(),
                  0);

      ASTBlock triedBlock =
	new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
      ASTTryStatement tryStmt = makeExceptionCatchBlock(triedBlock,
							false);
      parentChild(ifStSkipAss, tryStmt, 1);


      if (quantif) { // quantified assertion:
        // store the statements returned by the visit of the quantified assertion
        // (this visit returns an ASTBlock)
        for (int i = 0; i < visitedAss.jjtGetNumChildren(); i++) {
          parentChild(triedBlock, (Node)visitedAss.jjtGetChild(i), i);
        }
        insertChild = visitedAss.jjtGetNumChildren();
      }
      else {
        // pre-assertions code

        for (int i = 0; i < nbp; i++) {
          parentChild(triedBlock, (Node)preAssertion.elementAt(i), i);
        }
        insertChild = nbp;
    
        // tddRes = last tmp
        parentChild(triedBlock,
                    new StatementBuilder
                    ( new AssignmentBuilder("tddRes", "=", visitedAss))
                    .generatedAST(),
                    insertChild);
        insertChild += 1;
      }


      // if (debugMode) {
      if (!quantif) {

        // No debug mode for quantified assertions

        ASTIfStatement ifDebug = new
          ASTIfStatement(ADLJavaParserTreeConstants.JJTIFSTATEMENT);
        parentChild(ifDebug, makeName("debugMode"), 0);
        parentChild(triedBlock, ifDebug, insertChild);
        insertChild += 1;

        ASTBlock thenBlk2 = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
        parentChild(ifDebug, thenBlk2, 1);

        // infoline of sub-expressions

        Node preass; ASTLocalVariableDeclaration vardecl;
        ASTAdditiveExpression arg;
        int insertChild2 = 0;
        String[] ops = { "+" };

        for (int i = 0; i < nbp; i++) {
          preass = (Node)preAssertion.elementAt(i);
          if (preass instanceof ASTBlockStatement) {
            vardecl = (ASTLocalVariableDeclaration)preass.jjtGetChild(0);
            arg = new
              ASTAdditiveExpression(ADLJavaParserTreeConstants.JJTADDITIVEEXPRESSION);
            arg.setOps(ops);

            ASTStringLiteral left = new
              ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
            try {
              left.setValue("\"" + (vardecl.getProperty("tokenString")) + "\\t-> \"");
            }
            catch(PropertyAccessException e) {
              left.setValue("\"unknown string\"");
            }
            parentChild(arg, left, 0);
            parentChild(arg,                        
                        makeName(((SimpleNode)vardecl.jjtGetChild(1).jjtGetChild(0)).getName()),
                        1);
            parentChild(thenBlk2,
                        new StatementBuilder
                        ( new PrimaryExpressionBuilder
                          ( makeName("testContext.infoline"),
                            new ArgumentsBuilder(arg).generatedAST()))
                        .generatedAST(),
                        insertChild2);

            insertChild2 += 1;
          }
        }
      }

      // testContext.assertionResult((tddRes ? ADL_PASS : ADL_FAIL));

      ASTConditionalExpression passFail = new
        ASTConditionalExpression(ADLJavaParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(passFail, makeName("tddRes"), 0);
      parentChild(passFail, makeName("ADL_PASS"), 1);
      parentChild(passFail, makeName("ADL_FAIL"), 2);
      parentChild(triedBlock,
                  new StatementBuilder
                  ( new PrimaryExpressionBuilder
                    ( makeName("testContext.assertionResult"),
                      new ArgumentsBuilder(passFail).generatedAST())
                    .generatedAST())
                  .generatedAST(),
                  insertChild++);


      // else { testContext.assertionResult(ADL_UNEVALUATED); }
      ASTBlock elseBlk = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
      parentChild(ifStSkipAss, elseBlk, 2);
      parentChild(elseBlk,
                  new StatementBuilder
                  ( new PrimaryExpressionBuilder
                    ( makeName("testContext.assertionResult"),
                      new ArgumentsBuilder(makeName("ADL_UNEVALUATED"))
                      .generatedAST())
                    .generatedAST())
                  .generatedAST(),
                  0);

      parentChild(createdNode, ifStSkipAss, 1);
    }

    preAssertion = null;
    preAssIndex = 0;

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_Binding.
    * <BR>Generates a local variable declaration that corresponds to the
    * binding.
    * @param node the ASTADL_Binding to visit.
    * @param data unused.
    * @return the created node (ASTBlockStatement).
    */
  public Object visit(ASTADL_Binding node, Object data) { 

    if (debug) System.out.println("ASTADL_Binding");

    ASTFormalParameter params = 
      (ASTFormalParameter)node.jjtGetChild(0); // formal parameter
    Node expr = node.jjtGetChild(1); // expression

    isInBinding = true;
    isDecomposable = false;
    if ((expr instanceof ASTADL_AssertionGroup) &&
        (expr.jjtGetChild(0).jjtGetChild(0) instanceof ASTADL_TryStatement)) {
      ASTADL_TryStatement tryexpr =
        (ASTADL_TryStatement)expr.jjtGetChild(0).jjtGetChild(0);
      
      String bindingName = ((SimpleNode)params.jjtGetChild(1)).getName();

      ASTBlock createdNode = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);

      // <bindingType> <bindingName> = <bindingType defaultValue>;
      parentChild(createdNode,
                  new BlockStatementBuilder
                  ( new LocalVariableDeclarationBuilder
                    ( (ASTType)params.jjtGetChild(0),
                      bindingName,
                      TypeAux.getDefaultValue((ASTType)params.jjtGetChild(0)))
                    .generatedAST())
                  .generatedAST(),
                  0);

      // try { <bindingName> = <tryexpr.tryblk>; }
      // ( catch(<tryexpr.catch>) { <bindingName> = <tryexpr.catchblk>; } )+

      ASTTryStatement tryStmt = new
        ASTTryStatement(ADLJavaParserTreeConstants.JJTTRYSTATEMENT);
      parentChild(createdNode, tryStmt, 1);

      ASTBlock blk = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
      parentChild(blk,
                  new StatementBuilder
                  ( new AssignmentBuilder
                    ( bindingName, "=",
                      (Node)tryexpr.jjtGetChild(0).jjtAccept(this, null))
                    .generatedAST())
                  .generatedAST(),
                  0);
      parentChild(tryStmt, blk, 0);

      ASTFormalParameter exc;
      for (int i = 1; i < tryexpr.jjtGetNumChildren()-1; i += 2) {
        // formal parameter, e.g. "catch(Exc e)"

        exc = (ASTFormalParameter)tryexpr.jjtGetChild(i);
        parentChild(tryStmt, exc, i);

        blk = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
        parentChild(blk,
                    new StatementBuilder
                    ( new AssignmentBuilder
                      ( bindingName, "=",
                        (Node)tryexpr.jjtGetChild(i+1).jjtAccept(this, null))
                      .generatedAST())
                    .generatedAST(),
                    0);
        parentChild(tryStmt, blk, i+1);
      }
      isInBinding = false;
      return createdNode;
    }
    else {
      ASTBlockStatement createdNode = new BlockStatementBuilder
        ( new LocalVariableDeclarationBuilder
          ( makeType((ASTType)params.jjtGetChild(0)), // binding type
            ( (SimpleNode)params.jjtGetChild(1)).getName(),              // binding name
            (Node)expr.jjtAccept(this, data)))                           // binding expr
        .generatedAST();
      isInBinding = false;
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_Statement.
    * <BR>No special processing: just retransmits to the child node.
    * @param node the ASTADL_Statement to visit.
    * @param data unused.
    * @return the transformed node.
    */
  public Object visit(ASTADL_Statement node, Object data) { 

    if (debug) System.out.println("ASTADL_Statement");

    return (Node)node.jjtGetChild(0).jjtAccept(this, data);
  }


  /**
    * Visit pattern for a node of type ASTADL_TryStatement.
    * <P><B>Algorithm: </B><UL><LI><B>IF</B> inside an assertion, it is
    * considered as an expression (but not a decomposable one):
    * <BR>call makeTryStatement and return the Name node of the variable that
    * has been initialized with the value of the input node.
    * <LI><B>ELSE</B> generates a Java try/catch statement by visiting the
    * child nodes (try block and catch block(s)).  </UL>
    * @param node the ASTADL_TryStatement to visit.
    * @param data unused.
    * @return the created node (ASTName / ASTTryStatement).
    * @see #makeTryStatement
    */
  public Object visit(ASTADL_TryStatement node, Object data) { 

    if (debug) System.out.println("ASTADL_TryStatement");

    if (isInAssertion) { // nested
      String tmpStr = "", tmpName;
      if (! (isInCallState || isInQuantified)) {
	tmpStr = node.getExpressionImage();
      }

      isInTry = true; isDecomposable = false;
      tmpName = makeTryStatement(isInCallState, node, tmpStr);
      isInTry = false; isDecomposable = ! (isInCallState || isInQuantified);

      return makeName(tmpName);
    }
    else { // not nested
      ASTTryStatement createdNode = new
        ASTTryStatement(ADLJavaParserTreeConstants.JJTTRYSTATEMENT);

      // try assertion group
      parentChild(createdNode, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);

      int nbChildren = node.jjtGetNumChildren();
      for (int i = 1; i < nbChildren-1; i += 2) {
        // catch formal parameters
        parentChild(createdNode, node.jjtGetChild(i), i);
        // catch assertion group
        parentChild(createdNode, (Node)node.jjtGetChild(i+1).jjtAccept(this, data), i+1);
      }
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_IfStatement.
    * <P><B>Algorithm: </B><UL><LI><B>IF</B> inside an assertion, it is
    * considered as an expression and will generate an " ? : " conditional
    * expression by visiting the child nodes. If there is no else child node,
    * the else branch of the generated conditional expression will be the
    * boolean literal true.
    * <LI><B>ELSE</B> generates a Java if statement by visiting the child nodes
    * (condition, then node, else node if present).  </UL>
    * @param node the ASTADL_IfStatement to visit.
    * @param data unused.
    * @return the created node (ASTConditionalExpression / ASTIfStatement).
    */
  public Object visit(ASTADL_IfStatement node, Object data) { 

    if (debug) System.out.println("ASTADL_IfStatement");

    Node child; 
    if (isInAssertion) { // nested
      ASTParenthExpression parenth = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
      ASTConditionalExpression condNode = new
        ASTConditionalExpression(ADLJavaParserTreeConstants.JJTCONDITIONALEXPRESSION);
      parentChild(parenth, condNode, 0);

      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        parentChild(condNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
      }
      if (nbc == 2) {
        child = new
          ASTBooleanLiteral(ADLJavaParserTreeConstants.JJTBOOLEANLITERAL);
        ((SimpleNode)child).setValue(true);
        parentChild(condNode, child, 2);
      }
      parenth.setType(node.getType());
      if (isDecomposable) {
        return makeName(makePreAssertion(parenth));
      }
      else {
        return parenth;
      }
    }
    else { // not nested
      ASTIfStatement createdNode = new
        ASTIfStatement(ADLJavaParserTreeConstants.JJTIFSTATEMENT);
      
      int nbc = node.jjtGetNumChildren();
      for (int i = 0; i < nbc; i++) {
        parentChild(createdNode, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
      }
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_Label.
    * @param node the ASTADL_Label to visit.
    * @param data unused.
    * @return the name of the label (String)
    */
  public Object visit(ASTADL_Label node, Object data) { 

    if (debug) System.out.println("ASTADL_Label");

    return node.getName();
  }


  /**
    * Visit pattern for a node of type ASTADL_Tags.
    * @param node the ASTADL_Tags to visit.
    * @param data unused.
    * @return the assertion tags (String array).
    */
  public Object visit(ASTADL_Tags node, Object data) { 

    if (debug) System.out.println("ASTADL_Tags");

    // hypothesis: the only possible tag is "U"
    String[] assTags;
    try {
      assTags = (String[])node.getProperty("taglist");
    }
    catch(PropertyAccessException e) {
      assTags = new String[1];
      assTags[0] = "<no tag>";
    }
    return assTags;
  }


  /**
    * Visit pattern for a node of type ASTADL_QuantifiedAssertion.
    * <P>Generates the code that corresponds to a quantified assertion, i.e.
    * for instance for assertion
    * <BR>"forall (String s : ProductsBank.keys()) { Stock_balance(s) >= 0; }" :
<PRE>
|  testContext.assertionStart("forall(Strings:ProductsBank.keys()){Stock_balance(s)>=0;}");
|  if (!skipAssertions) {
|     tddRes = true;
|     java.util.Enumeration enum0 = ProductsBank.keys();
|     String s = (String)enum0.nextElement();
|     for (; tddRes && enum0.hasMoreElements(); s = (String)enum0.nextElement())
|       tddRes = StockBrokerObject.Stock_balance(s) >= 0;
|     testContext.assertionResult(( tddRes ? ADL_PASS : ADL_FAIL ));
|  }
|  else {
|     testContext.assertionResult(ADL_UNEVALUATED);
|  }
</PRE>
    * <P><B>Notes</B><UL>
    * <LI>a quantified assertion is considered as non-decomposable
    * <LI>in the example above, if "forall" is replaced by "exists", the
    * assignment "tddRes = true;" must be replaced by "tddRes = false;" and the
    * expression "tddRes && enum0.hasMoreElements()" by "!tddRes &&
    * enum0.hasMoreElements()".
    * <LI>if there are several enumeration domains, the generated code is an
    * imbrication of for loops on enum0, enum1, etc.  </UL>
    * @param node the ASTADL_QuantifiedAssertion to visit.
    * @param data unused.
    * @return the created node (ASTBlock).
    */
  public Object visit(ASTADL_QuantifiedAssertion node, Object data) { 

    if (debug) System.out.println("ASTADL_QuantifiedAssertion");

    ASTBlock createdNode = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);

    boolean isForall = ((ASTADL_Quantifier)node.jjtGetChild(0)).getName().equals("forall");
    int insertChild = 0;

    // tddRes = <isForall>;
    ASTBooleanLiteral tddResInit = new
      ASTBooleanLiteral(ADLJavaParserTreeConstants.JJTBOOLEANLITERAL);
    tddResInit.setValue(isForall);
    parentChild(createdNode, new StatementBuilder
                ( new AssignmentBuilder("tddRes", "=", tddResInit).generatedAST())
                .generatedAST(),
                insertChild++);

    int nbLoops = node.jjtGetChild(1).jjtGetNumChildren();
    ASTFormalParameter[] domFP = new ASTFormalParameter[nbLoops];
    Node[] domExpr = new Node[nbLoops];

    for (int i = 0; i < nbLoops; i++) {
      domFP[i] = (ASTFormalParameter)node.jjtGetChild(1).jjtGetChild(i).jjtGetChild(0);
      domExpr[i] = node.jjtGetChild(1).jjtGetChild(i).jjtGetChild(1);
    }

    ASTCastExpression castexpr; ASTType FPtyp1, FPtyp2;
    for (int i = 0; i < nbLoops; i++) {
      // java.util.Enumeration enum<i> = <domExpr[i]>;
      parentChild(createdNode,
                  new BlockStatementBuilder
                  ( new LocalVariableDeclarationBuilder
                    ( "java.util.Enumeration",
                      "enum" + i,
                      domExpr[i]))
                  .generatedAST(),
                  insertChild++);

      // <domFP[i].type> <domFP[i].name> = (<domFP[i].type>) enum<i>.nextElement();
      FPtyp1 = (ASTType)domFP[i].jjtGetChild(0);
      castexpr = new 
        ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
      parentChild(castexpr, FPtyp1, 0);
      parentChild(castexpr,
                  new PrimaryExpressionBuilder
                  ( makeName("enum" + i + ".nextElement"),
                    new ArgumentsBuilder().generatedAST())
                  .generatedAST(),
                  1);
      FPtyp2 = makeType(FPtyp1);
      parentChild(createdNode, 
                  new BlockStatementBuilder
                  ( new LocalVariableDeclarationBuilder
                    ( FPtyp2,
                      ((ASTVariableDeclaratorId)domFP[i].jjtGetChild(1)).getName(),
                      castexpr))
                  .generatedAST(),
                  insertChild++);
    }

    ASTForStatement forEnum = new
      ASTForStatement(ADLJavaParserTreeConstants.JJTFORSTATEMENT);
    parentChild(createdNode, forEnum, insertChild++);

    Node last;

    ASTConditionalAndExpression andexp;
    ASTExpression forExpr;
    ASTAssignment upd;
    ASTForUpdate forUpdate;

    for (int i = 0; i < nbLoops; i++) {
      // for (; [!]tddRes && enum<i>.hasMoreElements();
      //        <domFP[i].name> = (<domFP[i].name>) enum<i>.nextElement()) {
      forExpr = new
        ASTExpression(ADLJavaParserTreeConstants.JJTEXPRESSION);
      andexp = new
        ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
      parentChild(forExpr, andexp, 0);
      if (isForall) {
        parentChild(andexp, makeName("tddRes"), 0);
      }
      else { // exists
        parentChild(andexp,
                    new NotBuilder(makeName("tddRes"), false)
                    .generatedAST(),
                    0);
      }
      parentChild(andexp,
                  new PrimaryExpressionBuilder
                  ( makeName("enum" + i + ".hasMoreElements"),
                    new ArgumentsBuilder().generatedAST())
                  .generatedAST(),
                  1);

      forUpdate = new 
        ASTForUpdate(ADLJavaParserTreeConstants.JJTFORUPDATE);
      castexpr = new
        ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
      FPtyp1 = (ASTType)domFP[i].jjtGetChild(0);
      parentChild(castexpr, FPtyp1, 0);
      parentChild(castexpr,
                  new PrimaryExpressionBuilder
                  ( makeName("enum" + i + ".nextElement"),
                    new ArgumentsBuilder().generatedAST())
                  .generatedAST(),
                  1);

      upd = new AssignmentBuilder
        ( ((ASTVariableDeclaratorId)domFP[i].jjtGetChild(1)).getName(),
          "=", castexpr)
        .generatedAST();
      parentChild(forUpdate, upd, 0);

      parentChild(forEnum, forExpr, 0);
      parentChild(forEnum, forUpdate, 1);

      last = forEnum;

      if (i == nbLoops-1) {
        // visit the assertion-expression
        isInQuantified = true; isDecomposable = false;
        Node assGrp = (Node)node.jjtGetChild(2).jjtAccept(this, data);    
        isInQuantified = false;
        isDecomposable = isInAssertion && ! (isInCallState || isInTry || isInBinding);


        // tddRes = assGrp
        parentChild(last, new StatementBuilder
                    ( new AssignmentBuilder
                      ( "tddRes", "=", assGrp)).generatedAST(),
                    2);
      }
      else {
        forEnum = new
          ASTForStatement(ADLJavaParserTreeConstants.JJTFORSTATEMENT);
        parentChild(last, forEnum, 2);
      }
    }

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_ImplExpression.
    * <P><B>Algorithm: </B><UL><LI>visits the 2 operand expressions (which
    * return ret_exp1 and ret_exp2).
    * <LI><B>CASE</B> operator :<UL>
    * <LI><B>==&#62;</B> : return !(ret_exp1) || (ret_exp2)
    * <LI><B>&#60;==</B> : return (ret_exp1) || !(ret_exp2)
    * <LI><B>&#60;=&#62;</B> : return (ret_exp1) == (ret_exp2)
    * <LI><B>&#60;:&#62;</B> : return (ret_exp1) ? "abnormal" : !(ret_exp2 && "abnormal")
    * </UL></UL>
    * @param node the ASTADL_ImplExpression to visit.
    * @param data unused.
    * @return the created node.
    */
  public Object visit(ASTADL_ImplExpression node, Object data) { 

    if (debug) System.out.println("ASTADL_ImplExpression");

    TypedNode createdNode;
    Node ch1 = node.jjtGetChild(0),
      ch2 = node.jjtGetChild(2);

    ASTADL_ImplOp op = (ASTADL_ImplOp)node.jjtGetChild(1);
    String tmpStr = "", tmpName;

    if (isDecomposable) {
      tmpStr = node.getExpressionImage();
    }      
     
    if (op.getOp().equals("==>") || 
        op.getOp().equals("<==")) {
      if (op.getOp().equals("<==")) {  // exchange (exp1) and (exp2)
        Node tmp = ch1;
        ch1 = ch2;
        ch2 = ch1;
      }

      // e1 ==> e2 changed into !(e1) || (e2)
      createdNode = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      createdNode.setType(JBuiltinType.booleanT);

      parentChild(createdNode,
                  new NotBuilder(ch1,
                                 !(ch1 instanceof ASTParenthExpression))
                  .generatedAST(),
                  0);
      parentChild(createdNode, ch2, 1);
      createdNode = (TypedNode)createdNode.jjtAccept(this, data);
    }

    else if (op.getOp().equals("<=>")) {
      // e1 <=> e2 changed into (e1) == (e2)
      createdNode = new
        ASTEqualityExpression(ADLJavaParserTreeConstants.JJTEQUALITYEXPRESSION);
      createdNode.setType(JBuiltinType.booleanT);
      String[] ops = { "==" };
      ((SimpleNode)createdNode).setOps(ops);

      parentChild(createdNode, ch1, 0);
      parentChild(createdNode, ch2, 1);
      createdNode = (TypedNode)createdNode.jjtAccept(this, data);
    }

    else { // "<:>"
      /*
      // e1 <:> e2 changed into (e1) ? "abnormal" : !((e2) && "abnormal")
      createdNode = new
        ASTConditionalExpression(ADLJavaParserTreeConstants.JJTCONDITIONALEXPRESSION);

      parentChild(createdNode, parenth1, 0);
      
      ASTName subnode1 = makeName("abnormal");
      parentChild(createdNode, subnode1, 1);


      ASTConditionalAndExpression subnode2_1 = new
        ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
      parentChild(subnode2_1, parenth2, 0);
      parentChild(createdNode, new NotBuilder(subnode2_1).generatedAST(), 2);

      ASTName subnode2_1_2 = makeName("abnormal");
      parentChild(subnode2_1, subnode2_1_2, 1);
      */

      // e1 <:> e2 changed into ((!(e1) || "abnormal") && (!"abnormal" && !(e2) && (e1)))
      createdNode = new
        ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
      createdNode.setType(JBuiltinType.booleanT);

      ASTParenthExpression subnode1 = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);

      ASTConditionalOrExpression snode1 = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      snode1.setType(JBuiltinType.booleanT);

      parentChild(createdNode, subnode1, 0);
      parentChild(subnode1, snode1, 0);
      ch1 = (Node)ch1.jjtAccept(this, data);
      parentChild(snode1, new NotBuilder(ch1,
                                         !(ch1 instanceof ASTParenthExpression))
                  .generatedAST(), 0);
      parentChild(snode1,
                  makeName("abnormal"),
                  1);

      ASTParenthExpression child2 = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
      parentChild(createdNode, child2, 1);

      ASTConditionalOrExpression subnode2 = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      subnode2.setType(JBuiltinType.booleanT);
      parentChild(child2, subnode2, 0);

      Node p1bis = null;
      try {
        p1bis = (Node)((SimpleNode)ch1).clone();
      } catch(CloneNotSupportedException e) {}

      // lazy evaluation: 2nd child must be evaluated only when abnormal is true
      boolContextStack.push(new BoolContext("abnormal", true));
      ch2 = (Node)ch2.jjtAccept(this, data);
      boolContextStack.pop();

      parentChild(subnode2,
                  new NotBuilder(makeName("abnormal"), false).generatedAST(),
                  0);
      parentChild(subnode2, new NotBuilder(ch2, false).generatedAST(), 1);
      parentChild(subnode2, p1bis, 2);
    }

    if (isDecomposable) {
      return makeName(makePreAssertion(makeType("boolean"), createdNode, tmpStr));
    }
    else {
      return createdNode;
    }    
  }


  /**
    * Visit pattern for a node of type ASTConditionalExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConditionalExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see #decompose
    */
  public Object visit(ASTConditionalExpression node, Object data) { 

    if (debug) System.out.println("ASTConditionalExpression");

    // lazy evaluation:
    // b ? e1 : e2
    // --> t1 = visit(b);
    // t2 = t1 ? visit(e1) : <default value>;
    // t3 = t1 ? <default value> : visit(e2);
    // t4 = t1 ? t2 : t3

    parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
    // ATTENTION: we suppose that the boolean expression is ALWAYS decomposed,
    // even for instance for literal values
    int ind = preAssIndex-1;
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, true));
    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    boolContextStack.pop();
    boolContextStack.push(new BoolContext("__ADL_tmp_" + ind, false));
    parentChild(node, (Node)node.jjtGetChild(2).jjtAccept(this, data), 2);
    boolContextStack.pop();

    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTConditionalOrExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConditionalOrExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTConditionalOrExpression node, Object data) { 

    if (debug) System.out.println("ASTConditionalOrExpression");

    if (node.jjtGetNumChildren() > 2) {
      // to simplify lazy evaluation mechanism, we use associativity:
      // e1 || e2 || e3... is first changed in e1 || (e2 || e3...)
      ASTConditionalOrExpression newOr = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      newOr.setType(JBuiltinType.booleanT);
      newOr.copySourceTokens(node);

      parentChild(newOr, node.jjtGetChild(0), 0);
      ASTConditionalOrExpression newSubOr = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      newSubOr.setType(JBuiltinType.booleanT);

      // copy source tokens and skip first child tokens
      newSubOr.copySourceTokens(node);
      newSubOr.firstToken = ((SimpleNode)node.jjtGetChild(1)).firstToken;

      parentChild(newOr, newSubOr, 1);

      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        parentChild(newSubOr, node.jjtGetChild(i), i-1);
      }

      return newOr.jjtAccept(this, data);
    }

    // lazy evaluation:
    // e1 || e2
    // --> t1 = visit(e1);
    // t2 = t1 ? false : visit(e2)
    // t3 = t1 || t2
    parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
    boolContextStack.push(new BoolContext("__ADL_tmp_" + (preAssIndex-1), false));
    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    boolContextStack.pop();

    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTConditionalAndExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTConditionalAndExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTConditionalAndExpression node, Object data) { 

    if (debug) System.out.println("ASTConditionalAndExpression");

    if (node.jjtGetNumChildren() > 2) {
      // to simplify lazy evaluation mechanism, we use associativity:
      // e1 && e2 && e3... is first changed in e1 && (e2 && e3...)
      ASTConditionalAndExpression newAnd = new
        ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
      newAnd.setType(JBuiltinType.booleanT);
      newAnd.copySourceTokens(node);
      parentChild(newAnd, node.jjtGetChild(0), 0);

      ASTConditionalAndExpression newSubAnd = new
        ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
      newSubAnd.setType(JBuiltinType.booleanT);

      // copy source tokens and skip first child tokens
      newSubAnd.copySourceTokens(node);
      newSubAnd.firstToken = ((SimpleNode)node.jjtGetChild(1)).firstToken;

      parentChild(newAnd, newSubAnd, 1);

      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        parentChild(newSubAnd, node.jjtGetChild(i), i-1);
      }

      return newAnd.jjtAccept(this, data);
    }

    // lazy evaluation:
    // e1 && e2
    // --> t1 = visit(e1);
    // t2 = t1 ? visit(e2) : false
    // t3 = t1 && t2
    parentChild(node, (Node)node.jjtGetChild(0).jjtAccept(this, data), 0);
    boolContextStack.push(new BoolContext("__ADL_tmp_" + (preAssIndex-1), true));
    parentChild(node, (Node)node.jjtGetChild(1).jjtAccept(this, data), 1);
    boolContextStack.pop();

    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTInclusiveOrExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTInclusiveOrExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTInclusiveOrExpression node, Object data) { 

    if (debug) System.out.println("ASTInclusiveOrExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTExclusiveOrExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTExclusiveOrExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTExclusiveOrExpression node, Object data) { 

    if (debug) System.out.println("ASTExclusiveOrExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTAndExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTAndExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTAndExpression node, Object data) { 

    if (debug) System.out.println("ASTAndExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTEqualityExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTEqualityExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTEqualityExpression node, Object data) { 

    if (debug) System.out.println("ASTEqualityExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTRelationalExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTRelationalExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTRelationalExpression node, Object data) { 

    if (debug) System.out.println("ASTRelationalExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTShiftExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTShiftExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTShiftExpression node, Object data) { 

    if (debug) System.out.println("ASTShiftExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTAdditiveExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTAdditiveExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTAdditiveExpression node, Object data) { 

    if (debug) System.out.println("ASTAdditiveExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTMultiplicativeExpression: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTMultiplicativeExpression to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTMultiplicativeExpression node, Object data) { 

    if (debug) System.out.println("ASTMultiplicativeExpression");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTUnaryPlusMinus: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTUnaryPlusMinus to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTUnaryPlusMinus node, Object data) { 

    if (debug) System.out.println("ASTUnaryPlusMinus");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTUnaryNotPlusMinus: just a call
    * to decompose(node, data).
    * <BR>See the expression decomposition algorithm.
    * @param node the ASTUnaryNotPlusMinus to visit.
    * @param data unused.
    * @return the transformed node.
    * @see decompose
    */
  public Object visit(ASTUnaryNotPlusMinus node, Object data) { 

    if (debug) System.out.println("ASTUnaryNotPlusMinus");

    return decompose(node, data);
  }


  /**
    * Visit pattern for a node of type ASTCastExpression.
    * <BR>Just visit the expression child (nothing to do for the type child).
    * <BR>A cast expression is not decomposable.
    * @param node the ASTCastExpression to visit.
    * @param data unused.
    * @return the transformed node.
    */
  public Object visit(ASTCastExpression node, Object data) { 

    if (debug) System.out.println("ASTCastExpression");

    int lastnode = node.jjtGetNumChildren()-1;
    parentChild(node, (Node)node.jjtGetChild(lastnode).jjtAccept(this, data), lastnode);

    return node;
  }


  /**
    * Visit pattern for a node of type ASTADL_CallStateExpression.
    * <P><B>Algorithm: </B><OL>
    * <LI>call makeSaveBlock to create a "saveBlock", i.e. a declaration of a
    * temporary local variable that will be initialized with the type and value
    * of this node child expression <I>before</I> the call to the tested method.
    * <LI><UL><LI><B>IF</B> the expression is decomposable, call
    * "makePreAssertion" to create a "preAssertion", i.e. a declaration of a
    * temporary local variable that will be initialized with the type and value
    * of the precedently created variable, and return this second variable name
    * (that is considered as a sub-expression in the decomposition algorithm).
    * <LI><B>ELSE</B> return the name of the saveBlock variable.
    * </UL> </OL>
<PRE>
|  &#60;type of @expr&#62; __ADL_savTmp_i = &#60;@expr&#62;; // saveBlock
|  // ...
|  // &#60;call to tested method&#62;
|  // ...
|  // assertion:
|  &#60;type of @expr&#62; __ADL_tmp_j = __ADL_savTmp_i; // preAssertion
</PRE>
    * @param node the ASTADL_CallStateExpression to visit.
    * @param data unused.
    * @return the created node.
    * @see #makeSaveBlock
    * @see #makePreAssertion(ORG.opengroup.parsers.adljava.ASTType,
    * ORG.opengroup.parsers.adljava.Node, java.lang.String)
    */
  public Object visit(ASTADL_CallStateExpression node, Object data) { 

    if (debug) System.out.println("ASTADL_CallStateExpression");

    if (isInCallState) {
      return node.jjtGetChild(0).jjtAccept(this, data);
    }
    TypedNode child = (TypedNode)node.jjtGetChild(0);
    ASTName createdNode;
    String tmpStr = node.getExpressionImage();
    LType exprType = child.getType();

    isInCallState = true; isDecomposable = false;
    createdNode = makeName
      (makeSaveBlock(exprType, (Node)child.jjtAccept(this, null)));
    isInCallState = false;
    isDecomposable = isInAssertion && ! (isInTry || isInQuantified || isInBinding);
    if (isDecomposable) {
      return makeName(makePreAssertion(TypeAux.LType2ASTType(exprType),
                                       createdNode, tmpStr));
    }
    else {
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_Unchanged.
    * <P><B>Algorithm:</B><UL>
    * <LI>clone the node children (arguments of the "unchanged" expression).
    * <LI>copy the source tokens (that represents the 
    * code of the expression) on the cloned arguments.
    * <LI><B>IF</B> there is <I>one</I> argument "arg", create a node "(@arg ==
    * clone_arg)" and visit its 2 children nodes.
    * <LI><B>ELSE</B> (there are <I>n</I> arguments
    * "arg<SUB>1</SUB>"..."arg<SUB>n</SUB>"): create a node
    * "((@arg<SUB>1</SUB> == clone_arg<SUB>1</SUB>) && ... &&
    * (@arg<SUB>n</SUB> == clone_arg<SUB>n</SUB>))" and visit all
    * "@arg<SUB>i</SUB>" and "clone_arg<SUB>i</SUB>" child nodes. </UL>
    * @param node the ASTADL_Unchanged to visit.
    * @param data unused.
    * @return the created node.
    */
  public Object visit(ASTADL_Unchanged node, Object data) { 

    if (debug) System.out.println("ASTADL_Unchanged");

    if (node.jjtGetChild(0).jjtGetNumChildren() == 0) {
      // unchanged() ---> true
      ASTBooleanLiteral createdNode = new
        ASTBooleanLiteral(ADLJavaParserTreeConstants.JJTBOOLEANLITERAL);
      createdNode.setValue(true);
      createdNode.copySourceTokens(node);

      return createdNode;
    }

    ASTArgumentList argList = (ASTArgumentList)node.jjtGetChild(0).jjtGetChild(0);
    int nbexpr = argList.jjtGetNumChildren();
    
    ASTArgumentList argCopy = null;
    try {
      argCopy = (ASTArgumentList)argList.clone();
    } catch (CloneNotSupportedException e) {}

    if (nbexpr == 1) {
      // unchanged(expr) ---> (expr == @expr)
      
      SimpleNode expr1 = (SimpleNode)argList.jjtGetChild(0);
      SimpleNode expr2 = (SimpleNode)argCopy.jjtGetChild(0);

      ASTParenthExpression createdNode = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
      ASTEqualityExpression eq = new
        ASTEqualityExpression(ADLJavaParserTreeConstants.JJTEQUALITYEXPRESSION);
      ASTADL_CallStateExpression callstatexpr = new
        ASTADL_CallStateExpression(ADLJavaParserTreeConstants.JJTADL_CALLSTATEEXPRESSION);

      String[] ops = { "==" };
      eq.setOps(ops);
      eq.setType(JBuiltinType.booleanT);
      eq.copySourceTokens(node);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens(expr1);
      callstatexpr.prependToken(new Token("@"));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(createdNode, (Node)eq.jjtAccept(this, data), 0);

      return createdNode;
    }

    // unchanged(expr1, ..., exprn) ---> ((expr1 == @expr1) && ... && (exprn == @exprn))
    ASTParenthExpression createdNode = new
      ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
    ASTConditionalAndExpression andNode = new
      ASTConditionalAndExpression(ADLJavaParserTreeConstants.JJTCONDITIONALANDEXPRESSION);
    andNode.setType(JBuiltinType.booleanT);

    SimpleNode expr1, expr2; ASTParenthExpression parenth;
    ASTEqualityExpression eq; ASTADL_CallStateExpression callstatexpr;
    String[] ops = { "==" };

    for (int i = 0; i < nbexpr; i++) {
      expr1 = (SimpleNode)argList.jjtGetChild(i);
      expr2 = (SimpleNode)argCopy.jjtGetChild(i);

      parenth = new
        ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
      eq = new
        ASTEqualityExpression(ADLJavaParserTreeConstants.JJTEQUALITYEXPRESSION);
      callstatexpr = new
        ASTADL_CallStateExpression(ADLJavaParserTreeConstants.JJTADL_CALLSTATEEXPRESSION);

      eq.setType(JBuiltinType.booleanT);
      eq.setOps(ops);

      // copy expr source tokens and push a "@" token in front of these
      callstatexpr.copySourceTokens(expr1);
      callstatexpr.prependToken(new Token("@"));
      eq.copySourceTokens(callstatexpr);
      eq.prependToken(new Token(expr1.getExpressionImage() + " == "));

      parentChild(eq, expr1, 0);
      parentChild(callstatexpr, expr2, 0);
      parentChild(eq, callstatexpr, 1);
      parentChild(parenth, eq, 0);
      parentChild(andNode, parenth, i);
    }
    andNode.copySourceTokens(node);
    parentChild(createdNode, (Node)andNode.jjtAccept(this, data), 0);

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTADL_Normal
    * @param node the ASTADL_Normal to visit.
    * @param data unused.
    * @return the ASTName node "normal".
    */
  public Object visit(ASTADL_Normal node, Object data) { 

    if (debug) System.out.println("ASTADL_Normal");

    ASTName n = makeName("normal");
    n.setType(JBuiltinType.booleanT);
    n.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(n));
    }
    else {
      return n;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_Abnormal.
    * @param node the ASTADL_Abnormal to visit.
    * @param data unused.
    * @return the ASTName node "abnormal".
    */
  public Object visit(ASTADL_Abnormal node, Object data) { 

    if (debug) System.out.println("ASTADL_Abnormal");

    ASTName abn = makeName("abnormal");
    abn.setType(JBuiltinType.booleanT);
    abn.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(abn));
    }
    else {
      return abn;
    }
  }


  /**
    * Visit pattern for a node of type ASTADL_Return.
    * @param node the ASTADL_Return to visit.
    * @param data unused.
    * @return the ASTName node "__ADL_returned".
    */
  public Object visit(ASTADL_Return node, Object data) { 

    if (debug) System.out.println("ASTADL_Return");

    if (inlining && isInCallState) {
      //      throw new ParseException("\"return\" in call-state expression via an inline");
      errorMessage("\"return\" in call-state expression via an inline",
                   node.getLineNumber());
    }

    ASTName ret = makeName("__ADL_returned");
    ret.setType(node.getType());
    ret.copySourceTokens(node);

    if (isDecomposable) {
      return makeName(makePreAssertion(ret));
    }
    else {
      return ret;
    }
  }


  /**
    * Visit pattern for a node of type ASTBooleanLiteral.
    * @param node the ASTBooleanLiteral to visit.
    * @param data unused.
    * @return the node.
    */
  public Object visit(ASTBooleanLiteral node, Object data) { 

    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTAllocationExpression.
    * @param node the ASTAllocationExpression to visit.
    * @param data unused.
    * @return the node.
    */
  public Object visit(ASTAllocationExpression node, Object data) { 

    node.jjtGetChild(1).jjtAccept(this, data);

    if (isDecomposable) {
      return makeName(makePreAssertion(node));
    }
    else {
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTPrimaryExpression.
    * <P><B>Algorithm</B>
    * <UL><LI><B>IF</B> the first child is not a name:
    *       <BR>usual decomposition algorithm (visit children and <B>return</B>
    *       the node if isDecomposable==false, call makePreAssertion otherwise).
    * <LI><B>ELSE</B><I> (first child is a name)</I>
    *     <UL><LI><B>IF</B> the first child is an inline name:
    *         <BR>create an ASTPrimaryExpression node that is a clone of the
    *         body of the corresponding inline and visit this node with flag
    *         "inlining" true, to replace formal parameters by actual ones
    *         (arguments: second child of the input node).
    *         <BR>If the input node has only 2 nodes (inline name + arguments),
    *         <B>return</B> the visited created node; otherwise (there are
    *         other suffixes after the inline call: &#60;inlineName&#62;
    *         (&#60;inlineArgs&#62;) &#60;suffix&#62;<SUP>*</SUP>) add these
    *         input suffixes as suffixes of the created node, visit it and
    *         <B>return</B> it.
    *     <LI><B>ELSE</B><I> (not a call to an inline macro)</I>
    *         <UL><LI><B>IF</B> "inlining"
    *             <UL><LI><B>IF</B> the first child is the name of a formal
    *                 parameter, create an ASTPrimaryExpression node whose
    *                 first child is a clone of the corresponding actual
    *                 parameter and other children are the input node's suffix
    *                 children; visit it and <B>return</B> it.
    *             <LI><B>ELSE IF</B> the first child is a qualified name whose
    *                 first component is the name of a formal parameter of the
    *                 currently inlined macro
    *                 <UL><LI><B>IF</B> the clone of the corresponding actual
    *                     parameter is a name, create an ASTPrimaryExpression
    *                     node whose first child is an ASTName node with as name
    *                     the concatenation of the clone name and the rest of
    *                     the input qualified name, and other children are the
    *                     input node's suffix children; visit it and
    *                     <B>return</B> it.
    *                     <BR>&#60;&#60;Name: inlineFormalP&#62; . &#60;nameRest&#62;&#62;
    *                     &#60;suffix&#62;<SUP>*</SUP> ---&#62;
    *                     &#60;&#60;Name: inlineActualP&#62; . &#60;nameRest&#62;&#62;
    *                     &#60;suffix&#62;<SUP>*</SUP>
    *                 <LI><B>ELSE</B> create an ASTPrimaryExpression whose
    *                     first child is the clone of the corresponding actual
    *                     parameter, the second child is an ASTDotSuffix whose
    *                     name is the rest of the input qualified name, and
    *                     other children are the input node's suffix children;
    *                     visit it and <B>return</B> it.
    *                     <BR>&#60;&#60;Name: inlineFormalP&#62; . &#60;nameRest&#62;&#62;
    *                     &#60;suffix&#62;<SUP>*</SUP> ---&#62;
    *                     &#60;PrimaryPrefix: inlineActualP&#62; &#60;DotSuffix: nameRest&#62;
    *                     &#60;suffix&#62;<SUP>*</SUP>
    *                 </UL>
    *             </UL> 
    *         <LI>If the first child name corresponds to a component of the
    *             real object (i.e. it is not qualified, and this is neither a
    *             method parameter nor a binding name) then qualify this name
    *             by preceding it with the real object name (or the adlclass
    *             name in the case of a constructor).
    *         </UL>
    *     </UL>
    * <LI>Visit all suffix children.
    * <LI>If the expression is decomposable, call makePreAssertion else
    *     <B>return</B> the node.
    * </UL>
    * @param node the ASTPrimaryExpression to visit.
    * @param data unused.
    * @return the created node.
    * @see #makePreAssertion(ORG.opengroup.parsers.adljava.TypedNode) */
  public Object visit(ASTPrimaryExpression node, Object data) { 

    if (debug) System.out.println("ASTPrimaryExpression");

    int nbChildren = node.jjtGetNumChildren();
    SimpleNode child1 = (SimpleNode)node.jjtGetChild(0);

    if (child1 instanceof ASTName) {
      String child1Name = ((ASTName)child1).getName();

      boolean isInline;
      try {
        isInline = ((String)child1.getProperty("astNameKind")).equals("ADLInline");
      }
      catch(PropertyAccessException e) { isInline = false; }

      if (isInline) {
        inlineFormalParams = null;
        for (Enumeration e = inlineDeclaration.elements(); 
             e.hasMoreElements(); ) {
          currentInline = (ASTADL_InlineDeclaration)e.nextElement();
          if (((SimpleNode)currentInline.jjtGetChild(1)).getName().equals(child1Name)) {
            inlineFormalParams = (ASTFormalParameters)currentInline.jjtGetChild(1).jjtGetChild(0);
            break;
          }
        }
        if (node.jjtGetChild(1).jjtGetNumChildren() > 0)
          inlineActualParams = (ASTArgumentList)node.jjtGetChild(1).jjtGetChild(0);

        ASTADL_AssertionGroup inlineBody = null;
        try {
          inlineBody = (ASTADL_AssertionGroup)
            ((SimpleNode)currentInline.jjtGetChild(2)).clone();
        } catch (CloneNotSupportedException e) {}
	
	// copy original source tokens in inline body expression node
	SimpleNode statement =
	  (SimpleNode)inlineBody.jjtGetChild(0).jjtGetChild(0);
	if (statement instanceof ASTADL_Assertion) {
	  // ADLAssertion : reach the ADL_Expression node
	  statement = (SimpleNode)statement
	    .jjtGetChild(statement.jjtGetNumChildren() - 1);
	} else {
	  // ADL_IfStatement || ADL_TryStatement : we have the right node
	}
	// the next statement has no side-effect on the Token chain of
	// the original inline body node since it has first been cloned.
	statement.prependToken(new Token("[inline call: " +
					 node.getExpressionImage() +
					 "] "));

        if (node.jjtGetNumChildren() == 2) { // only one PrimarySuffix
          inlining = true;
          SimpleNode createdNode = (SimpleNode)inlineBody.jjtAccept(this, data);
          inlining = false;

	  // copy source tokens from original node
	  createdNode.copySourceTokens(node);

          return createdNode;
        }
        else { // more than one PrimarySuffix
          inlining = true;
          PrimaryExpressionBuilder primBuild = new
            PrimaryExpressionBuilder((Node)inlineBody.jjtAccept(this, data));
          inlining = false;

	  // now visit remaining suffixes
          for (int i = 2; i < node.jjtGetNumChildren(); i++) {
            primBuild.addSuffix((Node)node.jjtGetChild(i).jjtAccept(this, data));
          }

          ASTPrimaryExpression prim = primBuild.generatedAST();
          prim.setType(node.getType());

	  // copy source tokens from original node
	  prim.copySourceTokens(node);

          return prim;// visitedPrim;
        }
      }
    
      else { // child1 instanceof ASTName && ! isInline
        if (inlining) {
          int firstdot = child1Name.indexOf('.');
          String firstName;
          if (firstdot == -1)
            firstName = child1Name;
          else
            firstName = child1Name.substring(0, firstdot);
          int param = -1;
          for (int i = inlineFormalParams.jjtGetNumChildren()-1; i >= 0; i--) {
            if (firstName.equals(((SimpleNode)inlineFormalParams
                                  .jjtGetChild(i).jjtGetChild(1)).getName())) {
              param = i;
              break;
            }
          }
          if (param != -1) { // formal parameter
            ASTPrimaryExpression createdNode = new
              ASTPrimaryExpression(ADLJavaParserTreeConstants.JJTPRIMARYEXPRESSION);
            SimpleNode actual = null;
            try {
              actual = (SimpleNode)
                ((SimpleNode)inlineActualParams.jjtGetChild(param))
                .clone();
            }
            catch (CloneNotSupportedException e) {}

            int suff = 0;
            if (firstdot != -1) { // inlineParam.aName <aSuffix>
              if (actual instanceof ASTName) {
                // -->  inlineActualParam.aName <aSuffix>
                actual = makeName(actual.getName() + child1Name.substring(firstdot));
              }
              else {
                // -->  inlineActualParam <dotsuffix: aName> <aSuffix>
                ASTDotSuffix dotSuff = new
                  ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
                dotSuff.setName(child1Name.substring(firstdot+1));
                parentChild(createdNode, dotSuff, 1);
                suff = 1;
              }
            }

            parentChild(createdNode, actual, 0);

            createdNode.setType(node.getType());
	    createdNode.copySourceTokens(node);

            for (int i = 1; i < node.jjtGetNumChildren(); i++) {
              parentChild(createdNode, node.jjtGetChild(i), i+suff);
            }

            return createdNode.jjtAccept(this, data);
          }
        }

        if (isRealObjectComponent((ASTName)child1)) {
          if (isInConstructor) {
            child1.setName("__ADL_returned." + child1.getName());
          }
          else {
            child1.setName(ACO.getAdlClassName() + "Object." + child1.getName());
          }
        }
      }
    }
    
    else { // !(child1 instanceof ASTName)
      parentChild(node, (Node)child1.jjtAccept(this, data), 0);
    }

    for (int i = 1 ; i < node.jjtGetNumChildren(); i++) {
      parentChild(node, (Node)node.jjtGetChild(i).jjtAccept(this, data), i);
    }

    if (isDecomposable)
      return makeName(makePreAssertion(node));
    else
      return node;
  }


  /**
    * Visit pattern for a node of type ASTThisPrefix.
    * <BR>In a constructor, returns an ASTName that corresponds to the object
    * created by the constructor call, otherwise returns an ASTName that
    * corresponds to the "real object" name.
    * @param node the ASTThisPrefix to visit.
    * @param data unused.
    * @return an ASTName node.
    */
  public Object visit(ASTThisPrefix node, Object data) { 

    if (debug) System.out.println("ASTThisPrefix");

    if (isInConstructor)
      return makeName("__ADL_returned");
    else
      return makeName(ACO.getAdlClassName() + "Object");
  }


  /**
    * Visit pattern for a node of type ASTADL_ThrownExpression.
    * <P><B>Algorithm</B><UL>
    * <LI>Get the list of exception names of the "thrown" clause.
    * <LI><B>IF</B> there is <I>one</I> name, create a node for the code
    * "(adl_thrownException instanceof &#60;name&#62;)".
    * <LI><B>ELSE</B> create a node for the code "(adl_thrownException
    * instanceof &#60;name<SUB>1</SUB>&#62; || ... || adl_thrownException instanceof
    * &#60;name<SUB>n</SUB>&#62;)"
    * <LI>If the expression is decomposable, call makePreAssertion, else
    * return the created node.
    * </UL>
    * @param node the ASTADL_ThrownExpression to visit.
    * @param data unused.
    * @return the created node.
    */
  public Object visit(ASTADL_ThrownExpression node, Object data) { 

    if (debug) System.out.println("ASTADL_ThrownExpression");

    ASTNameList child = (ASTNameList)node.jjtGetChild(0);
    String[] names;
    ASTParenthExpression createdNode = new
      ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);

    try {
      names = (String[])child.getProperty("namelist");
    }
    catch(PropertyAccessException e) {
      names = null;
    }

    int nbChildren = names.length;

    if (nbChildren == 1) {
      ASTInstanceOfExpression inst = new
        ASTInstanceOfExpression(ADLJavaParserTreeConstants.JJTINSTANCEOFEXPRESSION);
      parentChild(createdNode, inst, 0);
      parentChild(inst, makeName("adl_thrownException"), 0);
      parentChild(inst, makeType(names[0]), 1);
    }
    else {
      ASTConditionalOrExpression orexp  = new
        ASTConditionalOrExpression(ADLJavaParserTreeConstants.JJTCONDITIONALOREXPRESSION);
      parentChild(createdNode, orexp, 0);
      ASTInstanceOfExpression subnode;
      for (int i = 0; i < nbChildren; i++) {
        subnode = new
          ASTInstanceOfExpression(ADLJavaParserTreeConstants.JJTINSTANCEOFEXPRESSION);
        parentChild(orexp, subnode, i);
        parentChild(subnode, makeName("adl_thrownException"), 0);
        parentChild(subnode, makeType(names[i]), 1);
      }
    }

    if (isDecomposable) {
      String tmpStr = node.getExpressionImage();
      return makeName(makePreAssertion(makeType("boolean"), createdNode, tmpStr));
    }
    else {
      return createdNode;
    }
  }


  /**
    * Visit pattern for a node of type ASTName.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> "inlining" and the name is the name of a formal parameter
    * of the inline that is currently being inlined, then create a node that is
    * a clone of the corresponding actual parameter, visit it and <B>return</B>
    * it.
    * <LI><B>IF</B> the name corresponds to a component of the real object
    * (i.e. it is not qualified, and this is neither a method parameter nor a
    * binding name) then qualify this name by preceding it with the real object
    * name (or the adlclass name in the case of a constructor).
    * <LI>If the expression is decomposable, call makePreAssertion, else
    * return the node.
    * </UL>
    * @param node the ASTName to visit.
    * @param data unused.
    * @return the created node.
    */
  public Object visit(ASTName node, Object data) { 

    if (debug) System.out.println("ASTName: " + node.getName());

    if (node.getName().equals("adl_thrownException"))
      return node;

    if (inlining && isInCallState) {
      String astNameKind = null;
      try {
        astNameKind = (String)node.getProperty("astNameKind");
      }
      catch(PropertyAccessException e) { }
      if (astNameKind.equals("ADLBinding")) {
        errorMessage("Binding in a call-state via an inline",
                     node.getLineNumber());
      }
    }

    String myName = node.getName();
    if (inlining) {
      int param = -1;
      for (int i = inlineFormalParams.jjtGetNumChildren()-1; i >= 0; i--) {
        if (myName.equals(((SimpleNode)inlineFormalParams.jjtGetChild(i).jjtGetChild(1)).getName())) {
          param = i;
          break;
        }
      }
      if (param != -1) {
        SimpleNode createdNode = null;
        try {
          inlining = false;
          createdNode = (SimpleNode)((SimpleNode)((SimpleNode)
            inlineActualParams.jjtGetChild(param)).clone()).jjtAccept(this, data);
          inlining = true;
        }
        catch (CloneNotSupportedException e) {}
        return createdNode;
      }
    }

    if (isRealObjectComponent(node)) {
      if (isInConstructor) {
        node.setName("__ADL_returned." + myName);
      }
      else {
        node.setName(ACO.getAdlClassName() + "Object." + myName);
      }
    }

    if (isDecomposable) {
      return makeName(makePreAssertion
                      ( TypeAux.LType2ASTType(((TypedNode)node).getType()),
                        node, myName));
    }
    return node;
  }


  /**
    * Print an error message.
    * @param msg the error message
    * @param lineNumber the line number of the error in the source file
    */
  private void errorMessage(String msg, int lineNumber) {

    if (lineNumber ==0) {
      System.err.println(ACO.getAdlClassName() + ": " + msg);
    }
    else {
      System.err.println(ACO.getAdlClassName() + ".adl:" + lineNumber + ": " + msg);
    }
    syntaxError = true;
  }


  /**
    * Print a message in case of an error internal to the ADLT compiler.
    * Should not happen!
    */
  private void internalError() {

    System.err.println("Internal error in ADLT 2.0 Compiler: please contact ADL team");
    syntaxError = true;
  }

}

// ***********************************************************************
// History Log:
// $Log: JADLTransformVisitor.java,v $
// Revision 1.25  1998/10/15 18:14:09  mallet
// Bug fixed: pr/153: encapsulated call-state expressions and assertions
// within a try/catch statement to trap any exception raised during their
// evaluation.
//
// Revision 1.24  1998/08/07 11:05:40  mallet
// Updated according to change in ADLJava.jjt: name of
// ADL_ClassDeclaration node is now a FQN.
// Bug pr/150 fixed: the return type for a constructor uses the
// implementation class name instead of the adl class one.
//
// Revision 1.23  1998/06/30 09:33:26  mallet
// Replaced import of PropertyAccessException in
// ORG.opengroup.adl.parsers by an import on demand (.*).
// Replaced access to expressionImage property by call to
// getExpressionImage().
// Replaced setting of expressionImage property on newly created nodes by
// use of copySourceTokens() and other token chain related methods.
// Fixed bugs in processing of ASTADL_Unchanged: was only visiting the
// children of the equivalent equality expression instead of the equality
// itself.
// Fixed bug in visit of ASTPrimaryExpression for inline calls: was
// incorrectly getting the inline body expression.
// Changed debug mode reporting for inline calls: now also reporting the
// initial inline call expression.
//
// Revision 1.22  1998/06/29 12:55:21  bareau
// Added support for allocation expressions
// (e.g. "(new Bank()).get_active_accounts() == 0;").
//
// Revision 1.21  1998/06/23 12:26:10  mallet
// Added support for FQ name of adlclass (like org.omg.CORBA.Object): the
// ACO class uses the short name for its name but the FQN for all
// variables and references to the implementation class.
//
// Revision 1.20  1998/06/23 08:10:36  bareau
// Bug fix: the visit of the actual argument of an inline call must be
// done with flag inlining false.
//
// Revision 1.19  1998/06/18 10:30:36  mallet
// Fixed incorrect use of type short name instead of full one (FQN).
//
// Revision 1.18  1998/06/16 14:37:35  bareau
// Bug fix: there was an error in the inlining of inlines with no parameter.
//
// Revision 1.17  1998/06/11 15:44:50  bareau
// Bug fix: the transformation for nested ADL_IfStatement was incorrect.
//
// Revision 1.16  1998/04/15 13:57:41  bareau
// Update according to the new NL mechanism: there are no NLD_* nodes.
//
// Revision 1.15  1998/03/31 12:39:25  bareau
// Use of "super.semantics" in a constructor: call first the method of the
// inherited ACO that corresponds to the overridden constructor (these
// methods are not overridden by default).
//
// Revision 1.14  1998/03/20 14:04:39  bareau
// Bug fix: an inline call may occur _before_ the corresponding
// inline definition.
//
// Unchanged expression: completed the decomposition (there was a
// missing infoline).
//
// Revision 1.13  1998/03/19 17:26:34  bareau
// Importations in case of unnamed packages are now dealt with using
// property "classesFromUnnamedPackage".
//
// Revision 1.12  1998/03/16 15:12:48  bareau
// Property expressionImage is now set in the jjt and not in the translator
// (nodes unchanged, (ab)normal, call state expression, etc.)
//
// Revision 1.11  1998/03/13 08:48:17  bareau
// Implementation of lazy evaluation for logical operations.
// See inner class BoolContext, the method makePreAssertion and the
// visit methods for ConditionalExpression, Conditional{And/Or}Expression
// and ASTADL_ImplExpression.
//
// Bug fix: there was a bug when an expression f(foo, foo.bar()) was used
// in a binding of parameter foo (incorrect cloning).
//
// Revision 1.10  1998/03/10 17:26:46  mallet
// Merged v1.1 with development trunk
//
// Revision 1.9.2.4  1998/03/09 17:33:27  bareau
// Precedent commit was incorrect.
//
// Revision 1.9.2.3  1998/03/09 17:17:30  bareau
// Bug fix: missing "!" in translation of "<:>".
//
// Revision 1.9.2.2  1998/03/04 12:35:00  bareau
// Reset of local prolog/epilog for each method.
//
// Revision 1.9.2.1  1998/03/04 10:06:14  bareau
// The entry point for adl/tdd visitors is now a method "translate"
// that takes as parameter the root node of the source tree, performs
// necessary initializations and launches the visit.
//
// Revision 1.9  1998/02/27 13:07:57  bareau
// Updated copyright notice.
//
// Revision 1.8  1998/02/24 18:33:35  bareau
// Changed the root node of generated trees as ASTCompilationUnit,
// with package "adltest".
//
// Revision 1.7  1998/02/12 16:03:56  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.6  1998/02/11 17:22:36  bareau
// Added the source line number to error messages.
//
// Revision 1.5  1998/02/11 10:13:41  bareau
// Syntax error check: presence of a "return" or a binding inside a
// call-state expression via an inline call.
// Added a rudimentary error mechanism to deal with such errors:
// the transformer returns null.
//
// NLD annotations nodes: just return the node without visiting its children.
//
// Revision 1.4  1998/01/26 15:12:45  bareau
// Minor bug fixes.
//
// Revision 1.3  1998/01/21 16:23:53  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:47  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:00  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:44  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:02:19  adl2ext
// Transformer visitor for ADL/Java.
// Initial revision.
//
// ***********************************************************************

