// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JBuilder.java,v $
// $Revision: 1.9 $
// $Date: 1998/10/14 08:12:47 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        JBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import java.util.Vector;

/**
  * Abstract class, ancestor of the builder classes for all TDD
  * declarations (datasets, factories, directives).
  * @author Cyrille Bareau
  * @see JDatasetBuilder
  * @see JDirectiveBuilder
  * @see JFactoryBuilder
  */

public abstract class JBuilder {

  /** Handle on the root of the generated Java AST */
  public ASTCompilationUnit compilUnit;


  /** Handle on the "import" declarations */
  protected Vector importDecls;

  /** Name of the feature declaration in the source tdd class */
  protected String featureName;

  /** Name of the source tdd class */
  protected String unitName; 

  /** Name of the generated Java class */
  protected String className;

  /** Handle on the generated class body declaration AST */
  protected ASTClassBodyDeclaration body;

  /** Handle on the generated class declaration AST */
  protected ASTClassDeclaration dirClass;


  /**
    * Main constructor.
    * @param node the current feature declaration node
    * @param imports the "import" declarations
    * @param unit the name of the source tdd class
    */
  public JBuilder (SimpleNode node, Vector imports, String prefix, String unit) {
    super();
    importDecls = imports;
    unitName = unit;
    int nbChildren = 0;

    if (node instanceof ASTTDD_DatasetDeclaration) {
      featureName = ((SimpleNode)node.jjtGetChild(0).jjtGetChild(1)).getName();
    }
    else if (node instanceof ASTTDD_TestDirective) {
      // the parser names each directive
      featureName = node.getName();
    }
    else if (node instanceof ASTTDD_FactoryDeclaration) {
      featureName = ((SimpleNode)node.jjtGetChild(0).jjtGetChild(1)).getName();
    }
    else {
      // error
    }
    className = prefix + unitName + "_" + featureName;
  }

  /** 
   * Name of the class constructed for this TDD entity (dataset, factory,
   * test directive).
   */
  public String getClassName() { return className; }

  /**
   * Builds the qualified name of the file that implements the feature
   * (dataset or factory) described by the parameters.
   * @param prefix "F_" for a factory, "S_" for a dataset.
   * @param TDDClassName the name of the TDD class defining the feature
   * @param name the feature name, qualified if imported.
   * @return the built file name
   */
  public static String makeQualifiedName(String prefix,
                                         String TDDClassName,
                                         String featureName) {

    return prefix + TDDClassName + "_" + featureName;
  }


  /**
    * Generic build method.
    * Creates the tree that is common to all builder-generated Java files,
    * namely the root ASTADL_CompilationUnit and some of its children:
    * importations (ASTImportDeclaration), class declaration
    * (ASTClassDeclaration), name of the feature, name of the inherited
    * class, name of the implemented interface.
    * @param parentName name of the inherited feature class (ADLLiteralSet, etc.)
    * @see #compilUnit
    * @see #importDecls
    * @see #className
    */
  protected void build(String parentName) {

    compilUnit = new
      ASTCompilationUnit(ADLJavaParserTreeConstants.JJTCOMPILATIONUNIT);

    ASTPackageDeclaration pack = new
      ASTPackageDeclaration(ADLJavaParserTreeConstants.JJTPACKAGEDECLARATION);
    JGenericVisitor.parentChild(pack, JGenericVisitor.makeName("adltest"), 0);
    JGenericVisitor.parentChild(compilUnit, pack, 0);

    for (int i = 0; i < importDecls.size(); i++) {
      JGenericVisitor.parentChild(compilUnit,
                                  (ASTImportDeclaration)importDecls.elementAt(i),
                                  i+1);
    }

    dirClass = new
      ASTClassDeclaration(ADLJavaParserTreeConstants.JJTCLASSDECLARATION);
    JGenericVisitor.parentChild(compilUnit, dirClass, importDecls.size()+1);
    try {
      int modif = JModifier.PUBLIC;
      dirClass.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    dirClass.setName(className);

    ASTName extName = JGenericVisitor.makeName(parentName);
    JGenericVisitor.parentChild(dirClass, extName, 0);

    ASTNameList implemList = new
      ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(implemList, JGenericVisitor.makeName(unitName + "_CST"), 0);
    JGenericVisitor.parentChild(dirClass, implemList, 1);
    

    body = new
      ASTClassBodyDeclaration(ADLJavaParserTreeConstants.JJTCLASSBODYDECLARATION);
    JGenericVisitor.parentChild(dirClass, body, 2);
  }

}

// ***********************************************************************
// History Log:
// $Log: JBuilder.java,v $
// Revision 1.9  1998/10/14 08:12:47  mallet
// Bugs fixed: pr/155, pr/156: removed useless constructor for anonymous
// test directives: the parser now names each directive.
//
// Revision 1.8  1998/06/05 14:17:04  mallet
// Changed location of method makeQualifiedName from NameAux to JBuilder.
// Removed class NameAux.
//
// Revision 1.7  1998/02/27 13:07:58  bareau
// Updated copyright notice.
//
// Revision 1.6  1998/02/24 18:33:36  bareau
// Changed the root node of generated trees as ASTCompilationUnit,
// with package "adltest".
//
// Revision 1.5  1998/02/12 16:03:56  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.4  1998/02/02 12:47:13  cinard
// Added a getClassName method.
//
// Revision 1.3  1998/01/21 16:23:54  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:47  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:01  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:44  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 18:59:21  adl2ext
// Generic dataset component builder.
//
// ***********************************************************************
