// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JDatasetBuilder.java,v $
// $Revision: 1.8 $
// $Date: 1998/06/12 14:59:53 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        JDatasetBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Creates a Java compilation unit AST that corresponds to a dataset
  * declaration.
  * <BR>Uses JDatasetExprVisitor; is used by JTDDTransformVisitor.
  * <P>Example: definition of dataset TEST in tdd class bankTest:
  * <BR>"dataset bank TEST = make_bank(DEPOSITS, BANKSIZE + DEPOSITS,
  * INT_VALUES);"
  * @author Cyrille Bareau
  * @see JDatasetExprVisitor
  * @see JTDDTransformVisitor
  */

public class JDatasetBuilder extends JBuilder {


  /** The prefix used to give names to the components of the dataset
    * expression of the dataset declaration. */
  private String strStackInit = "elt";

  /** The name of the method used to add components of the dataset expression.
    * (addElement for dataset concatenation, addParameter for factory call,
    * setValues for literal dataset). */
  private String addMethod;


  /**
    * Finds the class type of the AST to build (ADLConcatenationSet,
    * ADLLiteralSet or a factory type).
    * <BR>Sets the addMethod field.
    * @param exp the dataset expression
    * @return the name (String) of the class type
    */
  private String findType(ASTTDD_DatasetExpr exp) {

    if (exp.jjtGetNumChildren() > 1) {
      addMethod = "addElement";
      return "ADLConcatenationSet";
    }

    SimpleNode node = (SimpleNode)exp.jjtGetChild(0).jjtGetChild(0);

    if (node instanceof ASTTDD_DatasetLiteral) {
      addMethod = "setValues";
      return "ADLLiteralSet";
    }
    if (node instanceof ASTTDD_FactoryCall) {
      addMethod = "addParameter";
      String TDDClassName = null;
      try {
        TDDClassName = ((String)((SimpleNode)node.jjtGetChild(0))
                        .getProperty("enclosingConstructName"));
      }
      catch(PropertyAccessException e) { }
      return makeQualifiedName("F_",
			       TDDClassName,
			       ((SimpleNode)node.jjtGetChild(0)).getName());
    }
    if (node instanceof ASTTDD_DatasetSingleton) {
      if ((node.jjtGetChild(0) instanceof ASTIntegerLiteral) ||
          (node.jjtGetChild(0) instanceof ASTFloatLiteral) ||
          (node.jjtGetChild(0) instanceof ASTBooleanLiteral) ||
          (node.jjtGetChild(0) instanceof ASTCharacterLiteral) ||
          (node.jjtGetChild(0) instanceof ASTStringLiteral)) {
        addMethod = "setValues";
        return "ADLLiteralSet";
      }
      else { // Name
        addMethod = "addElement";
        return "ADLConcatenationSet";
      }
    }
    // error
    return null;
  }


  /**
    * Single constructor.
    * <BR>Initializes the unit name (that begins with "S_")
    * @param node the dataset declaration node
    * @param imports the "import" declarations handle
    * @param unit source unit name
    */
  public JDatasetBuilder (ASTTDD_DatasetDeclaration node, Vector imports, String unit) {
    super(node, imports, "S_", unit);
    build(node);
  }


  /**
    * The main method. Calls findType and the inherited JBuilder.build, then
    * calls makeConstructor.
    * @see JBuilder#build(String)
    * @param node the root node
    */
  private void build(ASTTDD_DatasetDeclaration node) {

    ASTTDD_DatasetExpr dsExpr = (ASTTDD_DatasetExpr)node.jjtGetChild(1);
    String classType = findType(dsExpr);

    // call to inherited method build(String): see JBuilder
    build(classType);

    ASTConstructorDeclaration constr = 
      makeConstructor(dsExpr);
    JGenericVisitor.parentChild(body, constr, 0);
  } 


  /**
    * Creates the AST constructor of the dataset class.
    * <BR>Initializes a JDatasetExprVisitor and launches it on the passed
    * dataset expression.
    * Example for dataset TEST:
<PRE>
| 1|  public  S_bankTest_TEST(ADLTest tc) {
| 2|     super(tc, "TEST");
| 3|     S_bankTest_DEPOSITS elt_p1 = new S_bankTest_DEPOSITS(tc);
| 4|     ADLConcatenationSet elt_p2 = new ADLConcatenationSet(tc, null);
| 5|     S_bankTest_BANKSIZE elt_p2_p1 = new S_bankTest_BANKSIZE(tc);
| 6|     S_bankTest_DEPOSITS elt_p2_p2 = new S_bankTest_DEPOSITS(tc);
| 7|     elt_p2.addElement(elt_p2_p1);
| 8|     elt_p2.addElement(elt_p2_p2);
| 9|     S_bankTest_INT_VALUES elt_p3 = new S_bankTest_INT_VALUES(tc);
|10|     addParameter(elt_p1);
|11|     addParameter(elt_p2);
|12|     addParameter(elt_p3);
|13|  }
</PRE>
    * The generated class S_bankTest_TEST inherits from F_bankTest_make_bank,
    * that implements the factory definition make_bank (F_bankTest_make_bank
    * inherits from ADLFactorySet).
    * <BR>In this example, lines 3 to 9 are generated by the DatasetExprVisitor
    * run on the dataset expression. The statements super(...) and addParameter
    * (lines 10-12) are generated by makeConstructor.
    * @param dsExpr the dataset expression of the dataset declaration
    * @return the ASTConstructorDeclaration node root of the constructor tree
    */
  private ASTConstructorDeclaration makeConstructor(ASTTDD_DatasetExpr dsExpr) {

    JDatasetExprVisitor dsExprVisitor =
      new JDatasetExprVisitor(strStackInit, "tc", false);
    dsExpr.jjtAccept(dsExprVisitor, null);

    Vector datasetComp = (Vector)dsExprVisitor.getDatasetComponents().peek();
    int nbComp = ((Integer)dsExprVisitor.getNbComponents().peek()).intValue();

    ASTConstructorDeclaration createdNode = new
      ASTConstructorDeclaration(ADLJavaParserTreeConstants.JJTCONSTRUCTORDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    createdNode.setName(className);
    int nbChildren = 0;

    ASTFormalParameters fps = new FormalParametersBuilder
      (new FormalParameterBuilder("ADLTest", "tc").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(createdNode, fps, nbChildren++);


    ASTExplicitConstructorInvocation sup = new
      ASTExplicitConstructorInvocation(ADLJavaParserTreeConstants.JJTEXPLICITCONSTRUCTORINVOCATION);
    sup.setName("super");
    JGenericVisitor.parentChild(createdNode, sup, nbChildren++);

    ASTStringLiteral arg2 = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    arg2.setValue("\"" + featureName + "\"");

    ASTArguments args = new
      ArgumentsBuilder(JGenericVisitor.makeName("tc"))
      .addArgument(arg2).generatedAST();
    JGenericVisitor.parentChild(sup, args, 0);

    for (Enumeration e = datasetComp.elements(); e.hasMoreElements(); ) {
      JGenericVisitor.parentChild(createdNode, (SimpleNode)e.nextElement(), nbChildren++);
    }

    String compName;
    for (int i = 1; i <= nbComp; i++) {
      compName = strStackInit + "_p" + i;

      JGenericVisitor.parentChild(createdNode, 
                                  new StatementBuilder
                                  ( new PrimaryExpressionBuilder
                                    ( JGenericVisitor.makeName(addMethod),
                                      new ArgumentsBuilder
                                      ( JGenericVisitor.makeName(compName))
                                      .generatedAST())
                                    .generatedAST())
                                  .generatedAST(),
                                  nbChildren++);
    }

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: JDatasetBuilder.java,v $
// Revision 1.8  1998/06/12 14:59:53  mallet
// Simplified code due to changes in JDatasetExprVisitor.
//
// Revision 1.7  1998/06/05 14:17:04  mallet
// Changed location of method makeQualifiedName from NameAux to JBuilder.
// Removed class NameAux.
//
// Revision 1.6  1998/03/19 17:27:43  bareau
// Integration of the "use" clause (just a different call to
// NameAux.makeQuailfiedName).
//
// Revision 1.5  1998/02/27 13:07:58  bareau
// Updated copyright notice.
//
// Revision 1.4  1998/02/12 16:03:57  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.3  1998/01/21 16:23:54  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:48  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:01  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:45  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:04:54  adl2ext
// Dataset builder, used by the TDD Transform visitor.
// Creates the AST that corresponds to a dataset definition.
// Initial revision.
//
// ***********************************************************************
