// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JDatasetExprVisitor.java,v $
// $Revision: 1.9 $
// $Date: 1998/06/12 14:57:11 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        JDatasetExprVisitor
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Dataset expression visitor.
  * Generates a vector of statements that correspond to the declaration
  * and initialization of a dataset expression (see examples in JDatasetBuilder
  * and JDirectiveBuilder).
  * <BR>Is used by JDatasetBuilder and by JDirectiveBuilder.
  * <P> For an overview of the algorithms used by this transformation, see the
  * ADL2.0 Translation System Design Specification, Chapter 5 (Code Generation),
  * paragraph 5.2.2.6.
  * @author Cyrille Bareau
  * @see JDatasetBuilder
  * @see JDirectiveBuilder
  */

public class JDatasetExprVisitor extends JGeneric {

  /** A stack of vector of statements. These statements correspond to the
    * declaration and initialization of the components of the current dataset
    * expression (these components are the arguments of a factory call or the
    * operands of a dataset concatenation). */
  private Stack datasetComponents;

  /** Number of components of the current dataset expression. */
  private Stack nbComponents;

  /** A stack of strings used to generate the identifiers for the
    * dataset expression components. */
  private Stack strStack;

  /** String for the testContext. <BR>"tc" from a dataset builder,
    * "this" from a directive builder. */
  private String testContext;

  /** Flag for anonymous dataset expressions. Anonymous dataset expressions
    * are arguments of a test directive and components of a dataset
    * expression. */
  private boolean anonymous;

  private static final boolean debug = false;


  /**
    * Single constructor.
    * @param strStackInit The string that initializes strStack.
    * @param tc The string that initializes testContext.
    */
  public JDatasetExprVisitor (String strStackInit, String tc, boolean anon) {
    super();
    testContext = tc;
    strStack = new Stack();
    strStack.push(strStackInit);
    datasetComponents = new Stack();
    datasetComponents.push(new Vector());
    nbComponents = new Stack();
    nbComponents.push(new Integer(0));
    anonymous = anon;

    if (debug)
      System.out.println("*** new JDatasetExprVisitor: " + anonymous);
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetExpr.
    * <P><UL><LI><B>If</B> the current node has several children,
    * then this is a dataset concatenation: call makeConcatComponents.
    * <LI><B>ELSE</B> just re-transmits the visitor to the child node.</UL>
    * @see #makeConcatComponents
    * @param node the ASTTDD_DatasetExpr node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetExpr node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetExpr (" +
			 node.jjtGetNumChildren() + ")");

    if (node.jjtGetNumChildren() > 1) {
      makeConcatComponents(node, data);
    } else {
      node.jjtGetChild(0).jjtAccept(this, data);
    }
    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetConcatenationExpr.
    * <BR>Just re-transmits the visitor to the child node.
    * @param node the ASTTDD_DatasetConcatenationExpr node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetConcatenationExpr");

    node.jjtGetChild(0).jjtAccept(this, data);
    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetSingleton.
    * <P><B>Algorithm</B><UL>
    * <li>If child is an ASTName:<ul>
    * <li>if it is a dataset:<ul>
    * <li>Visits child and gets dataset qualified name in returned node
    * <li>Creates a creation statement:
<PRE>
|  &lt;dataset qualified name> &lt;getComponentName()> =
|         new &lt;dataset qualified name>(&lt;testContext>);
</PRE>
    * <li>Adds this statement to the vector of components on top of
    *     datasetComponents
    * <li>Increments the number of components
    * </ul><li>else, builds a singleton literal from the result of the
    *      visit of the child
    * </ul><li>else, builds a singleton literal from the child
    * </ul>
    *
    * @see #makeSingletonLiteral
    * @param node the ASTTDD_DatasetSingleton node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetSingleton node, Object data) {
    TypedNode singleton = (TypedNode)node.jjtGetChild(0);
    if (singleton instanceof ASTName) {
      // Fix up ASTName by visiting it
      singleton.jjtAccept(this, data);

      // Can be a dataset or a constant
      boolean isDS;
      try {
	isDS = ((String)singleton.getProperty("astNameKind"))
	  .equals("TDDDataset");
      }
      catch(PropertyAccessException e) { isDS = false; }
      if (! isDS) {
	// that's a constant, build up a singleton literal expression
	makeSingletonLiteral(singleton, data);
      } else {
	// <datasetName> <getComponentName()> = new <datasetName>(<testContext>);
	String datasetName = singleton.getName();
	String componentName = getComponentName();

	ASTAllocationExpression alloc = new
	  ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
	parentChild(alloc, singleton, 0);
	
	ASTArguments args = new ArgumentsBuilder
	  (makeName(testContext)).generatedAST();
	parentChild(alloc, args, 1);
	
	ASTBlockStatement createdNode = new BlockStatementBuilder
	  ( new LocalVariableDeclarationBuilder
	    ( makeType(datasetName),
	      componentName,
	      alloc))
	  .generatedAST();
	
	((Vector)datasetComponents.peek()).addElement(createdNode);
	int nbComp = ((Integer)nbComponents.pop()).intValue();
	nbComponents.push(new Integer(nbComp+1));
      }
    } else {
      // Literal node
      makeSingletonLiteral(singleton, data);
    }
    return null;
  }


  /**
    * Visit pattern for a node of type ASTName.
    * <P><B>Algorithm</B><UL>
    * <li>Gets defining tdd class name
    * <LI><B>IF</B> the name is not a dataset name, then it is a
    *    constant:<ul>
    * <li>Sets name to
    *    <samp>&lt;defining tdd class name>_CST.&lt;constant name></samp>
    * </ul><LI><B>ELSE</B><UL>
    * <LI>Gets the qualified name of the dataset as
    *    <samp>S_&lt;defining tdd class name>_&lt;dataset name></samp>
    * <LI>Sets name to this qualified name
    * </UL></UL>
    * @see JBuilder#makeQualifiedName(String, String, String)
    * @param node the ASTName node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTName node, Object data) {

    if (debug)
      System.out.println("visiting ASTName : " + node.getName());

    String nameKind = "";
    String TDDClassName = null;
    try {
      TDDClassName = (String)node.getProperty("enclosingConstructName");
      nameKind = ((String)node.getProperty("astNameKind"));
    }
    catch(PropertyAccessException e) { }

    if (!nameKind.equals("TDDDataset")) { // singleton literal dataset
      if (TDDClassName != null && !TDDClassName.equals("")) {
	node.setName(TDDClassName + "_CST." + node.getName());
      }
    } else {
      // general case: dataset name, not as root expression
      String datasetName = JBuilder.makeQualifiedName("S_", TDDClassName, 
						      node.getName());

      node.setName(datasetName);
    }

    if (debug)
      System.out.println("\tnew ASTName : " + node.getName());

    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetLiteral.
    * <P><B>Algorithm</B><OL>
    * <LI>Creates an AST for the variable declaration:
<PRE>
|  ADLLiteralSet < &lt;getComponentName()> = new ADLLiteralSet(&lt;testContext>);
</PRE>
    * <LI>Pushes component name
    * <li>Adds created node to vector of components on top of datasetCompoents
    * <li>Visits children
    * </OL>
    * @param node the ASTTDD_DatasetLiteral node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetLiteral node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetLiteral");

    // ADLLiteralSet <getComponentName()> =
    //    new ADLLiteralSet(<testContext>);

    String componentName = getComponentName();
    pushName(componentName);

    ASTAllocationExpression alloc = new
      ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
    parentChild(alloc, makeName("ADLLiteralSet"), 0);

    parentChild(alloc,
                new ArgumentsBuilder(makeName(testContext))
                .generatedAST(),
                1);

    ASTBlockStatement createdNode = new BlockStatementBuilder
      ( new LocalVariableDeclarationBuilder
        ( makeType("ADLLiteralSet"),
          componentName,
          alloc))
      .generatedAST();

    ((Vector)datasetComponents.peek()).addElement(createdNode);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));

    node.childrenAccept(this, null);

    popName();

    return null;
  }
  

  /**
    * Visit pattern for a node of type ASTTDD_DatasetMember.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the current node has only one child: not a range:<ul>
    * <ul>Generate an AST for statement:
<pre>
|  &lt;peekName()>.addValue(&lt;visit child>);
</pre>
    * </ul><LI><B>ELSE</B> <I>(the current node has two children: range)</I>:<ul>
    * <li>Generate ASTs for statements:
<PRE>
|  &lt;getComponentName()>.addRangeValues(&lt;visit child1>, &lt;visit child2>);
</PRE>
    * </UL>
    * <li>Created statement is added to the vector on top of stack
    * datasetComponents.
    * </ul>
    * @param node the ASTTDD_DatasetMember node to visit
    * @param data null <I>or</I> the string that represents the Java wrapper
    * type of the literal type if it is primitive.
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetMember node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetMember");

    ASTStatement stmt;

    if (node.jjtGetNumChildren() == 1) { // simple case: no range
      // <peekName()>.addValue(<node.child1>);
      SimpleNode arg = (SimpleNode)node.jjtGetChild(0);

      // Fix up ASTName children by visiting arg
      arg.jjtAccept(this, data);

      stmt = new StatementBuilder
        ( new PrimaryExpressionBuilder
          ( makeName(peekName() + ".addValue"),
            new ArgumentsBuilder(arg).generatedAST())
          .generatedAST())
        .generatedAST();
    }
    else {
      // <peekName()>.addRangeValues(<node.child1>, <node.child2>);
      SimpleNode arg1 = (SimpleNode)node.jjtGetChild(0);
      SimpleNode arg2 = (SimpleNode)node.jjtGetChild(1);

      // Fix up ASTName children by visiting args
      arg1.jjtAccept(this, data);
      arg2.jjtAccept(this, data);

      stmt = new StatementBuilder
        ( new PrimaryExpressionBuilder
          ( makeName(peekName() + ".addRangeValues"),
            new ArgumentsBuilder(arg1).addArgument(arg2)
            .generatedAST())
          .generatedAST())
        .generatedAST();
    }      

    ((Vector)datasetComponents.peek()).addElement(stmt);

    return null;
  }


  /**
    * Creates the code for a singleton literal dataset.
<PRE>
|  ADLLiteralSet &lt;getComponentName()> = new
|     ADLLiteralSet(&lt;testContext>);    
|  &lt;getComponentName()>.addValue(&lt;literal expression>);
</PRE>
    */
  private void makeSingletonLiteral(TypedNode node, Object data) {

    if (debug)
      System.out.println("makeSingletonLiteral: " + node.getName());

    // ADLLiteralSet <getComponentName()> = new ADLLiteralSet(<testContext>);
    String componentName = getComponentName();
    ASTAllocationExpression alloc = new
      ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
    parentChild(alloc, makeName("ADLLiteralSet"), 0);

    parentChild(alloc,
                new ArgumentsBuilder(makeName(testContext))
                .generatedAST(),
                1);

    ASTBlockStatement createdNode = new BlockStatementBuilder
      ( new LocalVariableDeclarationBuilder
        ( makeType("ADLLiteralSet"),
          componentName,
          alloc))
      .generatedAST();

    ((Vector)datasetComponents.peek()).addElement(createdNode);
    int nbComp = ((Integer)nbComponents.pop()).intValue();
    nbComponents.push(new Integer(nbComp+1));

    // <componentName>.addValue(<node>);
    ASTStatement stmt = new StatementBuilder
      ( new PrimaryExpressionBuilder
        ( makeName(componentName + ".addValue"),
          new ArgumentsBuilder(node).generatedAST())
        .generatedAST())
      .generatedAST();

    ((Vector)datasetComponents.peek()).addElement(stmt);
  }


  /**
    * Create the code for a dataset concatenation expression.
    * <UL><LI><B>IF</B> anonymous expression:<ul>
    * <li>Adds statement to create an ADLConcatenationSet
<PRE>
|  ADLConcatenationSet &lt;getComponentName()> = new
|     ADLConcatenationSet(&lt;testContext>, null);
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushes a new Integer(0) on nbComponents
    * <li>Keeps value of <samp>anonymous</samp> and set it to <samp>true</samp>
    * </ul><li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Copies each of the components in this vector into vector on top
    *     of stack
    * <LI><B>IF</B> anonymous expression,<UL>
    * <LI>add statements to add the child element to the dataset:
<PRE>
|  &lt;peekName()>.addElement(&lt;peekName()>_p&lt;i>);
</PRE>
    * </UL></UL>
    * <li>Restores previous value of <samp>anonymous</samp>
    * <li><B>IF</B> anonymous expression,<UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul></ul>
    * @param node the ASTTDD_DatasetExpr node to visit
    * @param data unused
    */
  private void makeConcatComponents(ASTTDD_DatasetExpr node, Object data) {

    if (debug)
      System.out.println("makeConcatComponents: " + node + " " + node.getName());

    if (anonymous) {
      if (debug)
	System.out.println("makeConcatComponents: anonymous");
      
      // ADLConcatenationSet <getComponentName()> = new ADLConcatenationSet(<testContext>, null);
      String componentName = pushName(getComponentName());

      ASTAllocationExpression alloc = new
	ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
      parentChild(alloc, makeName("ADLConcatenationSet"), 0);
      
      ASTArguments args = new ArgumentsBuilder
	( makeName(testContext))
	.addArgument(makeName("null")).generatedAST();
      parentChild(alloc, args, 1);
      
      ASTBlockStatement createdNode = new BlockStatementBuilder
	( new LocalVariableDeclarationBuilder
	  ( makeType("ADLConcatenationSet"), componentName, alloc)
	  .generatedAST())
	.generatedAST();
    
      ((Vector)datasetComponents.peek()).addElement(createdNode);
      int nbComp = ((Integer)nbComponents.pop()).intValue();
      nbComponents.push(new Integer(nbComp+1));
      nbComponents.push(new Integer(0));
    }

    // visit elements of concatenation
    Node blkStmt;
    ASTPrimaryExpression prim;
    boolean myAnonymous = anonymous;
    
    anonymous = true;
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      if (debug)
	System.out.println("makeConcatComponents: element " + (i + 1));
    
      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, data);

      Vector elemComponents = (Vector)datasetComponents.pop();
      
      // add components resulting from visit of element to the list 
      // of components of the dataset expression
      Vector thetop = (Vector)datasetComponents.peek();
      for (Enumeration e = elemComponents.elements(); 
	   e.hasMoreElements(); ) {
	thetop.addElement((SimpleNode)e.nextElement());
      }
      
      if (myAnonymous) {
	// <peekName()>.addElement(<peekName()>_p<i+1>);
	prim = new PrimaryExpressionBuilder
	  ( makeName(peekName() + ".addElement"),
	    new ArgumentsBuilder
	    ( makeName(peekName() + "_p" + (i + 1))).generatedAST())
	  .generatedAST();
	
	blkStmt = new StatementBuilder(prim).generatedAST();
	((Vector)datasetComponents.peek()).addElement(blkStmt);
      }
    }
    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }
  }


  /**
    * Visit pattern for a node of type ASTTDD_FactoryCall.
    * <UL><LI><B>IF</B> anonymous expression:<ul>
    * <li>Adds statement to create an ADLConcatenationSet
<PRE>
|  &lt;factory name> &lt;getComponentName()> = new
|     &lt;factory name>(&lt;testContext>, null);
</PRE>
    * <li>Pushes current component name
    * <li>Adds new component and increments number of components
    * <li>Pushed a new Integer(0) on nbComponents
    * <li>Keeps value of <samp>anonymous</samp> and set it to <samp>true</samp>
    * </ul><li>for each child:<ul><li>
    * <LI>Pushes an empty vector on datasetComponents
    * <li>Visits child
    * <li>Pops previously pushed vector off datasetComponents
    * <li>Copies each of the components in this vector into vector on top
    *     of stack
    * <LI><B>IF</B> anonymous expression,<UL>
    * <LI>add statements to add the child element to the dataset:
<PRE>
|  &lt;peekName()>.addParameter(&lt;peekName()>_p&lt;i>);
</PRE>
    * </UL></UL>
    * <li>Restores previous value of <samp>anonymous</samp>
    * <li><B>IF</B> anonymous expression,<UL>
    * <li>pops integer off nbComponents
    * <li>pops component name
    * </ul></ul>
    *
    * @param node the ASTTDD_FactoryCall node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_FactoryCall node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_FactoryCall");

    String TDDClassName = null, factName;
    try {
      TDDClassName = (String)((SimpleNode)node.jjtGetChild(0))
                      .getProperty("enclosingConstructName");
    }
    catch(PropertyAccessException e) { }

    factName = 
      JBuilder.makeQualifiedName("F_", TDDClassName,
				 ((SimpleNode)node.jjtGetChild(0)).getName());

    if (anonymous) {
      if (debug)
	System.out.println("ASTTDD_FactoryCall: " + factName +
			   " : anonymous");
      
      // <factName> <getComponentName()> = new <factName>(<testContext>, null);
      String componentName = pushName(getComponentName());
      
      ASTAllocationExpression alloc = new
	ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
      parentChild(alloc, makeName(factName), 0);
      
      ASTArguments args = new ArgumentsBuilder
	( makeName(testContext))
	.addArgument(makeName("null")).generatedAST();
      parentChild(alloc, args, 1);
      
      ASTBlockStatement createdNode = new BlockStatementBuilder
	( new LocalVariableDeclarationBuilder
	  ( makeType(factName), componentName, alloc)
	  .generatedAST())
	.generatedAST();
      
      ((Vector)datasetComponents.peek()).addElement(createdNode);
      int nbComp = ((Integer)nbComponents.pop()).intValue();
      nbComponents.push(new Integer(nbComp+1));
      nbComponents.push(new Integer(0));
    }

    // visit parameters of factory call
    Node blkStmt;
    ASTPrimaryExpression prim;
    int numParameters = node.jjtGetNumChildren() - 1;
    boolean myAnonymous = anonymous;

    anonymous = true;
    for (int i = 1; i <= numParameters; i++) {
      if (debug)
	System.out.println("ASTTDD_FactoryCall: " + factName +
			   " : parameter " + i);
    
      datasetComponents.push(new Vector());

      node.jjtGetChild(i).jjtAccept(this, data);

      Vector paramComponents = (Vector)datasetComponents.pop();

      // add components resulting from visit of parameter to the list 
      // of components of the dataset expression
      Vector thetop = (Vector)datasetComponents.peek();
      for (Enumeration e = paramComponents.elements(); 
	   e.hasMoreElements(); ) {
	thetop.addElement((SimpleNode)e.nextElement());
      }
	
      if (myAnonymous) {
	// <peekName()>.addParameter(<peekName()>_p<i>);
	prim = new PrimaryExpressionBuilder
	  ( makeName(peekName() + ".addParameter"),
	    new ArgumentsBuilder( makeName(peekName() + "_p" + i))
	    .generatedAST())
	  .generatedAST();
	
	blkStmt = new StatementBuilder(prim).generatedAST();
	((Vector)datasetComponents.peek()).addElement(blkStmt);
      }
    }
    anonymous = myAnonymous;

    if (anonymous) {
      nbComponents.pop();
      popName();
    }

    return null;
  }


  /**
    * @return datasetComponents
    */
  public Stack getDatasetComponents() {
    return datasetComponents;
  }


  /**
    * @return nbComponents
    */
  public Stack getNbComponents() {
    return nbComponents;
  }

  private String getComponentName() {
    return (String)strStack.peek() +
      "_p" +
      (((Integer)nbComponents.peek()).intValue() + 1);
  }

  private String pushName(String levelName) {
    strStack.push(levelName);
    return levelName;
  }

  private String popName() {
    return (String)strStack.pop();
  }

  private String peekName() {
    return (String)strStack.peek();
  }
}

// ***********************************************************************
// History Log:
// $Log: JDatasetExprVisitor.java,v $
// Revision 1.9  1998/06/12 14:57:11  mallet
// Major revision of the code to fix reference to "use"d constants:
// 	- removed useless fields,
// 	- simplified code,
// 	- moved processing from visit of ASTName to visit of
// 	  TDD_DatasetSingleton,
// 	- removed visit methods for *Literal,
// 	- added case of "use"d constants in visit of ASTName,
// 	- revised component name management: new methods
// 	  getComponentName(), peekName(), pushName(), popName(),
// 	- updated javadoc comments.
//
// Revision 1.8  1998/06/05 14:17:05  mallet
// Changed location of method makeQualifiedName from NameAux to JBuilder.
// Removed class NameAux.
//
// Revision 1.7  1998/03/19 17:27:43  bareau
// Integration of the "use" clause (just a different call to
// NameAux.makeQuailfiedName).
//
// Revision 1.6  1998/03/02 17:40:25  bareau
// Changed code to take into account the addValue/addRangeValues methods
// introduced in the runtime.
// The "main" method in directive classes uses a dummy tmp variable
// "mainADLTest" to work-around a bug in JET/TET.
//
// Revision 1.5  1998/02/27 13:07:58  bareau
// Updated copyright notice.
//
// Revision 1.4  1998/02/12 16:03:57  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.3  1998/01/21 16:23:54  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:48  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:01  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:45  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:07:53  adl2ext
// Transformer visitor for dataset expressions, used by the
// dataset and the directive builders.
// Initial revision.
//
// ***********************************************************************
