// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JDirectiveBuilder.java,v $
// $Revision: 1.14 $
// $Date: 1998/10/14 08:14:21 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   translators.j
// NAME:        JDirectiveBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Creates a Java compilation unit AST that corresponds to a directive
  * declaration.
  * <BR>Uses JDatasetExprVisitor; is used by JTDDTransformVisitor.
  * <P>Example: definition of directive "mydir" in tdd class bankTest:
  * <BR>"mydir : test (bank b = B_COMBINE,int d = DEPOSITS,int w = WITHDRAWS+DEPOSITS)
  * test_bank (b,d,w);"
  * @author Cyrille Bareau
  * @see JDatasetExprVisitor
  * @see JTDDTransformVisitor
  */

public class JDirectiveBuilder extends JBuilder {

  /** Handle on the TDD single declarators used in domains in the directive
    * parameters */
  private Vector singleDecls;

  /** Names of the ACOs to generate. */
  private Vector acoNames = new Vector();
  

  public Vector getACONames() {
    return acoNames;
  }

  /**
    * Single constructor.
    * <BR>Initializes the unit name (that begins with "D_")
    * @param node the directive declaration node
    * @param imports the "import" declarations handle
    * @param unit source unit name
    */
  public JDirectiveBuilder (ASTTDD_TestDirective node, Vector imports,
                           String unit) {
    super(node, imports, "D_", unit);
    singleDecls = new Vector();
    build(node);
  }


  /**
    * Performs some initializations (call to inherited JBuilder.build) and
    * launches the methods that generate the code.
    * @see JBuilder#build(String)
    */
  private void build(ASTTDD_TestDirective node) {
    build(unitName + "TDD");

    JGenericVisitor.parentChild(body, makeConstructor(node), 0);
    JGenericVisitor.parentChild(body, makeMain(), 1);
    JGenericVisitor.parentChild(body, makeRunInstance(node), 2);
  }


  /**
    * Creates the constructor of the generated directive class.
    * Example for directive mydir:
<PRE>
| 1|  public  D_bankTest_mydir() {
| 2|     S_bankTest_B_COMBINE elt_p1 = new S_bankTest_B_COMBINE(this);
| 3|     S_bankTest_DEPOSITS elt_p2 = new S_bankTest_DEPOSITS(this);
| 4|     ADLConcatenationSet elt_p3 = new ADLConcatenationSet(this, null);
| 5|     S_bankTest_WITHDRAWS elt_p3_p1 = new S_bankTest_WITHDRAWS(this);
| 6|     S_bankTest_DEPOSITS elt_p3_p2 = new S_bankTest_DEPOSITS(this);
| 7|     elt_p3.addElement(elt_p3_p1);
| 8|     elt_p3.addElement(elt_p3_p2);
| 9|     addParameter(elt_p1);
|10|     addParameter(elt_p2);
|11|     addParameter(elt_p3);
|12|  }
</PRE>
    * The transformation of each directive parameter dataset expression is
    * performed by calling a JDatasetExprVisitor (in the example, lines 2
    * to 8).
    * @see JDatasetExprVisitor
    * @param testDir the test directive
    * @return The root of the constructor (ASTConstructorDeclaration).
    */
  private ASTConstructorDeclaration makeConstructor(ASTTDD_TestDirective testDir) {

    ASTConstructorDeclaration createdNode = new
      ASTConstructorDeclaration(ADLJavaParserTreeConstants.JJTCONSTRUCTORDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    createdNode.setName(className);
    
    ASTFormalParameters fps = new FormalParametersBuilder().generatedAST();
    JGenericVisitor.parentChild(createdNode, fps, 0);
    // no parameters

    String strStackInit = "elt";
    ASTTDD_DatasetDomain domain;
    JDatasetExprVisitor exprVisitor =
      new JDatasetExprVisitor(strStackInit, "this", true);
    int nbChildren = 1;

    // visit all test directive parameters to gather the statements 
    // that create them
    for (int i = 0; i < testDir.jjtGetNumChildren()-1; i++) {
      domain = (ASTTDD_DatasetDomain)testDir.jjtGetChild(i);
      if (domain.jjtGetNumChildren() > 1) {
        singleDecls.addElement(domain.jjtGetChild(0));
      }
      else { // void dataset
        singleDecls.addElement(null);
      }

      domain.jjtGetChild(domain.jjtGetNumChildren()-1)
	.jjtAccept(exprVisitor, null);
    }

    // now attach all the statements that create the directive
    // parameters (components of the DatasetExprVisitor) to the current
    // block
    for (Enumeration e = ((Vector)exprVisitor.getDatasetComponents().peek()).elements();
	 e.hasMoreElements(); ) {
      JGenericVisitor.parentChild(createdNode,
				  (SimpleNode)e.nextElement(),
				  nbChildren++);
    }
      
    // and also add the "addParameter" statements for the directive
    // parameters
    for (int i = 0; i < testDir.jjtGetNumChildren()-1; i++) {
      ASTArguments pArgs = new ArgumentsBuilder
        ( JGenericVisitor.makeName(strStackInit + "_p" + (i+1)) )
        .generatedAST();
      ASTPrimaryExpression prim = new PrimaryExpressionBuilder
        ( JGenericVisitor.makeName("addParameter"), pArgs )
        .generatedAST();
      ASTStatement blk = new StatementBuilder(prim).generatedAST();
      JGenericVisitor.parentChild(createdNode, blk, nbChildren++);
    }
    return createdNode;
  }


  /**
    * Creates the "main" method of the generated directive class.
    * Example for directive mydir:
<PRE>
|  static public  void main(String[] args) {
|     mainADLTest = new D_bankTest_mydir();
|     testCase.main("mydir", args, mainADLTest);
|  }
</PRE>
    * @return The root of the constructor (ASTConstructorDeclaration).
    */
  private ASTMethodDeclaration makeMain() {

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PUBLIC | JModifier.STATIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}

    JGenericVisitor.parentChild(createdNode, JGenericVisitor.makeType("void"), 0);

    ASTMethodDeclarator methDecl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    methDecl.setName("main");
    JGenericVisitor.parentChild(createdNode, methDecl, 1);

    ASTType objArr = JGenericVisitor.makeType("String");
    try {
      objArr.setProperty("arrayLevel", new Integer(1));
    }
    catch(PropertyAccessException e) {}

    ASTFormalParameters fp = new FormalParametersBuilder
      ( new FormalParameterBuilder(objArr, "args").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(methDecl, fp, 0);

    ASTBlock body = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    JGenericVisitor.parentChild(createdNode, body, 2);

    // mainADLTest = new <className>();

    ASTAllocationExpression alloc = new
      ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
    JGenericVisitor.parentChild
      (alloc, JGenericVisitor.makeName(className), 0);
    JGenericVisitor.parentChild
      (alloc, new ArgumentsBuilder().generatedAST(), 1);

    JGenericVisitor.parentChild(body,
                                new StatementBuilder
                                ( new AssignmentBuilder
                                  ( "mainADLTest", "=", alloc).generatedAST())
                                .generatedAST(),
                                0);

    // testCase.main("<featureName>", args, mainADLTest);

    ASTStringLiteral featName = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    featName.setValue("\"" + featureName + "\"");

    ASTArguments args = new ArgumentsBuilder(featName)
      .addArgument(JGenericVisitor.makeName("args"))
      .addArgument(JGenericVisitor.makeName("mainADLTest"))
      .generatedAST();

    JGenericVisitor.parentChild(body,
                                new StatementBuilder
                                ( new PrimaryExpressionBuilder
                                  ( JGenericVisitor.makeName("testCase.main"),
                                    args))
                                .generatedAST(),
                                1);
                                  
    return createdNode;
  }

   
  /**
    * Creates the "runInstance" method of the generated directive class.
    * Example for directive mydir:
<PRE>
|  protected  void runInstance(Object[] paramValues) throws Throwable {
|     bank b = (bank)paramValues[0];
|     infoline("Parameter bank b = " + b);
|     int d = ((Integer)paramValues[1]).intValue();
|     infoline("Parameter int d = " + d);
|     int w = ((Integer)paramValues[2]).intValue();
|     infoline("Parameter int w = " + w);
|     &#60;directive body&362;
|  }
</PRE>
    * <B>Note:</B> the directive body is visited by an RemoveADLVisitor
    * to transform the ADL/ADL_new expressions it contains.
    * @see RemoveADLVisitor
    * @return The root of the constructor (ASTConstructorDeclaration).
    */
  private ASTMethodDeclaration makeRunInstance(ASTTDD_TestDirective testDir) {

    // protected void runInstance(Object[] paramValues)

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PROTECTED;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}

    JGenericVisitor.parentChild(createdNode, JGenericVisitor.makeType("void"), 0);

    ASTMethodDeclarator decl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    decl.setName("runInstance");
    JGenericVisitor.parentChild(createdNode, decl, 1);

    ASTType t = JGenericVisitor.makeType("Object");
    try {
      t.setProperty("arrayLevel", new Integer(1));
    }
    catch(PropertyAccessException e) {}

    ASTFormalParameters params = new FormalParametersBuilder
      ( new FormalParameterBuilder(t, "paramValues").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(decl, params, 0);

    // throws Throwable
    ASTNameList thrl = new ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(thrl, JGenericVisitor.makeName("Throwable"), 0);    
    JGenericVisitor.parentChild(createdNode, thrl, 2);

    ASTBlock blk = new
      ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    JGenericVisitor.parentChild(createdNode, blk, 3);

    int nbChildren = 0; int dummy = 1;
    ASTTDD_SingleDeclarator single; Node blkStmt;
    ASTType typ1; String id; ASTCastExpression castExpr;
    ASTBracketSuffix suffix; ASTAdditiveExpression add;
    ASTPrimaryExpression prim; ASTArguments args;
    ASTStringLiteral str; ASTIntegerLiteral index;

    for (int indexSD = 0; indexSD < singleDecls.size(); indexSD++) {
      single = (ASTTDD_SingleDeclarator)singleDecls.elementAt(indexSD);

      if (single != null) { // not void dataset parameter
        typ1 = TypeAux.resultType2Type((ASTResultType)single.jjtGetChild(0));
        id = ((SimpleNode)single.jjtGetChild(1)).getName();

        if (TypeAux.isPrimitiveType(typ1)) {
          // <typ1> <id> = ((Number)paramValues[<i>]).<typ1>Value();
          castExpr = new
            ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
          JGenericVisitor.parentChild(castExpr,
//                                       JGenericVisitor.makeType(TypeAux.getJavaTypeName(typ1)), 0);
                                      JGenericVisitor.makeType("Number"), 0);
          suffix = new
            ASTBracketSuffix(ADLJavaParserTreeConstants.JJTBRACKETSUFFIX);
          index = new
            ASTIntegerLiteral(ADLJavaParserTreeConstants.JJTINTEGERLITERAL);
          index.setValue(indexSD);
          JGenericVisitor.parentChild(suffix, index, 0);

          prim = new PrimaryExpressionBuilder
            ( JGenericVisitor.makeName("paramValues"), suffix)
            .generatedAST();
          JGenericVisitor.parentChild(castExpr, prim, 1);

          ASTParenthExpression parenth = new
            ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
          JGenericVisitor.parentChild(parenth, castExpr, 0);

          ASTDotSuffix dot = new
            ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
          dot.setName(typ1.getName() + "Value");

          ASTPrimaryExpression prim2 = new PrimaryExpressionBuilder
            (parenth, dot)
            .addSuffix(new ArgumentsBuilder().generatedAST())
            .generatedAST();

          JGenericVisitor.parentChild(blk,
                                      new BlockStatementBuilder
                                      ( new LocalVariableDeclarationBuilder
                                        ( typ1, id, prim2 )).generatedAST(),
                                      nbChildren++);
        }
        else {
          // <typ1> <id> = (<typ1>)paramValues[<i>];
          castExpr = new
            ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
          JGenericVisitor.parentChild(castExpr,
                                      JGenericVisitor.makeName(typ1.getName()), 0);

          // primary expression
          suffix = new
            ASTBracketSuffix(ADLJavaParserTreeConstants.JJTBRACKETSUFFIX);
          index = new
            ASTIntegerLiteral(ADLJavaParserTreeConstants.JJTINTEGERLITERAL);
          index.setValue(indexSD);
          JGenericVisitor.parentChild(suffix, index, 0);

          prim = new PrimaryExpressionBuilder
            ( JGenericVisitor.makeName("paramValues"), suffix)
            .generatedAST();
          JGenericVisitor.parentChild(castExpr, prim, 1);

          JGenericVisitor.parentChild(blk,
                                      new BlockStatementBuilder
                                      ( new LocalVariableDeclarationBuilder(typ1, id, castExpr))
                                      .generatedAST(),
                                      nbChildren++);
        }

        // infoline (primary expression)
        add = new
          ASTAdditiveExpression(ADLJavaParserTreeConstants.JJTADDITIVEEXPRESSION);
        String[] ops = { "+" };
        add.setOps(ops);

        str = new
          ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
        str.setValue("\"Parameter " + typ1.getName() + " " + id + " = \"");
        JGenericVisitor.parentChild(add, str, 0);
        JGenericVisitor.parentChild(add, JGenericVisitor.makeName(id), 1);

        args = new ArgumentsBuilder(add).generatedAST();

        prim = new PrimaryExpressionBuilder
          ( JGenericVisitor.makeName("infoline"), args)
          .generatedAST();

        blkStmt = new StatementBuilder(prim).generatedAST();
        JGenericVisitor.parentChild(blk, blkStmt, nbChildren++);
      }
    }

    // user code:

    // visit the user code to change ADL/ADL_new expressions
    RemoveADLVisitor v = new RemoveADLVisitor();
    Node userStmt = (Node)testDir.jjtGetChild(testDir.jjtGetNumChildren()-1)
                                 .jjtAccept(v, null);

    // get back list of ACO names
    Enumeration e = v.getACONames().elements();
    while ( e.hasMoreElements() ) {
      String acoName = (String)e.nextElement();
      if ( ! acoNames.contains(acoName) ) {
	acoNames.addElement(acoName);
      }
    }

    if (userStmt instanceof ASTBlock) {
      for (int i = 0; i < userStmt.jjtGetNumChildren(); i++) {
        JGenericVisitor.parentChild(blk, userStmt.jjtGetChild(i), nbChildren++);
      }
    }
    else {
      JGenericVisitor.parentChild(blk, userStmt, nbChildren);
    }

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: JDirectiveBuilder.java,v $
// Revision 1.14  1998/10/14 08:14:21  mallet
// Bugs fixed: pr/155, pr/156: removed useless index of anonymous
// directive as argument to constructor.
//
// Revision 1.13  1998/06/18 10:30:37  mallet
// Fixed incorrect use of type short name instead of full one (FQN).
//
// Revision 1.12  1998/06/12 15:05:57  mallet
// Updated code due to changes in JDatasetExprVisitor: only one instance
// of this class is created for all parameters of a directive, the
// list of components (statements) generated for the parameters is also
// recovered once for all parameters, and these components are "added" at
// the end of the constructor for the directive.
//
// Revision 1.11  1998/06/05 08:36:35  mallet
// Added a new vector "acoNames" and accessor method to store the list of
// ACO used in a test directive (call to ADL(...) or ADL_new ...).
// Added code to get the list of used ACO from RemoveADLVisitor.
//
// Revision 1.10  1998/03/02 17:40:26  bareau
// Changed code to take into account the addValue/addRangeValues methods
// introduced in the runtime.
// The "main" method in directive classes uses a dummy tmp variable
// "mainADLTest" to work-around a bug in JET/TET.
//
// Revision 1.9  1998/02/27 13:07:59  bareau
// Updated copyright notice.
//
// Revision 1.8  1998/02/17 10:57:21  mallet
// Merged v1.0 with development trunk
//
// Revision 1.7.2.1  1998/02/16 21:06:26  bareau
// "Number" used only for casting, not for creation.
//
// Revision 1.7  1998/02/12 16:03:58  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.6  1998/02/03 14:35:04  bareau
// Bug fix for void datasets: generate a dummy SingleDeclarator.
//
// Revision 1.5  1998/02/02 13:49:12  bareau
// Changed the class "ADLremove" into "RemoveADL".
//
// Revision 1.4  1998/01/26 15:12:45  bareau
// Minor bug fixes.
//
// Revision 1.3  1998/01/21 16:23:55  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:48  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:01  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:45  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:05:49  adl2ext
// Directive builder, used by the TDD Transform visitor.
// Creates the AST that corresponds to a directive definition.
// Initial revision.
//
// ***********************************************************************
