// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JFactoryBuilder.java,v $
// $Revision: 1.7 $
// $Date: 1998/02/27 13:07:59 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        JFactoryBuilder
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import java.util.Vector;


/**
  * Class that creates the Java compilation unit AST that corresponds to a
  * factory declaration.
  * <BR>Is used by JTDDTransformVisitor.
  * <P>Example: factory "make_bank" defined in tdd class "bankTest":
<PRE>
  factory bank make_bank (int maxAccts, int numAcctsActive, int initDeposit) {
    bank ret = new bank(maxAccts);
    for (int x=0;x < maxAccts && x < numAcctsActive;x++) {
      try { bankAccount ba = ret.open_account(initDeposit); }
      catch (bankException be) { tdd_end_case(); }
    }
    return ret;
  }
  // optional relinquish clause
  relinquish (bank b) {
    if (numAcctsActive > 0) {
      try { b.close_all_accounts(); }
      catch(Throwable e) {}
    }
  }
</PRE>
  * @author Cyrille Bareau
  * @see JTDDTransformVisitor
  */

public class JFactoryBuilder extends JBuilder {

  /** The factory declaration */
  private ASTMethodDeclaration factDecl;

  /** The return type of the factory */
  private ASTResultType factType;

  /** The name of the factory */
  private String factName;

  /** The formal parameters of the factory */
  private ASTFormalParameters factParams;

  /** The "throws list" of the factory */
  private ASTNameList factThrowsList;

  /** The body of the factory */
  private ASTBlock factBlk;

  /** The body of the (optional) relinquish block */
  private ASTBlock relBlock = null;

  /** The parameter of the relinquish block (for non void factories) */
  private ASTFormalParameter relParam = null;

  /** Flag "the factory is void: no returned object" */
  private boolean voidFactory;

  /** Flag "no relinquish clause in the factory declaration */
  private boolean noRelinquish = true;

  static final private boolean debug = false;


  /**
    * Single constructor.
    * @param node The factory declaration AST node
    * @param imports The import declarations
    * @param unit The name of the adlclass file
    */
  public JFactoryBuilder(ASTTDD_FactoryDeclaration node, Vector imports, String unit) {

    super(node, imports, "F_", unit);
    build(node);
  }


  /**
    * Initializes the data and launches the methods code generation.
    */
  private void build(ASTTDD_FactoryDeclaration node) {

    build("ADLFactorySet");

    factDecl = (ASTMethodDeclaration)node.jjtGetChild(0);
    factType = (ASTResultType)factDecl.jjtGetChild(0);
    //    factType = JGenericVisitor.makeType(((SimpleNode)factDecl.jjtGetChild(0)).getName());
    factName = ((SimpleNode)factDecl.jjtGetChild(1)).getName();
    voidFactory = TypeAux.isVoid(factType);
    factParams = (ASTFormalParameters)factDecl.jjtGetChild(1).jjtGetChild(0);
    if (node.jjtGetNumChildren() > 1) {
      noRelinquish = false;
      if (node.jjtGetNumChildren() > 2) {
        relParam = (ASTFormalParameter)node.jjtGetChild(1);
        relBlock = (ASTBlock)node.jjtGetChild(2);
      }
      else {
        relBlock = (ASTBlock)node.jjtGetChild(1);
      }
    }
    
    if (factDecl.jjtGetNumChildren() == 2) {
      // error
    }
    else {
      if (factDecl.jjtGetChild(2) instanceof ASTNameList) {
        factThrowsList = (ASTNameList)factDecl.jjtGetChild(2);
      }
      factBlk = (ASTBlock)factDecl.jjtGetChild(factDecl.jjtGetNumChildren()-1);
    }
   
    JGenericVisitor.parentChild(body, makeConstructor(), 0);

    JGenericVisitor.parentChild(body, makeFactoryProvide(), 1);

    JGenericVisitor.parentChild(body, makeProvide(), 2);

    if (! noRelinquish) {
      JGenericVisitor.parentChild(body, makeFactoryRelinquish(), 3);

      JGenericVisitor.parentChild(body, makeRelinquish(), 4);
    }
  }


  /**
    * Creates the factory constructor.
    * Example for factory make_bank:
<PRE>
|  public  F_bankTest_make_bank(ADLTest tc, String label) {
|     super(tc, label);
|     factoryName = "make_bank";
|  }
</PRE>
    * @return the root of the constructor (ASTConstructorDeclaration).
    */
  private ASTConstructorDeclaration makeConstructor() {

    ASTConstructorDeclaration createdNode = new
      ASTConstructorDeclaration(ADLJavaParserTreeConstants.JJTCONSTRUCTORDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    createdNode.setName(className);

    ASTFormalParameters fps = new FormalParametersBuilder
      (new FormalParameterBuilder("ADLTest", "tc").generatedAST())
      .addParameter(new FormalParameterBuilder("String", "label").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(createdNode, fps, 0);

    ASTExplicitConstructorInvocation sup = new
      ASTExplicitConstructorInvocation(ADLJavaParserTreeConstants.JJTEXPLICITCONSTRUCTORINVOCATION);
    sup.setName("super");
    JGenericVisitor.parentChild(createdNode, sup, 1);

    ASTArguments args = new
      ArgumentsBuilder(JGenericVisitor.makeName("tc"))
      .addArgument(JGenericVisitor.makeName("label"))
      .generatedAST();
    JGenericVisitor.parentChild(sup, args, 0);


    // factoryName = "<factName>";
    ASTStringLiteral factn = new
      ASTStringLiteral(ADLJavaParserTreeConstants.JJTSTRINGLITERAL);
    factn.setValue("\"" + factName + "\"");
    JGenericVisitor.parentChild(createdNode,
                                new StatementBuilder
                                ( new AssignmentBuilder
                                  ( "factoryName", "=", factn)) // $$$$
                                .generatedAST(),
                                2);

    return createdNode;
  }


  /**
    * Creates the method "factory_provide".
    * Example for factory make_bank:
<PRE>
|  public  Object factory_provide(Object[] p) throws Throwable {
|     int elt0 = ((Integer)p[0]).intValue();
|     int elt1 = ((Integer)p[1]).intValue();
|     int elt2 = ((Integer)p[2]).intValue();
|     return provide_make_bank(elt0, elt1, elt2);
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTMethodDeclaration makeFactoryProvide() {

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    JGenericVisitor.parentChild(createdNode,
                                JGenericVisitor.makeType("Object"), 0);

    ASTMethodDeclarator methDecl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    methDecl.setName("factory_provide");
    JGenericVisitor.parentChild(createdNode, methDecl, 1);

    ASTType objArr = JGenericVisitor.makeType("Object");
    try {
      objArr.setProperty("arrayLevel", new Integer(1));
    }
    catch(PropertyAccessException e) {}

    ASTFormalParameters fp = new FormalParametersBuilder
      ( new FormalParameterBuilder(objArr, "p").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(methDecl, fp, 0);

    ASTNameList thrl = new ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(thrl, JGenericVisitor.makeName("Throwable"), 0);
    JGenericVisitor.parentChild(createdNode, thrl, 2);

    ASTBlock body = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    JGenericVisitor.parentChild(createdNode, body, 3);

    ASTType ptype; String pname; ASTBracketSuffix brack;
    ASTCastExpression castExpr; ASTPrimaryExpression prim, prim2;
    ASTParenthExpression parenth;
    ASTIntegerLiteral arrlit; ASTDotSuffix dot;

    int nbp = factParams.jjtGetNumChildren();
    for (int i = 0; i < nbp; i++) {
      ptype = (ASTType)factParams.jjtGetChild(i).jjtGetChild(0);
      pname = "elt" + i;
      arrlit = new
        ASTIntegerLiteral(ADLJavaParserTreeConstants.JJTINTEGERLITERAL);
      arrlit.setValue(i);
      brack = new
        ASTBracketSuffix(ADLJavaParserTreeConstants.JJTBRACKETSUFFIX);
      JGenericVisitor.parentChild(brack, arrlit, 0);
      prim = new PrimaryExpressionBuilder
        ( JGenericVisitor.makeName("p"),
          brack).generatedAST();

      if (TypeAux.isPrimitiveType(ptype)) {
        // <primtype> elt<i> = ((Number)p[<i>]).<primtype>Value();
        // e.g.  int elt<i> = ((Number)p[<i>]).intValue();
        castExpr = new
          ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
        JGenericVisitor.parentChild(castExpr,
//                                     JGenericVisitor.makeType(TypeAux.getJavaTypeName(ptype)), 0);
                                    JGenericVisitor.makeType("Number"), 0);
        JGenericVisitor.parentChild(castExpr, prim, 1);

        parenth = new
          ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
        JGenericVisitor.parentChild(parenth, castExpr, 0);

        dot = new
          ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
        dot.setName(ptype.getName() + "Value");

        prim2 = new PrimaryExpressionBuilder
          (parenth, dot)
          .addSuffix(new ArgumentsBuilder().generatedAST())
          .generatedAST();

        JGenericVisitor.parentChild(body,
                                    new BlockStatementBuilder
                                    ( new LocalVariableDeclarationBuilder
                                      ( ptype, pname, prim2 )).generatedAST(),
                                    i);
      }
      else { // object type
        // <type> elt<i> = (<type>)p[<i>];
        castExpr = new
          ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
        JGenericVisitor.parentChild(castExpr, ptype, 0);
        JGenericVisitor.parentChild(castExpr, prim, 1);
        JGenericVisitor.parentChild(body,
                                    new BlockStatementBuilder
                                    ( new LocalVariableDeclarationBuilder
                                      ( ptype, pname, castExpr )).generatedAST(),
                                    i);
      }
    }

    // return provide_<factName>(elt<0>, elt<1>, ...);
    // or, if void factory:
    // provide_<factName>(elt<0>, elt<1>, ...); return new Object();

    ArgumentsBuilder argb = new ArgumentsBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(JGenericVisitor.makeName("elt" + i));
    }
    ASTPrimaryExpression last = new PrimaryExpressionBuilder
      ( JGenericVisitor.makeName("provide_" + factName),
        argb.generatedAST())
      .generatedAST();

    if (voidFactory) {
      JGenericVisitor.parentChild
        (body, new StatementBuilder(last).generatedAST(), nbp);

      ASTAllocationExpression newObj = new
        ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);
      JGenericVisitor.parentChild(newObj, JGenericVisitor.makeName("Object"), 0);
      JGenericVisitor.parentChild(newObj, new ArgumentsBuilder().generatedAST(), 1);
      ASTReturnStatement ret = new
        ASTReturnStatement(ADLJavaParserTreeConstants.JJTRETURNSTATEMENT);
      JGenericVisitor.parentChild(ret, newObj, 0);
      JGenericVisitor.parentChild(body, ret, nbp+1);
    }
    else {
      ASTReturnStatement ret = new
        ASTReturnStatement(ADLJavaParserTreeConstants.JJTRETURNSTATEMENT);
      JGenericVisitor.parentChild(ret, last, 0);
      JGenericVisitor.parentChild(body, ret, nbp);
    }

    return createdNode;
  }


  /**
    * Creates the method "provide_&#60;factory name&#62;".
    * Example for factory make_bank:
<PRE>
|  public bank provide_make_bank(int maxAccts, int numAcctsActive, int initDeposit)
|     throws Throwable {
|     &#60;factory body&#62;
|  }
</PRE>
    * <B>Note:</B> if the factory returns a primitive type, one must create a
    * FactoryVisitor to replace returned values in the factory body by values
    * of the corresponding Java wrapper class.
    * @see FactoryVisitor
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTMethodDeclaration makeProvide () {

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}

    boolean primitiveType = TypeAux.isPrimitiveType(factType);
    if (primitiveType) {
      JGenericVisitor.parentChild(createdNode,
                                  JGenericVisitor.makeType(TypeAux.getJavaTypeName(factType)),
                                  0);
    }
    else {
      JGenericVisitor.parentChild(createdNode, factType, 0);
    }

    ASTMethodDeclarator methDecl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    methDecl.setName("provide_" + factName);
    JGenericVisitor.parentChild(createdNode, methDecl, 1);

    JGenericVisitor.parentChild(methDecl, factParams, 0);

    ASTNameList thrl = new ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(thrl, JGenericVisitor.makeName("Throwable"), 0);
    JGenericVisitor.parentChild(createdNode, thrl, 2);

    if (primitiveType) {
      // visit factory body to modify return statements
      FactoryVisitor factV = new FactoryVisitor();
      Node visited = (Node)factBlk.jjtAccept(factV,
                                             TypeAux.getJavaTypeName(factType));
      JGenericVisitor.parentChild(createdNode, visited, 3);
    }
    else { // nothing to modify in the factory body
      JGenericVisitor.parentChild(createdNode, factBlk, 3);
    }

    return createdNode;
  }


  /**
    * Creates the method "factory_relinquish".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  public  boolean factory_relinquish(Object[] p, Object obj) {
|     int elt0 = ((Integer)p[0]).intValue();
|     int elt1 = ((Integer)p[1]).intValue();
|     int elt2 = ((Integer)p[2]).intValue();
|     bank theObj = (bank)obj;
|     return relinquish_make_bank(elt0, elt1, elt2, theObj);
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTMethodDeclaration makeFactoryRelinquish() {

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    JGenericVisitor.parentChild(createdNode,
                                JGenericVisitor.makeType("boolean"), 0);

    ASTMethodDeclarator methDecl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    methDecl.setName("factory_relinquish");
    JGenericVisitor.parentChild(createdNode, methDecl, 1);

    ASTType objArr = JGenericVisitor.makeType("Object");
    try {
      objArr.setProperty("arrayLevel", new Integer(1));
    }
    catch(PropertyAccessException e) {}

    ASTFormalParameters fp = new FormalParametersBuilder
      ( new FormalParameterBuilder(objArr, "p").generatedAST())
      .addParameter( new FormalParameterBuilder("Object", "obj").generatedAST())
      .generatedAST();
    JGenericVisitor.parentChild(methDecl, fp, 0);

    ASTNameList thrl = new ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(thrl, JGenericVisitor.makeName("Throwable"), 0);
    JGenericVisitor.parentChild(methDecl, thrl, 1);

    ASTBlock body = new ASTBlock(ADLJavaParserTreeConstants.JJTBLOCK);
    JGenericVisitor.parentChild(createdNode, body, 2);

    ASTType ptype; String pname; ASTBracketSuffix arr; ASTDotSuffix dot;
    ASTCastExpression castExpr; ASTPrimaryExpression prim, prim2;
    ASTIntegerLiteral arrlit; ASTParenthExpression parenth;

    int nbp = factParams.jjtGetNumChildren();
    for (int i = 0; i < nbp; i++) {
      ptype = (ASTType)factParams.jjtGetChild(i).jjtGetChild(0);
      pname = "elt" + i;
      arrlit = new
        ASTIntegerLiteral(ADLJavaParserTreeConstants.JJTINTEGERLITERAL);
      arrlit.setValue(i);
      arr = new
        ASTBracketSuffix(ADLJavaParserTreeConstants.JJTBRACKETSUFFIX);
      JGenericVisitor.parentChild(arr, arrlit, 0);
      prim = new PrimaryExpressionBuilder
        ( JGenericVisitor.makeName("p"),
          arr).generatedAST();

      if (TypeAux.isPrimitiveType(ptype)) {
        // <primtype> elt<i> = ((Number)p[<i>]).<primtype>Value();
        // e.g.  int elt<i> = ((Number)p[<i>]).intValue();
        castExpr = new
          ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
//         JGenericVisitor.parentChild(castExpr, JGenericVisitor.makeName(TypeAux.getJavaTypeName(ptype)), 0);
        JGenericVisitor.parentChild(castExpr, JGenericVisitor.makeName("Number"), 0);
        JGenericVisitor.parentChild(castExpr, prim, 1);
        parenth = new
          ASTParenthExpression(ADLJavaParserTreeConstants.JJTPARENTHEXPRESSION);
        JGenericVisitor.parentChild(parenth, castExpr, 0);

        dot = new
          ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
        dot.setName(ptype.getName() + "Value");

        prim2 = new PrimaryExpressionBuilder(parenth, dot)
          .addSuffix(new ArgumentsBuilder().generatedAST())
          .generatedAST();
        JGenericVisitor.parentChild(body,
                                    new BlockStatementBuilder
                                    ( new LocalVariableDeclarationBuilder
                                      ( ptype, pname, prim2 )).generatedAST(),
                                    i);
      }
      else { // object type
        // <type> elt<i> = (<type>)p[<i>];
        castExpr = new
          ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
        JGenericVisitor.parentChild(castExpr, ptype, 0);
        JGenericVisitor.parentChild(castExpr, prim, 1);
        JGenericVisitor.parentChild(body,
                                    new BlockStatementBuilder
                                    ( new LocalVariableDeclarationBuilder
                                      ( ptype, pname, castExpr )).generatedAST(),
                                    i);
      }
    }

    // <factType> theObj = (<factType>)obj;

    if (! voidFactory) {
      castExpr = new
        ASTCastExpression(ADLJavaParserTreeConstants.JJTCASTEXPRESSION);
      JGenericVisitor.parentChild(castExpr, factType, 0);
      JGenericVisitor.parentChild(castExpr, JGenericVisitor.makeName("obj"), 1);
      JGenericVisitor.parentChild(body,
                                  new BlockStatementBuilder
                                  ( new LocalVariableDeclarationBuilder
                                    ( TypeAux.resultType2Type(factType),
                                      "theObj",
                                      castExpr ))
                                  .generatedAST(),
                                  nbp);
    }


    // return relinquish_<factName>(elt<0>, elt<1>, ..., theObj);
    // (or (elt<0>, ..., null) if void factory

    ArgumentsBuilder argb = new ArgumentsBuilder();
    for (int i = 0; i < nbp; i++) {
      argb.addArgument(JGenericVisitor.makeName("elt" + i));
    }
    if (voidFactory) {
      argb.addArgument(new ASTNullLiteral(ADLJavaParserTreeConstants.JJTNULLLITERAL));
    }
    else {
      argb.addArgument(JGenericVisitor.makeName("theObj"));
    }
    ASTPrimaryExpression last = new PrimaryExpressionBuilder
      ( JGenericVisitor.makeName("relinquish_" + factName),
        argb.generatedAST())
      .generatedAST();
    ASTReturnStatement ret = new
      ASTReturnStatement(ADLJavaParserTreeConstants.JJTRETURNSTATEMENT);
    JGenericVisitor.parentChild(ret, last, 0);
    JGenericVisitor.parentChild(body, ret, nbp+1);

    return createdNode;
  }


  /**
    * Creates the method "relinquish_&#60;factory name&#62;".
    * Example for factory make_bank of tdd class bankTest:
<PRE>
|  public  boolean relinquish_make_bank(int maxAccts, int numAcctsActive,
|                                       int initDeposit, bank b) {
|     &#60;relinquish body&#62;
|     return true;
|  }
</PRE>
    * @return the root of the method (ASTMethodDeclaration).
    */
  private ASTMethodDeclaration makeRelinquish() {

    ASTMethodDeclaration createdNode = new
      ASTMethodDeclaration(ADLJavaParserTreeConstants.JJTMETHODDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    JGenericVisitor.parentChild(createdNode, JGenericVisitor.makeType("boolean"), 0);

    ASTMethodDeclarator methDecl = new
      ASTMethodDeclarator(ADLJavaParserTreeConstants.JJTMETHODDECLARATOR);
    methDecl.setName("relinquish_" + factName);
    JGenericVisitor.parentChild(createdNode, methDecl, 1);

    FormalParametersBuilder relParamS = new FormalParametersBuilder();
    for (int i = 0; i < factParams.jjtGetNumChildren(); i++) {
      relParamS.addParameter((ASTFormalParameter)factParams.jjtGetChild(i));
    }
    if (voidFactory) {
      relParamS.addParameter(new FormalParameterBuilder("Object", "obj")
                             .generatedAST());
    }
    else {
      relParamS.addParameter(relParam);
    }

    JGenericVisitor.parentChild(methDecl, relParamS.generatedAST(), 0);

    ASTNameList thrl = new ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    JGenericVisitor.parentChild(thrl, JGenericVisitor.makeName("Throwable"), 0);
    JGenericVisitor.parentChild(methDecl, thrl, 1);

    JGenericVisitor.parentChild(createdNode, relBlock, 2);

    // return true;
    ASTBooleanLiteral trueLit = new
      ASTBooleanLiteral(ADLJavaParserTreeConstants.JJTBOOLEANLITERAL);
    trueLit.setValue(true);
    ASTReturnStatement retTrue = new
      ASTReturnStatement(ADLJavaParserTreeConstants.JJTRETURNSTATEMENT);
    JGenericVisitor.parentChild(retTrue, trueLit, 0);
    JGenericVisitor.parentChild(relBlock, retTrue, relBlock.jjtGetNumChildren());

    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: JFactoryBuilder.java,v $
// Revision 1.7  1998/02/27 13:07:59  bareau
// Updated copyright notice.
//
// Revision 1.6  1998/02/24 18:33:37  bareau
// Changed the root node of generated trees as ASTCompilationUnit,
// with package "adltest".
//
// Revision 1.5  1998/02/17 10:57:22  mallet
// Merged v1.0 with development trunk
//
// Revision 1.4.2.1  1998/02/16 21:06:27  bareau
// "Number" used only for casting, not for creation.
//
// Revision 1.4  1998/02/12 16:03:59  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.3  1998/01/21 16:23:55  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:49  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:02  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:46  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:06:28  adl2ext
// Factory builder, used by the TDD Transform visitor.
// Creates the AST that corresponds to a factory definition.
// Initial revision.
//
// ***********************************************************************
