// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/JTDDTransformVisitor.java,v $
// $Revision: 1.17 $
// $Date: 1998/10/14 08:17:26 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        JTDDTransformVisitor
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Nov 26, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.PropertyAccessException;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.j.JModifier;
import java.util.Vector;
import java.util.Enumeration;

/**
  * Transformation visitor.  Must change a TDD AST into Java ASTs.
  * <P> For an overview of the algorithms used by this transformation,
  * see the ADL2.0 Translation System Design Specification, Chapter 5
  * (Code Generation), paragraph 5.2.2.
  * @see JGenericVisitor
  * @see ADLJavaParserVisitor
  * @author Cyrille Bareau
 */

public class JTDDTransformVisitor extends JGeneric {

  /** Public handle for vector of generated Java ASTs that
    * correspond to datasets */
  public Vector datasets;

  /** Public handle for vector of generated Java ASTs that
    * correspond to factories */
  public Vector factories;

  /** Public handle for vector of generated Java ASTs that
    * correspond to directives */
  public Vector directives;

  /** Public handle for the generated Java AST that
    * corresponds to the interface for the TDD fields */
  public ASTCompilationUnit constantsInterface;

  /** Handle for the "import" declarations (used by all generated ASTs) */
  private Vector importDecls;

  /** Handle for the declaration of the TDD local fields
    * (will be used to construct constantsInterface)
    * @see #constantsInterface */
  private Vector constants;

  /** Name of the TDD class (i.e. name of the source file prefix) */
  private String unitName; // tddClassName

  /** Names of the ACOs to generate. */
  private Vector acoNames = null;

  /** Names of the TDD classes to generate. */
  private Vector tddClassNames = null;

  /** Names of the TDD constant classes to generate. */
  private Vector cstClassNames = null;

  /** Names of the TDD datasets to generates. */
  private Vector tddDatasetNames = null;

  /** Names of the TDD factories to generates. */
  private Vector tddFactoryNames = null;

  /** Names of the TDD test directives to generates. */
  private Vector tddTestDirectiveNames = null;


  static final private boolean debug = false;


  /** 
    * Entry point. Initializes needed handles and launches the visit.
    * @param topNode the root node of the tree to visit
    * @return the root node of the transformed tree
    */
  public Node translate(Node topNode) {

    init();
    return (Node)topNode.jjtAccept(this, null);
  }


  /**
    * Initialization of handles.
    */
  private void init() {

    datasets = new Vector();
    directives = new Vector();
    factories = new Vector();
    constants = new Vector();
    constantsInterface = null;

    importDecls = new Vector();

    /** Names of the ACOs to generate. */
    acoNames = new Vector();

    /** Names of the TDD classes to generate. */
    tddClassNames = new Vector();

    /** Names of the TDD constant classes to generate. */
    cstClassNames = new Vector();

    /** Names of the TDD datasets to generate. */
    tddDatasetNames = new Vector();

    /** Names of the TDD factories to generate. */
    tddFactoryNames = new Vector();

    /** Names of the TDD test directives to generate. */
    tddTestDirectiveNames = new Vector();
  }


  /** Names of the ACOs to generate. */
  public Vector getACONames() { return acoNames; }

  /** Names of the TDD classes to generate. */
  public Vector getTDDClassNames() { return tddClassNames; }

  /** Names of the TDD constant classes to generate. */
  public Vector getTDDConstantNames() { return cstClassNames; }
  
  /** Names of the TDD datasets to generates. */
  public Vector getTDDDatasetNames() { return tddDatasetNames; }

  /** Names of the TDD factories to generates. */
  public Vector getTDDFactoryNames() { return tddFactoryNames; }

  /** Names of the TDD test directives to generates. */
  public Vector getTDDTestDirectiveNames() { return tddTestDirectiveNames; }


  /**
    * Visit pattern for a node of type ASTADL_CompilationUnit.
    * <P><B>Algorithm</B><UL>
    * <LI>Stores the ASTImportDeclaration children nodes in importDecls;
    * adds one for "import ORG.opengroup.adl.runtime.*;"
    * <LI>Visits the ASTTDD_ClassDeclaration child node.
    * </UL>
    * @param node the ASTADL_CompilationUnit node to visit
    * @param data unused
    * @return the transformed ASTADL_CompilationUnit node
    */
  public Object visit(ASTADL_CompilationUnit node, Object data) {

    if (debug)
      System.out.println("visiting ASTADL_CompilationUnit");

    Node child, grandChild;
    ASTCompilationUnit createdNode = new
      ASTCompilationUnit(ADLJavaParserTreeConstants.JJTCOMPILATIONUNIT);

    ASTPackageDeclaration pack = new
      ASTPackageDeclaration(ADLJavaParserTreeConstants.JJTPACKAGEDECLARATION);
    parentChild(pack, makeName("adltest"), 0);
    parentChild(createdNode, pack, 0);

    ASTImportDeclaration adlruntime = new
      ASTImportDeclaration(ADLJavaParserTreeConstants.JJTIMPORTDECLARATION);
    parentChild(adlruntime, makeName("ORG.opengroup.adl.runtime"), 0);
    try {
      adlruntime.setProperty("onDemand", Boolean.TRUE);
    }
    catch ( PropertyAccessException e ) { }

    importDecls.addElement(adlruntime);
    parentChild(createdNode, adlruntime, 1);

    int nbChildren = 2;
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      child = node.jjtGetChild(i);

      if (child instanceof ASTImportDeclaration) {
        importDecls.addElement(child);
        parentChild(createdNode, child, nbChildren++);
      }
      if (child instanceof ASTTDD_ClassDeclaration) {
        // case of unnamed package
        String[] imports = null;
        try {
          imports = (String[])node.getProperty("classesFromUnnamedPackage");
        } catch(PropertyAccessException e) {}

        if (imports != null) {
          // The tested class is in an unnamed package: in the generated file(s),
          // any class (of this unnamed package) used by the tested class must be
          // explicitely imported
          ASTImportDeclaration impNode;
          for (int j = imports.length-1; j >= 0; j--) {
            impNode = new
              ASTImportDeclaration(ADLJavaParserTreeConstants.JJTIMPORTDECLARATION);
            JGenericVisitor.parentChild(impNode,
					JGenericVisitor.makeName(imports[j]),
					0);
            JGenericVisitor.parentChild(createdNode,
					impNode,
					nbChildren++);
	    importDecls.addElement(impNode);
          }
        }

        parentChild(createdNode, (Node)child.jjtAccept(this, data), nbChildren++);
      }
      if (child instanceof ASTTDD_UseDeclaration) {
        // "use" clause only used by stm
      }
    }
    return createdNode;
  }

  /**
    * Visit pattern for a node of type ASTTDD_ClassDeclaration.
    * <P><B>Algorithm</B><OL>
    * <LI>Creates an ASTClassDeclaration node for the class
    * "&#60;tddclass name&#62;TDD". This class has an implementation child
    * for interface "&#60;tddclass name&#62;_CST".
    * <LI>The input node children that correspond to method declarations
    * are added to the created node while the others are just visited
    * (they don't return anything but create the public handles datasets,
    * factories and directives).
    * <LI>Calls makeConstantsInterface to create the interface
    * "&#60;tddclass name&#62;_CST".
    * <LI>Returns the created node.
    * </OL>
    * @param node the ASTTDD_ClassDeclaration node to visit
    * @param data unused
    * @return the created node (ASTClassDeclaration)
    */
  public Object visit(ASTTDD_ClassDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_ClassDeclaration");

    ASTClassDeclaration createdNode = new
      ASTClassDeclaration(ADLJavaParserTreeConstants.JJTCLASSDECLARATION);
    try {
      int modif = JModifier.PUBLIC | JModifier.ABSTRACT;
      createdNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    parentChild(createdNode, makeName("ADLTest"), 0);
    
    unitName = node.getName();
    createdNode.setName(unitName + "TDD");

    // update tdd classes names list
    tddClassNames.addElement(unitName + "TDD");

    ASTNameList impl = new 
      ASTNameList(ADLJavaParserTreeConstants.JJTNAMELIST);
    parentChild(createdNode, impl, 1);
    parentChild(impl, makeName(unitName + "_CST"), 0);

    // update constant classes names list
    cstClassNames.addElement(unitName + "_CST");

    int insertChild = 2;
    for (int i = 0; i < node.jjtGetNumChildren(); i++) {
      Node vis = (Node)node.jjtGetChild(i).jjtGetChild(0).jjtAccept(this, data);
      if (vis != null) { // method declaration
        parentChild(createdNode, vis, insertChild++);
      }
    }

    makeConstantsInterface();
    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTTDD_ClassBodyDeclaration.
    * <BR>Just re-transmits the visitor to the child node.
    * @param node the ASTTDD_ClassBodyDeclaration node to visit
    * @param data unused
    * @return the transformed node
    */
  public Object visit(ASTTDD_ClassBodyDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_ClassBodyDeclaration");

    return node.jjtGetChild(0).jjtAccept(this, data);
  }


  /**
    * Visit pattern for a node of type ASTTDD_DatasetDeclaration.
    * <BR>Creates a JDatasetBuilder with as parameters the current node,
    * the tdd class name and the importations vector. Gets the
    * "compilUnit" of this builder, that corresponds as the root of the
    * AST for the class that implements the current dataset definition,
    * and adds this field in the "datasets" handle.
    * @see JDatasetBuilder
    * @param node the ASTTDD_DatasetDeclaration node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_DatasetDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_DatasetDeclaration");
    JDatasetBuilder jdb = new JDatasetBuilder((ASTTDD_DatasetDeclaration)node,
                                              importDecls,
                                              unitName);
    // update tdd dataset name list
    tddDatasetNames.addElement(jdb.getClassName());

    datasets.addElement(jdb.compilUnit);
    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_FactoryDeclaration.
    * <BR>Creates a JFactoryBuilder with as parameters the current node,
    * the tdd class name and the importations vector. Gets the
    * "compilUnit" of this builder, that corresponds as the root of the
    * AST for the class that implements the current factory definition,
    * and adds this field in the "factories" handle.
    * @see JFactoryBuilder
    * @param node the ASTTDD_FactoryDeclaration node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_FactoryDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_FactoryDeclaration");

    JFactoryBuilder jfb = new JFactoryBuilder
                              ((ASTTDD_FactoryDeclaration)node,
                               importDecls,
                               unitName);

    // update tdd factory name list
    tddFactoryNames.addElement(jfb.getClassName());    

    factories.addElement(jfb.compilUnit);

    return null;
  }


  /**
    * Visit pattern for a node of type ASTTDD_TestDirective.
    * <BR>Creates a JDirectiveBuilder with as parameters the current node,
    * the tdd class name and the importations vector. Gets the
    * "compilUnit" of this builder, that corresponds as the root of the
    * AST for the class that implements the current directive definition,
    * and adds this field in the "directives" handle.
    * @see JDirectiveBuilder
    * @param node the ASTTDD_TestDirective node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_TestDirective node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_TestDirective");

    JDirectiveBuilder jdb = new JDirectiveBuilder
                                ((ASTTDD_TestDirective)node,
                                 importDecls,
                                 unitName);

    // update tdd test directive names list
    tddTestDirectiveNames.addElement(jdb.getClassName());

    directives.addElement(jdb.compilUnit);

    // update list of ACO names
    Enumeration e = jdb.getACONames().elements();
    while ( e.hasMoreElements() ) {
      String acoName = (String)e.nextElement();
      if ( ! acoNames.contains(acoName) ) {
	acoNames.addElement(acoName);
      }
    }

    return null;
  }
  

  /**
    * Visit pattern for a node of type ASTTDD_FieldDeclaration.
    * <BR>Creates Java field declaration that corresponds to the TDD field
    * declaration, and adds it to the "constants" handle.
    * @see #makeConstantsInterface
    * @param node the ASTTDD_FieldDeclaration node to visit
    * @param data unused
    * @return nothing
    */
  public Object visit(ASTTDD_FieldDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_FieldDeclaration");

    ASTFieldDeclaration constNode = new
      ASTFieldDeclaration(ADLJavaParserTreeConstants.JJTFIELDDECLARATION);
    try {
      int modif = JModifier.PUBLIC;
      constNode.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    parentChild(constNode, node.jjtGetChild(0), 0);
    ASTVariableDeclarator varDecl;
    for(int j = 1; j < node.jjtGetNumChildren(); j += 2) {
      SimpleNode initExpr = (SimpleNode)node.jjtGetChild(j+1);
      // visit initExpr to get fully qualified name for fields
      initExpr.jjtAccept(this, data);
      varDecl = new
        ASTVariableDeclarator(ADLJavaParserTreeConstants.JJTVARIABLEDECLARATOR);
      parentChild(constNode, varDecl, (int)((j+1)/2));
      parentChild(varDecl, node.jjtGetChild(j), 0);
      parentChild(varDecl, initExpr, 1);
    }
    constants.addElement(constNode);
    return null;
  }
  
  
  /**
    * Visit pattern for a node of type ASTMethodDeclaration.
    * <BR>Returns the node unchanged except for last child (ASTBlock)
    * which is visited by an RemoveADLVisitor to transform its casual
    * ADL/ADL_new expressions.
    * @see RemoveADLVisitor
    * @param node the ASTMethodDeclaration node to visit
    * @param data unused
    * @return the transformed node
    */
  public Object visit(ASTMethodDeclaration node, Object data) {

    if (debug)
      System.out.println("visiting ASTMethodDeclaration");

    RemoveADLVisitor v = new RemoveADLVisitor();
    node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(v, null);
    
    Enumeration e = v.getACONames().elements();
    while ( e.hasMoreElements() )
    {
       String acoName = (String)e.nextElement();
       if ( ! acoNames.contains(acoName) )
       {
          acoNames.addElement(acoName);
       }
    }

    return node;
  }


  /**
    * Visit pattern for a node of type ASTName.
    * <P><B>Algorithm</B><UL>
    * <li>Gets defining tdd class name ("enclosingConstructName" property)
    * <li>Gets name kind ("astNameKind" property)
    * <LI><B>IF</B> the name is an attribute name, then it is a
    *    constant:<ul>
    * <LI><B>IF</B> the defining tdd class name is not null:<ul>
    * <li>Sets name in ASTName as
    *    <samp>&lt;defining tdd class name>_CST.&lt;constant name></samp>
    * </ul></ul></ul>
    *
    * @param node the ASTName node to visit
    * @param data unused
    * @return null
    */
  public Object visit(ASTName node, Object data) {

    if (debug)
      System.out.println("visiting ASTName : " + node.getName());

    String nameKind = "";
    String TDDClassName = null;
    try {
      TDDClassName = (String)node.getProperty("enclosingConstructName");
      nameKind = (String)node.getProperty("astNameKind");
    }
    catch(PropertyAccessException e) { }

    if (nameKind != null && nameKind.equals("LAttribute")) {  // Field
      if (TDDClassName != null && !TDDClassName.equals("")) { // TDD Field
	node.setName(TDDClassName + "_CST." + node.getName());
      }
    }

    return null;
  }


  /**
    * Creates the AST for interface "&#60;tddclass name&#62;_CST", that
    * gathers the tdd field declarations stored in the "constants" handle.
    * This interface will be implemented by all created classes.
    * @see #visit(ORG.opengroup.parsers.adljava.ASTTDD_FieldDeclaration, Object)
    */
  private void makeConstantsInterface() {

    if (debug)
      System.out.println("makeConstantsInterface");

    constantsInterface = new
      ASTCompilationUnit(ADLJavaParserTreeConstants.JJTCOMPILATIONUNIT);

    ASTPackageDeclaration pack = new
      ASTPackageDeclaration(ADLJavaParserTreeConstants.JJTPACKAGEDECLARATION);
    parentChild(pack, makeName("adltest"), 0);
    parentChild(constantsInterface, pack, 0);

    for (int i = 0; i < importDecls.size(); i++) {
      JGenericVisitor.parentChild(constantsInterface,
                                  (ASTImportDeclaration)importDecls.elementAt(i),
                                  i+1);
    }

    ASTInterfaceDeclaration interf = new
      ASTInterfaceDeclaration(ADLJavaParserTreeConstants.JJTINTERFACEDECLARATION);
    parentChild(constantsInterface, interf, importDecls.size()+1);
    try {
      int modif = JModifier.PUBLIC;
      interf.setProperty("modifier", new Integer(modif));
    }
    catch(PropertyAccessException e) {}
    interf.setName(unitName + "_CST");
    ASTInterfaceMemberDeclaration member;
    for (int i = 0; i < constants.size(); i++) {
      member = new
        ASTInterfaceMemberDeclaration(ADLJavaParserTreeConstants.JJTINTERFACEMEMBERDECLARATION);
      parentChild(interf, member, i);
      parentChild(member, (SimpleNode)constants.elementAt(i), 0);
    }
  }
}

// ***********************************************************************
// History Log:
// $Log: JTDDTransformVisitor.java,v $
// Revision 1.17  1998/10/14 08:17:26  mallet
// Bugs fixed: pr/155, pr/156: removed useless index of anonymous test
// directives.
//
// Revision 1.16  1998/07/02 08:49:27  mallet
// Importations in case of unnamed packages are now dealt with using
// property "classesFromUnnamedPackage".
//
// Revision 1.15  1998/06/12 15:08:35  mallet
// Added visit of initializer for TDD_FieldDeclaration to fix up name of
// ASTName nodes in case of a "use"d TDD field.
// Added visit method for ASTName to fix name in case of a "use"d TDD field.
//
// Revision 1.14  1998/06/05 14:17:05  mallet
// Changed location of method makeQualifiedName from NameAux to JBuilder.
// Removed class NameAux.
//
// Revision 1.13  1998/06/05 08:38:40  mallet
// Added code to get the list of used ACOs from the JDirectiveBuilder:
// this fixes a bug where those ACOs were not listed in the dependencies
// of the TDD class.
//
// Revision 1.12  1998/04/15 13:57:43  bareau
// Update according to the new NL mechanism: there are no NLD_* nodes.
//
// Revision 1.11  1998/03/10 17:26:47  mallet
// Merged v1.1 with development trunk
//
// Revision 1.10.2.1  1998/03/04 10:06:15  bareau
// The entry point for adl/tdd visitors is now a method "translate"
// that takes as parameter the root node of the source tree, performs
// necessary initializations and launches the visit.
//
// Revision 1.10  1998/02/27 13:08:00  bareau
// Updated copyright notice.
//
// Revision 1.9  1998/02/24 18:33:38  bareau
// Changed the root node of generated trees as ASTCompilationUnit,
// with package "adltest".
//
// Revision 1.8  1998/02/12 16:04:00  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.7  1998/02/10 14:48:45  bareau
// Corrected the ADD_CLASSPATH in the Makefile.
// Minor code cleanup in JTDDTransformVisitor.
//
// Revision 1.6  1998/02/10 14:13:49  cinard
// Reset vectors in reset() method.
//
// Revision 1.5  1998/02/02 13:49:12  bareau
// Changed the class "ADLremove" into "RemoveADL".
//
// Revision 1.4  1998/02/02 12:46:50  cinard
// Collect information for Makefile generation. This information is
// available via methods :
// getACONames(), getTDDClassNames(), getTDDConstantNames(),
// getTDDDatasetNames(), getTDDFactoryNames(), getTDDTestDirectiveNames().
//
// Revision 1.3  1998/01/21 16:23:55  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:50  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:18:03  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// Revision 1.2  1997/12/03 17:41:46  bareau
// Complete revision WITHOUT any tests.
//
// Revision 1.1  1997/11/26 19:08:26  adl2ext
// Transformer visitor for TDD/Java.
// Initial revision.
//
// ***********************************************************************
