// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
// 
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/translators/j/RemoveADLVisitor.java,v $
// $Revision: 1.3 $
// $Date: 1998/02/27 13:08:02 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:  .translators.j
// NAME:        RemoveADLVisitor.java
//
// HISTORY:
//      Cyrille Bareau, TOG-RI - Dec 16, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.translators.j;

import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.parsers.*;
import java.util.Vector;

/**
 * Transformation visitor.  Must change an ADL or ADLnew expression.
 * <BR>
 * Used in directives and test methods.
 * @see JDirectiveBuilder
 * @author Cyrille Bareau
 */

public class RemoveADLVisitor extends JGenericVisitor {

  private static final boolean debug = false;

  /** Names of the ACOs to generate. */
  private Vector acoNames = new Vector();
  

  public Vector getACONames() {
    return acoNames;
  }


  /**
    * Visit pattern for a node of type ASTPrimaryExpression.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the node's first child if a "ADL_new" expression, create
    * an ASTPrimaryExpression that is the return of the visit to this child,
    * and add as suffixes to this node the suffixes of the input node.
    * <BR>"ADL_new &#60;name&#62; &#60;suffix&#62;<SUP>*</SUP>  --&#62;
    * "new &#60;name&#62;ACO(null, this)&#60;suffix&#62;<SUP>*</SUP>"
    * <LI><B>ELSE</B> visit the children nodes.
    * </UL>
    * @param node the ASTPrimaryExpression to visit.
    * @param data unused.
    * @return the transformed node.
    */
  public Object visit(ASTPrimaryExpression node, Object data) {

    Node child = node.jjtGetChild(0);
    if (child instanceof ASTTDD_ADLnewExpression) {
      ASTPrimaryExpression createdNode = 
        (ASTPrimaryExpression)child.jjtAccept(this, data);
      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        parentChild(createdNode,
                    (Node)node.jjtGetChild(i).jjtAccept(this, data),
                    i+1);
      }
      return createdNode;
    }
    else {
      for (int i = 0; i < node.jjtGetNumChildren(); i++) {
        parentChild(node,
                    (Node)node.jjtGetChild(i).jjtAccept(this, data),
                    i);
      }
      return node;
    }
  }


  /**
    * Visit pattern for a node of type ASTTDD_ADLnewExpression.
    * <BR>Transforms the code "ADL_new &#60;name&#62;" into
    * "new &#60;name&#62;ACO(null, this).&#60;name&#62;"
    * @param node the ASTTDD_ADLnewExpression to visit.
    * @param data unused.
    * @return an ASTPrimaryExpression.
    */
  public Object visit(ASTTDD_ADLnewExpression node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_ADLnewExpression");
    
    ASTAllocationExpression alloc = new
      ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);

    String acoName = ((SimpleNode)node.jjtGetChild(0)).getName() + "ACO";

    // update aco names
    if ( ! acoNames.contains(acoName) ) {
      acoNames.addElement(acoName);
    }

    parentChild(alloc, makeName(acoName), 0);
    ASTArguments args = new ArgumentsBuilder
      ( new ASTNullLiteral(ADLJavaParserTreeConstants.JJTNULLLITERAL) )
      .addArgument(makeName("this"))
      .generatedAST();
    parentChild(alloc, args, 1);

    ASTDotSuffix dotsuffix = new
      ASTDotSuffix(ADLJavaParserTreeConstants.JJTDOTSUFFIX);
    dotsuffix.setName(((SimpleNode)node.jjtGetChild(0)).getName());

    ASTPrimaryExpression createdNode = new
      ASTPrimaryExpression(ADLJavaParserTreeConstants.JJTPRIMARYEXPRESSION);
    parentChild(createdNode, alloc, 0);
    parentChild(createdNode, dotsuffix, 1);

    return createdNode;
  }


  /**
    * Visit pattern for a node of type ASTTDD_ADLExpression.
    * <P><B>Algorithm</B><UL>
    * <LI><B>IF</B> the argument of the ADL expression is a class name
    * ("ADL(&#60;className&#62;)": call to a checked static method), returns an
    * ASTPrimaryExpression that is the root for code
    * "new &#60;className&#62;ACO(null, this)"
    * <LI><B>ELSE</B> visits the expression node, gets its type, and returns an
    * ASTPrimaryExpression that is the root for code
    * "new &#60;expressionType&#62;ACO(&#60;visited expression&#62;, this)"
    * </UL>
    * @param node the ASTTDD_ADLExpression to visit.
    * @param data unused.
    * @return the transformed node.
    */
  public Object visit(ASTTDD_ADLExpression node, Object data) {

    if (debug)
      System.out.println("visiting ASTTDD_ADLExpression. TypeName: " + node.getTypeName());

    ASTPrimaryExpression createdNode = new
      ASTPrimaryExpression(ADLJavaParserTreeConstants.JJTPRIMARYEXPRESSION);

    ASTAllocationExpression aco = new
      ASTAllocationExpression(ADLJavaParserTreeConstants.JJTALLOCATIONEXPRESSION);

    String acoName = node.getTypeName() + "ACO";

    // update aco names
    if ( ! acoNames.contains(acoName) ) {
      acoNames.addElement(acoName);
    }

    parentChild(aco,makeName(acoName) , 0);
    parentChild(createdNode, aco, 0);

    boolean isClass;
    try {
      isClass = ((Boolean)((SimpleNode)node.jjtGetChild(0))
                 .getProperty("isClass")).booleanValue();
    }
    catch(PropertyAccessException e) { isClass = false; }

    if (isClass) {
      // ADL(AClass) : call to a static method of a class
      // -> new AClassACO(null, this)
      parentChild(aco,
                  new ArgumentsBuilder(makeName("null"))
                  .addArgument(makeName("this"))
                  .generatedAST(),
                  1);
    }
    else {     // new <type>ACO(<expr>, this)
      parentChild(aco,
                  new ArgumentsBuilder((Node)node.jjtGetChild(0).jjtAccept(this, data))
                  .addArgument(makeName("this"))
                  .generatedAST(),
                  1);
    }
    return createdNode;
  }

}

// ***********************************************************************
// History Log:
// $Log: RemoveADLVisitor.java,v $
// Revision 1.3  1998/02/27 13:08:02  bareau
// Updated copyright notice.
//
// Revision 1.2  1998/02/12 16:04:02  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/02 13:49:12  bareau
// Changed the class "ADLremove" into "RemoveADL".
//
// Revision 1.4  1998/02/02 12:46:02  cinard
// Collect information for Makefile generation. This information is
// available via method getACONames().
//
// Revision 1.3  1998/01/21 16:23:52  bareau
// Javadoc comments + a few bug fixes.
//
// Revision 1.2  1998/01/08 16:32:44  bareau
// Version OK for bank/stock examples.
// Need more comments.
//
// Revision 1.1  1997/12/19 15:17:58  bareau
// Reorganisation of the translator package, with subpackages
// (only the "j" subdirectory for now).
// Added Makefiles.
//
// ***********************************************************************
