// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/DTD.java,v $
// $Revision: 1.3 $
// $Date: 1998/06/05 14:08:39 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   Unparser
// NAME:        DTD.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Apr 22, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Hashtable;
import com.objectspace.jgl.*;

/**
  * Abstract class that deals with the generation of a sgml documentation file
  * that matches the DocBook 3.0 DTD (document template definition).
  * See http://www.ora.com/davenport for information on DocBook.
  * <P>
  * The principle is as follows: the class template (aco.tmpl or tdd.tmpl)
  * is read and represented as a list of strings: each component of this list
  * is either an SGML entity or any string that ranges from an entity to the
  * following one. Then the unparsing consists in replacing these SGML entities.
  * <UL><LI>For simple entities, the replacement is just a string replacement.
  * <P>Example: replace entity &amp;signature; by the signature of the currently
  * documented method.
  * <LI>Some entities represent in fact a link to an other template (for
  * instance &amp;aco-funcs; has to be replaced by the template aco_func.tmpl).
  * When unparsing an ACO method, a list is created that corresponds to the
  * aco_func template: this is the "currentCompTmpl" field. Any entity
  * replacement is therefore performed on this list. When the unparsing for
  * this method is over, the list currentCompTmpl is itself spliced into
  * the main list classTmpl that corresponds to the aco template.
  * </UL>
  */
public abstract class DTD {

  /** The character for line separation. */
  private String linesep = System.getProperty("line.separator");

  /** Simple tabulation for display of indented assertions (if/try).
      Default: 3 blanks. */
  protected final String tab = "   ";

  /** Length of the tabulation. */
  protected final int tablength = tab.length();

  /** Indentation (a multiple of tabs). */
  protected String codeIndent;

  /** Hash table to store SGML entities (user-defined ones and those
    * defined in template files). */
  protected Hashtable entities;

  /** DTD template list for the current component.
    * Can be a representation of factory.tmpl, test_func.tmpl,
    * dataset.tmpl, test_directive.tmpl or aco_func.tmpl
    * (in $ADL2HOME/lib/nl/<locale>). */
  protected DList currentCompTmpl;

  /** DTD template list for the current ADL/TDD class.
    * Can be a representation of aco.tmpl or tdd.tmpl,
    * (in $ADL2HOME/lib/nl/<locale>). */
  protected DList classTmpl;

  /** Name of the current ADL/TDD class. */
  protected String className;

  /** Path name of the directory $ADL2HOME/lib/nl/<locale> */
  private String nlHome;


  /**
    * Single constructor. Just initializes nlHome and entities fields.
    * @param userEntities the hash table that contains SGML entities defined
    * by the user in nld blocks (comes from the parsing).
    */
  public DTD(Hashtable userEntities) {

    nlHome = System.getProperty("adl.home") + File.separator + "lib" +
      File.separator + "nl" + File.separator + getLocale() + File.separator;
    entities = (Hashtable)userEntities.clone();
  }


  /** Returns the class name. */
  public String getName() {
    return className;
  }


  /**
    * Generates an SGML "local link" (a reference to a tagged region
    * in the current SGML file).
    * @param className the current class name.
    * @param regionName the tag of the component to reference in the SGML file.
    * @param linkName the name of the component that is referenced.
    * @return the string <B>&lt;link linkend="<I>mangled-className</I>regionName"&gt;
    * linkName &lt;/link&gt;</B>
    */
  protected String makeLocalLink(String className, String regionName,
                                 String linkName) {

    return "<link linkend=\"" + SGMLNormalize(className) +
           regionName + "\">" + linkName + "</link>";
  }

  /**
    * Generates an SGML "local link" (a reference to a tagged region
    * in the current SGML file) towards a function/factory component.
    * @param methodSignature the signature of the component to reference.
    * @return the string <B>&lt;link linkend="<I>mangled-methodSignature</I>"&gt;
    * methodSignature &lt;/link&gt;</B>
    */
  protected String makeLocalLinkSignature(String methodSignature) {

    return "<link linkend=\"" +
           SGMLNormalizeSignature(methodSignature) + "\">" +
           methodSignature + "</link>\n";
  }


  /**
    * Generates an SGML "external link" (a reference to a tagged region
    * in a different SGML file).
    * @param className the class name of the component to reference.
    * @param regionName the tag of the component to reference in its SGML file.
    * @param linkName the name of the component to reference.
    * @param extension suffix of the referenced SGML file ("ACO" or "TDD").
    * @return the string <B>&lt;ulink url="<I>mangled-className+extension</I>.html#<I>mangled-regionName</I>"&gt;
    * className.linkName &lt;/ulink&gt;</B>
    */
  protected String makeExternalLink(String className, String regionName,
                                    String linkName, String extension) {
    
    return "<ulink url=\"" + SGMLNormalizeFileName(className + extension) + ".html#" +
           SGMLNormalize(regionName) + "\">" + className + "." + linkName + "</ulink>";
  }

  /**
    * Generates an SGML "external link" (a reference to a different SGML file).
    * @param className the class name of the component to reference.
    * @param extension suffix of the referenced SGML file ("ACO" or "TDD").
    * @return the string <B>&lt;ulink url="<I>mangled-className+extension</I>.html"&gt;
    * className &lt;/ulink&gt;</B>
    */
  protected String makeExternalLink(String className, String extension) {
    
    return "<ulink url=\"" + SGMLNormalizeFileName(className + extension) +
           ".html\">" + className + "</ulink>";
  }

  /**
    * Generates an SGML "external link" (a reference to a different SGML file)
    * towards a function/factory component.
    * @param className the class name of the component to reference.
    * @param methodSignature the signature of the component to reference.
    * @return the string <B>&lt;ulink url="<I>mangled-(className-methodSignature)</I>.html"&gt;
    * className.methodSignature &lt;/ulink&gt;</B>
    */
  protected String makeExternalMethodLink(String className,
                                          String methodSignature) {

    return "<ulink url=\"" +
           SGMLNormalizeFileName(className + "-" +
                                 SGMLNormalizeSignature(methodSignature)) +
           ".html\">" + className + "." + methodSignature + "</ulink>";
  }


  /**
    * Replace entity &amp;description; in the current template by the string
    * &lt;para&gt;<I>parameter</I>&lt;/para&gt;
    */
  public void emitComponentDescription(String entity) {

    currentCompTmpl.replace("&description;", "<para>" + entity + "</para>");
  }

  /**
    * Replace entity &amp;purpose; in the current template by the parameter.
    */
  public void emitComponentPurpose(String entity) {

    currentCompTmpl.replace("&purpose;", entity);
  }


  /**
    * if parameter entityName is equal to "description",
    * replace entity &amp;description; in the class template by the string
    * &lt;para&gt;<I>entity</I>&lt;/para&gt;
    * @param entityName name of the entity
    * @param entity value of the entity
    */
  public void emitClassEntity(String entity, String entityName) {

    if (entityName.equals("description")) {
      classTmpl.replace("&description;", "<para>" + entity + "</para>");
    }
  }


  /**
    * Replace entity &amp;return-type; in the current template by the parameter.
    */
  public void emitResultType(String resultType) {

    currentCompTmpl.replace("&return-type;", resultType);
  }


  /**
    * Replace entity &amp;thrown-exceptions; in the current template by
    * the string &lt;funcsynopsisinfo&gt;throws <I>thrownList</I>
    * &lt;/funcsynopsisinfo&gt;
    */
  public void emitThrownList(String thrownList) {

    currentCompTmpl.replace("&thrown-exceptions;",
                           "<funcsynopsisinfo>throws " + thrownList +
                           "</funcsynopsisinfo>");
  }


  /**
    * Remove component &amp;parameters-synopsis; and &amp;parameters; of the
    * current template.
    */
  public void removeEntities() {

    classTmpl.remove("&parameters-synopsis;");
    classTmpl.remove("&parameters;");
  }


  public String tmpl2string() {

    StringBuffer ret = new StringBuffer();

    Enumeration enum = classTmpl.elements();
    while (enum.hasMoreElements()) {
      ret.append((String)enum.nextElement());
    }
    return ret.toString();
  }


  public void emitParam(String paramType, String paramName, String paramDesc) {

    //    DListIterator paramdef = currentCompTmpl.find("&param-def*;");
    DListIterator paramdef = currentCompTmpl.find("&parameters-synopsis;");
    currentCompTmpl.insert(paramdef, "<paramdef>");
    currentCompTmpl.insert(paramdef, "&ptype;");
    currentCompTmpl.insert(paramdef, "<parameter> ");
    currentCompTmpl.insert(paramdef, "&pname;");
    currentCompTmpl.insert(paramdef, "</parameter></paramdef>");

    DListIterator parameter = currentCompTmpl.find("&parameters;");
    currentCompTmpl.insert(parameter, "<variablelist><varlistentry> <term><parameter>");
    currentCompTmpl.insert(parameter, "&pname;");
    currentCompTmpl.insert(parameter, "</parameter></term>\n<listitem><para>");
    currentCompTmpl.insert(parameter, "&short-description;");
    currentCompTmpl.insert(parameter, "</para></listitem>\n</varlistentry></variablelist>");

    currentCompTmpl.replace("&ptype;", paramType);
    currentCompTmpl.replace("&pname;", paramName);
    currentCompTmpl.replace("&short-description;", paramDesc);
  }


  protected String openFile(String fileName) {

    String tmplPath = nlHome + fileName;
    File tmplFile = new File(tmplPath);
    if (! tmplFile.exists())
      /*throw new FileNotFoundException*/
      System.err.println("No DTD Template file found: " +
                         tmplFile.getAbsolutePath());

    BufferedReader tmplReader = null;
    try {
      tmplReader = new BufferedReader(new FileReader(tmplFile));
    }
    catch(FileNotFoundException e) {}
    StringBuffer ret = new StringBuffer();
    try {
      while(tmplReader.ready()) {
        ret.append(tmplReader.readLine()).append(linesep);
      }
    }
    catch(IOException e) {
      System.err.println("I/O Error: Cannot read the DTD Template file:" +
                         tmplFile.getAbsolutePath());
      //throw e;
    }
    return ret.toString();
  }


  protected DList makeDList(String nl_class) {

    String entityName, entityValue;
    int userEntitiesIndex = entities.size();
    DList ret = new DList();
    int index2 = 0, index1 = nl_class.indexOf("<!ENTITY ", index2);;

    if (index1 != -1) {
      String str = "", val;
      for (Enumeration values = entities.elements(),
             keys = entities.keys();
           keys.hasMoreElements(); ) {
        val = (String)values.nextElement();
        str += "\n<!ENTITY " + (String)keys.nextElement() + " \"" +
          val.substring(1, val.length()-1) + "\">";
      }

      do {
        index1 = nl_class.indexOf("<!ENTITY ", index2);
        if (index1 != -1) {
          index2 = nl_class.indexOf(' ', index1 + 9);
          entityName = '&' + nl_class.substring(index1 + 9, index2) + ';';
          index1 = nl_class.indexOf('\"', index2);
          index2 = nl_class.indexOf('\"', index1 + 1);
          entityValue = nl_class.substring(index1+1, index2);
          entities.put(entityName, entityValue);
        }
      } while (index1 != -1);

      index2 = nl_class.indexOf("]>", index2);
      ret.add(nl_class.substring(0, index2) + str);
    }

    do {
      index1 = nl_class.indexOf('&', index2);
      if (index1 != -1) {
        ret.add(nl_class.substring(index2, index1));
        index2 = nl_class.indexOf(';', index1+1) + 1;
        if (index2 != 0)
          ret.add(nl_class.substring(index1, index2));
      }
    } while (! (index2 == 0 || index1 == -1));
    if (index1 == -1) {
      ret.add(nl_class.substring(index2));
    }
    return ret;
  }


  protected void replace(DList buf, String str, String newstr) {

    DListIterator iter = buf.find(str);
    if (iter.equals(buf.end())) {
      buf.pushBack(newstr);
    }
    else {
      buf.put(iter.index(), newstr);
    }
  }


  protected void emitSGMLNormalizedName(String name) {

    currentCompTmpl.replace("&mangled-name;",
                            SGMLNormalize(name));
  }


  protected String SGMLNormalize(String s) {

    StringBuffer ret = new StringBuffer();
    int n = s.length(); char c;
    for (int i = 0; i < n; i++) {
      c = s.charAt(i);
      if (Character.isLetterOrDigit(c))
        ret.append(Character.toUpperCase(c));
      else if (c == '_')
        ret.append('-');
      else if (c == '-')
        ret.append(c);
    }
    return ret.toString();
  }

  private String SGMLNormalizeType(String type) {

    int i = type.indexOf('[');
    if (i == -1)
      return SGMLNormalize(type);
    else return '1' + SGMLNormalize(type.substring(0, i));
  }

  protected String SGMLNormalizeFileName(String s) {

    StringBuffer ret = new StringBuffer();
    int n = s.length(); char c;
    for (int i = 0; i < n; i++) {
      c = s.charAt(i);
      if (Character.isLetterOrDigit(c))
        ret.append(Character.toLowerCase(c));
      else if (c == '_')
        ret.append('-');
      else if (c == '-')
        ret.append(c);
    }
    return ret.toString();
  }

  protected String SGMLNormalizeSignature(String signature) {

    String funcName = signature.substring(0, signature.indexOf('('));
    String params = signature.substring(funcName.length(), signature.length());
    
    if (params.length() == 2) // 0 parameter
      return SGMLNormalize(funcName);

    int index1 = params.indexOf(',');
    if (index1 == -1) { // 1 parameter
      return SGMLNormalize(funcName) + "-" +
        SGMLNormalizeType(params.substring(1, params.length()-1));
    }

    // at least 2 parameters
    int index2;
    String ret = SGMLNormalize(funcName) + "-" +
      SGMLNormalizeType(params.substring(1, index1));
    do {
      index2 = index1 + 2;
      index1 = params.indexOf(',', index2);
      if (index1 == -1)
        ret += '-' + SGMLNormalizeType(params.substring(index2, params.length()-1));
      else
        ret += '-' + SGMLNormalizeType(params.substring(index2, index1));
    } while(index1 != -1);
    return ret;
  }


  /** Return the current locale */
  private String getLocale() {

    String locale = System.getProperty("adl.locale");
    if ((locale == null) || (locale.equals("")))
      locale = "C";
    return locale;
  }

}

// ***********************************************************************
// History Log:
// $Log: DTD.java,v $
// Revision 1.3  1998/06/05 14:08:39  bareau
// Added support for multiple inheritance in C++.
// Began to javadoc.
//
// Revision 1.2  1998/05/26 15:24:35  bareau
// Improved the APIs for the creation of local/external links.
// Improved the SGML naming normalization.
//
// Revision 1.1  1998/05/15 13:49:29  bareau
// Added classes to deal with SGML docbook template.
// NLDGenericUnparser is the ancestor for NLD unparsers (this is the
// class that deals with DTD classes).
//
//
// ***********************************************************************
