// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/FileSectionExtractor.java,v $
// $Revision: 1.4 $
// $Date: 1998/03/10 17:26:47 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/Java Compiler
// NAME:        FileSectionExtractor.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Feb 09, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers;

import java.util.Vector;
import java.util.Enumeration;

import java.io.BufferedReader;

/** 
 *  Supplies methods to extract sections from files.
 */
public final class FileSectionExtractor
{
  /**
   * Constructs a FileSectionExtractor.
   * @param bufferedReader buffered reader from which the sections shall be 
   *                       read.
   */
  public FileSectionExtractor(BufferedReader bufferedReader)
  {
     _bufferedReader = bufferedReader;
  }

  /**
   * Extracts first section from file. 
   * Section starts at the first line of the file and extends to the 
   * first line containing supplied pattern (this last line is not included).
   * In case supplied pattern is not found section is considered to be
   * the whole file.
   * @param endPattern pattern delimiting the end of first section. This
   *                   pattern is just a string that will be searched for.
   * @return lines from the section.
   */

  public String[] extractFirstSection (String endPattern)
  {
     Vector section = new Vector();


     try
     {
        // get lines from section
        // note behaviour is optomistic : we fill section and hope endPattern 
        // will be found
        String currentLine = _bufferedReader.readLine();
        while (currentLine != null && currentLine.indexOf(endPattern) == -1)
        {
           section.addElement(currentLine);
           currentLine = _bufferedReader.readLine();
        }
        
     }
     catch (java.io.IOException e) { section.removeAllElements(); }

     // return section

     String[] tmp = new String[section.size()];
     section.copyInto(tmp);
     return tmp;
  }

  /**
   * Extracts middle section from file. 
   * Section starts after the next line of the file containing endPattern,
   * and extends to the first line containing startPattern (this last line 
   * is not included).
   * In case endPattern is not found section is void. In case startPattern
   * is not found section extends from endPattern (not included) up to end
   * of file.
   * @param startPattern pattern delimiting the beginning of the section. This
   *                     pattern is just a string that will be searched for.
   * @param endPattern pattern delimiting the end of first section. This
   *                   pattern is just a string that will be searched for.
   * @return lines from the section.
   */

  public String[] extractMiddleSection (String startPattern,String endPattern)
  {
     Vector section = new Vector();

     try 
     {
        // skip lines till endPattern
        String currentLine = _bufferedReader.readLine();
        while (   currentLine != null 
               && currentLine.indexOf(startPattern) == -1)
        {
           currentLine = _bufferedReader.readLine();
        }

        if (currentLine != null)
        {
           // start pattern found - put remaining lines till in section
           currentLine = _bufferedReader.readLine();
           while (    currentLine != null 
                   && currentLine.indexOf(endPattern) == -1 )
           {
              section.addElement(currentLine);
              currentLine = _bufferedReader.readLine();
           }
        }
     }
     catch (java.io.IOException e) { section.removeAllElements(); }

     String[] tmp = new String[section.size()];
     section.copyInto(tmp);
     return tmp;
  }

  /**
   * Extracts last section from file. 
   * Section starts at the first line containing supplied pattern (this line
   * is not included in section) and extends to the last line of the file.
   * In case supplied pattern is not found section is considered void.
   * @param startPattern pattern delimiting the beginning of the section. This
   *                     pattern is just a string that will be searched for.
   * @return lines from the section.
   */

  public String[] extractLastSection (String startPattern)
  {
     Vector section = new Vector();
     
     try 
     {
        // skip lines till startPattern
        String currentLine = _bufferedReader.readLine();
        while (currentLine != null && currentLine.indexOf(startPattern) == -1)
        {
           currentLine = _bufferedReader.readLine();
        }

        if (currentLine != null)
        {
           // start pattern found - put all remaining lines in section
           currentLine = _bufferedReader.readLine();
           while ( currentLine != null )
           {
              section.addElement(currentLine);
              currentLine = _bufferedReader.readLine();
           }
        }
     }
     catch (java.io.IOException e) { section.removeAllElements(); }

     // return section

     String[] tmp = new String[section.size()];
     section.copyInto(tmp);
     return tmp;
  }

  private BufferedReader _bufferedReader;
  
}

// ***********************************************************************
// History Log:
// $Log: FileSectionExtractor.java,v $
// Revision 1.4  1998/03/10 17:26:47  mallet
// Merged v1.1 with development trunk
//
// Revision 1.3.2.1  1998/03/04 18:47:12  cinard
// Analyzed files may now contain more than one tagged section.
//
// Revision 1.3  1998/02/27 13:02:52  cinard
// Updated copyright notice.
//
// Revision 1.2  1998/02/12 16:04:03  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/10 14:30:59  cinard
// Supplies methods to extract sections in files.
//
// ***********************************************************************
