// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/GenericUnparserVisitor.java,v $
// $Revision: 1.14 $
// $Date: 1998/07/28 15:58:24 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   Unparser
// NAME:        GenericUnparserVisitor.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Dec 9, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers;

import java.util.StringTokenizer;
import java.io.PrintWriter;
import java.io.File;

/** 
 * The GenericUnparserVisitor abstract class defines generic methods
 * to be used by language-specific unparser visitors.
 */
public abstract class GenericUnparserVisitor {
  private static final boolean debug = false;

  /** The name of the tool using the unparser visitor */
  protected String toolName;

  /** The version of the tool using the unparser visitor */
  protected String toolVersion;

  /** The standard header for generated files (valid for Java and C++) */
  protected String headerString;

  /** The standard header for generated Makefiles */
  protected String makefileHeaderString;

  /** PrintWriter to which the code is emitted ;
   * it is opened by the visitor when it encounters a node representing a
   * file.
   */
  protected PrintWriter out = null;

  /** Directory in which the unparser should generate its output */
  protected String outputDirectory = null;

  /** A StringBuffer maintaining the string emitted for an
   * <code>indent()</code> */
  private StringBuffer currentIndent = new StringBuffer();

  // formatting constants
  /** Sub-expression indentation level */
  protected int subIndent = 2;
  /** Regular statement indentation level */
  protected int regIndent = 3;
  /** Desired maximum number of characters per output line */
  protected int lineLimit = 65;

  /** Current position during unparsing (to be compared with 
   * <code>lineLimit</code>) */
  protected int currentColumn = 0;

  /**
   * Sets the unparser output directory.
   *
   * <p><b>Algorithm:</b> verifies that the directory exists and is a
   * directory, then set <code>outputDirectory</code>, otherwise don't
   * change.
   *
   * @param directory the output directory to use
   */
  public void setDirectory(String directory) {
    // if this is the same directory as previous one, just ignore request
    if (outputDirectory == null || !outputDirectory.equals(directory)) {
      // be sure the directory exists
      File file = new File(directory);
      if (file.exists() && file.isDirectory()) {
	outputDirectory = directory;
      } else {
	try {
	  if (file.mkdir()) {
	    System.out.println(toolName + ": Creating new directory '" +
			       directory + "'");
	    outputDirectory = directory;
	  } else {
	    System.err.println(toolName +
			       ": Failed to create new directory '" +
			       directory + "': generating in '.'");
	    outputDirectory = ".";
	  }
	} catch (SecurityException e) {
	  System.err.println(toolName + ": No such directory '" +
			     directory + "': generating in '.'");
	  outputDirectory = ".";
	}
      }
    }
  }

  /**
   * Default constructor: initialize toolName, toolVersion, headerString
   * and get default values for constants.
   *
   * @param toolname the name of the tool using the unparser
   * @param version   the RCS revision string of that tool
   */
  protected GenericUnparserVisitor(String toolname, String version) {
    toolName = toolname;

    StringTokenizer st = new StringTokenizer(version);
    if (st.countTokens() > 1) {
      st.nextToken();
      toolVersion = st.nextToken();
    }

    headerString = "// Generated by " + toolName + 
      " (" + toolVersion + ")";

    makefileHeaderString = "## Generated by " + toolName + 
      " (" + toolVersion + ")";
  }

  /**
   * Basic constructor: call default constructor and initialize constants
   *
   * @param toolname  the name of the tool using the unparser
   * @param version   the RCS revision string of that tool
   * @param linelimit the desired maximum length of an emitted line
   * @param regindent the regular statement indentation level
   * @param subindent the sub-expression indentation level
   */
  protected GenericUnparserVisitor(String toolname, String version,
				   int linelimit, int regindent,
				   int subindent) {
    this(toolname, version);

    if (linelimit > 0) lineLimit = linelimit;
    if (regindent > 0) regIndent = regindent;
    if (subindent > 0) subIndent = subindent;
  }

  /**
   * Emits an error message when the node type is not what is expected.
   *
   * @param node the faulty node
   */
  protected void incorrectNode(Object node) {
    System.err.println(toolName + ": Incorrect node type '" + 
		       node.getClass().getName() + "'");
  }

  /**
   * Verify that a node type is compatible with what was expected.
   *
   * <p><b>Algorithm:</b> the check is performed by using the method
   * <code>java.lang.Class.isAssignableFrom</code>.
   *
   * @param node  the node whose type should be verified
   * @param clazz the expected class
   * @return true if types are compatible, false otherwise
   *
   * @see java.lang.Class#isAssignableFrom
   */
  protected boolean assertNodeType(Object node, Class clazz) {
    return clazz.isAssignableFrom(node.getClass());
  }

  /**
   * Emit a string on the opened PrintWriter stream and advance the
   * current column marker. 
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if the string is null, ignore it
   * <li>else, emit the string
   * <li>increment <code>currentColumn</code> by the length of the initial
   *     string
   * </ul>
   * 
   * <p>This method may be redefined in sub classes if filtering through
   * addUnicodeEscapes is not desired.
   *
   * @param str the string to emit
   */
  protected void emit(String str) {
    if (str != null) {
      if (debug) System.out.print(str);

      // print out the string
      out.print(str);
      // and take care of current column
      currentColumn += str.length();
    }
  }

  /**
   * Emits a newline character (system dependent) on the opened
   * PrintWriter stream and reset current column marker.
   *
   * @see java.io.PrintWriter#println
   */
  protected void newline() {
    if (debug) System.out.println();

    // end line
    out.println();
    // and reset current column
    currentColumn = 0;
  }

  /**
   * Emits the current indentation string.
   */
  protected void indent() {
    // emit current indent
    emit(currentIndent.toString());
  }

  /**
   * Returns the current value of indentation level.
   */
  protected int getIndent() {
    return currentIndent.length();
  }

  /**
   * Sets the current value of the indentation level: according to the new
   * indentation level, proceed with <code>addIndent</code> or
   * <code>delIndent</code>. 
   *
   * @param newIndent the new indentation level
   */
  protected void setIndent(int newIndent) {
    // determine what to do according to diff with current indent level
    int indentDiff = newIndent - currentIndent.length();
    if (indentDiff > 0) addIndent(indentDiff);
    else if (indentDiff < 0) delIndent(-indentDiff);
  }

  /**
   * Adds an extra indent to the current indentation level: append
   * whitespace characters to the end of the <code>currentIndent</code>
   * buffer.
   *
   * @param extraIndent the extra indent to add
   */
  protected void addIndent(int extraIndent) {
    // just append the extra indent to current indent
    for (int i = 0; i < extraIndent; ++i) {
      currentIndent.append(" ");
    }
  }

  /**
   * Removes an extra indent from the current indentation level: reset
   * the length of the <code>currentIndent</code> buffer appropriately.
   *
   * @param extraIndent the extra indent to remove
   */
  protected void delIndent(int extraIndent) {
    // be sure current indent is at least as large as indent to remove
    int newLength = currentIndent.length() - extraIndent;
    if (newLength >= 0) {
      currentIndent.setLength(newLength);
    }
  }

  /** Get current column. */
  public int getCurrentColumn() {
     return currentColumn;
  }


}

// ***********************************************************************
// History Log:
// $Log: GenericUnparserVisitor.java,v $
// Revision 1.14  1998/07/28 15:58:24  mallet
// Added a debug mode where the text is emitted to stdout as well.
//
// Revision 1.13  1998/06/05 14:08:40  bareau
// Added support for multiple inheritance in C++.
// Began to javadoc.
//
// Revision 1.12  1998/05/15 13:49:30  bareau
// Added classes to deal with SGML docbook template.
// NLDGenericUnparser is the ancestor for NLD unparsers (this is the
// class that deals with DTD classes).
//
// Revision 1.11  1998/05/11 12:18:08  cinard
// Added getCurrentColumn() method.
//
// Revision 1.10  1998/03/19 09:10:03  mallet
// Fixed bug in generated code for C++: Unicode escapes should not be
// generated. Moved addUnicodeEscapes and redefined emit to call it in
// JUnparserVisitor. emit in GenericUnparserVisitor no longer calls
// addUnicodeEscapes.
//
// Revision 1.9  1998/02/27 13:02:52  cinard
// Updated copyright notice.
//
// Revision 1.8  1998/02/25 08:18:11  bareau
// Field toolName made protected (used by JUnparser).
//
// Revision 1.7  1998/02/25 07:58:17  mallet
// Changed the setDirectory method to try to create the directory when it
// does not exist: if it cannot be created, the outputDirectory is set to
// '.' (bug fixed).
//
// Revision 1.6  1998/02/12 16:04:04  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.5  1998/02/02 12:51:34  cinard
// Added header string for generated Makefiles (makefileHeaderString field).
//
// Revision 1.4  1998/01/07 15:11:30  mallet
// Updated and fixed javadoc comments.
//
// Revision 1.3  1998/01/05 16:33:36  mallet
// Added outputDirectory field and setDirectory method to specify the
// directory where files will be generated.
//
// Revision 1.2  1997/12/17 11:06:44  mallet
// Fixed some javadoc comments using cross-reference.
//
// Revision 1.1  1997/12/09 16:45:55  mallet
// Abstract generic Unparser class providing language-independent unparsing
// services.
//
// ***********************************************************************
