// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/LineCutter.java,v $
// $Revision: 1.3 $
// $Date: 1998/02/27 13:02:52 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/Java Compiler
// NAME:        LineCutter.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Jan 28, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers;

import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;

/** 
 * LineCutter supplies methods to cut long strings into shorter ones.
 */
public final class LineCutter
{
  /**
   * Constructs a LineCutter object.
   * <p>A line cutter can cut lines into shorter lines whose length is at
   * most lineLimit.
   * <p>Note tabs ('\t') are considered to be 8 characters long.
   * @param lineLimit   limit for the line length.
   */

  public LineCutter(int  lineLimit)
  {
     this(lineLimit,"","");
  }

  /**
   * Constructs a LineCutter object.
   * <p>A line cutter can cut lines into shorter lines whose length is at
   * most lineLimit. Additional formatting information is supplied :
   * beforeNewline and afterNewline which are simply strings to add before
   * and after each <new line> character when a line is cutted.
   * <p>Note tabs ('\t') are considered to be 8 characters long.
   * @param lineLimit   limit for the line length.
   * @param beforeNewline  prefix for next lines.
   * @param afterNewline  suffix for next lines.
   */

  public LineCutter(int    lineLimit,
                    String beforeNewline,
                    String afterNewline)
  {
     _lineLimit     = lineLimit;
     _beforeNewline = beforeNewline;
     _afterNewline  = afterNewline;
  }

  /**
   * Cut a long line into sevaral shorter lines. Each resulting line
   * is at most 'lineLimit' character long.
   * @param longLine string containing the long line to cut into several 
   *        shorter lines
   * @return a string made up of the resulting lines, separated by newlines
   *         character.
   */
  public final String cutLine(String longLine)
  {
     StringBuffer ret             = new StringBuffer();

     StringBuffer shortLine       = new StringBuffer();
     int          shortLineLength = 0;

     int    beforeNewlineLength   = _length(_beforeNewline);

     int    afterNewlineLength    = _length(_afterNewline);

     StringTokenizer st           = new StringTokenizer(longLine," \t",true);

     String token                 = null;

     boolean skipSpaceToken       = false; // flag indicating that space tokens
                                           // shall be skipped

     while ( st.hasMoreTokens() )
     {
        token       = st.nextToken();

        // check if line limit reached
        if (   shortLineLength + _length(token) + beforeNewlineLength 
             > _lineLimit )
        {
           // line limit reached

           // remove spaces at end of short line
           removeSpacesAtEnd(shortLine);

           // add new short line to the string to return
           ret.append(shortLine.toString() + _beforeNewline + "\n");

           // create a new short line
           shortLine       = new StringBuffer(_afterNewline);   
           shortLineLength = afterNewlineLength;

           // skip space tokens at beginning of new line
           skipSpaceToken = true;
        }

        if ( skipSpaceToken )
        {
           if  ( token.equals(" ") || token.equals("\t") )
           { 
              // space token, skip!
              continue; 
           }
           else 
           { 
              // not a space token, stop skipping tokens
              skipSpaceToken = false; 
           }
        }

        // append token to current short line
        shortLine.append(token);
        shortLineLength += _length(token);
     }

     ret.append(shortLine.append("\n").toString());
     return ret.toString();
  }

  /**
   * Remove spaces at the end of a string buffer.
   * @param sb a string buffer.
   */
  private final void removeSpacesAtEnd(StringBuffer sb)
  {
     int length = sb.length();
     int c = sb.charAt(length - 1);
     while ( ( length > 0 ) && ( c == ' ' || c == '\t' ) )
     {
        length = length - 1;
        c = sb.charAt(length - 1);
     }
     sb.setLength(length);
  }

  /**
   * Compute length of a string, counting 8 characters for a tab.
   */
  private final int _length(String s)
  {
     return s.length() + countChar(s,'\t') * ( 8 - 1 );
  }

  /**
   * Count occurences of a specified char in a string.
   * @param s the string.
   * @param c the char.
   * @return number of occurences of c in s.
   */
  private final int countChar(String s,int c)
  {
     int count = 0;
     int index = s.indexOf(c,0);
     while ( index != -1 )
     {
        count++;
        index = s.indexOf(c,index + 1);
     }
     return count;
  }

  // -- private fields used for cutted lines formatting --

  /** Limit for line length */
  private int _lineLimit        = 80;

  /** String to insert before a <new line>. */
  private String _beforeNewline  = "";

  /** String to insert after a <new line>. */
  private String _afterNewline  = "";

}

// ***********************************************************************
// History Log:
// $Log: LineCutter.java,v $
// Revision 1.3  1998/02/27 13:02:52  cinard
// Updated copyright notice.
//
// Revision 1.2  1998/02/12 16:04:04  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.1  1998/02/10 14:31:38  cinard
// Supplies methods to cut long lines in shorter ones.
//
// ***********************************************************************
