// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/NLDGenericUnparser.java,v $
// $Revision: 1.6 $
// $Date: 1998/06/12 13:22:15 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   Unparser
// NAME:        NLDGenericUnparser.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Apr 22, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers;

import java.util.StringTokenizer;
import java.util.Vector;
import java.io.*;
import ORG.opengroup.adl.nlengine.NLEngine;
import ORG.opengroup.adl.nlengine.NLException;
import ORG.opengroup.adl.symboltable.Scope;
import ORG.opengroup.adl.symboltable.LCallable;

/** 
 * The NLDGenericUnparser abstract class defines generic methods
 * to be used by language-specific NLD unparser visitors.
 */
public abstract class NLDGenericUnparser extends GenericUnparserVisitor {

  /** The version number of this makefile generator */
  static private String versionId = "$Revision: 1.6 $";

  /** The NL engine to deal with the Prolog interpreter
    * (for translating assertions). */
  protected NLEngine nlengine;

  /** A handle on the class that deals with the DTD template for
    * generated documentation of ACOs (ADL classes). */
  private DTD_ACO aco;

  /** A handle on the class that deals with the DTD template for
    * generated documentation of TDD classes. */
  private DTD_TDD tdd;

  /** The value of the handle either aco/tdd, casted on DTD,
    * the ancestor class of both DTD_ACO/DTD_TDD. Just a trick
    * to avoid dynamic type lookup and dynamic casting, when
    * using directly methods of DTD. */
  private DTD dtd;

  /** Scope of the currently unparsed method. */
  private Scope currentMethodScope;

  /** Scope of the currently unparsed class. */
  private Scope classScope;

  static private final boolean debug = true;


  /**
    * Default constructor: initialize toolName, versionId and
    * the NL engine
    *
    * @param toolname the name of the tool using the unparser.
    * @param versionId the RCS revision string of that tool.
    * @param nl the NL engine (interface with the Prolog engine).
    */
  protected NLDGenericUnparser(String toolname, String versionId, NLEngine nl) {

    super(toolname, versionId);
    nlengine = nl;
  }

  /**
    * Set aco and dtd handles.
    * @param className name of the ADL class
    */
  protected void initialize_aco(String className) {
    aco = new DTD_ACO(className, nlengine.getEntities());
    dtd = (DTD)aco;
  }

  /**
    * Set tdd and dtd handles.
    * @param className name of the TDD class
    */
  protected void initialize_tdd(String className) {
    tdd = new DTD_TDD(className, nlengine.getEntities());
    dtd = (DTD)tdd;
  }


  /**
    * Send a query to the NL engine to translate an expression
    * in a given scope.
    * @param expr the expression to translate.
    * @param scope the scope of the expression (to find the right
    * mapping for idents used in the expression).
    * @return the answer from the NL engine.
    */
  protected String query(String expr, Scope scope) {

    String ret = "";
    try {
      ret = nlengine.query(expr, scope);
    }
    catch(NLException e) {
      System.err.println("Error occurred while generating documentation: cannot translate " +
                         expr);
    }
    return ret;
  }


  /**
    * Send a query to the NL engine to find tyhe property translation
    * for an expression in a given scope.
    * @param expr the expression to translate.
    * @param scope the scope of the expression (to find the right
    * mapping for idents used in the expression).
    * @param property The property to search a mapping for ("description",
    * "includes", "seeAlso" or "purpose").
    * @return the answer from the NL engine.
    */
  protected String query(String expression, Scope scope, String property) {

    String ret = "";
    try {
      ret = nlengine.query(expression, scope, property);
    }
    catch(NLException e) {
      System.err.println("Error occurred while generating documentation: cannot translate " +
                         expression);
    }
    return ret;
  }


  /**
    * Flush the templat buffer in the output file and close it.
    */
  protected void close() {

    dtd.removeEntities();
    out.print(dtd.tmpl2string());
    out.flush();
    out.close();
    out = null;
  }


  /**
    * Search the description translation for the current class
    * and emit the corresponding header.
    * @param scope the current class scope
    * @see DTD#emitClassEntity
    */
  protected void emitClassEntity(Scope scope) {

    classScope = scope;
    String desc = query(dtd.getName(), scope, "description");
    if (! desc.equals(""))
      dtd.emitClassEntity(desc, "description");
  }


  /**
    * Search the description and purpose translations for the current method
    * or test function and emit the corresponding header.
    * @param methodName the name of the current method
    * @param scope the current scope
    * @see DTD#emitComponentDescription
    * @see DTD#emitComponentPurpose
    */
  protected void emitMethodEntity(String methodName, Scope scope) {

    String sign = methodName + ((LCallable)scope).getNLSignature();
    currentMethodScope = scope;

    String str = query(sign, scope, "description");
    if (! str.equals(""))
      dtd.emitComponentDescription(str);

    str = query(sign, scope, "purpose");
    if (! str.equals(""))
      dtd.emitComponentPurpose(str);
  }


  /**
    * Search the description translation for the current method parameter
    * and emit the corresponding entities.
    * @param paramType the type of the parameter
    * @param paramName the name of the parameter
    * @see DTD#emitParam
    */
  protected void emitParam(String paramType, String paramName) {

    if (paramType.equals("")) {
      dtd.emitParam("", "", "");
    }
    else {
      String paramDesc = query(paramName, currentMethodScope, "description");    
      dtd.emitParam(paramType, paramName, paramDesc);
    }
  }


  /**
    * @see DTD#emitResultType
    */
  protected void emitResultType(String resultType) {

    dtd.emitResultType(resultType);
  }


  /**
    * @see DTD#emitThrownList
    */
  protected void emitThrownList(String thrownList) {

    dtd.emitThrownList(thrownList);
  }


  /**
    * @see DTD#makeLocalLink
    */
  protected String makeLocalLink(String className, String regionName,
                                 String linkName) {

    return dtd.makeLocalLink(className, regionName, linkName);
  }

  /**
    * @see DTD#makeLocalLink
    */
  protected String makeLocalLink(String className, String linkName) {

    return dtd.makeLocalLink(className, "", linkName);
  }

  /**
    * @see DTD#makeExternalLink
    */
  protected String makeTDDExternalLink(String className, String regionName,
                                       String linkName) {
    
    return dtd.makeExternalLink(className, regionName, linkName, "TDD");
  }


  /************************************************
     ACO
  ************************************************/

  /**
    * Emit header of a Java class
    * @see DTD_ACO#emitClassHeader
    */
  protected void emitClassHeader(String fullName, String modif,
                                 String extendsName, String implementsName) {

    aco.emitClassHeader(fullName, modif, extendsName, implementsName);
  }

  /**
    * Emit header of a C++ class
    * @see DTD_ACO#emitClassHeader
    */
  protected void emitClassHeader(String fullName, String modif,
                                 String[] extendsNames) {

    aco.emitClassHeader(fullName, modif, extendsNames);
  }

  /**
    * @see DTD_ACO#emitClassInherited
    */
  protected void emitClassInherited(String inhClassName) {

    aco.emitClassInherited(inhClassName);
  }

  /**
    * @see DTD_ACO#emitClassInherited
    */
  protected void emitClassInherited(Vector inhClassNames) {

    aco.emitClassInherited(inhClassNames);
  }


  /**
    * @see DTD_ACO#emitImport
    */
  protected void emitImport(StringBuffer importedFile, boolean onDemand) {

  }

  /**
   * @see DTD_ACO#emitInclude
   */
  protected void emitInclude(String imcludedFileName) {
    // TBD cia - implementation should be put here
  }



  /**
    * @see DTD_ACO#beginMethodTemplate
    */
  protected void beginMethodTemplate(String methodName) {

    aco.beginMethodTemplate(methodName);
  }

  /**
    * @see DTD_ACO#endMethodTemplate
    */
  protected void endMethodTemplate() {

    aco.endMethodTemplate();
  }


  /**
    * @see DTD_ACO#endPrototype
    */
  protected void endPrototype() {

    aco.endPrototype();
  }


  /**
    * @see DTD_ACO#emitSuperSemantics
    */
  protected void emitSuperSemantics(boolean constructor) {

    aco.emitSuperSemantics(null, constructor);
  }

  /**
    * @see DTD_ACO#emitSuperSemantics
    */
  protected void emitSuperSemantics(String parent, boolean constructor) {

    aco.emitSuperSemantics(parent, constructor);
  }


  /**
    * @see DTD_ACO#emitNormal
    */
  protected void emitNormal(String expr, boolean normal) {

    aco.emitNormal(expr, normal);
  }


  /**
    * @see DTD_ACO#emitAssertion
    */
  protected void emitAssertion(String ass, String label,
                               String image, boolean excBhv) {

    aco.emitAssertion(ass, label, image, excBhv);
  }


  /**
    * @see DTD_ACO#emitCondition
    */
  protected void emitCondition(String ass, String image, boolean withElse) {

    aco.emitCondition(ass, image, withElse);
  }

  /**
    * @see DTD_ACO#emitEndCondition
    */
  protected void emitEndCondition() {

    aco.emitEndCondition();
  }

  /**
    * @see DTD_ACO#emitElse
    */
  protected void emitElse() {

    aco.emitElse();
  }


  /**
    * @see DTD_ACO#emitTry
    */
  protected void emitTry(int nbCatch) {

    aco.emitTry(nbCatch);
  }

  /**
    * @see DTD_ACO#emitEndTry
    */
  protected void emitEndTry() {

    aco.emitEndTry();
  }

  /**
    * @see DTD_ACO#emitCatch
    */
  protected void emitCatch(String excType, String excName) {

    aco.emitCatch(excType, excName);
  }

  /**
    * @see DTD_ACO#emitCatch
    */
  protected void emitCatches(String excType, String excName) {

    // TBD
//     aco.emitCatch(excType, excName);
  }


  /**
    * @see DTD_ACO#emitQuantifier
    */
  protected void emitQuantifier(String label, String quantifier, String image) {

    aco.emitQuantifier(label, quantifier, image);
  }

  /**
    * @see DTD_ACO#emitDomain
    */
  protected void emitDomain(String name, String domain, int index) {

    aco.emitDomain(name, domain, index);
  }

  /**
    * @see DTD_ACO#emitEndQuantified
    */
  protected void emitEndQuantified() {

    aco.emitEndQuantified();
  }


  /************************************************
     TDD
  ************************************************/

  /**
    * @see DTD_TDD#emitUseClass
    */
  protected void emitUseClass(String usedClass) {

    tdd.emitUseClass(usedClass);
  }


  /**
    * Search the description translation for the current constant definition
    * and call the tdd emitConstant method.
    * @see DTD_TDD#emitConstant
    */
  protected void emitConstant(String type, String name, String expr) {

    tdd.emitConstant(type, name, expr,
                     query(name, classScope, "description"));
  }


  /**
    * Search the description translation for the current dataset definition
    * and call the tdd emitConstant method.
    * @see DTD_TDD#emitDataset
    */
  protected void emitDataset(String dsName, String dsType, String expr) {

    tdd.emitDataset(dsName, dsType, expr,
                    query(dsName, classScope, "description"));
  }


  /**
    * @see DTD_TDD#beginFactoryTemplate
    */
  protected void beginFactoryTemplate(String factName) {

    tdd.beginFactoryTemplate(factName);
  }

  /**
    * @see DTD_TDD#endFactoryTemplate
    */
  protected void endFactoryTemplate() {

    tdd.endFactoryTemplate();
  }


  /**
    * @see DTD_TDD#beginTestFuncTemplate
    */
  protected void beginTestFuncTemplate(String funcName, String signature) {

    tdd.beginTestFuncTemplate(funcName, signature);
  }

  /**
    * @see DTD_TDD#endTestFuncTemplate
    */
  protected void endTestFuncTemplate() {

    tdd.endTestFuncTemplate();
  }


  /**
    * @see DTD_TDD#beginDirectiveTemplate
    */
  protected void beginDirectiveTemplate(String dirName) {

    tdd.beginDirectiveTemplate(dirName,
                               query(dirName, classScope, "purpose"));
  }

  /**
    * @see DTD_TDD#endDirectiveTemplate
    */
  protected void endDirectiveTemplate() {

    tdd.endDirectiveTemplate();
  }


  /**
    * @see DTD_TDD#emitCode
    */
  protected void emitCode(String code) {

    tdd.emitCode(code);
  }


  /**
    * @see DTD_TDD#emitDomain
    */
  protected void emitDomain(String dataset) {

    tdd.emitDomain(dataset);
  }


  /**
    * @see DTD_TDD#emitACO
    */
  protected void emitACO(String acoName) {

    tdd.emitACO(acoName);
  }


  /**
    * @see DTD_TDD#emitTestFunction
    */
  protected void emitTestFunction(String funcName, String signature) {

    tdd.emitTestFunction(funcName, signature);
  }

}

// ***********************************************************************
// History Log:
// $Log: NLDGenericUnparser.java,v $
// Revision 1.6  1998/06/12 13:22:15  bareau
// Added support for NL unparsing of quantified assertions.
//
// Revision 1.5  1998/06/10 15:52:57  bareau
// Bug fix: when the behavior of a constructor overrides the inherited
// one, the created link was incorrect.
//
// Revision 1.4  1998/06/08 15:10:42  cinard
// Added emitInclude() method.
//
// Revision 1.3  1998/06/05 14:08:40  bareau
// Added support for multiple inheritance in C++.
// Began to javadoc.
//
// Revision 1.2  1998/05/26 15:31:22  bareau
// Added comments and new APIs (references to inherited classes,
// used classes, etc.)
//
// Revision 1.1  1998/05/15 13:49:31  bareau
// Added classes to deal with SGML docbook template.
// NLDGenericUnparser is the ancestor for NLD unparsers (this is the
// class that deals with DTD classes).
//
//
// ***********************************************************************
