// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/c/CNLDUnparserVisitor.java,v $
// $Revision: 1.7 $
// $Date: 1998/08/06 14:10:57 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/C Compiler
// NAME:        CUnparserVisitor.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Apr 29, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.c;

import java.io.*;
import java.util.Vector;

import com.objectspace.jgl.Pair;

import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.unparsers.NLDGenericUnparser;
import ORG.opengroup.adl.unparsers.DTD;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import ORG.opengroup.adl.nlengine.NLEngine;

/** 
  * This class defines a visitor which can be applied to an AST
  * generated from the ADLCParser grammar. It does <B>NOT</B>
  * modify this AST: its only effect is to produce an SGML file
  * that corresponds to the documentation of the visited file.
  *
  * <p>The basic algorithm is to walk through the AST using the "visitor"
  * pattern (see Design Patterns from Gamma et al.) and emit the code on
  * an output stream maintained by the visitor itself.
  *
  * <p>The unparsing starts by calling the <a href="#unparse">unparse()</a>
  * method passing the top of AST to unparse as parameter.
  */

public class CNLDUnparserVisitor extends NLDGenericUnparser
                                 implements ADLCParserVisitor

{
  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.7 $";

  /** Temporary buffer for the currently unparsed assertion. */
  private StringBuffer assertionBuffer;

  /** Flag set when visiting an assertion. */
  private boolean inAssertion = false;

  /** Flag set when visiting a TDD definition (constant, dataset, factory). */
  private boolean inTddDef = false;

  /** Flag set when the current assertion is an "exception behavior:
    * assertion. */
  private boolean exceptionBhv;

  /** Name of the ADL/TDD module. */
  private String moduleName;

  /** Index for anonymous directives. */
  private int anonDirectiveIndex;
  
  /** Level of nesting inside adl assertion groups */
  private int assertionGroupLevel = 0;

  static private final boolean debug = false;

  /**
   * Default constructor: call superclass constructor.
   *
   * @param toolname the name of the tool using this unparser visitor
   * @param nlengine the NL engine to deal with the Prolog interpreter
   */
  public CNLDUnparserVisitor(String toolname, NLEngine nlengine) {

    super(toolname, versionId, nlengine);
  }


  /**
   * Starts unparsing an AST.
   *
   * If the passed AST root is not an CompilationUnit, an error is
   * reported and the AST is not unparsed. Otherwise the AST is
   * visited.
   *
   * @param ast the root node of the AST to unparse
   */

  public void unparse(ASTADL_TranslationUnit ast) {

    if (debug)
      System.out.println("NLD unparse");

    ast.jjtAccept(this, null);
  }

  /**
   * Visit SimpleNode.
   * <p> Just called in case acceptor not unimplemented in subclass.
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(SimpleNode node, Object data)
  {
     System.out.println( node + ": acceptor not unimplemented in subclass?");
     node.childrenAccept(this,data);
     return data;
  }


  

  /**
   * Visits an TranslationUnit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>not relevant for nld unparsing
   * </ul>
   *
   * @param node the visited TranslationUnit
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTranslationUnit node, Object data) 
  {
    throw new Error("Should not be visited!");
  }

  

  /**
   * Visits an ADL_TranslationUnit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the name of the declared adl module and open
   *     the corresponding stream
   * <li>visit children
   * </ul>
   *
   * @param node the visited ADL_TranslationUnit
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_TranslationUnit node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_TranslationUnit");

    // get the name of the declared adlmodule and open
    // the corresponding stream.
    
    // The declaration node is always the last child
    SimpleNode decl = null;
    if (node.jjtGetNumChildren() >= 1) {
      decl = (SimpleNode)node.jjtGetChild(node.jjtGetNumChildren() - 1);
    }

    if ( ! ((decl instanceof ASTADL_AnnotatedDeclaration) ||
	    (decl instanceof ASTTDD_AnnotatedDeclaration))) {
      System.err.println(toolName + ": can't get name of file: " +
			 "no adl or tdd class declaration");
      return data;
    }

    // now we know the decl node is an adl or tdd module
    moduleName = decl.getName();

    if (decl instanceof ASTADL_AnnotatedDeclaration)
      initialize_aco(moduleName);
    else 
      initialize_tdd(moduleName);

    // try to open the output stream
    try {
      // Try to get best performance through a BufferedWriter on top of
      // a FileWriter
      String fileName = outputDirectory + File.separator + moduleName 
	              + ((decl instanceof ASTADL_AnnotatedDeclaration) 
			 ? "ACO" 
			 : "TDD") 
	              + ".sgml";

      if (Adlt.verboseMode()) {
	System.out.println("+++ generating " + fileName);
      }

      FileWriter outStream = new FileWriter(fileName);
      out = new PrintWriter(new BufferedWriter(outStream));
    } catch (IOException exc) {
      System.err.println("Error: " + exc.getMessage());
      return null;
    }

    // and visit children
    data = node.childrenAccept(this, data);

    // close stream and reset handle
    close();

    return data;
  }

  

  /**
   * Visits an IncludeFileList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited IncludeFileList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTIncludeFileList node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  

  /**
   * Visits an IncludeFileDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit include
   * </ul>
   *
   * @param node the visited IncludeFileDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTIncludeFileDeclaration node, Object data) 
  {
    emitInclude(node.getName());
    return data;
  }

  

  /**
   * Visits an ADL_AnnotatedDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit annotated declaration
   * </ul>
   *
   * @param node the visited ADL_AnnotatedDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_AnnotatedDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_AnnotatedDeclaration");

    ADLClass adlModule = (ADLClass)node.getScope();
    
    String fullName = adlModule.getFullName();

    String modif = adlModule.getModifier().toString();

    // TBD check for two last parameters
    emitClassHeader(fullName, modif, null, null);

    emitClassEntity(node.getScope());

    int nbChildren = node.jjtGetNumChildren();
    for (int i = 0; i < nbChildren; i++) {
      if (node.jjtGetChild(i) instanceof ASTADL_BehaviorDeclaration)
        node.jjtGetChild(i).jjtAccept(this, data);
    }
    return data;
  }

  

  /**
   * Visits an ADL_InlineDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>not relevant
   * </ul>
   *
   * @param node the visited ADL_InlineDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_InlineDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_InlineDeclaration");

    return data;
  }

  // TBD - improve code for following methods in java version :
  //       manipulate function symbol instead of ast nodes!

  

  /**
   * Visits an ADL_BehaviorDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit inforamation for behavior declaration signature
   * <li>visit behavior specification
   * </ul>
   *
   * @param node the visited ADL_BehaviorDeclaration
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_BehaviorDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_BehaviorDeclaration");
    
    LCallable adlFunction = (LCallable)node.getScope();

    String methodName = adlFunction.getName();
    String resultType = adlFunction.getReturnType().getName();
    LParameter[] parameters = adlFunction.getParameters();

    beginMethodTemplate(methodName);

    emitMethodEntity(methodName, node.getScope());

    emitResultType(resultType);

    if ( parameters.length == 0 && ! adlFunction.lastParameterIsEllipsis() ) {
      emitParam("","");
    }
    else for (int i = 0; i < parameters.length; i++) {
      emitParam(parameters[i].getType().getName(), parameters[i].getName());
    }

    if ( adlFunction.lastParameterIsEllipsis() ) {
      emitParam("","...");
    }

    endPrototype();

    Node child = null;
    int index = 0;
    do {
      child = node.jjtGetChild(index++);
    } while (! ( child instanceof ASTADL_BehaviorSpecification));

    child.jjtAccept(this, data);

    endMethodTemplate();

    return data;
  }

  

  /**
   * Visits an ADL_Prolog.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>should not be visited ("pure" imperative Java)
   * </ul>
   *
   * @param node the visited ADL_Prolog
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_Prolog node, Object data) 
  {
    throw new Error("should not be visited");
  }

  

  /**
   * Visits an ADL_Epilog.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>should not be visited ("pure" imperative Java)
   * </ul>
   *
   * @param node the visited ADL_Epilog
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_Epilog node, Object data) 
  {
    throw new Error("should not be visited");
  }

  

  /**
   * Visits an ADL_BehaviorSpecification.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_BehaviorSpecification
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_BehaviorSpecification node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_BehaviorSpecification");
    return node.childrenAccept(this, data);
  }

  

  /**
   * Visits an ADL_BehaviorClassification.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_BehaviorClassification
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_BehaviorClassification node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  

  /**
   * Visits an ADL_NormalBehavior.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit normal 
   * </ul>
   *
   * @param node the visited ADL_NormalBehavior
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_NormalBehavior node, Object data) 
  {
    inAssertion = true;
    String assertion = "'ADL#Normal' = "
                     + node.jjtGetChild(0).jjtAccept(this, data);
    inAssertion = false;
    emitNormal(query(assertion, node.getScope()), true);
    return data;
  }

  

  /**
   * Visits an ADL_AbnormalBehavior.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit normal 
   * </ul>
   *
   * @param node the visited ADL_AbnormalBehavior
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_AbnormalBehavior node, Object data)
  {
    inAssertion = true;
    String assertion = "'ADL#Abnormal' = "
                     + node.jjtGetChild(0).jjtAccept(this, data);
    inAssertion = false;
    emitNormal(query(assertion, node.getScope()), false);
    return data;
  }

  

  /**
   * Visits an ADL_AssertionGroup.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_AssertionGroup
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_AssertionGroup node, Object data) 
  {
    if ( debug )
      System.out.println("visit ASTADL_AssertionGroup");

    if ( inAssertion ) assertionGroupLevel++;

    String ret = "";
    if ( isTopLevelAssertion() ) 
    {
      ret += join(node,0," ");
    }
    else 
    {
      ret += "'ADL#Group'([" + join(node,0,", ") + "])";
    }

    if ( inAssertion ) assertionGroupLevel--;

    return ret;
  }

  

  /**
   * Visits an ADL_Binding.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>not relevant
   * </ul>
   *
   * @param node the visited ADL_Binding
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_Binding node, Object data) 
  {
    return data;
  }

  /**
   * Visits an ADL_NamedParamList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>
   * </ul>
   *
   * @param node the visited ADL_NamedParamList
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_NamedParamList node, Object data) 
  {
    // TBD
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ADL_NamedParam.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_NamedParam
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_NamedParam node, Object data) 
  {
    // TBD
    return node.childrenAccept(this, data);
  }

  

  /**
   * Visits an ADL_Statement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_Statement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_Statement node, Object data) 
  {
    if ( debug )
      System.out.println("visit ASTADL_Statement");

    return join(node,0,"");
  }

  

  /**
   * Visits an ADL_IfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit condition
   * <li>visit "if" statement
   * <li>visit "else" statement
   * </ul>
   *
   * @param node the visited ADL_IfStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_IfStatement node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_IfStatement");
    
    if ( isTopLevelAssertion() )
    {
      // visit the condition
      inAssertion = true;
      String assertion = (String)node.jjtGetChild(0).jjtAccept(this, data); 
      inAssertion = false;

      emitCondition(query(assertion, node.getScope()),
		    ((SimpleNode)node.jjtGetChild(0)).getExpressionImage(),
		    node.jjtGetNumChildren() > 2);

      // visit "if" statement
      String ifAssertion = (String) node.jjtGetChild(1).jjtAccept(this, data); 

      // if required visit "else" statement
      if (node.jjtGetNumChildren() > 2) 
      {
	emitElse();
	String elseAssertion = (String)
	                       node.jjtGetChild(2).jjtAccept(this, data);
      }

      emitEndCondition();

    }
    else // nested assertion
    {
      String ret = (String) node.jjtGetChild(0).jjtAccept(this, data)
                 + " '?' "
	         + (String) node.jjtGetChild(1).jjtAccept(this, data)
                 + " ':' "
                 + ( (node.jjtGetNumChildren() == 3) 
		   ? (String) node.jjtGetChild(2).jjtAccept(this, data)
		   : "true"
		   );
      data = ret;
    }
    return data;
  }

  

  /**
   * Visits an ADL_Assertion.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>
   * </ul>
   *
   * @param node the visited ADL_Assertion
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_Assertion node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_Assertion");

    if ( isTopLevelAssertion() )
    {
      String label = "";
      if (node.jjtGetChild(0) instanceof ASTADL_Labels) 
      {
	label += node.jjtGetChild(0).jjtAccept(this, data);
      }
      Node lastChild = node.jjtGetChild(node.jjtGetNumChildren()-1);
      if (lastChild instanceof ASTADL_QuantifiedAssertion) 
      {
        lastChild.jjtAccept(this, label);
      }
      else // expression
      {
	inAssertion = true;
        exceptionBhv = false;
	String assertion = (String)lastChild.jjtAccept(this, data);
	inAssertion = false;
	emitAssertion(query(assertion, node.getScope()),
                      label,
		      node.getExpressionImage(), exceptionBhv);
      }
    }
    else // nested assertion
    {
      data = node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept
	                                                  (this, data);
    }
    return data;
  }

  

  /**
   * Visits an ADL_Labels.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return label
   * </ul>
   *
   * @param node the visited ADL_Labels
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_Labels node, Object data) 
  {
    return node.getName();
  }

  

  /**
   * Visits an ADL_Tags.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return list of tags separated by ','
   * </ul>
   *
   * @param node the visited ADL_Tags
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_Tags node, Object data) 
  {
    String ret;
    try { 
      String[] tags = (String[])node.getProperty("taglist");
      ret = join(tags,", ");
    }
    catch ( PropertyAccessException e ) { ret = ""; } 
    return ret;
  }

  

  /**
   * Visits an ADL_QuantifiedAssertion.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit quantified assertion
   * <li>visit  assertion group
   * </ul>
   *
   * @param node the visited ADL_QuantifiedAssertion
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_QuantifiedAssertion node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_QuantifiedAssertion.");

    String label = (String)data;
    ASTADL_Domain dom;
    ASTADL_DomainList domlist = (ASTADL_DomainList)node.jjtGetChild(1);

    emitQuantifier(label, 
		   ((ASTADL_Quantifier)node.jjtGetChild(0)).getName(),
		   domlist.getExpressionImage());
    String name, expr;
    for (int i = 0; i < domlist.jjtGetNumChildren(); i++) 
    {
      dom = (ASTADL_Domain)domlist.jjtGetChild(i);
      inAssertion = true; 
      String assertion = (String)dom.jjtGetChild(1).jjtAccept(this, null);
      expr = query(assertion, node.getScope());
      inAssertion = false; 
      emitDomain(dom.getName(),
                 expr,
		 i);
    }
    String quantifiedAssertion = (String)
                                 node.jjtGetChild(2).jjtAccept(this, null);
    emitEndQuantified();

    return data;
  }

  

  /**
   * Visits an ADL_Quantifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_Quantifier
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_Quantifier node, Object data) 
  {
    return node.getName();
  }

  

  /**
   * Visits an ADL_DomainList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_DomainList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_DomainList node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  

  /**
   * Visits an ADL_Domain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>join children
   * </ul>
   *
   * @param node the visited ADL_Domain
   * @param data unused
   * @return adl domain string
   */

  public Object visit(ASTADL_Domain node, Object data) 
  {
    return join(node,0," ':' ");
  }

  

  /**
   * Visits an ADL_DomainExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_Domain
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTADL_DomainExpression node, Object data) 
  {
    // get range function
    LFunction range = (LFunction)node.getScope();
    return "'ADL#Func'(" 
           +  "['" + range.getName() + range.getNLSignature() + "', "
           + "[" + nlengine.NLScopeScopeList(range.getParent()) + "]"
           + "], "
           + "[" + join(node,0,", ") + "]" 
           + ")";

  }

  

  /**
   * Visits an ADL_ImplExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return children visit strings
   * </ul>
   *
   * @param node the visited ADL_ImplExpression
   * @param data unused
   * @return the imple expression string
   */

  public Object visit(ASTADL_ImplExpression node, Object data) 
  {
    if ( debug )
      System.out.println("visit ASTADL_ImplExpression");

    return join(node,0,"");
  }

  
  /**
   * Visits an ADL_ImplOp.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ADL_ImplOp
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTADL_ImplOp node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_ImplOp");

    String ret = null; 
    try 
    { 
      String op = (String)node.getProperty("op"); 
      ret = " '" + op + "' ";
      if (op.equals("<:>")) 
      {
        exceptionBhv = true;
      }
    }
    catch (PropertyAccessException e) {}
    return ret;
  }

  

  /**
   * Visits an ADL_CallStateExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return '@' followed by expression string
   * </ul>
   *
   * @param node the visited ADL_CallStateExpression
   * @param data unused
   * @return the expression string
   */

  public Object visit(ASTADL_CallStateExpression node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_CallStateExpression");

    return "'@' " + node.jjtGetChild(0).jjtAccept(this, data);
  }

  

  /**
   * Visits an ADL_Normal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return normal expression
   * </ul>
   *
   * @param node the visited ADL_Normal
   * @param data unused
   * @return the expression string
   */

  public Object visit(ASTADL_Normal node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_Normal");

    return "'ADL#Normal'";
  }

  

  /**
   * Visits an ADL_Abnormal.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return abnormal expression
   * </ul>
   *
   * @param node the visited ADL_Abnormal
   * @param data unused
   * @return the expression string
   */
  public Object visit(ASTADL_Abnormal node, Object data) 
  {
    return "'ADL#Abnormal'";
  }

  
  /**
   * Visits an ADL_Unchanged.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return unchanged expression
   * </ul>
   *
   * @param node the visited ADL_Unchanged
   * @param data unused
   * @return the expression string
   */

  public Object visit(ASTADL_Unchanged node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTADL_Unchanged");

    return "'ADL#Unchanged'"
           + "([" + node.jjtGetChild(0).jjtAccept(this, data) + "])";
  }

  

  /**
   * Visits an ADL_ArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>,</code>
   * </ul>
   *
   * @param node the visited ADL_ArgumentList
   * @param data unused
   * @return the argument list string
   */
  public Object visit(ASTADL_ArgumentList node, Object data) 
  {
    return join(node,0,", ");
  }


  

  /**
   * Visits an Name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit name, with suitable additional information
   * </ul>
   *
   * @param node the visited Name
   * @param data unused
   * @return the name string
   */

  public Object visit(ASTName node, Object data) 
  {
    if (debug) {
      System.out.println("visit ASTName: " + node.getName());
    }

    String ret = null;

    if (inAssertion) 
    {
      ret = "'" + node.getName() + "'";
    }
    else if (inTddDef) 
    {
      try 
      {
        String cstClass = (String)node.getProperty("enclosingConstructName");
        if (cstClass.equals(moduleName)) 
	{
          ret = makeLocalLink
	        (moduleName, "TDD-CSTS", node.getName());
        }
        else 
	{
          ret = makeTDDExternalLink
	        (cstClass, cstClass + "TDD-CSTS", node.getName());
        }
      } 
      catch(PropertyAccessException e)
      {
        ret = node.getName();
      }
    }
    return ret;
  }

  

  /**
   * Visits an ExternalDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ExternalDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTExternalDeclaration node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an FunctionDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited FunctionDefinition
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTFunctionDefinition node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an Declaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited Declaration
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDeclaration node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an DeclarationSpecifiers.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited DeclarationSpecifiers
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDeclarationSpecifiers node, Object data) 
  {
    return join(node,0," ");
  }

  /**
   * Visits an TypeQualifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return qualifier image
   * </ul>
   *
   * @param node the visited TypeQualifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTTypeQualifier node, Object data) 
  {
    return node.getName();
  }

  /**
   * Visits an StorageStructSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return the specifier image
   * </ul>
   *
   * @param node the visited StorageStructSpecifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTStorageStructSpecifier node, Object data) 
  {
    return node.getName();
  }

  /**
   * Visits an BuiltinTypeSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return specifier image
   * </ul>
   *
   * @param node the visited BuiltinTypeSpecifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBuiltinTypeSpecifier node, Object data) 
  {
    return node.getName();
  }

  /**
   * Visits an InitDeclaratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited InitDeclaratorList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTInitDeclaratorList node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an InitDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited InitDeclarator
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTInitDeclarator node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an StructSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set up a string buffer
   * <li>append struct specifier kind (struct or union) and optional tag
   * <li>if required appen struct members inside braces
   * </ul>
   *
   * @param node the visited StructSpecifier
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTStructSpecifier node, Object data) {
    String ret = node.getName();
    // if there are members, append them inside braces
    if ( node.jjtGetNumChildren() > 0 ) {
      ret += " { " + join(node,0," ") + " } ";
    }
    return ret;
  }

  /**
   * Visits an MemberDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>join strings obtained from children visit
   * <li>if member declaration is not a simple declaration, append ";"
   * </ul>
   *
   * @param node the visited MemberDeclaration
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTMemberDeclaration node, Object data) 
  {
    // join strings obtained from children visit
    String ret = join(node,0," ");

    // if member declaration is not a simple declaration, append ";"
    if ( node.jjtGetNumChildren() == 0 ) {
      ret += ";";
    }
    else if ( ! ( node.jjtGetChild(0) instanceof ASTDeclaration ) ) {
      ret += ";";
    }
    return ret;
  }

  /**
   * Visits an MemberDeclaratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited MemberDeclaratorList
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTMemberDeclaratorList node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an EnumSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set up a string buffer
   * <li>append 'enum' and optional tag
   * <li>if required append members inside braces
   * <li>return resulting string
   * </ul>
   *
   * @param node the visited EnumSpecifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTEnumSpecifier node, Object data) 
  {
    String ret = node.getName();
    // if there are members, append them inside braces
    if ( node.jjtGetNumChildren() > 0 ) {
      ret += " { " + join(node,0," ") + " } ";
    }
    return ret;
  }

  /**
   * Visits an EnumeratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>returned string obtained from child visit serparated by ','
   * </ul>
   *
   * @param node the visited EnumeratorList
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTEnumeratorList node, Object data) 
  {
    return join(node,0,", ");
  }

  /**
   * Visits an Enumerator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return enumarator name
   * </ul>
   *
   * @param node the visited Enumerator
   * @param data unused
   * @return the enumarator name
   */
  public Object visit(ASTEnumerator node, Object data) 
  {
    return node.getName();
  }

  /**
   * Visits an PtrOperator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return pointer image, that is "*" followed by the concatenation 
   *     of strings obtained from child visit
   * </ul>
   *
   * @param node the visited PtrOperator
   * @param data unused
   * @return pointer image
   */
  public Object visit(ASTPtrOperator node, Object data) 
  {
    return "*" + " " + join(node,0," ");
  }

  /**
   * Visits an CvQualifierSeq.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return qualifier image
   * </ul>
   *
   * @param node the visited CvQualifierSeq
   * @param data unused
   * @return the qualifier image
   */
  public Object visit(ASTCvQualifierSeq node, Object data) 
  {
    return node.getName();
  }

  /**
   * Visits an Declarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>join  strings obtain from children's visit with a <code>blank</code> 
   *     separator 
   * </ul>
   *
   * @param node the visited Declarator
   * @param data unused
   * @return declarator image
   */
  public Object visit(ASTDeclarator node, Object data) 
  {
    return join(node,0," ");
  }

  /**
   * Visits an DirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if DirectDeclarator's first child is a nested Declarator
   *     emit it inside parenthesis, otherwise simply emit name
   * <li>visit DeclaratorSuffixes
   * </ul>
   *
   * @param node the visited DirectDeclarator
   * @param data unused
   * @return the direct declarator image
   */
  public Object visit(ASTDirectDeclarator node, Object data) 
  {
    String ret = "";

    boolean nestedDeclarator = node.jjtGetChild(0) instanceof ASTDeclarator;

     if ( nestedDeclarator ) {
        ret += "( ";
     }

     ret += node.jjtGetChild(0).jjtAccept(this, data);     

     if (  nestedDeclarator ) {
        ret += " )";
     }

     // visit DeclaratorSuffixes
     ret += join(node,1,"");

     return ret;
  }

  /**
   * Visits an DeclaratorSuffixes.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>in case suffix is bracket suffixes, return brackets and subscripts
   * <li>in case suffix is a parameter list, return parameters inside 
   *     parenthesis 
   * <li>in case suffix is a bit field : return ":" before constant
   * </ul>
   *
   * @param node the visited DeclaratorSuffixes
   * @param data unused
   * @return declarator suffixes image
   */
  public Object visit(ASTDeclaratorSuffixes node, Object data) 
  {
    String ret = "";
    if ( node.getName().equals("[]") )
      {
        // suffix bracket suffixes, unparse brackets and subscripts
        try
        {
           ret += "[";
           SimpleNode[] subscripts = (SimpleNode[])node.getProperty
                                                        ("subscripts");
           ret += join(subscripts,"][");        
           ret += "]";
        }
        catch ( PropertyAccessException e ) { }
     }
     else if ( node.getName().equals("()") )
     {
        // parameter list, unparse inside parenthesis
        ret += "(";
        ret += node.jjtGetChild(0).jjtAccept(this, data);
        ret += ")";
     }
     else if ( node.getName().equals(":") )
     {
        // bit field : unparse ":" before constant
        ret += ": ";
        ret += node.jjtGetChild(0).jjtAccept(this, data);
     }
     return ret;
  }

  /**
   * Visits an FunctionDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited FunctionDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTFunctionDeclarator node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an FunctionDirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited FunctionDirectDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTFunctionDirectDeclarator node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ParameterList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> return parameter declaration list
   * <li> optional final ellipsis can be concatened
   * </ul>
   *
   * @param node the visited ParameterList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTParameterList node, Object data) 
  {
    String ret = "";
    try 
    {
      // append parameter declaration list
      ret += join(node,0,", ");
      Boolean lastParameterIsEllipsis = (Boolean)node.getProperty
                                        ("lastParameterIsEllipsis");
      if ( Boolean.TRUE.equals(lastParameterIsEllipsis) )
      {
	// if ellipsis is not only parameter emit separator
	if ( ret.length() > 0 ) ret += ", ";
	// emit ellipsis
	ret += "...";
      }
    }
    catch ( PropertyAccessException e ) { }
     
    return ret;
  }

  /**
   * Visits an ParameterDeclarationList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>returns children strings separated by blanks
   * </ul>
   *
   * @param node the visited ParameterDeclarationList
   * @param data unused
   * @return the declaration list string
   */
  public Object visit(ASTParameterDeclarationList node, Object data) 
  {
    return join(node,0," ");
  }

  /**
   * Visits an ParameterDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return parameters separated by a ","
   * </ul>
   *
   * @param node the visited ParameterDeclaration
   * @param data unused
   * @return the parameter declaration
   */
  public Object visit(ASTParameterDeclaration node, Object data) 
  {
    return join(node,0,", ");
  }

  /**
   * Visits an Initializer.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited Initializer
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTInitializer node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an TypeName.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>returns children strings separated by blanks
   * </ul>
   *
   * @param node the visited TypeName
   * @param data unused
   * @return the type name
   */

  public Object visit(ASTTypeName node, Object data) 
  {
    return join(node,0,", ");
  }

  /**
   * Visits an AbstractDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for pointer abstract declarator, return children separated by blanks
   * <li>for nested abstract declarator and suffixes, return nested abstract
   *     declarator inside parenthesis and then suffixes
   * <li>for bracket suffixes, return brackets and subscripts
   * </ul>
   *
   * @param node the visited AbstractDeclarator
   * @param data unused
   * @return the abstract declarator image
   */

  public Object visit(ASTAbstractDeclarator node, Object data) 
  {
    String ret = "";
    if ( node.getName().equals("") )
    {
      // pointer abstract declarator, emit children separated by blanks
      ret += join(node,0," ");
    }
    else if ( node.getName().equals("()") )
    {
      // nested abstract declarator and suffixes, unparse nested abstract
      // declarator inside parenthesis and then suffixes
      ret += "(";
      ret += node.jjtGetChild(0).jjtAccept(this, data);
      ret += ")";
      ret += join(node,1," ");
    }
    else if ( node.getName().equals("[]") )
    {
      // suffix bracket suffixes, unparse brackets and subscripts
      try
      {
	ret += "[";
	SimpleNode[] subscripts = (SimpleNode[])node.getProperty
                                                     ("subscripts");
	ret += join(subscripts, "][");
        ret += "]";
      }
      catch ( PropertyAccessException e ) { }
    }
    return ret;
  }

  /**
   * Visits an AbstractDeclaratorSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for bracket suffixe, return brackets and subscript
   * <li>for parameter list, return inside parenthesis
   * </ul>
   *
   * @param node the visited AbstractDeclaratorSuffix
   * @param data unused
   * @return the suffix image
   */

  public Object visit(ASTAbstractDeclaratorSuffix node, Object data) 
  {
    String ret = "";
    if ( node.getName().equals("[]") )
    {
      // bracket suffixe, unparse brackets and subscript
      ret += "[";
      if ( node.jjtGetNumChildren() > 0 )
      {
	ret += node.jjtGetChild(0).jjtAccept(this, data);
      }
      ret += "]";
    }
    else if ( node.getName().equals("()") )
    {
      // parameter list, unparse inside parenthesis
      ret += "(";
      if ( node.jjtGetNumChildren() > 0 )
      {
	ret += node.jjtGetChild(0).jjtAccept(this, data);
      }
      ret += ")";
    }
    return ret;
  }

  // TBD - add throw new Error("Should not be visited!);
  //       in statements

  /**
   * Visits a Statement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited Statement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an LabeledStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited LabeledStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTLabeledStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an CaseLabelStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited CaseLabelStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTCaseLabelStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an DefaultLabelStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited DefaultLabelStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDefaultLabelStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an CompoundStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited CompoundStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTCompoundStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an IfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited IfStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTIfStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an SwitchStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited SwitchStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTSwitchStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an WhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited WhileStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTWhileStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an DoWhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited DoWhileStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDoWhileStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ForStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTForStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an GotoStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited GotoStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTGotoStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ContinueStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ContinueStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTContinueStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an BreakStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited BreakStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBreakStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ReturnStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited ReturnStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTReturnStatement node, Object data) 
  {
    return node.childrenAccept(this, data);
  }

  
  /**
   * Visits an Expression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return concatenation of children expression image
   * </ul>
   *
   * @param node the visited Expression
   * @param data unused
   * @return concatenation of children expression image
   */
  public Object visit(ASTExpression node, Object data) 
  {
    if (debug)
      System.out.println("Visiting ASTExpression");
    return join(node,0,", ");
  }

  
  /**
   * Visits an AssignmentExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from child visits separated by suitable
   *     assignment operator
   * </ul>
   *
   * @param node the visited AssignmentExpression
   * @param data unused
   * @return the assignment string
   */

  public Object visit(ASTAssignmentExpression node, Object data) 
  {
    // get array of operator
    String operator = null;
    try 
    {
      operator = (String)node.getProperty("operator");
    } 
    catch (PropertyAccessException e) {}

    return join(node,0,operator);
  }

  
  /**
   * Visits an ConditionalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return concatenation of string obtained from children 
   *     visit with suitable separators
   * </ul>
   *
   * @param node the visited ConditionalExpression
   * @param data unused
   * @return concatenation of string obtained from children visit
   */
  public Object visit(ASTConditionalExpression node, Object data) 
  {
    String[] separators = { " '?' " , " ':' " };
    return join(node,0,separators);
  }

  

  /**
   * Visits an ConstantExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from children visit
   * </ul>
   *
   * @param node the visited ConstantExpression
   * @param data unused
   * @return string obtained from children visit
   */

  public Object visit(ASTConstantExpression node, Object data) 
  {
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  

  /**
   * Visits an LogicalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited LogicalOrExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTLogicalOrExpression node, Object data) 
  {
    return join(node, 0, quoteIfRequired("||"));
  }

  

  /**
   * Visits an LogicalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited LogicalAndExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTLogicalAndExpression node, Object data) 
  {
    return join(node, 0, quoteIfRequired("&&"));
  }

  
  /**
   * Visits an InclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited InclusiveOrExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTInclusiveOrExpression node, Object data) 
  {
    return join(node, 0, quoteIfRequired("|"));
  }

  
  /**
   * Visits an ExclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited ExclusiveOrExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTExclusiveOrExpression node, Object data) 
  {
    return join(node, 0, quoteIfRequired("^"));
  }

  
  /**
   * Visits an AndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited AndExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTAndExpression node, Object data) 
  {
    return join(node, 0, quoteIfRequired("&"));
  }

  
  /**
   * Visits an EqualityExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited EqualityExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTEqualityExpression node, Object data) 
  {
    // get array of operators
    String operators[] = null;
    try 
    {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    return join(node,0,quoteIfRequired(operators));
  }

  

  /**
   * Visits an RelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited RelationalExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTRelationalExpression node, Object data) 
  {
    // get array of operators
    String operators[] = null;
    try 
    {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    return join(node, 0,quoteIfRequired(operators));
  }

  

  /**
   * Visits an ShiftExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited ShiftExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */

  public Object visit(ASTShiftExpression node, Object data) 
  {
    // get array of operators
    String operators[] = null;
    try 
    {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    return join(node, 0,quoteIfRequired(operators));
  }

  

  /**
   * Visits an AdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited AdditiveExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */

  public Object visit(ASTAdditiveExpression node, Object data) 
  {
    // get array of operators
    String operators[] = null;
    try 
    {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    return join(node, 0,quoteIfRequired(operators));
  }

  
  /**
   * Visits an MultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited MultiplicativeExpression
   * @param data unused
   * @return string obtained from unparsing children with suitable separator 
   */
  public Object visit(ASTMultiplicativeExpression node, Object data) 
  {
    // get array of operators
    String operators[] = null;
    try 
    {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    return join(node, 0,quoteIfRequired(operators));
  }

  // TBD - check this
  /**
   * Visits an CastExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing first child inside 
   *     parenthesis and second child
   * </ul>
   *
   * @param node the visited CastExpression
   * @param data unused
   * @return unparsed string
   */

  public Object visit(ASTCastExpression node, Object data) 
  {
    return   "(" 
           + node.jjtGetChild(0).jjtAccept(this, data) 
           + ") "
           + node.jjtGetChild(1).jjtAccept(this, data);
  }

  // TBD - check this
  /**
   * Visits an PreIncrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable operator
   * </ul>
   *
   * @param node the visited PreIncrementExpression
   * @param data unused
   * @return unparsed string
   */
  public Object visit(ASTPreIncrementExpression node, Object data) 
  {
    return quoteIfRequired("++") + node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits an PreDecrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable operator
   * </ul>
   *
   * @param node the visited PreDecrementExpression
   * @param data unused
   * @return unparsed string
   */
  public Object visit(ASTPreDecrementExpression node, Object data) 
  {
    return quoteIfRequired("--") + node.jjtGetChild(0).jjtAccept(this, data);
  }

  
  /**
   * Visits an UnaryOperatorExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from unparsing children 
   *     with suitable separator 
   * </ul>
   *
   * @param node the visited UnaryOperatorExpression
   * @param data unused
   * @return unparsing string
   */
  public Object visit(ASTUnaryOperatorExpression node, Object data) 
  {
    String ret = "";
    try
    {
      ret = quoteIfRequired((String)node.getProperty("operator"))
	  + node.jjtGetChild(0).jjtAccept(this, data);
    }
    catch (PropertyAccessException e) {}
    return ret;
  }

  // TBD - check this

  /**
   * Visits an SizeOfExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return <code>sizeof</code> and child expression in parenthesis
   * </ul>
   *
   * @param node the visited SizeOfExpression
   * @param data unused
   * @return unparsed string
   */
  public Object visit(ASTSizeOfExpression node, Object data) 
  {
    return " 'sizeof' " + node.jjtGetChild(0).jjtAccept(this, data);
  }
   
  // TBD check for inTdddef and final childrenAccept

  /**
   * Visits an PostfixExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get unparsed string from prefix child (first child)
   * <li>visit suffix children , each visit in charge of updating 
   *     above string
   * <li>return resulting string
   * </ul>
   *
   * @param node the visited PostfixExpression
   * @param data unused
   * @return unparsed string.
   */

  public Object visit(ASTPostfixExpression node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTPostfixExpression");

    // special processing for test functions
    if ( inTddDef )
    {
      try 
      {
	Scope exprScope = (Scope)((SimpleNode)node.jjtGetChild(0))
	                         .getProperty("NLDScope");
	if (    (exprScope instanceof LMethod) 
	     && (exprScope.getParent() instanceof TDDClass)) 
	{
	  emitTestFunction(((LMethod)exprScope).getName(),
			   ((LMethod)exprScope).getNLSignature());
	}
      }
      catch(PropertyAccessException e) {}
      return null;
    }

    // otherwise normal processing

    // initialize string to be return
    String ret = (String)node.jjtGetChild(0).jjtAccept(this, data);

    // update string to be returned
    for ( int i = 1; i < node.jjtGetNumChildren(); i++ )
    {
      SimpleNode child = (SimpleNode)node.jjtGetChild(i);

      // normal visit
      ret = (String)child.jjtAccept(this, ret);
    }

    return ret;
  }

  /**
   * Visits an ArraySuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>'ADL#Array' and arguments in case an assertion is visited
   * <li>normal suffix unparsing otherwise
   * </ul>
   *
   * @param node the visited ArraySuffix
   * @param data string containing the result of unparsing of the 
   *             primary expr till this suffix
   * @return updated unparsed string
   */
  public Object visit(ASTArraySuffix node, Object data) 
  {
    String prefixAndFormerSuffixes = (String)data;
    String ret = "";
    if ( inAssertion )
      return   "'ADL#Array'"
             + "(" + node.jjtGetChild(0).jjtAccept(this,null) 
	     + "," + prefixAndFormerSuffixes 
             + ")";
    else
      return   prefixAndFormerSuffixes 
             + "[" +  node.jjtGetChild(0).jjtAccept(this,null) + "]";
  }

  /**
   * Visits an ArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>'ADL#Func' and arguments in case an assertion is visited
   * <li>normal suffix unparsing otherwise
   * </ul>
   *
   * @param node the visited ArgumentList
   * @param data string containing the result of unparsing of the 
   *             primary expr till this suffix
   * @return the updated unparsed string
   */

  public Object visit(ASTArgumentList node, Object data) 
  {
    String ret = "";
    String prefixAndFormerSuffixes = (String)data;

    if ( inAssertion )
    {
      try
      {
	LCallable methScope = (LCallable)node.getProperty("NLDScope");
	ret += "'ADL#Func'(" 
	    +  "['" + methScope.getName() + methScope.getNLSignature() + "', "
	        + "[" + nlengine.NLScopeScopeList(methScope.getParent()) + "]"
	     + "], "
	     + "[" + join(node,0,", ") + "]" 
	     + ")";
      }
      catch(PropertyAccessException e) {}
    }
    else // inTddDef
    {
      ret += prefixAndFormerSuffixes 
	  + "(";
      if ( node.jjtGetNumChildren() > 0 ) 
	ret += node.jjtGetChild(0).jjtAccept(this,null);
      ret += ")";
    }

    return ret;
  }
  
  /**
   * Visits an DotAccessSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>'ADL#Field' and arguments in case an assertion is visited
   * <li>normal suffix unparsing otherwise

   * </ul>
   *
   * @param node the visited DotAccessSuffix
   * @param data string containing the result of unparsing of the 
   *             primary expr till this suffix
   * @return updated unparsed string
   */

  public Object visit(ASTDotAccessSuffix node, Object data) 
  {
    String prefixAndFormerSuffixes = (String)data;
    String ret = "";
    if ( inAssertion )
    {
      ret += "'ADL#Field'('" 
          + ((ASTName)node.jjtGetChild(0)).getName() + "," 
          + prefixAndFormerSuffixes 
          + ")";
    }
    else
    {
      ret += prefixAndFormerSuffixes 
	  + "." 
          + ((ASTName)node.jjtGetChild(0)).getName();
    }
    return ret;
  }

  /**
   * Visits an RefAccessSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>no secial processing in case an assertion is visited
   * <li>normal suffix unparsing otherwise
   * </ul>
   *
   * @param node the visited RefAccessSuffix
   * @param data string containing the result of unparsing of the 
   *             primary expr till this suffix
   * @return the updated unparsed string
   */

  public Object visit(ASTRefAccessSuffix node, Object data) 
  {
    String prefixAndFormerSuffixes = (String)data;
    String ret = "";
    if ( inAssertion )
    {
      ret += prefixAndFormerSuffixes;
    }
    else
    {
      ret += prefixAndFormerSuffixes 
	  + "->" 
          + ((ASTName)node.jjtGetChild(0)).getName();
    }
    return ret;

  }

  /**
   * Visits an PostDeIncrement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>no secial processing in case an assertion is visited
   * <li>normal suffix unparsing otherwise
   * </ul>
   *
   * @param node the visited PostDeIncrement
   * @param data string containing the result of unparsing of the 
   *             primary expr till this suffix
   * @return the updated unparsed string
   */

  public Object visit(ASTPostDeIncrement node, Object data) 
  {
    String prefixAndFormerSuffixes = (String)data;
    String ret = "";
    try
    {
      String operator = (String)node.getProperty("operator");
      if ( inAssertion )
      {
	// TBD check this
	ret += prefixAndFormerSuffixes + operator;
      }
      else
      {
	ret += prefixAndFormerSuffixes + operator;
      }
    }
    catch ( PropertyAccessException e ) { }
    return ret;
  }

  /**
   * Visits an String.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string image in between simple quotes
   * </ul>
   *
   * @param node the visited String
   * @param data unused
   * @return the unparsed string
   */
  public Object visit(ASTString node, Object data) 
  {
    return "\'\"" + node.getName() + "\"\'";
  }
  
  /**
   * Visits an ADL_Return.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if visiting adl assertion return 'ADL#Return'
   * <li>otherwise return 'return'
   * </ul>
   *
   * @param node the visited ADL_Return
   * @param data unused
   * @return the unparsed string
   */
  public Object visit(ASTADL_Return node, Object data) 
  {
    return   inAssertion ? "'ADL#Return'" : "'return'";
  }
  
  /**
   * Visits an ParentheizedExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return children unparsed string in between parenthesis
   * </ul>
   *
   * @param node the visited ParentheizedExpression
   * @param data unused
   * @return the unparsed string
   */
  public Object visit(ASTParentheizedExpression node, Object data) 
  {
    return "(" + node.jjtGetChild(0).jjtAccept(this, data) + ")";
  }

  /**
   * Visits an UnaryPlusMinusConstant.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return signed constant
   * </ul>
   *
   * @param node the visited UnaryPlusMinusConstant
   * @param data unused
   * @return the unparsed string
   */
  public Object visit(ASTUnaryPlusMinusConstant node, Object data) 
  {
    String ret = "";
    try
    {
      String operator = (String)node.getProperty("operator");
      ret = operator + node.jjtGetChild(0).jjtAccept(this,data);
    }
    catch ( PropertyAccessException e ) { }
    return ret;
  }
  
  /**
   * Visits an Constant.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return constant value image
   * </ul>
   *
   * @param node the visited Constant
   * @param data unused
   * @return the constant value image
   */

  public Object visit(ASTConstant node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTConstant");
    
    String ret = node.getName();
    
    // Prolog does not recognize final letter ('L', 'D', etc.)
    if (    ! ret.equals("true") 
	 && ! ret.equals("false")
	 && Character.isLetter(ret.charAt(ret.length()-1)))
    {
      ret = ret.substring(0, ret.length()-1);
    }
    
    return ret;
  }

  /**
   * Visits an TDD_AnnotatedDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit class entity
   * <li>visit children
   * </ul>
   *
   * @param node the visited TDD_AnnotatedDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_AnnotatedDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTTDD_AnnotatedDeclaration");

    emitClassEntity(node.getScope());

    return node.childrenAccept(this, data);
  }

  /**
   * Visits an TDD_ImportDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit aco name
   * </ul>
   *
   * @param node the visited TDD_ImportDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_ImportDeclaration node, Object data) 
  {
    return data;
  }

  /**
   * Visits an TDD_UseDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit used class
   * </ul>
   *
   * @param node the visited TDD_UseDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_UseDeclaration node, Object data) 
  {

    emitUseClass(node.getName());
    return data;
  }

  /**
   * Visits an TDD_Declaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>join children strings and return resulting string
   * </ul>
   *
   * @param node the visited TDD_Declaration
   * @param data unused
   * @return the declaration string
   */

  public Object visit(ASTTDD_Declaration node, Object data) 
  {
    return join(node,0,"");
  }

  /**
   * Visits an TDD_FieldDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited TDD_FieldDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_FieldDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTTDD_FieldDeclaration");
    
    try
    {
      LExpression[] fields = (LExpression[])node.getProperty("fields");
      inTddDef = true;
      for (int i = 0; i < fields.length; i++)
      {
	// constant expression index is 2 ( i + 1 ) among node children 
	emitConstant(fields[i].getType().getName(),
		     fields[i].getName(),
		     (String)node.jjtGetChild(2 * i + 2).jjtAccept(this, data)
		    );
      }
      inTddDef = false;
    }
    catch ( PropertyAccessException e ) {}    
    return data;
  }

  /**
   * Visits an TDD_DatasetDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get dataset name, type 
   * <li>emit dataset
   * </ul>
   *
   * @param node the visited TDD_DatasetDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_DatasetDeclaration node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTTDD_DatasetDeclaration");

    try 
    {
      String expr;
      inTddDef = true;
      expr = (String)node.jjtGetChild(1).jjtAccept(this, data);
      inTddDef = false;

      String name = (String)node.getProperty("datasetName");
      LType  type = (LType)node.getProperty("datasetType");
      emitDataset(name, type.getName(), expr); 
    }
    catch ( PropertyAccessException e ) {}

    return data;

  }

  /**
   * Visits an TDD_FactoryDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit factory information
   * </ul>
   *
   * @param node the visited TDD_FactoryDefinition
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_FactoryDefinition node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTTDD_FactoryDeclaration");

    TDDFactory factory = (TDDFactory)node.getScope();

    String factoryName = factory.getName();
    String resultType = factory.getReturnType().getName();
    LParameter[] parameters = factory.getParameters();


    beginFactoryTemplate(factoryName);
    emitMethodEntity(factoryName, factory);
    emitResultType(resultType);

    if ( parameters.length == 0 )
    {
      emitParam("","");
    }
    else for (int i = 0; i < parameters.length; i++) 
    {
      emitParam(parameters[i].getType().getName(), parameters[i].getName());
    }

    emitCode(node.getSourceText());

    endFactoryTemplate();

    return data;
  }

  /**
   * Visits an TDD_TestDirective.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>set up a directive template
   * <li>visit children (second child to child before last)
   * <li>visit directive statement
   * </ul>
   *
   * @param node the visited TDD_TestDirective
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_TestDirective node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_TestDirective");

     String  dirName = node.getName();
     if ( dirName == null ) 
     {
        dirName = "directive_" + anonDirectiveIndex++;
     }

     beginDirectiveTemplate(dirName);

     // visit tdd dataset domains
     int nbChildren = node.jjtGetNumChildren();
     for(int i = 0; i < nbChildren-1; i++) 
     {
        node.jjtGetChild(i).jjtAccept(this, data);
     }

     emitCode(node.getSourceText());


     // visit directive statement
     inTddDef = true;
     node.jjtGetChild(nbChildren-1).jjtAccept(this, data);
     inTddDef = false;

     endDirectiveTemplate();

     return data;
  }

  /**
   * Visits an TDD_DatasetDomain.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit dataset expression
   * </ul>
   *
   * @param node the visited TDD_DatasetDomain
   * @param data unused
   * @return the unused data
  */
  public Object visit(ASTTDD_DatasetDomain node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_DatasetDomain");

     inTddDef = true;
     String ds = (String)node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept
                 (this, data);
     inTddDef = false;

     emitDomain(ds);

     return data;
  }

  
  /**
   * Visits an TDD_DatasetLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return children strings separated by "," inside braces
   * </ul>
   *
   * @param node the visited TDD_DatasetLiteral
   * @param data unused
   * @return the dataset literal string
   */
  public Object visit(ASTTDD_DatasetLiteral node, Object data) 
  {
    if (debug)
      System.out.println("visit ASTTDD_DatasetLiteral");

    String ret = "{ " + join(node,0,", ") + " }";
    
    return ret + " }";
  }

  /**
   * Visits an TDD_DatasetMember.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return children strings separated by ".."
   * </ul>
   *
   * @param node the visited TDD_DatasetMember
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTTDD_DatasetMember node, Object data) 
  {
    return join(node,0," .. ");
  }
  
  /**
   * Visits an TDD_DatasetExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return children strings separated by "+" 
   * </ul>
   *
   * @param node the visited TDD_DatasetExpression
   * @param data unused
   * @return the dataset expression string
   */
  public Object visit(ASTTDD_DatasetExpression node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_DatasetExpr");

     return join(node,0, " + ");
  }

  /**
   * Visits an TDD_DatasetConcatenationExpr.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return string obtained from children visit
   * </ul>
   *
   * @param node the visited TDD_DatasetConcatenationExpr
   * @param data unused
   * @return string obtained from children visit
   */
  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data) 
  {
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits an TDD_DatasetSingleton.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if expression refers to a dataset or a constant name 
   *     return a suitable tdd link 
   * <li>otherwise return string obtained from visiting children
   * </ul>
   *
   * @param node the visited TDD_DatasetSingleton
   * @param data unused
   * @return string for tdd link or constant
   */

  public Object visit(ASTTDD_DatasetSingleton node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_DatasetSingleton");

     String ret = null;

     try
     {
        SimpleNode child = (SimpleNode)node.jjtGetChild(0);

        if (child instanceof ASTName) 
        {
           String dsName = child.getName();
           String kind = (String)child.getProperty("astNameKind");
           String dsClass = (String)child.getProperty
                                          ("enclosingConstructName");

           if (kind.equals("TDDDataset")) 
           {
              if (dsClass.equals(moduleName)) 
              {
                 ret = makeLocalLink(dsName, dsName);
              }
              else 
              {
                 ret = makeTDDExternalLink(dsClass, dsName, dsName);
              }
           }
           else // constant
           { 
              if (dsClass.equals(moduleName)) 
              {
                 ret = makeLocalLink(moduleName, "TDD-CSTS", dsName);
              }
              else 
              {
                 ret = makeTDDExternalLink
                       (dsClass, dsClass + "TDD-CSTS", dsName);
              }
           }
        }
        else // child is a UnaryPlusMinusConstant or a Constant
        {
           ret = (String)child.jjtAccept(this, data);
        }
     } 
     catch(PropertyAccessException e) { }

     return ret;
  }

  /**
   * Visits an TDD_FactoryCall.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>return factory call image
   * </ul>
   *
   * @param node the visited TDD_FactoryCall
   * @param data unused
   * @return factory call image
   */
  public Object visit(ASTTDD_FactoryCall node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_FactoryCall");

     String ret = null;
     
     try 
     {
        String factName = node.getName(); 
        String factModule = (String)node.getProperty("enclosingConstructName");

        if (factModule.equals(moduleName)) 
        {
           ret = makeLocalLink(factName, factName);
        }
        else 
        {
           ret = makeTDDExternalLink(factModule, factName, factName);
        }
        ret += " ( ";
        ret += join(node,1,", ");
        ret += " )";
     } 
     catch(PropertyAccessException e) {}

     return ret;
  }

  /**
   * Visits an TDD_ADLExpression: emit the name of the ACO used in
   * this expression.
   *
   * @param node the visited TDD_ADLExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTDD_ADLExpression node, Object data) 
  {
     if (debug)
        System.out.println("visit ASTTDD_ADLExpression");
     
     // the grammar says:
     // TDD_ADLExpression ::= "ADL" "(" Name ArgumentList ")"
     // and creates a PostfixExpression to hold the Name and ArgumentList
     Node postfix = node.jjtGetChild(0); // PostfixExpression
     SimpleNode astName = (SimpleNode)postfix.jjtGetChild(0); // Name
    
     try {
       String acoName = (String)astName.getProperty("acoName");
       // we do have an acoName property set, emit that ACO name
       emitACO(acoName);
     } catch (PropertyAccessException e) {}

     return data;
  }

  // -- helper methods --

  /** Are we in a top level assertion? */
  public boolean isTopLevelAssertion()
  {
    return assertionGroupLevel <= 1;
  }
      
  /** Are we in a nested assertion? */
  public boolean isAssertionNested()
  {
    return assertionGroupLevel > 1;
  }

  /**
   * Returns a string obtained from the concatenation of some strings,
   * a separtor is inserted in between each of the strings.
   */

  private String join(String[] strings,String separator) 
  {
     String ret = "";
     for ( int i = 0 ; i < strings.length; i++ ) 
     {
        if ( i != 0 ) ret += separator;
        ret += strings[i];
     }
     return ret;
  }

  /**
   * Returns a string obtained from the concatenation of some strings.
   * Separators are picked up among supplied separators.
   */

  private String join(String[] strings,String[] separators) 
  {
     String ret = "";
     for ( int i = 0 ; i < strings.length; i++ )
     {
        if ( i != 0 ) ret += separators[i];
        ret += strings[i];
     }
     return ret;
  }

  /**
   * Return a string obtained by joining the strings resulting from unparsing
   * a node's children. The strings are separated by supplied separator.
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separator the string that separate children
   * @return the concatenation string.
   */

  private String join(Node node, 
                      int firstChild, 
                      String separator)
  {
    String ret = "";

    int numChildren = node.jjtGetNumChildren();

    for (int i = firstChild; i < numChildren; i++) 
    {
      if ( i != firstChild ) ret += separator;
      ret += node.jjtGetChild(i).jjtAccept(this, null);
    }
    return ret;
  }

  /**
   * Return a string obtained by joining the strings resulting from unparsing
   * a node's children. The strings are separated by supplied separators.
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separators strings that separate children
   * @return the concatenation string.
   */

  private String join(Node node, 
                      int firstChild, 
                      String[] separators)
  {
    String ret = "";

    int numChildren = node.jjtGetNumChildren();

    for (int i = firstChild; i< numChildren; i++) 
    {
      if ( i != firstChild ) ret += separators[i-firstChild-1];
      ret += node.jjtGetChild(i).jjtAccept(this, null);
    }
    return ret;
  }

  /**
   * Return a string obtained by joining the strings resulting from unparsing
   * some node. The strings are separated by supplied separator.
   *
   * @param nodes some nodes 
   * @param separator the string that separate children
   * @return the concatenation string.
   */

  private String join(SimpleNode[] nodes, String separator)
  {
    String ret = "";
    for (int i = 0; i < nodes.length; i++) 
    {
      if ( i != 0 ) 
         ret += separator;
      if ( nodes[i] != null )
         ret += nodes[i].jjtAccept(this, null);
    }
    return ret;
  }

  /**
   * Quote string if required. 
   * Useful for quoting operators in assertions. Note only non blank
   * strings are likely to be quoted.
   */

  private String quoteIfRequired(String string)
  {
    return inAssertion ? ( " '" + string + "' " )
                       : ( string ) ;
  }

  /**
   * Quote strings if required. 
   * Useful for quoting operators in assertions. Note only non blank
   * strings are likely to be quoted.
   */

  private String[] quoteIfRequired(String[] strings)
  {
    String[] ret = new String[strings.length];
    for (int i = 0; i < strings.length; i++) 
    {
      ret[i] = quoteIfRequired(strings[i]);
    }
    return ret;
  }

  

}
