// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/c/CUnparserVisitor.java,v $
// $Revision: 1.7 $
// $Date: 1998/07/29 14:24:38 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/C Compiler
// NAME:        CUnparserVisitor.java
//
// HISTORY:
//      Cyril Inard d'Argence, The Open Group RI - Apr 29, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.c;

import java.io.*;
import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlc.*;
import ORG.opengroup.adl.symboltable.LType;
import ORG.opengroup.adl.symboltable.c.CModifier;
import java.util.Vector;
import java.util.Enumeration ;

/** 
 * This class defines a visitor which can be applied to an AST
 * generated from the ADLCParser grammar. It produces pure C
 * source files by asserting that no ADL_*, TDD_* and NLD_* nodes
 * remain in the final AST. The only exception to this are the
 * NLD_Annotation nodes.
 *
 * <p>The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and emit the code on
 * an output stream maintained by the visitor itself.
 *
 * <p>The unparsing starts by calling the <a href="#unparse">unparse()</a>
 * method passing the top of AST to unparse as parameter.
 */

public class CUnparserVisitor extends GenericUnparserVisitor
implements ADLCParserVisitor
{
  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.7 $";

  /**
   * Default constructor: call superclass constructor
   *
   * @param toolname the name of the tool using this unparser visitor
   */   
  public CUnparserVisitor(String toolname) 
  {
    super(toolname, versionId);
  }  

  /**
   * Basic constructor specifying unparser parameters: call superclass
   * constructor.
   *
   * @param toolname the name of the tool using this unparser visitor
   * @param linelimit the desired maximum length of lines in output file
   * @param regindent the regular indentation level
   * @param subindent the sub-expression indentation level.
   */
  public CUnparserVisitor(String toolname,
                          int linelimit,
                          int regindent,
                          int subindent)
  {
     super(toolname, versionId, linelimit, regindent, subindent);
  }

  /**
   * Starts unparsing an AST.
   *
   * @param ast The root node of the AST to unparse
   * @see #visit(ORG.opengroup.adl.parsers.adlc.ASTTranslationUnit,java.lang.Object)
   */
  
  public void unparse(ASTTranslationUnit ast)
  {
     ast.jjtAccept(this, null);
  }

  /** Should not be visited! */

  public Object visit(SimpleNode node, Object data) 
  {
     incorrectNode(node);
     return data;
  }

  /**
   * Visits an TranslationUnit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the name of the module saved in the node
   * <li>open the destination file (<code>outputDirectory</code>/<em>name</em>
   * .h or c),
   * <li> reset current column and indentation,
   * <li> emit the standard header,
   * <li> visit children nodes,
   * <li> and finally close the file.
   * </ul>
   *
   * @param node the TranslationUnit to visit
   * @param data unused. Always null
   * @return data returned by last child visited or
   * null if the file cannot be opened.
   * @exception ORG.opengroup.adl.parsers.cpp.ParseException thrown when parse
   * error occur.
   * @see ORG.opengroup.adl.unparsers.GenericUnparserVisitor#outputDirectory
   */

  public Object visit(ASTTranslationUnit node, Object data)
  {
     try 
     {
        // get the name of the module and open the corresponding stream.
        // get file name stored as a node property.
        String moduleName = node.getName();
        String fileExtension = (String)node.getProperty("fileExtension");

        // try to open the output stream
        String fileName = outputDirectory + File.separator 
                        + moduleName
                        + "." + fileExtension;

        if (Adlt.verboseMode()) 
        {
           System.out.println("+++ generating " + fileName);
        }

        out = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));

        // now reset currentColumn and currentIndent
        currentColumn = 0;
        setIndent(0);

        // emit general comment
        emit(headerString);
        newline();
        newline();

        // if if not def lines.
        String defineName = moduleName+"_"+ fileExtension ;
        emit("#ifndef "+ defineName);
        newline();
        emit("#define "+defineName) ;
        newline() ;
        newline();

        // Visit children
        data = node.childrenAccept(this, data );
        newline() ;
        newline();

        // emit end if
        // close stream and reset handle
        emit("#endif ");
        newline() ;
        out.close();
        out = null;
     } 
     catch ( IOException exc ) 
     {
        System.err.println("Error: " + exc.getMessage());
        return null;
     }
     catch ( PropertyAccessException e ) {}
     
     return data;
  }

  
  /** Should not be visited! */
  
  public Object visit(ASTADL_TranslationUnit node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /**
   * Visits an IncludeFileList.
   * <p><b>Algorithm:</b><ul>
   * Simply viist children.
   * </ul>
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTIncludeFileList node, Object data)
  {
    data= node.childrenAccept( this , data );
    newline() ;
    return data ;
  }

  /**
   * Visits an IncludeFileDeclaration.
   *
   * <p> This node handles the included filenames
   * <p><b>Algorithm:</b><ul>
   * <li> emit the "# include" keywords
   * <li> emit the file name saved as a node property.
   * <li> emit " ;"
   * </ul>
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  
  public Object visit(ASTIncludeFileDeclaration node, Object data) 
  {
    // # include line, emit keyword , then emit filename ";"
    emit("#include ");
    emit( node.getName()  );
    newline();

    return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_AnnotatedDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_InlineDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Prolog node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Epilog node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorSpecification node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorClassification node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_NormalBehavior node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_AbnormalBehavior node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_AssertionGroup node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Binding node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_NamedParamList node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_NamedParam node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Statement node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_IfStatement node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Assertion node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Labels node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Tags node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_QuantifiedAssertion node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Quantifier node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_DomainList node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Domain node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */

  public Object visit(ASTADL_DomainExpression node, Object data) 
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_ImplExpression node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_ImplOp node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_CallStateExpression node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Normal node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Abnormal node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Unchanged node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_ArgumentList node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /**
   * Visits an Name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit name
   * </ul>
   *
   * @param node the visited Name
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTName node, Object data) 
  {
    indentIfRequired();
    emit( node.getName()  );
    return data;
  }

  /**
   * Visits an ExternalDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children</code>
   * <li>in case children do not make a FunctionDefinition or a Declaration
   *     emit a terminating ";"
   * </ul>
   *
   * @param node the visited ExternalDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTExternalDeclaration node, Object data) 
  {
     // visit children
     node.childrenAccept( this,data);

     // in case external declaration is not a FunctionDefinition nor 
     // a Declaration emit a terminating ";" 
     if ( node.jjtGetNumChildren() > 0 )
     {
        Node child = node.jjtGetChild(0);
        if (  ! ( child instanceof ASTFunctionDefinition )
           && ! ( child instanceof ASTDeclaration )
           )
        {
           emit(";");
        }
     }

     newline();
     newline();

     return data;
  }

  /**
   * Visits an FunctionDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children
   * <li>if function definition (it's just a declaration) has no body emit ";"
   * </ul>
   *
   * @param node the visited FunctionDefinition
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTFunctionDefinition node, Object data) 
  {
     indent();

     // visit children
     unparseList(node,0," ",false,data);

     // if function definition  has no body (it's just a declaration)
     // emit ";"
     if ( ! ( node.jjtGetChild(node.jjtGetNumChildren() - 1) 
            instanceof ASTCompoundStatement ) )
     {
        emit(";");
     }

     newline();
     newline();

     return data;
  }

  /**
   * Visits a Declaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children
   * <li>emit <code>;</code>
   * </ul>
   *
   * @param node the visited Declaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDeclaration node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     // visit children
     unparseList(node,0," ",false,data);
     emit(";");
     return data;
  }

  /**
   * Visits a DeclarationSpecifiers.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children
   * </ul>
   *
   * @param node the visited DeclarationSpecifiers
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDeclarationSpecifiers node, Object data) 
  {
     // visit children
     unparseList(node,0," ",false,data);
     return data;
  }

  /**
   * Visits a TypeQualifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> emit type qualifier
   * </ul>
   *
   * @param node the visited TypeQualifier
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTypeQualifier node, Object data) 
  {
     // emit type qualifier
     emit(node.getName());
     return data;
  }

  /**
   * Visits a StorageStructSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit storage struct specifier
   * </ul>
   *
   * @param node the visited StorageStructSpecifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTStorageStructSpecifier node, Object data) 
  {
     // emit storage struct specifier
     emit(node.getName());
     return data;
  }

  /**
   * Visits an BuiltinTypeSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit builtin type specifier
   * </ul>
   *
   * @param node the visited BuiltinTypeSpecifier
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBuiltinTypeSpecifier node, Object data) 
  {
     // emit builtin type specifier
     emit(node.getName());
     return data;
  }

  /**
   * Visits an InitDeclaratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children with a <code>,</code> separator
   * </ul>
   *
   * @param node the visited InitDeclaratorList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTInitDeclaratorList node, Object data) 
  {
     unparseList(node,0,", ",false,data);
     return data;
  }


  /**
   * Visits an InitDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit declarator
   * <li>if there is an initializer, emit <code>=</code> and visit it
   * </ul>
   *
   * @param node the visited InitDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTInitDeclarator node, Object data) 
  {
    // visit declarator
     node.jjtGetChild(0).jjtAccept(this, data);
     // if there is an initializer, emit <code>=</code> and visit it
     if ( node.jjtGetNumChildren() == 2 )
     {
        emit(" = ");
        node.jjtGetChild(1).jjtAccept(this, data);
     }
     return data;
  }

  /**
   * Visits an StructSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit node's name (struct or union followed by the optional tag)
   * <li>if there are members, emit them in brackets
   * </ul>
   *
   * @param node the visited StructSpecifier
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTStructSpecifier node, Object data) 
  {
     // set up a line watcher
     LineWatcher lw = new LineWatcher();

     // emit struct or union name
     emit(node.getName());
     
     // if there are members, emit them inside braces
     int nbChildren = node.jjtGetNumChildren();
     if ( nbChildren > 0 )
     {
        emit(" { ");
        addIndent(regIndent);
        unparseList(node,0," ",false,data);
        delIndent(regIndent);
        if ( lw.newlineEmmited() )
        {
           newline();
           indent();
        }
        else
        {
           emit(" ");
        }
        emit("}");
     }
     return data;
  }

  /**
   * Visits an MemberDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children
   * <li>if member declaration is not a simple declaration, emit ";"
   * </ul>
   *
   * @param node the visited MemberDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTMemberDeclaration node, Object data) 
  {
     // visit children
     node.childrenAccept(this,data);

     // if member declaration is not a simple declaration, emit ";"
     if ( node.jjtGetNumChildren() == 0 )
     {
        emit(";");
     }
     else if ( ! ( node.jjtGetChild(0) instanceof ASTDeclaration ) )
     {
        emit(";");
     }

     return data;
  }

  /**
   * Visits an MemberDeclaratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children with a <code>,</code> separator
   * </ul>
   *
   * @param node the visited MemberDeclaratorList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTMemberDeclaratorList node, Object data) 
  {
     unparseList(node,0,", ",false,data);
     return data;
  }

  /**
   * Visits an EnumSpecifier.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for each child,<ul>
   * <li>emit <code></code>
   * <li>visit the child
   * <li>emit <code></code></ul>
   * </ul>
   * </ul>
   *
   * @param node the visited EnumSpecifier
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTEnumSpecifier node, Object data) 
  {
     // set up a line watcher
     LineWatcher lw = new LineWatcher();

     // emit enum name
     emit(node.getName());
     
     // if there are members, emit them inside braces
     int nbChildren = node.jjtGetNumChildren();
     if ( nbChildren > 0 )
     {
        emit(" { ");
        addIndent(regIndent);
        unparseList(node,0," ",false,data);
        delIndent(regIndent);
        if ( lw.newlineEmmited() )
        {
           newline();
           indent();
        }
        else
        {
           emit(" ");
        }
        emit("}");
     }
     return data;
  }

  /**
   * Visits an EnumeratorList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children with a <code>,</code> separator
   * </ul>
   *
   * @param node the visited EnumeratorList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTEnumeratorList node, Object data) 
  {
     unparseList(node,0,", ",false,data);
     return data;
  }

  /**
   * Visits an Enumerator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit enumerator id
   * <li>if there is an initializer, emit <code>=</code> and visit it
   * </ul>
   *
   * @param node the visited Enumerator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTEnumerator node, Object data) 
  {
     // emit enumerator id
     emit(node.getName());

     // if there is an initializer, emit <code>=</code> and visit it
     if ( node.jjtGetNumChildren() > 0 )
     {
        emit(" = ");
        node.jjtGetChild(0).jjtAccept(this, data);
     }
     return data;
  }

  /**
   * Visits an PtrOperator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>*</code>
   * <li>visit CvQualifierSeq
   * </ul>
   *
   * @param node the visited PtrOperator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTPtrOperator node, Object data) 
  {
     // emit <code>*</code>
     emit("*");

     // visit children
     node.childrenAccept( this,data);
     
     return data;
  }

  /**
   * Visits an CvQualifierSeq.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit node's name
   * </ul>
   *
   * @param node the visited CvQualifierSeq
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTCvQualifierSeq node, Object data) 
  {
     emit (node.getName() );
     return data;
  }

  /**
   * Visits an Declarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children with a <code>blank</code> separator 
   * </ul>
   *
   * @param node the visited Declarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDeclarator node, Object data) 
  {
     unparseList(node,0," ",true,data);
     return data;
  }

  /**
   * Visits an DirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if DirectDeclarator's first child is a nested Declarator
   *     emit it inside parenthesis, otherwise simply emit name
   * <li>visit DeclaratorSuffixes
   * </ul>
   *
   * @param node the visited DirectDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDirectDeclarator node, Object data) 
  {
     // if DirectDeclarator's first child is a nested Declarator
     // emit it inside parenthesis, otherwise simply emit name

     boolean nestedDeclarator = node.jjtGetChild(0) instanceof ASTDeclarator;

     if ( nestedDeclarator ) 
     {
        emit("( ");
     }

     node.jjtGetChild(0).jjtAccept(this, data);     

     if (  nestedDeclarator ) 
     {
        emit(" )");
     }

     // visit DeclaratorSuffixes
     int nbChildren = node.jjtGetNumChildren();
     for (int i = 1; i < nbChildren ; i++)
     {
       node.jjtGetChild(i).jjtAccept(this, data); 
     }

     return data;
  }

  /**
   * Visits an DeclaratorSuffixes.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>in case suffix is bracket suffixes, unparse brackets and subscripts
   * <li>in case suffix is a parameter list, unparse parameters inside 
   *     parenthesis
   * <li>in case suffix is a bit field : unparse ":" before constant
   * </ul>
   *
   * @param node the visited DeclaratorSuffixes
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDeclaratorSuffixes node, Object data) 
  {
     if ( node.getName().equals("[]") )
     {
        // suffix bracket suffixes, unparse brackets and subscripts
        try
        {
           emit("[");
           SimpleNode[] subscripts = (SimpleNode[])node.getProperty
                                                      ("subscripts");
           unparseList(subscripts, "][", data);        
           emit("]");
        }
        catch ( PropertyAccessException e ) { }
     }
     else if ( node.getName().equals("()") )
     {
        // parameter list, unparse inside parenthesis
        emit("(");
        node.jjtGetChild(0).jjtAccept(this, data);
        emit(")");
     }
     else if ( node.getName().equals(":") )
     {
        // bit field : unparse ":" before constant
        emit(": ");
        node.jjtGetChild(0).jjtAccept(this, data);
     }
     return data;
  }

  /**
   * Visits an FunctionDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children, separating them with a blank
   * </ul>
   *
   * @param node the visited FunctionDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTFunctionDeclarator node, Object data) 
  {
     // visit children
     unparseList(node,0," ",true,data);
     return data;
  }

  /**
   * Visits an FunctionDirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit function name
   * <li>emit parameter list inside parenthesis
   * </ul>
   *
   * @param node the visited FunctionDirectDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTFunctionDirectDeclarator node, Object data) 
  {
     node.jjtGetChild(0).jjtAccept(this, data);
     emit("(");
     if ( node.jjtGetNumChildren() > 1 )
     {
        node.jjtGetChild(1).jjtAccept(this, data);
     }
     emit(")");
     return data;
  }

  /**
   * Visit ASTParameterList.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> emit parameter declaration list
   * <li> emit final ellipsis if there is one
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTParameterList node, Object data)
  {
     try
     {
        // emit parameter declaration list
        if ( node.jjtGetNumChildren() > 0 )
        {
           node.jjtGetChild(0).jjtAccept(this,data);
        }
        
        Boolean lastParameterIsEllipsis = (Boolean)node.getProperty
                                          ("lastParameterIsEllipsis");
        if ( Boolean.TRUE.equals(lastParameterIsEllipsis) )
        {
           // if ellipsis is not only parameter emit separator
           if ( node.jjtGetNumChildren() > 0 )
           {
              emit(", ");
           }
           // emit ellipsis
           emit("...");
        }
     }
     catch ( PropertyAccessException e ) { }
     
     return data;
  }


  /**
   * Visits an ParameterDeclarationList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit parameters separated by a ","
   * </ul>
   *
   * @param node the visited ParameterDeclarationList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTParameterDeclarationList node, Object data) 
  {
     // emit parameters separated by a ","
     unparseList(node,0,", ",true,data);
     return data;
  }

  /**
   * Visits an ParameterDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children separated by blanks
   * </ul>
   *
   * @param node the visited ParameterDeclaration
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTParameterDeclaration node, Object data) 
  {
     // emit children separated by blanks
     unparseList(node,0," ",true,data);
     return data;
  }

  /**
   * Visits an Initializer.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children separated by a "," and inside bracket
   * </ul>
   *
   * @param node the visited Initializer
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTInitializer node, Object data) 
  {
     // set up a line watcher
     LineWatcher lw = new LineWatcher();

     emit("{ ");
     addIndent(regIndent);
     unparseList(node,0,", ",true,data);
     delIndent(regIndent);
     if ( lw.newlineEmmited() )
     {
        newline();
        indent();
     }
     else
     {
        emit(" ");
     }
     emit("}");
     return data;
  }

  /**
   * Visits an TypeName.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children separated by blanks
   * </ul>
   *
   * @param node the visited TypeName
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTTypeName node, Object data) 
  {
     // emit children separated by blanks
     unparseList(node,0," ",true,data);
     return data;
  }

  /**
   * Visits an AbstractDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for pointer abstract declarator, emit children separated by blanks
   * <li>for nested abstract declarator and suffixes, unparse nested abstract
   *     declarator inside parenthesis and then suffixes
   * <li>for suffix bracket suffixes, unparse brackets and subscripts
   * </ul>
   *
   * @param node the visited AbstractDeclarator
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTAbstractDeclarator node, Object data) 
  {
     if ( node.getName().equals("") )
     {
        // pointer abstract declarator, emit children separated by blanks
        unparseList(node,0," ",true,data);
     }
     else if ( node.getName().equals("()") )
     {
        // nested abstract declarator and suffixes, unparse nested abstract
        // declarator inside parenthesis and then suffixes
        emit("(");
        node.jjtGetChild(0).jjtAccept(this, data);
        emit(")");
        unparseList(node,1," ",true,data);
     }
     else if ( node.getName().equals("[]") )
     {
        // suffix bracket suffixes, unparse brackets and subscripts
        try
        {
           emit("[");
           SimpleNode[] subscripts = (SimpleNode[])node.getProperty
                                                      ("subscripts");
           unparseList(subscripts, "][", data);        
           emit("]");
        }
        catch ( PropertyAccessException e ) { }
     }
     return data;
  }

  /**
   * Visits an AbstractDeclaratorSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for bracket suffixe, unparse brackets and subscript
   * <li>for parameter list, unparse inside parenthesis
   * </ul>
   *
   * @param node the visited AbstractDeclaratorSuffix
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTAbstractDeclaratorSuffix node, Object data) 
  {
     if ( node.getName().equals("[]") )
     {
        // bracket suffixe, unparse brackets and subscript
        emit("[");
        if ( node.jjtGetNumChildren() > 0 )
        {
           node.jjtGetChild(0).jjtAccept(this, data);
        }
        emit("]");
     }
     else if ( node.getName().equals("()") )
     {
        // parameter list, unparse inside parenthesis
        emit("(");
        if ( node.jjtGetNumChildren() > 0 )
        {
           node.jjtGetChild(0).jjtAccept(this, data);
        }
        emit(")");
     }
     return data;
  }

  /**
   * Visit ASTStatement.
   * <p> <b> Algorithm : </b> <ul> 
   * <li> visit children 
   * <li> emit final ";" if necessary
   * <li> emit new line
   * </ul>
   * @param node node to visit.
   * @param data arbitrary data received from the caller.
   */

  public Object visit(ASTStatement node, Object data)
  {
     try
     {
        // emit children
        node.childrenAccept(this,data);

        // emit final ";" if required
        Boolean endsWithSemiColon = 
                (Boolean) node.getProperty("endsWithSemiColon");

        if ( Boolean.TRUE.equals(endsWithSemiColon) )
        {
           emit(";");
        }
     }
     catch ( PropertyAccessException e ) { }
     
     return data;
  }


  /**
   * Visits an LabeledStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> emit "label :"
   * <li> emit child expression
   * </ul>
   *
   * @param node the visited LabeledStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTLabeledStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     // emit "label :"
     emit(node.getName());
     emit(" : ");

     // emit children
     node.childrenAccept(this,data);
     return data;
  }

  /**
   * Visits an CaseLabelStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit "case"
   * <li>emit constant expression
   * <li>emit ":"
   * <li>emit statement
   * </ul>
   *
   * @param node the visited CaseLabelStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTCaseLabelStatement node, Object data) 
  {
     // emit case statement a little before regular statement
     delIndent(subIndent);
     indentIfRequired();

     emit("case ");

     // emit constant expression
     node.jjtGetChild(0).jjtAccept(this, data);

     emit(" : ");

     // emit statement
     node.jjtGetChild(1).jjtAccept(this, data);

     // restore indentation
     addIndent(subIndent);
     return data;
  }

  /**
   * Visits an DefaultLabelStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit "default :"
   * <li>emit statement
   * </ul>
   *
   * @param node the visited DefaultLabelStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDefaultLabelStatement node, Object data) 
  {
     // emit case statement a little before regular statement
     delIndent(subIndent);
     indentIfRequired();

     emit("default : ");

     // emit statement
     node.jjtGetChild(0).jjtAccept(this, data);

     // restore indent
     addIndent(subIndent);

     return data;
  }

  /**
   * Visits an CompoundStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit children inside braces
   * </ul>
   *
   * @param node the visited CompoundStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTCompoundStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     // set up a line watcher
     LineWatcher lw = new LineWatcher();

     emit("{");

     addIndent(regIndent);
     
     int nbChildren = node.jjtGetNumChildren();
     for ( int i = 0; i < nbChildren; i++ ) 
     {
        newline();
        node.jjtGetChild(i).jjtAccept(this, data);
     }

     // restore current indentation
     delIndent(regIndent);

     // emit final brace

     if ( lw.newlineEmmited() )
     {
        newline();
        indent();
     }
     else
     {
        emit(" ");
     }

     emit("}");

     return data;
  }

  /**
   * Visits an IfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit if
   * <li>emit condition inside parenthesis
   * <li>emit if statement
   * <li>if required emit else statement
   * </ul>
   *
   * @param node the visited IfStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTIfStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("if ( ");
     // emit condition
     node.jjtGetChild(0).jjtAccept(this, data);
     emit(" ) ");

     // emit if statement
     node.jjtGetChild(1).jjtAccept(this, data);

     // if required emit else statement
     if ( node.jjtGetNumChildren() > 2 )
     {
        emit(" else ");
        node.jjtGetChild(2).jjtAccept(this, data);
     }

     return data;
  }

  /**
   * Visits an SwitchStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit switch
   * <li>emit condition inside parenthesis
   * <li>emit statement
   * </ul>
   *
   * @param node the visited SwitchStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTSwitchStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("switch ( ");

     // emit condition
     node.jjtGetChild(0).jjtAccept(this, data);
     emit(" ) ");

     // emit statement
     node.jjtGetChild(1).jjtAccept(this, data);

     return data;
  }

  /**
   * Visits an WhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit while
   * <li>emit condition inside parenthesis
   * <li>emit statement
   * </ul>
   *
   * @param node the visited WhileStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTWhileStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("while ( ");
     // emit condition
     node.jjtGetChild(0).jjtAccept(this, data);
     emit(" ) ");

     // emit statement
     node.jjtGetChild(1).jjtAccept(this, data);

     return data;
  }

  /**
   * Visits an DoWhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit do
   * <li>emit statement
   * <li>emit while
   * <li>emit condition inside parenthesis
   * </ul>
   *
   * @param node the visited DoWhileStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDoWhileStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     // emit statement
     emit("do ");
     node.jjtGetChild(0).jjtAccept(this, data);
     emit(" while ( ");

     // emit condition
     node.jjtGetChild(1).jjtAccept(this, data);
     emit(" );");
     
     return data;
  }

  /**
   * Visits an ForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit children taking into account the for statmement structure
   * <li>visit last child (statement)
   * </ul>
   *
   * @param node the visited ForStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTForStatement node, Object data) 
  {
     try
     {
        // always indent a statement
        indentIfRequired();

        int childIndex = 0;

        boolean[] hasExpressionAtPosition = 
                  (boolean[]) node.getProperty("hasExpressionAtPosition");

        SimpleNode[] forExpressions = new SimpleNode[3]; 
        for (int i = 0; i < 3 ; i++ )
        {
           // update for expression array and current child index
            forExpressions[i] =  hasExpressionAtPosition[i] 
                              ?  (SimpleNode)(node.jjtGetChild(childIndex++))
                              : null;
        }

        emit("for ( ");
        unparseList(forExpressions,
                    " ; ",
                    true, // separate before new line
                    data);
        emit(" ) ");

        // visit last child
        node.jjtGetChild(node.jjtGetNumChildren() - 1).jjtAccept(this,data);
     }
     catch ( PropertyAccessException e ) { }

     return data;
  }

  /**
   * Visits an GotoStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>goto</code>, lable and <code>;</code></ul>
   * </ul>
   *
   * @param node the visited GotoStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTGotoStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("goto ");
     emit(node.getName());
     emit(";");
     return data;
  }

  /**
   * Visits an ContinueStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>continue;</code>
   * </ul>
   *
   * @param node the visited ContinueStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTContinueStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("continue;");
     return data;
  }

  /**
   * Visits an BreakStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit<code>break;</code>
   * </ul>
   *
   * @param node the visited BreakStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTBreakStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("break;");
     return data;
  }

  /**
   * Visits an ReturnStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>return</code>
   * <li>if there is one, emit expression
   * </ul>
   *
   * @param node the visited ReturnStatement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTReturnStatement node, Object data) 
  {
     // always indent a statement
     indentIfRequired();

     emit("return");
     if ( node.jjtGetNumChildren() > 0 )
     {
        emit(" ");
        node.jjtGetChild(0).jjtAccept(this,data);
     }
     emit(";");
     return data;
  }

  /**
   * Visits an Expression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit each child expression separated by a comma
   * </ul>
   *
   * @param node the visited Expression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTExpression node, Object data) 
  {
     indentIfRequired();
     unparseList(node,0,", ",true,data);
     return data;
  }

  /**
   * Visits an AssignmentExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit left hand expression
   * <li>emit operator
   * <li>emit right hand expression
   * </ul>
   *
   * @param node the visited AssignmentExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTAssignmentExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        node.jjtGetChild(0).jjtAccept(this, data);
        emit(" " + (String)node.getProperty("operator") + " ");
        node.jjtGetChild(1).jjtAccept(this, data);
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits a ConditionalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit first expression
   * <li>emit "?"
   * <li>emit second expression
   * <li>emit ":"
   * <li>emit last expression
   * </ul>
   *
   * @param node the visited ConditionalExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTConditionalExpression node, Object data) 
  {
     indentIfRequired();
     String[] operators = { " ? " , " : " };
     Indenter indenter = new Indenter();
     unparseList(node,0,operators,false,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an ConstantExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expression
   * </ul>
   *
   * @param node the visited ConstantExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTConstantExpression node, Object data) 
  {
     indentIfRequired();
     node.jjtGetChild(0).jjtAccept(this, data);
     return data;
  }

  /**
   * Visits an LogicalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operator
   * </ul>
   *
   * @param node the visited LogicalOrExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTLogicalOrExpression node, Object data) 
  {
     indentIfRequired();
     Indenter indenter = new Indenter();
     unparseList(node,0," || ",true,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an LogicalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operator
   * </ul>
   *
   * @param node the visited LogicalAndExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTLogicalAndExpression node, Object data) 
  {
     indentIfRequired();
     Indenter indenter = new Indenter();
     unparseList(node,0," && ",true,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an InclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operator
   * </ul>
   *
   * @param node the visited InclusiveOrExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTInclusiveOrExpression node, Object data) 
  {
     indentIfRequired();
     Indenter indenter = new Indenter();
     unparseList(node,0," | ",true,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an ExclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operator
   * </ul>
   *
   * @param node the visited ExclusiveOrExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTExclusiveOrExpression node, Object data) 
  {
     indentIfRequired();
     Indenter indenter = new Indenter();
     unparseList(node,0," ^ ",true,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an AndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operator
   * </ul>
   *
   * @param node the visited AndExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTAndExpression node, Object data) 
  {
     indentIfRequired();
     Indenter indenter = new Indenter();
     unparseList(node,0," & ",true,data);
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an EqualityExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operators
   * </ul>
   *
   * @param node the visited EqualityExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTEqualityExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String[] operators = (String[])node.getProperty("operators");
        Indenter indenter = new Indenter();
        unparseList(node,0,operators,true,data);
        indenter.restoreFormerIndentation();
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an RelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operators
   * </ul>
   *
   * @param node the visited RelationalExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTRelationalExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String[] operators = (String[])node.getProperty("operators");
        Indenter indenter = new Indenter();
        unparseList(node,0,operators,true,data);
        indenter.restoreFormerIndentation();
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an ShiftExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operators
   * </ul>
   *
   * @param node the visited ShiftExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTShiftExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String[] operators = (String[])node.getProperty("operators");
        Indenter indenter = new Indenter();
        unparseList(node,0,operators,true,data);
        indenter.restoreFormerIndentation();
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an AdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operators
   * </ul>
   *
   * @param node the visited AdditiveExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTAdditiveExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String[] operators = (String[])node.getProperty("operators");
        Indenter indenter = new Indenter();
        unparseList(node,0,operators,true,data);
        indenter.restoreFormerIndentation();
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an MultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expressions separated by operators
   * </ul>
   *
   * @param node the visited MultiplicativeExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTMultiplicativeExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String[] operators = (String[])node.getProperty("operators");
        Indenter indenter = new Indenter();
        unparseList(node,0,operators,true,data);
        indenter.restoreFormerIndentation();
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an CastExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit cast type in between parenthesis
   * <li>emit child expression
   * </ul>
   *
   * @param node the visited CastExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTCastExpression node, Object data) 
  {
     indentIfRequired();
     emit("(");
     node.jjtGetChild(0).jjtAccept(this, data);
     emit(") ");
     node.jjtGetChild(1).jjtAccept(this, data);
     return data;
  }

  /**
   * Visits an PreIncrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit "++" and child expression
   * </ul>
   *
   * @param node the visited PreIncrementExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTPreIncrementExpression node, Object data) 
  {
     indentIfRequired();
     emit("++");
     node.jjtGetChild(0).jjtAccept(this, data);
     return data;
  }

  /**
   * Visits an PreDecrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit "--" and child expression
   * </ul>
   *
   * @param node the visited PreDecrementExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTPreDecrementExpression node, Object data) 
  {
     indentIfRequired();
     emit("--");
     node.jjtGetChild(0).jjtAccept(this, data);
     return data;
  }

  /**
   * Visits an UnaryOperatorExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator and child expression
   * </ul>
   *
   * @param node the visited UnaryOperatorExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTUnaryOperatorExpression node, Object data) 
  {
     try
     {
        indentIfRequired();
        String operator = (String)node.getProperty("operator");
        emit(operator);
        node.jjtGetChild(0).jjtAccept(this, data);
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }

  /**
   * Visits an SizeOfExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>sizeof</code>
   * <li>if child is a type name emit it in between parenthesis
   * <li>otherwise emit child expression alone
   * </ul>
   *
   * @param node the visited SizeOfExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTSizeOfExpression node, Object data) 
  {
     indentIfRequired();
     emit("sizeof ");
     SimpleNode child = (SimpleNode)node.jjtGetChild(0);

     if ( child instanceof ASTTypeName )
     {
        emit("(");
     }

     child.jjtAccept(this, data);

     if ( child instanceof ASTTypeName )
     {
        emit(")");
     }
     return data;
  }

  /**
   * Visits an PostfixExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>just visit children
   * </ul>
   *
   * @param node the visited PostfixExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTPostfixExpression node, Object data) 
  {
     indentIfRequired();
     data = node.childrenAccept(this, data );
     return data;
  }


  /**
   * Visits an ArraySuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expression in between brackets
   * </ul>
   *
   * @param node the visited ArraySuffix
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTArraySuffix node, Object data) 
  {
     emit("[");
     node.jjtGetChild(0).jjtAccept(this,data);
     emit("]");
     return data;
  }

  /**
   * Visits an ArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expression in between parenthesis
   * </ul>
   *
   * @param node the visited ArgumentList
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTArgumentList node, Object data) 
  {
     emit("(");
     Indenter indenter = new Indenter();
     data = node.childrenAccept(this, data );
     emit(")");
     indenter.restoreFormerIndentation();
     return data;
  }

  /**
   * Visits an DotAccessSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit dot and identifier
   * </ul>
   *
   * @param node the visited DotAccessSuffix
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTDotAccessSuffix node, Object data) 
  {
     emit(".");
     emit(node.getName());
     return data;
  }

  /**
   * Visits an RefAccessSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit reference access and identifier
   * </ul>
   *
   * @param node the visited RefAccessSuffix
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTRefAccessSuffix node, Object data) 
  {
     emit("->");
     emit(node.getName());
     return data;
  }

  /**
   * Visits an PostDeIncrement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator
   * </ul>
   *
   * @param node the visited PostDeIncrement
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTPostDeIncrement node, Object data) 
  {
     try
     {
        String operator = (String)node.getProperty("operator");
        emit(operator);
     }
     catch ( PropertyAccessException e ) { }
     return data;
  }


  /**
   * Visits an String.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit string image
   * </ul>
   *
   * @param node the visited String
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTString node, Object data) 
  {
     indentIfRequired();
     emit(node.getName());
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Return node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /**
   * Visits an ParentheizedExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit child expression in between parenthesis
   * </ul>
   *
   * @param node the visited ParentheizedExpression
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTParentheizedExpression node, Object data) 
  {
     indentIfRequired();
     emit("(");
     node.jjtGetChild(0).jjtAccept(this,data);
     emit(")");
     return data;
  }

  /**
   * Visits an UnaryPlusMinusConstant.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator and child constant
   * </ul>
   *
   * @param node the visited UnaryPlusMinusConstant
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTUnaryPlusMinusConstant node, Object data) 
  {
     try
     {
        indentIfRequired();
        String operator = (String)node.getProperty("operator");
        emit(operator);
        node.jjtGetChild(0).jjtAccept(this,data);
     }
     catch ( PropertyAccessException e ) { }

     return data;
  }

  /**
   * Visits an Constant.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit constant image
   * </ul>
   *
   * @param node the visited Constant
   * @param data unused
   * @return the unused data
   */

  public Object visit(ASTConstant node, Object data) 
  {
     indentIfRequired();
     emit(node.getName());
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_AnnotatedDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_ImportDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_UseDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_Declaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_FieldDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetDeclaration node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_FactoryDefinition node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_TestDirective node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetDomain node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetLiteral node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetMember node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetExpression node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetSingleton node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_FactoryCall node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDD_ADLExpression node, Object data)
  {
     incorrectNode(node);
     return data;
  }

  // -- private methods --

  /**
   * Unparse a list of <em>something</em> each having its own separator,
   * possibly wrapping on the next line if required.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get current indentation level
   * <li>set current indentation to current column so that wrapped
   *     statements are all aligned with the first one
   * <li>get the number of children from the parent node
   * <li>visit first child in the list (<code>firstChild</code>)
   * <li>for all remaining children of the parent node:<ul>
   * <li>if the separator should be emitted before wrapping, emit it
   * <li>if current column is already below the line limit:<ul>
   * <li>go to next line
   * <li>indent
   * </ul><li>else emit a space
   * <li>if the separator should be emitted after wrapping, emit it
   * <li>visit the child
   * </ul>
   * <li>reset current indentation level
   * </ul>
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separators the array of strings that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */

  private Object unparseList(Node node,
                             int firstChild, 
                             String[] separators,
                             boolean sepBefore,
                             Object data)
  {
     // get number of children
     int numChildren = node.jjtGetNumChildren();

     // visit first child, then loop on remaining ones
     node.jjtGetChild(firstChild).jjtAccept(this, data);
     for (int childNum = firstChild + 1; childNum < numChildren; childNum++)
     {
        // if requested, emit the separator before wrapping
        if (sepBefore) emit(" " + separators[childNum - 1] + " ");

        // determine if we have to wrap
        if (currentColumn >= lineLimit)
        {
           newline();
           indent();
        } 

        // if requested, emit the separator after wrapping
        if (!sepBefore)
        {
           emit(" " + separators[childNum - 1] + " ");
        }

        node.jjtGetChild(childNum).jjtAccept(this, data);
     }

     return data;
  }

  /**
   * Unparse a list of <em>something</em> supplying a single separator,
   * possibly wrapping on the next line if required.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get current indentation level
   * <li>set current indentation to current column so that wrapped
   *     statements are all aligned with the first one
   * <li>get the number of children from the parent node
   * <li>visit first child in the list (<code>firstChild</code>)
   * <li>for all remaining children of the parent node:<ul>
   * <li>if the separator should be emitted before wrapping, emit it
   * <li>if current column is already below the line limit:<ul>
   * <li>go to next line
   * <li>indent
   * </ul><li>else emit a space
   * <li>if the separator should be emitted after wrapping, emit it
   * <li>visit the child
   * </ul>
   * <li>reset current indentation level
   * </ul>
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separator the string that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */

  private Object unparseList(Node node, 
                             int firstChild, 
                             String separator,
                             boolean sepBefore,
                             Object data)
  {
     // get number of children
     int numChildren = node.jjtGetNumChildren();

     // visit first child, then loop on remaining ones
     if ( numChildren > firstChild )
     {
        node.jjtGetChild(firstChild).jjtAccept(this, data);
        for (int childNum = firstChild + 1; childNum < numChildren; childNum++)
        {
           // if requested, emit the separator before wrapping
           if (sepBefore) emit(separator);

           // determine if we have to wrap
           if (currentColumn >= lineLimit) 
           {
              newline();
              indent();
           } 

           // if requested, emit the separator after wrapping
           if (!sepBefore) 
           {
              emit(separator);
           }

           node.jjtGetChild(childNum).jjtAccept(this, data);
        }
     }
     
     return data;
  }

  /**
   * Unparse a list of <em>something</em> supplying a single separator,
   * possibly wrapping on the next line if required.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get current indentation level
   * <li>set current indentation to current column so that wrapped
   *     statements are all aligned with the first one
   * <li>get the number of children from the parent node
   * <li>visit first child in the list (<code>firstChild</code>)
   * <li>for all remaining children of the parent node:<ul>
   * <li>if the separator should be emitted before wrapping, emit it
   * <li>if current column is already below the line limit:<ul>
   * <li>go to next line
   * <li>indent
   * <li>if the separator should be emitted after wrapping, emit it
   * <li>visit the child
   * </ul>
   * <li>reset current indentation level
   * </ul>
   *
   * @param nodes nodes to visit (zero length string is emmited for nulls)
   * @param separator the string that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */

  private Object unparseList(SimpleNode[] nodes, 
                             String separator,
                             boolean sepBefore,
                             Object data)
  {
     // visit first child, then loop on remaining ones
     if ( nodes[0] != null )
     {
        nodes[0].jjtAccept(this, data);
     }

     for (int childNum = 1; childNum < nodes.length; childNum++)
     {
        // if requested, emit the separator before wrapping
        if (sepBefore) emit(separator);
        
        // determine if we have to wrap
        if (currentColumn >= lineLimit) 
        {
           newline();
           indent();
        } 

        // if requested, emit the separator after wrapping
        if (!sepBefore) 
        {
           emit(separator);
        }

        if ( nodes[childNum] != null )
        {
           nodes[childNum].jjtAccept(this, data);
        }
     }
     
     return data;
  }

  /**
   * Unparse a list of nodes supplying a single separator. For null nodes
   * a zero length string is emmited (that is nothing!).
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for each node:
   * <li>emit separator except for first node
   * <li>emit node
   * </ul>
   *
   * @param nodes nodes to unparse
   * @param separator the string that separate children
   * @param data unused
   * @return the unused data
   */

  private Object unparseList(SimpleNode[] nodes, 
                             String separator,
                             Object data)
  {
     for (int i = 0; i < nodes.length; i++)
     {
        if ( i > 0 )
        {
           emit(separator);
        }

        if ( nodes[i] != null )
        {
           nodes[i].jjtAccept(this, data);
        }
     }
     return data;
  }
  
  /** Indent only in case current column is zero. */
  private void indentIfRequired()
  {
     if ( currentColumn == 0 )
     {
        indent();
     }
  }

 
  /**
   * Emits a newline character (system dependent) on the opened
   * PrintWriter stream and reset current column marker.
   *
   * @see java.io.PrintWriter#println
   */
  protected void newline() 
  {
     super.newline();
     _currentLine++;
  }

  private long _currentLine = 1;

  /**
   * Sets the current value of the indentation level: according to the new
   * indentation level, proceed with <code>addIndent</code> or
   * <code>delIndent</code>. 
   *
   * @param newIndent the new indentation level
   */

  public void setIndent(int newIndent) { super.setIndent(newIndent); }

  /**
   * Adds an extra indent to the current indentation level: append
   * whitespace characters to the end of the <code>currentIndent</code>
   * buffer.
   *
   * @param extraIndent the extra indent to add
   */

  public void addIndent(int extraIndent) { super.addIndent(extraIndent); }

  /**
   * Returns the current value of indentation level.
   */

  public int getIndent() { return super.getIndent(); }

  /**
   * Enable to control wether a new line has been emmited 
   * during some duration. 
   * Example, if you want a pretty formatting for a struct :
   * <code>
   *    emit("struct {");
   *    LineWatcher lw = new LineWatcher(); 
   *    // emit struct member in some way
   *    if ( lw.newlineEmmited() )
   *    {
   *       newline();
   *       indent();
   *    }
   *    else
   *    {
   *       emit(" ");
   *    }
   *    emit("}");
   * </code>
   */
     
  class LineWatcher
  {
    /** Has a new line been emmited since line watcher construction. */

    boolean newlineEmmited()
    {
       return _watchedLine < _currentLine ;
    }

    /**
     * Watched line number, that is the current line number 
     * when the line watcher is constructed.
     */

    private final long _watchedLine = _currentLine;
  }

  /** 
   * Enable to handle indentation.
   * TBD - an indent mechanism should be added to the generic unparser
   */

  class Indenter
  {
    /** Default constructor. Set indentation to current column. */
    Indenter()
    {
       setIndent(getCurrentColumn());
    }
    /** Constructs indenter to add supplied column to indentation. */
    Indenter(int added)
    {
       addIndent(added);
    }
    void restoreFormerIndentation()
    {
       setIndent(_savedIndent);
    }
    private final int _savedIndent = getIndent();
  }

}
