// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/cpp/CPPNLDUnparserVisitor.java,v $
// $Revision: 1.10 $
// $Date: 1998/08/06 14:10:58 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   NLD/C++ Unparser
// NAME:        CPPNLDUnparserVisitor.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - May 28, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.cpp;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;
import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.unparsers.NLDGenericUnparser;
import ORG.opengroup.adl.unparsers.DTD;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import ORG.opengroup.adl.nlengine.NLEngine;

/** 
  * This class defines a visitor which can be applied to an AST
  * generated from the ADLCPPParser grammar. It does <B>NOT</B>
  * modify this AST: its only effect is to produce an SGML file
  * that corresponds to the documentation of the visited file.
  *
  * <p>The basic algorithm is to walk through the AST using the "visitor"
  * pattern (see Design Patterns from Gamma et al.) and emit the code on
  * an output stream maintained by the visitor itself.
  *
  * <p>The unparsing starts by calling the <a href="#unparse">unparse()</a>
  * method passing the top of AST to unparse as parameter.
  */

public class CPPNLDUnparserVisitor extends NLDGenericUnparser
  implements ADLCPPParserVisitor
{

  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.10 $";

  /** Temporary buffer for the currently unparsed assertion. */
  private StringBuffer assertionBuffer = null;

  /** Flag set when visiting an assertion. */
  private boolean inAssertion = false;

  /** Flag set when visiting a TDD definition (constant, dataset, factory). */
  private boolean inTddDef = false;

  /** Flag set when the current assertion is an "exception behavior:
    * assertion. */
  private boolean exceptionBhv;

  /** Name of the ADL/TDD class. */
  private String className;

  /** Index for anonymous directives. */
  private int anonDirectiveIndex;

  private int nbParenth;

  /** Flag for constructor behaviors. */
  private boolean constructor;

  static private final boolean debug = false;


  /**
    * Default constructor: call superclass constructor.
    *
    * @param toolname the name of the tool using this unparser visitor
    * @param nlengine the NL engine to deal with the Prolog interpreter
    */
  public CPPNLDUnparserVisitor(String toolname, NLEngine nlengine) {

    super(toolname, versionId, nlengine);
  }


  /**
    * Starts unparsing an AST.
    *
    * If the passed AST root is not an CompilationUnit, an error is
    * reported and the AST is not unparsed. Otherwise the AST is
    * visited.
    *
    * @param ast the root node of the AST to unparse
    * @see #visit(ORG.opengroup.adl.parsers.adljava.ASTCompilationUnit, java.lang.Object)
    */
  public void unparse(Node ast) throws ParseException {

    if (debug)
      System.out.println("NLD unparse");

    if (assertNodeType(ast, ASTTranslationUnit.class)) {
      ast.jjtAccept(this, null);
    } else {
      incorrectNode(ast);
      System.err.println("\twas expecting an adl or tdd class");
    }
  }


  /**
    * Visits a ADL_CompilationUnit.
    *
    * <p><b>Algorithm:</b><ul>
    * <li>get the name of the declared adlclass from the
    *     declaration node (always last one),
    * <li>open the source file for the class or interface (<code>outputDirectory</code>/<em>name</em>.sgml),
    * <li>emit the standard header,
    * <li>visit the child nodes,
    * <li>and finally close the file.
    * </ul>
    *
    * @param node the ADL_CompilationUnit to visit
    * @param data unused
    * @return data returned by last child visited or
    * null if the file cannot be opened.
    * @see ORG.opengroup.adl.unparsers.GenericUnparserVisitor#outputDirectory
    *
    */
  public Object visit(ASTTranslationUnit node, Object data) throws ParseException {

    if (debug)
      System.out.println("visit ASTTranslationUnit");

    // save former stm context
    Object savedStmContext = SimpleNode.stm.getContext();

    // set stm context for this ast
    try {
      Object stmContext = node.getProperty("stmContext");
      SimpleNode.stm.setContext(stmContext);
    }
    catch ( PropertyAccessException e ) { }

    // get the name of the declared adlclass and open
    // the corresponding stream.
    
    // The declaration node is always the last child
    SimpleNode decl = null;
    if (node.jjtGetNumChildren() >= 1) {
      decl = (SimpleNode)node.jjtGetChild(node.jjtGetNumChildren() - 1);
    }

    if ((decl == null) ||
	!((decl instanceof ASTADLClass) ||
	  (decl instanceof ASTTDDClassDecl))) {
      System.err.println(toolName + ": Can't get name of file: " +
			 "no adl or tdd class declaration");
      return data;
    }

    // now we know the decl node is an adl or tdd class

    className = node.getName();

    if (decl instanceof ASTADLClass) {
      initialize_aco(className);
    }
    else {
      initialize_tdd(className);
    }

    // try to open the output stream
    try {
      // Try to get best performance through a BufferedWriter on top of
      // a FileWriter
      String fileName = outputDirectory + File.separator + className +
        ((decl instanceof ASTADLClass) ? "ACO" : "TDD") + ".sgml";

      if (Adlt.verboseMode()) {
	System.out.println("+++ generating " + fileName);
      }

      FileWriter outStream = new FileWriter(fileName);
      out = new PrintWriter(new BufferedWriter(outStream));
    } catch (IOException exc) {
      System.err.println("Error: " + exc.getMessage());
      return null;
    }

    // and visit children
    data = node.childrenAccept(this, data);

    // close stream and reset handle
    close();

    // restore former stm context
    SimpleNode.stm.setContext(savedStmContext);

    return data;
  }


  public Object visit(ASTTDDUseDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("visit ASTTDDUseDeclaration");

    Vector used = node.getUsedFiles();
    for (Enumeration e = used.elements();
         e.hasMoreElements(); ) {
      emitUseClass((String)e.nextElement());
    }
    return data;
  }


  public Object visit(ASTADLClass node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLClass");

    LClass classScope = (LClass)((ADLClass)node.getScope()).getLClass();

    String fullName = classScope.getFullName();

    LClass[] parents = classScope.getSuperClasses();
    int nbSuper = parents.length;
    String[] extendsNames = new String[nbSuper];
    for (int i = 0; i < nbSuper; i++) {
      extendsNames[i] = parents[i].getName();
    }

    String modif = classScope.getModifier().toString();

    emitClassHeader(fullName, modif, extendsNames);

    emitClassEntity(node.getScope());

    if (node.jjtGetChild(0) instanceof ASTADLSuperClass) {
      emitClassInherited(((ASTADLSuperClass)node.jjtGetChild(0)).getSuperClasses());
    }

    int nbChildren = node.jjtGetNumChildren();
    for (int i = 0; i < nbChildren; i++) {
      if (node.jjtGetChild(i) instanceof ASTADLBhvDecl)
        node.jjtGetChild(i).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTADLBhvDecl node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLBhvDecl");

    int childIndex = 0;
    String resultType = "";
    String methodName;
    Node child = node.jjtGetChild(0), params = null, thrownExcs = null;

    if (child instanceof ASTCtorDeclarator) {
      constructor = true;
      methodName = ((ASTQualifiedId)child.jjtGetChild(0)).getFullName();
      if (child.jjtGetNumChildren() == 2)
        params = child.jjtGetChild(1);
      if (node.jjtGetChild(1) instanceof ASTExceptionSpec) {
        thrownExcs = node.jjtGetChild(1);
        childIndex++;
      }
    }
    else {
      constructor = false;
      if (child instanceof ASTDeclarationSpecifiers) {
        resultType += ((ASTDeclarationSpecifiers)child).getType().getName();
        child = node.jjtGetChild(1);
        childIndex++;
      }
      while (child instanceof ASTFunctionDeclarator) {
        resultType += ((ASTPtrOperator)child.jjtGetChild(0)).getPtrstring();
        child = child.jjtGetChild(1);
      }
      methodName = ((ASTQualifiedId)child.jjtGetChild(0)).getFullName();
      if (child.jjtGetNumChildren() > 1) {
        if ((child.jjtGetChild(1) instanceof ASTParameterDeclarationList) ||
            (child.jjtGetChild(1) instanceof ASTParameterDeclaration))
          params = child.jjtGetChild(1);
        if (child.jjtGetChild(child.jjtGetNumChildren()-1)
            instanceof ASTExceptionSpec)
          thrownExcs = child.jjtGetChild(child.jjtGetNumChildren()-1);
      }
    }

    beginMethodTemplate(methodName);

    emitMethodEntity(methodName, node.getScope());

    if (! constructor) {
      emitResultType(resultType);
    }


    if (params == null) {
      emitParam("", "");
    }
    else {
      if (params instanceof ASTParameterDeclaration) {
        LParameter p = ((ASTParameterDeclaration)params).getFctParameter();
        emitParam(p.getType().getName(), p.getName());
      }
      else { // params instanceof of ASTParameterDeclarationList
        LParameter[] pp = ((ASTParameterDeclarationList)params).getFctParameters();
        for (int i = 0; i < pp.length; i++) {
          emitParam(pp[i].getType().getName(), pp[i].getName());
        }
      }
    }


    if (thrownExcs != null) {
      if (thrownExcs.jjtGetNumChildren() == 0) {
        emitThrownList("");
      }
      else {
        thrownExcs = thrownExcs.jjtGetChild(0);
        String thrownList;
        if (thrownExcs instanceof ASTParameterDeclaration) {
          thrownList = ((ASTParameterDeclaration)thrownExcs).getFctParameter().getType().getName();
        }
        else { // thrownExcs instanceof of ASTParameterDeclarationList
          LParameter[] p = ((ASTParameterDeclarationList)thrownExcs).getFctParameters();
          thrownList = p[0].getType().getName();
          for (int i = 1; i < p.length; i++) {
            thrownList += ", " + p[i].getType().getName();
          }
        }
        emitThrownList(thrownList);
      }
    }

    endPrototype();


    while (! (child instanceof ASTADLBhvSpec)) {
      child = node.jjtGetChild(++childIndex);
    }

    child.jjtAccept(this, data);

    endMethodTemplate();

    return data;
  }


  public Object visit(ASTADLBhvSpec node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLBhvSpec");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLBhvClassification node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLBhvClassification");

    boolean normalFirst = node.getFirstChildKind().equals("NORMAL");
    if (normalFirst) {
      assertionBuffer = new StringBuffer("'ADL#Normal' = ");
    } else {
      assertionBuffer = new StringBuffer("'ADL#Abnormal' = ");
    }
    inAssertion = true;
    node.jjtGetChild(0).jjtAccept(this, data);
    inAssertion = false;
    emitNormal(query(assertionBuffer.toString(), node.getScope()), normalFirst);
    assertionBuffer = null;

    if (node.jjtGetNumChildren() > 1) {
      if (normalFirst) {
        assertionBuffer = new StringBuffer("'ADL#Abnormal' = ");
      } else {
        assertionBuffer = new StringBuffer("'ADL#Normal' = ");
      }
      inAssertion = true;
      node.jjtGetChild(1).jjtAccept(this, data);
      inAssertion = false;
      emitNormal(query(assertionBuffer.toString(), node.getScope()), ! normalFirst);
      assertionBuffer = null;
    }

    return data;
  }


  public Object visit(ASTADLAssGrp node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLAssGrp");

    if (assertionBuffer == null) {
      node.childrenAccept(this, data);
    }
    else {
      assertionBuffer.append("'ADL#Group'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        assertionBuffer.append(", ");
        node.jjtGetChild(i).jjtAccept(this, data);
      }
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADLInhSemantics node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLInhSemantics");

    emitSuperSemantics(node.getSuperClassName(), constructor);

    return data;
  }


  public Object visit(ASTADLBinding node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLBinding");

    return data;
  }


  public Object visit(ASTADLInlineDecl node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLInlineDecl");

    return data;
  }


  public Object visit(ASTADLIfStmt node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLIfStmt");

    if (assertionBuffer == null) {
      assertionBuffer = new StringBuffer();
      inAssertion = true;
      node.jjtGetChild(0).jjtAccept(this, data); // visit the condition
      inAssertion = false;
      emitCondition(query(assertionBuffer.toString(), node.getScope()),
                    ((SimpleNode)node.jjtGetChild(0)).getExpressionImage(),
                    node.jjtGetNumChildren() > 2);
      assertionBuffer = null;

      node.jjtGetChild(1).jjtAccept(this, data); // "then" branch
      if (node.jjtGetNumChildren() > 2) {
        emitElse();
        node.jjtGetChild(2).jjtAccept(this, data); // "else" branch
      }
      emitEndCondition();
    }
    else { // nested if
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(" '?' ");
      node.jjtGetChild(1).jjtAccept(this, data);
      assertionBuffer.append(" ':' ");
      if (node.jjtGetNumChildren() == 3)
        node.jjtGetChild(2).jjtAccept(this, data);
      else
        assertionBuffer.append("true");
    }
    return data;
  }


  public Object visit(ASTADLTryStmt node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLTryStmt");

    int nbChildren = node.jjtGetNumChildren();
    if (assertionBuffer == null) {
      emitTry((int)(nbChildren-1)/2);
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < nbChildren; i += 2) {
        ASTExceptionDeclaration exc = (ASTExceptionDeclaration)node.jjtGetChild(i);
        if (exc.jjtGetNumChildren() == 0) {
          emitCatch("...", "");
        }
        else if (exc.jjtGetChild(0) instanceof ASTParameterDeclaration) {
          LParameter p = ((ASTParameterDeclaration)exc.jjtGetChild(0)).getFctParameter();
          emitCatch(p.getType().getName(), p.getName());
        }
        else {
          LParameter[] pp = ((ASTParameterDeclarationList)exc.jjtGetChild(0)).getFctParameters();
          for (int j = 0; j < pp.length; j++) {
            emitCatches(pp[j].getType().getName(), pp[j].getName());
          }
        }
        node.jjtGetChild(i+1).jjtAccept(this, data);
      }
      emitEndTry();
    }
    else { // try in nested
      assertionBuffer.append("'ADL#TryCatch'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < nbChildren; i += 2) {
        assertionBuffer.append(", [['");
        ASTExceptionDeclaration exc = (ASTExceptionDeclaration)node.jjtGetChild(i);
        if (exc.jjtGetNumChildren() == 0) {
          assertionBuffer.append("...'], ");
        }
        else if (exc.jjtGetChild(0) instanceof ASTParameterDeclaration) {
          LParameter p = ((ASTParameterDeclaration)exc.jjtGetChild(0)).getFctParameter();
          assertionBuffer.append(p.getType().getName())
            .append("', '")
            .append(p.getName())
            .append("'], ");
        }
        else {
          LParameter[] pp = ((ASTParameterDeclarationList)exc.jjtGetChild(0)).getFctParameters();
          for (int j = 0; j < pp.length-1; j++) {
            assertionBuffer.append(pp[j].getType().getName())
              .append("', '")
              .append(pp[j].getName())
              .append("', '");
          }
          assertionBuffer.append(pp[pp.length-1].getType().getName())
              .append("', '")
              .append(pp[pp.length-1].getName())
            .append("'], ");
        }
        node.jjtGetChild(i+1).jjtAccept(this, data);
        assertionBuffer.append(']');
      }
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADLLabel node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLLabel");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLTags node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLTags");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLAssertion node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLAssertion");

    if (assertionBuffer == null) {
      String label = "";
      if (node.jjtGetChild(0) instanceof ASTADLLabel) {
        label = ((ASTADLLabel)node.jjtGetChild(0)).getName();
      }
      // omit tags???

      Node lastChild = node.jjtGetChild(node.jjtGetNumChildren()-1);
      if (lastChild instanceof ASTADLQuantifiedAssertion) {
        lastChild.jjtAccept(this, label);
      }
      else { // expression
        assertionBuffer = new StringBuffer();
        inAssertion = true;
        exceptionBhv = false;
        lastChild.jjtAccept(this, data);
        inAssertion = false;
        emitAssertion(query(assertionBuffer.toString(), node.getScope()),
                      label, node.getExpressionImage(), exceptionBhv);
        assertionBuffer = null;
      }
    }
    else { // nested assertion
      node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTADLImplExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLImplExpr");

    node.jjtGetChild(0).jjtAccept(this, data);

    String op = node.getOp();
    assertionBuffer.append(" '" + op + "' ");
    if (op.equals("<:>")) {
      exceptionBhv = true;
    }

    node.jjtGetChild(1).jjtAccept(this, data);

    return data;
  }


  public Object visit(ASTCondExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCondExpr");

    node.jjtGetChild(0).jjtAccept(this, data);
    if (inAssertion)
      assertionBuffer.append(" '?' ");
    node.jjtGetChild(1).jjtAccept(this, data);
    if (inAssertion)
      assertionBuffer.append(" ':' ");
    node.jjtGetChild(2).jjtAccept(this, data);

    return data;
  }


  public Object visit(ASTBinaryLogicalExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBinaryLogicalExpr");

    return unparseList(node, 0, node.getOperatorsVector(), data);
  }

  public Object visit(ASTBinaryArithExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBinaryArithExpr");

    return unparseList(node, 0, node.getOperatorsVector(), data);
  }

  public Object visit(ASTPmExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTPmExpr");

    return unparseList(node, 0, node.getOperatorsVector(), data);
  }


  public Object visit(ASTUnaryExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTUnaryExpr");

    if (inAssertion) {
      assertionBuffer.append(" '")
        .append(node.getOp())
        .append("' ");
    }
    // just one child
    return node.jjtGetChild(0).jjtAccept(this, data);
  }


  public Object visit(ASTADLCallStateExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLCallStateExpr");

    assertionBuffer.append("'@' ");
    node.jjtGetChild(0).jjtAccept(this, data);
    return data;
  }


  public Object visit(ASTNormalExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTNormalExpr");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Normal'");
    }
    return data;
  }


  public Object visit(ASTAbnormalExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTAbnormalExpr");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Abnormal'");
    }
    return data;
  }


  public Object visit(ASTUnchangedExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTUnchangedExpr");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Unchanged'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADLThrownExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLThrownExpr");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Thrown'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append("])");
    }

    return data;
  }


  public Object visit(ASTReturn node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTReturn");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Return'");
    }
    return data;
  }


  public Object visit(ASTTDDADLnewExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDADLnewExpr");

    emitACO(((ASTIdExpr)node.jjtGetChild(0)).getIdName());

    return data;
  }


  public Object visit(ASTTDDADLExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDADLExpr");

    emitACO(node.getType().getName());

    return data;
  }


  public Object visit(ASTThis node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTThis");

    if (inAssertion) {
      assertionBuffer.append("'ADL#This'");
    }
    return data;
  }


  public Object visit(ASTParenthExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTParenthExpr");

    if (inAssertion) {
      assertionBuffer.append('(');
      data = node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(')');
    }
    return data;
  }



  /*****************************************/

  /**
   * Unparse a list of <em>something</em> with a single separator.
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node. But we trust 
   * our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separator the string that separates children
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     String separator, Object data)
    throws ParseException {

    int numChildren = node.jjtGetNumChildren();

    // visit first child, then loop on remaining ones
    node.jjtGetChild(firstChild).jjtAccept(this, data);
    for (int childNum = firstChild + 1; childNum < numChildren; childNum++) {
      assertionBuffer.append(" '");
      assertionBuffer.append(separator);
      assertionBuffer.append("' ");
      node.jjtGetChild(childNum).jjtAccept(this, data);
    }
    return data;
  }

  /**
   * Unparse a list of <em>something</em> each having its own separator.
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node and when the size
   * of the array of separators does not match the number of children to
   * visit - 1. But we trust our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separators the vector of strings that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     Vector separators, Object data)
    throws ParseException {

    if (inAssertion) {
      // get number of children
      int numChildren = node.jjtGetNumChildren();

      // visit first child, then loop on remaining ones
      node.jjtGetChild(firstChild).jjtAccept(this, data);
      for (int childNum = firstChild + 1; childNum < numChildren; childNum++) {
        assertionBuffer.append(" '");
        assertionBuffer.append((String)separators.elementAt(childNum - 1));
        assertionBuffer.append("' ");
        node.jjtGetChild(childNum).jjtAccept(this, data);
      }
    }
    return data;
  }


  public Object visit(SimpleNode node, Object data) throws ParseException {

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTIncludeFileList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTIncludeFileList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTIncludeFileDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTIncludeFileDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTLinkageSpecification node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTLinkageSpecification");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDeclarationSpecifiers node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDeclarationSpecifiers");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTQualifiedId node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTQualifiedId");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTQualifiedType node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTQualifiedType");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTFullyScopedType node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTFullyScopedType");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTBuiltinTypeSpec node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBuiltinTypeSpec");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTInitDeclaratorList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTInitDeclaratorList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTInitDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTInitDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTClassSpecifier node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTClassSpecifier");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTClassMemberList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTClassMemberList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTBaseClause node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBaseClause");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTBaseSpecifier node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBaseSpecifier");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTAccessSpecifier node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTAccessSpecifier");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTMemberDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTMemberDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTMemberDeclaratorList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTMemberDeclaratorList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTEnumSpecifier node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTEnumSpecifier");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTEnumeratorList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTEnumeratorList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTEnumerator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTEnumerator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTPtrOperator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTPtrOperator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDirectDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDirectDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDeclaratorSuffixes node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDeclaratorSuffixes");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTFunctionDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTFunctionDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTFunctionDirectDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTFunctionDirectDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDtorDefinition node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDtorDefinition");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCtorDefinition node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCtorDefinition");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCtorDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCtorDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCtorInitializer node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCtorInitializer");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTSuperclassInit node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTSuperclassInit");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDtorDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDtorDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTParameterDeclarationList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTParameterDeclarationList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTParameterDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTParameterDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTInitializer node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTInitializer");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTypeNameList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTypeNameList");

    if (inAssertion) {
      assertionBuffer.append(((ASTTypeName)node.jjtGetChild(0)).getType().getName());
      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        assertionBuffer.append(", ")
          .append(((ASTTypeName)node.jjtGetChild(i)).getType().getName());
      }
    }
    return data;
  }


  public Object visit(ASTTypeName node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTypeName");
    // should not be visited!!!
    return data;
  }


  public Object visit(ASTAbstractDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTAbstractDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTAbstractDeclaratorSuffix node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTAbstractDeclaratorSuffix");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTStatement node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTStatement");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTLabelNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTLabelNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCaseExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCaseExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCompoundStatement node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCompoundStatement");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTElseNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTElseNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTIfNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTIfNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTSwitchNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTSwitchNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTWhileNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTWhileNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDoWhileNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDoWhileNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTForNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTForNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTGotoNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTGotoNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTContinueNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTContinueNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTBreakNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTBreakNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTryNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTryNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCatchNode node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCatchNode");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTExceptionDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTExceptionDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTThrowStmt node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTThrowStmt");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTAssignmentExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTAssignmentExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTConstantExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTConstantExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTCastExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTCastExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTNewExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTNewExpr");

    if (inAssertion &&
        (node.jjtGetChild(node.jjtGetNumChildren()-1) instanceof ASTArgumentList)) {
      LCallable constrScope = null;
      try {
        constrScope = (LCallable)node.getProperty("NLDScope");
      } catch(PropertyAccessException e){}
      assertionBuffer.append("'ADL#Func'(['")
        .append(constrScope.getName())
        .append(constrScope.getNLSignature())
        .append("', [")
        .append(nlengine.NLScopeScopeList(constrScope.getParent()))
        .append("]], [");
      node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);
      assertionBuffer.append("])");
      return data;
    }
    return node.childrenAccept(this, data);
  }


  public Object visit(ASTNewTypeId node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTNewTypeId");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTNewDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTNewDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDirectNewDeclarator node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDirectNewDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTDeleteExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDeleteExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTPostfixExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTPostfixExpr");

    Node prefix = node.jjtGetChild(0);
    String prefixName = null;
    if (prefix instanceof ASTIdExpr) {
      prefixName = ((ASTIdExpr)prefix).getIdName();
    }

    if (inAssertion) {
      Node child;
      int j, nbParenth = 0, nbChildren = node.jjtGetNumChildren();
      boolean comma = false, staticMethod = false;
      for (int i = nbChildren-1; i > 0; i--) {
        if (comma) {
          assertionBuffer.append(", ");
          comma = false;
        }

        child = node.jjtGetChild(i);
        if ((child instanceof ASTDotAccessSuffix) ||
            (child instanceof ASTRefAccessSuffix) ||
            (child instanceof ASTArraySuffix)) {
          child.jjtAccept(this, data);
          nbParenth += 1;
          comma = true;
        }
        else { // (child instanceof ASTArgumentList)
          LCallable methScope = null;
          j = i;
          do { j -= 1; }
          while (! ((node.jjtGetChild(j) instanceof ASTDotAccessSuffix) ||
                    (node.jjtGetChild(j) instanceof ASTRefAccessSuffix) ||
                    (j == 0)));// should never happen!!!
          try {
            methScope = (LCallable)((SimpleNode)node.jjtGetChild(j))
              .getProperty("NLDScope");
          }
          catch(PropertyAccessException e){
            if (debug)
              System.err.println("No NLDScope property");
          }
          staticMethod = (nbChildren == 2) &&
            ((CModifier)methScope.getModifier()).isStatic();
          i -= 1; // ignore the preceding access suffix
          assertionBuffer.append("'ADL#Func'(");
          assertionBuffer.append("['")
            .append(methScope.getName())
            .append(methScope.getNLSignature())
            .append("', [")
            .append(nlengine.NLScopeScopeList(methScope.getParent()))
            .append("]]");
          assertionBuffer.append(", [");
          child.jjtAccept(this, data);
          assertionBuffer.append("])");
          comma = true;
        }
      }

      if (node.jjtGetChild(1) instanceof ASTArgumentList) {
        if (! (staticMethod || (prefixName.indexOf(':') == -1))) {
          // Should never happen!!!
          assertionBuffer.append(", '")
            .append(prefixName.substring(0, prefixName.lastIndexOf("::")))
            .append("'");
          throw new ParseException("Internal Error");
        }
      }
      else {
        if (comma)
          assertionBuffer.append(", ");
        node.jjtGetChild(0).jjtAccept(this, data);
      }

      for (int i = nbParenth; i > 0; i--) {
        assertionBuffer.append(")");
      }

      return data;
    } // end if (inAssertion)

    else if (inTddDef && (prefixName != null)) {
      Scope exprScope = null;
      try {
        exprScope = (Scope)((SimpleNode)node.jjtGetChild(0)).getProperty("NLDScope");
      } catch(PropertyAccessException e) {
}
      if ((exprScope instanceof LMethod) &&
          (exprScope.getParent() instanceof TDDClass)) {
        emitTestFunction(((LMethod)exprScope).getName(),
                         ((LMethod)exprScope).getNLSignature());
      }
    }

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTPrimaryExpr node, Object data) throws ParseException {
    
    if (debug)
      System.out.println("Visiting ASTPrimaryExpr");
    
    Node prefix = node.jjtGetChild(0);
    String prefixName = null;
    if (prefix instanceof ASTIdExpr) {
      prefixName = ((ASTIdExpr)prefix).getIdName();
    }
    if (inAssertion) {
      nbParenth = 0;
      boolean staticMethod;
      if (prefix instanceof ASTIdExpr) {
        LCallable methScope = null;
        try {
          methScope = (LCallable)node.getProperty("NLDScope");
        }
        catch(PropertyAccessException e){}
        staticMethod = ((CModifier)methScope.getModifier()).isStatic();
        if ((! staticMethod) &&
            ((prefixName == null) || (prefixName.indexOf(':') != -1))) {
          assertionBuffer.append("'ADL#Object'(");
          nbParenth += 1;
        }
        assertionBuffer.append("'ADL#Func'(");
        assertionBuffer.append("['")
          .append(methScope.getName())
          .append(methScope.getNLSignature())
          .append("', [")
          .append(nlengine.NLScopeScopeList(methScope.getParent()))
          .append("]]");
        assertionBuffer.append(", [");
        node.jjtGetChild(1).jjtAccept(this, data);
        assertionBuffer.append("])");
        if ((! staticMethod) && (prefixName.indexOf(':') != -1)) {
          assertionBuffer.append(", '")
            .append(prefixName.substring(0, prefixName.lastIndexOf("::")))
            .append("'");
        }
        for (int i = nbParenth; i > 0; i--) {
          assertionBuffer.append(")");
        }

        return data;
      } // end if (prefix instanceof ASTIdExpr)
    } // end if (inAssertion)

    else if (inTddDef && (prefixName != null)) {
      Scope exprScope = null;
      try {
        exprScope = (Scope)node.getProperty("NLDScope");
      } catch(PropertyAccessException e) {}
      if ((exprScope instanceof LMethod) &&
          (((LMethod)exprScope).isTestMethod())) {// &&
//           (! ((LMethod)exprScope).isRuntimeMethod())) {
        emitTestFunction(((LMethod)exprScope).getName(),
                         ((LMethod)exprScope).getNLSignature());
      }
    }

    return node.childrenAccept(this, data);
  }


  private void fctCall(SimpleNode node) throws ParseException {

    // node (IdExpr, ArgumentList)
    LCallable methScope = null;
    try {
      methScope = (LCallable)node.getProperty("NLDScope");
    }
    catch(PropertyAccessException e){
      if (debug)
        System.err.println("No NLDScope property");
    }
    boolean staticMethod = ((CModifier)methScope.getModifier()).isStatic();
    if (! staticMethod) {
      assertionBuffer.append("'ADL#Object'(");
      nbParenth += 1;
    }
    assertionBuffer.append("'ADL#Func'(");
    assertionBuffer.append("['")
      .append(methScope.getName())
      .append(methScope.getNLSignature())
      .append("', [")
      .append(nlengine.NLScopeScopeList(methScope.getParent()))
      .append("]]");
    assertionBuffer.append(", [");
    node.jjtGetChild(1).jjtAccept(this, null);
    assertionBuffer.append("])");
  }

  public Object visit(ASTDotAccessSuffix node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTDotAccessSuffix");

    if (node.jjtGetNumChildren() == 1) {
      assertionBuffer.append("'ADL#Field'('" + ((ASTIdExpr)node.jjtGetChild(0)).getIdName() + "', ");
    }
    else {
      fctCall(node);
    }

    return data;
  }


  public Object visit(ASTRefAccessSuffix node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTRefAccessSuffix");

    if (inAssertion) {
      if (node.jjtGetNumChildren() == 1) {
        assertionBuffer.append("'ADL#FieldPtr'('" + ((ASTIdExpr)node.jjtGetChild(0)).getIdName() + "', ");
      }
      else {
        fctCall(node);
      }
    }

    return data;
  }


  public Object visit(ASTArraySuffix node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTArraySuffix");

    assertionBuffer.append("'ADL#Array'(");
    node.jjtGetChild(0).jjtAccept(this, data);
    assertionBuffer.append(", ");

    return data;
  }


  public Object visit(ASTPostDeIncrement node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTPostDeIncrement");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTArgumentList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTArgumentList");

    if (inAssertion) {
      // visit first child, then loop on remaining ones
      int nbChildren = node.jjtGetNumChildren();
      if (nbChildren > 0) {
        node.jjtGetChild(0).jjtAccept(this, data);
        for (int i = 1; i < nbChildren; i++) {
          assertionBuffer.append(", ");
          node.jjtGetChild(i).jjtAccept(this, data);
        }
      }
    }
    return data;
  }


  public Object visit(ASTIdExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTIdExpr");

    // A name is just a sequence of ids separated by "."
    // the full sequence is accessible via getName()
    if (inAssertion) {
      assertionBuffer.append("'").append(node.getIdName()).append("'");
    }
    else if (inTddDef) {
      String cstClass = null, ret;
      try {
        cstClass = (String)node.getProperty("enclosingConstructName");
        if (cstClass.equals(className)) {
          ret = makeLocalLink(className, "TDD-CSTS", node.getIdName());
        }
        else {
          ret = makeTDDExternalLink(cstClass, cstClass + "TDD-CSTS", node.getIdName());
        }
      } catch(PropertyAccessException e) {
        ret = node.getIdName();
      }
      return ret;
      //      return makeLocalLink(className, "TDD-CSTS", node.getName());
    }

    return data;
  }


  public Object visit(ASTReturnStmt node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTReturnStmt");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTExpr");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTConstant node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTConstant");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      String val = node.getConstantValue();
      // Prolog does not recognize final letter ('L', 'D', etc.)
      if (Character.isDigit(val.charAt(0)) &&
          Character.isLetter(val.charAt(val.length()-1))) {
        val = val.substring(0, val.length()-1);
      }
      if (inAssertion)
        assertionBuffer.append(val);
      else return val;
    }
    return data;
  }


  public Object visit(ASTString node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTString");

    if (inAssertion || inTddDef) {
      String val = node.getName();
      val = "\'\"" + val/*.substring(1, val.length()-1)*/ + "\"\'";
      if (inAssertion)
        assertionBuffer.append(val);
      else return val;
    }
    return data;
  }


  public Object visit(ASTExceptionSpec node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTExceptionSpec");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLNamedParamList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLNamedParamList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLSuperClass node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLSuperClass");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLQuantifiedAssertion node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLQuantifiedAssertion");

    String label = (data == null) ? "" : (String)data;
    ASTADLDomain dom;
    ASTADLDomainList domlist = (ASTADLDomainList)node.jjtGetChild(0);

    emitQuantifier(label, node.getQuantifier(),
		   domlist.getExpressionImage());

    String name, expr;
    for (int i = 0; i < domlist.jjtGetNumChildren(); i++) {
      dom = (ASTADLDomain)domlist.jjtGetChild(i);
      inAssertion = true; assertionBuffer = new StringBuffer();
      dom.jjtGetChild(1).jjtAccept(this, null);
      expr = query(assertionBuffer.toString(), node.getScope());
      emitDomain(((ASTDeclaration)dom.jjtGetChild(0)).getExpressionDecl().getName(),
                 expr, i);
      inAssertion = false; assertionBuffer = null;
    }
    node.jjtGetChild(1).jjtAccept(this, null);
    emitEndQuantified();

    return data;
  }


  public Object visit(ASTADLDomainList node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLDomainList");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLDomain node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLDomain");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLProlog node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLProlog");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADLEpilog node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTADLEpilog");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDDClassDecl node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDClassDecl");

    emitClassEntity(node.getScope());

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDDFieldDeclaration node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDFieldDeclaration");

    Node child; String name;
    inTddDef = true;

    Vector types = node.getTypes();
    for (int i = 0; i < types.size(); i++) {
      child = node.jjtGetChild(2*i+1);
      while ((! (child instanceof ASTQualifiedId)) && child.jjtGetNumChildren() > 0)
        child = child.jjtGetChild(0);
      if (child instanceof ASTQualifiedId) 
        name = ((ASTQualifiedId)child).getName();
      else
        name = "~" + ((ASTDirectDeclarator)child).getDestructorName();
      emitConstant(((LType)types.elementAt(i)).getName(),
                   name,
                   (String)node.jjtGetChild(2*i+2).jjtAccept(this, data));
    }

    inTddDef = false;
    return data;
  }


  public Object visit(ASTTDDDatasetDecl node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetDecl");

    TDDDataset ds = node.getDataset();
    String expr;
    inTddDef = true;
    expr = (String)node.jjtGetChild(1).jjtAccept(this, data);
    inTddDef = false;

    emitDataset(ds.getName(), ds.getType().getName(), expr); 

    return data;
  }


  public Object visit(ASTTDDFactoryDef node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDFactoryDef");

    Node fctDecl = node.jjtGetChild(1), params = null, thrownExcs = null;
    while (! (fctDecl instanceof ASTFunctionDirectDeclarator))
      fctDecl = fctDecl.jjtGetChild(1);
    String factName = ((ASTQualifiedId)fctDecl.jjtGetChild(0)).getFullName();

    beginFactoryTemplate(factName);
    emitMethodEntity(factName, node.getFactoryScope());

    emitResultType(node.getFactoryType().getName());

    if (fctDecl.jjtGetNumChildren() > 1) {
      if ((fctDecl.jjtGetChild(1) instanceof ASTParameterDeclaration) ||
          (fctDecl.jjtGetChild(1) instanceof ASTParameterDeclarationList))
        params = fctDecl.jjtGetChild(1);
      if (fctDecl.jjtGetChild(fctDecl.jjtGetNumChildren()-1)
          instanceof ASTExceptionSpec)
        thrownExcs = fctDecl.jjtGetChild(fctDecl.jjtGetNumChildren()-1);
    }

    if (params == null) {
      emitParam("", "");
    }
    else {
      if (params instanceof ASTParameterDeclaration) {
        LParameter p = ((ASTParameterDeclaration)params).getFctParameter();
        emitParam(p.getType().getName(), p.getName());
      }
      else { // params instanceof of ASTParameterDeclarationList
        LParameter[] pp = ((ASTParameterDeclarationList)params).getFctParameters();
        for (int i = 0; i < pp.length; i++) {
          emitParam(pp[i].getType().getName(), pp[i].getName());
        }
      }
    }

    if (thrownExcs != null) {
      if (thrownExcs.jjtGetNumChildren() == 0) {
        emitThrownList("");
      }
      else {
        thrownExcs = thrownExcs.jjtGetChild(0);
        String thrownList;
        if (thrownExcs instanceof ASTParameterDeclaration) {
          thrownList = ((ASTParameterDeclaration)thrownExcs).getFctParameter().getType().getName();
        }
        else { // thrownExcs instanceof of ASTParameterDeclarationList
          LParameter[] p = ((ASTParameterDeclarationList)thrownExcs).getFctParameters();
          thrownList = p[0].getType().getName();
          for (int i = 1; i < p.length; i++) {
            thrownList += ", " + p[i].getType().getName();
          }
        }
        emitThrownList(thrownList);
      }
    }

    emitCode(node.getSourceText());

    endFactoryTemplate();

    return data;
  }


  public Object visit(ASTFunctionDefinition node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTFunctionDefinition");
    String funcName, resultType = null;
    Node child = node.jjtGetChild(0), params = null, thrownExcs = null;

    if (child instanceof ASTDeclarationSpecifiers) {
      resultType += ((ASTDeclarationSpecifiers)child).getType().getName();
      child = node.jjtGetChild(1);
    }
    while (child instanceof ASTFunctionDeclarator) {
      resultType += ((ASTPtrOperator)child.jjtGetChild(0)).getPtrstring();
      child = child.jjtGetChild(1);
    }
    funcName = ((ASTQualifiedId)child.jjtGetChild(0)).getFullName();
    if (child.jjtGetNumChildren() > 1) {
      if ((child.jjtGetChild(1) instanceof ASTParameterDeclarationList) ||
          (child.jjtGetChild(1) instanceof ASTParameterDeclaration))
        params = child.jjtGetChild(1);
      if (child.jjtGetChild(child.jjtGetNumChildren()-1)
          instanceof ASTExceptionSpec)
        thrownExcs = child.jjtGetChild(child.jjtGetNumChildren()-1);
    }

    beginTestFuncTemplate(funcName,
                          node.storedFunction.getNLSignature());
    emitMethodEntity(funcName, node.storedFunction);
    emitResultType(resultType);

    if (params == null) {
      emitParam("", "");
    }
    else {
      if (params instanceof ASTParameterDeclaration) {
        LParameter p = ((ASTParameterDeclaration)params).getFctParameter();
        emitParam(p.getType().getName(), p.getName());
      }
      else { // params instanceof of ASTParameterDeclarationList
        LParameter[] pp = ((ASTParameterDeclarationList)params).getFctParameters();
        for (int i = 0; i < pp.length; i++) {
          emitParam(pp[i].getType().getName(), pp[i].getName());
        }
      }
    }

    if (thrownExcs != null) {
      if (thrownExcs.jjtGetNumChildren() == 0) {
        emitThrownList("");
      }
      else {
        thrownExcs = thrownExcs.jjtGetChild(0);
        String thrownList;
        if (thrownExcs instanceof ASTParameterDeclaration) {
          thrownList = ((ASTParameterDeclaration)thrownExcs).getFctParameter().getType().getName();
        }
        else { // thrownExcs instanceof of ASTParameterDeclarationList
          LParameter[] p = ((ASTParameterDeclarationList)thrownExcs).getFctParameters();
          thrownList = p[0].getType().getName();
          for (int i = 1; i < p.length; i++) {
            thrownList += ", " + p[i].getType().getName();
          }
        }
        emitThrownList(thrownList);
      }
    }

    emitCode(node.getSourceText());

    //    node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);

    endTestFuncTemplate();

    return data;
  }


  public Object visit(ASTTDDTestDirective node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDTestDirective");

    String dirName;
    dirName = node.getLabelName();
    if (dirName.equals("")) {
      dirName = "directive_" + anonDirectiveIndex;
      anonDirectiveIndex++;
    }
    beginDirectiveTemplate(dirName);

    int nbChildren = node.jjtGetNumChildren();
    for(int i = 0; i < nbChildren-1; i++) {
      node.jjtGetChild(i).jjtAccept(this, data);
    }

    emitCode(node.getSourceText());

    inTddDef = true;
    node.jjtGetChild(nbChildren-1).jjtAccept(this, data);
    inTddDef = false;

    endDirectiveTemplate();

    return data;
  }


  public Object visit(ASTTDDDatasetDomain node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetDomain");

    inTddDef = true;
    String ds = (String)node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);
    inTddDef = false;

    emitDomain(ds);

    return data;
  }


  public Object visit(ASTTDDDatasetLiteral node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetLiteral");

    String ret = "{ ";
    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 0) {
      return ret + "}";
    }
    ret += (String)node.jjtGetChild(0).jjtAccept(this, data);
    for (int i = 1; i < nbChildren; i++) {
      ret += ", " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }
    return ret + " }";
  }


  public Object visit(ASTTDDDatasetMember node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetMember");

    String ret = (String)node.jjtGetChild(0).jjtAccept(this, data);
    if (node.jjtGetNumChildren() == 2)
      ret += " .. " +
        (String)node.jjtGetChild(1).jjtAccept(this, data);

    return ret;
  }


  public Object visit(ASTTDDDatasetExpr node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetExpr");

    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 1) {
      return node.jjtGetChild(0).jjtAccept(this, data);
    }

    String ret = (String)node.jjtGetChild(0).jjtAccept(this, data);
    for (int i = 1; i < nbChildren; i++) {
      ret += " + " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }

    return ret;
  }


  public Object visit(ASTTDDDatasetSingleton node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDDatasetSingleton");

    if (node.jjtGetChild(0) instanceof ASTIdExpr) {
      ASTIdExpr child = (ASTIdExpr)node.jjtGetChild(0);
      String dsName = child.getIdName(), dsClass = null, kind = null, ret;

      try {
        dsClass = (String)child.getProperty("enclosingConstructName");
      } catch(PropertyAccessException e) {}

      if (dsClass.equals(className)) {
        if (child.isDataset)
          ret = makeLocalLink(dsName, dsName);
        else
          ret = makeLocalLink(className, "TDD-CSTS", dsName);
      }
      else {
        if (child.isDataset)
          ret = makeTDDExternalLink(dsClass, dsName, dsName);
        else
          ret = makeTDDExternalLink(dsClass, dsClass + "TDD-CSTS", dsName);
      }
      return ret;
    }
    else {
      node.jjtGetChild(0).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTTDDFactoryCall node, Object data) throws ParseException {

    if (debug)
      System.out.println("Visiting ASTTDDFactoryCall");

    String factClass = null;

    try {
      factClass = (String)((SimpleNode)node.jjtGetChild(0)).getProperty("enclosingConstructName");
    } catch(PropertyAccessException e) {}

    String factName = ((ASTQualifiedId)node.jjtGetChild(0)).getName(); 

    String ret;
    if (factClass.equals(className)) {
      ret = makeLocalLink(factName, factName) + " ( ";
    }
    else {
      ret = makeTDDExternalLink(factClass, factName, factName) + " ( ";
    }

    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 1) {
      return ret + ")";
    }
    ret += (String)node.jjtGetChild(1).jjtAccept(this, data);
    for (int i = 2; i < nbChildren; i++) {
      ret += ", " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }
    return ret + " )";
  }

}

// ********************************************************************** {
// History Log
// $Log: CPPNLDUnparserVisitor.java,v $
// Revision 1.10  1998/08/06 14:10:58  mallet
// Merged branch v1_2-patches back onto the trunk.
//
// Revision 1.9.2.1  1998/07/08 13:54:58  mallet
// Fixed bug #143: having a field which identifier starts with an
// uppercase letter causes the Prolog engine to enter an infinite
// loop. Simple quotes have been added around an ADL#Field identifier.
//
// Revision 1.9  1998/06/30 18:23:48  bareau
// Unparsing of ADLnew expressions and allocation expressions.
//
// Revision 1.8  1998/06/30 15:57:49  mallet
// Replaced OutputStreamWriter over FileOutputStream by FileWriter.
// Replaced access to expressionImage property by call to
// getExpressionImage() or getSourceText() depending on usage in
// generated documentation.
//
// Revision 1.7  1998/06/15 13:14:28  bareau
// Addes support for quantified assertions.
//
// Revision 1.6  1998/06/12 16:19:18  bareau
// Minor modifications related to the changes in the grammar at production
// DatasetSingleton.
//
// Revision 1.5  1998/06/11 16:06:15  bareau
// Added support for nested if expressions (they are considered exactly
// like "?:" conditional expressions).
//
// Revision 1.4  1998/06/11 13:48:13  bareau
// Added support for nested try blocks.
//
// Revision 1.3  1998/06/11 10:21:55  bareau
// Added support for nested assertion groups.
//
// Revision 1.2  1998/06/10 15:57:54  bareau
// Added processing for documentation generation for tdd classes.
// Several minor bugs fixes for adl processing.
//
// Revision 1.1  1998/06/05 14:10:44  bareau
// NLD unparser for C++, initial revision.
//
// ***********************************************************************
