// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/cpp/CPPUnparserVisitor.java,v $
// $Revision: 1.23 $
// $Date: 1998/06/30 15:53:38 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/Java Compiler
// NAME:        CPPUnparserVisitor.java
//
// HISTORY:
//    Wassel BEARRAYANA, The Open Group RI - December, 97: Created
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.cpp;

import java.io.*;
import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.parsers.ParseException;
import ORG.opengroup.adl.parsers.adlcpp.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.c.*;
import java.util.Vector;
import java.util.Enumeration ;


/**
 * class CPPUnparserVisitor
 *
 * The CPPUnparserVisitor class defines a visitor which can be applied
 * to an AST generated from the ADLCPPParser grammar. It produces pure
 * C++ source files by asserting that no ADL*, TDD* and NLD* nodes
 * remain in the final AST. 
 *
 * <p> The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and emit the code on
 * an output stream maintained by the visitor itself.
 *
 * <p> The unparsing starts by calling the <a href="#unparse"> unparse()</a> 
 * method passing the top of AST to unparse as parameter.
 *
 */
public class CPPUnparserVisitor extends GenericUnparserVisitor
implements ADLCPPParserVisitor
{
  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.23 $";

  /**
   * Default constructor: call superclass constructor
   *
   * @param toolname the name of the tool using this unparser visitor
   */   
  public CPPUnparserVisitor(String toolname) {
    super(toolname, versionId);
  }

  /**
   * Basic constructor specifying unparser parameters: call superclass
   * constructor.
   *
   * @param toolname the name of the tool using this unparser visitor
   * @param linelimit the desired maximum length of lines in output file
   * @param regindent the regular indentation level
   * @param subindent the sub-expression indentation level.
   */
  public CPPUnparserVisitor(String toolname, int linelimit, int regindent,
		   int subindent) {
    super(toolname, versionId, linelimit, regindent, subindent);
  }

  /**
   * Starts unparsing an AST.
   *
   *If the passed AST is not TranslationUnit, an error is reported and the AST
   * is not unparsed. Otherwise the AST is visited.
   *
   * @param ast The root node of the AST to unparse
   * @see #visit(ORG.opengroup.adl.parsers.adljava.ASTTranslationUnit,java.lang.Object)
   */
  public void unparse(Node ast) throws ParseException {
  
    if (assertNodeType(ast, ASTTranslationUnit.class)) {
      ( (ASTTranslationUnit)ast).jjtAccept(this, null);
    } else {
      incorrectNode(ast);
      System.err.println("\twas expecting TranslationUnit");
    }
  }


  /**
   * Method making easier separated display of children 
   *
   * <p> visit children of the node and separate them with a separator String
   * @param node the node to visit
   * @param separator String emitted between node children
   */ 
  public Object childrenEmit ( SimpleNode node, String separator ) throws ParseException {
    Object data = null;
    int childNum = node.jjtGetNumChildren();
    if( childNum == 0)
      return data;

    for ( int i =0; i< childNum -1; i++ ) {
      data = node.jjtGetChild(i).jjtAccept( this, data );
      emit(separator);
    }
    // visit last child 
    data = node.jjtGetChild(childNum -1).jjtAccept( this, data );
    return data;
  }
 


  /** Default visit for SimplenOde. */

  public Object visit(SimpleNode node, Object data) throws ParseException {
   return  node.childrenAccept(this, data);
  }


  /**
   * Visits an TranslationUnit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the name of the declared class saved in the node
   * <li>open the destination file (<code>outputDirectory</code>/<em>name</em>
   * .hh or cpp),
   * <li> reset current column and indentation,
   * <li> emit the standard header,
   * <li>visit children nodes,
   * <li> and finally close the file.
   * </ul>
   *
   * @param node the TranslationUnit to visit
   * @param data unused. Always null
   * @return data returned by last child visited or
   * null if the file cannot be opened.
   * @exception ORG.opengroup.adl.parsers.cpp.ParseException thrown when parse
   * error occur.
   * @see ORG.opengroup.adl.unparsers.GenericUnparserVisitor#outputDirectory
   */

  public Object visit(ASTTranslationUnit node, Object data)
   throws ParseException {

    // get the name of the declared class and open the corresponding stream.
    // get file name stored as a node property.
    String className = node.getName();
    // get from node an indication if we should generate a file.cc or file.hh
    String fileExtension = node.getNodeExtension();

    // try to open the output stream
    try {
      String fileName = outputDirectory + File.separator + className +
	"." + fileExtension;

      if (Adlt.verboseMode()) {
	System.out.println("+++ generating " + fileName);
      }

      FileWriter outStream = new FileWriter(fileName);
      out = new PrintWriter(new BufferedWriter(outStream));
    } catch (IOException exc) {
      System.err.println("Error: " + exc.getMessage());
      return null;
    }

    // now reset currentColumn and currentIndent
    currentColumn = 0;
    setIndent(0);
    // emit general comment
    emit(headerString);
    newline();
    newline();
    // if if not def lines.
    String defineName = className+"_"+ fileExtension ;
    emit("#ifndef "+ defineName);
    newline();
    emit("#define "+defineName) ;
    newline() ;
    newline();
    // Visit children
    data = node.childrenAccept(this, data );
    newline() ;
    newline();
    // emit end if
    // close stream and reset handle
    emit("#endif ");
    newline() ;
    out.close();
    out = null;
    return data;
  }

  /**
   * Visits an IncludeFileList.
   * <p><b>Algorithm:</b><ul>
   * Simply viist children.
   * </ul>
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTIncludeFileList node, Object data)
  throws ParseException {
    data= node.childrenAccept( this , data );
    newline() ;
    return data ;
  }



  /**
   * Visits an IncludeFileDeclaration.
   *
   * <p> This node handles the included filenames
   * <p><b>Algorithm:</b><ul>
   * <li> emit the "# include" keywords
   * <li> emit the file name saved as a node property.
   * <li> emit " ;"
   * </ul>
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTIncludeFileDeclaration node, Object data) throws ParseException {
    // # include line, emit keyword , then emit filename ";"
    emit("#include ");
    String filename = node.getFileName();
    emit( filename  );
    newline();

    return data;
  }



  /** 
   * Visits Declaration Node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> visits children.
   * set Indent if node parent is not a ForNode.
   * <li> emits ";" and a new line at the end of Declaration if node parent is 
   * not ForNode
   *</ul>
   * @param node the Declaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTDeclaration node, Object data) throws ParseException {

    if ( ! ( node.jjtGetParent() instanceof ASTForNode ) )
      indent();
    // Visit children
    node.childrenAccept( this,data);
    emit(";");
    if ( ! ( node.jjtGetParent() instanceof ASTForNode ) )
      newline() ;

    return data;
  }


  /** 
   * Visits LinkageSpecification Node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Set indent and emit keyword <code>extern</code>.
   * <li> get and emit the linkage style.
   * <li> Emit "{" then add indent.
   * <li> visit children and emit "}".
   * </ul>
   * @param node the LinkageSpecification to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTLinkageSpecification node, Object data)
  throws ParseException { 

    indent();
    emit("extern");
    emit( node.getStyle()+" {" );
    newline();
    addIndent(regIndent) ;
    // Visit children
    node.childrenAccept( this,data);
    newline() ;
    delIndent( regIndent ) ;
    indent();
    emit("} ;");
    newline();

    return data;
  }


  /**
   * Visits DeclarationSpecifiers.
   * 
   *<p>  DeclarationSpecifeirs have 5 kinds of child specifying the type: Fully
   *ScopedType, ClassSpecifier, QualifiedType, EnumSpecifier or BuiltinTypes.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> get declaration modifier saved as an integer within the node.
   * <li> Construct an opengroup.adl.symboltable.c.CModifier  and uses its
   * toString() method to emit the node modifier.  
   * <li> If type children are builtinTypes, emit node builtin type name saved
   * as a property during parsing time. When such property not found loop on
   * children ( all builtintypes in this case ) to emit their type names. 
   * <li> Else, visit the unique node child. 
   * </ul>
   *
   * @param node the DeclarationSpecifiers to visit
   * @param data unused
   * @return data returned by the visited child
   */
   
  public Object visit( ASTDeclarationSpecifiers node, Object data)
    throws ParseException {

    int modif;
    int nbchild = node.jjtGetNumChildren();

    // display modifiers first
    modif = node.getModifier();
    CModifier Cmodif = new CModifier( modif);
    emit ( Cmodif.toString());
    if ( nbchild == 0)
      return data;

    Node child0 = node.jjtGetChild(0);
    if( child0 instanceof ASTBuiltinTypeSpec ) {
      // in this case all children are builtintypes
      // see if node tells the recovered builtin type at parsing time
      String builtName = node.getBuiltinTypeName();
      if( builtName != null) {
	emit( builtName + " ");
      }
      else {
	// loop on all builtintypes and display them
	for (int i =0; i< nbchild;i++ ) {
	  Node child = node.jjtGetChild(i);
	  if ( child instanceof ASTBuiltinTypeSpec )
	    emit( ( (ASTBuiltinTypeSpec)child).getTypeName() + " ");
	}
      }
    }
    else {
      // in all the other cases visit the child
      child0.jjtAccept(this,data);
    }
    return data;

  }


  /**
   * Visits EnumSpecifier.
   *
   * <p> Node has 1 or 0 children.
   * <p><b>Algorithm:</b><ul>
   * <li> emit "enum" keyword.
   * <li> emit Enum name node property. may be empty string if not-named enum.
   * <li> If child Enumaerator or EnumeratorList exists then emit "{", add
   * indent, visit the child, delete the added indent and emit"{"
   * </ul>
   *
   * @param node EnumSpecifier to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit( ASTEnumSpecifier node, Object data)
    throws ParseException {
    // get number of children => should be either 0 or 1
    int nbchild = node.jjtGetNumChildren();
    emit("enum ");
    String enumName = node.getEnumName();
    emit(enumName+ " ");
    
    if( nbchild ==1 ) {

      Node child0 = node.jjtGetChild(0);
      emit( "{");
      // set Indent to have enumerators displayed one a line
      addIndent( regIndent);
      // add new line
      newline();
      // visit child0. It is either an Enumerator or EnumeratorList
      child0.jjtAccept(this,data);
      delIndent( regIndent);
      indent();
      emit("} ");
    }

    return data;
  }
  
  /**
   * Visits EnumeratorList node.
   * <p> Node has at least 2 children.
   * <p><b>Algorithm:</b><ul>
   * <li> loop to visit all children
   * <li> emit "," between each 2 children.
   * </ul>
   * @param node EnumeratorList to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit( ASTEnumeratorList node, Object data) throws ParseException {
    Node child;
    int nbchild = node.jjtGetNumChildren();    
    // when created, ASTEnumeratorList has at least 1 child
    for ( int i= 0; i < nbchild -1; i++ ) {
      child = node.jjtGetChild(i);
      child.jjtAccept ( this, data);
      emit (", ");
      newline();
    }

    child = node.jjtGetChild( nbchild- 1);
    child.jjtAccept ( this, data);
    newline();
    //delete the added indent
    
    return data;

  }

  /** 
   * Visits the Enumarator node.
   *
   * <p> Enumerator has 0/1 child. Child exists when initialized  enumerator.
   * <p><b>Algorithm:</b><ul>
   * <li> emit the enumerator ID name saved in the node.
   * <li> Emit "=" and Visit the child if exits.
   * </ul>
   * @param node the Enumerator to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit( ASTEnumerator node, Object data) throws ParseException {
   
    int nbchild = node.jjtGetNumChildren();
    indent();
    emit( node.getEnumId() + " ");

    if ( nbchild > 0) {

      emit("= ");
      Node child0 = node.jjtGetChild(0);
      // visit the initialization expression
      child0.jjtAccept( this, data);
    }

    return data;
  }

  /**
   * Visits ClassSpecifier.
   * <p> Node may have no children : case of Elaborated type name
   * or 1 child in case  class definition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> emit the class kind that may be "class:, "struct", or union.
   * <li> <ul>
   *   <li>In case of C style struct definition ( when name starts with
   * "?adl_anonymous_struct"), do not emit the name.
   *   <li> Else, emit the class/struct/union name saved within the node.
   * </ul>
   * <li> visit child ClassMemberList when exits.
   * </ul>
   *
   * @param node the ClassSpecifier to visit
   * @param data unused
   * @return data returned by the visited child
   */

  public Object visit( ASTClassSpecifier node, Object data) throws ParseException {
    // get number of children => should be either 0 or 1
    int nbchild = node.jjtGetNumChildren();
    String classKind = node.getClassKind();
 

    emit(classKind + " ");
    // display class name
    String clazz = node.getName();
    if(!  clazz.startsWith("?adl_anonymous_struct")  )
       emit( clazz +" ");
    data = node.childrenAccept( this, data);

    return data;
  }


  /**
   * Visits ClassMemberList.
   *
   * <p>Node ClassMemberList tells super ClassSpecifier is a class definition.
   * Children are the class members.
   * <p><b>Algorithm:</b><ul>
   * <li> emits "{", adds an indent and a new line.
   * <li> visits children
   * <li> delete added indent and emit "}".
   * </ul>
   *
   * @param node the ClassMemberList to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit( ASTClassMemberList node , Object data ) throws ParseException {
    emit("{");
    addIndent( regIndent );
    newline();
    // visit children 
    data = node.childrenAccept( this, data);
    delIndent( regIndent);
    newline();
    indent();
    emit("} ");

    return data;
  }


  /**
   * Visits BaseClause.
   * <p> Node BaseClause has at least 2 children.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> emits C++ inheritance keyword ":"
   * <li>Emit children ( calling <a href="#childrenEmit"> childrenEmit</a> )
   * with ", " separating them.
   * </ul>
   * @param node the BaseClause to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit( ASTBaseClause node, Object data) throws ParseException {
  
    Node child;
    emit(" : ");
    childrenEmit( node , ", ");
    return data;
  }

  /**
   * Visits BaseSpecifier.
   * 
   * <p> Node BaseSpecifier may have or not a BaseClause super class.
   * <p><b>Algorithm:</b><ul>
   * <li> If super class is not BaseClause emit inheritance symbol ":"
   * <li> call node method isVirtualParent() and Emit "virtual" if super class
   * is virtual
   * <li> get and emit node inheritance access specifier.
   * </ul>
   * @param node the BaseSpecifier to visit
   * @param data unused
   * @return data returned by the visited child
   */

  public Object visit( ASTBaseSpecifier node, Object data) throws ParseException {

    // If parent is not BaseClause => emit ":"
    Node parent = node.jjtGetParent();
    if (!( parent instanceof ASTBaseClause) )
      emit(" : ");

    String superName = node.getName();
    if ( node.isVirtualParent())
      emit("virtual ");

    emit ( node.getAccessSpec()+" ");
    emit( node.getName()+" ");

    return data;
  }
  
  /**
   * Visits a DirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> If number children = 0 or number of children =1 and the child is a
   * DeclaratorSuffixes <=> case of expansion -> ~ ID (DeclaratorSuffixes)? :
   * Emit Destructor name and visit children.
   * <li> Else if child 0 is a QualifiedId, visit children
   * <li> Else <=> Case of -> "(" Declarator ")" ( DeclaratorSuffixes)? :
   * Emit "(", visit first child, emit ")" and visit 2nd child when exists.
   * </ul>
   * @param node the DirectDeclrator to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTDirectDeclarator node, Object data)
    throws ParseException { 
    int childNum = node.jjtGetNumChildren();

    // case of destructor
    if(((childNum ==1) && node.jjtGetChild(0) instanceof ASTDeclaratorSuffixes)
                                                   || childNum ==0   ) {
      
      emit("~");
      emit( node.getDestructorName()+" ");

      data = node.childrenAccept(this, data);
    }
    else if ( node.jjtGetChild(0) instanceof ASTQualifiedId ) {
      
      data = node.childrenAccept(this, data);
    }
    else {
      // case of "(" Declarator ")" ( DeclaratorSuffixes)?
      emit("( ");
      node.jjtGetChild(0).jjtAccept( this, data );
      emit(") ");
      if( childNum ==2 ) {
	node.jjtGetChild(1).jjtAccept( this, data );
      }
    }

    return data ;


  }

 /** Visits an ExceptionDeclaration node.
   * <p><b>Algorithm:</b><ul>
   * <li> If node has no children <=> case of "..." : Emit "...".
   * <li> Else visit children.
   * </ul>
   * @param node the ExceptionDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */ 

  public Object visit(ASTExceptionDeclaration node, Object data)
   throws ParseException {

     if ( node.jjtGetNumChildren() == 0 )
       emit("... ");
     else
       data = node.childrenAccept(this, data);

     return data ;
  }


  /** Visits a Declarator.
   * <p> Simply visits children.
   * @param node the Declarator to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTDeclarator node, Object data) throws ParseException {
    return node.childrenAccept(this, data);
  }

  /** Visits a QualifiedId.
   * <p> Simply emits the ID name.
   * @param node the QualifiedId to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTQualifiedId node, Object data) throws ParseException {
  
    emit( node.getName() + " ");
    return data;
  }

  /** 
   * Visits a MemberDeclaration.
   * <p> Simply visit children and emit ";" and changes line at the end.
   * @param node the QualifiedId to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTMemberDeclaration node, Object data) throws ParseException {
    indent();
    data= node.childrenAccept(this, data);
    emit(";");
    newline();
    return data;
  }
  /** 
   * Visits a MemberDeclaratorList.
   * <p> Node has at least (Initialized or not ) child.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> get node children initialization value vector.
   * <li> Visit first child 
   * <li> If child initialized, emit "=" and the initialization vector element.
   * <li> Do the same for remainig children separating them with ",".
   * </ul>
   * @param node the MemberDeclaratorList  to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTMemberDeclaratorList node, Object data)
    throws ParseException {
    
    int childNum = node.jjtGetNumChildren();
    Vector initValues = node.getValuesVector();
    String value;
    // visit  child (0)
    Node child = node.jjtGetChild(0);
    child.jjtAccept( this, data );
    // see if the member declarator was initialized
    value = (String) initValues.elementAt(0);
    if (! value.equals("<EMPTY>")) { 
      emit("= " + value + " ");
    }
    //do almost the same with other children and separate them with ","   
   
    for ( int i=1; i< childNum ;i++ ) {
      emit(", ");
      child = node.jjtGetChild(i);
      child.jjtAccept( this, data );
      value = (String) initValues.elementAt(i);
      if (! value.equals("<EMPTY>")) { 
	emit("= " + value + " ");
      }
    }

    return data;
  } 

  /**
   * Visit an InitDeclaratorList.
   * <p><b>Algorithm:</b><ul>
   * Emit children ( calling <a href="#childrenEmit"> childrenEmit</a> )
   * with ", " separating them.
   * </ul>
   * @param node the InitDeclaratorList to visit
   * @param data unused
   * @return data returned by the visited child
   */

  public Object visit(ASTInitDeclaratorList node, Object data) throws ParseException { 
    data = childrenEmit( node,", ");
    return data;
  }

  /**
   * Visits a InitDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Emit both children with "=" between them if node is initialized.
   * <li> Else, visit 1st child, emit "(", visit last child then emit ")".
   * </ul>
   * @param node the InitDeclarator to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTInitDeclarator node, Object data)
    throws ParseException { 
    // Node has 2 children visit then one after one.
    node.jjtGetChild(0).jjtAccept( this ,data );
    if ( node.notInitialized )
      emit("( ");
    else
      emit(" = ");
     node.jjtGetChild(1).jjtAccept( this ,data );
     if ( node.notInitialized )
       emit(" )");

    return data;
  }


  /**
   * Visits an Initializer node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>  Emit "{", visit children separating them with "," the emit "}".
   * </ul>
   * @param node the FunctionDefinition to visit
   * @param data unused
   * @return data returned by the visited child
   */      
  public Object visit(ASTInitializer node, Object data) throws ParseException {

    // simply viist child when aussi 
    emit("{ ");
    childrenEmit( node , ", ");
    emit("} ");
    return data ;
  }
	 
    
  /**
   * Visits a FunctionDefinition.
   * <p><b>Algorithm:</b><ul>
   * <li> Emit an indent and visit all children
   * <li> If last child is not CompoundStatement <=> function declaration  :
   * emit ";".
   * <li> Add a new line.
   * </ul>
   * @param node the FunctionDefinition to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTFunctionDefinition node, Object data) throws ParseException {

    indent();
    int nbchild= node.jjtGetNumChildren();
   
    // visit children
    data = node.childrenAccept(this, data);
    // if last child is not a CompoundStatement => display semicolon
    if ( !(node.jjtGetChild( nbchild-1 ) instanceof ASTCompoundStatement ) )
      emit(" ; ");
    newline();
    newline();

    return data;
  }

  /**
   * Visits an Exception specification node ExceptionSpec.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Emit <code>throw </code> keyword then "(".
   * <li> Visit child if exits.
   * <li> Close paranthesis.
   * </ul>
   * @param node the ExceptionSpec to visit
   * @param data unused
   * @return data returned by the visited child
   */   

  public Object visit(ASTExceptionSpec node, Object data)
    throws ParseException {

    emit (" throw ");
    emit("( ");
    if ( node.jjtGetNumChildren() > 0)
      node.jjtGetChild(0).jjtAccept( this , data ) ; 
    emit(" ) ");
    return data ;
  }

  /**
   * Visits a CompoundStatement.
   * <p><b>Algorithm:</b><ul>
   * <li> emit an indent, emit "{" increase indent and changes of line
   * <li> emit an indent and visits children.
   * <li> decrease the added indent and emits "}"
   * </ul>
   * @param node the CompoundStatement to visit
   * @param data unused
   * @return data returned by the visited child
   */   
 public Object visit(ASTCompoundStatement node, Object data)
   throws ParseException { 
      indent();
      emit("{");
      addIndent(regIndent);
      newline();
      // visit Statements children
      int nbchild= node.jjtGetNumChildren();
      for ( int i =0; i < nbchild ; i++ ) {
	node.jjtGetChild(i).jjtAccept ( this, data);
	newline();
      }
    
      delIndent(regIndent);
      indent();
      emit("}");

      return data;
  
 }

  /**
   * Visits Statement Node.
   * <p> Node exits only when the Statement is an expression statement
   * <p><b>Algorithm:</b><ul>
   * Visits the child and then emits <code>";"</code>
   * </ul>
   * @param node the Statement to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTStatement node,Object data) throws ParseException { 
    indent();
    // visit child
    node.jjtGetChild(0).jjtAccept( this , data);
    // emit semicolon
    emit(" ; ");
	
    return data;
  }

  /**
   * Visits an IdExpr.
   * <p><b>Algorithm:</b><ul>
   * Emits the fully scoped identifier name saved in the node.
   * </ul>
   * @param node the IdExpr to visit
   * @param data unused
   * @return data returned by the visited child
   */   
  public Object visit(ASTIdExpr node, Object data) throws ParseException {
    emit( node.getIdName() + " ");
    return data;
  }


  /**
   * Visits a NewExpr.
   *
   * @param node the NewExpr to visit
   * @param data unused
   * @return data returned by the visited child
   */ 

  public Object visit(ASTNewExpr node, Object data) throws ParseException {
    emit("new ");
    if ( node.jjtGetChild(0) instanceof ASTTypeName ) {
      emit("(");
      node.jjtGetChild(0).jjtAccept( this ,data ) ;
      emit(")");
    }
    else
      node.jjtGetChild(0).jjtAccept( this ,data ) ;
    // now visit remaining children
    if ( node.jjtGetNumChildren() == 2 )
      node.jjtGetChild(1).jjtAccept( this ,data ) ;

    return data ;
  }

  /**
   * Visits a NewTypeId.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children 
   * </ul>
   * @param node the NewTypeId to visit
   * @param data unused
   * @return data returned by the visited child
   */     
  public Object visit(ASTNewTypeId node, Object data)
  throws ParseException {

   return node.childrenAccept( this , data );
  }


  /**
   * Visits a NewDeclarator.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children 
   * </ul>
   * @param node to visit
   * @param data unused
   * @return data returned by the visited child
   */     
  public Object visit(ASTNewDeclarator node, Object data)
  throws ParseException {

   return node.childrenAccept( this , data );
  }

  /**
   * Visits a DirectNewDeclarator.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children with "[" and "]" enclosing them 
   * </ul>
   * @param node to visit
   * @param data unused
   * @return data returned by the visited child
   */     
  public Object visit(ASTDirectNewDeclarator node, Object data)
  throws ParseException {
    emit("[ ");
   node.childrenAccept( this , data );
   emit(" ] ");
   return data ;
  }


 /**
   * Visits a DeleteExpr.
   * 
   * @param node to visit
   * @param data unused
   * @return data returned by the visited child
   */     
  public Object visit(ASTDeleteExpr node, Object data ) throws ParseException {

    if ( node.withOverride )
      emit(" ::");
    emit("delete ");
    if ( node.withBraces )
      emit("[]");
    // visit child
    return node.childrenAccept( this , data );
  }

  /**
   * Visits a destructor declarator DtorDeclarator.
   * <p><b>Algorithm:</b><ul>
   * <lI> Emit destructor name saved in node preceeded with "~".
   * <li> Emits <code>" ( ) "</code> for destructor invocation.
   * </ul>
   * @param node the DtorDeclarator to visit
   * @param data unused
   * @return data returned by the visited child
   */     

  public Object visit(ASTDtorDeclarator node, Object data)
    throws ParseException {
    emit("~"+ node.getDestructorName()+ " ( ) " );
    return data;
  }

  /**
   * Visits a destructor definition node DtorDefinition.
   * <p><b>Algorithm:</b><ul>
   * Visits children, emits <code>";"</code> the adds a new line.
   * </ul>
   * @param node the DtorDefinition to visit
   * @param data unused
   * @return data returned by the visited child
   */   
 public Object visit(ASTDtorDefinition node, Object data) throws ParseException {
   indent();
   int mdf = node.getModifier() ;
   CModifier modif = new CModifier(mdf);
   emit (modif.toString() + " ");
   data = node.childrenAccept(this, data);
   emit(" ;");
   newline();

   return data;
 }


  /**
   * Visits a constrcutor initializer node CtorInitializer.
   *
   * <p><b>Algorithm:</b><ul>
   * Emit ":" and calls <a href="#childrenEmit"> childrenEmit</a>
   * </ul>
   * @param node the CtorInitializer to visit
   * @param data unused
   * @return data returned by the visited child
   */      
  public Object visit(ASTCtorInitializer node, Object data)
    throws ParseException {
      
    emit(" : ");
    childrenEmit(node, ", ") ;
    return data ;
  }

  /**
   * Visits a super class initializer SuperclassInit.
   * 
   * <p><b>Algorithm:</b><ul>
   * <li> Visit 1st child.
   * <li> visit remainig child if they exist.
   * </ul>
   * @param node the SuperclassInit to visit
   * @param data unused
   * @return data returned by the visited child
   */         
  public Object visit(ASTSuperclassInit node, Object data)
    throws ParseException {
 
    node.jjtGetChild(0).jjtAccept( this , data );
    node.childrenAccept(this, data, 1);
    return data ;
  }
  /**
   * Visits a Constrcutor Definition node CtorDefinition.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Get the constructor modifier integer. Construct the CModifier object
   * and emit it using its toString() method.
   * <li> Visit children.
   * <li> Emit ";" if last child is not CompoundStatement.
   * </ul>
   * @param node the CtorDefinition to visit
   * @param data unused
   * @return data returned by the visited child
   *
   * @see ORG.opengroup.adl.symboltable.c.CModifier.
   */     
  public Object visit(ASTCtorDefinition node, Object data)
    throws ParseException {

    int nbchild = node.jjtGetNumChildren() ;
    int mdf = node.getModifier();
    CModifier modif = new CModifier(mdf) ;
    indent() ;
    emit ( modif.toString() + " ");
    node.childrenAccept( this, data );
    if ( ! (node.jjtGetChild(nbchild -1 ) instanceof ASTCompoundStatement ) )
      emit(";") ;
    // add new line 
    newline();
    return data ;
  }

  /**
   * Visits a constrcutor declarator node CtorDeclarator.
   * <p><b>Algorithm:</b><ul>
   * <li> Visit 1st child.
   * <li> emit "(", visit 2nd child if exists the emit ")"
   * </ul>
   * @param node the CtorDeclarator to visit
   * @param data unused
   * @return data returned by the visited child.
   */
  public Object visit(ASTCtorDeclarator node, Object data)
    throws ParseException {

    int nbchild = node.jjtGetNumChildren() ;
    node.jjtGetChild(0).jjtAccept( this , data );
    emit ("( ");
    if ( nbchild > 1 )
      node.jjtGetChild(1).jjtAccept( this , data );
    emit (" )");

    return data ;
  }


 


  /**
   * Visits FunctionDirectDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Visit first child then emit "(".
   * <li> <ul>
   * <li> In case of function with non-empty parameters ( case second child 
   * exist and is a  ParameterDeclaration or ParameterDeclarationList ) :
   * visit 2nd child and emit ")"
   * <li> Else, emit ")"
   * </ul>
   * <li> Visit remaining children.
   * </ul>
   * @param node the FunctionDirectDeclarator to visit
   * @param data unused
   * @return data returned by the visited child
   */     

  public Object visit(ASTFunctionDirectDeclarator node, Object data)
    throws ParseException {
   
   int curchild =0;
   int nbchild= node.jjtGetNumChildren();

   node.jjtGetChild(0).jjtAccept( this,data );
   emit ("( ");
   // case of function with arguments
   if ( (nbchild >1) && 
	( (node.jjtGetChild(1) instanceof ASTParameterDeclarationList) ||
	  ( node.jjtGetChild(1) instanceof ASTParameterDeclaration ) ) )  {

     node.jjtGetChild(1).jjtAccept( this, data );
     curchild = 1;
   }
   
   emit(" ) ");
   // visit the remaining children
   for ( int i = curchild +1 ; i< nbchild ; i++ ) {
     node.jjtGetChild(i).jjtAccept( this, data );
   }
   
   String initValue = node.getInitValueImage() ;
   if ( !initValue.equals(" ") )
     emit (" = "+initValue) ;

   return data;
  }      

  /** 
   * Visits a DeclaratorSuffixes.
   * <p> Node may be array suffixes or a function signatures.
   * <p><b>Algorithm:</b><ul>
   * Get array size of the node which is the  number of suffix balanced braces
   * <ul>
   * <li> If size = 0 => case of function signature :
   *<ul>
   * <li> Emit "(" , visit parameter declaration child if exits, then emit ")"
   * <li> Visit remaining children.
   * </ul>
   * <li> Else => case of Multidimensional array :
   * <ul>
   * <li>If node array size > number of children=> emit empty brace pair "[ ]".
   * <li> For each child, emit "[", visit the child and emit "]".
   * </ul> </ul>
   * @param node DeclaratorSuffixes to visit.
   * @param data unused
   * @return the data returned by the visited child
   */
  public Object visit(ASTDeclaratorSuffixes node, Object data) throws ParseException {

    int arraySize = node.getArraySize();
    int childNum = node.jjtGetNumChildren();

    // arraysize tells if the node is an array definition or method declaration
    if ( arraySize == 0 ) {

      emit("( ");

      if ( (childNum > 0) && 
         ( (node.jjtGetChild(0) instanceof ASTParameterDeclarationList) || 
            ( node.jjtGetChild(0) instanceof ASTParameterDeclaration ) ) ) {

	node.jjtGetChild(0).jjtAccept( this, data );
	emit(" )");
	data = node.childrenAccept(this, data, 1);
      } else {

	emit (" )");
	data = node.childrenAccept(this, data);
      }      
    }
    else {
      // case of array suffix
      if ( arraySize > childNum ) {
	// case of first undefinite bound 
	emit("[ ]");
      }
      for ( int i=0; i < childNum;i++ ) {
	emit("[ ");
	node.jjtGetChild(i).jjtAccept( this, data );
	emit("]");
      }
    }

    return data;
  }


  /** 
   * Visits a PtrOperator.
   * <P> Node kind can either be a "Pointer To" opeartor, "Reference To"
   * operator, or "Pointer To Member" operator.
   * <p><b>Algorithm:</b><ul>
   * <li> get the pointer kind from node.
   *<ul>
   * <li> In case of "PointerTo", emit "*"
   * <li> In case of "ReferenceTo", emit "&"
   * <li> In case of "PointerToMember", emit node enclosing scopeoverride name 
   * and emit "*".
   * </ul>
   * <li> get node pointer qualifier saved as integer in the node.
   * Construct a ORG.opengroup.adl.symboltable.c.CModifier for that modifier.
   * Call its method toString() to emit the node qualifier.
   * </ul>
   * @param node the visited PtrOperator
   * @param data unused
   * @return the data returned by the visited child
   */   
  public Object visit(ASTPtrOperator node, Object data) throws ParseException { 
    String ptrKind = node.getPtrkind();
    if ( ptrKind.equals("ReferenceTo") ) {

      emit("& ");
    } else if (ptrKind.equals("PointerTo")) {
      
      emit("* ");
    }
    else if( ptrKind.equals("PointerToMember")) {
     emit( (node.getScopedName()).Name );
     emit(" * ");
    }

    CModifier ptrModif = new CModifier( node.getQualifier() );
    emit( ptrModif.toString()+" " );

    return data;
  }

  /**
   * Visits a ParameterDeclarationList.
   * 
   * <p><b>Algorithm:</b><ul>
   * Emit children ( calling <a href="#childrenEmit"> childrenEmit</a> )
   * with ", " separating them.
   * </ul>
   * @param node the ParameterDeclarationList to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTParameterDeclarationList node, Object data) throws ParseException {

    childrenEmit( node , ", ");
    return data;
  }


  /**
   * Visits a ParameterDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> Visit first and scond child.
   * <li>If number of children =3 <=>declaring parameter with default argument:
   * emit "=" and visit third child.
   * </ul>
   * @param node ParameterDeclaration to visit
   * @param data unused
   * @return the data returned by the visited child
   */   
  public Object visit(ASTParameterDeclaration node, Object data)
    throws ParseException {
    // get number of children
    int childNum = node.jjtGetNumChildren();

    // visit child(0) and child(1)
    node.jjtGetChild(0).jjtAccept(this , data );
    node.jjtGetChild(1).jjtAccept(this , data );

    // childNum may either be 2 or 3. If three => Default arguments
    if ( childNum == 3 ) {
      emit(" = ");
      node.jjtGetChild(2).jjtAccept(this , data );
    }

    return data;
  }
 

  /**
   * Visits AbstractDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li> If node has more than one child and first child is PtrOperator =>
   * Case of production -> PtrOperator() AbstractDeclarator() : visit children.
   * <li> Else, If node array number = 0 => case of grammar production 
   * -> "(" AbstractDeclarator ")" (AbstarctDeclaratorSuffix )? :
   * emit "(", visit first child, emit ")" and visit remaining children.
   * <li> Else => case of Multidimensional array.
   * <ul>
   * <li> If node array number >  node children number => case of empty first 
   * brace of multidimensional array : emit an empty braces pair ( "[]" ).
   * <li> for each child : emit "[", visit the child and emit "]"
   * </ul> </ul>
   *
   * @param node the visited AbstractDeclarator
   * @param data unused
   * @return the data returned by the visited child
   */
   
  public Object visit(ASTAbstractDeclarator node, Object data)
    throws ParseException { 
 
   // get number of children
    int childNum = node.jjtGetNumChildren();

    // case of pointer into abstract declarator or no children
    if ( ( (childNum > 0) && (node.jjtGetChild(0) instanceof ASTPtrOperator ))
	 || ( (node.getNumberArrays() == 0) &&  (childNum==0)) ) {

      data = node.childrenAccept(this, data);
      return data;
    }			
    // array numbers tell if the abstract declarator is an array or not
    if ( node.getNumberArrays() == 0) {
      // case of "(" abstractDeclarator ")" (AbstarctDeclaratorSuffix )?
      emit("( ");
      node.jjtGetChild(0).jjtAccept(this,data );
      emit(") ");
      // visit remaining children
      for ( int i=1; i < childNum; i++ ) {	
	node.jjtGetChild(i).jjtAccept(this,data );
      }
    }
    else {
      // case of empty fisrt array braces
      if( childNum > node.getNumberArrays() ) 
	emit("[]");
      
      // visit remainig children
      for( int i=0; i < childNum; i++ ) {
       emit("[ ");
       node.jjtGetChild(i).jjtAccept(this,data );
       emit("] ");
      }	
    }
    return data;
  }

  /**
   * Visits an AbstractDeclaratorSuffix.
   * <p><b>Algorithm:</b><ul>
   * Node may be a ( multiple) array suffix or "(" [ParameterList]")
   * according to node array suffix indicator.
   * <li> If node array size = 0 =>case of  production ->"(" [ParameterList]"):
   * emit "(", visit ParameterDEclaration(List) child when exists then emit ")"
   * <li> If node array size > 0 => node is a multiple array suffix
   * <ul>
   * <li>If array size > children number, emit an empty pair of braces ("[]")
   * <li> For each child : emit "[", visit the child the emit "]".
   *</ul></ul>
   * @param node the visited AbstractDeclaratorSuffix
   * @param data unused
   * @return the data returned by the visited child
   */
  public Object visit(ASTAbstractDeclaratorSuffix node, Object data) throws ParseException { 
    int arraySize = node.getArraySize();
    int childNum = node.jjtGetNumChildren();

    // arraysize tells if the node is an array definition or method declaration
    if ( arraySize == 0 ) {

      emit("( ");

      if ((childNum > 0) && 
	  ( (node.jjtGetChild(0) instanceof ASTParameterDeclarationList)
	    || ( node.jjtGetChild(0) instanceof ASTParameterDeclaration ) ) ){

	node.jjtGetChild(0).jjtAccept( this, data );
	emit(" )");
      } else {

	emit (" )");
	data = node.childrenAccept(this, data);
      }      
    }
    else {
      if ( arraySize > childNum ) {
	// case of first undefinite bound 
	emit("[ ]");
      }
      // case of array suffix
      for ( int i=0; i < childNum;i++ ) {
	emit("[ ");
	node.jjtGetChild(i).jjtAccept( this, data );
	emit("]");
      }
    }

    return data;
  }

  /**
   * Visits an AccessSpecifier.
   * <p><b>Algorithm:</b><ul>
   * <li> Reduce the indent to be inferior to class members indent.
   * <li> get and emit the node access specifier name. 
   * <li> add a new line and restore class member indent level.
   * </ul>
   * @param node the AccessSpecifier to visit
   * @param data unused
   * @return unused data
   */
  public Object visit(ASTAccessSpecifier node, Object data) 
    throws ParseException { 

    // acces specifier display indent prior to members indent
    delIndent( regIndent-1);
    indent();
    emit( node.getAccessSpecifierName() +" :");
    newline();
    // restore previous indent
    addIndent( regIndent-1);

    return data;
 }

  /**
   * Visits an IfNode.
   * <p><b>Algorithm:</b><ul>
   * <li> emit keyword <code>if</code> and "(".
   * <li> Visit first child.
   * <li> emit ")", add a newline, increase the indent if satatement child is
   * not a compound statement, visit second child,
   * <li> reduce back the added indent.
   * <li> If children number = 3 => case of ElseNode child : visit that child.
   * </ul>
   * @param node the IfNode to visit
   * @param data unused
   * @return unused data
   */   
  public Object visit(ASTIfNode node, Object data) throws ParseException {

    int childNum = node.jjtGetNumChildren();
    int added =0 ;
    // childNum can either be  2 or 3
    if ( ! (node.jjtGetParent() instanceof ASTElseNode ) )
      indent() ;
    emit("if ( ");
    Node child = node.jjtGetChild(0);
    // visit first child : expression child
    child.jjtAccept ( this ,data );
    emit (") ");
    newline();
    child = node.jjtGetChild(1);
    if ( !( child instanceof ASTCompoundStatement )) { 
      addIndent(regIndent) ;
      added = regIndent ;
    }

    // visit satement child
    child.jjtAccept ( this ,data );
    delIndent(added);
    newline();
    if ( childNum == 3) {
      child = node.jjtGetChild(2);
      // visit ElseNode
      child.jjtAccept ( this ,data );
    }

    return data;
  }

  /**
   * Visits an ElseNode.
   * <p><b>Algorithm:</b><ul>
   * <li> emit <code> else </code> keyword.
   * <li> If child is CompoundStatement, simply change of line.
   * <li>If node child is not an IfNode child, add new line and add indent.
   * <li> Visit child.
   * <li> reduce the indent if added any.
   * @param node the ElseNode to visit
   * @param data unused
   * @return unused data
   */   
  public Object visit(ASTElseNode node, Object data) throws ParseException {
    // node has only one child
    indent();
    emit("else ");
    // indent to be added for else statement
    int indentToAdd =0; 
    Node child = node.jjtGetChild(0) ;
    // if child is IfNode don't change of line nor add indent
    if ( child instanceof ASTCompoundStatement )
      newline();
    else if (!(child instanceof ASTIfNode ) ) {
      newline();
      addIndent(regIndent) ;
      indentToAdd = regIndent ;
    }
    // visit child 
    child.jjtAccept ( this ,data ) ;
    delIndent( indentToAdd ) ;
    
    return data;
  }


  /**
   * Visits an Expr. 
   * <p> Node has at lesat two children.
   * <p><b>Algorithm:</b><ul>
   * Emit children ( calling <a href="#childrenEmit"> childrenEmit</a> )
   * with ", " separating them.
   * </ul>
   * @param node the Expr to visit
   * @param data unused
   * @return unused data
   */   
  public Object visit(ASTExpr node, Object data) throws ParseException {
    
    data = childrenEmit( node,", ");
    return data;
  }

  /**
   * Visits an AssignmentExpr.
   * <p><b>Algorithm:</b><ul>
   * Emit left hand child, emit node operator then emit right hand child.
   * </ul>
   * @param node the AssignmentExpr to visit
   * @param data unused
   * @return unused data
   */      
  public Object visit(ASTAssignmentExpr node, Object data) throws ParseException { 
    // node has exactly two children
    Node child = node.jjtGetChild(0);
    child.jjtAccept ( this ,data );
    emit( node.getOp() + " ");
    child = node.jjtGetChild(1);
    child.jjtAccept ( this ,data );

    return data;
  }

  /**
   * Visits a CondExpr.
   * 
   * <p><b>Algorithm:</b><ul>
   * <li> Visit first child then emit "?".
   * <li> Visit second and emit ":"
   * <li> Visit remainig child.
   * </ul>
   * @param node the CondExpr to visit
   * @param data unused
   * @return unused data
   */   
  public Object visit(ASTCondExpr node, Object data) throws ParseException { 
    // node has 3 children
    node.jjtGetChild(0).jjtAccept ( this ,data );
    emit(" ? ");
    node.jjtGetChild(1).jjtAccept ( this ,data );
    emit(":");
    node.jjtGetChild(2).jjtAccept ( this ,data );

    return data;
 }

  /**
   * Visits BinaryLogicalExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> get from node the operators vector.
   * <li> Visit first child.
   * <li> loop on remainig children index emitting related vector operator
   * element and visiting the current children.
   * </ul>
   * @param node the BinaryLogicalExpr to visit
   * @param data unused
   * @return unused data
   */   
  public Object visit(ASTBinaryLogicalExpr node, Object data) throws ParseException   {
    int childNum = node.jjtGetNumChildren();
    Vector operators = node.getOperatorsVector() ;
    node.jjtGetChild(0).jjtAccept ( this ,data );
    
    for ( int i =1; i < childNum; i++ ) {
      emit( " "+(String) operators.elementAt(i-1)+" " );
      node.jjtGetChild(i).jjtAccept ( this ,data );
    }

    return data;
  }

  public Object visit(ASTConstant node, Object data) throws ParseException { 

    emit(node.getConstantValue() + " ");
    return data;
  }


  /** 
   * Visits a BinaryArithExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> get from node the operators vector.
   * <li> Visit first child.
   * <li> loop on remainig children index emitting related vector operator
   * element and visiting the current children.
   * </ul>
   * @param node the BinaryArithExpr to visit
   * @param data unused
   * @return unused data
   */     

  public Object visit(ASTBinaryArithExpr node, Object data) throws ParseException {
    int childNum = node.jjtGetNumChildren();
    Vector operators = node.getOperatorsVector() ;
    node.jjtGetChild(0).jjtAccept ( this ,data );
    
    for ( int i =1; i < childNum; i++ ) {
      emit( " "+(String) operators.elementAt(i-1)+" " );
      node.jjtGetChild(i).jjtAccept ( this ,data );
    }

    return data;
  }

  /**
   * Visits a CastExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> emit "(", visit first child, then ")".
   * <li> Visit second child.
   * </ul>
   * @param node the CastExpr to visit
   * @param data unused
   * @return unused data
   */     
  public Object visit(ASTCastExpr node, Object data) throws ParseException {
    // Node has exactly 2 children
    emit("( ");
    node.jjtGetChild(0).jjtAccept ( this ,data );
    emit(" ) ");
    node.jjtGetChild(1).jjtAccept ( this ,data );

    return data;
  }

  /**
   * Visits an UnaryExpr.
   * <p> Node has only one child.
   * <p><b>Algorithm:</b><ul>
   * <li> get the node operator.
   * <ul>
   * <li>If operator is "sizeof" : emit "sizeof (", visit child and emit ")".
   * <li> Else, emit the operator then visit the child.
   * </ul> </ul>
   * @param node the UnaryExpr to visit
   * @param data unused
   * @return unused data
   */     
  public Object visit(ASTUnaryExpr node, Object data) throws ParseException {
    // get the unique node child
    Node child = node.jjtGetChild(0);
    // get the node operator: one of { ++,--,*,&, sizeof, UnaryOp}
    String operator = node.getOp();
    // first case -> sizeof ( TypeName ) 
    if ( operator.equals("sizeof") && ( child instanceof ASTTypeName )) {
      emit(operator );
      emit("( ");
      child.jjtAccept( this ,data );
      emit(" ) ");
    }
    else {
      // in otehr cases just emit the operator then the child
      emit(operator+" ");
      child.jjtAccept( this ,data );
    }

    return data;
  }

  /**
   * Visits a PostfixExpr.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children.
   * </ul>
   * @param node the PostfixExpr to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTPostfixExpr node, Object data) throws ParseException {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits PostDeIncrement.
   * <p><b>Algorithm:</b><ul>
   * Simply emit the operator ( one of "++" or "--").
   * @param node the PostDeIncrementxpr to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTPostDeIncrement node, Object data) throws ParseException {
    
    emit( node.getOp() );
    return data;
  }

  /**
   * Visits an ArraySuffix.
   * <p><b>Algorithm:</b><ul>
   * Emit "[", visit the chid then emit "]".
   * </ul>
   * @param node the ArraySuffix to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTArraySuffix node, Object data) throws ParseException {
    emit("[ ");
    node.jjtGetChild(0).jjtAccept( this, data );
    emit(" ]");
    return data;
  }
  /**
   * Visits an DotAccessSuffix.
   * <p><b>Algorithm:</b><ul>
   * Emit "." and  visit the children.
   * </ul>
   * @param node the DotAccessSuffix to visit
   * @param data unused
   * @return unused data
   */
  public Object visit(ASTDotAccessSuffix node, Object data)
    throws ParseException {
    emit(".");
    data = node.childrenAccept(this, data);
    return data ;
  }

  /**
   * Visits an RefAccessSuffix.
   * <p><b>Algorithm:</b><ul>
   * Emit "->" and  visit the children.
   * </ul>
   * @param node the RefAccessSuffix to visit
   * @param data unused
   * @return unused data
   */

  public Object visit(ASTRefAccessSuffix node, Object data)
    throws ParseException {
    emit("->");
    data =  node.childrenAccept(this, data);
    return data ;
  }


  /**
   * Visits an ArgumentList.
   * <p><b>Algorithm:</b><ul>
   * <li> emit "("
   * <li>  Emit children ( calling <a href="#childrenEmit"> childrenEmit</a> )
   * with ", " separating them.
   * <li> Emit ")".
   * </ul>
   * @param node the RefAccessSuffix to visit
   * @param data unused
   * @return unused data
   */
  public Object visit(ASTArgumentList node, Object data)
    throws ParseException {
    emit("(");
    data = childrenEmit(node,", ");
    emit(")");
    return data;
  }

  /**
   * Visits a PrimaryExpr.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children.
   * </ul>
   * @param node the PrimaryExpr to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTPrimaryExpr node, Object data) throws ParseException {
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a TypeName.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children.
   * </ul>
   * @param node the TypeName to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTTypeName node, Object data) throws ParseException {
    return node.childrenAccept(this, data);
  }

 /**
   * Visits a TypeNameList.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children.
   * </ul>
   * @param node the TypeNameList to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTTypeNameList node, Object data) throws ParseException {
    return node.childrenAccept(this, data);
  }


  /**
   * Visits a ParenthExpr.
   * <p><b>Algorithm:</b><ul>
   * Emit "(", visits the child then emit ")".
   * </ul>
   * @param node the ParenthExpr to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTParenthExpr node, Object data) throws ParseException { 
    emit("( ");
    node.jjtGetChild(0).jjtAccept( this ,data );
    emit(" )");
    return data;
  }



  /**
   * Visits a FunctionDeclarator.
   * <p><b>Algorithm:</b><ul>
   * Simply visit children.
   * </ul>
   * @param node the FunctionDeclarator to visit
   * @param data unused
   * @return unused data
   */ 

 public Object visit(ASTFunctionDeclarator node, Object data) throws ParseException { 
    return node.childrenAccept(this, data);
 }





  /**
   * Visits a QualifiedType.
   * <p><b>Algorithm:</b><ul>
   * Emit the qualified type name.
   * </ul>
   * @param node the  QualifiedType to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTQualifiedType node, Object data) throws ParseException {
    emit( node.getTypeName() + " " );
    return data;
  }

  /**
   * Visits a FullyScopedType.
   * <p><b>Algorithm:</b><ul>
   * Emit the fully scoped type name.
   * </ul>
   * @param node the FullyScopedType to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTFullyScopedType node, Object data) throws ParseException { 
    emit(node.getFullTypeName()+" ");
    return data;;
 }

  /**
   * Visits a BuiltinTypeSpec.
   * <p><b>Algorithm:</b><ul>
   * Emit builtintype name.
   * </ul>
   * @param node the BuiltinTypeSpec to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTBuiltinTypeSpec node, Object data)
    throws ParseException {

    emit( node.getTypeName()+" ");
    return data;
  }

  /**
   * Visits a LabelNode.
   * <p><b>Algorithm:</b><ul>
   * <li> emit the label name and ":".
   * <li> increase the indent indent and a new line.
   * <li> Visit the child, reduce back the indent and add a new line.. 
   * </ul>
   * @param node the LabelNode to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTLabelNode node, Object data) throws ParseException { 
    indent();
    // emit the node label : an identifier or the keyword "default"
    emit(node.getName()+" : ");
    addIndent( regIndent);
    newline();
    // visit the unique node child 
    node.jjtGetChild(0).jjtAccept( this, data );
    // delete the added indent
    delIndent( regIndent );
    newline();
    return data;
  }

  /**
   * Visits a caseExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> emit the <code> case </code> keyword.
   * <li> Visit the first child, emit ":", increase the indent and add new line
   * <li> Visit second child, delete added indent and add a newline.
   * </ul>
   * @param node the CaseExpr to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTCaseExpr node, Object data) throws ParseException {
    indent();
    int added = 0 ;
    emit("case ");
    // visit first child
    node.jjtGetChild(0).jjtAccept( this, data );
    emit(" : ");
    if ( ! ( node.jjtGetChild(0) instanceof ASTCompoundStatement ) ) {
      addIndent( regIndent);
      added =  regIndent ;
    }
    newline();
    // visit second child
    node.jjtGetChild(1).jjtAccept( this, data );
    // delete the added indent
    delIndent( added );
    newline();

    return data;
  }

  /**
   * Visits a ReturnStmt.
   * <p><b>Algorithm:</b><ul>
   * <li> emit keyword <code> return </code>.
   * <li> Visit the child when exists.
   * <li> Emit ";" and a new line. 
   * </ul>
   * @param node the ReturnStmt to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTReturnStmt node, Object data)
    throws ParseException { 
    int childNum = node.jjtGetNumChildren();
    indent();
    emit("return ");
    // visit child if exists
    if ( childNum == 1) {
      node.jjtGetChild(0).jjtAccept( this, data );
    }
    emit(";");
    newline();

    return data;
  }

  /**
   * Visits a ThrowStmt.
   * <p><b>Algorithm:</b><ul>
   * <li> emit keyword <code> throw </code>.
   * <li> Visit the unique child if any.
   * <li> Emit ";" and a new line. 
   * </ul>
   * @param node the ThrowStmt to visit
   * @param data unused
   * @return unused data
   */ 
 public Object visit(ASTThrowStmt node, Object data) throws ParseException { 

   indent();
   emit ("throw ");
   // visit child if any.
   if ( node.jjtGetNumChildren() > 0)
     node.jjtGetChild(0).jjtAccept( this, data );
   emit(" ;");
   newline();

   return data ;
 }

  /**
   * Visits a SwitchNode .
   * <p><b>Algorithm:</b><ul>
   * <li> set indent, emit keyword <code> switch </code> and emit "("
   * <li> Visit first child then emit ")".
   * <li> if second child is not a cvompound increase the indent.
   * <li> Add a new line and visit second child.
   * <li> Decrease the indent if added any and add a new line.
   * </ul>
   * @param node the SwitchNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTSwitchNode node, Object data)
    throws ParseException {

    indent();
    emit("switch (");
    // visit first child
    node.jjtGetChild(0).jjtAccept( this, data );
    emit(" ) ");
    int added = regIndent;
    Node child = node.jjtGetChild(1);
    if( child instanceof ASTCompoundStatement )
      added = 0;
    // add indent
    addIndent( added);
    newline();
    // visit second child
    child.jjtAccept( this, data );
    delIndent( added);
    newline();

    return data;
  }

  /**
   * Visits a WhileNode.
   * <p><b>Algorithm:</b><ul>
   * <li> set an indent, emit "while (" and visit first child.
   * <li> emit ")", increase the indent, visit second child then reduce indent.
   * </ul>
   * @param node the WhileNode to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTWhileNode node, Object data) throws ParseException {

    indent();
    int added = 0 ;
    emit("while ( ");
    node.jjtGetChild(0).jjtAccept( this, data );
    emit(" ) ");
    // add indent
    if ( ! ( node.jjtGetChild(1) instanceof ASTCompoundStatement ) ) {
      addIndent( regIndent);
      added = regIndent ;
    }
    newline();
    // visit second child
    node.jjtGetChild(1).jjtAccept( this, data );
    delIndent( added);
    newline();

    return data;
  }

 /**
   * Visits a DoWhileNode.
   * <p><b>Algorithm:</b><ul>  
   * <li> Emit keyword <code> do </code> and visit first child.
   * <li> Add a new line, set an indent and  emit ""while ("
   * <li> visit second child, emit ")" and add a newline.
   * </ul>
   * @param node the DoWhileNode to visit
   * @param data unused
   * @return unused data
   */ 

  public Object visit(ASTDoWhileNode node, Object data) throws ParseException {
    
    indent();
    emit("do ");
    if (  node.jjtGetChild(0) instanceof ASTCompoundStatement )
      newline();
    // visit first child 
    node.jjtGetChild(0).jjtAccept( this, data );
    newline();
    indent();
    emit("while ( ");
    // visit last child
    node.jjtGetChild(1).jjtAccept( this, data );
    emit(" ) ;");
    newline();

    return data;
  }
  /**
   * Visits a ForNode.
   * <p> A Vector stored within the node tells the children  kind ( one of 
   * Expr, Declaration or None) of the for (? ; ? ; ? ) statement. 
   * <p><b>Algorithm:</b><ul> 
   * <li> Emit "for (".
   * <li> Visit first child and increment currentchild counter when kind of 
   * first part of for satement not "EMPTY". 
   * <li>Emit ";" if first kind is not a declaration.
   * <li>Visit current child and increment counter when 2nd part kind ! EMPTY
   * <li>Emit ";" and do the same for before for 3rd part kind.
   * <li> Emit ")", increase the indent and change of line.
   * <li> Visit last child and decrease indent again.
   * </ul>
   * @param node the ForNode to visit
   * @param data unused
   * @return unused data
   *
   */ 

  public Object visit(ASTForNode node, Object data) throws ParseException {

    int childNum = node.jjtGetNumChildren();
    int currentChild = 0;
    indent();
    emit("for ( ");

    // visit ForNode initialisation part if exits
    if (! (node.getChildKind(0).equals("EMPTY") )) { 
      node.jjtGetChild(0).jjtAccept( this, data );
      currentChild = 1;
    }
    if ( ! node.getChildKind(0).equals("Declaration") )
      emit("; ");
    // visit ForNode condition child if exists
    if (! (node.getChildKind( 1).equals("EMPTY") )) {
      node.jjtGetChild(currentChild).jjtAccept( this, data );
      currentChild++;
    }
    emit("; ");
    // visit the iteration child if exists
    if (! (node.getChildKind(2).equals("EMPTY") )) {
      node.jjtGetChild(currentChild).jjtAccept( this, data );
      currentChild++;
    }

    emit(" )");
    int added = regIndent;
    if ( node.jjtGetChild(currentChild) instanceof ASTCompoundStatement )
      added= 0;
    addIndent( added);
    newline();
    node.jjtGetChild(childNum-1).jjtAccept( this, data );
    delIndent( added );

    return  data;
  }


  /**
   * Visits a GotoNode.
   * <p><b>Algorithm:</b><ul> 
   * Set indent. Emit keyword <code>goto</code>, node destination name and ";" 
   * </ul>
   * @param node the GotoNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTGotoNode node, Object data) throws ParseException {
    indent();
    emit("goto "+ node.getName()+" ;");
    newline();
    return data;
  }

  /**
   * Visits a ContinueNode.
   * <p><b>Algorithm:</b><ul>
   * Set an indent and  emit keyword <code> continue</code> and ";".
   * </ul>
   * @param node the ContinueNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTContinueNode node, Object data) throws ParseException { 
    indent();
    emit("continue ;");
    return data;
  }

  /**
   * Visits a BreakNode.
   * <p><b>Algorithm:</b><ul>
   * Set an indent and  emit keyword <code> break</code> and ";".
   * </ul>
   * @param node the BreakNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTBreakNode node, Object data) throws ParseException {
    indent();
    emit("break ;");
    return data;
    
  }
  /**
   * Visits a TryNode.
   * <p><b>Algorithm:</b><ul>
   * <li> Set indent and emit keyword <code>try</code>.
   * <li> Visit children.
   * </ul>
   * @param node the TryNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTTryNode node, Object data) throws ParseException { 
    indent();
    emit("try ") ;
    newline();
    data= node.childrenAccept(this, data);
    newline();
    return data;
  }

  /**
   * Visits a CatchNode.
   * <p><b>Algorithm:</b><ul>
   * <li> Set indent and emit "catch (".
   * <li> Visit first child, add new line then visit second child..
   * </ul>
   * @param node the CatchNode to visit
   * @param data unused
   * @return unused data
   */ 
  public Object visit(ASTCatchNode node, Object data) throws ParseException {
    indent();
    emit("catch (");
    node.jjtGetChild(0).jjtAccept( this , data );
    emit(") ");
    newline();
    node.jjtGetChild(1).jjtAccept( this , data );

    return data;
  }

  /**
   * Visits a ConstantExpr.
   * <p><b>Algorithm:</b><ul>
   * Visit node children.
   * </ul>
   * @param node the ConstantExpr to visit
   * @param data unused
   * @return unused data
   */    
  public Object visit(ASTConstantExpr node, Object data)
  throws ParseException { 
     data = node.childrenAccept(this, data); 
     return data ;
    
  }
  /**
   * Visits a pointer to member expression node PmExpr.
   * <p><b>Algorithm:</b><ul>
   * <li> get from node the operators vector.
   * <li> Visit first child.
   * <li> loop on remainig children index emitting related vector operator
   * element and visiting the current children.
   * </ul>
   * @param node the PmExpr to visit
   * @param data unused
   * @return unused data
   */   
  
  public Object visit(ASTPmExpr node, Object data) throws ParseException { 

    int childNum = node.jjtGetNumChildren();
    Vector operators = node.getOperatorsVector() ;
    node.jjtGetChild(0).jjtAccept ( this ,data );
    
    for ( int i =1; i < childNum; i++ ) {
      emit( " "+(String) operators.elementAt(i-1)+" " );
      node.jjtGetChild(i).jjtAccept ( this ,data );
    }

    return data;
  }

  /**
   * Visits a This node.
   * <p><b>Algorithm:</b><ul>
   * Emit keyword <code>this</code>.
   * </ul>
   * @param node the ConstantExpr to visit
   * @param data unused
   * @return unused data
   */    

  public Object visit(ASTThis node, Object data) throws ParseException {
    emit("this");
    return data;
  }
  /**
   * Visits a String node.
   * <p><b>Algorithm:</b><ul>
   * Emit the string name.
   * </ul>
   * @param node the String to visit
   * @param data unused
   * @return unused data
   */    
  public Object visit(ASTString node, Object data) throws ParseException {
    emit("\"" + node.getName() + "\" ");
    return data;
  }


  // All following AST nodes are incorrect since they should have
  // disappeared during AST transformation


  /** Should not be visited! */
  public Object visit(ASTADLImplExpr node, Object data) throws ParseException {
    incorrectNode ( node) ; 
    return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLDomainList node, Object data)
    throws ParseException {
    incorrectNode ( node) ; 
    return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLCallStateExpr node, Object data) throws ParseException { 
    incorrectNode ( node) ; return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLClass node, Object data) throws ParseException {

    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLBhvDecl node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLBhvSpec node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLBhvClassification node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */

  public Object visit(ASTADLLabel node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLTags node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }


  /** Should not be visited! */
  public Object visit(ASTADLProlog node, Object data) throws ParseException   { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLEpilog node, Object data) throws ParseException   { 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLAssGrp node, Object data) throws ParseException   { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLInhSemantics node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLBinding node, Object data) throws ParseException {
    incorrectNode( node ); return data; 
  }
  /** Should not be visited! */
  public Object visit(ASTADLInlineDecl node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLSuperClass node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLIfStmt node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLTryStmt node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLAssertion node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTReturn node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLThrownExpr node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTNormalExpr node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTAbnormalExpr node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTUnchangedExpr node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADLQuantifiedAssertion node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLNamedParamList node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADLDomain node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDUseDeclaration node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDFieldDeclaration node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDADLnewExpr node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }

 /** Should not be visited! */
  public Object visit(ASTTDDADLExpr node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }


  /** Should not be visited! */
  public Object visit(ASTTDDClassDecl node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDDDatasetDecl node, Object data) throws ParseException { 
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDDFactoryDef node,Object data) throws ParseException{ 
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDDatasetExpr node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDDDatasetSingleton node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDDFactoryCall node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDTestDirective node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }


  /** Should not be visited! */
  public Object visit(ASTTDDDatasetDomain node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDDatasetLiteral node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

  /** Should not be visited! */
  public Object visit(ASTTDDDatasetMember node, Object data) throws ParseException {
    incorrectNode( node ); return data;
  }

}



// ***********************************************************************
// History Log:
// $Log: CPPUnparserVisitor.java,v $
// Revision 1.23  1998/06/30 15:53:38  mallet
// Replaced OutputStreamWriter over FileOutputStream by FileWriter.
//
// Revision 1.22  1998/06/30 13:36:08  mallet
// Added import of ORG.opengroup.adl.parsers.ParseException.
//
// Revision 1.21  1998/04/15 14:08:47  bareau
// Update according to the new NL mechanism: there are no NLD_* nodes.
//
// Revision 1.20  1998/03/31 14:15:49  wassel
// Merged back portable-code into trunk
//
// Revision 1.19.2.4  1998/03/31 08:31:43  wassel
// Removed visit to removed nodes PreProcessedUserFile and PreProcessedSystemFile.
//
// Revision 1.19.2.3  1998/03/30 08:19:55  wassel
// Added default visit to PreProcessedSystemFile node.
//
// Revision 1.19.2.2  1998/03/20 10:13:45  wassel
// In IncludeFileDeclaration visit, brakets are not added any more to the included
// file name. This is because the IncludeFileDeclaration definition is extended :
//   IncludeFileDeclaration - > #include "userfile"
//                          |   #include <userfile>
//
// Revision 1.19.2.1  1998/03/17 18:24:53  wassel
// Added visit to node ASTPreProcessedUserFiles.
//
// Revision 1.19  1998/03/10 17:26:49  mallet
// Merged v1.1 with development trunk
//
// Revision 1.18.2.1  1998/03/04 15:10:04  wassel
// Fixed bug in ForNode visit.
//
// Revision 1.18  1998/03/02 16:08:14  wassel
// Removed useless method indexChildrenEmit.
// Improved visit of ForNode and Declaration to prevent having newline when Declaration is the first child of ForNode.
//
// Revision 1.15  1998/02/26 10:54:35  bareau
// Bug fix: suppressed a blank between $(TARGET) and $(O).
// Variables DEP are now empty (no generated .o).
//
// Revision 1.14  1998/02/25 13:24:04  wassel
// Fixed bug in ForNode visit.
//
// Revision 1.13  1998/02/25 08:20:50  bareau
// Syntactic bug fixed in CPPMakefileGenerator.
// Changed a newline in CPPUnparserVisitor visit method for ASTDeclaration.
//
// Revision 1.12  1998/02/23 15:29:53  wassel
// Fixed bug in the previous commit of InitDeclarator visit.
//
// Revision 1.11  1998/02/19 09:59:46  wassel
// Changement in visit of Initializer and InitDeclarator to match jjt changements.
//
// Revision 1.10  1998/02/12 16:04:06  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.9  1998/02/12 13:04:39  mallet
// Fixed bug: generated files are now created in the correct output
// directory.
//
// Revision 1.8  1998/02/12 10:29:08  wassel
// Improved visit of statement nodes.
// In the visit of TranslationUnit, emit #ifndef/define/endif lines to prevent including generated file more than once.
//
// Revision 1.7  1998/02/11 09:46:02  wassel
// Fixed bug in CatchNode visit.
// Added visit to new nodes.
//
// Revision 1.6  1998/02/06 15:58:14  wassel
// Added visit to IncludeFileList.
//
// Revision 1.5  1998/02/05 16:18:38  wassel
// Added visit for Initializer.
// In TranslationUnit visit and after visit of IncludeFileDeclaration children, emit "#include filename" lines of node-specified included files.
//
// Revision 1.4  1998/01/30 12:34:38  wassel
// Added an entry for new ASTDDFieldDeclaration node.
//
// Revision 1.3  1998/01/27 16:35:43  wassel
// Changed visit to ExceptionDeclaration.
// Implmented visit to TypeNameList, ThrowStmt, ADLDomainList ADLNamedParamList ADLDomain.
// Removed visit to removed nodes ExternalDeclaration and FullyScopedId.
//
// Revision 1.2  1998/01/19 15:41:36  wassel
// Removed node ASTDDLongTest. Test directives can not be short test directive any more and are always long test directive now.
//
// Revision 1.1  1998/01/14 11:18:05  wassel
// C++ unparser visitor.
//
// ***********************************************************************

