// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/j/JNLDUnparserVisitor.java,v $
// $Revision: 1.11 $
// $Date: 1998/08/06 14:10:58 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   NLD/Java Unparser
// NAME:        JNLDUnparserVisitor.java
//
// HISTORY:
//      Cyrille Bareau, The Open Group RI - Apr 22, 1998: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.j;

import java.io.*;
import java.util.Vector;
import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.unparsers.NLDGenericUnparser;
import ORG.opengroup.adl.unparsers.DTD;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.*;
import ORG.opengroup.adl.symboltable.j.*;
import ORG.opengroup.adl.nlengine.NLEngine;

/** 
  * This class defines a visitor which can be applied to an AST
  * generated from the ADLJavaParser grammar. It does <B>NOT</B>
  * modify this AST: its only effect is to produce an SGML file
  * that corresponds to the documentation of the visited file.
  *
  * <p>The basic algorithm is to walk through the AST using the "visitor"
  * pattern (see Design Patterns from Gamma et al.) and emit the code on
  * an output stream maintained by the visitor itself.
  *
  * <p>The unparsing starts by calling the <a href="#unparse">unparse()</a>
  * method passing the top of AST to unparse as parameter.
  */

public class JNLDUnparserVisitor extends NLDGenericUnparser
  implements ADLJavaParserVisitor
{

  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.11 $";

  /** Temporary buffer for the currently unparsed assertion. */
  private StringBuffer assertionBuffer = null;

  /** Flag set when visiting an assertion. */
  private boolean inAssertion = false;

  /** Flag set when visiting a TDD definition (constant, dataset, factory). */
  private boolean inTddDef = false;

  /** Flag set when the current assertion is an "exception behavior:
    * assertion. */
  private boolean exceptionBhv;

  /** Name of the ADL/TDD class. */
  private String className;

  /** Index for anonymous directives. */
  private int anonDirectiveIndex;

  /** Flag for constructor behaviors. */
  private boolean constructor;

  static private final boolean debug = false;


  /**
    * Default constructor: call superclass constructor.
    *
    * @param toolname the name of the tool using this unparser visitor
    * @param nlengine the NL engine to deal with the Prolog interpreter
    */
  public JNLDUnparserVisitor(String toolname, NLEngine nlengine) {

    super(toolname, versionId, nlengine);
  }


  /**
    * Starts unparsing an AST.
    *
    * If the passed AST root is not an CompilationUnit, an error is
    * reported and the AST is not unparsed. Otherwise the AST is
    * visited.
    *
    * @param ast the root node of the AST to unparse
    * @see #visit(ORG.opengroup.adl.parsers.adljava.ASTCompilationUnit, java.lang.Object)
    */
  public void unparse(Node ast) {

    if (debug)
      System.out.println("NLD unparse");

    if ((assertNodeType(ast, ASTADL_CompilationUnit.class)) ||
        (assertNodeType(ast, ASTTDD_ClassDeclaration.class))) {
      ast.jjtAccept(this, null);
    } else {
      incorrectNode(ast);
      System.err.println("\twas expecting an adl or tdd class");
    }
  }


  /**
    * Visits a ADL_CompilationUnit.
    *
    * <p><b>Algorithm:</b><ul>
    * <li>get the name of the declared adlclass from the
    *     declaration node (always last one),
    * <li>open the source file for the class or interface (<code>outputDirectory</code>/<em>name</em>.sgml),
    * <li>emit the standard header,
    * <li>visit the child nodes,
    * <li>and finally close the file.
    * </ul>
    *
    * @param node the ADL_CompilationUnit to visit
    * @param data unused
    * @return data returned by last child visited or
    * null if the file cannot be opened.
    * @see ORG.opengroup.adl.unparsers.GenericUnparserVisitor#outputDirectory
    *
    */
  public Object visit(ASTADL_CompilationUnit node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_CompilationUnit");

    // get the name of the declared adlclass and open
    // the corresponding stream.
    
    // The declaration node is always the last child
    SimpleNode decl = null;
    if (node.jjtGetNumChildren() >= 1) {
      decl = (SimpleNode)node.jjtGetChild(node.jjtGetNumChildren() - 1);
    }

    if ((decl == null) ||
	!((decl instanceof ASTADL_ClassDeclaration) ||
	  (decl instanceof ASTTDD_ClassDeclaration))) {
      System.err.println(toolName + ": Can't get name of file: " +
			 "no adl or tdd class declaration");
      return data;
    }

    // now we know the decl node is an adl or tdd class
    className = decl.getName();

    if (decl instanceof ASTADL_ClassDeclaration)
      initialize_aco(className);
    else initialize_tdd(className);

    // try to open the output stream
    try {
      // Try to get best performance through a BufferedWriter on top of
      // a FileWriter
      String fileName = outputDirectory + File.separator + className +
        ((decl instanceof ASTADL_ClassDeclaration) ? "ACO" : "TDD") + ".sgml";

      if (Adlt.verboseMode()) {
	System.out.println("+++ generating " + fileName);
      }

      FileWriter outStream = new FileWriter(fileName);
      out = new PrintWriter(new BufferedWriter(outStream));
    } catch (IOException exc) {
      System.err.println("Error: " + exc.getMessage());
      return null;
    }

    // and visit children
    data = node.childrenAccept(this, data);

    // close stream and reset handle
    close();

    return data;
  }


  /**
   * Visits an ImportDeclaration.
   *
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTImportDeclaration node, Object data) {

    if (debug)
      System.out.println("ASTImportDeclaration");

    StringBuffer imported = (StringBuffer)node.jjtGetChild(0).jjtAccept(this, data);

    // if it's a package import, emit .*
    boolean onDemand = false;
    try {
      onDemand = ((Boolean)node.getProperty("onDemand")).booleanValue();
    } catch (PropertyAccessException e) {
      // ignore raised exception
    }

    emitImport(imported, onDemand);

    return data;
  }


  public Object visit(ASTTDD_UseDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_UseDeclaration");

    emitUseClass(node.getName());

    return data;
  }


  public Object visit(ASTADL_ClassDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_ClassDeclaration");

    LClass classScope = (LClass)((ADLClass)node.getScope()).getLClass();

    String fullName = classScope.getFullName();
    String extendsName = (classScope.getSuperClasses())[0].getName();
    String implementsName = null;
    LClass[] interf = classScope.getInterfaces();
    if (! (interf == null || interf.length == 0)) {
      implementsName = interf[0].getName();
      for (int i = 1; i < interf.length; i++) {
        implementsName += ", " + interf[i].getName();
      }
    }
    String modif = classScope.getModifier().toString();

    emitClassHeader(fullName, modif, extendsName, implementsName);

    emitClassEntity(node.getScope());

    if (node.jjtGetChild(1) instanceof ASTName) {
      emitClassInherited(((SimpleNode)node.jjtGetChild(1)).getName());
    }

    int nbChildren = node.jjtGetNumChildren();
    for (int i = 0; i < nbChildren; i++) {
      if (node.jjtGetChild(i) instanceof ASTADL_BehaviorDeclaration)
        node.jjtGetChild(i).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTADL_InlineDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_InlineDeclaration");

    return data;
  }


  public Object visit(ASTADL_BehaviorDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_BehaviorDeclaration");

    int childIndex = 0; boolean constructor = true;
    String resultType = null;
    Node child = node.jjtGetChild(0);

    if (child instanceof ASTResultType) {
      constructor = false;
      resultType = ((SimpleNode)child).getName();
      childIndex = 1;
      child = node.jjtGetChild(1);
    }

    String methodName = ((SimpleNode)child).getName();
    beginMethodTemplate(methodName);

    emitMethodEntity(methodName, node.getScope());

    if (! constructor) {
      emitResultType(resultType);
    }

    child = node.jjtGetChild(++childIndex); // formal parameters

    String paramType, paramName;
    int nbparam = child.jjtGetNumChildren();
    if (nbparam == 0) {
      emitParam("", "");
    }
    else for (int i = 0; i < nbparam; i++) {
      paramType = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(0)).getName();
      paramName = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(1)).getName();
      emitParam(paramType, paramName);
    }

    child = node.jjtGetChild(++childIndex);
    if (child instanceof ASTNameList) {
      String thrownList = ((SimpleNode)child.jjtGetChild(0)).getName();
      for (int i = 1; i < child.jjtGetNumChildren(); i++) {
        thrownList += ", " + ((SimpleNode)child.jjtGetChild(i)).getName();
      }
      emitThrownList(thrownList);
    }

    endPrototype();

    while (! (child instanceof ASTADL_BehaviorSpecification)) {
      child = node.jjtGetChild(++childIndex);
    }

    child.jjtAccept(this, data);

    endMethodTemplate();

    return data;
  }


  public Object visit(ASTADL_BehaviorSpecification node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_BehaviorSpecification");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADL_BehaviorClassification node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_BehaviorClassification");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADL_NormalBehavior node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_NormalBehavior");

    assertionBuffer = new StringBuffer("'ADL#Normal' = ");
    inAssertion = true;
    node.jjtGetChild(0).jjtAccept(this, data);
    inAssertion = false;
    emitNormal(query(assertionBuffer.toString(), node.getScope()), true);
    assertionBuffer = null;
    return data;
  }


  public Object visit(ASTADL_AbNormalBehavior node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_AbNormalBehavior");

    assertionBuffer = new StringBuffer("'ADL#Abnormal' = ");
    inAssertion = true;
    node.jjtGetChild(0).jjtAccept(this, data);
    inAssertion = false;
    emitNormal(query(assertionBuffer.toString(), node.getScope()), false);
    assertionBuffer = null;
    return data;
  }


  public Object visit(ASTADL_AssertionGroup node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_AssertionGroup");

    if (assertionBuffer == null) {
      node.childrenAccept(this, data);
    }
    else {
      assertionBuffer.append("'ADL#Group'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        assertionBuffer.append(", ");
        node.jjtGetChild(i).jjtAccept(this, data);
      }
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADL_SuperSemantics node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_SuperSemantics");

    emitSuperSemantics(constructor);

    return data;
  }


  public Object visit(ASTADL_Binding node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Binding");

    return data;
  }


  public Object visit(ASTADL_Statement node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Statement");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADL_IfStatement node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_IfStatement");

    if (assertionBuffer == null) {
      assertionBuffer = new StringBuffer();
      inAssertion = true;
      node.jjtGetChild(0).jjtAccept(this, data); // visit the condition
      inAssertion = false;

      emitCondition(query(assertionBuffer.toString(), node.getScope()),
                    ((SimpleNode)node.jjtGetChild(0)).getExpressionImage(),
                    node.jjtGetNumChildren() > 2);

      assertionBuffer = null;

      node.jjtGetChild(1).jjtAccept(this, data); // "then" branch
      if (node.jjtGetNumChildren() > 2) {
        emitElse();
        node.jjtGetChild(2).jjtAccept(this, data); // "else" branch
      }
      emitEndCondition();
    }
    else { // nested if
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(" '?' ");
      node.jjtGetChild(1).jjtAccept(this, data);
      assertionBuffer.append(" ':' ");
      if (node.jjtGetNumChildren() == 3)
        node.jjtGetChild(2).jjtAccept(this, data);
      else
        assertionBuffer.append("true");
    }
    return data;
  }


  public Object visit(ASTADL_TryStatement node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_TryStatement");

    int nbChildren = node.jjtGetNumChildren();
    if (assertionBuffer == null) {
      emitTry((int)(nbChildren-1)/2);
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < nbChildren; i += 2) {
        emitCatch(((SimpleNode)node.jjtGetChild(i).jjtGetChild(0)).getName(),
                  ((SimpleNode)node.jjtGetChild(i).jjtGetChild(1)).getName());
        node.jjtGetChild(i+1).jjtAccept(this, data);
      }
      emitEndTry();
    }
    else { // try in nested
      assertionBuffer.append("'ADL#TryCatch'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < nbChildren; i += 2) {
        assertionBuffer.append(", [['")
          .append(((SimpleNode)node.jjtGetChild(i).jjtGetChild(0)).getName())
          .append("', '")
          .append(((SimpleNode)node.jjtGetChild(i).jjtGetChild(1)).getName())
          .append("'], ");
        node.jjtGetChild(i+1).jjtAccept(this, data);
        assertionBuffer.append(']');
      }
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADL_Quantifier node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Quantifier");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADL_Assertion node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Assertion");

    if (assertionBuffer == null) {
      String label = "";
      if (node.jjtGetChild(0) instanceof ASTADL_Label) {
        label = ((ASTADL_Label)node.jjtGetChild(0)).getName();
      }
      // omit tags???

      Node lastChild = node.jjtGetChild(node.jjtGetNumChildren()-1);
      if (lastChild instanceof ASTADL_QuantifiedAssertion) {
        lastChild.jjtAccept(this, label);
      }
      else { // expression
        assertionBuffer = new StringBuffer();
        inAssertion = true;
        exceptionBhv = false;
        lastChild.jjtAccept(this, data);
        inAssertion = false;
        emitAssertion(query(assertionBuffer.toString(), node.getScope()),
                      label, node.getExpressionImage(), exceptionBhv);
        assertionBuffer = null;
      }
    }
    else { // nested assertion
      node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTADL_Tags node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Tags");

    return data;
  }


  public Object visit(ASTADL_QuantifiedAssertion node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_QuantifiedAssertion: " + inAssertion + " " + assertionBuffer);

    String label = (String)data;
    ASTADL_Domain dom;
    ASTADL_DomainList domlist = (ASTADL_DomainList)node.jjtGetChild(1);

    emitQuantifier(label,
		   ((ASTADL_Quantifier)node.jjtGetChild(0)).getName(),
		   domlist.getExpressionImage());
    String name, expr;
    for (int i = 0; i < domlist.jjtGetNumChildren(); i++) {
      dom = (ASTADL_Domain)domlist.jjtGetChild(i);
      inAssertion = true; assertionBuffer = new StringBuffer();
      dom.jjtGetChild(1).jjtAccept(this, null);
      expr = query(assertionBuffer.toString(), node.getScope());
      inAssertion = false; assertionBuffer = null;
      emitDomain(((SimpleNode)dom.jjtGetChild(0).jjtGetChild(1)).getName(),
                 expr, i);
    }
    node.jjtGetChild(2).jjtAccept(this, null);
    emitEndQuantified();

    return data;
  }


  public Object visit(ASTADL_ImplExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_ImplExpression");

    node.childrenAccept(this, data);
    return data;
  }


  public Object visit(ASTADL_ImplOp node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_ImplOp");

    if (inAssertion) {
      String op = node.getOp();
      assertionBuffer.append(" '" + op + "' ");
      if (op.equals("<:>")) {
        exceptionBhv = true;
      }
    }
    return data;
  }


  /**
   * Visits a ConditionalExpression.
   *
   * Note: this node is only created when there are 3 children
   * 
   * @param node the visited ConditionalExpression
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTConditionalExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTConditionalExpression");

    if (inAssertion) {
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(" '?' ");
      node.jjtGetChild(1).jjtAccept(this, data);
      assertionBuffer.append(" ':' ");
      node.jjtGetChild(2).jjtAccept(this, data);
    }
    else if (inTddDef) {
      return node.jjtGetChild(0).jjtAccept(this, data) + " '?' " +
             node.jjtGetChild(1).jjtAccept(this, data) + " ':' " +
             node.jjtGetChild(2).jjtAccept(this, data);
    }
    return data;
  }

  /**
   * Visits a ConditionalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>||</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ConditionalOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTConditionalOrExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTConditionalOrExpression");

    // just unparse children separated by "||"
    return unparseList(node, 0, "||", data);
  }

  /**
   * Visits a ConditionalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>&&</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ConditionalAndExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTConditionalAndExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTConditionalAndExpression");

    // just unparse children separated by "&&"
    return unparseList(node, 0, "&&", data);
  }

  /**
   * Visits an InclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>|</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited InclusiveOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTInclusiveOrExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTInclusiveOrExpression");

    // just unparse children separated by "|"
    return unparseList(node, 0, "|", data);
  }

  /**
   * Visits an ExclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>^</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ExclusiveOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTExclusiveOrExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTExclusiveOrExpression");

    // just unparse children separated by "^"
    return unparseList(node, 0, "^", data);
  }

  /**
   * Visits an AndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>&</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited AndExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTAndExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTAndExpression");

    // just unparse children separated by "&"
    return unparseList(node, 0, "&", data);
  }

  /**
   * Visits an EqualityExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited EqualityExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], java.lang.Object)
   */
  public Object visit(ASTEqualityExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTEqualityExpression");

    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}
    
    // unparse list of expression children
    return unparseList(node, 0, operators, data);
  }

  /**
   * Visits an InstanceOfExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child
   * <li>emit <code>instanceof</code>
   * <li>visit second child
   * </ul>
   *
   * Note: this node is only created when there are 2 children
   *
   * @param node the visited InstanceOfExpression
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTInstanceOfExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTInstanceOfExpression");

    // there are always 2 children
    
    if (inAssertion) {
      assertionBuffer.append("'ADL#InstanceOf'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(", ");
      node.jjtGetChild(1).jjtAccept(this, data);
      assertionBuffer.append("])");
    }
    return data;
  }

  /**
   * Visits a RelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited RelationalExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], java.lang.Object)
   */
  public Object visit(ASTRelationalExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTRelationalExpression");

    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, data);
  }

  /**
   * Visits a ShiftExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ShiftExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], java.lang.Object)
   */
  public Object visit(ASTShiftExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTShiftExpression");

    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, data);
  }

  /**
   * Visits an AdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited AdditiveExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], java.lang.Object)
   */
  public Object visit(ASTAdditiveExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTAdditiveExpression");

    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, data);
  }

  /**
   * Visits a MultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited MultiplicativeExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], java.lang.Object)
   */
  public Object visit(ASTMultiplicativeExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTMultiplicativeExpression");

    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, data);
  }

  /**
   * Visits an UnaryPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator, accessible via the <code>op</code> property
   * <li>visit first and only child
   * </ul>
   *
   * @param node the visited UnaryPlusMinus
   * @param data unused
   * @return the data by visited child
   */
  public Object visit(ASTUnaryPlusMinus node, Object data) {

    if (debug)
      System.out.println("visit ASTUnaryPlusMinus");

    if (inAssertion) {
      try { assertionBuffer.append(" '")
              .append((String)node.getProperty("op"))
              .append("' "); }
      catch (PropertyAccessException e) { emit("<MissingOperator>"); }
    }
    // just one child
    node.jjtGetChild(0).jjtAccept(this, data);
    return data;
  }

  /**
   * Visits an UnaryNotPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator, accessible via the <code>op</code> property
   * <li>visit first and only child
   * </ul>
   *
   * @param node the visited UnaryNotPlusMinus
   * @param data unused
   * @return the data by visited child
   */
  public Object visit(ASTUnaryNotPlusMinus node, Object data) {

    if (debug)
      System.out.println("visit ASTUnaryNotPlusMinus");

    if (inAssertion) {
      try { assertionBuffer.append(" '")
              .append((String)node.getProperty("op"))
              .append("' "); }
      catch (PropertyAccessException e) { emit("<MissingOperator>"); }
    }
    // just one child
    node.jjtGetChild(0).jjtAccept(this, data);
    return data;
  }


  public Object visit(ASTADL_CallStateExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_CallStateExpression");

    assertionBuffer.append("'@' ");
    node.jjtGetChild(0).jjtAccept(this, data);
    return data;
  }


  /**
   * Visits a CastExpression.
   *
   * @param node the visited CastExpression
   * @param data unused
   * @return the data returned by visiting second child
   */
  public Object visit(ASTCastExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTCastExpression");

    return node.childrenAccept(this,data);
  }


  public Object visit(ASTADL_Unchanged node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Unchanged");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Unchanged'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append("])");
    }
    return data;
  }


  public Object visit(ASTADL_Normal node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Normal");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Normal'");
    }
    return data;
  }


  public Object visit(ASTADL_Abnormal node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Abnormal");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Abnormal'");
    }
    return data;
  }


  public Object visit(ASTADL_Return node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Return");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Return'");
    }
    return data;
  }


  /**
   * Visits a PrimaryExpression.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTPrimaryExpression node, Object data) {

    if (debug) {
      System.out.println("visit ASTPrimaryExpression");
    }

    Node prefix = node.jjtGetChild(0);
    String prefixName = null;
    if (prefix instanceof ASTName) {
      prefixName = ((ASTName)prefix).getName();
    }

    if (inAssertion) {
      Node child;
      int j, nbParenth = 0, nbChildren = node.jjtGetNumChildren();
      boolean comma = false, staticMethod = false;
      LCallable methScope;
      for (int i = nbChildren-1; i > 0; i--) {
        if (comma) {
          assertionBuffer.append(", ");
          comma = false;
        }

        child = node.jjtGetChild(i);
        if (child instanceof ASTBracketSuffix) {
          assertionBuffer.append("'ADL#Array'(");
          child.jjtGetChild(0).jjtAccept(this, data);
          assertionBuffer.append(", ");
          nbParenth += 1;
        }
        else if (child instanceof ASTDotSuffix) {
          assertionBuffer.append("'ADL#Field'('" + ((ASTDotSuffix)child).getName() + "', ");
          nbParenth += 1;
        }
        else { // (child instanceof ASTArguments)
          methScope = null;
          j = i;
          do { j -= 1; }
          while (! ((node.jjtGetChild(j) instanceof ASTDotSuffix) ||
                    (j == 0)));
          try {
            methScope = (LCallable)((SimpleNode)node.jjtGetChild(j))
              .getProperty("NLDScope");
          }
          catch(PropertyAccessException e){} 
          staticMethod = (nbChildren == 2) &&
            ((JModifier)methScope.getModifier()).isStatic();
          if (! (staticMethod ||
                 ((nbChildren == 2) &&
                  (prefixName != null) &&
                  (prefixName.indexOf('.') == -1)))) {
            // ignore the case "method(<args>)" (with method unqualified)
            assertionBuffer.append("'ADL#Object'(");
            nbParenth += 1;
          }
          i -= 1; // ignore the preceding dotsuffix
          assertionBuffer.append("'ADL#Func'(");
          assertionBuffer.append("['")
            .append(methScope.getName())
            .append(methScope.getNLSignature())
            .append("', [")
            .append(nlengine.NLScopeScopeList(methScope.getParent()))
            .append("]]");
          assertionBuffer.append(", [");
          child.jjtAccept(this, data);
          assertionBuffer.append("])");
          comma = true;
        }
      }

      if (node.jjtGetChild(1) instanceof ASTArguments) {
        if (! (staticMethod || (prefixName.indexOf('.') == -1))) {
          assertionBuffer.append(", '")
            .append(prefixName.substring(0, prefixName.lastIndexOf('.')))
            .append("'");
        }
      }
      else {
        if (comma)
          assertionBuffer.append(", ");
        node.jjtGetChild(0).jjtAccept(this, data);
      }

      for (int i = nbParenth; i > 0; i--) {
        assertionBuffer.append(")");
      }

      return data;
    } // end if (inAssertion)

    else if (inTddDef && (prefixName != null)) {
      Scope exprScope = null;
      try {
        exprScope = (Scope)((SimpleNode)node.jjtGetChild(0)).getProperty("NLDScope");
      } catch(PropertyAccessException e) {}
      if ((exprScope instanceof LMethod) &&
          (((LMethod)exprScope).isTestMethod())) {
        emitTestFunction(((LMethod)exprScope).getName(),
                         ((LMethod)exprScope).getNLSignature());
      }
    }

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDD_ADLnewExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_ADLnewExpression");

    emitACO(((ASTName)node.jjtGetChild(0)).getName());

    return data;
  }


  public Object visit(ASTTDD_ADLExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_ADLExpression");

    emitACO(node.getTypeName());

    return data;
  }


  public Object visit(ASTThisPrefix node, Object data) {

    if (debug)
      System.out.println("visit ASTThisPrefix");

    if (inAssertion) {
      assertionBuffer.append("'ADL#This'");
    }
    return data;
  }


  /**
   * Visits a ParenthExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>visit first and only child
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited ParenthExpression
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTParenthExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTParenthExpression");

    if (inAssertion) {
      assertionBuffer.append('(');
      data = node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append(')');
    }
    return data;
  }

  /**
   * Visits a BracketSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>[</code>
   * <li>visit first and only child
   * <li>emit <code>]</code>
   * </ul>
   *
   * @param node the visited BracketSuffix
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTBracketSuffix node, Object data) {

    if (debug)
      System.out.println("visit ASTBracketSuffix");

    // there is always one single child
    if (inAssertion) {
      assertionBuffer.append(" 'ADL#Array' ");
      data = node.jjtGetChild(0).jjtAccept(this, data);
    }
    return data;
  }

  /**
   * Visits a DotSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>.</code> followed by identifier, accessible via
   *     <code>getName()</code>
   * </ul>
   *
   * @param node the visited 
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDotSuffix node, Object data) {

    if (debug)
      System.out.println("visit ASTDotSuffix");

    return data;
  }

  /**
   * Visits a NullLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>null</code>
   * </ul>
   *
   * @param node the visited NullLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTNullLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTNullLiteral");

    // value of a null literal is "null"
    if (inAssertion)
      assertionBuffer.append("'null'");
    else if (inTddDef)
      return "'null'";
    return data;
  }

  /**
   * Visits an IntegerLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited IntegerLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTIntegerLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTIntegerLiteral");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      try {
        String val = (String)node.getProperty("value");
        // Prolog does not recognize final letter ('L', 'D', etc.)
        if (Character.isLetter(val.charAt(val.length()-1))) {
          val = val.substring(0, val.length()-1);
        }
        if (inAssertion)
          assertionBuffer.append(val);
        else return val;
      }
      catch (PropertyAccessException e) { emit("<MissingValue>"); }
    }
    return data;
  }

  /**
   * Visits an FloatLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited FloatLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTFloatLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTFloatLiteral");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      try {
        String val = (String)node.getProperty("value");
        // Prolog does not recognize final letter ('L', 'D', etc.)
        if (Character.isLetter(val.charAt(val.length()-1))) {
          val = val.substring(0, val.length()-1);
        }
        if (inAssertion)
          assertionBuffer.append(val);
        else return val;
      }
      catch (PropertyAccessException e) { emit("<MissingValue>"); }
    }
    return data;
  }

  /**
   * Visits an CharacterLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited CharacterLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTCharacterLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTCharacterLiteral");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      try {
        String val = (String)node.getProperty("value");
        if (inAssertion)
          assertionBuffer.append(val);
        else return val;
      }
      catch (PropertyAccessException e) { emit("<MissingValue>"); }
    }
    return data;
  }

  /**
   * Visits an StringLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited StringLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTStringLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTStringLiteral");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      String strvalue = null;
      try { strvalue = (String)node.getProperty("value"); }
      catch (PropertyAccessException e) { }
      if (inAssertion)
        assertionBuffer.append("\'\"")
          .append(strvalue.substring(1, strvalue.length()-1))
          .append("\"\'");
      else return "\'\"" + strvalue.substring(1, strvalue.length()-1) + "\"\'";
    }
    return data;
  }

  /**
   * Visits an BooleanLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited BooleanLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBooleanLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTBooleanLiteral");

    // value of a literal is accessible via the value property
    if (inAssertion || inTddDef) {
      try {
        String val = (String)node.getProperty("value");
        if (inAssertion)
          assertionBuffer.append(val);
        else return val;
      }
      catch (PropertyAccessException e) { }
    }
    return data;
  }

  /**
   * Visits an Arguments.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>if there is one child, visit it
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited Arguments
   * @param data unused
   * @return the unused data or the data returned by visited child
   */
  public Object visit(ASTArguments node, Object data) {

    if (debug)
      System.out.println("visit ASTArguments");

    node.childrenAccept(this, data);
    return data;
  }

  /**
   * Visits an ArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>,</code>
   * </ul>
   *
   * @param node the visited ArgumentList
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTArgumentList node, Object data) {

    if (debug)
      System.out.println("visit ASTArgumentList");

    if (inAssertion) {
      // visit first child, then loop on remaining ones
      node.jjtGetChild(0).jjtAccept(this, data);
      for (int i = 1; i < node.jjtGetNumChildren(); i++) {
        assertionBuffer.append(", ");
        node.jjtGetChild(i).jjtAccept(this, data);
      }
    }
    return data;
  }

  /**
   * Visits an AllocationExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>new</code>
   * <li>visit children
   * </ul>
   *
   * @param node the visited AllocationExpression
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTAllocationExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTAllocationExpression");

    if (inAssertion && (node.jjtGetChild(1) instanceof ASTArguments)) {
      LCallable constrScope = null;
      try {
        constrScope = (LCallable)node.getProperty("NLDScope");
      }
      catch(PropertyAccessException e) {}
      assertionBuffer.append("'ADL#Func'(")
        .append("['")
        .append(constrScope.getName())
        .append(constrScope.getNLSignature())
        .append("', [")
        .append(nlengine.NLScopeScopeList(constrScope.getParent()))
        .append("]], [");
      node.jjtGetChild(1).jjtAccept(this, data);
      assertionBuffer.append("])");
      return data;
    }

    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ArrayDimensions.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for each child (dimension with an expressed value),<ul>
   * <li>emit <code>[</code>
   * <li>visit the child
   * <li>emit <code>]</code></ul>
   * <li>get <code>arrayLevel</code> property for array type, substract
   * the number of children, and while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited ArrayDimensions
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTArrayDimensions node, Object data) {

    if (debug)
      System.out.println("visit ASTArrayDimensions");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTADL_ThrownExpression node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_ThrownExpression");

    if (inAssertion) {
      assertionBuffer.append("'ADL#Thrown'([");
      node.jjtGetChild(0).jjtAccept(this, data);
      assertionBuffer.append("])");
    }

    return data;
  }


  public Object visit(ASTADL_DomainList node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_DomainList");

    return data;
  }


  public Object visit(ASTADL_Domain node, Object data) {

    if (debug)
      System.out.println("visit ASTADL_Domain");

    return data;
  }


  public Object visit(ASTVariableDeclaratorId node, Object data) {

    if (debug)
      System.out.println("visit ASTVariableDeclaratorId");

    return data;
  }


  /**
   * Visits a FormalParameters.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>if there is at least one child, unparse list of children
   *     separated by <code>,</code>
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited FormalParameters
   * @param data unused
   * @return the data returned by last visited child or unused data
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTFormalParameters node, Object data) {

    if (debug)
      System.out.println("visit ASTFormalParameters");

    return data;
  }

  /**
   * Visits a FormalParameter.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child for the type of the parameter
   * <li>emit a space
   * <li>visit second child for the VariableDeclaratorId
   * </ul>
   *
   * @param node the visited FormalParameter
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTFormalParameter node, Object data) {

    if (debug)
      System.out.println("visit ASTFormalParameter");

    return node.childrenAccept(this, data);
  }

  /**
   * Visits a Type.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit type name, accessible via <code>getName()</code>
   * <li>get <code>arrayLevel</code> property for array type, while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited Type
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTType node, Object data) {

    if (debug)
      System.out.println("visit ASTType");

    return node.childrenAccept(this, data);
  }

  /**
   * Visits a ResultType.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if there is one child, visit it
   * <li>otherwise emit <code>void</code>
   * </ul>
   *
   * @param node the visited 
   * @param data unused
   * @return the data returned by visited child or unused data
   */
  public Object visit(ASTResultType node, Object data) {

    if (debug)
      System.out.println("visit ASTResultType");

    return node.childrenAccept(this, data);
  }

  /**
   * Visits a Name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit node name, accessible via <code>getName()</code>
   * </ul>
   *
   * @param node the visited Name
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTName node, Object data) {

    if (debug) {
      System.out.println("visit ASTName: " + node.getName());
    }

    // A name is just a sequence of ids separated by "."
    // the full sequence is accessible via getName()
    if (inAssertion) {
      assertionBuffer.append("'").append(node.getName()).append("'");
    }
    else if (inTddDef) {
      String cstClass = null, ret;
      try {
        cstClass = (String)node.getProperty("enclosingConstructName");
        if (cstClass.equals(className)) {
          ret = makeLocalLink(className, "TDD-CSTS", node.getName());
        }
        else {
          ret = makeTDDExternalLink(cstClass, cstClass + "TDD-CSTS", node.getName());
        }
      } catch(PropertyAccessException e) {
        ret = node.getName();
      }
      return ret;
      //      return makeLocalLink(className, "TDD-CSTS", node.getName());
    }
    return data;
  }

  /**
   * Visits a NameList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>,</code>
   * </ul>
   *
   * Note: there is always at least one child
   *
   * @param node the visited NameList
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, java.lang.Object)
   */
  public Object visit(ASTNameList node, Object data) {

    if (debug)
      System.out.println("visit ASTNameList");

    int nbChildren = node.jjtGetNumChildren();
    node.jjtGetChild(0).jjtAccept(this, data);
    for (int i = 1; i < nbChildren; i++) {
      assertionBuffer.append(", ");
      node.jjtGetChild(i).jjtAccept(this, data);
    }
    return data;
  }


  public Object visit(ASTTDD_ClassDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_ClassDeclaration");

    emitClassEntity(node.getScope());

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDD_ClassBodyDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_ClassBodyDeclaration");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDD_FieldDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_FieldDeclaration");

    String typename = ((SimpleNode)node.jjtGetChild(0)).getName();
    inTddDef = true;
    for (int i = 1; i < node.jjtGetNumChildren(); i += 2) {
      emitConstant(typename,
                   ((SimpleNode)node.jjtGetChild(i)).getName(),
                   (String)node.jjtGetChild(i+1).jjtAccept(this, data));
    }
    inTddDef = false;
    return data;
  }


  public Object visit(ASTTDD_DatasetDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetDeclaration");

    String expr;
    inTddDef = true;
    expr = (String)node.jjtGetChild(1).jjtAccept(this, data);
    inTddDef = false;

    emitDataset(((SimpleNode)node.jjtGetChild(0).jjtGetChild(1)).getName(), // name
                ((SimpleNode)node.jjtGetChild(0).jjtGetChild(0)).getName(), // type
                expr); 

    return data;
  }


  public Object visit(ASTTDD_FactoryDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_FactoryDeclaration");

    Node methDecl = node.jjtGetChild(0);
    String factName = ((SimpleNode)methDecl.jjtGetChild(1)).getName();
    beginFactoryTemplate(factName);
    emitMethodEntity(factName, ((SimpleNode)methDecl).getScope());

    emitResultType((String)((SimpleNode)methDecl.jjtGetChild(0)).getName());

    Node child = methDecl.jjtGetChild(1).jjtGetChild(0); // params

    String paramType, paramName;
    int nbparam = child.jjtGetNumChildren();
    if (nbparam == 0) {
      emitParam("", "");
    }
    else for (int i = 0; i < nbparam; i++) {
      paramType = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(0)).getName();
      paramName = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(1)).getName();
      emitParam(paramType, paramName);
    }

    child = methDecl.jjtGetChild(2); // thrown list???
    if (child instanceof ASTNameList) {
      String thrownList = ((SimpleNode)child.jjtGetChild(0)).getName();
      for (int i = 1; i < child.jjtGetNumChildren()-1; i++) {
        thrownList += ", " + ((SimpleNode)child.jjtGetChild(i)).getName();
      }
      emitThrownList(thrownList);
    }

    emitCode(node.getSourceText());

    endFactoryTemplate();

    return data;
  }


  public Object visit(ASTMethodDeclaration node, Object data) {

    if (debug)
      System.out.println("visit ASTMethodDeclaration");

    String funcName = (String)((SimpleNode)node.jjtGetChild(1)).getName();
    beginTestFuncTemplate(funcName,
                         ((LCallable)node.getScope()).getNLSignature());
    emitMethodEntity(funcName, node.getScope());

    emitResultType((String)((SimpleNode)node.jjtGetChild(0)).getName());

    Node child = node.jjtGetChild(1).jjtGetChild(0); // params

    String paramType, paramName;
    int nbparam = child.jjtGetNumChildren();
    if (nbparam == 0) {
      emitParam("", "");
    }
    else for (int i = 0; i < nbparam; i++) {
      paramType = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(0)).getName();
      paramName = ((SimpleNode)child.jjtGetChild(i).jjtGetChild(1)).getName();
      emitParam(paramType, paramName);
    }

    child = node.jjtGetChild(2); // thrown list???
    if (child instanceof ASTNameList) {
      String thrownList = ((SimpleNode)child.jjtGetChild(0)).getName();
      for (int i = 1; i < child.jjtGetNumChildren()-1; i++) {
        thrownList += ", " + ((SimpleNode)child.jjtGetChild(i)).getName();
      }
      emitThrownList(thrownList);
    }

    emitCode(node.getSourceText());

    //    node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);

    endTestFuncTemplate();

    return data;
  }


  public Object visit(ASTTDD_TestDirective node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_TestDirective");

    String dirName;
    try {
      dirName = (String)node.getProperty("name");
    } catch(PropertyAccessException e) {
      dirName = "directive_" + anonDirectiveIndex;
      anonDirectiveIndex++;
    }
    beginDirectiveTemplate(dirName);

    int nbChildren = node.jjtGetNumChildren();
    for(int i = 0; i < nbChildren-1; i++) {
      node.jjtGetChild(i).jjtAccept(this, data);
    }

    emitCode(node.getSourceText());

    inTddDef = true;
    node.jjtGetChild(nbChildren-1).jjtAccept(this, data);
    inTddDef = false;

    endDirectiveTemplate();

    return data;
  }


  public Object visit(ASTTDD_DatasetDomain node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetDomain");

    inTddDef = true;
    String ds = (String)node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);
    inTddDef = false;

    emitDomain(ds);

    return data;
  }


  public Object visit(ASTTDD_DatasetLiteral node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetLiteral");

    String ret = "{ ";
    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 0) {
      return ret + "}";
    }
    ret += (String)node.jjtGetChild(0).jjtAccept(this, data);
    for (int i = 1; i < nbChildren; i++) {
      ret += ", " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }
    return ret + " }";
  }


  public Object visit(ASTTDD_SingleDeclarator node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_SingleDeclarator");

    return node.childrenAccept(this, data);
  }


  public Object visit(ASTTDD_DatasetMember node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetMember");

    String ret = (String)node.jjtGetChild(0).jjtAccept(this, data);
    if (node.jjtGetNumChildren() == 2)
      ret += " .. " +
        (String)node.jjtGetChild(1).jjtAccept(this, data);

    return ret;
  }


  public Object visit(ASTTDD_DatasetExpr node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetExpr");

    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 1) {
      return node.jjtGetChild(0).jjtAccept(this, data);
    }

    String ret = (String)node.jjtGetChild(0).jjtAccept(this, data);
    for (int i = 1; i < nbChildren; i++) {
      ret += " + " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }

    return ret;
  }


  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetConcatenationExpr");

    return node.jjtGetChild(0).jjtAccept(this, data);
  }


  public Object visit(ASTTDD_DatasetSingleton node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_DatasetSingleton");

    SimpleNode child = (SimpleNode)node.jjtGetChild(0);
    if (child instanceof ASTName) {
      String dsName = child.getName(), dsClass = null, kind = null, ret;
      try{
        kind = (String)child.getProperty("astNameKind");
        dsClass = (String)child.getProperty("enclosingConstructName");
      } catch(PropertyAccessException e) { }

      if (kind.equals("TDDDataset")) {
        if (dsClass.equals(className)) {
          ret = makeLocalLink(dsName, dsName);
        }
        else {
          ret = makeTDDExternalLink(dsClass, dsName, dsName);
        }
        return ret;
      }
      else { // constant name

        if (dsClass.equals(className)) {
          ret = makeLocalLink(className, "TDD-CSTS", dsName);
        }
        else {
          ret = makeTDDExternalLink(dsClass, dsClass + "TDD-CSTS", dsName);
        }
        return ret;
      }
    }
    else return child.jjtAccept(this, data);
  }


  public Object visit(ASTTDD_FactoryCall node, Object data) {

    if (debug)
      System.out.println("visit ASTTDD_FactoryCall");

    String factClass = null;
    try {
      factClass = (String)((SimpleNode)node.jjtGetChild(0)).getProperty("enclosingConstructName");
    } catch(PropertyAccessException e) {}

    String factName = ((SimpleNode)node.jjtGetChild(0)).getName(); 

    String ret;
    if (factClass.equals(className)) {
      ret = makeLocalLink(factName, factName) + " ( ";
    }
    else {
      ret = makeTDDExternalLink(factClass, factName, factName) + " ( ";
    }

    int nbChildren = node.jjtGetNumChildren();
    if (nbChildren == 1) {
      return ret + ")";
    }
    ret += (String)node.jjtGetChild(1).jjtAccept(this, data);
    for (int i = 2; i < nbChildren; i++) {
      ret += ", " + (String)node.jjtGetChild(i).jjtAccept(this, data);
    }
    return ret + " )";
  }


  public Object visit(ASTExpression node, Object data) {

    return node.jjtGetChild(0).jjtAccept(this, data);
  }


  public Object visit(ASTVariableInitializer node, Object data) {

    return node.jjtGetChild(0).jjtAccept(this, data);
  }


  // Nodes that should not be visited ("pure" imperative Java).
  // Recall: prologues and epilogues children are not visited.


  public Object visit(ASTADL_Prologue node, Object data) {

    return node.childrenAccept(this, data);;
  }

  public Object visit(ASTADL_Epilogue node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTADL_Label node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTCompilationUnit node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTMethodDeclarator node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTClassBodyDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTInterfaceDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTInterfaceMemberDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTFieldDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTStaticInitializer node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTConstructorDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTVariableDeclarator node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTBlock node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTExplicitConstructorInvocation node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTBlockStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTLabeledStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTEmptyStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTSwitchStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTSwitchLabel node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTIfStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTWhileStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTDoStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTForStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTForInit node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTBreakStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTContinueStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTReturnStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTThrowStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTSynchronizedStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTTryStatement node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTPreIncrementExpression node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTPreDecrementExpression node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTLocalVariableDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTAssignment node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTPostfixExpression node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTStatementExpressionList node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTForUpdate node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(SimpleNode node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTPackageDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTSuperPrefix node, Object data) {

    return node.childrenAccept(this, data);
  }

  public Object visit(ASTClassDeclaration node, Object data) {

    return node.childrenAccept(this, data);
  }


  /**
   * Unparse a list of <em>something</em> with a single separator.
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node. But we trust 
   * our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separator the string that separates children
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     String separator, Object data) {

    int numChildren = node.jjtGetNumChildren();

    // visit first child, then loop on remaining ones
    node.jjtGetChild(firstChild).jjtAccept(this, data);
    for (int childNum = firstChild + 1; childNum < numChildren; childNum++) {
      assertionBuffer.append(" '");
      assertionBuffer.append(separator);
      assertionBuffer.append("' ");
      node.jjtGetChild(childNum).jjtAccept(this, data);
    }
    return data;
  }

  /**
   * Unparse a list of <em>something</em> each having its own separator.
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node and when the size
   * of the array of separators does not match the number of children to
   * visit - 1. But we trust our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separators the array of strings that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     String[] separators, Object data) {

    if (inAssertion) {
      // get number of children
      int numChildren = node.jjtGetNumChildren();

      // visit first child, then loop on remaining ones
      node.jjtGetChild(firstChild).jjtAccept(this, data);
      for (int childNum = firstChild + 1; childNum < numChildren; childNum++) {
        assertionBuffer.append(" '");
        assertionBuffer.append(separators[childNum - 1]);
        assertionBuffer.append("' ");
        node.jjtGetChild(childNum).jjtAccept(this, data);
      }
    }
    return data;
  }

}

// ***********************************************************************
// History Log:
// $Log: JNLDUnparserVisitor.java,v $
// Revision 1.11  1998/08/06 14:10:58  mallet
// Merged branch v1_2-patches back onto the trunk.
//
// Revision 1.10.2.1  1998/07/08 13:54:59  mallet
// Fixed bug #143: having a field which identifier starts with an
// uppercase letter causes the Prolog engine to enter an infinite
// loop. Simple quotes have been added around an ADL#Field identifier.
//
// Revision 1.10  1998/06/30 15:57:50  mallet
// Replaced OutputStreamWriter over FileOutputStream by FileWriter.
// Replaced access to expressionImage property by call to
// getExpressionImage() or getSourceText() depending on usage in
// generated documentation.
//
// Revision 1.9  1998/06/30 10:14:12  bareau
// Unparsing of ADLnew expressions and allocation expressions.
//
// Revision 1.8  1998/06/12 13:22:16  bareau
// Added support for NL unparsing of quantified assertions.
//
// Revision 1.7  1998/06/11 16:06:14  bareau
// Added support for nested if expressions (they are considered exactly
// like "?:" conditional expressions).
//
// Revision 1.6  1998/06/11 13:48:12  bareau
// Added support for nested try blocks.
//
// Revision 1.5  1998/06/11 10:21:56  bareau
// Added support for nested assertion groups.
//
// Revision 1.4  1998/06/10 15:55:31  bareau
// Minor bugs fixes.
//
// Revision 1.3  1998/06/03 07:49:05  bareau
// Minor bug fixes. Suppressed unicode processing.
//
// Revision 1.2  1998/05/26 15:33:39  bareau
// Added links to external constants, inherited classes, inherited behavior
// and "used" tdd classes.
//
// Revision 1.1  1998/05/15 13:45:03  bareau
// Unparser for generating documentation for ADL/TDD Java classes.
//
//
// ***********************************************************************
