// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
// 
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
// 
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/ORG/opengroup/adl/unparsers/j/JUnparserVisitor.java,v $
// $Revision: 1.17 $
// $Date: 1998/06/30 15:53:39 $
// PRODUCT:     Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:   ADL/Java Compiler
// NAME:        JUnparserVisitor.java
//
// HISTORY:
//      Lionel Mallet, The Open Group RI - Nov 25, 1997: Created
//      See Log at end of file
//
// ***********************************************************************

package ORG.opengroup.adl.unparsers.j;

import java.io.*;
import ORG.opengroup.adl.compiler.Adlt;
import ORG.opengroup.adl.unparsers.GenericUnparserVisitor;
import ORG.opengroup.adl.parsers.*;
import ORG.opengroup.adl.parsers.adljava.*;
import ORG.opengroup.adl.symboltable.LType;
import ORG.opengroup.adl.symboltable.j.JModifier;

/** 
 * This class defines a visitor which can be applied to an AST
 * generated from the ADLJavaParser grammar. It produces pure Java
 * source files by asserting that no ADL_*, TDD_* and NLD_* nodes
 * remain in the final AST. The only exception to this are the
 * NLD_Annotation nodes.
 *
 * <p>The basic algorithm is to walk through the AST using the "visitor"
 * pattern (see Design Patterns from Gamma et al.) and emit the code on
 * an output stream maintained by the visitor itself.
 *
 * <p>The unparsing starts by calling the <a href="#unparse">unparse()</a>
 * method passing the top of AST to unparse as parameter.
 */
public class JUnparserVisitor extends GenericUnparserVisitor
implements ADLJavaParserVisitor
{
  /** The version number of this unparser */
  static private String versionId = "$Revision: 1.17 $";

  /**
   * Default constructor: call superclass constructor.
   *
   * @param toolname the name of the tool using this unparser visitor
   */
  public JUnparserVisitor(String toolname) {
    super(toolname, versionId);
  }

  /**
   * Basic constructor specifying unparser parameters: call superclass
   * constructor.
   *
   * @param toolname the name of the tool using this unparser visitor
   * @param linelimit the desired maximum length of lines in output file
   * @param regindent the regular indentation level
   * @param subindent the sub-expression indentation level
   */
  public JUnparserVisitor(String toolname, int linelimit, int regindent,
		   int subindent) {
    super(toolname, versionId, linelimit, regindent, subindent);
  }

  /**
   * Starts unparsing an AST.
   *
   * If the passed AST root is not an CompilationUnit, an error is
   * reported and the AST is not unparsed. Otherwise the AST is
   * visited.
   *
   * @param ast the root node of the AST to unparse
   * @see #visit(ORG.opengroup.adl.parsers.adljava.ASTCompilationUnit, java.lang.Object)
   */
  public void unparse(SimpleNode ast) {
  
    if (assertNodeType(ast, ASTCompilationUnit.class)) {
      ast.jjtAccept(this, null);
    } else {
      incorrectNode(ast);
      System.err.println("\twas expecting CompilationUnit");
    }
  }

  /**
   * Visits a CompilationUnit.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the name of the declared class or interface from the
   *     declaration node (always last one),
   * <li>open the source file for the class or interface (<code>outputDirectory</code>/<em>name</em>.java),
   * <li>reset current column and indentation,
   * <li>emit the standard header,
   * <li>visit the child nodes,
   * <li>and finally close the file.
   * </ul>
   *
   * @param node the CompilationUnit to visit
   * @param data unused
   * @return data returned by last child visited or
   * null if the file cannot be opened.
   * @see ORG.opengroup.adl.unparsers.GenericUnparserVisitor#outputDirectory
   *
   */
  public Object visit(ASTCompilationUnit node, Object data) {
    // get the name of the declared class or interface and open
    // the corresponding stream.

    // The declaration node is always the last child
    SimpleNode decl = null;
    if (node.jjtGetNumChildren() >= 1) {
      decl = (SimpleNode)node.jjtGetChild(node.jjtGetNumChildren() - 1);
    }

    if ((decl == null) ||
	!((decl instanceof ASTClassDeclaration) ||
	  (decl instanceof ASTInterfaceDeclaration))) {
      System.err.println(toolName + ": Can't get name of file: " +
			 "no class or interface declaration");
      return data;
    }

    // now we now the decl node is a class or interface declaration
    String className = decl.getName();

    // try to open the output stream
    try {
      // Try to get best performance through a BufferedWriter on top of
      // a FileWriter
      String fileName = outputDirectory + File.separator + className +
	".java";

      if (Adlt.verboseMode()) {
	System.out.println("+++ generating " + fileName);
      }

      FileWriter outStream = new FileWriter(fileName);
      out = new PrintWriter(new BufferedWriter(outStream));
    } catch (IOException exc) {
      System.err.println("Error: " + exc.getMessage());
      return null;
    }

    // now reset currentColumn and currentIndent
    currentColumn = 0;
    setIndent(0);

    // emit general comment
    emit(headerString);
    newline();
    newline();

    // and visit children
    data = node.childrenAccept(this, data);

    // close stream and reset handle
    out.close();
    out = null;

    return data;
  }

  /**
   * Visits an PackageDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the "package" keyword
   * <li>visit the only child node which gives the name of the package
   * </ul>
   * @param node the PackageDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTPackageDeclaration node, Object data) {
    // package line, emit keyword, then ask child to emit its name
    emit("package ");
    data = node.jjtGetChild(0).jjtAccept(this, data);

    emit(";");
    newline();

    return data;
  }

  /**
   * Visits an ImportDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the "import" keyword
   * <li>visit the only child node which gives the name of the imported
   *     class or package
   * <li>get the "onDemand" property which speicifies whether we are
   *     importing a class (if not set) or a package (if set)
   * <li>when we are importing a package, emit ".*"
   * </ul>
   * @param node the ImportDeclaration to visit
   * @param data unused
   * @return data returned by the visited child
   */
  public Object visit(ASTImportDeclaration node, Object data) {
    // import line, emit keyword, then ask child to emit its name
    emit("import ");
    data = node.jjtGetChild(0).jjtAccept(this, data);

    // if it's a package import, emit .*
    try {
      boolean onDemand = ((Boolean)node.getProperty("onDemand")).booleanValue();
      if (onDemand) {
	emit(".*");
      }
    } catch (PropertyAccessException e) {
      // ignore raised exception
    }

    emit(";");
    newline();

    return data;
  }

  /**
   * Visits a ClassDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the modifier (call <code>emitModifier</code>)
   * <li>emit the <code>class</code> keyword
   * <li>emit the name of the class accessible through <code>getName()</code>
   * <li>if the first optional children is a Name, emit <code>extends</code>,
   *     visit that child, and go to next child
   * <li>if the next optional children is a NameList, emit <code>implements</code>,
   *     visit that child and go to next child
   * <li>emit <code>{</code> to open class declaration
   * <li>for all remaining children (they should be ClassBodyDeclaration),
   *     visit the child
   * <li>emit <code>}</code> to close class declaration
   * </ul>
   *
   * @param node the ClassDeclaration to visit
   * @param data unused
   * @return unused data
   * 
   * @see #emitModifier
   */
  public Object visit(ASTClassDeclaration node, Object data) {
    // add more vertical space
    newline();

    // emit modifier
    emitModifier(node);

    // emit class name
    emit("class " + node.getName());
    
    int numChildren = node.jjtGetNumChildren();
    int childNum = 0; // first optional child index
    Node child;

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTName) {
	// first child is a name, i.e., of extended class
	emit(" extends ");
	child.jjtAccept(this, data);
	
	// go to next child
	childNum++;
      } 
    }

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTNameList) {
	// next child is a name list, i.e., of implemented interfaces
	newline();
	addIndent(regIndent);
	indent();
	delIndent(regIndent);
	emit("implements ");
	child.jjtAccept(this, data);
	
	// go to next child
	childNum++;
      }
    }

    // now emit { and go to next line
    emit(" {");
    newline();
    addIndent(regIndent);

    // visit class body declarations
    while (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      child.jjtAccept(this, data);
      childNum++;
    }

    // revert indentation
    delIndent(regIndent);
    emit("}");
    newline();

    return data;
  }

  /**
   * Visits an InterfaceDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the modifier (call <code>emitModifier</code>)
   * <li>emit the <code>interface</code> keyword
   * <li>if first optional child is a NameList, that is of extended
   *     interfaces, emit <code>extends</code>, visit this child
   *     then go to next child
   * <li>emit <code>{</code> to open interface declaration body
   * <li>visit remaining children, that is interface body declarations
   * <li>emit <code>}</code> to close interface declaration body
   * </ul>
   *
   * @param node the InterfaceDeclaration to visit
   * @param data unused
   * @return unused data
   *
   * @see #emitModifier
   */
  public Object visit(ASTInterfaceDeclaration node, Object data) {
    // add more vertical space
    newline();

    // emit modifier
    emitModifier(node);

    // emit public interface (all generated interfaces are public
    emit("interface " + node.getName());
    
    int numChildren = node.jjtGetNumChildren();
    int childNum = 0; // first optional child index
    Node child;

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTNameList) {
	// next child is a name list, i.e., of extended interfaces
	emit(" extends ");
	child.jjtAccept(this, data);
	
	// go to next child
	childNum++;
      }
    }

    // now emit { and go to next line
    emit(" {");
    newline();
    addIndent(regIndent);

    // visit interface body declarations
    while (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      child.jjtAccept(this, data);
      childNum++;
    }

    // revert indentation
    delIndent(regIndent);
    emit("}");
    newline();

    return data;
  }

  /**
   * Visits a ClassBodyDeclaration.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTClassBodyDeclaration node, Object data) {
    // nothing specific to do, visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an InterfaceMemberDeclaration.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTInterfaceMemberDeclaration node, Object data) {
    // nothing specific to do, visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a StaticInitializer.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the <code>static</code> keyword
   * <li>and visit the Block child
   * </ul>
   *
   * @param node the visited StaticInitializer
   * @param data unused
   * @return the data returned by the visited Block child
   */
  public Object visit(ASTStaticInitializer node, Object data) {
    indent();
    emit("static ");
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits a ConstructorDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the modifier (call <code>emitModifier</code>)
   * <li>emit the constructor name (accessible through <code>getName()</code>
   * <li>visit the FormalParameters child (first one)
   * <li>if next child is a NameList, emit <code>throws</code> and visit
   *     this child
   * <li>emit <code>{</code> to open constructor body
   * <li>visit remaining children (constructor body)
   * <li>emit <code>}</code> to close constructor body
   * </ul>
   *
   * @param node the visited ConstructorDeclaration
   * @param data unused
   * @return the unused data
   *
   * @see #emitModifier
   */
  public Object visit(ASTConstructorDeclaration node, Object data) {
    indent();

    // emit modifier
    emitModifier(node);

    emit(node.getName());

    int numChildren = node.jjtGetNumChildren();
    int childNum = 1; // first optional child index
    Node child;

    // visit FormalParameters node (first child)
    node.jjtGetChild(0).jjtAccept(this, data);

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTNameList) {
	// next child is a name list, i.e., of thrown exceptions
	emit(" throws ");
	child.jjtAccept(this, data);
	
	// go to next child
	childNum++;
      }
    }

    // now emit { and go to next line
    emit(" {");
    newline();
    addIndent(regIndent);

    // visit constructor body
    while (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      child.jjtAccept(this, data);
      childNum++;
    }

    // revert indentation
    delIndent(regIndent);
    indent();
    emit("}");
    newline();
    newline();

    return data;
  }

  /**
   * Visits a MethodDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the modifier (call <code>emitModifier</code>)
   * <li>emit the returned type by visiting the first child
   * <li>visit the MethodDeclarator child (second child)
   * <li>if next child is a NameList, emit <code>throws</code> and visit
   *     this child
   * <li>if there is a remaining child, it is a Block defining the method
   *     body, visit it
   * <li>else the method is empty, emit <code>;</code>
   * </ul>
   *
   * @param node the visited MethodDeclaration
   * @param data unused
   * @return the unused data
   *
   * @see #emitModifier
   */
  public Object visit(ASTMethodDeclaration node, Object data) {
    indent();

    // emit modifier
    emitModifier(node);

    int numChildren = node.jjtGetNumChildren();
    int childNum = 2; // first optional child index
    Node child;

    // visit ResultType node (first child)
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(" "); // separator between type and name of method

    // visit MethodDeclarator node (second child)
    node.jjtGetChild(1).jjtAccept(this, data);

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTNameList) {
	// next child is a name list, i.e., of thrown exceptions
	emit(" throws ");
	child.jjtAccept(this, data);
	
	// go to next child
	childNum++;
      }
    }

    if (childNum < numChildren) {
      // now if there is still another child, it is a Block
      emit(" ");
      child = node.jjtGetChild(childNum);
      child.jjtAccept(this, data);
    } else {
      // there is no other child, so the method is empty
      emit(";");
      newline();
    }
    newline();

    return data;
  }

  /**
   * Visits a FieldDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the modifier (call <code>emitModifier</code>)
   * <li>emit the type of the field by visiting the first child
   * <li>visit the remaining VariableDeclarator children (call
   *     <code>unparseList</code>)
   * <li>finally emit <code>;</code> to end the field declaration
   * </ul>
   *
   * @param node the visited FieldDeclaration
   * @param data unused
   * @return the unused data
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   * @see #emitModifier
   */
  public Object visit(ASTFieldDeclaration node, Object data) {
    indent();

    // emit modifier
    emitModifier(node);

    // now visit Type child always first child
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(" "); // separator between type and name of field

    // starting from second child for first variable, unparse list
    // of children
    data = unparseList(node, 1, ",", true, data);

    emit(";");
    newline();
    newline();

    return data;
  }

  /**
   * Visits a Statement, actually a StatementExpression statement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit the StatementExpresion child (first child)
   * <li>finally emit <code>;</code> to end the statement
   * </ul>
   * <p>Note that all other statements have their own AST node type.
   * 
   * @param node the visited Statement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTStatement node, Object data) {
    // this is actually a StatementExpression statement, so
    // there is always only one child
    Node child = node.jjtGetChild(0);

    // always indent a statement
    indent();

    // now visit child
    data = child.jjtAccept(this, data);

    emit(";");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a VariableDeclarator.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child to emit variable name
   * <li>then if there is a second child, emit <code>=</code> and visit it
   * </ul>
   *
   * @param node the visited VariableDeclarator
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTVariableDeclarator node, Object data) {
    // visit VariableDeclaratorId child
    node.jjtGetChild(0).jjtAccept(this, data);

    if (node.jjtGetNumChildren() > 1) {
      emit(" = ");
      // visit VariableInitializer now
      node.jjtGetChild(1).jjtAccept(this, data);
    }

    return data;
  }

  /**
   * Visits a Block.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>{</code> to open block
   * <li>add <code>regIndent</code> to current indentation
   * <li>visit all children
   * <li>remove <code>regIndent</code> to current indentation
   * <li>emit <code>}</code> to close block
   * </ul>
   *
   * @param node the visited Block
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTBlock node, Object data) {
    emit("{");
    addIndent(regIndent);
    newline();

    // visit BlockStatement children
    data = node.childrenAccept(this, data);

    delIndent(regIndent);
    indent();
    emit("}");

    // after a block, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a ExplicitConstructorInvocation.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>whether it's <code>this</code> or <code>super</code> is
   *     accessible from <code>getName()</code>, emit that
   * <li>visit all children (Arguments)
   * <li>emit <code>;</code> to end statement
   * </ul>
   *
   * @param node the visited ExplicitConstructorInvocation
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTExplicitConstructorInvocation node, Object data) {
    indent();

    // whether it's "this" or "super" is accessible via getName()
    emit(node.getName());

    // visit the Arguments child
    data = node.childrenAccept(this, data);

    // and end line
    emit(";");
    newline();

    return data;
  }

  /**
   * Visits a BlockStatement, actually a LocalVariableDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit only child which is a LocalVariableDeclaration
   * <li>emit <code>;</code> to end statement
   * </ul>
   *
   * @param node the visited BlockStatement
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTBlockStatement node, Object data) {
    // there is always only one child, which is a LocalVariableDeclaration
    Node child = node.jjtGetChild(0);

    indent();

    // visit the LocalVariableDeclaration
    data = child.jjtAccept(this, data);

    // and end line
    emit(";");
    newline();

    return data;
  }

  /**
   * Visits a LabeledStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get label name which is accessible via <code>getName()</code>
   * <li>remove label length + 2 (for colon and space following) from
   *     current indent so that the statement following the label is
   *     indented like any other previous or following statements
   * <li>indent and readjust current indentation back to previous value
   * <li>emit the label name followed by a colon and a space
   * <li>and visit only child
   * </ul>
   *
   * @param node the visited LabeledStatement
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTLabeledStatement node, Object data) {
    // label is accessible via getName()
    // always indent a statement
    String labelString = node.getName();
    int labelLength = labelString.length();

    // adjust indent level so that statement following the label is
    // indented like any other statement
    delIndent(labelLength + 2);
    indent();
    addIndent(labelLength + 2);

    emit(node.getName() + ": ");
    data = node.jjtGetChild(0).jjtAccept(this, data);

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits an EmptyStatement: just indent and emit <code>;</code>
   *
   * @param node the visited EmptyStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTEmptyStatement node, Object data) {
    // just emit the empty statement
    indent();
    emit(";");
    newline();

    return data;
  }

  /**
   * Visits a SwitchStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>switch (</code>
   * <li>visit first child which is the switch Expression
   * <li>emit <code>) {</code> and regular indent for following block
   * <li>for all remaining children:<ul>
   * <li>if child is a SwitchLabel:<ul>
   * <li>if it is not the first label, remove extra indent
   * <li>indent and visit label child
   * <li>go to next line and add extra indent for following statements
   * </ul><li>else:<ul>
   * <li>if child is a Block, indent (other statements indent themselves)
   * <li>visit statement child
   * </ul>
   * </ul>
   * <li>remove regular indent, indent and emit <code>}</code> to close
   *     block
   * <li>finally go to next line
   * </ul>
   *
   * @param node the visited SwitchStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTSwitchStatement node, Object data) {
    // always indent a statement
    indent();

    // start by emitting the "switch" keyword
    emit("switch (");
    
    // then visit first child
    node.jjtGetChild(0).jjtAccept(this, data);

    // end switch expression
    emit(") {");
    newline();
    addIndent(regIndent);
    
    // visit remaining children
    int numChildren = node.jjtGetNumChildren();
    boolean firstLabel = true;
    for (int childNum = 1; childNum < numChildren; childNum++) {
      Node child = node.jjtGetChild(childNum);
      if (child instanceof ASTSwitchLabel) {
	// after a label we add an extra indent, this requires some
	// extra processing
	if (!firstLabel)
	  // remove the extra indent before visiting the label
	  delIndent(subIndent);
	else
	  firstLabel = false;

	// indent for switch label
	indent();

	// visit the label
	child.jjtAccept(this, data);

	newline();            // BlockStatement always start on next line
	addIndent(subIndent); // with an extra indent

      } else {
	// indent for Block only, other statements indent themselves
	if (child instanceof ASTBlock) indent();

	// visit BlockStatement
	child.jjtAccept(this, data);
      }
    }

    delIndent(regIndent);
    indent();
    emit("}");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a SwitchLabel.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if there is no child, emit <code>default</code>
   * <li>else:<ul>
   * <li>emit <code>case </code>
   * <li>visit child node to emit label name
   * </ul>
   * <li>finally emit <code>:</code>
   * </ul>
   *
   * @param node the visited SwitchLabel
   * @param data unused
   * @return the data returned by visited child or the unused data
   */
  public Object visit(ASTSwitchLabel node, Object data) {
    if (node.jjtGetNumChildren() > 0) {
      emit("case ");
      data = node.jjtGetChild(0).jjtAccept(this, data);
    } else {
      emit("default");
    }
    emit(":");

    return data;
  }

  /**
   * Visits an IfStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>if (</code>
   * <li>than visit first child for condition expression
   * <li>emit <code>)</code>
   * <li>visit second child (call <code>visitStatementChild</code>)
   * <li>if there is one more child, visit it (call
   *     <code>visitStatementchild</code>)
   * </ul>
   *
   * @param node the visited IfStatement
   * @param data unused
   * @return the unused data
   * @see #visitStatementChild
   */
  public Object visit(ASTIfStatement node, Object data) {
    // always indent a statement
    indent();

    emit("if (");

    // first child is condition expression
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(")");

    // fisrt if statement
    data = visitStatementChild(node.jjtGetChild(1), data);

    // now else statement if it exists
    if (node.jjtGetNumChildren() > 2) {
      indent();
      emit("else");
      data = visitStatementChild(node.jjtGetChild(2), data);
    }

    return data;
  }

  /**
   * Visits a WhileStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>while (</code>
   * <li>than visit first child for condition expression
   * <li>emit <code>)</code>
   * <li>visit second child (call <code>visitStatementChild</code>)
   * </ul>
   *
   * @param node the visited WhileStatement
   * @param data unused
   * @return the data returned by second child visited
   * @see #visitStatementChild
   */
  public Object visit(ASTWhileStatement node, Object data) {
    // always indent a statement
    indent();

    emit("while (");

    // first child is condition expression
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(")");

    // while statement
    return visitStatementChild(node.jjtGetChild(1), data);
  }

  /**
   * Visits a DoStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>do</code>
   * <li>visit first child (call <code>visitStatementChild</code>)
   * <li>indent and emit <code>while (</code>
   * <li>than visit second child for condition expression
   * <li>emit <code>);</code>
   * <li>go to next line
   * </ul>
   *
   * @param node the visited DoStatement
   * @param data unused
   * @return the data returned by second child visited
   * @see #visitStatementChild
   */
  public Object visit(ASTDoStatement node, Object data) {
    // always indent a statement
    indent();

    emit("do");

    // do statement
    visitStatementChild(node.jjtGetChild(0), data);

    indent();
    emit("while (");

    // first child is condition expression
    data = node.jjtGetChild(1).jjtAccept(this, data);

    emit(");");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a ForStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>for (</code>
   * <li>if first child is a ForInit, visit it and go to next child
   * <li>emit <code>;</code>
   * <li>if next child is an Expression:<ul>
   * <li>emit a space
   * <li>visit child
   * <li>go to next child
   * </ul>
   * <li>emit <code>;</code>
   * <li>if next child is a ForUpdate:<ul>
   * <li>emit a space
   * <li>visit child
   * <li>go to next child
   * </ul>
   * <li>emit <code>)</code>
   * <li>visit last child (call <code>visitStatementChild</code>)
   * </ul>
   *
   * @param node the visited ForStatement
   * @param data unused
   * @return the data returned by last child visited
   * @see #visitStatementChild
   */
  public Object visit(ASTForStatement node, Object data) {
    // always indent a statement
    indent();

    emit("for (");

    int numChildren = node.jjtGetNumChildren();
    int childNum = 0; // first optional child index
    Node child;

    child = node.jjtGetChild(childNum);
    if (child instanceof ASTForInit) {
      // there is a for init, visit the child and go to next one
      child.jjtAccept(this, data);
      childNum++;
    }
    emit(";");

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTExpression) {
	// there is a for test, visit the child and go to next one
	emit(" ");
	child.jjtAccept(this, data);
	childNum++;
      }
    }
    emit(";");

    if (childNum < numChildren) {
      child = node.jjtGetChild(childNum);
      if (child instanceof ASTForUpdate) {
	// there is a for update, visit the child and go to next one
	emit(" ");
	child.jjtAccept(this, data);
	childNum++;
      }
    }
    emit(")");

    // now we know there is this child because it's not optional!
    return visitStatementChild(node.jjtGetChild(childNum), data);
  }

  /**
   * Visits a ForInit.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTForInit node, Object data) {
    // nothing specific to do, visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a ForUpdate.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTForUpdate node, Object data) {
    // nothing specific to do, visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a SynchronizedStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>synchronized (</code>
   * <li>visit first child, that is Expression
   * <li>emit <code>) </code>
   * <li>visit second child, that is Block
   * </ul>
   *
   * @param node the visited SynchronizedStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTSynchronizedStatement node, Object data) {
    // always indent a statement
    indent();

    emit("synchronized (");

    // visit first child: Expression
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(") ");

    // visit second child: Block
    node.jjtGetChild(1).jjtAccept(this, data);

    return data;
  }

  /**
   * Visits a TryStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>try</code>
   * <li>visit first child which is tried Block
   * <li>for all remaining children:<ul>
   * <li>indent, then, 
   * <li>if we have a couple of FormalParameter, Block:<ul>
   * <li>emit <code>catch (</code>
   * <li>visit FormalParameter child
   * <li>emit <code>)</code>
   * <li>visit Block child
   * </ul>
   * <li>else (only a Block):<ul>
   * <li>emit <code>finally</code>
   * <li>visit child
   * </ul>
   * </ul>
   * </ul>
   *
   * @param node the visited TryStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTTryStatement node, Object data) {
    // always indent a statement
    indent();

    emit("try ");

    // visit first child: Block
    node.jjtGetChild(0).jjtAccept(this, data);

    // visit remaining children
    int numChildren = node.jjtGetNumChildren();
    for (int childNum = 1; childNum < numChildren; childNum++) {
      Node child = node.jjtGetChild(childNum);
      indent();
      if (child instanceof ASTFormalParameter) {
	// that's a catch clause, next child is associated block
	emit("catch (");
	child.jjtAccept(this, data);
	emit(") ");
	node.jjtGetChild(++childNum).jjtAccept(this, data);
      } else if (child instanceof ASTBlock) {
	// that's the finally clause
	emit("finally ");
	child.jjtAccept(this, data);
      }
    }

    return data;
  }

  /**
   * Visits a BreakStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>try to get id of breaked statement, via <code>getName()</code>
   * <li>emit <code>break</code>
   * <li>if id is not null, emit it
   * <li>emit <code>;</code>
   * <li>go to next line
   * </ul>
   *
   * @param node the visited BreakStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBreakStatement node, Object data) {
    // always indent a statement
    indent();

    // id of statement to break out of is accessible via getName()
    String id = node.getName();

    emit("break");
    if (id != null) emit(" " + id);
    emit(";");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a ContinueStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>try to get id of continued statement, via <code>getName()</code>
   * <li>emit <code>continue</code>
   * <li>if id is not null, emit it
   * <li>emit <code>;</code>
   * <li>go to next line
   * </ul>
   *
   * @param node the visited ContinueStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTContinueStatement node, Object data) {
    // always indent a statement
    indent();

    // id of statement to continue from is accessible via getName()
    String id = node.getName();

    emit("continue");
    if (id != null) emit(" " + id);
    emit(";");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a ReturnStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>return</code>
   * <li>if there is a child, visit it
   * <li>emit <code>;</code>
   * <li>go to next line
   * </ul>
   *
   * @param node the visited ReturnStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTReturnStatement node, Object data) {
    // always indent a statement
    indent();

    emit("return");

    // visit child if there is one
    if (node.jjtGetNumChildren() > 0) {
      emit(" ");
      node.jjtGetChild(0).jjtAccept(this, data);
    }

    emit(";");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits a ThrowStatement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>as for any statement, indent
   * <li>emit <code>throw</code>
   * <li>visit first and only child
   * <li>emit <code>;</code>
   * <li>go to next line
   * </ul>
   *
   * @param node the visited ThrowStatement
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTThrowStatement node, Object data) {
    // always indent a statement
    indent();

    emit("throw ");

    // visit child
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(";");

    // after a statement, always go to next line
    newline();

    return data;
  }

  /**
   * Visits an Expression.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTExpression node, Object data) {
    // nothing specific to do, just visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a VariableInitializer.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTVariableInitializer node, Object data) {
    // nothing specific to do, just visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a LocalVariableDeclaration.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child for type
   * <li>emit a space
   * <li>unparse list of variable declarator (call <code>unparseList</code>)
   * </ul>
   *
   * @param node the visited LocalVariableDeclaration
   * @param data unused
   * @return the unused data
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTLocalVariableDeclaration node, Object data) {
    // now visit Type child always first child
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(" "); // separator between type and name of field

    // starting from second child for first variable, unparse list
    // of children
    return unparseList(node, 1, ",", true, data);
  }

  /**
   * Visits an Assignment.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child
   * <li>get operator from property <code>op</code> (default to =)
   * <li>emit operator
   * <li>visit second child
   * </ul>
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTAssignment node, Object data) {
    // nothing specific to do, just visit children
    node.jjtGetChild(0).jjtAccept(this, data);
    try { emit(" " + (String)node.getProperty("op") + " "); }
    catch (PropertyAccessException e) { emit(" <MissingOperator> "); }

    return node.jjtGetChild(1).jjtAccept(this, data);
  }

  /**
   * Visits a PostfixExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first and only child
   * <li>if not empty, emit the operator, accessible via the
   *     <code>op</code> property
   * </ul>
   *
   * @param node the visited PostfixExpression
   * @param data unused
   * @return the data returned by the visited child
   */
  public Object visit(ASTPostfixExpression node, Object data) {
    // there is a single child and postfix operator is accessible via
    // the op property
    String operator;
    try { operator = (String)node.getProperty("op"); }
    catch (PropertyAccessException e) { operator = "<MissingOperator>"; }
    data = node.jjtGetChild(0).jjtAccept(this, data);
    if (operator != null && !operator.equals(""))
      emit(operator);

    return data;
  }

  /**
   * Visits a StatementExpressionList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of statement expression (call <code>unparseList</code>)
   * </ul>
   *
   * @param node the visited StatementExpressionList
   * @param data unused
   * @return the data returned by the last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTStatementExpressionList node, Object data) {
    // just unparse list
    return unparseList(node, 0, ",", true, data);
  }

  /**
   * Visits a ConditionalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>check if parent node is a ParenthExpression
   * <li>if not, emit <code>(</code>
   * <li>visit first child (condition)
   * <li>emit <code>?</code>
   * <li>visit second child
   * <li>emit <code>:</code>
   * <li>visit third child
   * <li>if required emit <code>)</code>
   * </ul>
   *
   * Note: this node is only created when there are 3 children
   * 
   * @param node the visited ConditionalExpression
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTConditionalExpression node, Object data) {
    // there are always 3 children
    
    // determine type of father to see if we have to put
    // the expression between parenthesis
    boolean childOfParenthExpression = false;
    if (node.jjtGetParent() instanceof ASTParenthExpression)
      childOfParenthExpression = true;

    if (!childOfParenthExpression) emit("( ");
    node.jjtGetChild(0).jjtAccept(this, data);
    emit(" ? ");
    node.jjtGetChild(1).jjtAccept(this, data);
    emit(" : ");
    node.jjtGetChild(2).jjtAccept(this, data);
    if (!childOfParenthExpression) emit(" )");

    return data;
  }

  /**
   * Visits a ConditionalOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>||</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ConditionalOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTConditionalOrExpression node, Object data) {
    // just unparse children separated by "||"
    return unparseList(node, 0, "||", false, data);
  }

  /**
   * Visits a ConditionalAndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>&&</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ConditionalAndExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTConditionalAndExpression node, Object data) {
    // just unparse children separated by "&&"
    return unparseList(node, 0, "&&", false, data);
  }

  /**
   * Visits an InclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>|</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited InclusiveOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTInclusiveOrExpression node, Object data) {
    // just unparse children separated by "|"
    return unparseList(node, 0, "|", false, data);
  }

  /**
   * Visits an ExclusiveOrExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>^</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ExclusiveOrExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTExclusiveOrExpression node, Object data) {
    // just unparse children separated by "^"
    return unparseList(node, 0, "^", false, data);
  }

  /**
   * Visits an AndExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>&</code>
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited AndExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTAndExpression node, Object data) {
    // just unparse children separated by "&"
    return unparseList(node, 0, "&", false, data);
  }

  /**
   * Visits an EqualityExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited EqualityExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], boolean, java.lang.Object)
   */
  public Object visit(ASTEqualityExpression node, Object data) {
    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, false, data);
  }

  /**
   * Visits an InstanceOfExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child
   * <li>emit <code>instanceof</code>
   * <li>visit second child
   * </ul>
   *
   * Note: this node is only created when there are 2 children
   *
   * @param node the visited InstanceOfExpression
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTInstanceOfExpression node, Object data) {
    // there are always 2 children
    
    // always put the expression between parenthesis
    node.jjtGetChild(0).jjtAccept(this, data);
    emit(" instanceof ");
    return node.jjtGetChild(1).jjtAccept(this, data);
  }

  /**
   * Visits a RelationalExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited RelationalExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], boolean, java.lang.Object)
   */
  public Object visit(ASTRelationalExpression node, Object data) {
    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, false, data);
  }

  /**
   * Visits a ShiftExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited ShiftExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], boolean, java.lang.Object)
   */
  public Object visit(ASTShiftExpression node, Object data) {
    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, false, data);
  }

  /**
   * Visits an AdditiveExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited AdditiveExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], boolean, java.lang.Object)
   */
  public Object visit(ASTAdditiveExpression node, Object data) {
    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, false, data);
  }

  /**
   * Visits a MultiplicativeExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get list of operators, accessible in the <code>operators</code>
   *     property
   * <li>unparse list of children separated by their own operator
   * </ul>
   *
   * Note: this node is only created when there are 2+ children
   *
   * @param node the visited MultiplicativeExpression
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String[], boolean, java.lang.Object)
   */
  public Object visit(ASTMultiplicativeExpression node, Object data) {
    // get array of operators
    String operators[] = null;
    try {
      operators = (String[])node.getProperty("operators");
    } catch (PropertyAccessException e) {}

    // unparse list of expression children
    return unparseList(node, 0, operators, false, data);
  }

  /**
   * Visits an UnaryPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator, accessible via the <code>op</code> property
   * <li>visit first and only child
   * </ul>
   *
   * @param node the visited UnaryPlusMinus
   * @param data unused
   * @return the data by visited child
   */
  public Object visit(ASTUnaryPlusMinus node, Object data) {
    try { emit((String)node.getProperty("op")); }
    catch (PropertyAccessException e) { emit("<MissingOperator>"); }

    // just one child
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits a PreIncrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>++</code>
   * <li>first first and only child
   * </ul>
   *
   * @param node the visited PreIncrementExpression
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTPreIncrementExpression node, Object data) {
    // just one child
    emit("++");
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits a PreDecrementExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>--</code>
   * <li>first first and only child
   * </ul>
   *
   * @param node the visited PreDecrementExpression
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTPreDecrementExpression node, Object data) {
    // just one child
    emit("--");
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits an UnaryNotPlusMinus.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit operator, accessible via the <code>op</code> property
   * <li>visit first and only child
   * </ul>
   *
   * @param node the visited UnaryNotPlusMinus
   * @param data unused
   * @return the data by visited child
   */
  public Object visit(ASTUnaryNotPlusMinus node, Object data) {
    try { emit((String)node.getProperty("op")); }
    catch (PropertyAccessException e) { emit("<MissingOperator>"); }

    // just one child
    return node.jjtGetChild(0).jjtAccept(this, data);
  }

  /**
   * Visits a CastExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>if there is more than one child:<ul>
   * <li>visit first child for casted type (not a primitive type)
   * </ul><li>else:<ul>
   * <li>get casted type from the <code>type</code> property and emit it
   * </ul>
   * <li>get <code>arrayLevel</code> property for array type, while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * <li>emit <code>)</code>
   * <li>visit second child
   * </ul>
   *
   * @param node the visited CastExpression
   * @param data unused
   * @return the data returned by visiting second child
   */
  public Object visit(ASTCastExpression node, Object data) {
    emit("(");

    // if casted type is a primitive type, there is no type children
    if (node.jjtGetNumChildren() > 1) {
      // this is a Name child
      node.jjtGetChild(0).jjtAccept(this, data);
    } else {
      // cast to a primitive type
      try {
	LType castedType = (LType)node.getProperty("type");
        String castedTypeName = castedType.getName();
	emit(castedTypeName);
      } catch (PropertyAccessException e) {
	emit("<unknown type>");
      }
    }

    // emit potential array indicators
    try {
      int arrayLevel =
	((Integer)node.getProperty("arrayLevel")).intValue();
      while (arrayLevel-- > 0) {
	emit("[]");
      }
    } catch (PropertyAccessException e) {}

    emit(")");
    data = node.jjtGetChild(node.jjtGetNumChildren()-1).jjtAccept(this, data);

    return data;
  }

  /**
   * Visits a PrimaryExpression.
   *
   * <p><b>Algorithm:</b> nothing specific to do here,
   * just visit children.
   *
   * @return data returned by last visited children
   */
  public Object visit(ASTPrimaryExpression node, Object data) {
    // nothing specific to do, just visit children
    return node.childrenAccept(this, data);
  }

  /**
   * Visits a ParenthExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>visit first and only child
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited ParenthExpression
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTParenthExpression node, Object data) {
    emit("(");
    data = node.jjtGetChild(0).jjtAccept(this, data);
    emit(")");
    return data;
  }

  /**
   * Visits a BracketSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>[</code>
   * <li>visit first and only child
   * <li>emit <code>]</code>
   * </ul>
   *
   * @param node the visited BracketSuffix
   * @param data unused
   * @return the data returned by visited child
   */
  public Object visit(ASTBracketSuffix node, Object data) {
    // there is always one single child
    emit("[");
    data = node.jjtGetChild(0).jjtAccept(this, data);
    emit("]");

    return data;
  }

  /**
   * Visits a DotSuffix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>.</code> followed by identifier, accessible via
   *     <code>getName()</code>
   * </ul>
   *
   * @param node the visited 
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTDotSuffix node, Object data) {
    // following identifier is accessible via getName()
    emit("." + node.getName());

    return data;
  }

  /**
   * Visits a NullLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>null</code>
   * </ul>
   *
   * @param node the visited NullLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTNullLiteral node, Object data) {
    // value of a null literal is "null"
    emit("null");
    return data;
  }

  /**
   * Visits an IntegerLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited IntegerLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTIntegerLiteral node, Object data) {
    // value of a literal is accessible via the value property
    try { emit((String)node.getProperty("value")); }
    catch (PropertyAccessException e) { emit("<MissingValue>"); }
    return data;
  }

  /**
   * Visits an FloatLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited FloatLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTFloatLiteral node, Object data) {
    // value of a literal is accessible via the value property
    try { emit((String)node.getProperty("value")); }
    catch (PropertyAccessException e) { emit("<MissingValue>"); }
    return data;
  }

  /**
   * Visits an CharacterLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited CharacterLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTCharacterLiteral node, Object data) {
    // value of a literal is accessible via the value property
    try { emit((String)node.getProperty("value")); }
    catch (PropertyAccessException e) { emit("<MissingValue>"); }
    return data;
  }

  /**
   * Visits an StringLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited StringLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTStringLiteral node, Object data) {
    // value of a literal is accessible via the value property
    try { emit((String)node.getProperty("value")); }
    catch (PropertyAccessException e) { emit("<MissingValue>"); }
    return data;
  }

  /**
   * Visits an BooleanLiteral.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit literal value, accessible via <code>getValue()</code>
   * </ul>
   *
   * @param node the visited BooleanLiteral
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTBooleanLiteral node, Object data) {
    // value of a literal is accessible via the value property
    try { emit((String)node.getProperty("value")); }
    catch (PropertyAccessException e) { emit("<MissingValue>"); }
    return data;
  }

  /**
   * Visits an Arguments.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>if there is one child, visit it
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited Arguments
   * @param data unused
   * @return the unused data or the data returned by visited child
   */
  public Object visit(ASTArguments node, Object data) {
    emit("(");
    if (node.jjtGetNumChildren() > 0) 
      data = node.jjtGetChild(0).jjtAccept(this, data);
    emit(")");

    return data;
  }

  /**
   * Visits an ArgumentList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>,</code>
   * </ul>
   *
   * @param node the visited ArgumentList
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTArgumentList node, Object data) {
    // unparse list of arguments
    return unparseList(node, 0, ",", true, data);
  }

  /**
   * Visits an AllocationExpression.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>new</code>
   * <li>visit children
   * </ul>
   *
   * @param node the visited AllocationExpression
   * @param data unused
   * @return the data returned by last visited child
   */
  public Object visit(ASTAllocationExpression node, Object data) {
    emit("new ");
    return node.childrenAccept(this, data);
  }

  /**
   * Visits an ArrayDimensions.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>for each child (dimension with an expressed value),<ul>
   * <li>emit <code>[</code>
   * <li>visit the child
   * <li>emit <code>]</code></ul>
   * <li>get <code>arrayLevel</code> property for array type, substract
   * the number of children, and while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited ArrayDimensions
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTArrayDimensions node, Object data) {
    // visit children: values of expressed dimension size
    int nbChildren = node.jjtGetNumChildren();
    for (int i = 0; i < nbChildren; i++) {
      emit("[");
      node.jjtGetChild(i).jjtAccept(this, data);
      emit("]");
    }
    // emit potential array indicators
    try {
      int arrayLevel =
	((Integer)node.getProperty("arrayLevel")).intValue() - nbChildren;
      while (arrayLevel-- > 0) {
	emit("[]");
      }
    } catch (PropertyAccessException e) {}
    return data;
  }

  /**
   * Visits a VariableDeclaratorId.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the variable name, accessible via <code>getName()</code>
   * <li>get <code>arrayLevel</code> property for array type, while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited VariableDeclaratorId
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTVariableDeclaratorId node, Object data) {
    // jut emit name of variable
    emit(node.getName());

    // emit potential array indicators
    try {
      int arrayLevel =
	((Integer)node.getProperty("arrayLevel")).intValue();
      while (arrayLevel-- > 0) {
	emit("[]");
      }
    } catch (PropertyAccessException e) {}

    return data;
  }

  /**
   * Visits a MethodDeclarator node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the name of the method, accessible via <code>getName()</code>
   * <li>visit the first child for the formal parameters
   * <li>get <code>arrayLevel</code> property for array type, while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited MethodDeclarator node
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTMethodDeclarator node, Object data) {
    // the name of the method is accessible via getName()
    emit(node.getName());

    // now visit the FormalParameters node
    node.jjtGetChild(0).jjtAccept(this, data);

    // emit potential array indicators
    try {
      int arrayLevel =
	((Integer)node.getProperty("arrayLevel")).intValue();
      while (arrayLevel-- > 0) {
	emit("[]");
      }
    } catch (PropertyAccessException e) {}

    return data;
  }

  /**
   * Visits a FormalParameters.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>(</code>
   * <li>if there is at least one child, unparse list of children
   *     separated by <code>,</code>
   * <li>emit <code>)</code>
   * </ul>
   *
   * @param node the visited FormalParameters
   * @param data unused
   * @return the data returned by last visited child or unused data
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTFormalParameters node, Object data) {
    emit("(");

    if (node.jjtGetNumChildren() > 0) {
      data = unparseList(node, 0, ",", true, data);
    }

    emit(")");

    return data;
  }

  /**
   * Visits a FormalParameter.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>visit first child for the type of the parameter
   * <li>emit a space
   * <li>visit second child for the VariableDeclaratorId
   * </ul>
   *
   * @param node the visited FormalParameter
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTFormalParameter node, Object data) {
    // first child is Type
    node.jjtGetChild(0).jjtAccept(this, data);

    emit(" "); // don't forget this separator

    // second child is VariableDeclaratorId
    node.jjtGetChild(1).jjtAccept(this, data);

    return data;
  }

  /**
   * Visits a Type.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit type name, accessible via <code>getName()</code>
   * <li>get <code>arrayLevel</code> property for array type, while
   *     it is greater than 0:<ul>
   * <li>emit <code>[]</code>
   * </ul>
   * </ul>
   *
   * @param node the visited Type
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTType node, Object data) {
    // there is no need to recurse down to children
    // whether the type is primitive or not, its name is
    // accessible via getName()
    emit(node.getName());

    // emit potential array indicators
    try {
      int arrayLevel =
	((Integer)node.getProperty("arrayLevel")).intValue();
      while (arrayLevel-- > 0) {
	emit("[]");
      }
    } catch (PropertyAccessException e) {}

    return data;
  }

  /**
   * Visits a ResultType.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if there is one child, visit it
   * <li>otherwise emit <code>void</code>
   * </ul>
   *
   * @param node the visited 
   * @param data unused
   * @return the data returned by visited child or unused data
   */
  public Object visit(ASTResultType node, Object data) {
    if (node.jjtGetNumChildren() > 0) {
      return node.jjtGetChild(0).jjtAccept(this, data);
    } else {
      // void type
      emit("void");
      return data;
    }
  }

  /**
   * Visits a Name.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit node name, accessible via <code>getName()</code>
   * </ul>
   *
   * @param node the visited Name
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTName node, Object data) {
    // A name is just a sequence of ids separated by "."
    // the full sequence is accessible via getName()
    emit(node.getName());
    return data;
  }

  /**
   * Visits a NameList.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>unparse list of children separated by <code>,</code>
   * </ul>
   *
   * Note: there is always at least one child
   *
   * @param node the visited NameList
   * @param data unused
   * @return the data returned by last visited child
   * @see #unparseList(ORG.opengroup.adl.parsers.adljava.Node, int, java.lang.String, boolean, java.lang.Object)
   */
  public Object visit(ASTNameList node, Object data) {
    // unparse list of names
    return unparseList(node, 0, ",", true, data);
  }

  /**
   * Visits a ThisPrefix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit <code>this</code>
   * </ul>
   *
   * @param node the visited ThisPrefix
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTThisPrefix node, Object data) {
    // although its name is accessible via getName(), it's also
    // always this!
    emit("this");

    return data;
  }

  /**
   * Visits a SuperPrefix.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>emit the full prefix, i.e., super.<id>, obtained via
   *     <code>getName()</code>
   * </ul>
   *
   * @param node the visited SuperPrefix
   * @param data unused
   * @return the unused data
   */
  public Object visit(ASTSuperPrefix node, Object data) {
    // the full prefix is accessible via getName()
    emit(node.getName());

    return data;
  }

  // All following AST nodes are incorrect since they should have
  // disappeared during AST transformation

  /** Should not be visited! */
  public Object visit(ASTADL_CompilationUnit node, Object data) {
    incorrectNode(node);
    return data;
  }

  /** Should not be visited! */
  public Object visit(SimpleNode node, Object data) {
    incorrectNode(node);
    return data;
  }

  /** Should not be visited! */
  public Object visit(ASTADL_Unchanged node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_SuperSemantics node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_UseDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_ClassDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_InlineDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Prologue node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Epilogue node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorSpecification node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_BehaviorClassification node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_NormalBehavior node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_AbNormalBehavior node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_AssertionGroup node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Binding node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Statement node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_IfStatement node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_TryStatement node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Quantifier node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Assertion node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Label node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Tags node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_QuantifiedAssertion node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_ImplExpression node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_ImplOp node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_CallStateExpression node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Normal node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Abnormal node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Return node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_ThrownExpression node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_DomainList node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTADL_Domain node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_ClassDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_ClassBodyDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_FieldDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_FactoryDeclaration node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_TestDirective node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetDomain node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetLiteral node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_SingleDeclarator node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetMember node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetExpr node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetConcatenationExpr node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_DatasetSingleton node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_FactoryCall node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_ADLnewExpression node, Object data) {
    incorrectNode(node);
    return data;
  }
  /** Should not be visited! */
  public Object visit(ASTTDD_ADLExpression node, Object data) {
    incorrectNode(node);
    return data;
  }

  /**
   * Visits a child which is a statement, that is either a Block or
   * any other statement.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if child is a Block:<ul>
   * <li>emit a space and visit the Block child
   * </ul><li>else:<ul>
   * <li>add sub-expression indent
   * <li>go to next line
   * <li>visit the statement child
   * <li>remove extra indent
   * </ul></ul>
   *
   * @param node the statement child to visit
   * @param data unused data to be passed to visited node
   * @return data returned by visited node
   */
  private Object visitStatementChild(Node node, Object data) {
    if (node instanceof ASTBlock) {
      // it's a block, proceed on same line
      emit(" ");
      data = node.jjtAccept(this, data);
    } else {
      // it's another statement, go to next line and indent
      addIndent(subIndent);
      newline();
      data = node.jjtAccept(this, data);
      delIndent(subIndent);
    }

    return data;
  }


  /**
   * Unparse a list of <em>something</em> with a single separator,
   * possibly wrapping on the next line if required.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get current indentation level
   * <li>set current indentation to current column so that wrapped
   *     statements are all aligned with the first one
   * <li>get the number of children from the parent node
   * <li>visit first child in the list (<code>firstChild</code>)
   * <li>for all remaining children of the parent node:<ul>
   * <li>if the separator should be emitted before wrapping, emit it
   * <li>if current column is already below the line limit:<ul>
   * <li>go to next line
   * <li>indent
   * </ul><li>else emit a space
   * <li>if the separator should be emitted after wrapping, emit it
   * <li>visit the child
   * </ul>
   * <li>reset current indentation level
   * </ul>
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node. But we trust 
   * our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separator the string that separates children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     String separator, boolean sepBefore,
			     Object data) {
    int currentLevel = getIndent();
    setIndent(currentColumn);
    
    // get number of children
    int numChildren = node.jjtGetNumChildren();

    // visit first child, then loop on remaining ones
    node.jjtGetChild(firstChild).jjtAccept(this, data);
    for (int childNum = firstChild + 1; childNum < numChildren;
	 childNum++) {

      // if requested, emit the separator before wrapping
      if (sepBefore) emit(separator);

      // determine if we have to wrap
      if (currentColumn >= lineLimit) {
	newline();
	indent();
      } else
	emit(" ");

      // if requested, emit the separator after wrapping
      if (!sepBefore) emit(separator + " ");

      // visit node
      node.jjtGetChild(childNum).jjtAccept(this, data);
    }
    
    // now reset indent level
    setIndent(currentLevel);

    return data;
  }

  /**
   * Unparse a list of <em>something</em> each having its own separator,
   * possibly wrapping on the next line if required.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get current indentation level
   * <li>set current indentation to current column so that wrapped
   *     statements are all aligned with the first one
   * <li>get the number of children from the parent node
   * <li>visit first child in the list (<code>firstChild</code>)
   * <li>for all remaining children of the parent node:<ul>
   * <li>if the separator should be emitted before wrapping, emit it
   * <li>if current column is already below the line limit:<ul>
   * <li>go to next line
   * <li>indent
   * </ul><li>else emit a space
   * <li>if the separator should be emitted after wrapping, emit it
   * <li>visit the child
   * </ul>
   * <li>reset current indentation level
   * </ul>
   *
   * Note: we should throw an exception when firstChild is greater than
   * the index of the last children of the parent node and when the size
   * of the array of separators does not match the number of children to
   * visit - 1. But we trust our own code!
   *
   * @param node the parent node of the list 
   * @param firstChild the index of the first child of the list
   * @param separators the array of strings that separate children
   * @param sepBefore whether to emit the separator before wrapping or not
   * @param data unused
   * @return the unused data
   */
  private Object unparseList(Node node, int firstChild, 
			     String[] separators, boolean sepBefore,
			     Object data) {
    // get current indent level
    int currentLevel = getIndent();
    setIndent(currentColumn + subIndent);

    // get number of children
    int numChildren = node.jjtGetNumChildren();

    // visit first child, then loop on remaining ones
    node.jjtGetChild(firstChild).jjtAccept(this, data);
    for (int childNum = firstChild + 1; childNum < numChildren;
	 childNum++) {

      // if requested, emit the separator before wrapping
      if (sepBefore) emit(separators[childNum - 1]);

      // determine if we have to wrap
      if (currentColumn >= lineLimit) {
	newline();
	indent();
      } else 
	emit(" ");

      // if requested, emit the separator after wrapping
      if (!sepBefore) emit(separators[childNum - 1] + " ");

      node.jjtGetChild(childNum).jjtAccept(this, data);
    }

    // restore current indentation
    setIndent(currentLevel);

    return data;
  }

  /**
   * Emit the modifier string associated to a given node.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>get the <code>modifier</code> property
   * <li>create a JModifier object out of it and call <code>toString()</code>
   * <li>if the modifier string is not null, emit it
   * <li>otherwise emit default modifier (<code>public</code>)
   * </ul>
   *
   * @param node the node onto which a modifier property is attached
   */
  private void emitModifier(SimpleNode node) {
    // get modifier string
    int modifier;
    String modifierString = null;
    try {
      modifier = ((Integer)node.getProperty("modifier")).intValue();
      modifierString = new JModifier(modifier).toString();
      } catch (PropertyAccessException e) {}
    if (modifierString != null) emit(modifierString + " ");
    else emit("public "); // default modifier: public
  }

  /**
   * Emit a string with Unicode escapes added.
   *
   * <p><b>Algorithm:</b><ul>
   * <li>if the string is null, ignore it
   * <li>else, filter the string to add appropriate unicode escapes to
   *     generate a full ASCII file and emit it
   * <li>increment <code>currentColumn</code> by the length of the initial
   *     string
   * </ul>
   * 
   * <p>This method may be redefined in sub classes if filtering through
   * addUnicodeEscapes is not desired.
   *
   * @param str the string to emit
   * @see #addUnicodeEscapes
   */
  protected void emit(String str) {
    super.emit(addUnicodeEscapes(str));
  }

  /**
   * Filters a string replacing non ASCII characters by their equivalent
   * using the Java Unicode escape sequence "\\u...".
   *
   * @param str the string to filter
   * @return a pure ASCII string
   */
  private String addUnicodeEscapes(String str) {
    String retval = "";
    char ch;
    for (int i = 0; i < str.length(); i++) {
      ch = str.charAt(i);
      if ((ch < 0x20 || ch > 0x7e) &&
	  ch != '\t' && ch != '\n' && ch != '\r' && ch != '\f') {
  	String s = "0000" + Integer.toString(ch, 16);
  	retval += "\\u" + s.substring(s.length() - 4, s.length());
      } else {
        retval += ch;
      }
    }
    return retval;
  }
}

// ***********************************************************************
// History Log:
// $Log: JUnparserVisitor.java,v $
// Revision 1.17  1998/06/30 15:53:39  mallet
// Replaced OutputStreamWriter over FileOutputStream by FileWriter.
//
// Revision 1.16  1998/04/15 13:58:39  bareau
// Update according to the new NL mechanism: there are no NLD_* nodes.
//
// Revision 1.15  1998/03/19 09:10:04  mallet
// Fixed bug in generated code for C++: Unicode escapes should not be
// generated. Moved addUnicodeEscapes and redefined emit to call it in
// JUnparserVisitor. emit in GenericUnparserVisitor no longer calls
// addUnicodeEscapes.
//
// Revision 1.14  1998/02/27 13:02:56  cinard
// Updated copyright notice.
//
// Revision 1.13  1998/02/24 17:00:09  mallet
// Added visit for new node types ASTCompilationUnit (code from
// visit(ASTADL_CompilationUnit, Object)) and ASTPackageDeclaration.
// Changed unparse method to expect an ASTCompilationUnit.
// Changed visit for ASTADL_CompilationUnit to report an unexpected node.
// Made visit for ASTCompilationUnit more robust when there is no class
// or interface declaration (which should be an compiler error).
//
// Revision 1.12  1998/02/12 16:04:09  mallet
// Fixed copyright notice: TOG RI -> TOG, year extended to 1998 for TOG
// and IPA.
//
// Revision 1.11  1998/02/11 09:57:06  mallet
// Changed visit(ASTNLD_Annotation) to avoid reporting an incorrect node:
// NLD annotations may pass through the whole chain but are harmless.
//
// Revision 1.10  1998/01/27 11:14:14  bareau
// Bug fix in ArrayDimensions: has children to visit, and arrayLevel property
// denotes all array dimensions, empty or not.
//
// Revision 1.9  1998/01/26 15:30:44  mallet
// Added reporting of generated file in verbose mode only.
//
// Revision 1.8  1998/01/08 13:39:18  bareau
// Bug fix in CastExpression.
//
// Revision 1.7  1998/01/07 13:29:42  mallet
// Fixed javadoc comments.
//
// Revision 1.6  1998/01/05 16:35:08  mallet
// Changed name of start method: now called unparse.
// Fixed processing of CastExpression node: when casted type is a
// primitive type there is no child for it.
//
// Revision 1.5  1997/12/22 13:34:42  mallet
// Fixed processing of modifier by using JModifier class of symbol table
// package.
//
// Revision 1.4  1997/12/19 17:59:12  mallet
// Added visit method for two new nodes: ADL_NormalBehavior and
// ADL_AbNormalBehavior.
//
// Revision 1.3  1997/12/17 17:14:18  mallet
// Changed case of all properties to start with a lowercase letter.
// Replaced getOp and getValue by getProperty.
//
// Revision 1.2  1997/12/17 11:22:30  mallet
// Added javadoc comments.
// Added modifier retrieval through the Modifier property.
// Fixed a bug in SwitchLabel node: was not checking for default label.
// Updated processing of Assignment since AssignmentOperator is now void.
// Fixed NullLiteral processing: don't need to call getValue().C
// Added generic methods visitStatementChild, to visit a statement child
// and format output properly when the child is a block or not, and
// unparseList, to unparse a list of children with either a fixed
// separator or a specific separator for each child, while taking care of
// line wrapping.
//
// Revision 1.1  1997/12/10 15:22:08  mallet
// Java Unparser Visitor: first working version.
// TBD: javadoc comments, replace getProperty calls by get* when these
// methods are implemented in appropriate nodes, improve modifiers
// access.
//
// ***********************************************************************
