//
// Copyright (c) 1996-1998 Information-technology Promotion Agency, Japan
// Copyright (c) 1996-1998 Mitsubishi Research Institute, Inc.
// Copyright (c) 1996-1998 The Open Group
//
// All rights reserved.
// This technology is a result of the Advanced Software  Enrichment 
// Project of Information-technology Promotion Agency, Japan (IPA).
//
// Permissions  to  use,  copy, modify and distribute this software
// are governed by the terms and conditions set forth in  the  file 
// COPYRIGHT, located in the top directory of this software.
//
// ***********************************************************************
//
// $Source: /vobstore/cvsstore/adl/adl2/src/runtime/c/ADLParameterSet.c,v $
// $Revision: 1.3 $
// $Date: 1998/05/11 12:54:09 $
// PRODUCT:	Assertion Definition Language Translation System
//              (ADLTS) Version 2.0
// COMPONENT:	C Runtime Environment - Test directive and factory 
//                                      parameter set class
// NAME:	ADLParameterSet.h
// 	
// HISTORY:
// 	Shane McCarron, The Open Group - 11 April 1998: Converted
//      Lionel Mallet, The Open Group RI  - Oct  26, 1997: Created
//      See Log at end of file
// 		
// ***********************************************************************

#include "ADLParameterSet.h"
#include "ADLMemory.h"

// Internal Routines

static char buf[ADL_ExceptionLen];

/* checkBounds - ensure that the index is in range.
 *
 * Returns 1 if the value is in range.
 * Returns 0 if it is not.
 */

static int checkBounds(ADLParameterSet *p, int index) {
  if (index < 0 || index >= p->numValues) {
    sprintf(buf, "ADLParameterSet: index out of range (%d/%d)", index, p->numValues);
    ADL_RaiseException(ADLRuntime, buf);
    return 0;
  }
  return 1;
}
static int *computeArrayofMaxInstances(ADLParameterSet *p) {

  ADLDataSet *data;
  int	x;

  // allocate returned array (freed by caller)
  int *elementInstanceValues = ADL__calloc(p->numElements, sizeof(int *));
  
  if (elementInstanceValues) {

      data = p->elementSet;
      for (x = 0; x < p->numElements; x++) {
	elementInstanceValues[x] = data->numValues;
	data = data->nextSet;
      }
  }
  return elementInstanceValues;
}

// Note that instance number should be starting at 0
static int *computeIndexes(ADLParameterSet *p, int instance) {

  int j;
  int *maxInstanceArray = NULL;
  int *ret = NULL;

  if (!checkBounds(p, instance)) return NULL;

  if (p->numElements == 0) {
    return NULL;
  }

  // compute the array of element max instance values (freed at end of
  // method)
  maxInstanceArray = computeArrayofMaxInstances(p);

  // allocate array of computed indexes (never freed in provide)
  ret = ADL__calloc(p->numElements, sizeof(int));

  // perform magic computation
  for (j = p->numElements - 1; j >= 0; j--) {
    ret[j] = (instance % maxInstanceArray[j]);
    instance /= maxInstanceArray[j];
  }

  // free array of max instances
  ADL__free(maxInstanceArray);

  // final check
  if (instance != 0) {
    ADL_RaiseException(ADLRuntime, 
			      "ADLParameterSet: instance != 0");
  }

  return ret;
}

/* Allocate a parameter set and then initialize the important elements
 */

ADLParameterSet *ADL_ParameterSetInit() {
    ADLParameterSet *ptr;

    ptr = ADL__calloc(1, sizeof(ADLParameterSet));
    ptr->elementSet = NULL;
    ptr->numElements = 0;
    ptr->numValues = 1;
    ptr->elementIndexes = NULL;
    return(ptr);
}

void ADL__addElement(ADLParameterSet *p, ADLDataSet *e) {
  // Add an element to the link list chain

    ADLDataSet *ptr = p->elementSet;

    if (ptr == NULL) {
	p->elementSet = e;
    } else {
	while (ptr->nextSet) {
	    ptr = ptr->nextSet;
	}
	ptr->nextSet = e;
    }

    p->numElements++;
    p->numValues *= e->numValues;
}

int ADL__getNumValues(ADLParameterSet *p) {
    return p->numValues;
}

int ADL__getNumElements(ADLParameterSet *p) {
    return p->numElements;
}


void **ADL__provide(ADLParameterSet *p, int index) {

  ADLDataSet *data;
  void **ret = NULL;
  int x;

  if (!checkBounds(p, index)) return NULL;

  // free previously alloced array if not null
  if (p->elementIndexes != NULL) free(p->elementIndexes);
  p->elementIndexes = computeIndexes(p, index);

  // allocate returned array (freed by caller)
  ret = ADL__calloc(p->numElements, sizeof(void *));

  data = p->elementSet;
  for (x = 0; x < p->numElements; x++) {
    ret[x] = data->provide(data, p->elementIndexes[x]);
    data = data->nextSet;
  }

  return ret;
}

int ADL__relinquish(ADLParameterSet *p, int index) {
  
  ADLDataSet *data;
  int result = 1;
  int x;

  if (checkBounds(p, index)) {

      data = p->elementSet;
      for (x = 0; x < p->numElements; x++) {
	result &= data->relinquish(data, p->elementIndexes[x]);
	data = data->nextSet;
      }
  }

  return result;
}


// ***********************************************************************
// History Log:
// $Log: ADLParameterSet.c,v $
// Revision 1.3  1998/05/11 12:54:09  shane
// Changed _ADL_* symbols to ADL__* since our naming convention says that all ADL
// symbols should start with ADL or tdd.
//
// Revision 1.2  1998/05/08 20:21:07  shane
// Various modifications, including:
//
// Changed to print certain information only when debugMode is true.
// Changed to actually set debugMode based upon ADL_RPT_MODE in configuration.
// Fixed how pointers are used (man am I stupid).
//
// Revision 1.1  1998/04/16 12:09:12  shane
// First checkin of complete set.
//
//
// ***********************************************************************
